/* Generated by Cython 0.19.1 on Mon Jul 29 00:29:58 2013 */

#define PY_SSIZE_T_CLEAN
#ifndef CYTHON_USE_PYLONG_INTERNALS
#ifdef PYLONG_BITS_IN_DIGIT
#define CYTHON_USE_PYLONG_INTERNALS 0
#else
#include "pyconfig.h"
#ifdef PYLONG_BITS_IN_DIGIT
#define CYTHON_USE_PYLONG_INTERNALS 1
#else
#define CYTHON_USE_PYLONG_INTERNALS 0
#endif
#endif
#endif
#include "Python.h"
#ifndef Py_PYTHON_H
    #error Python headers needed to compile C extensions, please install development version of Python.
#elif PY_VERSION_HEX < 0x02040000
    #error Cython requires Python 2.4+.
#else
#include <stddef.h> /* For offsetof */
#ifndef offsetof
#define offsetof(type, member) ( (size_t) & ((type*)0) -> member )
#endif
#if !defined(WIN32) && !defined(MS_WINDOWS)
  #ifndef __stdcall
    #define __stdcall
  #endif
  #ifndef __cdecl
    #define __cdecl
  #endif
  #ifndef __fastcall
    #define __fastcall
  #endif
#endif
#ifndef DL_IMPORT
  #define DL_IMPORT(t) t
#endif
#ifndef DL_EXPORT
  #define DL_EXPORT(t) t
#endif
#ifndef PY_LONG_LONG
  #define PY_LONG_LONG LONG_LONG
#endif
#ifndef Py_HUGE_VAL
  #define Py_HUGE_VAL HUGE_VAL
#endif
#ifdef PYPY_VERSION
#define CYTHON_COMPILING_IN_PYPY 1
#define CYTHON_COMPILING_IN_CPYTHON 0
#else
#define CYTHON_COMPILING_IN_PYPY 0
#define CYTHON_COMPILING_IN_CPYTHON 1
#endif
#if PY_VERSION_HEX < 0x02050000
  typedef int Py_ssize_t;
  #define PY_SSIZE_T_MAX INT_MAX
  #define PY_SSIZE_T_MIN INT_MIN
  #define PY_FORMAT_SIZE_T ""
  #define CYTHON_FORMAT_SSIZE_T ""
  #define PyInt_FromSsize_t(z) PyInt_FromLong(z)
  #define PyInt_AsSsize_t(o)   __Pyx_PyInt_AsInt(o)
  #define PyNumber_Index(o)    ((PyNumber_Check(o) && !PyFloat_Check(o)) ? PyNumber_Int(o) : \
                                (PyErr_Format(PyExc_TypeError, \
                                              "expected index value, got %.200s", Py_TYPE(o)->tp_name), \
                                 (PyObject*)0))
  #define __Pyx_PyIndex_Check(o) (PyNumber_Check(o) && !PyFloat_Check(o) && \
                                  !PyComplex_Check(o))
  #define PyIndex_Check __Pyx_PyIndex_Check
  #define PyErr_WarnEx(category, message, stacklevel) PyErr_Warn(category, message)
  #define __PYX_BUILD_PY_SSIZE_T "i"
#else
  #define __PYX_BUILD_PY_SSIZE_T "n"
  #define CYTHON_FORMAT_SSIZE_T "z"
  #define __Pyx_PyIndex_Check PyIndex_Check
#endif
#if PY_VERSION_HEX < 0x02060000
  #define Py_REFCNT(ob) (((PyObject*)(ob))->ob_refcnt)
  #define Py_TYPE(ob)   (((PyObject*)(ob))->ob_type)
  #define Py_SIZE(ob)   (((PyVarObject*)(ob))->ob_size)
  #define PyVarObject_HEAD_INIT(type, size) \
          PyObject_HEAD_INIT(type) size,
  #define PyType_Modified(t)
  typedef struct {
     void *buf;
     PyObject *obj;
     Py_ssize_t len;
     Py_ssize_t itemsize;
     int readonly;
     int ndim;
     char *format;
     Py_ssize_t *shape;
     Py_ssize_t *strides;
     Py_ssize_t *suboffsets;
     void *internal;
  } Py_buffer;
  #define PyBUF_SIMPLE 0
  #define PyBUF_WRITABLE 0x0001
  #define PyBUF_FORMAT 0x0004
  #define PyBUF_ND 0x0008
  #define PyBUF_STRIDES (0x0010 | PyBUF_ND)
  #define PyBUF_C_CONTIGUOUS (0x0020 | PyBUF_STRIDES)
  #define PyBUF_F_CONTIGUOUS (0x0040 | PyBUF_STRIDES)
  #define PyBUF_ANY_CONTIGUOUS (0x0080 | PyBUF_STRIDES)
  #define PyBUF_INDIRECT (0x0100 | PyBUF_STRIDES)
  #define PyBUF_RECORDS (PyBUF_STRIDES | PyBUF_FORMAT | PyBUF_WRITABLE)
  #define PyBUF_FULL (PyBUF_INDIRECT | PyBUF_FORMAT | PyBUF_WRITABLE)
  typedef int (*getbufferproc)(PyObject *, Py_buffer *, int);
  typedef void (*releasebufferproc)(PyObject *, Py_buffer *);
#endif
#if PY_MAJOR_VERSION < 3
  #define __Pyx_BUILTIN_MODULE_NAME "__builtin__"
  #define __Pyx_PyCode_New(a, k, l, s, f, code, c, n, v, fv, cell, fn, name, fline, lnos) \
          PyCode_New(a, l, s, f, code, c, n, v, fv, cell, fn, name, fline, lnos)
#else
  #define __Pyx_BUILTIN_MODULE_NAME "builtins"
  #define __Pyx_PyCode_New(a, k, l, s, f, code, c, n, v, fv, cell, fn, name, fline, lnos) \
          PyCode_New(a, k, l, s, f, code, c, n, v, fv, cell, fn, name, fline, lnos)
#endif
#if PY_MAJOR_VERSION < 3 && PY_MINOR_VERSION < 6
  #define PyUnicode_FromString(s) PyUnicode_Decode(s, strlen(s), "UTF-8", "strict")
#endif
#if PY_MAJOR_VERSION >= 3
  #define Py_TPFLAGS_CHECKTYPES 0
  #define Py_TPFLAGS_HAVE_INDEX 0
#endif
#if (PY_VERSION_HEX < 0x02060000) || (PY_MAJOR_VERSION >= 3)
  #define Py_TPFLAGS_HAVE_NEWBUFFER 0
#endif
#if PY_VERSION_HEX < 0x02060000
  #define Py_TPFLAGS_HAVE_VERSION_TAG 0
#endif
#if PY_VERSION_HEX > 0x03030000 && defined(PyUnicode_KIND)
  #define CYTHON_PEP393_ENABLED 1
  #define __Pyx_PyUnicode_READY(op)       (likely(PyUnicode_IS_READY(op)) ? \
                                              0 : _PyUnicode_Ready((PyObject *)(op)))
  #define __Pyx_PyUnicode_GET_LENGTH(u)   PyUnicode_GET_LENGTH(u)
  #define __Pyx_PyUnicode_READ_CHAR(u, i) PyUnicode_READ_CHAR(u, i)
  #define __Pyx_PyUnicode_READ(k, d, i)   PyUnicode_READ(k, d, i)
#else
  #define CYTHON_PEP393_ENABLED 0
  #define __Pyx_PyUnicode_READY(op)       (0)
  #define __Pyx_PyUnicode_GET_LENGTH(u)   PyUnicode_GET_SIZE(u)
  #define __Pyx_PyUnicode_READ_CHAR(u, i) ((Py_UCS4)(PyUnicode_AS_UNICODE(u)[i]))
  #define __Pyx_PyUnicode_READ(k, d, i)   ((k=k), (Py_UCS4)(((Py_UNICODE*)d)[i]))
#endif
#if PY_MAJOR_VERSION >= 3
  #define PyBaseString_Type            PyUnicode_Type
  #define PyStringObject               PyUnicodeObject
  #define PyString_Type                PyUnicode_Type
  #define PyString_Check               PyUnicode_Check
  #define PyString_CheckExact          PyUnicode_CheckExact
#endif
#if PY_VERSION_HEX < 0x02060000
  #define PyBytesObject                PyStringObject
  #define PyBytes_Type                 PyString_Type
  #define PyBytes_Check                PyString_Check
  #define PyBytes_CheckExact           PyString_CheckExact
  #define PyBytes_FromString           PyString_FromString
  #define PyBytes_FromStringAndSize    PyString_FromStringAndSize
  #define PyBytes_FromFormat           PyString_FromFormat
  #define PyBytes_DecodeEscape         PyString_DecodeEscape
  #define PyBytes_AsString             PyString_AsString
  #define PyBytes_AsStringAndSize      PyString_AsStringAndSize
  #define PyBytes_Size                 PyString_Size
  #define PyBytes_AS_STRING            PyString_AS_STRING
  #define PyBytes_GET_SIZE             PyString_GET_SIZE
  #define PyBytes_Repr                 PyString_Repr
  #define PyBytes_Concat               PyString_Concat
  #define PyBytes_ConcatAndDel         PyString_ConcatAndDel
#endif
#if PY_MAJOR_VERSION >= 3
  #define __Pyx_PyBaseString_Check(obj) PyUnicode_Check(obj)
  #define __Pyx_PyBaseString_CheckExact(obj) PyUnicode_CheckExact(obj)
#else
  #define __Pyx_PyBaseString_Check(obj) (PyString_CheckExact(obj) || PyUnicode_CheckExact(obj) || \
                                         PyString_Check(obj) || PyUnicode_Check(obj))
  #define __Pyx_PyBaseString_CheckExact(obj) (Py_TYPE(obj) == &PyBaseString_Type)
#endif
#if PY_VERSION_HEX < 0x02060000
  #define PySet_Check(obj)             PyObject_TypeCheck(obj, &PySet_Type)
  #define PyFrozenSet_Check(obj)       PyObject_TypeCheck(obj, &PyFrozenSet_Type)
#endif
#ifndef PySet_CheckExact
  #define PySet_CheckExact(obj)        (Py_TYPE(obj) == &PySet_Type)
#endif
#define __Pyx_TypeCheck(obj, type) PyObject_TypeCheck(obj, (PyTypeObject *)type)
#if PY_MAJOR_VERSION >= 3
  #define PyIntObject                  PyLongObject
  #define PyInt_Type                   PyLong_Type
  #define PyInt_Check(op)              PyLong_Check(op)
  #define PyInt_CheckExact(op)         PyLong_CheckExact(op)
  #define PyInt_FromString             PyLong_FromString
  #define PyInt_FromUnicode            PyLong_FromUnicode
  #define PyInt_FromLong               PyLong_FromLong
  #define PyInt_FromSize_t             PyLong_FromSize_t
  #define PyInt_FromSsize_t            PyLong_FromSsize_t
  #define PyInt_AsLong                 PyLong_AsLong
  #define PyInt_AS_LONG                PyLong_AS_LONG
  #define PyInt_AsSsize_t              PyLong_AsSsize_t
  #define PyInt_AsUnsignedLongMask     PyLong_AsUnsignedLongMask
  #define PyInt_AsUnsignedLongLongMask PyLong_AsUnsignedLongLongMask
#endif
#if PY_MAJOR_VERSION >= 3
  #define PyBoolObject                 PyLongObject
#endif
#if PY_VERSION_HEX < 0x03020000
  typedef long Py_hash_t;
  #define __Pyx_PyInt_FromHash_t PyInt_FromLong
  #define __Pyx_PyInt_AsHash_t   PyInt_AsLong
#else
  #define __Pyx_PyInt_FromHash_t PyInt_FromSsize_t
  #define __Pyx_PyInt_AsHash_t   PyInt_AsSsize_t
#endif
#if (PY_MAJOR_VERSION < 3) || (PY_VERSION_HEX >= 0x03010300)
  #define __Pyx_PySequence_GetSlice(obj, a, b) PySequence_GetSlice(obj, a, b)
  #define __Pyx_PySequence_SetSlice(obj, a, b, value) PySequence_SetSlice(obj, a, b, value)
  #define __Pyx_PySequence_DelSlice(obj, a, b) PySequence_DelSlice(obj, a, b)
#else
  #define __Pyx_PySequence_GetSlice(obj, a, b) (unlikely(!(obj)) ? \
        (PyErr_SetString(PyExc_SystemError, "null argument to internal routine"), (PyObject*)0) : \
        (likely((obj)->ob_type->tp_as_mapping) ? (PySequence_GetSlice(obj, a, b)) : \
            (PyErr_Format(PyExc_TypeError, "'%.200s' object is unsliceable", (obj)->ob_type->tp_name), (PyObject*)0)))
  #define __Pyx_PySequence_SetSlice(obj, a, b, value) (unlikely(!(obj)) ? \
        (PyErr_SetString(PyExc_SystemError, "null argument to internal routine"), -1) : \
        (likely((obj)->ob_type->tp_as_mapping) ? (PySequence_SetSlice(obj, a, b, value)) : \
            (PyErr_Format(PyExc_TypeError, "'%.200s' object doesn't support slice assignment", (obj)->ob_type->tp_name), -1)))
  #define __Pyx_PySequence_DelSlice(obj, a, b) (unlikely(!(obj)) ? \
        (PyErr_SetString(PyExc_SystemError, "null argument to internal routine"), -1) : \
        (likely((obj)->ob_type->tp_as_mapping) ? (PySequence_DelSlice(obj, a, b)) : \
            (PyErr_Format(PyExc_TypeError, "'%.200s' object doesn't support slice deletion", (obj)->ob_type->tp_name), -1)))
#endif
#if PY_MAJOR_VERSION >= 3
  #define PyMethod_New(func, self, klass) ((self) ? PyMethod_New(func, self) : PyInstanceMethod_New(func))
#endif
#if PY_VERSION_HEX < 0x02050000
  #define __Pyx_GetAttrString(o,n)   PyObject_GetAttrString((o),((char *)(n)))
  #define __Pyx_SetAttrString(o,n,a) PyObject_SetAttrString((o),((char *)(n)),(a))
  #define __Pyx_DelAttrString(o,n)   PyObject_DelAttrString((o),((char *)(n)))
#else
  #define __Pyx_GetAttrString(o,n)   PyObject_GetAttrString((o),(n))
  #define __Pyx_SetAttrString(o,n,a) PyObject_SetAttrString((o),(n),(a))
  #define __Pyx_DelAttrString(o,n)   PyObject_DelAttrString((o),(n))
#endif
#if PY_VERSION_HEX < 0x02050000
  #define __Pyx_NAMESTR(n) ((char *)(n))
  #define __Pyx_DOCSTR(n)  ((char *)(n))
#else
  #define __Pyx_NAMESTR(n) (n)
  #define __Pyx_DOCSTR(n)  (n)
#endif
#ifndef CYTHON_INLINE
  #if defined(__GNUC__)
    #define CYTHON_INLINE __inline__
  #elif defined(_MSC_VER)
    #define CYTHON_INLINE __inline
  #elif defined (__STDC_VERSION__) && __STDC_VERSION__ >= 199901L
    #define CYTHON_INLINE inline
  #else
    #define CYTHON_INLINE
  #endif
#endif
#ifndef CYTHON_RESTRICT
  #if defined(__GNUC__)
    #define CYTHON_RESTRICT __restrict__
  #elif defined(_MSC_VER)
    #define CYTHON_RESTRICT __restrict
  #elif defined (__STDC_VERSION__) && __STDC_VERSION__ >= 199901L
    #define CYTHON_RESTRICT restrict
  #else
    #define CYTHON_RESTRICT
  #endif
#endif
#ifdef NAN
#define __PYX_NAN() ((float) NAN)
#else
static CYTHON_INLINE float __PYX_NAN() {
  /* Initialize NaN. The sign is irrelevant, an exponent with all bits 1 and
   a nonzero mantissa means NaN. If the first bit in the mantissa is 1, it is
   a quiet NaN. */
  float value;
  memset(&value, 0xFF, sizeof(value));
  return value;
}
#endif


#if PY_MAJOR_VERSION >= 3
  #define __Pyx_PyNumber_Divide(x,y)         PyNumber_TrueDivide(x,y)
  #define __Pyx_PyNumber_InPlaceDivide(x,y)  PyNumber_InPlaceTrueDivide(x,y)
#else
  #define __Pyx_PyNumber_Divide(x,y)         PyNumber_Divide(x,y)
  #define __Pyx_PyNumber_InPlaceDivide(x,y)  PyNumber_InPlaceDivide(x,y)
#endif

#ifndef __PYX_EXTERN_C
  #ifdef __cplusplus
    #define __PYX_EXTERN_C extern "C"
  #else
    #define __PYX_EXTERN_C extern
  #endif
#endif

#if defined(WIN32) || defined(MS_WINDOWS)
#define _USE_MATH_DEFINES
#endif
#include <math.h>
#define __PYX_HAVE__sklearn__tree___tree
#define __PYX_HAVE_API__sklearn__tree___tree
#include "string.h"
#include "stdio.h"
#include "stdlib.h"
#include "numpy/arrayobject.h"
#include "numpy/ufuncobject.h"
#include "math.h"
#ifdef _OPENMP
#include <omp.h>
#endif /* _OPENMP */

#ifdef PYREX_WITHOUT_ASSERTIONS
#define CYTHON_WITHOUT_ASSERTIONS
#endif

#ifndef CYTHON_UNUSED
# if defined(__GNUC__)
#   if !(defined(__cplusplus)) || (__GNUC__ > 3 || (__GNUC__ == 3 && __GNUC_MINOR__ >= 4))
#     define CYTHON_UNUSED __attribute__ ((__unused__))
#   else
#     define CYTHON_UNUSED
#   endif
# elif defined(__ICC) || (defined(__INTEL_COMPILER) && !defined(_MSC_VER))
#   define CYTHON_UNUSED __attribute__ ((__unused__))
# else
#   define CYTHON_UNUSED
# endif
#endif
typedef struct {PyObject **p; char *s; const Py_ssize_t n; const char* encoding;
                const char is_unicode; const char is_str; const char intern; } __Pyx_StringTabEntry; /*proto*/

#define __PYX_DEFAULT_STRING_ENCODING_IS_ASCII 0
#define __PYX_DEFAULT_STRING_ENCODING_IS_DEFAULT 0
#define __PYX_DEFAULT_STRING_ENCODING ""
#define __Pyx_PyObject_FromString __Pyx_PyBytes_FromString
#define __Pyx_PyObject_FromStringAndSize __Pyx_PyBytes_FromStringAndSize
static CYTHON_INLINE char* __Pyx_PyObject_AsString(PyObject*);
static CYTHON_INLINE char* __Pyx_PyObject_AsStringAndSize(PyObject*, Py_ssize_t* length);
#define __Pyx_PyBytes_FromString        PyBytes_FromString
#define __Pyx_PyBytes_FromStringAndSize PyBytes_FromStringAndSize
static CYTHON_INLINE PyObject* __Pyx_PyUnicode_FromString(char*);
#if PY_MAJOR_VERSION < 3
    #define __Pyx_PyStr_FromString        __Pyx_PyBytes_FromString
    #define __Pyx_PyStr_FromStringAndSize __Pyx_PyBytes_FromStringAndSize
#else
    #define __Pyx_PyStr_FromString        __Pyx_PyUnicode_FromString
    #define __Pyx_PyStr_FromStringAndSize __Pyx_PyUnicode_FromStringAndSize
#endif
#define __Pyx_PyObject_AsUString(s)    ((unsigned char*) __Pyx_PyObject_AsString(s))
#define __Pyx_PyObject_FromUString(s)  __Pyx_PyObject_FromString((char*)s)
#define __Pyx_PyBytes_FromUString(s)   __Pyx_PyBytes_FromString((char*)s)
#define __Pyx_PyStr_FromUString(s)     __Pyx_PyStr_FromString((char*)s)
#define __Pyx_PyUnicode_FromUString(s) __Pyx_PyUnicode_FromString((char*)s)
#if PY_MAJOR_VERSION < 3
static CYTHON_INLINE size_t __Pyx_Py_UNICODE_strlen(const Py_UNICODE *u)
{
    const Py_UNICODE *u_end = u;
    while (*u_end++) ;
    return u_end - u - 1;
}
#else
#define __Pyx_Py_UNICODE_strlen Py_UNICODE_strlen
#endif
#define __Pyx_PyUnicode_FromUnicode(u)       PyUnicode_FromUnicode(u, __Pyx_Py_UNICODE_strlen(u))
#define __Pyx_PyUnicode_FromUnicodeAndLength PyUnicode_FromUnicode
#define __Pyx_PyUnicode_AsUnicode            PyUnicode_AsUnicode
#define __Pyx_Owned_Py_None(b) (Py_INCREF(Py_None), Py_None)
#define __Pyx_PyBool_FromLong(b) ((b) ? (Py_INCREF(Py_True), Py_True) : (Py_INCREF(Py_False), Py_False))
static CYTHON_INLINE int __Pyx_PyObject_IsTrue(PyObject*);
static CYTHON_INLINE PyObject* __Pyx_PyNumber_Int(PyObject* x);
static CYTHON_INLINE Py_ssize_t __Pyx_PyIndex_AsSsize_t(PyObject*);
static CYTHON_INLINE PyObject * __Pyx_PyInt_FromSize_t(size_t);
static CYTHON_INLINE size_t __Pyx_PyInt_AsSize_t(PyObject*);
#if CYTHON_COMPILING_IN_CPYTHON
#define __pyx_PyFloat_AsDouble(x) (PyFloat_CheckExact(x) ? PyFloat_AS_DOUBLE(x) : PyFloat_AsDouble(x))
#else
#define __pyx_PyFloat_AsDouble(x) PyFloat_AsDouble(x)
#endif
#define __pyx_PyFloat_AsFloat(x) ((float) __pyx_PyFloat_AsDouble(x))
#if PY_MAJOR_VERSION < 3 && __PYX_DEFAULT_STRING_ENCODING_IS_ASCII
static int __Pyx_sys_getdefaultencoding_not_ascii;
static int __Pyx_init_sys_getdefaultencoding_params() {
    PyObject* sys = NULL;
    PyObject* default_encoding = NULL;
    PyObject* ascii_chars_u = NULL;
    PyObject* ascii_chars_b = NULL;
    sys = PyImport_ImportModule("sys");
    if (sys == NULL) goto bad;
    default_encoding = PyObject_CallMethod(sys, (char*) (const char*) "getdefaultencoding", NULL);
    if (default_encoding == NULL) goto bad;
    if (strcmp(PyBytes_AsString(default_encoding), "ascii") == 0) {
        __Pyx_sys_getdefaultencoding_not_ascii = 0;
    } else {
        const char* default_encoding_c = PyBytes_AS_STRING(default_encoding);
        char ascii_chars[128];
        int c;
        for (c = 0; c < 128; c++) {
            ascii_chars[c] = c;
        }
        __Pyx_sys_getdefaultencoding_not_ascii = 1;
        ascii_chars_u = PyUnicode_DecodeASCII(ascii_chars, 128, NULL);
        if (ascii_chars_u == NULL) goto bad;
        ascii_chars_b = PyUnicode_AsEncodedString(ascii_chars_u, default_encoding_c, NULL);
        if (ascii_chars_b == NULL || strncmp(ascii_chars, PyBytes_AS_STRING(ascii_chars_b), 128) != 0) {
            PyErr_Format(
                PyExc_ValueError,
                "This module compiled with c_string_encoding=ascii, but default encoding '%s' is not a superset of ascii.",
                default_encoding_c);
            goto bad;
        }
    }
    Py_XDECREF(sys);
    Py_XDECREF(default_encoding);
    Py_XDECREF(ascii_chars_u);
    Py_XDECREF(ascii_chars_b);
    return 0;
bad:
    Py_XDECREF(sys);
    Py_XDECREF(default_encoding);
    Py_XDECREF(ascii_chars_u);
    Py_XDECREF(ascii_chars_b);
    return -1;
}
#endif
#if __PYX_DEFAULT_STRING_ENCODING_IS_DEFAULT && PY_MAJOR_VERSION >= 3
#define __Pyx_PyUnicode_FromStringAndSize(c_str, size) PyUnicode_DecodeUTF8(c_str, size, NULL)
#else
#define __Pyx_PyUnicode_FromStringAndSize(c_str, size) PyUnicode_Decode(c_str, size, __PYX_DEFAULT_STRING_ENCODING, NULL)
#if __PYX_DEFAULT_STRING_ENCODING_IS_DEFAULT
static char* __PYX_DEFAULT_STRING_ENCODING;
static int __Pyx_init_sys_getdefaultencoding_params() {
    PyObject* sys = NULL;
    PyObject* default_encoding = NULL;
    char* default_encoding_c;
    sys = PyImport_ImportModule("sys");
    if (sys == NULL) goto bad;
    default_encoding = PyObject_CallMethod(sys, (char*) (const char*) "getdefaultencoding", NULL);
    if (default_encoding == NULL) goto bad;
    default_encoding_c = PyBytes_AS_STRING(default_encoding);
    __PYX_DEFAULT_STRING_ENCODING = (char*) malloc(strlen(default_encoding_c));
    strcpy(__PYX_DEFAULT_STRING_ENCODING, default_encoding_c);
    Py_DECREF(sys);
    Py_DECREF(default_encoding);
    return 0;
bad:
    Py_XDECREF(sys);
    Py_XDECREF(default_encoding);
    return -1;
}
#endif
#endif


#ifdef __GNUC__
  /* Test for GCC > 2.95 */
  #if __GNUC__ > 2 || (__GNUC__ == 2 && (__GNUC_MINOR__ > 95))
    #define likely(x)   __builtin_expect(!!(x), 1)
    #define unlikely(x) __builtin_expect(!!(x), 0)
  #else /* __GNUC__ > 2 ... */
    #define likely(x)   (x)
    #define unlikely(x) (x)
  #endif /* __GNUC__ > 2 ... */
#else /* __GNUC__ */
  #define likely(x)   (x)
  #define unlikely(x) (x)
#endif /* __GNUC__ */

static PyObject *__pyx_m;
static PyObject *__pyx_d;
static PyObject *__pyx_b;
static PyObject *__pyx_empty_tuple;
static PyObject *__pyx_empty_bytes;
static int __pyx_lineno;
static int __pyx_clineno = 0;
static const char * __pyx_cfilenm= __FILE__;
static const char *__pyx_filename;

#if !defined(CYTHON_CCOMPLEX)
  #if defined(__cplusplus)
    #define CYTHON_CCOMPLEX 1
  #elif defined(_Complex_I)
    #define CYTHON_CCOMPLEX 1
  #else
    #define CYTHON_CCOMPLEX 0
  #endif
#endif
#if CYTHON_CCOMPLEX
  #ifdef __cplusplus
    #include <complex>
  #else
    #include <complex.h>
  #endif
#endif
#if CYTHON_CCOMPLEX && !defined(__cplusplus) && defined(__sun__) && defined(__GNUC__)
  #undef _Complex_I
  #define _Complex_I 1.0fj
#endif


static const char *__pyx_f[] = {
  "_tree.pyx",
  "_tree.pxd",
  "numpy.pxd",
  "type.pxd",
};
#define IS_UNSIGNED(type) (((type) -1) > 0)
struct __Pyx_StructField_;
#define __PYX_BUF_FLAGS_PACKED_STRUCT (1 << 0)
typedef struct {
  const char* name; /* for error messages only */
  struct __Pyx_StructField_* fields;
  size_t size;     /* sizeof(type) */
  size_t arraysize[8]; /* length of array in each dimension */
  int ndim;
  char typegroup; /* _R_eal, _C_omplex, Signed _I_nt, _U_nsigned int, _S_truct, _P_ointer, _O_bject, c_H_ar */
  char is_unsigned;
  int flags;
} __Pyx_TypeInfo;
typedef struct __Pyx_StructField_ {
  __Pyx_TypeInfo* type;
  const char* name;
  size_t offset;
} __Pyx_StructField;
typedef struct {
  __Pyx_StructField* field;
  size_t parent_offset;
} __Pyx_BufFmt_StackElem;
typedef struct {
  __Pyx_StructField root;
  __Pyx_BufFmt_StackElem* head;
  size_t fmt_offset;
  size_t new_count, enc_count;
  size_t struct_alignment;
  int is_complex;
  char enc_type;
  char new_packmode;
  char enc_packmode;
  char is_valid_array;
} __Pyx_BufFmt_Context;


/* "numpy.pxd":723
 * # in Cython to enable them only on the right systems.
 * 
 * ctypedef npy_int8       int8_t             # <<<<<<<<<<<<<<
 * ctypedef npy_int16      int16_t
 * ctypedef npy_int32      int32_t
 */
typedef npy_int8 __pyx_t_5numpy_int8_t;

/* "numpy.pxd":724
 * 
 * ctypedef npy_int8       int8_t
 * ctypedef npy_int16      int16_t             # <<<<<<<<<<<<<<
 * ctypedef npy_int32      int32_t
 * ctypedef npy_int64      int64_t
 */
typedef npy_int16 __pyx_t_5numpy_int16_t;

/* "numpy.pxd":725
 * ctypedef npy_int8       int8_t
 * ctypedef npy_int16      int16_t
 * ctypedef npy_int32      int32_t             # <<<<<<<<<<<<<<
 * ctypedef npy_int64      int64_t
 * #ctypedef npy_int96      int96_t
 */
typedef npy_int32 __pyx_t_5numpy_int32_t;

/* "numpy.pxd":726
 * ctypedef npy_int16      int16_t
 * ctypedef npy_int32      int32_t
 * ctypedef npy_int64      int64_t             # <<<<<<<<<<<<<<
 * #ctypedef npy_int96      int96_t
 * #ctypedef npy_int128     int128_t
 */
typedef npy_int64 __pyx_t_5numpy_int64_t;

/* "numpy.pxd":730
 * #ctypedef npy_int128     int128_t
 * 
 * ctypedef npy_uint8      uint8_t             # <<<<<<<<<<<<<<
 * ctypedef npy_uint16     uint16_t
 * ctypedef npy_uint32     uint32_t
 */
typedef npy_uint8 __pyx_t_5numpy_uint8_t;

/* "numpy.pxd":731
 * 
 * ctypedef npy_uint8      uint8_t
 * ctypedef npy_uint16     uint16_t             # <<<<<<<<<<<<<<
 * ctypedef npy_uint32     uint32_t
 * ctypedef npy_uint64     uint64_t
 */
typedef npy_uint16 __pyx_t_5numpy_uint16_t;

/* "numpy.pxd":732
 * ctypedef npy_uint8      uint8_t
 * ctypedef npy_uint16     uint16_t
 * ctypedef npy_uint32     uint32_t             # <<<<<<<<<<<<<<
 * ctypedef npy_uint64     uint64_t
 * #ctypedef npy_uint96     uint96_t
 */
typedef npy_uint32 __pyx_t_5numpy_uint32_t;

/* "numpy.pxd":733
 * ctypedef npy_uint16     uint16_t
 * ctypedef npy_uint32     uint32_t
 * ctypedef npy_uint64     uint64_t             # <<<<<<<<<<<<<<
 * #ctypedef npy_uint96     uint96_t
 * #ctypedef npy_uint128    uint128_t
 */
typedef npy_uint64 __pyx_t_5numpy_uint64_t;

/* "numpy.pxd":737
 * #ctypedef npy_uint128    uint128_t
 * 
 * ctypedef npy_float32    float32_t             # <<<<<<<<<<<<<<
 * ctypedef npy_float64    float64_t
 * #ctypedef npy_float80    float80_t
 */
typedef npy_float32 __pyx_t_5numpy_float32_t;

/* "numpy.pxd":738
 * 
 * ctypedef npy_float32    float32_t
 * ctypedef npy_float64    float64_t             # <<<<<<<<<<<<<<
 * #ctypedef npy_float80    float80_t
 * #ctypedef npy_float128   float128_t
 */
typedef npy_float64 __pyx_t_5numpy_float64_t;

/* "numpy.pxd":747
 * # The int types are mapped a bit surprising --
 * # numpy.int corresponds to 'l' and numpy.long to 'q'
 * ctypedef npy_long       int_t             # <<<<<<<<<<<<<<
 * ctypedef npy_longlong   long_t
 * ctypedef npy_longlong   longlong_t
 */
typedef npy_long __pyx_t_5numpy_int_t;

/* "numpy.pxd":748
 * # numpy.int corresponds to 'l' and numpy.long to 'q'
 * ctypedef npy_long       int_t
 * ctypedef npy_longlong   long_t             # <<<<<<<<<<<<<<
 * ctypedef npy_longlong   longlong_t
 * 
 */
typedef npy_longlong __pyx_t_5numpy_long_t;

/* "numpy.pxd":749
 * ctypedef npy_long       int_t
 * ctypedef npy_longlong   long_t
 * ctypedef npy_longlong   longlong_t             # <<<<<<<<<<<<<<
 * 
 * ctypedef npy_ulong      uint_t
 */
typedef npy_longlong __pyx_t_5numpy_longlong_t;

/* "numpy.pxd":751
 * ctypedef npy_longlong   longlong_t
 * 
 * ctypedef npy_ulong      uint_t             # <<<<<<<<<<<<<<
 * ctypedef npy_ulonglong  ulong_t
 * ctypedef npy_ulonglong  ulonglong_t
 */
typedef npy_ulong __pyx_t_5numpy_uint_t;

/* "numpy.pxd":752
 * 
 * ctypedef npy_ulong      uint_t
 * ctypedef npy_ulonglong  ulong_t             # <<<<<<<<<<<<<<
 * ctypedef npy_ulonglong  ulonglong_t
 * 
 */
typedef npy_ulonglong __pyx_t_5numpy_ulong_t;

/* "numpy.pxd":753
 * ctypedef npy_ulong      uint_t
 * ctypedef npy_ulonglong  ulong_t
 * ctypedef npy_ulonglong  ulonglong_t             # <<<<<<<<<<<<<<
 * 
 * ctypedef npy_intp       intp_t
 */
typedef npy_ulonglong __pyx_t_5numpy_ulonglong_t;

/* "numpy.pxd":755
 * ctypedef npy_ulonglong  ulonglong_t
 * 
 * ctypedef npy_intp       intp_t             # <<<<<<<<<<<<<<
 * ctypedef npy_uintp      uintp_t
 * 
 */
typedef npy_intp __pyx_t_5numpy_intp_t;

/* "numpy.pxd":756
 * 
 * ctypedef npy_intp       intp_t
 * ctypedef npy_uintp      uintp_t             # <<<<<<<<<<<<<<
 * 
 * ctypedef npy_double     float_t
 */
typedef npy_uintp __pyx_t_5numpy_uintp_t;

/* "numpy.pxd":758
 * ctypedef npy_uintp      uintp_t
 * 
 * ctypedef npy_double     float_t             # <<<<<<<<<<<<<<
 * ctypedef npy_double     double_t
 * ctypedef npy_longdouble longdouble_t
 */
typedef npy_double __pyx_t_5numpy_float_t;

/* "numpy.pxd":759
 * 
 * ctypedef npy_double     float_t
 * ctypedef npy_double     double_t             # <<<<<<<<<<<<<<
 * ctypedef npy_longdouble longdouble_t
 * 
 */
typedef npy_double __pyx_t_5numpy_double_t;

/* "numpy.pxd":760
 * ctypedef npy_double     float_t
 * ctypedef npy_double     double_t
 * ctypedef npy_longdouble longdouble_t             # <<<<<<<<<<<<<<
 * 
 * ctypedef npy_cfloat      cfloat_t
 */
typedef npy_longdouble __pyx_t_5numpy_longdouble_t;

/* "sklearn/tree/_tree.pxd":11
 * cimport numpy as np
 * 
 * ctypedef np.npy_float32 DTYPE_t          # Type of X             # <<<<<<<<<<<<<<
 * ctypedef np.npy_float64 DOUBLE_t         # Type of y, sample_weight
 * ctypedef np.npy_intp SIZE_t              # Type for indices and counters
 */
typedef npy_float32 __pyx_t_7sklearn_4tree_5_tree_DTYPE_t;

/* "sklearn/tree/_tree.pxd":12
 * 
 * ctypedef np.npy_float32 DTYPE_t          # Type of X
 * ctypedef np.npy_float64 DOUBLE_t         # Type of y, sample_weight             # <<<<<<<<<<<<<<
 * ctypedef np.npy_intp SIZE_t              # Type for indices and counters
 * 
 */
typedef npy_float64 __pyx_t_7sklearn_4tree_5_tree_DOUBLE_t;

/* "sklearn/tree/_tree.pxd":13
 * ctypedef np.npy_float32 DTYPE_t          # Type of X
 * ctypedef np.npy_float64 DOUBLE_t         # Type of y, sample_weight
 * ctypedef np.npy_intp SIZE_t              # Type for indices and counters             # <<<<<<<<<<<<<<
 * 
 * 
 */
typedef npy_intp __pyx_t_7sklearn_4tree_5_tree_SIZE_t;
#if CYTHON_CCOMPLEX
  #ifdef __cplusplus
    typedef ::std::complex< float > __pyx_t_float_complex;
  #else
    typedef float _Complex __pyx_t_float_complex;
  #endif
#else
    typedef struct { float real, imag; } __pyx_t_float_complex;
#endif

#if CYTHON_CCOMPLEX
  #ifdef __cplusplus
    typedef ::std::complex< double > __pyx_t_double_complex;
  #else
    typedef double _Complex __pyx_t_double_complex;
  #endif
#else
    typedef struct { double real, imag; } __pyx_t_double_complex;
#endif


/*--- Type declarations ---*/
struct __pyx_obj_7sklearn_4tree_5_tree_Tree;
struct __pyx_obj_7sklearn_4tree_5_tree_Splitter;
struct __pyx_obj_7sklearn_4tree_5_tree_RandomSplitter;
struct __pyx_obj_7sklearn_4tree_5_tree_Criterion;
struct __pyx_obj_7sklearn_4tree_5_tree_ClassificationCriterion;
struct __pyx_obj_7sklearn_4tree_5_tree_Gini;
struct __pyx_obj_7sklearn_4tree_5_tree_Entropy;
struct __pyx_obj_7sklearn_4tree_5_tree_RegressionCriterion;
struct __pyx_obj_7sklearn_4tree_5_tree_MSE;
struct __pyx_obj_7sklearn_4tree_5_tree_BestSplitter;

/* "numpy.pxd":762
 * ctypedef npy_longdouble longdouble_t
 * 
 * ctypedef npy_cfloat      cfloat_t             # <<<<<<<<<<<<<<
 * ctypedef npy_cdouble     cdouble_t
 * ctypedef npy_clongdouble clongdouble_t
 */
typedef npy_cfloat __pyx_t_5numpy_cfloat_t;

/* "numpy.pxd":763
 * 
 * ctypedef npy_cfloat      cfloat_t
 * ctypedef npy_cdouble     cdouble_t             # <<<<<<<<<<<<<<
 * ctypedef npy_clongdouble clongdouble_t
 * 
 */
typedef npy_cdouble __pyx_t_5numpy_cdouble_t;

/* "numpy.pxd":764
 * ctypedef npy_cfloat      cfloat_t
 * ctypedef npy_cdouble     cdouble_t
 * ctypedef npy_clongdouble clongdouble_t             # <<<<<<<<<<<<<<
 * 
 * ctypedef npy_cdouble     complex_t
 */
typedef npy_clongdouble __pyx_t_5numpy_clongdouble_t;

/* "numpy.pxd":766
 * ctypedef npy_clongdouble clongdouble_t
 * 
 * ctypedef npy_cdouble     complex_t             # <<<<<<<<<<<<<<
 * 
 * cdef inline object PyArray_MultiIterNew1(a):
 */
typedef npy_cdouble __pyx_t_5numpy_complex_t;
struct __pyx_opt_args_7sklearn_4tree_5_tree_4Tree__resize;
struct __pyx_opt_args_7sklearn_4tree_5_tree_4Tree_build;
struct __pyx_opt_args_7sklearn_4tree_5_tree_4Tree_compute_feature_importances;

/* "sklearn/tree/_tree.pxd":136
 *                                 double impurity,
 *                                 SIZE_t n_node_samples)
 *     cdef void _resize(self, SIZE_t capacity=*)             # <<<<<<<<<<<<<<
 * 
 *     cpdef build(self, np.ndarray X,
 */
struct __pyx_opt_args_7sklearn_4tree_5_tree_4Tree__resize {
  int __pyx_n;
  __pyx_t_7sklearn_4tree_5_tree_SIZE_t capacity;
};

/* "sklearn/tree/_tree.pxd":138
 *     cdef void _resize(self, SIZE_t capacity=*)
 * 
 *     cpdef build(self, np.ndarray X,             # <<<<<<<<<<<<<<
 *                       np.ndarray y,
 *                       np.ndarray sample_weight=*)
 */
struct __pyx_opt_args_7sklearn_4tree_5_tree_4Tree_build {
  int __pyx_n;
  PyArrayObject *sample_weight;
};

/* "sklearn/tree/_tree.pxd":144
 *     cpdef predict(self, np.ndarray[DTYPE_t, ndim=2] X)
 *     cpdef apply(self, np.ndarray[DTYPE_t, ndim=2] X)
 *     cpdef compute_feature_importances(self, normalize=*)             # <<<<<<<<<<<<<<
 */
struct __pyx_opt_args_7sklearn_4tree_5_tree_4Tree_compute_feature_importances {
  int __pyx_n;
  PyObject *normalize;
};

/* "sklearn/tree/_tree.pyx":27
 * np.import_array()
 * 
 * cdef enum:             # <<<<<<<<<<<<<<
 *     # Max value for our rand_r replacement (near the bottom).
 *     # We don't use RAND_MAX because it's different across platforms and
 */
enum  {
  __pyx_e_7sklearn_4tree_5_tree_RAND_R_MAX = 0x7FFFFFFF
};

/* "sklearn/tree/_tree.pxd":102
 * # =============================================================================
 * 
 * cdef class Tree:             # <<<<<<<<<<<<<<
 *     # Input/Output layout
 *     cdef public SIZE_t n_features        # Number of features in X
 */
struct __pyx_obj_7sklearn_4tree_5_tree_Tree {
  PyObject_HEAD
  struct __pyx_vtabstruct_7sklearn_4tree_5_tree_Tree *__pyx_vtab;
  __pyx_t_7sklearn_4tree_5_tree_SIZE_t n_features;
  __pyx_t_7sklearn_4tree_5_tree_SIZE_t *n_classes;
  __pyx_t_7sklearn_4tree_5_tree_SIZE_t n_outputs;
  __pyx_t_7sklearn_4tree_5_tree_SIZE_t max_n_classes;
  __pyx_t_7sklearn_4tree_5_tree_SIZE_t value_stride;
  struct __pyx_obj_7sklearn_4tree_5_tree_Splitter *splitter;
  __pyx_t_7sklearn_4tree_5_tree_SIZE_t max_depth;
  __pyx_t_7sklearn_4tree_5_tree_SIZE_t min_samples_split;
  __pyx_t_7sklearn_4tree_5_tree_SIZE_t min_samples_leaf;
  PyObject *random_state;
  __pyx_t_7sklearn_4tree_5_tree_SIZE_t node_count;
  __pyx_t_7sklearn_4tree_5_tree_SIZE_t capacity;
  __pyx_t_7sklearn_4tree_5_tree_SIZE_t *children_left;
  __pyx_t_7sklearn_4tree_5_tree_SIZE_t *children_right;
  __pyx_t_7sklearn_4tree_5_tree_SIZE_t *feature;
  double *threshold;
  double *value;
  double *impurity;
  __pyx_t_7sklearn_4tree_5_tree_SIZE_t *n_node_samples;
};


/* "sklearn/tree/_tree.pxd":57
 * # Splitter
 * # =============================================================================
 * cdef class Splitter:             # <<<<<<<<<<<<<<
 *     # Internal structures
 *     cdef public Criterion criterion      # Impurity criterion
 */
struct __pyx_obj_7sklearn_4tree_5_tree_Splitter {
  PyObject_HEAD
  struct __pyx_vtabstruct_7sklearn_4tree_5_tree_Splitter *__pyx_vtab;
  struct __pyx_obj_7sklearn_4tree_5_tree_Criterion *criterion;
  __pyx_t_7sklearn_4tree_5_tree_SIZE_t max_features;
  __pyx_t_7sklearn_4tree_5_tree_SIZE_t min_samples_leaf;
  PyObject *random_state;
  unsigned int rand_r_state;
  __pyx_t_7sklearn_4tree_5_tree_SIZE_t *samples;
  __pyx_t_7sklearn_4tree_5_tree_SIZE_t n_samples;
  __pyx_t_7sklearn_4tree_5_tree_SIZE_t *features;
  __pyx_t_7sklearn_4tree_5_tree_SIZE_t n_features;
  __pyx_t_7sklearn_4tree_5_tree_SIZE_t start;
  __pyx_t_7sklearn_4tree_5_tree_SIZE_t end;
  PyArrayObject *X;
  __pyx_t_7sklearn_4tree_5_tree_DOUBLE_t *y;
  __pyx_t_7sklearn_4tree_5_tree_SIZE_t y_stride;
  __pyx_t_7sklearn_4tree_5_tree_DOUBLE_t *sample_weight;
};


/* "sklearn/tree/_tree.pyx":1054
 * 
 * 
 * cdef class RandomSplitter(Splitter):             # <<<<<<<<<<<<<<
 *     """Splitter for finding the best random split."""
 *     def __reduce__(self):
 */
struct __pyx_obj_7sklearn_4tree_5_tree_RandomSplitter {
  struct __pyx_obj_7sklearn_4tree_5_tree_Splitter __pyx_base;
};


/* "sklearn/tree/_tree.pxd":20
 * # =============================================================================
 * 
 * cdef class Criterion:             # <<<<<<<<<<<<<<
 *     # Internal structures
 *     cdef DOUBLE_t* y                     # Values of y
 */
struct __pyx_obj_7sklearn_4tree_5_tree_Criterion {
  PyObject_HEAD
  struct __pyx_vtabstruct_7sklearn_4tree_5_tree_Criterion *__pyx_vtab;
  __pyx_t_7sklearn_4tree_5_tree_DOUBLE_t *y;
  __pyx_t_7sklearn_4tree_5_tree_SIZE_t y_stride;
  __pyx_t_7sklearn_4tree_5_tree_DOUBLE_t *sample_weight;
  __pyx_t_7sklearn_4tree_5_tree_SIZE_t *samples;
  __pyx_t_7sklearn_4tree_5_tree_SIZE_t start;
  __pyx_t_7sklearn_4tree_5_tree_SIZE_t pos;
  __pyx_t_7sklearn_4tree_5_tree_SIZE_t end;
  __pyx_t_7sklearn_4tree_5_tree_SIZE_t n_outputs;
  __pyx_t_7sklearn_4tree_5_tree_SIZE_t n_node_samples;
  double weighted_n_node_samples;
  double weighted_n_left;
  double weighted_n_right;
};


/* "sklearn/tree/_tree.pyx":89
 * 
 * 
 * cdef class ClassificationCriterion(Criterion):             # <<<<<<<<<<<<<<
 *     """Abstract criterion for classification."""
 *     cdef SIZE_t* n_classes
 */
struct __pyx_obj_7sklearn_4tree_5_tree_ClassificationCriterion {
  struct __pyx_obj_7sklearn_4tree_5_tree_Criterion __pyx_base;
  __pyx_t_7sklearn_4tree_5_tree_SIZE_t *n_classes;
  __pyx_t_7sklearn_4tree_5_tree_SIZE_t label_count_stride;
  double *label_count_left;
  double *label_count_right;
  double *label_count_total;
};


/* "sklearn/tree/_tree.pyx":402
 * 
 * 
 * cdef class Gini(ClassificationCriterion):             # <<<<<<<<<<<<<<
 *     """Gini Index impurity criteria.
 * 
 */
struct __pyx_obj_7sklearn_4tree_5_tree_Gini {
  struct __pyx_obj_7sklearn_4tree_5_tree_ClassificationCriterion __pyx_base;
};


/* "sklearn/tree/_tree.pyx":314
 * 
 * 
 * cdef class Entropy(ClassificationCriterion):             # <<<<<<<<<<<<<<
 *     """Cross Entropy impurity criteria.
 * 
 */
struct __pyx_obj_7sklearn_4tree_5_tree_Entropy {
  struct __pyx_obj_7sklearn_4tree_5_tree_ClassificationCriterion __pyx_base;
};


/* "sklearn/tree/_tree.pyx":492
 * 
 * 
 * cdef class RegressionCriterion(Criterion):             # <<<<<<<<<<<<<<
 *     """Abstract criterion for regression.
 * 
 */
struct __pyx_obj_7sklearn_4tree_5_tree_RegressionCriterion {
  struct __pyx_obj_7sklearn_4tree_5_tree_Criterion __pyx_base;
  double *mean_left;
  double *mean_right;
  double *mean_total;
  double *sq_sum_left;
  double *sq_sum_right;
  double *sq_sum_total;
  double *var_left;
  double *var_right;
};


/* "sklearn/tree/_tree.pyx":748
 *             dest[k] = mean_total[k]
 * 
 * cdef class MSE(RegressionCriterion):             # <<<<<<<<<<<<<<
 *     """Mean squared error impurity criterion.
 * 
 */
struct __pyx_obj_7sklearn_4tree_5_tree_MSE {
  struct __pyx_obj_7sklearn_4tree_5_tree_RegressionCriterion __pyx_base;
};


/* "sklearn/tree/_tree.pyx":892
 * 
 * 
 * cdef class BestSplitter(Splitter):             # <<<<<<<<<<<<<<
 *     """Splitter for finding the best split."""
 *     def __reduce__(self):
 */
struct __pyx_obj_7sklearn_4tree_5_tree_BestSplitter {
  struct __pyx_obj_7sklearn_4tree_5_tree_Splitter __pyx_base;
};



/* "sklearn/tree/_tree.pyx":1196
 * # =============================================================================
 * 
 * cdef class Tree:             # <<<<<<<<<<<<<<
 *     """Struct-of-arrays representation of a binary decision tree.
 * 
 */

struct __pyx_vtabstruct_7sklearn_4tree_5_tree_Tree {
  __pyx_t_7sklearn_4tree_5_tree_SIZE_t (*_add_node)(struct __pyx_obj_7sklearn_4tree_5_tree_Tree *, __pyx_t_7sklearn_4tree_5_tree_SIZE_t, int, int, __pyx_t_7sklearn_4tree_5_tree_SIZE_t, double, double, __pyx_t_7sklearn_4tree_5_tree_SIZE_t);
  void (*_resize)(struct __pyx_obj_7sklearn_4tree_5_tree_Tree *, struct __pyx_opt_args_7sklearn_4tree_5_tree_4Tree__resize *__pyx_optional_args);
  PyObject *(*build)(struct __pyx_obj_7sklearn_4tree_5_tree_Tree *, PyArrayObject *, PyArrayObject *, int __pyx_skip_dispatch, struct __pyx_opt_args_7sklearn_4tree_5_tree_4Tree_build *__pyx_optional_args);
  PyObject *(*predict)(struct __pyx_obj_7sklearn_4tree_5_tree_Tree *, PyArrayObject *, int __pyx_skip_dispatch);
  PyObject *(*apply)(struct __pyx_obj_7sklearn_4tree_5_tree_Tree *, PyArrayObject *, int __pyx_skip_dispatch);
  PyObject *(*compute_feature_importances)(struct __pyx_obj_7sklearn_4tree_5_tree_Tree *, int __pyx_skip_dispatch, struct __pyx_opt_args_7sklearn_4tree_5_tree_4Tree_compute_feature_importances *__pyx_optional_args);
};
static struct __pyx_vtabstruct_7sklearn_4tree_5_tree_Tree *__pyx_vtabptr_7sklearn_4tree_5_tree_Tree;


/* "sklearn/tree/_tree.pyx":52
 * # =============================================================================
 * 
 * cdef class Criterion:             # <<<<<<<<<<<<<<
 *     """Interface for impurity criteria."""
 * 
 */

struct __pyx_vtabstruct_7sklearn_4tree_5_tree_Criterion {
  void (*init)(struct __pyx_obj_7sklearn_4tree_5_tree_Criterion *, __pyx_t_7sklearn_4tree_5_tree_DOUBLE_t *, __pyx_t_7sklearn_4tree_5_tree_SIZE_t, __pyx_t_7sklearn_4tree_5_tree_DOUBLE_t *, __pyx_t_7sklearn_4tree_5_tree_SIZE_t *, __pyx_t_7sklearn_4tree_5_tree_SIZE_t, __pyx_t_7sklearn_4tree_5_tree_SIZE_t);
  void (*reset)(struct __pyx_obj_7sklearn_4tree_5_tree_Criterion *);
  void (*update)(struct __pyx_obj_7sklearn_4tree_5_tree_Criterion *, __pyx_t_7sklearn_4tree_5_tree_SIZE_t);
  double (*node_impurity)(struct __pyx_obj_7sklearn_4tree_5_tree_Criterion *);
  double (*children_impurity)(struct __pyx_obj_7sklearn_4tree_5_tree_Criterion *);
  void (*node_value)(struct __pyx_obj_7sklearn_4tree_5_tree_Criterion *, double *);
};
static struct __pyx_vtabstruct_7sklearn_4tree_5_tree_Criterion *__pyx_vtabptr_7sklearn_4tree_5_tree_Criterion;


/* "sklearn/tree/_tree.pyx":492
 * 
 * 
 * cdef class RegressionCriterion(Criterion):             # <<<<<<<<<<<<<<
 *     """Abstract criterion for regression.
 * 
 */

struct __pyx_vtabstruct_7sklearn_4tree_5_tree_RegressionCriterion {
  struct __pyx_vtabstruct_7sklearn_4tree_5_tree_Criterion __pyx_base;
};
static struct __pyx_vtabstruct_7sklearn_4tree_5_tree_RegressionCriterion *__pyx_vtabptr_7sklearn_4tree_5_tree_RegressionCriterion;


/* "sklearn/tree/_tree.pyx":748
 *             dest[k] = mean_total[k]
 * 
 * cdef class MSE(RegressionCriterion):             # <<<<<<<<<<<<<<
 *     """Mean squared error impurity criterion.
 * 
 */

struct __pyx_vtabstruct_7sklearn_4tree_5_tree_MSE {
  struct __pyx_vtabstruct_7sklearn_4tree_5_tree_RegressionCriterion __pyx_base;
};
static struct __pyx_vtabstruct_7sklearn_4tree_5_tree_MSE *__pyx_vtabptr_7sklearn_4tree_5_tree_MSE;


/* "sklearn/tree/_tree.pyx":89
 * 
 * 
 * cdef class ClassificationCriterion(Criterion):             # <<<<<<<<<<<<<<
 *     """Abstract criterion for classification."""
 *     cdef SIZE_t* n_classes
 */

struct __pyx_vtabstruct_7sklearn_4tree_5_tree_ClassificationCriterion {
  struct __pyx_vtabstruct_7sklearn_4tree_5_tree_Criterion __pyx_base;
};
static struct __pyx_vtabstruct_7sklearn_4tree_5_tree_ClassificationCriterion *__pyx_vtabptr_7sklearn_4tree_5_tree_ClassificationCriterion;


/* "sklearn/tree/_tree.pyx":402
 * 
 * 
 * cdef class Gini(ClassificationCriterion):             # <<<<<<<<<<<<<<
 *     """Gini Index impurity criteria.
 * 
 */

struct __pyx_vtabstruct_7sklearn_4tree_5_tree_Gini {
  struct __pyx_vtabstruct_7sklearn_4tree_5_tree_ClassificationCriterion __pyx_base;
};
static struct __pyx_vtabstruct_7sklearn_4tree_5_tree_Gini *__pyx_vtabptr_7sklearn_4tree_5_tree_Gini;


/* "sklearn/tree/_tree.pyx":790
 * # =============================================================================
 * 
 * cdef class Splitter:             # <<<<<<<<<<<<<<
 *     def __cinit__(self, Criterion criterion,
 *                         SIZE_t max_features,
 */

struct __pyx_vtabstruct_7sklearn_4tree_5_tree_Splitter {
  void (*init)(struct __pyx_obj_7sklearn_4tree_5_tree_Splitter *, PyArrayObject *, PyArrayObject *, __pyx_t_7sklearn_4tree_5_tree_DOUBLE_t *);
  void (*node_reset)(struct __pyx_obj_7sklearn_4tree_5_tree_Splitter *, __pyx_t_7sklearn_4tree_5_tree_SIZE_t, __pyx_t_7sklearn_4tree_5_tree_SIZE_t, double *);
  void (*node_split)(struct __pyx_obj_7sklearn_4tree_5_tree_Splitter *, __pyx_t_7sklearn_4tree_5_tree_SIZE_t *, __pyx_t_7sklearn_4tree_5_tree_SIZE_t *, double *);
  void (*node_value)(struct __pyx_obj_7sklearn_4tree_5_tree_Splitter *, double *);
};
static struct __pyx_vtabstruct_7sklearn_4tree_5_tree_Splitter *__pyx_vtabptr_7sklearn_4tree_5_tree_Splitter;


/* "sklearn/tree/_tree.pyx":314
 * 
 * 
 * cdef class Entropy(ClassificationCriterion):             # <<<<<<<<<<<<<<
 *     """Cross Entropy impurity criteria.
 * 
 */

struct __pyx_vtabstruct_7sklearn_4tree_5_tree_Entropy {
  struct __pyx_vtabstruct_7sklearn_4tree_5_tree_ClassificationCriterion __pyx_base;
};
static struct __pyx_vtabstruct_7sklearn_4tree_5_tree_Entropy *__pyx_vtabptr_7sklearn_4tree_5_tree_Entropy;


/* "sklearn/tree/_tree.pyx":1054
 * 
 * 
 * cdef class RandomSplitter(Splitter):             # <<<<<<<<<<<<<<
 *     """Splitter for finding the best random split."""
 *     def __reduce__(self):
 */

struct __pyx_vtabstruct_7sklearn_4tree_5_tree_RandomSplitter {
  struct __pyx_vtabstruct_7sklearn_4tree_5_tree_Splitter __pyx_base;
};
static struct __pyx_vtabstruct_7sklearn_4tree_5_tree_RandomSplitter *__pyx_vtabptr_7sklearn_4tree_5_tree_RandomSplitter;


/* "sklearn/tree/_tree.pyx":892
 * 
 * 
 * cdef class BestSplitter(Splitter):             # <<<<<<<<<<<<<<
 *     """Splitter for finding the best split."""
 *     def __reduce__(self):
 */

struct __pyx_vtabstruct_7sklearn_4tree_5_tree_BestSplitter {
  struct __pyx_vtabstruct_7sklearn_4tree_5_tree_Splitter __pyx_base;
};
static struct __pyx_vtabstruct_7sklearn_4tree_5_tree_BestSplitter *__pyx_vtabptr_7sklearn_4tree_5_tree_BestSplitter;
#ifndef CYTHON_REFNANNY
  #define CYTHON_REFNANNY 0
#endif
#if CYTHON_REFNANNY
  typedef struct {
    void (*INCREF)(void*, PyObject*, int);
    void (*DECREF)(void*, PyObject*, int);
    void (*GOTREF)(void*, PyObject*, int);
    void (*GIVEREF)(void*, PyObject*, int);
    void* (*SetupContext)(const char*, int, const char*);
    void (*FinishContext)(void**);
  } __Pyx_RefNannyAPIStruct;
  static __Pyx_RefNannyAPIStruct *__Pyx_RefNanny = NULL;
  static __Pyx_RefNannyAPIStruct *__Pyx_RefNannyImportAPI(const char *modname); /*proto*/
  #define __Pyx_RefNannyDeclarations void *__pyx_refnanny = NULL;
#ifdef WITH_THREAD
  #define __Pyx_RefNannySetupContext(name, acquire_gil) \
          if (acquire_gil) { \
              PyGILState_STATE __pyx_gilstate_save = PyGILState_Ensure(); \
              __pyx_refnanny = __Pyx_RefNanny->SetupContext((name), __LINE__, __FILE__); \
              PyGILState_Release(__pyx_gilstate_save); \
          } else { \
              __pyx_refnanny = __Pyx_RefNanny->SetupContext((name), __LINE__, __FILE__); \
          }
#else
  #define __Pyx_RefNannySetupContext(name, acquire_gil) \
          __pyx_refnanny = __Pyx_RefNanny->SetupContext((name), __LINE__, __FILE__)
#endif
  #define __Pyx_RefNannyFinishContext() \
          __Pyx_RefNanny->FinishContext(&__pyx_refnanny)
  #define __Pyx_INCREF(r)  __Pyx_RefNanny->INCREF(__pyx_refnanny, (PyObject *)(r), __LINE__)
  #define __Pyx_DECREF(r)  __Pyx_RefNanny->DECREF(__pyx_refnanny, (PyObject *)(r), __LINE__)
  #define __Pyx_GOTREF(r)  __Pyx_RefNanny->GOTREF(__pyx_refnanny, (PyObject *)(r), __LINE__)
  #define __Pyx_GIVEREF(r) __Pyx_RefNanny->GIVEREF(__pyx_refnanny, (PyObject *)(r), __LINE__)
  #define __Pyx_XINCREF(r)  do { if((r) != NULL) {__Pyx_INCREF(r); }} while(0)
  #define __Pyx_XDECREF(r)  do { if((r) != NULL) {__Pyx_DECREF(r); }} while(0)
  #define __Pyx_XGOTREF(r)  do { if((r) != NULL) {__Pyx_GOTREF(r); }} while(0)
  #define __Pyx_XGIVEREF(r) do { if((r) != NULL) {__Pyx_GIVEREF(r);}} while(0)
#else
  #define __Pyx_RefNannyDeclarations
  #define __Pyx_RefNannySetupContext(name, acquire_gil)
  #define __Pyx_RefNannyFinishContext()
  #define __Pyx_INCREF(r) Py_INCREF(r)
  #define __Pyx_DECREF(r) Py_DECREF(r)
  #define __Pyx_GOTREF(r)
  #define __Pyx_GIVEREF(r)
  #define __Pyx_XINCREF(r) Py_XINCREF(r)
  #define __Pyx_XDECREF(r) Py_XDECREF(r)
  #define __Pyx_XGOTREF(r)
  #define __Pyx_XGIVEREF(r)
#endif /* CYTHON_REFNANNY */
#define __Pyx_CLEAR(r)    do { PyObject* tmp = ((PyObject*)(r)); r = NULL; __Pyx_DECREF(tmp);} while(0)
#define __Pyx_XCLEAR(r)   do { if((r) != NULL) {PyObject* tmp = ((PyObject*)(r)); r = NULL; __Pyx_DECREF(tmp);}} while(0)

#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyObject_GetAttrStr(PyObject* obj, PyObject* attr_name) {
    PyTypeObject* tp = Py_TYPE(obj);
    if (likely(tp->tp_getattro))
        return tp->tp_getattro(obj, attr_name);
#if PY_MAJOR_VERSION < 3
    if (likely(tp->tp_getattr))
        return tp->tp_getattr(obj, PyString_AS_STRING(attr_name));
#endif
    return PyObject_GetAttr(obj, attr_name);
}
#else
#define __Pyx_PyObject_GetAttrStr(o,n) PyObject_GetAttr(o,n)
#endif

static PyObject *__Pyx_GetBuiltinName(PyObject *name); /*proto*/

static void __Pyx_RaiseArgtupleInvalid(const char* func_name, int exact,
    Py_ssize_t num_min, Py_ssize_t num_max, Py_ssize_t num_found); /*proto*/

static void __Pyx_RaiseDoubleKeywordsError(const char* func_name, PyObject* kw_name); /*proto*/

static int __Pyx_ParseOptionalKeywords(PyObject *kwds, PyObject **argnames[], \
    PyObject *kwds2, PyObject *values[], Py_ssize_t num_pos_args, \
    const char* function_name); /*proto*/

static int __Pyx_ArgTypeTest(PyObject *obj, PyTypeObject *type, int none_allowed,
    const char *name, int exact); /*proto*/

static CYTHON_INLINE int  __Pyx_GetBufferAndValidate(Py_buffer* buf, PyObject* obj,
    __Pyx_TypeInfo* dtype, int flags, int nd, int cast, __Pyx_BufFmt_StackElem* stack);
static CYTHON_INLINE void __Pyx_SafeReleaseBuffer(Py_buffer* info);

#define __Pyx_BufPtrStrided1d(type, buf, i0, s0) (type)((char*)buf + i0 * s0)
static CYTHON_INLINE void __Pyx_ErrRestore(PyObject *type, PyObject *value, PyObject *tb); /*proto*/
static CYTHON_INLINE void __Pyx_ErrFetch(PyObject **type, PyObject **value, PyObject **tb); /*proto*/

static void __Pyx_WriteUnraisable(const char *name, int clineno,
                                  int lineno, const char *filename); /*proto*/

static CYTHON_INLINE int __Pyx_TypeTest(PyObject *obj, PyTypeObject *type); /*proto*/

#define __Pyx_BufPtrCContig2d(type, buf, i0, s0, i1, s1) ((type)((char*)buf + i0 * s0) + i1)
static CYTHON_INLINE PyObject *__Pyx_GetModuleGlobalName(PyObject *name); /*proto*/

static void __Pyx_RaiseBufferFallbackError(void); /*proto*/

#define __Pyx_BufPtrStrided2d(type, buf, i0, s0, i1, s1) (type)((char*)buf + i0 * s0 + i1 * s1)
#define __Pyx_BufPtrStrided3d(type, buf, i0, s0, i1, s1, i2, s2) (type)((char*)buf + i0 * s0 + i1 * s1 + i2 * s2)
static void __Pyx_Raise(PyObject *type, PyObject *value, PyObject *tb, PyObject *cause); /*proto*/

static CYTHON_INLINE void __Pyx_RaiseTooManyValuesError(Py_ssize_t expected);

static CYTHON_INLINE void __Pyx_RaiseNeedMoreValuesError(Py_ssize_t index);

static CYTHON_INLINE void __Pyx_RaiseNoneNotIterableError(void);

static CYTHON_INLINE int __Pyx_IterFinish(void); /*proto*/

static int __Pyx_IternextUnpackEndCheck(PyObject *retval, Py_ssize_t expected); /*proto*/

static int __Pyx_SetVtable(PyObject *dict, void *vtable); /*proto*/

static PyObject* __Pyx_ImportFrom(PyObject* module, PyObject* name); /*proto*/

static CYTHON_INLINE PyObject *__Pyx_PyInt_to_py_Py_intptr_t(Py_intptr_t);

static CYTHON_INLINE Py_intptr_t __Pyx_PyInt_from_py_Py_intptr_t(PyObject *);

static PyObject *__Pyx_Import(PyObject *name, PyObject *from_list, int level); /*proto*/

typedef struct {
  Py_ssize_t shape, strides, suboffsets;
} __Pyx_Buf_DimInfo;
typedef struct {
  size_t refcount;
  Py_buffer pybuffer;
} __Pyx_Buffer;
typedef struct {
  __Pyx_Buffer *rcbuffer;
  char *data;
  __Pyx_Buf_DimInfo diminfo[8];
} __Pyx_LocalBuf_ND;

#if PY_MAJOR_VERSION < 3
    static int __Pyx_GetBuffer(PyObject *obj, Py_buffer *view, int flags);
    static void __Pyx_ReleaseBuffer(Py_buffer *view);
#else
    #define __Pyx_GetBuffer PyObject_GetBuffer
    #define __Pyx_ReleaseBuffer PyBuffer_Release
#endif


static Py_ssize_t __Pyx_zeros[] = {0, 0, 0, 0, 0, 0, 0, 0};
static Py_ssize_t __Pyx_minusones[] = {-1, -1, -1, -1, -1, -1, -1, -1};

static CYTHON_INLINE long __Pyx_pow_long(long, long); /* proto */

#if CYTHON_CCOMPLEX
  #ifdef __cplusplus
    #define __Pyx_CREAL(z) ((z).real())
    #define __Pyx_CIMAG(z) ((z).imag())
  #else
    #define __Pyx_CREAL(z) (__real__(z))
    #define __Pyx_CIMAG(z) (__imag__(z))
  #endif
#else
    #define __Pyx_CREAL(z) ((z).real)
    #define __Pyx_CIMAG(z) ((z).imag)
#endif
#if defined(_WIN32) && defined(__cplusplus) && CYTHON_CCOMPLEX
    #define __Pyx_SET_CREAL(z,x) ((z).real(x))
    #define __Pyx_SET_CIMAG(z,y) ((z).imag(y))
#else
    #define __Pyx_SET_CREAL(z,x) __Pyx_CREAL(z) = (x)
    #define __Pyx_SET_CIMAG(z,y) __Pyx_CIMAG(z) = (y)
#endif

static CYTHON_INLINE __pyx_t_float_complex __pyx_t_float_complex_from_parts(float, float);

#if CYTHON_CCOMPLEX
    #define __Pyx_c_eqf(a, b)   ((a)==(b))
    #define __Pyx_c_sumf(a, b)  ((a)+(b))
    #define __Pyx_c_difff(a, b) ((a)-(b))
    #define __Pyx_c_prodf(a, b) ((a)*(b))
    #define __Pyx_c_quotf(a, b) ((a)/(b))
    #define __Pyx_c_negf(a)     (-(a))
  #ifdef __cplusplus
    #define __Pyx_c_is_zerof(z) ((z)==(float)0)
    #define __Pyx_c_conjf(z)    (::std::conj(z))
    #if 1
        #define __Pyx_c_absf(z)     (::std::abs(z))
        #define __Pyx_c_powf(a, b)  (::std::pow(a, b))
    #endif
  #else
    #define __Pyx_c_is_zerof(z) ((z)==0)
    #define __Pyx_c_conjf(z)    (conjf(z))
    #if 1
        #define __Pyx_c_absf(z)     (cabsf(z))
        #define __Pyx_c_powf(a, b)  (cpowf(a, b))
    #endif
 #endif
#else
    static CYTHON_INLINE int __Pyx_c_eqf(__pyx_t_float_complex, __pyx_t_float_complex);
    static CYTHON_INLINE __pyx_t_float_complex __Pyx_c_sumf(__pyx_t_float_complex, __pyx_t_float_complex);
    static CYTHON_INLINE __pyx_t_float_complex __Pyx_c_difff(__pyx_t_float_complex, __pyx_t_float_complex);
    static CYTHON_INLINE __pyx_t_float_complex __Pyx_c_prodf(__pyx_t_float_complex, __pyx_t_float_complex);
    static CYTHON_INLINE __pyx_t_float_complex __Pyx_c_quotf(__pyx_t_float_complex, __pyx_t_float_complex);
    static CYTHON_INLINE __pyx_t_float_complex __Pyx_c_negf(__pyx_t_float_complex);
    static CYTHON_INLINE int __Pyx_c_is_zerof(__pyx_t_float_complex);
    static CYTHON_INLINE __pyx_t_float_complex __Pyx_c_conjf(__pyx_t_float_complex);
    #if 1
        static CYTHON_INLINE float __Pyx_c_absf(__pyx_t_float_complex);
        static CYTHON_INLINE __pyx_t_float_complex __Pyx_c_powf(__pyx_t_float_complex, __pyx_t_float_complex);
    #endif
#endif

static CYTHON_INLINE __pyx_t_double_complex __pyx_t_double_complex_from_parts(double, double);

#if CYTHON_CCOMPLEX
    #define __Pyx_c_eq(a, b)   ((a)==(b))
    #define __Pyx_c_sum(a, b)  ((a)+(b))
    #define __Pyx_c_diff(a, b) ((a)-(b))
    #define __Pyx_c_prod(a, b) ((a)*(b))
    #define __Pyx_c_quot(a, b) ((a)/(b))
    #define __Pyx_c_neg(a)     (-(a))
  #ifdef __cplusplus
    #define __Pyx_c_is_zero(z) ((z)==(double)0)
    #define __Pyx_c_conj(z)    (::std::conj(z))
    #if 1
        #define __Pyx_c_abs(z)     (::std::abs(z))
        #define __Pyx_c_pow(a, b)  (::std::pow(a, b))
    #endif
  #else
    #define __Pyx_c_is_zero(z) ((z)==0)
    #define __Pyx_c_conj(z)    (conj(z))
    #if 1
        #define __Pyx_c_abs(z)     (cabs(z))
        #define __Pyx_c_pow(a, b)  (cpow(a, b))
    #endif
 #endif
#else
    static CYTHON_INLINE int __Pyx_c_eq(__pyx_t_double_complex, __pyx_t_double_complex);
    static CYTHON_INLINE __pyx_t_double_complex __Pyx_c_sum(__pyx_t_double_complex, __pyx_t_double_complex);
    static CYTHON_INLINE __pyx_t_double_complex __Pyx_c_diff(__pyx_t_double_complex, __pyx_t_double_complex);
    static CYTHON_INLINE __pyx_t_double_complex __Pyx_c_prod(__pyx_t_double_complex, __pyx_t_double_complex);
    static CYTHON_INLINE __pyx_t_double_complex __Pyx_c_quot(__pyx_t_double_complex, __pyx_t_double_complex);
    static CYTHON_INLINE __pyx_t_double_complex __Pyx_c_neg(__pyx_t_double_complex);
    static CYTHON_INLINE int __Pyx_c_is_zero(__pyx_t_double_complex);
    static CYTHON_INLINE __pyx_t_double_complex __Pyx_c_conj(__pyx_t_double_complex);
    #if 1
        static CYTHON_INLINE double __Pyx_c_abs(__pyx_t_double_complex);
        static CYTHON_INLINE __pyx_t_double_complex __Pyx_c_pow(__pyx_t_double_complex, __pyx_t_double_complex);
    #endif
#endif

static CYTHON_INLINE unsigned char __Pyx_PyInt_AsUnsignedChar(PyObject *);

static CYTHON_INLINE unsigned short __Pyx_PyInt_AsUnsignedShort(PyObject *);

static CYTHON_INLINE unsigned int __Pyx_PyInt_AsUnsignedInt(PyObject *);

static CYTHON_INLINE char __Pyx_PyInt_AsChar(PyObject *);

static CYTHON_INLINE short __Pyx_PyInt_AsShort(PyObject *);

static CYTHON_INLINE int __Pyx_PyInt_AsInt(PyObject *);

static CYTHON_INLINE signed char __Pyx_PyInt_AsSignedChar(PyObject *);

static CYTHON_INLINE signed short __Pyx_PyInt_AsSignedShort(PyObject *);

static CYTHON_INLINE signed int __Pyx_PyInt_AsSignedInt(PyObject *);

static CYTHON_INLINE int __Pyx_PyInt_AsLongDouble(PyObject *);

static CYTHON_INLINE unsigned long __Pyx_PyInt_AsUnsignedLong(PyObject *);

static CYTHON_INLINE unsigned PY_LONG_LONG __Pyx_PyInt_AsUnsignedLongLong(PyObject *);

static CYTHON_INLINE long __Pyx_PyInt_AsLong(PyObject *);

static CYTHON_INLINE PY_LONG_LONG __Pyx_PyInt_AsLongLong(PyObject *);

static CYTHON_INLINE signed long __Pyx_PyInt_AsSignedLong(PyObject *);

static CYTHON_INLINE signed PY_LONG_LONG __Pyx_PyInt_AsSignedLongLong(PyObject *);

static int __Pyx_check_binary_version(void);

#if !defined(__Pyx_PyIdentifier_FromString)
#if PY_MAJOR_VERSION < 3
  #define __Pyx_PyIdentifier_FromString(s) PyString_FromString(s)
#else
  #define __Pyx_PyIdentifier_FromString(s) PyUnicode_FromString(s)
#endif
#endif

static PyObject *__Pyx_ImportModule(const char *name); /*proto*/

static PyTypeObject *__Pyx_ImportType(const char *module_name, const char *class_name, size_t size, int strict);  /*proto*/

typedef struct {
    int code_line;
    PyCodeObject* code_object;
} __Pyx_CodeObjectCacheEntry;
struct __Pyx_CodeObjectCache {
    int count;
    int max_count;
    __Pyx_CodeObjectCacheEntry* entries;
};
static struct __Pyx_CodeObjectCache __pyx_code_cache = {0,0,NULL};
static int __pyx_bisect_code_objects(__Pyx_CodeObjectCacheEntry* entries, int count, int code_line);
static PyCodeObject *__pyx_find_code_object(int code_line);
static void __pyx_insert_code_object(int code_line, PyCodeObject* code_object);

static void __Pyx_AddTraceback(const char *funcname, int c_line,
                               int py_line, const char *filename); /*proto*/

static int __Pyx_InitStrings(__Pyx_StringTabEntry *t); /*proto*/


/* Module declarations from 'cpython.buffer' */

/* Module declarations from 'cpython.ref' */

/* Module declarations from 'libc.string' */

/* Module declarations from 'libc.stdio' */

/* Module declarations from 'cpython.object' */

/* Module declarations from '__builtin__' */

/* Module declarations from 'cpython.type' */
static PyTypeObject *__pyx_ptype_7cpython_4type_type = 0;

/* Module declarations from 'libc.stdlib' */

/* Module declarations from 'numpy' */

/* Module declarations from 'numpy' */
static PyTypeObject *__pyx_ptype_5numpy_dtype = 0;
static PyTypeObject *__pyx_ptype_5numpy_flatiter = 0;
static PyTypeObject *__pyx_ptype_5numpy_broadcast = 0;
static PyTypeObject *__pyx_ptype_5numpy_ndarray = 0;
static PyTypeObject *__pyx_ptype_5numpy_ufunc = 0;
static CYTHON_INLINE char *__pyx_f_5numpy__util_dtypestring(PyArray_Descr *, char *, char *, int *); /*proto*/

/* Module declarations from 'libc.math' */

/* Module declarations from 'sklearn.tree._tree' */
static PyTypeObject *__pyx_ptype_7sklearn_4tree_5_tree_Tree = 0;
static PyTypeObject *__pyx_ptype_7sklearn_4tree_5_tree_Splitter = 0;
static PyTypeObject *__pyx_ptype_7sklearn_4tree_5_tree_RandomSplitter = 0;
static PyTypeObject *__pyx_ptype_7sklearn_4tree_5_tree_Criterion = 0;
static PyTypeObject *__pyx_ptype_7sklearn_4tree_5_tree_ClassificationCriterion = 0;
static PyTypeObject *__pyx_ptype_7sklearn_4tree_5_tree_Gini = 0;
static PyTypeObject *__pyx_ptype_7sklearn_4tree_5_tree_Entropy = 0;
static PyTypeObject *__pyx_ptype_7sklearn_4tree_5_tree_RegressionCriterion = 0;
static PyTypeObject *__pyx_ptype_7sklearn_4tree_5_tree_MSE = 0;
static PyTypeObject *__pyx_ptype_7sklearn_4tree_5_tree_BestSplitter = 0;
static double __pyx_v_7sklearn_4tree_5_tree_INFINITY;
static __pyx_t_7sklearn_4tree_5_tree_SIZE_t __pyx_v_7sklearn_4tree_5_tree__TREE_LEAF;
static __pyx_t_7sklearn_4tree_5_tree_SIZE_t __pyx_v_7sklearn_4tree_5_tree__TREE_UNDEFINED;
static void __pyx_f_7sklearn_4tree_5_tree_sort(PyArrayObject *, __pyx_t_7sklearn_4tree_5_tree_SIZE_t, __pyx_t_7sklearn_4tree_5_tree_SIZE_t *, __pyx_t_7sklearn_4tree_5_tree_SIZE_t); /*proto*/
static CYTHON_INLINE int __pyx_f_7sklearn_4tree_5_tree_our_rand_r(unsigned int *); /*proto*/
static CYTHON_INLINE PyArrayObject *__pyx_f_7sklearn_4tree_5_tree_sizet_ptr_to_ndarray(__pyx_t_7sklearn_4tree_5_tree_SIZE_t *, __pyx_t_7sklearn_4tree_5_tree_SIZE_t); /*proto*/
static CYTHON_INLINE PyArrayObject *__pyx_f_7sklearn_4tree_5_tree_double_ptr_to_ndarray(double *, __pyx_t_7sklearn_4tree_5_tree_SIZE_t); /*proto*/
static CYTHON_INLINE __pyx_t_7sklearn_4tree_5_tree_SIZE_t __pyx_f_7sklearn_4tree_5_tree_rand_int(__pyx_t_7sklearn_4tree_5_tree_SIZE_t, unsigned int *); /*proto*/
static CYTHON_INLINE double __pyx_f_7sklearn_4tree_5_tree_rand_double(unsigned int *); /*proto*/
static CYTHON_INLINE double __pyx_f_7sklearn_4tree_5_tree_log(double); /*proto*/
static __Pyx_TypeInfo __Pyx_TypeInfo_nn___pyx_t_7sklearn_4tree_5_tree_SIZE_t = { "SIZE_t", NULL, sizeof(__pyx_t_7sklearn_4tree_5_tree_SIZE_t), { 0 }, 0, IS_UNSIGNED(__pyx_t_7sklearn_4tree_5_tree_SIZE_t) ? 'U' : 'I', IS_UNSIGNED(__pyx_t_7sklearn_4tree_5_tree_SIZE_t), 0 };
static __Pyx_TypeInfo __Pyx_TypeInfo_nn___pyx_t_7sklearn_4tree_5_tree_DTYPE_t = { "DTYPE_t", NULL, sizeof(__pyx_t_7sklearn_4tree_5_tree_DTYPE_t), { 0 }, 0, 'R', 0, 0 };
static __Pyx_TypeInfo __Pyx_TypeInfo_nn___pyx_t_7sklearn_4tree_5_tree_DOUBLE_t = { "DOUBLE_t", NULL, sizeof(__pyx_t_7sklearn_4tree_5_tree_DOUBLE_t), { 0 }, 0, 'R', 0, 0 };
static __Pyx_TypeInfo __Pyx_TypeInfo_nn___pyx_t_5numpy_float64_t = { "float64_t", NULL, sizeof(__pyx_t_5numpy_float64_t), { 0 }, 0, 'R', 0, 0 };
static __Pyx_TypeInfo __Pyx_TypeInfo_nn___pyx_t_5numpy_int32_t = { "int32_t", NULL, sizeof(__pyx_t_5numpy_int32_t), { 0 }, 0, IS_UNSIGNED(__pyx_t_5numpy_int32_t) ? 'U' : 'I', IS_UNSIGNED(__pyx_t_5numpy_int32_t), 0 };
#define __Pyx_MODULE_NAME "sklearn.tree._tree"
int __pyx_module_is_main_sklearn__tree___tree = 0;

/* Implementation of 'sklearn.tree._tree' */
static PyObject *__pyx_builtin_MemoryError;
static PyObject *__pyx_builtin_ValueError;
static PyObject *__pyx_builtin_range;
static PyObject *__pyx_builtin_RuntimeError;
static int __pyx_pf_7sklearn_4tree_5_tree_23ClassificationCriterion___cinit__(struct __pyx_obj_7sklearn_4tree_5_tree_ClassificationCriterion *__pyx_v_self, __pyx_t_7sklearn_4tree_5_tree_SIZE_t __pyx_v_n_outputs, PyArrayObject *__pyx_v_n_classes); /* proto */
static void __pyx_pf_7sklearn_4tree_5_tree_23ClassificationCriterion_2__dealloc__(struct __pyx_obj_7sklearn_4tree_5_tree_ClassificationCriterion *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_7sklearn_4tree_5_tree_23ClassificationCriterion_4__reduce__(struct __pyx_obj_7sklearn_4tree_5_tree_ClassificationCriterion *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_7sklearn_4tree_5_tree_23ClassificationCriterion_6__getstate__(CYTHON_UNUSED struct __pyx_obj_7sklearn_4tree_5_tree_ClassificationCriterion *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_7sklearn_4tree_5_tree_23ClassificationCriterion_8__setstate__(CYTHON_UNUSED struct __pyx_obj_7sklearn_4tree_5_tree_ClassificationCriterion *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v_d); /* proto */
static int __pyx_pf_7sklearn_4tree_5_tree_19RegressionCriterion___cinit__(struct __pyx_obj_7sklearn_4tree_5_tree_RegressionCriterion *__pyx_v_self, __pyx_t_7sklearn_4tree_5_tree_SIZE_t __pyx_v_n_outputs); /* proto */
static void __pyx_pf_7sklearn_4tree_5_tree_19RegressionCriterion_2__dealloc__(struct __pyx_obj_7sklearn_4tree_5_tree_RegressionCriterion *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_7sklearn_4tree_5_tree_19RegressionCriterion_4__reduce__(struct __pyx_obj_7sklearn_4tree_5_tree_RegressionCriterion *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_7sklearn_4tree_5_tree_19RegressionCriterion_6__getstate__(CYTHON_UNUSED struct __pyx_obj_7sklearn_4tree_5_tree_RegressionCriterion *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_7sklearn_4tree_5_tree_19RegressionCriterion_8__setstate__(CYTHON_UNUSED struct __pyx_obj_7sklearn_4tree_5_tree_RegressionCriterion *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v_d); /* proto */
static int __pyx_pf_7sklearn_4tree_5_tree_8Splitter___cinit__(struct __pyx_obj_7sklearn_4tree_5_tree_Splitter *__pyx_v_self, struct __pyx_obj_7sklearn_4tree_5_tree_Criterion *__pyx_v_criterion, __pyx_t_7sklearn_4tree_5_tree_SIZE_t __pyx_v_max_features, __pyx_t_7sklearn_4tree_5_tree_SIZE_t __pyx_v_min_samples_leaf, PyObject *__pyx_v_random_state); /* proto */
static void __pyx_pf_7sklearn_4tree_5_tree_8Splitter_2__dealloc__(struct __pyx_obj_7sklearn_4tree_5_tree_Splitter *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_7sklearn_4tree_5_tree_8Splitter_4__getstate__(CYTHON_UNUSED struct __pyx_obj_7sklearn_4tree_5_tree_Splitter *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_7sklearn_4tree_5_tree_8Splitter_6__setstate__(CYTHON_UNUSED struct __pyx_obj_7sklearn_4tree_5_tree_Splitter *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v_d); /* proto */
static PyObject *__pyx_pf_7sklearn_4tree_5_tree_8Splitter_9criterion___get__(struct __pyx_obj_7sklearn_4tree_5_tree_Splitter *__pyx_v_self); /* proto */
static int __pyx_pf_7sklearn_4tree_5_tree_8Splitter_9criterion_2__set__(struct __pyx_obj_7sklearn_4tree_5_tree_Splitter *__pyx_v_self, PyObject *__pyx_v_value); /* proto */
static int __pyx_pf_7sklearn_4tree_5_tree_8Splitter_9criterion_4__del__(struct __pyx_obj_7sklearn_4tree_5_tree_Splitter *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_7sklearn_4tree_5_tree_8Splitter_12max_features___get__(struct __pyx_obj_7sklearn_4tree_5_tree_Splitter *__pyx_v_self); /* proto */
static int __pyx_pf_7sklearn_4tree_5_tree_8Splitter_12max_features_2__set__(struct __pyx_obj_7sklearn_4tree_5_tree_Splitter *__pyx_v_self, PyObject *__pyx_v_value); /* proto */
static PyObject *__pyx_pf_7sklearn_4tree_5_tree_8Splitter_16min_samples_leaf___get__(struct __pyx_obj_7sklearn_4tree_5_tree_Splitter *__pyx_v_self); /* proto */
static int __pyx_pf_7sklearn_4tree_5_tree_8Splitter_16min_samples_leaf_2__set__(struct __pyx_obj_7sklearn_4tree_5_tree_Splitter *__pyx_v_self, PyObject *__pyx_v_value); /* proto */
static PyObject *__pyx_pf_7sklearn_4tree_5_tree_12BestSplitter___reduce__(struct __pyx_obj_7sklearn_4tree_5_tree_BestSplitter *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_7sklearn_4tree_5_tree_14RandomSplitter___reduce__(struct __pyx_obj_7sklearn_4tree_5_tree_RandomSplitter *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_7sklearn_4tree_5_tree_4Tree_9n_classes___get__(struct __pyx_obj_7sklearn_4tree_5_tree_Tree *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_7sklearn_4tree_5_tree_4Tree_13children_left___get__(struct __pyx_obj_7sklearn_4tree_5_tree_Tree *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_7sklearn_4tree_5_tree_4Tree_14children_right___get__(struct __pyx_obj_7sklearn_4tree_5_tree_Tree *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_7sklearn_4tree_5_tree_4Tree_7feature___get__(struct __pyx_obj_7sklearn_4tree_5_tree_Tree *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_7sklearn_4tree_5_tree_4Tree_9threshold___get__(struct __pyx_obj_7sklearn_4tree_5_tree_Tree *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_7sklearn_4tree_5_tree_4Tree_5value___get__(struct __pyx_obj_7sklearn_4tree_5_tree_Tree *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_7sklearn_4tree_5_tree_4Tree_8impurity___get__(struct __pyx_obj_7sklearn_4tree_5_tree_Tree *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_7sklearn_4tree_5_tree_4Tree_14n_node_samples___get__(struct __pyx_obj_7sklearn_4tree_5_tree_Tree *__pyx_v_self); /* proto */
static int __pyx_pf_7sklearn_4tree_5_tree_4Tree___cinit__(struct __pyx_obj_7sklearn_4tree_5_tree_Tree *__pyx_v_self, int __pyx_v_n_features, PyArrayObject *__pyx_v_n_classes, int __pyx_v_n_outputs, struct __pyx_obj_7sklearn_4tree_5_tree_Splitter *__pyx_v_splitter, __pyx_t_7sklearn_4tree_5_tree_SIZE_t __pyx_v_max_depth, __pyx_t_7sklearn_4tree_5_tree_SIZE_t __pyx_v_min_samples_split, __pyx_t_7sklearn_4tree_5_tree_SIZE_t __pyx_v_min_samples_leaf, PyObject *__pyx_v_random_state); /* proto */
static void __pyx_pf_7sklearn_4tree_5_tree_4Tree_2__dealloc__(struct __pyx_obj_7sklearn_4tree_5_tree_Tree *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_7sklearn_4tree_5_tree_4Tree_4__reduce__(struct __pyx_obj_7sklearn_4tree_5_tree_Tree *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_7sklearn_4tree_5_tree_4Tree_6__getstate__(struct __pyx_obj_7sklearn_4tree_5_tree_Tree *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_7sklearn_4tree_5_tree_4Tree_8__setstate__(struct __pyx_obj_7sklearn_4tree_5_tree_Tree *__pyx_v_self, PyObject *__pyx_v_d); /* proto */
static PyObject *__pyx_pf_7sklearn_4tree_5_tree_4Tree_10build(struct __pyx_obj_7sklearn_4tree_5_tree_Tree *__pyx_v_self, PyArrayObject *__pyx_v_X, PyArrayObject *__pyx_v_y, PyArrayObject *__pyx_v_sample_weight); /* proto */
static PyObject *__pyx_pf_7sklearn_4tree_5_tree_4Tree_12predict(struct __pyx_obj_7sklearn_4tree_5_tree_Tree *__pyx_v_self, PyArrayObject *__pyx_v_X); /* proto */
static PyObject *__pyx_pf_7sklearn_4tree_5_tree_4Tree_14apply(struct __pyx_obj_7sklearn_4tree_5_tree_Tree *__pyx_v_self, PyArrayObject *__pyx_v_X); /* proto */
static PyObject *__pyx_pf_7sklearn_4tree_5_tree_4Tree_16compute_feature_importances(struct __pyx_obj_7sklearn_4tree_5_tree_Tree *__pyx_v_self, PyObject *__pyx_v_normalize); /* proto */
static PyObject *__pyx_pf_7sklearn_4tree_5_tree_4Tree_10n_features___get__(struct __pyx_obj_7sklearn_4tree_5_tree_Tree *__pyx_v_self); /* proto */
static int __pyx_pf_7sklearn_4tree_5_tree_4Tree_10n_features_2__set__(struct __pyx_obj_7sklearn_4tree_5_tree_Tree *__pyx_v_self, PyObject *__pyx_v_value); /* proto */
static PyObject *__pyx_pf_7sklearn_4tree_5_tree_4Tree_9n_outputs___get__(struct __pyx_obj_7sklearn_4tree_5_tree_Tree *__pyx_v_self); /* proto */
static int __pyx_pf_7sklearn_4tree_5_tree_4Tree_9n_outputs_2__set__(struct __pyx_obj_7sklearn_4tree_5_tree_Tree *__pyx_v_self, PyObject *__pyx_v_value); /* proto */
static PyObject *__pyx_pf_7sklearn_4tree_5_tree_4Tree_13max_n_classes___get__(struct __pyx_obj_7sklearn_4tree_5_tree_Tree *__pyx_v_self); /* proto */
static int __pyx_pf_7sklearn_4tree_5_tree_4Tree_13max_n_classes_2__set__(struct __pyx_obj_7sklearn_4tree_5_tree_Tree *__pyx_v_self, PyObject *__pyx_v_value); /* proto */
static PyObject *__pyx_pf_7sklearn_4tree_5_tree_4Tree_12value_stride___get__(struct __pyx_obj_7sklearn_4tree_5_tree_Tree *__pyx_v_self); /* proto */
static int __pyx_pf_7sklearn_4tree_5_tree_4Tree_12value_stride_2__set__(struct __pyx_obj_7sklearn_4tree_5_tree_Tree *__pyx_v_self, PyObject *__pyx_v_value); /* proto */
static PyObject *__pyx_pf_7sklearn_4tree_5_tree_4Tree_8splitter___get__(struct __pyx_obj_7sklearn_4tree_5_tree_Tree *__pyx_v_self); /* proto */
static int __pyx_pf_7sklearn_4tree_5_tree_4Tree_8splitter_2__set__(struct __pyx_obj_7sklearn_4tree_5_tree_Tree *__pyx_v_self, PyObject *__pyx_v_value); /* proto */
static int __pyx_pf_7sklearn_4tree_5_tree_4Tree_8splitter_4__del__(struct __pyx_obj_7sklearn_4tree_5_tree_Tree *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_7sklearn_4tree_5_tree_4Tree_9max_depth___get__(struct __pyx_obj_7sklearn_4tree_5_tree_Tree *__pyx_v_self); /* proto */
static int __pyx_pf_7sklearn_4tree_5_tree_4Tree_9max_depth_2__set__(struct __pyx_obj_7sklearn_4tree_5_tree_Tree *__pyx_v_self, PyObject *__pyx_v_value); /* proto */
static PyObject *__pyx_pf_7sklearn_4tree_5_tree_4Tree_17min_samples_split___get__(struct __pyx_obj_7sklearn_4tree_5_tree_Tree *__pyx_v_self); /* proto */
static int __pyx_pf_7sklearn_4tree_5_tree_4Tree_17min_samples_split_2__set__(struct __pyx_obj_7sklearn_4tree_5_tree_Tree *__pyx_v_self, PyObject *__pyx_v_value); /* proto */
static PyObject *__pyx_pf_7sklearn_4tree_5_tree_4Tree_16min_samples_leaf___get__(struct __pyx_obj_7sklearn_4tree_5_tree_Tree *__pyx_v_self); /* proto */
static int __pyx_pf_7sklearn_4tree_5_tree_4Tree_16min_samples_leaf_2__set__(struct __pyx_obj_7sklearn_4tree_5_tree_Tree *__pyx_v_self, PyObject *__pyx_v_value); /* proto */
static PyObject *__pyx_pf_7sklearn_4tree_5_tree_4Tree_12random_state___get__(struct __pyx_obj_7sklearn_4tree_5_tree_Tree *__pyx_v_self); /* proto */
static int __pyx_pf_7sklearn_4tree_5_tree_4Tree_12random_state_2__set__(struct __pyx_obj_7sklearn_4tree_5_tree_Tree *__pyx_v_self, PyObject *__pyx_v_value); /* proto */
static int __pyx_pf_7sklearn_4tree_5_tree_4Tree_12random_state_4__del__(struct __pyx_obj_7sklearn_4tree_5_tree_Tree *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_7sklearn_4tree_5_tree_4Tree_10node_count___get__(struct __pyx_obj_7sklearn_4tree_5_tree_Tree *__pyx_v_self); /* proto */
static int __pyx_pf_7sklearn_4tree_5_tree_4Tree_10node_count_2__set__(struct __pyx_obj_7sklearn_4tree_5_tree_Tree *__pyx_v_self, PyObject *__pyx_v_value); /* proto */
static PyObject *__pyx_pf_7sklearn_4tree_5_tree_4Tree_8capacity___get__(struct __pyx_obj_7sklearn_4tree_5_tree_Tree *__pyx_v_self); /* proto */
static int __pyx_pf_7sklearn_4tree_5_tree_4Tree_8capacity_2__set__(struct __pyx_obj_7sklearn_4tree_5_tree_Tree *__pyx_v_self, PyObject *__pyx_v_value); /* proto */
static int __pyx_pf_5numpy_7ndarray___getbuffer__(PyArrayObject *__pyx_v_self, Py_buffer *__pyx_v_info, int __pyx_v_flags); /* proto */
static void __pyx_pf_5numpy_7ndarray_2__releasebuffer__(PyArrayObject *__pyx_v_self, Py_buffer *__pyx_v_info); /* proto */
static PyObject *__pyx_tp_new_7sklearn_4tree_5_tree_Tree(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_7sklearn_4tree_5_tree_Splitter(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_7sklearn_4tree_5_tree_RandomSplitter(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_7sklearn_4tree_5_tree_Criterion(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_7sklearn_4tree_5_tree_ClassificationCriterion(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_7sklearn_4tree_5_tree_Gini(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_7sklearn_4tree_5_tree_Entropy(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_7sklearn_4tree_5_tree_RegressionCriterion(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_7sklearn_4tree_5_tree_MSE(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_7sklearn_4tree_5_tree_BestSplitter(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static char __pyx_k_2[] = "compute_feature_importances";
static char __pyx_k_3[] = "ndarray is not C contiguous";
static char __pyx_k_5[] = "ndarray is not Fortran contiguous";
static char __pyx_k_7[] = "Non-native byte order not supported";
static char __pyx_k_9[] = "unknown dtype code in numpy.pxd (%d)";
static char __pyx_k_10[] = "Format string allocated too short, see comment in numpy.pxd";
static char __pyx_k_13[] = "Format string allocated too short.";
static char __pyx_k__B[] = "B";
static char __pyx_k__C[] = "C";
static char __pyx_k__H[] = "H";
static char __pyx_k__I[] = "I";
static char __pyx_k__L[] = "L";
static char __pyx_k__O[] = "O";
static char __pyx_k__Q[] = "Q";
static char __pyx_k__X[] = "X";
static char __pyx_k__b[] = "b";
static char __pyx_k__d[] = "d";
static char __pyx_k__f[] = "f";
static char __pyx_k__g[] = "g";
static char __pyx_k__h[] = "h";
static char __pyx_k__i[] = "i";
static char __pyx_k__l[] = "l";
static char __pyx_k__q[] = "q";
static char __pyx_k__y[] = "y";
static char __pyx_k__Zd[] = "Zd";
static char __pyx_k__Zf[] = "Zf";
static char __pyx_k__Zg[] = "Zg";
static char __pyx_k__np[] = "np";
static char __pyx_k__inf[] = "inf";
static char __pyx_k__max[] = "max";
static char __pyx_k__sum[] = "sum";
static char __pyx_k__DTYPE[] = "DTYPE";
static char __pyx_k__apply[] = "apply";
static char __pyx_k__build[] = "build";
static char __pyx_k__dtype[] = "dtype";
static char __pyx_k__flags[] = "flags";
static char __pyx_k__int32[] = "int32";
static char __pyx_k__numpy[] = "numpy";
static char __pyx_k__order[] = "order";
static char __pyx_k__range[] = "range";
static char __pyx_k__value[] = "value";
static char __pyx_k__zeros[] = "zeros";
static char __pyx_k__DOUBLE[] = "DOUBLE";
static char __pyx_k__asarray[] = "asarray";
static char __pyx_k__feature[] = "feature";
static char __pyx_k__float32[] = "float32";
static char __pyx_k__float64[] = "float64";
static char __pyx_k__predict[] = "predict";
static char __pyx_k__randint[] = "randint";
static char __pyx_k____main__[] = "__main__";
static char __pyx_k____test__[] = "__test__";
static char __pyx_k__capacity[] = "capacity";
static char __pyx_k__impurity[] = "impurity";
static char __pyx_k__itemsize[] = "itemsize";
static char __pyx_k__splitter[] = "splitter";
static char __pyx_k__TREE_LEAF[] = "TREE_LEAF";
static char __pyx_k__criterion[] = "criterion";
static char __pyx_k__max_depth[] = "max_depth";
static char __pyx_k__n_classes[] = "n_classes";
static char __pyx_k__n_outputs[] = "n_outputs";
static char __pyx_k__normalize[] = "normalize";
static char __pyx_k__threshold[] = "threshold";
static char __pyx_k__ValueError[] = "ValueError";
static char __pyx_k____import__[] = "__import__";
static char __pyx_k__contiguous[] = "contiguous";
static char __pyx_k__n_features[] = "n_features";
static char __pyx_k__node_count[] = "node_count";
static char __pyx_k__MemoryError[] = "MemoryError";
static char __pyx_k__RuntimeError[] = "RuntimeError";
static char __pyx_k____getstate__[] = "__getstate__";
static char __pyx_k__max_features[] = "max_features";
static char __pyx_k__random_state[] = "random_state";
static char __pyx_k__children_left[] = "children_left";
static char __pyx_k__sample_weight[] = "sample_weight";
static char __pyx_k__TREE_UNDEFINED[] = "TREE_UNDEFINED";
static char __pyx_k____pyx_vtable__[] = "__pyx_vtable__";
static char __pyx_k__children_right[] = "children_right";
static char __pyx_k__n_node_samples[] = "n_node_samples";
static char __pyx_k____pyx_getbuffer[] = "__pyx_getbuffer";
static char __pyx_k__min_samples_leaf[] = "min_samples_leaf";
static char __pyx_k__min_samples_split[] = "min_samples_split";
static char __pyx_k____pyx_releasebuffer[] = "__pyx_releasebuffer";
static PyObject *__pyx_kp_u_10;
static PyObject *__pyx_kp_u_13;
static PyObject *__pyx_n_s_2;
static PyObject *__pyx_kp_u_3;
static PyObject *__pyx_kp_u_5;
static PyObject *__pyx_kp_u_7;
static PyObject *__pyx_kp_u_9;
static PyObject *__pyx_n_s__C;
static PyObject *__pyx_n_s__DOUBLE;
static PyObject *__pyx_n_s__DTYPE;
static PyObject *__pyx_n_s__MemoryError;
static PyObject *__pyx_n_s__RuntimeError;
static PyObject *__pyx_n_s__TREE_LEAF;
static PyObject *__pyx_n_s__TREE_UNDEFINED;
static PyObject *__pyx_n_s__ValueError;
static PyObject *__pyx_n_s__X;
static PyObject *__pyx_n_s____getstate__;
static PyObject *__pyx_n_s____import__;
static PyObject *__pyx_n_s____main__;
static PyObject *__pyx_n_s____pyx_getbuffer;
static PyObject *__pyx_n_s____pyx_releasebuffer;
static PyObject *__pyx_n_s____pyx_vtable__;
static PyObject *__pyx_n_s____test__;
static PyObject *__pyx_n_s__apply;
static PyObject *__pyx_n_s__asarray;
static PyObject *__pyx_n_s__build;
static PyObject *__pyx_n_s__capacity;
static PyObject *__pyx_n_s__children_left;
static PyObject *__pyx_n_s__children_right;
static PyObject *__pyx_n_s__contiguous;
static PyObject *__pyx_n_s__criterion;
static PyObject *__pyx_n_s__dtype;
static PyObject *__pyx_n_s__feature;
static PyObject *__pyx_n_s__flags;
static PyObject *__pyx_n_s__float32;
static PyObject *__pyx_n_s__float64;
static PyObject *__pyx_n_s__impurity;
static PyObject *__pyx_n_s__inf;
static PyObject *__pyx_n_s__int32;
static PyObject *__pyx_n_s__itemsize;
static PyObject *__pyx_n_s__max;
static PyObject *__pyx_n_s__max_depth;
static PyObject *__pyx_n_s__max_features;
static PyObject *__pyx_n_s__min_samples_leaf;
static PyObject *__pyx_n_s__min_samples_split;
static PyObject *__pyx_n_s__n_classes;
static PyObject *__pyx_n_s__n_features;
static PyObject *__pyx_n_s__n_node_samples;
static PyObject *__pyx_n_s__n_outputs;
static PyObject *__pyx_n_s__node_count;
static PyObject *__pyx_n_s__normalize;
static PyObject *__pyx_n_s__np;
static PyObject *__pyx_n_s__numpy;
static PyObject *__pyx_n_s__order;
static PyObject *__pyx_n_s__predict;
static PyObject *__pyx_n_s__randint;
static PyObject *__pyx_n_s__random_state;
static PyObject *__pyx_n_s__range;
static PyObject *__pyx_n_s__sample_weight;
static PyObject *__pyx_n_s__splitter;
static PyObject *__pyx_n_s__sum;
static PyObject *__pyx_n_s__threshold;
static PyObject *__pyx_n_s__value;
static PyObject *__pyx_n_s__y;
static PyObject *__pyx_n_s__zeros;
static PyObject *__pyx_int_0;
static PyObject *__pyx_int_neg_1;
static PyObject *__pyx_int_neg_2;
static PyObject *__pyx_int_15;
static PyObject *__pyx_k_1;
static PyObject *__pyx_k_tuple_4;
static PyObject *__pyx_k_tuple_6;
static PyObject *__pyx_k_tuple_8;
static PyObject *__pyx_k_tuple_11;
static PyObject *__pyx_k_tuple_12;
static PyObject *__pyx_k_tuple_14;

/* "sklearn/tree/_tree.pyx":55
 *     """Interface for impurity criteria."""
 * 
 *     cdef void init(self, DOUBLE_t* y,             # <<<<<<<<<<<<<<
 *                          SIZE_t y_stride,
 *                          DOUBLE_t* sample_weight,
 */

static void __pyx_f_7sklearn_4tree_5_tree_9Criterion_init(CYTHON_UNUSED struct __pyx_obj_7sklearn_4tree_5_tree_Criterion *__pyx_v_self, CYTHON_UNUSED __pyx_t_7sklearn_4tree_5_tree_DOUBLE_t *__pyx_v_y, CYTHON_UNUSED __pyx_t_7sklearn_4tree_5_tree_SIZE_t __pyx_v_y_stride, CYTHON_UNUSED __pyx_t_7sklearn_4tree_5_tree_DOUBLE_t *__pyx_v_sample_weight, CYTHON_UNUSED __pyx_t_7sklearn_4tree_5_tree_SIZE_t *__pyx_v_samples, CYTHON_UNUSED __pyx_t_7sklearn_4tree_5_tree_SIZE_t __pyx_v_start, CYTHON_UNUSED __pyx_t_7sklearn_4tree_5_tree_SIZE_t __pyx_v_end) {

}

/* "sklearn/tree/_tree.pyx":65
 *         pass
 * 
 *     cdef void reset(self) nogil:             # <<<<<<<<<<<<<<
 *         """Reset the criterion at pos=start."""
 *         pass
 */

static void __pyx_f_7sklearn_4tree_5_tree_9Criterion_reset(CYTHON_UNUSED struct __pyx_obj_7sklearn_4tree_5_tree_Criterion *__pyx_v_self) {

}

/* "sklearn/tree/_tree.pyx":69
 *         pass
 * 
 *     cdef void update(self, SIZE_t new_pos) nogil:             # <<<<<<<<<<<<<<
 *         """Update the collected statistics by moving samples[pos:new_pos] from
 *            the right child to the left child."""
 */

static void __pyx_f_7sklearn_4tree_5_tree_9Criterion_update(CYTHON_UNUSED struct __pyx_obj_7sklearn_4tree_5_tree_Criterion *__pyx_v_self, CYTHON_UNUSED __pyx_t_7sklearn_4tree_5_tree_SIZE_t __pyx_v_new_pos) {

}

/* "sklearn/tree/_tree.pyx":74
 *         pass
 * 
 *     cdef double node_impurity(self) nogil:             # <<<<<<<<<<<<<<
 *         """Evaluate the impurity of the current node, i.e. the impurity of
 *            samples[start:end]."""
 */

static double __pyx_f_7sklearn_4tree_5_tree_9Criterion_node_impurity(CYTHON_UNUSED struct __pyx_obj_7sklearn_4tree_5_tree_Criterion *__pyx_v_self) {
  double __pyx_r;

  __pyx_r = 0;
  return __pyx_r;
}

/* "sklearn/tree/_tree.pyx":79
 *         pass
 * 
 *     cdef double children_impurity(self) nogil:             # <<<<<<<<<<<<<<
 *         """Evaluate the impurity in children nodes, i.e. the impurity of
 *            samples[start:pos] + the impurity of samples[pos:end]."""
 */

static double __pyx_f_7sklearn_4tree_5_tree_9Criterion_children_impurity(CYTHON_UNUSED struct __pyx_obj_7sklearn_4tree_5_tree_Criterion *__pyx_v_self) {
  double __pyx_r;

  __pyx_r = 0;
  return __pyx_r;
}

/* "sklearn/tree/_tree.pyx":84
 *         pass
 * 
 *     cdef void node_value(self, double* dest) nogil:             # <<<<<<<<<<<<<<
 *         """Compute the node value of samples[start:end] into dest."""
 *         pass
 */

static void __pyx_f_7sklearn_4tree_5_tree_9Criterion_node_value(CYTHON_UNUSED struct __pyx_obj_7sklearn_4tree_5_tree_Criterion *__pyx_v_self, CYTHON_UNUSED double *__pyx_v_dest) {

}

/* Python wrapper */
static int __pyx_pw_7sklearn_4tree_5_tree_23ClassificationCriterion_1__cinit__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static int __pyx_pw_7sklearn_4tree_5_tree_23ClassificationCriterion_1__cinit__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  __pyx_t_7sklearn_4tree_5_tree_SIZE_t __pyx_v_n_outputs;
  PyArrayObject *__pyx_v_n_classes = 0;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__cinit__ (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s__n_outputs,&__pyx_n_s__n_classes,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s__n_outputs)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s__n_classes)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__cinit__", 1, 2, 2, 1); {__pyx_filename = __pyx_f[0]; __pyx_lineno = 97; __pyx_clineno = __LINE__; goto __pyx_L3_error;}
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "__cinit__") < 0)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 97; __pyx_clineno = __LINE__; goto __pyx_L3_error;}
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_n_outputs = __Pyx_PyInt_from_py_Py_intptr_t(values[0]); if (unlikely((__pyx_v_n_outputs == (npy_intp)-1) && PyErr_Occurred())) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 97; __pyx_clineno = __LINE__; goto __pyx_L3_error;}
    __pyx_v_n_classes = ((PyArrayObject *)values[1]);
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__cinit__", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); {__pyx_filename = __pyx_f[0]; __pyx_lineno = 97; __pyx_clineno = __LINE__; goto __pyx_L3_error;}
  __pyx_L3_error:;
  __Pyx_AddTraceback("sklearn.tree._tree.ClassificationCriterion.__cinit__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return -1;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_n_classes), __pyx_ptype_5numpy_ndarray, 1, "n_classes", 0))) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 97; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __pyx_r = __pyx_pf_7sklearn_4tree_5_tree_23ClassificationCriterion___cinit__(((struct __pyx_obj_7sklearn_4tree_5_tree_ClassificationCriterion *)__pyx_v_self), __pyx_v_n_outputs, __pyx_v_n_classes);
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "sklearn/tree/_tree.pyx":97
 *     cdef double* label_count_total
 * 
 *     def __cinit__(self, SIZE_t n_outputs, np.ndarray[SIZE_t, ndim=1] n_classes):             # <<<<<<<<<<<<<<
 *         # Default values
 *         self.y = NULL
 */

static int __pyx_pf_7sklearn_4tree_5_tree_23ClassificationCriterion___cinit__(struct __pyx_obj_7sklearn_4tree_5_tree_ClassificationCriterion *__pyx_v_self, __pyx_t_7sklearn_4tree_5_tree_SIZE_t __pyx_v_n_outputs, PyArrayObject *__pyx_v_n_classes) {
  __pyx_t_7sklearn_4tree_5_tree_SIZE_t __pyx_v_k;
  __pyx_t_7sklearn_4tree_5_tree_SIZE_t __pyx_v_label_count_stride;
  __Pyx_LocalBuf_ND __pyx_pybuffernd_n_classes;
  __Pyx_Buffer __pyx_pybuffer_n_classes;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  __pyx_t_7sklearn_4tree_5_tree_SIZE_t __pyx_t_2;
  __pyx_t_7sklearn_4tree_5_tree_SIZE_t __pyx_t_3;
  __pyx_t_7sklearn_4tree_5_tree_SIZE_t __pyx_t_4;
  __pyx_t_7sklearn_4tree_5_tree_SIZE_t __pyx_t_5;
  int __pyx_t_6;
  int __pyx_t_7;
  int __pyx_t_8;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__cinit__", 0);
  __pyx_pybuffer_n_classes.pybuffer.buf = NULL;
  __pyx_pybuffer_n_classes.refcount = 0;
  __pyx_pybuffernd_n_classes.data = NULL;
  __pyx_pybuffernd_n_classes.rcbuffer = &__pyx_pybuffer_n_classes;
  {
    __Pyx_BufFmt_StackElem __pyx_stack[1];
    if (unlikely(__Pyx_GetBufferAndValidate(&__pyx_pybuffernd_n_classes.rcbuffer->pybuffer, (PyObject*)__pyx_v_n_classes, &__Pyx_TypeInfo_nn___pyx_t_7sklearn_4tree_5_tree_SIZE_t, PyBUF_FORMAT| PyBUF_STRIDES, 1, 0, __pyx_stack) == -1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 97; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  }
  __pyx_pybuffernd_n_classes.diminfo[0].strides = __pyx_pybuffernd_n_classes.rcbuffer->pybuffer.strides[0]; __pyx_pybuffernd_n_classes.diminfo[0].shape = __pyx_pybuffernd_n_classes.rcbuffer->pybuffer.shape[0];

  /* "sklearn/tree/_tree.pyx":99
 *     def __cinit__(self, SIZE_t n_outputs, np.ndarray[SIZE_t, ndim=1] n_classes):
 *         # Default values
 *         self.y = NULL             # <<<<<<<<<<<<<<
 *         self.y_stride = 0
 *         self.sample_weight = NULL
 */
  __pyx_v_self->__pyx_base.y = NULL;

  /* "sklearn/tree/_tree.pyx":100
 *         # Default values
 *         self.y = NULL
 *         self.y_stride = 0             # <<<<<<<<<<<<<<
 *         self.sample_weight = NULL
 * 
 */
  __pyx_v_self->__pyx_base.y_stride = 0;

  /* "sklearn/tree/_tree.pyx":101
 *         self.y = NULL
 *         self.y_stride = 0
 *         self.sample_weight = NULL             # <<<<<<<<<<<<<<
 * 
 *         self.samples = NULL
 */
  __pyx_v_self->__pyx_base.sample_weight = NULL;

  /* "sklearn/tree/_tree.pyx":103
 *         self.sample_weight = NULL
 * 
 *         self.samples = NULL             # <<<<<<<<<<<<<<
 *         self.start = 0
 *         self.pos = 0
 */
  __pyx_v_self->__pyx_base.samples = NULL;

  /* "sklearn/tree/_tree.pyx":104
 * 
 *         self.samples = NULL
 *         self.start = 0             # <<<<<<<<<<<<<<
 *         self.pos = 0
 *         self.end = 0
 */
  __pyx_v_self->__pyx_base.start = 0;

  /* "sklearn/tree/_tree.pyx":105
 *         self.samples = NULL
 *         self.start = 0
 *         self.pos = 0             # <<<<<<<<<<<<<<
 *         self.end = 0
 * 
 */
  __pyx_v_self->__pyx_base.pos = 0;

  /* "sklearn/tree/_tree.pyx":106
 *         self.start = 0
 *         self.pos = 0
 *         self.end = 0             # <<<<<<<<<<<<<<
 * 
 *         self.n_outputs = n_outputs
 */
  __pyx_v_self->__pyx_base.end = 0;

  /* "sklearn/tree/_tree.pyx":108
 *         self.end = 0
 * 
 *         self.n_outputs = n_outputs             # <<<<<<<<<<<<<<
 *         self.n_node_samples = 0
 *         self.weighted_n_node_samples = 0.0
 */
  __pyx_v_self->__pyx_base.n_outputs = __pyx_v_n_outputs;

  /* "sklearn/tree/_tree.pyx":109
 * 
 *         self.n_outputs = n_outputs
 *         self.n_node_samples = 0             # <<<<<<<<<<<<<<
 *         self.weighted_n_node_samples = 0.0
 *         self.weighted_n_left = 0.0
 */
  __pyx_v_self->__pyx_base.n_node_samples = 0;

  /* "sklearn/tree/_tree.pyx":110
 *         self.n_outputs = n_outputs
 *         self.n_node_samples = 0
 *         self.weighted_n_node_samples = 0.0             # <<<<<<<<<<<<<<
 *         self.weighted_n_left = 0.0
 *         self.weighted_n_right = 0.0
 */
  __pyx_v_self->__pyx_base.weighted_n_node_samples = 0.0;

  /* "sklearn/tree/_tree.pyx":111
 *         self.n_node_samples = 0
 *         self.weighted_n_node_samples = 0.0
 *         self.weighted_n_left = 0.0             # <<<<<<<<<<<<<<
 *         self.weighted_n_right = 0.0
 * 
 */
  __pyx_v_self->__pyx_base.weighted_n_left = 0.0;

  /* "sklearn/tree/_tree.pyx":112
 *         self.weighted_n_node_samples = 0.0
 *         self.weighted_n_left = 0.0
 *         self.weighted_n_right = 0.0             # <<<<<<<<<<<<<<
 * 
 *         # Count labels for each output
 */
  __pyx_v_self->__pyx_base.weighted_n_right = 0.0;

  /* "sklearn/tree/_tree.pyx":115
 * 
 *         # Count labels for each output
 *         self.n_classes = <SIZE_t*> malloc(n_outputs * sizeof(SIZE_t))             # <<<<<<<<<<<<<<
 *         if self.n_classes == NULL:
 *             raise MemoryError()
 */
  __pyx_v_self->n_classes = ((__pyx_t_7sklearn_4tree_5_tree_SIZE_t *)malloc((__pyx_v_n_outputs * (sizeof(__pyx_t_7sklearn_4tree_5_tree_SIZE_t)))));

  /* "sklearn/tree/_tree.pyx":116
 *         # Count labels for each output
 *         self.n_classes = <SIZE_t*> malloc(n_outputs * sizeof(SIZE_t))
 *         if self.n_classes == NULL:             # <<<<<<<<<<<<<<
 *             raise MemoryError()
 * 
 */
  __pyx_t_1 = ((__pyx_v_self->n_classes == NULL) != 0);
  if (__pyx_t_1) {

    /* "sklearn/tree/_tree.pyx":117
 *         self.n_classes = <SIZE_t*> malloc(n_outputs * sizeof(SIZE_t))
 *         if self.n_classes == NULL:
 *             raise MemoryError()             # <<<<<<<<<<<<<<
 * 
 *         cdef SIZE_t k = 0
 */
    PyErr_NoMemory(); {__pyx_filename = __pyx_f[0]; __pyx_lineno = 117; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    goto __pyx_L3;
  }
  __pyx_L3:;

  /* "sklearn/tree/_tree.pyx":119
 *             raise MemoryError()
 * 
 *         cdef SIZE_t k = 0             # <<<<<<<<<<<<<<
 *         cdef SIZE_t label_count_stride = 0
 * 
 */
  __pyx_v_k = 0;

  /* "sklearn/tree/_tree.pyx":120
 * 
 *         cdef SIZE_t k = 0
 *         cdef SIZE_t label_count_stride = 0             # <<<<<<<<<<<<<<
 * 
 *         for k from 0 <= k < n_outputs:
 */
  __pyx_v_label_count_stride = 0;

  /* "sklearn/tree/_tree.pyx":122
 *         cdef SIZE_t label_count_stride = 0
 * 
 *         for k from 0 <= k < n_outputs:             # <<<<<<<<<<<<<<
 *             self.n_classes[k] = n_classes[k]
 * 
 */
  __pyx_t_2 = __pyx_v_n_outputs;
  for (__pyx_v_k = 0; __pyx_v_k < __pyx_t_2; __pyx_v_k++) {

    /* "sklearn/tree/_tree.pyx":123
 * 
 *         for k from 0 <= k < n_outputs:
 *             self.n_classes[k] = n_classes[k]             # <<<<<<<<<<<<<<
 * 
 *             if n_classes[k] > label_count_stride:
 */
    __pyx_t_3 = __pyx_v_k;
    (__pyx_v_self->n_classes[__pyx_v_k]) = (*__Pyx_BufPtrStrided1d(__pyx_t_7sklearn_4tree_5_tree_SIZE_t *, __pyx_pybuffernd_n_classes.rcbuffer->pybuffer.buf, __pyx_t_3, __pyx_pybuffernd_n_classes.diminfo[0].strides));

    /* "sklearn/tree/_tree.pyx":125
 *             self.n_classes[k] = n_classes[k]
 * 
 *             if n_classes[k] > label_count_stride:             # <<<<<<<<<<<<<<
 *                 label_count_stride = n_classes[k]
 * 
 */
    __pyx_t_4 = __pyx_v_k;
    __pyx_t_1 = (((*__Pyx_BufPtrStrided1d(__pyx_t_7sklearn_4tree_5_tree_SIZE_t *, __pyx_pybuffernd_n_classes.rcbuffer->pybuffer.buf, __pyx_t_4, __pyx_pybuffernd_n_classes.diminfo[0].strides)) > __pyx_v_label_count_stride) != 0);
    if (__pyx_t_1) {

      /* "sklearn/tree/_tree.pyx":126
 * 
 *             if n_classes[k] > label_count_stride:
 *                 label_count_stride = n_classes[k]             # <<<<<<<<<<<<<<
 * 
 *         self.label_count_stride = label_count_stride
 */
      __pyx_t_5 = __pyx_v_k;
      __pyx_v_label_count_stride = (*__Pyx_BufPtrStrided1d(__pyx_t_7sklearn_4tree_5_tree_SIZE_t *, __pyx_pybuffernd_n_classes.rcbuffer->pybuffer.buf, __pyx_t_5, __pyx_pybuffernd_n_classes.diminfo[0].strides));
      goto __pyx_L6;
    }
    __pyx_L6:;
  }

  /* "sklearn/tree/_tree.pyx":128
 *                 label_count_stride = n_classes[k]
 * 
 *         self.label_count_stride = label_count_stride             # <<<<<<<<<<<<<<
 * 
 *         # Allocate counters
 */
  __pyx_v_self->label_count_stride = __pyx_v_label_count_stride;

  /* "sklearn/tree/_tree.pyx":131
 * 
 *         # Allocate counters
 *         self.label_count_left = <double*> calloc(n_outputs * label_count_stride, sizeof(double))             # <<<<<<<<<<<<<<
 *         self.label_count_right = <double*> calloc(n_outputs * label_count_stride, sizeof(double))
 *         self.label_count_total = <double*> calloc(n_outputs * label_count_stride, sizeof(double))
 */
  __pyx_v_self->label_count_left = ((double *)calloc((__pyx_v_n_outputs * __pyx_v_label_count_stride), (sizeof(double))));

  /* "sklearn/tree/_tree.pyx":132
 *         # Allocate counters
 *         self.label_count_left = <double*> calloc(n_outputs * label_count_stride, sizeof(double))
 *         self.label_count_right = <double*> calloc(n_outputs * label_count_stride, sizeof(double))             # <<<<<<<<<<<<<<
 *         self.label_count_total = <double*> calloc(n_outputs * label_count_stride, sizeof(double))
 * 
 */
  __pyx_v_self->label_count_right = ((double *)calloc((__pyx_v_n_outputs * __pyx_v_label_count_stride), (sizeof(double))));

  /* "sklearn/tree/_tree.pyx":133
 *         self.label_count_left = <double*> calloc(n_outputs * label_count_stride, sizeof(double))
 *         self.label_count_right = <double*> calloc(n_outputs * label_count_stride, sizeof(double))
 *         self.label_count_total = <double*> calloc(n_outputs * label_count_stride, sizeof(double))             # <<<<<<<<<<<<<<
 * 
 *         # Check for allocation errors
 */
  __pyx_v_self->label_count_total = ((double *)calloc((__pyx_v_n_outputs * __pyx_v_label_count_stride), (sizeof(double))));

  /* "sklearn/tree/_tree.pyx":136
 * 
 *         # Check for allocation errors
 *         if ((self.label_count_left == NULL) or             # <<<<<<<<<<<<<<
 *             (self.label_count_right == NULL) or
 *             (self.label_count_total == NULL)):
 */
  __pyx_t_1 = ((__pyx_v_self->label_count_left == NULL) != 0);
  if (!__pyx_t_1) {

    /* "sklearn/tree/_tree.pyx":137
 *         # Check for allocation errors
 *         if ((self.label_count_left == NULL) or
 *             (self.label_count_right == NULL) or             # <<<<<<<<<<<<<<
 *             (self.label_count_total == NULL)):
 *             free(self.n_classes)
 */
    __pyx_t_6 = ((__pyx_v_self->label_count_right == NULL) != 0);
    if (!__pyx_t_6) {

      /* "sklearn/tree/_tree.pyx":138
 *         if ((self.label_count_left == NULL) or
 *             (self.label_count_right == NULL) or
 *             (self.label_count_total == NULL)):             # <<<<<<<<<<<<<<
 *             free(self.n_classes)
 *             free(self.label_count_left)
 */
      __pyx_t_7 = ((__pyx_v_self->label_count_total == NULL) != 0);
      __pyx_t_8 = __pyx_t_7;
    } else {
      __pyx_t_8 = __pyx_t_6;
    }
    __pyx_t_6 = __pyx_t_8;
  } else {
    __pyx_t_6 = __pyx_t_1;
  }
  if (__pyx_t_6) {

    /* "sklearn/tree/_tree.pyx":139
 *             (self.label_count_right == NULL) or
 *             (self.label_count_total == NULL)):
 *             free(self.n_classes)             # <<<<<<<<<<<<<<
 *             free(self.label_count_left)
 *             free(self.label_count_right)
 */
    free(__pyx_v_self->n_classes);

    /* "sklearn/tree/_tree.pyx":140
 *             (self.label_count_total == NULL)):
 *             free(self.n_classes)
 *             free(self.label_count_left)             # <<<<<<<<<<<<<<
 *             free(self.label_count_right)
 *             free(self.label_count_total)
 */
    free(__pyx_v_self->label_count_left);

    /* "sklearn/tree/_tree.pyx":141
 *             free(self.n_classes)
 *             free(self.label_count_left)
 *             free(self.label_count_right)             # <<<<<<<<<<<<<<
 *             free(self.label_count_total)
 *             raise MemoryError()
 */
    free(__pyx_v_self->label_count_right);

    /* "sklearn/tree/_tree.pyx":142
 *             free(self.label_count_left)
 *             free(self.label_count_right)
 *             free(self.label_count_total)             # <<<<<<<<<<<<<<
 *             raise MemoryError()
 * 
 */
    free(__pyx_v_self->label_count_total);

    /* "sklearn/tree/_tree.pyx":143
 *             free(self.label_count_right)
 *             free(self.label_count_total)
 *             raise MemoryError()             # <<<<<<<<<<<<<<
 * 
 *     def __dealloc__(self):
 */
    PyErr_NoMemory(); {__pyx_filename = __pyx_f[0]; __pyx_lineno = 143; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    goto __pyx_L7;
  }
  __pyx_L7:;

  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  { PyObject *__pyx_type, *__pyx_value, *__pyx_tb;
    __Pyx_ErrFetch(&__pyx_type, &__pyx_value, &__pyx_tb);
    __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_n_classes.rcbuffer->pybuffer);
  __Pyx_ErrRestore(__pyx_type, __pyx_value, __pyx_tb);}
  __Pyx_AddTraceback("sklearn.tree._tree.ClassificationCriterion.__cinit__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  goto __pyx_L2;
  __pyx_L0:;
  __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_n_classes.rcbuffer->pybuffer);
  __pyx_L2:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static void __pyx_pw_7sklearn_4tree_5_tree_23ClassificationCriterion_3__dealloc__(PyObject *__pyx_v_self); /*proto*/
static void __pyx_pw_7sklearn_4tree_5_tree_23ClassificationCriterion_3__dealloc__(PyObject *__pyx_v_self) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__dealloc__ (wrapper)", 0);
  __pyx_pf_7sklearn_4tree_5_tree_23ClassificationCriterion_2__dealloc__(((struct __pyx_obj_7sklearn_4tree_5_tree_ClassificationCriterion *)__pyx_v_self));
  __Pyx_RefNannyFinishContext();
}

/* "sklearn/tree/_tree.pyx":145
 *             raise MemoryError()
 * 
 *     def __dealloc__(self):             # <<<<<<<<<<<<<<
 *         """Destructor."""
 *         free(self.n_classes)
 */

static void __pyx_pf_7sklearn_4tree_5_tree_23ClassificationCriterion_2__dealloc__(struct __pyx_obj_7sklearn_4tree_5_tree_ClassificationCriterion *__pyx_v_self) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__dealloc__", 0);

  /* "sklearn/tree/_tree.pyx":147
 *     def __dealloc__(self):
 *         """Destructor."""
 *         free(self.n_classes)             # <<<<<<<<<<<<<<
 *         free(self.label_count_left)
 *         free(self.label_count_right)
 */
  free(__pyx_v_self->n_classes);

  /* "sklearn/tree/_tree.pyx":148
 *         """Destructor."""
 *         free(self.n_classes)
 *         free(self.label_count_left)             # <<<<<<<<<<<<<<
 *         free(self.label_count_right)
 *         free(self.label_count_total)
 */
  free(__pyx_v_self->label_count_left);

  /* "sklearn/tree/_tree.pyx":149
 *         free(self.n_classes)
 *         free(self.label_count_left)
 *         free(self.label_count_right)             # <<<<<<<<<<<<<<
 *         free(self.label_count_total)
 * 
 */
  free(__pyx_v_self->label_count_right);

  /* "sklearn/tree/_tree.pyx":150
 *         free(self.label_count_left)
 *         free(self.label_count_right)
 *         free(self.label_count_total)             # <<<<<<<<<<<<<<
 * 
 *     def __reduce__(self):
 */
  free(__pyx_v_self->label_count_total);

  __Pyx_RefNannyFinishContext();
}

/* Python wrapper */
static PyObject *__pyx_pw_7sklearn_4tree_5_tree_23ClassificationCriterion_5__reduce__(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static PyObject *__pyx_pw_7sklearn_4tree_5_tree_23ClassificationCriterion_5__reduce__(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__reduce__ (wrapper)", 0);
  __pyx_r = __pyx_pf_7sklearn_4tree_5_tree_23ClassificationCriterion_4__reduce__(((struct __pyx_obj_7sklearn_4tree_5_tree_ClassificationCriterion *)__pyx_v_self));
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "sklearn/tree/_tree.pyx":152
 *         free(self.label_count_total)
 * 
 *     def __reduce__(self):             # <<<<<<<<<<<<<<
 *         return (ClassificationCriterion,
 *                 (self.n_outputs,
 */

static PyObject *__pyx_pf_7sklearn_4tree_5_tree_23ClassificationCriterion_4__reduce__(struct __pyx_obj_7sklearn_4tree_5_tree_ClassificationCriterion *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__reduce__", 0);

  /* "sklearn/tree/_tree.pyx":153
 * 
 *     def __reduce__(self):
 *         return (ClassificationCriterion,             # <<<<<<<<<<<<<<
 *                 (self.n_outputs,
 *                  sizet_ptr_to_ndarray(self.n_classes, self.n_outputs)),
 */
  __Pyx_XDECREF(__pyx_r);

  /* "sklearn/tree/_tree.pyx":154
 *     def __reduce__(self):
 *         return (ClassificationCriterion,
 *                 (self.n_outputs,             # <<<<<<<<<<<<<<
 *                  sizet_ptr_to_ndarray(self.n_classes, self.n_outputs)),
 *                 self.__getstate__())
 */
  __pyx_t_1 = __Pyx_PyInt_to_py_Py_intptr_t(__pyx_v_self->__pyx_base.n_outputs); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 154; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_1);

  /* "sklearn/tree/_tree.pyx":155
 *         return (ClassificationCriterion,
 *                 (self.n_outputs,
 *                  sizet_ptr_to_ndarray(self.n_classes, self.n_outputs)),             # <<<<<<<<<<<<<<
 *                 self.__getstate__())
 * 
 */
  __pyx_t_2 = ((PyObject *)__pyx_f_7sklearn_4tree_5_tree_sizet_ptr_to_ndarray(__pyx_v_self->n_classes, __pyx_v_self->__pyx_base.n_outputs)); if (unlikely(!__pyx_t_2)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 155; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = PyTuple_New(2); if (unlikely(!__pyx_t_3)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 154; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_3);
  PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_t_2);
  __Pyx_GIVEREF(__pyx_t_2);
  __pyx_t_1 = 0;
  __pyx_t_2 = 0;

  /* "sklearn/tree/_tree.pyx":156
 *                 (self.n_outputs,
 *                  sizet_ptr_to_ndarray(self.n_classes, self.n_outputs)),
 *                 self.__getstate__())             # <<<<<<<<<<<<<<
 * 
 *     def __getstate__(self):
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s____getstate__); if (unlikely(!__pyx_t_2)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 156; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_1 = PyObject_Call(__pyx_t_2, ((PyObject *)__pyx_empty_tuple), NULL); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 156; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = PyTuple_New(3); if (unlikely(!__pyx_t_2)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 153; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_INCREF(((PyObject *)((PyObject*)__pyx_ptype_7sklearn_4tree_5_tree_ClassificationCriterion)));
  PyTuple_SET_ITEM(__pyx_t_2, 0, ((PyObject *)((PyObject*)__pyx_ptype_7sklearn_4tree_5_tree_ClassificationCriterion)));
  __Pyx_GIVEREF(((PyObject *)((PyObject*)__pyx_ptype_7sklearn_4tree_5_tree_ClassificationCriterion)));
  PyTuple_SET_ITEM(__pyx_t_2, 1, ((PyObject *)__pyx_t_3));
  __Pyx_GIVEREF(((PyObject *)__pyx_t_3));
  PyTuple_SET_ITEM(__pyx_t_2, 2, __pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __pyx_t_3 = 0;
  __pyx_t_1 = 0;
  __pyx_r = ((PyObject *)__pyx_t_2);
  __pyx_t_2 = 0;
  goto __pyx_L0;

  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("sklearn.tree._tree.ClassificationCriterion.__reduce__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_7sklearn_4tree_5_tree_23ClassificationCriterion_7__getstate__(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static PyObject *__pyx_pw_7sklearn_4tree_5_tree_23ClassificationCriterion_7__getstate__(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__getstate__ (wrapper)", 0);
  __pyx_r = __pyx_pf_7sklearn_4tree_5_tree_23ClassificationCriterion_6__getstate__(((struct __pyx_obj_7sklearn_4tree_5_tree_ClassificationCriterion *)__pyx_v_self));
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "sklearn/tree/_tree.pyx":158
 *                 self.__getstate__())
 * 
 *     def __getstate__(self):             # <<<<<<<<<<<<<<
 *         return {}
 * 
 */

static PyObject *__pyx_pf_7sklearn_4tree_5_tree_23ClassificationCriterion_6__getstate__(CYTHON_UNUSED struct __pyx_obj_7sklearn_4tree_5_tree_ClassificationCriterion *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__getstate__", 0);

  /* "sklearn/tree/_tree.pyx":159
 * 
 *     def __getstate__(self):
 *         return {}             # <<<<<<<<<<<<<<
 * 
 *     def __setstate__(self, d):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = PyDict_New(); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 159; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(((PyObject *)__pyx_t_1));
  __pyx_r = ((PyObject *)__pyx_t_1);
  __pyx_t_1 = 0;
  goto __pyx_L0;

  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("sklearn.tree._tree.ClassificationCriterion.__getstate__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_7sklearn_4tree_5_tree_23ClassificationCriterion_9__setstate__(PyObject *__pyx_v_self, PyObject *__pyx_v_d); /*proto*/
static PyObject *__pyx_pw_7sklearn_4tree_5_tree_23ClassificationCriterion_9__setstate__(PyObject *__pyx_v_self, PyObject *__pyx_v_d) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__setstate__ (wrapper)", 0);
  __pyx_r = __pyx_pf_7sklearn_4tree_5_tree_23ClassificationCriterion_8__setstate__(((struct __pyx_obj_7sklearn_4tree_5_tree_ClassificationCriterion *)__pyx_v_self), ((PyObject *)__pyx_v_d));
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "sklearn/tree/_tree.pyx":161
 *         return {}
 * 
 *     def __setstate__(self, d):             # <<<<<<<<<<<<<<
 *         pass
 * 
 */

static PyObject *__pyx_pf_7sklearn_4tree_5_tree_23ClassificationCriterion_8__setstate__(CYTHON_UNUSED struct __pyx_obj_7sklearn_4tree_5_tree_ClassificationCriterion *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v_d) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__setstate__", 0);

  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "sklearn/tree/_tree.pyx":164
 *         pass
 * 
 *     cdef void init(self, DOUBLE_t* y,             # <<<<<<<<<<<<<<
 *                          SIZE_t y_stride,
 *                          DOUBLE_t* sample_weight,
 */

static void __pyx_f_7sklearn_4tree_5_tree_23ClassificationCriterion_init(struct __pyx_obj_7sklearn_4tree_5_tree_ClassificationCriterion *__pyx_v_self, __pyx_t_7sklearn_4tree_5_tree_DOUBLE_t *__pyx_v_y, __pyx_t_7sklearn_4tree_5_tree_SIZE_t __pyx_v_y_stride, __pyx_t_7sklearn_4tree_5_tree_DOUBLE_t *__pyx_v_sample_weight, __pyx_t_7sklearn_4tree_5_tree_SIZE_t *__pyx_v_samples, __pyx_t_7sklearn_4tree_5_tree_SIZE_t __pyx_v_start, __pyx_t_7sklearn_4tree_5_tree_SIZE_t __pyx_v_end) {
  double __pyx_v_weighted_n_node_samples;
  __pyx_t_7sklearn_4tree_5_tree_SIZE_t __pyx_v_n_outputs;
  __pyx_t_7sklearn_4tree_5_tree_SIZE_t *__pyx_v_n_classes;
  __pyx_t_7sklearn_4tree_5_tree_SIZE_t __pyx_v_label_count_stride;
  double *__pyx_v_label_count_total;
  __pyx_t_7sklearn_4tree_5_tree_SIZE_t __pyx_v_i;
  __pyx_t_7sklearn_4tree_5_tree_SIZE_t __pyx_v_p;
  __pyx_t_7sklearn_4tree_5_tree_SIZE_t __pyx_v_k;
  __pyx_t_7sklearn_4tree_5_tree_SIZE_t __pyx_v_c;
  __pyx_t_7sklearn_4tree_5_tree_DOUBLE_t __pyx_v_w;
  __pyx_t_7sklearn_4tree_5_tree_SIZE_t __pyx_v_offset;
  __pyx_t_7sklearn_4tree_5_tree_SIZE_t __pyx_t_1;
  __pyx_t_7sklearn_4tree_5_tree_SIZE_t *__pyx_t_2;
  double *__pyx_t_3;
  __pyx_t_7sklearn_4tree_5_tree_SIZE_t __pyx_t_4;
  int __pyx_t_5;
  int __pyx_t_6;

  /* "sklearn/tree/_tree.pyx":173
 *            children samples[start:start] and samples[start:end]."""
 *         # Initialize fields
 *         self.y = y             # <<<<<<<<<<<<<<
 *         self.y_stride = y_stride
 *         self.sample_weight = sample_weight
 */
  __pyx_v_self->__pyx_base.y = __pyx_v_y;

  /* "sklearn/tree/_tree.pyx":174
 *         # Initialize fields
 *         self.y = y
 *         self.y_stride = y_stride             # <<<<<<<<<<<<<<
 *         self.sample_weight = sample_weight
 *         self.samples = samples
 */
  __pyx_v_self->__pyx_base.y_stride = __pyx_v_y_stride;

  /* "sklearn/tree/_tree.pyx":175
 *         self.y = y
 *         self.y_stride = y_stride
 *         self.sample_weight = sample_weight             # <<<<<<<<<<<<<<
 *         self.samples = samples
 *         self.start = start
 */
  __pyx_v_self->__pyx_base.sample_weight = __pyx_v_sample_weight;

  /* "sklearn/tree/_tree.pyx":176
 *         self.y_stride = y_stride
 *         self.sample_weight = sample_weight
 *         self.samples = samples             # <<<<<<<<<<<<<<
 *         self.start = start
 *         self.end = end
 */
  __pyx_v_self->__pyx_base.samples = __pyx_v_samples;

  /* "sklearn/tree/_tree.pyx":177
 *         self.sample_weight = sample_weight
 *         self.samples = samples
 *         self.start = start             # <<<<<<<<<<<<<<
 *         self.end = end
 *         self.n_node_samples = end - start
 */
  __pyx_v_self->__pyx_base.start = __pyx_v_start;

  /* "sklearn/tree/_tree.pyx":178
 *         self.samples = samples
 *         self.start = start
 *         self.end = end             # <<<<<<<<<<<<<<
 *         self.n_node_samples = end - start
 *         cdef double weighted_n_node_samples = 0.0
 */
  __pyx_v_self->__pyx_base.end = __pyx_v_end;

  /* "sklearn/tree/_tree.pyx":179
 *         self.start = start
 *         self.end = end
 *         self.n_node_samples = end - start             # <<<<<<<<<<<<<<
 *         cdef double weighted_n_node_samples = 0.0
 * 
 */
  __pyx_v_self->__pyx_base.n_node_samples = (__pyx_v_end - __pyx_v_start);

  /* "sklearn/tree/_tree.pyx":180
 *         self.end = end
 *         self.n_node_samples = end - start
 *         cdef double weighted_n_node_samples = 0.0             # <<<<<<<<<<<<<<
 * 
 *         # Initialize label_count_total and weighted_n_node_samples
 */
  __pyx_v_weighted_n_node_samples = 0.0;

  /* "sklearn/tree/_tree.pyx":183
 * 
 *         # Initialize label_count_total and weighted_n_node_samples
 *         cdef SIZE_t n_outputs = self.n_outputs             # <<<<<<<<<<<<<<
 *         cdef SIZE_t* n_classes = self.n_classes
 *         cdef SIZE_t label_count_stride = self.label_count_stride
 */
  __pyx_t_1 = __pyx_v_self->__pyx_base.n_outputs;
  __pyx_v_n_outputs = __pyx_t_1;

  /* "sklearn/tree/_tree.pyx":184
 *         # Initialize label_count_total and weighted_n_node_samples
 *         cdef SIZE_t n_outputs = self.n_outputs
 *         cdef SIZE_t* n_classes = self.n_classes             # <<<<<<<<<<<<<<
 *         cdef SIZE_t label_count_stride = self.label_count_stride
 *         cdef double* label_count_total = self.label_count_total
 */
  __pyx_t_2 = __pyx_v_self->n_classes;
  __pyx_v_n_classes = __pyx_t_2;

  /* "sklearn/tree/_tree.pyx":185
 *         cdef SIZE_t n_outputs = self.n_outputs
 *         cdef SIZE_t* n_classes = self.n_classes
 *         cdef SIZE_t label_count_stride = self.label_count_stride             # <<<<<<<<<<<<<<
 *         cdef double* label_count_total = self.label_count_total
 * 
 */
  __pyx_t_1 = __pyx_v_self->label_count_stride;
  __pyx_v_label_count_stride = __pyx_t_1;

  /* "sklearn/tree/_tree.pyx":186
 *         cdef SIZE_t* n_classes = self.n_classes
 *         cdef SIZE_t label_count_stride = self.label_count_stride
 *         cdef double* label_count_total = self.label_count_total             # <<<<<<<<<<<<<<
 * 
 *         cdef SIZE_t i = 0
 */
  __pyx_t_3 = __pyx_v_self->label_count_total;
  __pyx_v_label_count_total = __pyx_t_3;

  /* "sklearn/tree/_tree.pyx":188
 *         cdef double* label_count_total = self.label_count_total
 * 
 *         cdef SIZE_t i = 0             # <<<<<<<<<<<<<<
 *         cdef SIZE_t p = 0
 *         cdef SIZE_t k = 0
 */
  __pyx_v_i = 0;

  /* "sklearn/tree/_tree.pyx":189
 * 
 *         cdef SIZE_t i = 0
 *         cdef SIZE_t p = 0             # <<<<<<<<<<<<<<
 *         cdef SIZE_t k = 0
 *         cdef SIZE_t c = 0
 */
  __pyx_v_p = 0;

  /* "sklearn/tree/_tree.pyx":190
 *         cdef SIZE_t i = 0
 *         cdef SIZE_t p = 0
 *         cdef SIZE_t k = 0             # <<<<<<<<<<<<<<
 *         cdef SIZE_t c = 0
 *         cdef DOUBLE_t w = 1.0
 */
  __pyx_v_k = 0;

  /* "sklearn/tree/_tree.pyx":191
 *         cdef SIZE_t p = 0
 *         cdef SIZE_t k = 0
 *         cdef SIZE_t c = 0             # <<<<<<<<<<<<<<
 *         cdef DOUBLE_t w = 1.0
 *         cdef SIZE_t offset = 0
 */
  __pyx_v_c = 0;

  /* "sklearn/tree/_tree.pyx":192
 *         cdef SIZE_t k = 0
 *         cdef SIZE_t c = 0
 *         cdef DOUBLE_t w = 1.0             # <<<<<<<<<<<<<<
 *         cdef SIZE_t offset = 0
 * 
 */
  __pyx_v_w = 1.0;

  /* "sklearn/tree/_tree.pyx":193
 *         cdef SIZE_t c = 0
 *         cdef DOUBLE_t w = 1.0
 *         cdef SIZE_t offset = 0             # <<<<<<<<<<<<<<
 * 
 *         for k from 0 <= k < n_outputs:
 */
  __pyx_v_offset = 0;

  /* "sklearn/tree/_tree.pyx":195
 *         cdef SIZE_t offset = 0
 * 
 *         for k from 0 <= k < n_outputs:             # <<<<<<<<<<<<<<
 *             for c from 0 <= c < n_classes[k]:
 *                 label_count_total[offset + c] = 0
 */
  __pyx_t_1 = __pyx_v_n_outputs;
  for (__pyx_v_k = 0; __pyx_v_k < __pyx_t_1; __pyx_v_k++) {

    /* "sklearn/tree/_tree.pyx":196
 * 
 *         for k from 0 <= k < n_outputs:
 *             for c from 0 <= c < n_classes[k]:             # <<<<<<<<<<<<<<
 *                 label_count_total[offset + c] = 0
 *             offset += label_count_stride
 */
    __pyx_t_4 = (__pyx_v_n_classes[__pyx_v_k]);
    for (__pyx_v_c = 0; __pyx_v_c < __pyx_t_4; __pyx_v_c++) {

      /* "sklearn/tree/_tree.pyx":197
 *         for k from 0 <= k < n_outputs:
 *             for c from 0 <= c < n_classes[k]:
 *                 label_count_total[offset + c] = 0             # <<<<<<<<<<<<<<
 *             offset += label_count_stride
 * 
 */
      (__pyx_v_label_count_total[(__pyx_v_offset + __pyx_v_c)]) = 0.0;
    }

    /* "sklearn/tree/_tree.pyx":198
 *             for c from 0 <= c < n_classes[k]:
 *                 label_count_total[offset + c] = 0
 *             offset += label_count_stride             # <<<<<<<<<<<<<<
 * 
 *         for p from start <= p < end:
 */
    __pyx_v_offset = (__pyx_v_offset + __pyx_v_label_count_stride);
  }

  /* "sklearn/tree/_tree.pyx":200
 *             offset += label_count_stride
 * 
 *         for p from start <= p < end:             # <<<<<<<<<<<<<<
 *             i = samples[p]
 * 
 */
  __pyx_t_1 = __pyx_v_end;
  for (__pyx_v_p = __pyx_v_start; __pyx_v_p < __pyx_t_1; __pyx_v_p++) {

    /* "sklearn/tree/_tree.pyx":201
 * 
 *         for p from start <= p < end:
 *             i = samples[p]             # <<<<<<<<<<<<<<
 * 
 *             if sample_weight != NULL:
 */
    __pyx_v_i = (__pyx_v_samples[__pyx_v_p]);

    /* "sklearn/tree/_tree.pyx":203
 *             i = samples[p]
 * 
 *             if sample_weight != NULL:             # <<<<<<<<<<<<<<
 *                 w = sample_weight[i]
 * 
 */
    __pyx_t_5 = ((__pyx_v_sample_weight != NULL) != 0);
    if (__pyx_t_5) {

      /* "sklearn/tree/_tree.pyx":204
 * 
 *             if sample_weight != NULL:
 *                 w = sample_weight[i]             # <<<<<<<<<<<<<<
 * 
 *             for k from 0 <= k < n_outputs:
 */
      __pyx_v_w = (__pyx_v_sample_weight[__pyx_v_i]);
      goto __pyx_L9;
    }
    __pyx_L9:;

    /* "sklearn/tree/_tree.pyx":206
 *                 w = sample_weight[i]
 * 
 *             for k from 0 <= k < n_outputs:             # <<<<<<<<<<<<<<
 *                 c = <SIZE_t> y[i * y_stride + k]
 *                 label_count_total[k * label_count_stride + c] += w
 */
    __pyx_t_4 = __pyx_v_n_outputs;
    for (__pyx_v_k = 0; __pyx_v_k < __pyx_t_4; __pyx_v_k++) {

      /* "sklearn/tree/_tree.pyx":207
 * 
 *             for k from 0 <= k < n_outputs:
 *                 c = <SIZE_t> y[i * y_stride + k]             # <<<<<<<<<<<<<<
 *                 label_count_total[k * label_count_stride + c] += w
 * 
 */
      __pyx_v_c = ((__pyx_t_7sklearn_4tree_5_tree_SIZE_t)(__pyx_v_y[((__pyx_v_i * __pyx_v_y_stride) + __pyx_v_k)]));

      /* "sklearn/tree/_tree.pyx":208
 *             for k from 0 <= k < n_outputs:
 *                 c = <SIZE_t> y[i * y_stride + k]
 *                 label_count_total[k * label_count_stride + c] += w             # <<<<<<<<<<<<<<
 * 
 *             weighted_n_node_samples += w
 */
      __pyx_t_6 = ((__pyx_v_k * __pyx_v_label_count_stride) + __pyx_v_c);
      (__pyx_v_label_count_total[__pyx_t_6]) = ((__pyx_v_label_count_total[__pyx_t_6]) + __pyx_v_w);
    }

    /* "sklearn/tree/_tree.pyx":210
 *                 label_count_total[k * label_count_stride + c] += w
 * 
 *             weighted_n_node_samples += w             # <<<<<<<<<<<<<<
 * 
 *         self.weighted_n_node_samples = weighted_n_node_samples
 */
    __pyx_v_weighted_n_node_samples = (__pyx_v_weighted_n_node_samples + __pyx_v_w);
  }

  /* "sklearn/tree/_tree.pyx":212
 *             weighted_n_node_samples += w
 * 
 *         self.weighted_n_node_samples = weighted_n_node_samples             # <<<<<<<<<<<<<<
 * 
 *         # Reset to pos=start
 */
  __pyx_v_self->__pyx_base.weighted_n_node_samples = __pyx_v_weighted_n_node_samples;

  /* "sklearn/tree/_tree.pyx":215
 * 
 *         # Reset to pos=start
 *         self.reset()             # <<<<<<<<<<<<<<
 * 
 *     cdef void reset(self) nogil:
 */
  ((struct __pyx_vtabstruct_7sklearn_4tree_5_tree_ClassificationCriterion *)__pyx_v_self->__pyx_base.__pyx_vtab)->__pyx_base.reset(((struct __pyx_obj_7sklearn_4tree_5_tree_Criterion *)__pyx_v_self));

}

/* "sklearn/tree/_tree.pyx":217
 *         self.reset()
 * 
 *     cdef void reset(self) nogil:             # <<<<<<<<<<<<<<
 *         """Reset the criterion at pos=start."""
 *         self.pos = self.start
 */

static void __pyx_f_7sklearn_4tree_5_tree_23ClassificationCriterion_reset(struct __pyx_obj_7sklearn_4tree_5_tree_ClassificationCriterion *__pyx_v_self) {
  __pyx_t_7sklearn_4tree_5_tree_SIZE_t __pyx_v_n_outputs;
  __pyx_t_7sklearn_4tree_5_tree_SIZE_t *__pyx_v_n_classes;
  __pyx_t_7sklearn_4tree_5_tree_SIZE_t __pyx_v_label_count_stride;
  double *__pyx_v_label_count_total;
  double *__pyx_v_label_count_left;
  double *__pyx_v_label_count_right;
  __pyx_t_7sklearn_4tree_5_tree_SIZE_t __pyx_v_k;
  __pyx_t_7sklearn_4tree_5_tree_SIZE_t __pyx_v_c;
  __pyx_t_7sklearn_4tree_5_tree_SIZE_t __pyx_v_offset;
  __pyx_t_7sklearn_4tree_5_tree_SIZE_t __pyx_t_1;
  double __pyx_t_2;
  __pyx_t_7sklearn_4tree_5_tree_SIZE_t *__pyx_t_3;
  double *__pyx_t_4;
  __pyx_t_7sklearn_4tree_5_tree_SIZE_t __pyx_t_5;

  /* "sklearn/tree/_tree.pyx":219
 *     cdef void reset(self) nogil:
 *         """Reset the criterion at pos=start."""
 *         self.pos = self.start             # <<<<<<<<<<<<<<
 * 
 *         self.weighted_n_left = 0.0
 */
  __pyx_t_1 = __pyx_v_self->__pyx_base.start;
  __pyx_v_self->__pyx_base.pos = __pyx_t_1;

  /* "sklearn/tree/_tree.pyx":221
 *         self.pos = self.start
 * 
 *         self.weighted_n_left = 0.0             # <<<<<<<<<<<<<<
 *         self.weighted_n_right = self.weighted_n_node_samples
 * 
 */
  __pyx_v_self->__pyx_base.weighted_n_left = 0.0;

  /* "sklearn/tree/_tree.pyx":222
 * 
 *         self.weighted_n_left = 0.0
 *         self.weighted_n_right = self.weighted_n_node_samples             # <<<<<<<<<<<<<<
 * 
 *         cdef SIZE_t n_outputs = self.n_outputs
 */
  __pyx_t_2 = __pyx_v_self->__pyx_base.weighted_n_node_samples;
  __pyx_v_self->__pyx_base.weighted_n_right = __pyx_t_2;

  /* "sklearn/tree/_tree.pyx":224
 *         self.weighted_n_right = self.weighted_n_node_samples
 * 
 *         cdef SIZE_t n_outputs = self.n_outputs             # <<<<<<<<<<<<<<
 *         cdef SIZE_t* n_classes = self.n_classes
 *         cdef SIZE_t label_count_stride = self.label_count_stride
 */
  __pyx_t_1 = __pyx_v_self->__pyx_base.n_outputs;
  __pyx_v_n_outputs = __pyx_t_1;

  /* "sklearn/tree/_tree.pyx":225
 * 
 *         cdef SIZE_t n_outputs = self.n_outputs
 *         cdef SIZE_t* n_classes = self.n_classes             # <<<<<<<<<<<<<<
 *         cdef SIZE_t label_count_stride = self.label_count_stride
 *         cdef double* label_count_total = self.label_count_total
 */
  __pyx_t_3 = __pyx_v_self->n_classes;
  __pyx_v_n_classes = __pyx_t_3;

  /* "sklearn/tree/_tree.pyx":226
 *         cdef SIZE_t n_outputs = self.n_outputs
 *         cdef SIZE_t* n_classes = self.n_classes
 *         cdef SIZE_t label_count_stride = self.label_count_stride             # <<<<<<<<<<<<<<
 *         cdef double* label_count_total = self.label_count_total
 *         cdef double* label_count_left = self.label_count_left
 */
  __pyx_t_1 = __pyx_v_self->label_count_stride;
  __pyx_v_label_count_stride = __pyx_t_1;

  /* "sklearn/tree/_tree.pyx":227
 *         cdef SIZE_t* n_classes = self.n_classes
 *         cdef SIZE_t label_count_stride = self.label_count_stride
 *         cdef double* label_count_total = self.label_count_total             # <<<<<<<<<<<<<<
 *         cdef double* label_count_left = self.label_count_left
 *         cdef double* label_count_right = self.label_count_right
 */
  __pyx_t_4 = __pyx_v_self->label_count_total;
  __pyx_v_label_count_total = __pyx_t_4;

  /* "sklearn/tree/_tree.pyx":228
 *         cdef SIZE_t label_count_stride = self.label_count_stride
 *         cdef double* label_count_total = self.label_count_total
 *         cdef double* label_count_left = self.label_count_left             # <<<<<<<<<<<<<<
 *         cdef double* label_count_right = self.label_count_right
 * 
 */
  __pyx_t_4 = __pyx_v_self->label_count_left;
  __pyx_v_label_count_left = __pyx_t_4;

  /* "sklearn/tree/_tree.pyx":229
 *         cdef double* label_count_total = self.label_count_total
 *         cdef double* label_count_left = self.label_count_left
 *         cdef double* label_count_right = self.label_count_right             # <<<<<<<<<<<<<<
 * 
 *         cdef SIZE_t k = 0
 */
  __pyx_t_4 = __pyx_v_self->label_count_right;
  __pyx_v_label_count_right = __pyx_t_4;

  /* "sklearn/tree/_tree.pyx":231
 *         cdef double* label_count_right = self.label_count_right
 * 
 *         cdef SIZE_t k = 0             # <<<<<<<<<<<<<<
 *         cdef SIZE_t c = 0
 *         cdef SIZE_t offset = 0
 */
  __pyx_v_k = 0;

  /* "sklearn/tree/_tree.pyx":232
 * 
 *         cdef SIZE_t k = 0
 *         cdef SIZE_t c = 0             # <<<<<<<<<<<<<<
 *         cdef SIZE_t offset = 0
 * 
 */
  __pyx_v_c = 0;

  /* "sklearn/tree/_tree.pyx":233
 *         cdef SIZE_t k = 0
 *         cdef SIZE_t c = 0
 *         cdef SIZE_t offset = 0             # <<<<<<<<<<<<<<
 * 
 *         for k from 0 <= k < n_outputs:
 */
  __pyx_v_offset = 0;

  /* "sklearn/tree/_tree.pyx":235
 *         cdef SIZE_t offset = 0
 * 
 *         for k from 0 <= k < n_outputs:             # <<<<<<<<<<<<<<
 *             for c from 0 <= c < n_classes[k]:
 *                 # Reset left label counts to 0
 */
  __pyx_t_1 = __pyx_v_n_outputs;
  for (__pyx_v_k = 0; __pyx_v_k < __pyx_t_1; __pyx_v_k++) {

    /* "sklearn/tree/_tree.pyx":236
 * 
 *         for k from 0 <= k < n_outputs:
 *             for c from 0 <= c < n_classes[k]:             # <<<<<<<<<<<<<<
 *                 # Reset left label counts to 0
 *                 label_count_left[offset + c] = 0
 */
    __pyx_t_5 = (__pyx_v_n_classes[__pyx_v_k]);
    for (__pyx_v_c = 0; __pyx_v_c < __pyx_t_5; __pyx_v_c++) {

      /* "sklearn/tree/_tree.pyx":238
 *             for c from 0 <= c < n_classes[k]:
 *                 # Reset left label counts to 0
 *                 label_count_left[offset + c] = 0             # <<<<<<<<<<<<<<
 *                 # Reset right label counts to the initial counts
 *                 label_count_right[offset + c] = label_count_total[offset + c]
 */
      (__pyx_v_label_count_left[(__pyx_v_offset + __pyx_v_c)]) = 0.0;

      /* "sklearn/tree/_tree.pyx":240
 *                 label_count_left[offset + c] = 0
 *                 # Reset right label counts to the initial counts
 *                 label_count_right[offset + c] = label_count_total[offset + c]             # <<<<<<<<<<<<<<
 *             offset += label_count_stride
 * 
 */
      (__pyx_v_label_count_right[(__pyx_v_offset + __pyx_v_c)]) = (__pyx_v_label_count_total[(__pyx_v_offset + __pyx_v_c)]);
    }

    /* "sklearn/tree/_tree.pyx":241
 *                 # Reset right label counts to the initial counts
 *                 label_count_right[offset + c] = label_count_total[offset + c]
 *             offset += label_count_stride             # <<<<<<<<<<<<<<
 * 
 *     cdef void update(self, SIZE_t new_pos) nogil:
 */
    __pyx_v_offset = (__pyx_v_offset + __pyx_v_label_count_stride);
  }

}

/* "sklearn/tree/_tree.pyx":243
 *             offset += label_count_stride
 * 
 *     cdef void update(self, SIZE_t new_pos) nogil:             # <<<<<<<<<<<<<<
 *         """Update the collected statistics by moving samples[pos:new_pos] from
 *             the right child to the left child."""
 */

static void __pyx_f_7sklearn_4tree_5_tree_23ClassificationCriterion_update(struct __pyx_obj_7sklearn_4tree_5_tree_ClassificationCriterion *__pyx_v_self, __pyx_t_7sklearn_4tree_5_tree_SIZE_t __pyx_v_new_pos) {
  __pyx_t_7sklearn_4tree_5_tree_DOUBLE_t *__pyx_v_y;
  __pyx_t_7sklearn_4tree_5_tree_SIZE_t __pyx_v_y_stride;
  __pyx_t_7sklearn_4tree_5_tree_DOUBLE_t *__pyx_v_sample_weight;
  __pyx_t_7sklearn_4tree_5_tree_SIZE_t *__pyx_v_samples;
  __pyx_t_7sklearn_4tree_5_tree_SIZE_t __pyx_v_pos;
  __pyx_t_7sklearn_4tree_5_tree_SIZE_t __pyx_v_n_outputs;
  CYTHON_UNUSED __pyx_t_7sklearn_4tree_5_tree_SIZE_t *__pyx_v_n_classes;
  __pyx_t_7sklearn_4tree_5_tree_SIZE_t __pyx_v_label_count_stride;
  CYTHON_UNUSED double *__pyx_v_label_count_total;
  double *__pyx_v_label_count_left;
  double *__pyx_v_label_count_right;
  double __pyx_v_weighted_n_left;
  double __pyx_v_weighted_n_right;
  __pyx_t_7sklearn_4tree_5_tree_SIZE_t __pyx_v_i;
  __pyx_t_7sklearn_4tree_5_tree_SIZE_t __pyx_v_p;
  __pyx_t_7sklearn_4tree_5_tree_SIZE_t __pyx_v_k;
  __pyx_t_7sklearn_4tree_5_tree_SIZE_t __pyx_v_label_index;
  __pyx_t_7sklearn_4tree_5_tree_DOUBLE_t __pyx_v_w;
  __pyx_t_7sklearn_4tree_5_tree_DOUBLE_t *__pyx_t_1;
  __pyx_t_7sklearn_4tree_5_tree_SIZE_t __pyx_t_2;
  __pyx_t_7sklearn_4tree_5_tree_SIZE_t *__pyx_t_3;
  double *__pyx_t_4;
  double __pyx_t_5;
  int __pyx_t_6;
  __pyx_t_7sklearn_4tree_5_tree_SIZE_t __pyx_t_7;
  __pyx_t_7sklearn_4tree_5_tree_SIZE_t __pyx_t_8;

  /* "sklearn/tree/_tree.pyx":246
 *         """Update the collected statistics by moving samples[pos:new_pos] from
 *             the right child to the left child."""
 *         cdef DOUBLE_t* y = self.y             # <<<<<<<<<<<<<<
 *         cdef SIZE_t y_stride = self.y_stride
 *         cdef DOUBLE_t* sample_weight = self.sample_weight
 */
  __pyx_t_1 = __pyx_v_self->__pyx_base.y;
  __pyx_v_y = __pyx_t_1;

  /* "sklearn/tree/_tree.pyx":247
 *             the right child to the left child."""
 *         cdef DOUBLE_t* y = self.y
 *         cdef SIZE_t y_stride = self.y_stride             # <<<<<<<<<<<<<<
 *         cdef DOUBLE_t* sample_weight = self.sample_weight
 * 
 */
  __pyx_t_2 = __pyx_v_self->__pyx_base.y_stride;
  __pyx_v_y_stride = __pyx_t_2;

  /* "sklearn/tree/_tree.pyx":248
 *         cdef DOUBLE_t* y = self.y
 *         cdef SIZE_t y_stride = self.y_stride
 *         cdef DOUBLE_t* sample_weight = self.sample_weight             # <<<<<<<<<<<<<<
 * 
 *         cdef SIZE_t* samples = self.samples
 */
  __pyx_t_1 = __pyx_v_self->__pyx_base.sample_weight;
  __pyx_v_sample_weight = __pyx_t_1;

  /* "sklearn/tree/_tree.pyx":250
 *         cdef DOUBLE_t* sample_weight = self.sample_weight
 * 
 *         cdef SIZE_t* samples = self.samples             # <<<<<<<<<<<<<<
 *         cdef SIZE_t pos = self.pos
 * 
 */
  __pyx_t_3 = __pyx_v_self->__pyx_base.samples;
  __pyx_v_samples = __pyx_t_3;

  /* "sklearn/tree/_tree.pyx":251
 * 
 *         cdef SIZE_t* samples = self.samples
 *         cdef SIZE_t pos = self.pos             # <<<<<<<<<<<<<<
 * 
 *         cdef SIZE_t n_outputs = self.n_outputs
 */
  __pyx_t_2 = __pyx_v_self->__pyx_base.pos;
  __pyx_v_pos = __pyx_t_2;

  /* "sklearn/tree/_tree.pyx":253
 *         cdef SIZE_t pos = self.pos
 * 
 *         cdef SIZE_t n_outputs = self.n_outputs             # <<<<<<<<<<<<<<
 *         cdef SIZE_t* n_classes = self.n_classes
 *         cdef SIZE_t label_count_stride = self.label_count_stride
 */
  __pyx_t_2 = __pyx_v_self->__pyx_base.n_outputs;
  __pyx_v_n_outputs = __pyx_t_2;

  /* "sklearn/tree/_tree.pyx":254
 * 
 *         cdef SIZE_t n_outputs = self.n_outputs
 *         cdef SIZE_t* n_classes = self.n_classes             # <<<<<<<<<<<<<<
 *         cdef SIZE_t label_count_stride = self.label_count_stride
 *         cdef double* label_count_total = self.label_count_total
 */
  __pyx_t_3 = __pyx_v_self->n_classes;
  __pyx_v_n_classes = __pyx_t_3;

  /* "sklearn/tree/_tree.pyx":255
 *         cdef SIZE_t n_outputs = self.n_outputs
 *         cdef SIZE_t* n_classes = self.n_classes
 *         cdef SIZE_t label_count_stride = self.label_count_stride             # <<<<<<<<<<<<<<
 *         cdef double* label_count_total = self.label_count_total
 *         cdef double* label_count_left = self.label_count_left
 */
  __pyx_t_2 = __pyx_v_self->label_count_stride;
  __pyx_v_label_count_stride = __pyx_t_2;

  /* "sklearn/tree/_tree.pyx":256
 *         cdef SIZE_t* n_classes = self.n_classes
 *         cdef SIZE_t label_count_stride = self.label_count_stride
 *         cdef double* label_count_total = self.label_count_total             # <<<<<<<<<<<<<<
 *         cdef double* label_count_left = self.label_count_left
 *         cdef double* label_count_right = self.label_count_right
 */
  __pyx_t_4 = __pyx_v_self->label_count_total;
  __pyx_v_label_count_total = __pyx_t_4;

  /* "sklearn/tree/_tree.pyx":257
 *         cdef SIZE_t label_count_stride = self.label_count_stride
 *         cdef double* label_count_total = self.label_count_total
 *         cdef double* label_count_left = self.label_count_left             # <<<<<<<<<<<<<<
 *         cdef double* label_count_right = self.label_count_right
 * 
 */
  __pyx_t_4 = __pyx_v_self->label_count_left;
  __pyx_v_label_count_left = __pyx_t_4;

  /* "sklearn/tree/_tree.pyx":258
 *         cdef double* label_count_total = self.label_count_total
 *         cdef double* label_count_left = self.label_count_left
 *         cdef double* label_count_right = self.label_count_right             # <<<<<<<<<<<<<<
 * 
 *         cdef double weighted_n_left = self.weighted_n_left
 */
  __pyx_t_4 = __pyx_v_self->label_count_right;
  __pyx_v_label_count_right = __pyx_t_4;

  /* "sklearn/tree/_tree.pyx":260
 *         cdef double* label_count_right = self.label_count_right
 * 
 *         cdef double weighted_n_left = self.weighted_n_left             # <<<<<<<<<<<<<<
 *         cdef double weighted_n_right = self.weighted_n_right
 * 
 */
  __pyx_t_5 = __pyx_v_self->__pyx_base.weighted_n_left;
  __pyx_v_weighted_n_left = __pyx_t_5;

  /* "sklearn/tree/_tree.pyx":261
 * 
 *         cdef double weighted_n_left = self.weighted_n_left
 *         cdef double weighted_n_right = self.weighted_n_right             # <<<<<<<<<<<<<<
 * 
 *         cdef SIZE_t i
 */
  __pyx_t_5 = __pyx_v_self->__pyx_base.weighted_n_right;
  __pyx_v_weighted_n_right = __pyx_t_5;

  /* "sklearn/tree/_tree.pyx":267
 *         cdef SIZE_t k
 *         cdef SIZE_t label_index
 *         cdef DOUBLE_t w = 1.0             # <<<<<<<<<<<<<<
 * 
 *         # Note: We assume start <= pos < new_pos <= end
 */
  __pyx_v_w = 1.0;

  /* "sklearn/tree/_tree.pyx":271
 *         # Note: We assume start <= pos < new_pos <= end
 * 
 *         for p from pos <= p < new_pos:             # <<<<<<<<<<<<<<
 *             i = samples[p]
 * 
 */
  __pyx_t_2 = __pyx_v_new_pos;
  for (__pyx_v_p = __pyx_v_pos; __pyx_v_p < __pyx_t_2; __pyx_v_p++) {

    /* "sklearn/tree/_tree.pyx":272
 * 
 *         for p from pos <= p < new_pos:
 *             i = samples[p]             # <<<<<<<<<<<<<<
 * 
 *             if sample_weight != NULL:
 */
    __pyx_v_i = (__pyx_v_samples[__pyx_v_p]);

    /* "sklearn/tree/_tree.pyx":274
 *             i = samples[p]
 * 
 *             if sample_weight != NULL:             # <<<<<<<<<<<<<<
 *                 w  = sample_weight[i]
 * 
 */
    __pyx_t_6 = ((__pyx_v_sample_weight != NULL) != 0);
    if (__pyx_t_6) {

      /* "sklearn/tree/_tree.pyx":275
 * 
 *             if sample_weight != NULL:
 *                 w  = sample_weight[i]             # <<<<<<<<<<<<<<
 * 
 *             for k from 0 <= k < n_outputs:
 */
      __pyx_v_w = (__pyx_v_sample_weight[__pyx_v_i]);
      goto __pyx_L5;
    }
    __pyx_L5:;

    /* "sklearn/tree/_tree.pyx":277
 *                 w  = sample_weight[i]
 * 
 *             for k from 0 <= k < n_outputs:             # <<<<<<<<<<<<<<
 *                 label_index = (k * label_count_stride +
 *                                <SIZE_t> y[i * y_stride + k])
 */
    __pyx_t_7 = __pyx_v_n_outputs;
    for (__pyx_v_k = 0; __pyx_v_k < __pyx_t_7; __pyx_v_k++) {

      /* "sklearn/tree/_tree.pyx":279
 *             for k from 0 <= k < n_outputs:
 *                 label_index = (k * label_count_stride +
 *                                <SIZE_t> y[i * y_stride + k])             # <<<<<<<<<<<<<<
 *                 label_count_left[label_index] += w
 *                 label_count_right[label_index] -= w
 */
      __pyx_v_label_index = ((__pyx_v_k * __pyx_v_label_count_stride) + ((__pyx_t_7sklearn_4tree_5_tree_SIZE_t)(__pyx_v_y[((__pyx_v_i * __pyx_v_y_stride) + __pyx_v_k)])));

      /* "sklearn/tree/_tree.pyx":280
 *                 label_index = (k * label_count_stride +
 *                                <SIZE_t> y[i * y_stride + k])
 *                 label_count_left[label_index] += w             # <<<<<<<<<<<<<<
 *                 label_count_right[label_index] -= w
 * 
 */
      __pyx_t_8 = __pyx_v_label_index;
      (__pyx_v_label_count_left[__pyx_t_8]) = ((__pyx_v_label_count_left[__pyx_t_8]) + __pyx_v_w);

      /* "sklearn/tree/_tree.pyx":281
 *                                <SIZE_t> y[i * y_stride + k])
 *                 label_count_left[label_index] += w
 *                 label_count_right[label_index] -= w             # <<<<<<<<<<<<<<
 * 
 *             weighted_n_left += w
 */
      __pyx_t_8 = __pyx_v_label_index;
      (__pyx_v_label_count_right[__pyx_t_8]) = ((__pyx_v_label_count_right[__pyx_t_8]) - __pyx_v_w);
    }

    /* "sklearn/tree/_tree.pyx":283
 *                 label_count_right[label_index] -= w
 * 
 *             weighted_n_left += w             # <<<<<<<<<<<<<<
 *             weighted_n_right -= w
 * 
 */
    __pyx_v_weighted_n_left = (__pyx_v_weighted_n_left + __pyx_v_w);

    /* "sklearn/tree/_tree.pyx":284
 * 
 *             weighted_n_left += w
 *             weighted_n_right -= w             # <<<<<<<<<<<<<<
 * 
 *         self.weighted_n_left = weighted_n_left
 */
    __pyx_v_weighted_n_right = (__pyx_v_weighted_n_right - __pyx_v_w);
  }

  /* "sklearn/tree/_tree.pyx":286
 *             weighted_n_right -= w
 * 
 *         self.weighted_n_left = weighted_n_left             # <<<<<<<<<<<<<<
 *         self.weighted_n_right = weighted_n_right
 * 
 */
  __pyx_v_self->__pyx_base.weighted_n_left = __pyx_v_weighted_n_left;

  /* "sklearn/tree/_tree.pyx":287
 * 
 *         self.weighted_n_left = weighted_n_left
 *         self.weighted_n_right = weighted_n_right             # <<<<<<<<<<<<<<
 * 
 *         self.pos = new_pos
 */
  __pyx_v_self->__pyx_base.weighted_n_right = __pyx_v_weighted_n_right;

  /* "sklearn/tree/_tree.pyx":289
 *         self.weighted_n_right = weighted_n_right
 * 
 *         self.pos = new_pos             # <<<<<<<<<<<<<<
 * 
 *     cdef double node_impurity(self) nogil:
 */
  __pyx_v_self->__pyx_base.pos = __pyx_v_new_pos;

}

/* "sklearn/tree/_tree.pyx":291
 *         self.pos = new_pos
 * 
 *     cdef double node_impurity(self) nogil:             # <<<<<<<<<<<<<<
 *         pass
 * 
 */

static double __pyx_f_7sklearn_4tree_5_tree_23ClassificationCriterion_node_impurity(CYTHON_UNUSED struct __pyx_obj_7sklearn_4tree_5_tree_ClassificationCriterion *__pyx_v_self) {
  double __pyx_r;

  __pyx_r = 0;
  return __pyx_r;
}

/* "sklearn/tree/_tree.pyx":294
 *         pass
 * 
 *     cdef double children_impurity(self) nogil:             # <<<<<<<<<<<<<<
 *         pass
 * 
 */

static double __pyx_f_7sklearn_4tree_5_tree_23ClassificationCriterion_children_impurity(CYTHON_UNUSED struct __pyx_obj_7sklearn_4tree_5_tree_ClassificationCriterion *__pyx_v_self) {
  double __pyx_r;

  __pyx_r = 0;
  return __pyx_r;
}

/* "sklearn/tree/_tree.pyx":297
 *         pass
 * 
 *     cdef void node_value(self, double* dest) nogil:             # <<<<<<<<<<<<<<
 *         """Compute the node value of samples[start:end] into dest."""
 *         cdef SIZE_t n_outputs = self.n_outputs
 */

static void __pyx_f_7sklearn_4tree_5_tree_23ClassificationCriterion_node_value(struct __pyx_obj_7sklearn_4tree_5_tree_ClassificationCriterion *__pyx_v_self, double *__pyx_v_dest) {
  __pyx_t_7sklearn_4tree_5_tree_SIZE_t __pyx_v_n_outputs;
  __pyx_t_7sklearn_4tree_5_tree_SIZE_t *__pyx_v_n_classes;
  __pyx_t_7sklearn_4tree_5_tree_SIZE_t __pyx_v_label_count_stride;
  double *__pyx_v_label_count_total;
  __pyx_t_7sklearn_4tree_5_tree_SIZE_t __pyx_v_k;
  __pyx_t_7sklearn_4tree_5_tree_SIZE_t __pyx_v_c;
  __pyx_t_7sklearn_4tree_5_tree_SIZE_t __pyx_v_offset;
  __pyx_t_7sklearn_4tree_5_tree_SIZE_t __pyx_t_1;
  __pyx_t_7sklearn_4tree_5_tree_SIZE_t *__pyx_t_2;
  double *__pyx_t_3;
  __pyx_t_7sklearn_4tree_5_tree_SIZE_t __pyx_t_4;

  /* "sklearn/tree/_tree.pyx":299
 *     cdef void node_value(self, double* dest) nogil:
 *         """Compute the node value of samples[start:end] into dest."""
 *         cdef SIZE_t n_outputs = self.n_outputs             # <<<<<<<<<<<<<<
 *         cdef SIZE_t* n_classes = self.n_classes
 *         cdef SIZE_t label_count_stride = self.label_count_stride
 */
  __pyx_t_1 = __pyx_v_self->__pyx_base.n_outputs;
  __pyx_v_n_outputs = __pyx_t_1;

  /* "sklearn/tree/_tree.pyx":300
 *         """Compute the node value of samples[start:end] into dest."""
 *         cdef SIZE_t n_outputs = self.n_outputs
 *         cdef SIZE_t* n_classes = self.n_classes             # <<<<<<<<<<<<<<
 *         cdef SIZE_t label_count_stride = self.label_count_stride
 *         cdef double* label_count_total = self.label_count_total
 */
  __pyx_t_2 = __pyx_v_self->n_classes;
  __pyx_v_n_classes = __pyx_t_2;

  /* "sklearn/tree/_tree.pyx":301
 *         cdef SIZE_t n_outputs = self.n_outputs
 *         cdef SIZE_t* n_classes = self.n_classes
 *         cdef SIZE_t label_count_stride = self.label_count_stride             # <<<<<<<<<<<<<<
 *         cdef double* label_count_total = self.label_count_total
 * 
 */
  __pyx_t_1 = __pyx_v_self->label_count_stride;
  __pyx_v_label_count_stride = __pyx_t_1;

  /* "sklearn/tree/_tree.pyx":302
 *         cdef SIZE_t* n_classes = self.n_classes
 *         cdef SIZE_t label_count_stride = self.label_count_stride
 *         cdef double* label_count_total = self.label_count_total             # <<<<<<<<<<<<<<
 * 
 *         cdef SIZE_t k
 */
  __pyx_t_3 = __pyx_v_self->label_count_total;
  __pyx_v_label_count_total = __pyx_t_3;

  /* "sklearn/tree/_tree.pyx":306
 *         cdef SIZE_t k
 *         cdef SIZE_t c
 *         cdef SIZE_t offset = 0             # <<<<<<<<<<<<<<
 * 
 *         for k from 0 <= k < n_outputs:
 */
  __pyx_v_offset = 0;

  /* "sklearn/tree/_tree.pyx":308
 *         cdef SIZE_t offset = 0
 * 
 *         for k from 0 <= k < n_outputs:             # <<<<<<<<<<<<<<
 *             for c from 0 <= c < n_classes[k]:
 *                 dest[offset + c] = label_count_total[offset + c]
 */
  __pyx_t_1 = __pyx_v_n_outputs;
  for (__pyx_v_k = 0; __pyx_v_k < __pyx_t_1; __pyx_v_k++) {

    /* "sklearn/tree/_tree.pyx":309
 * 
 *         for k from 0 <= k < n_outputs:
 *             for c from 0 <= c < n_classes[k]:             # <<<<<<<<<<<<<<
 *                 dest[offset + c] = label_count_total[offset + c]
 *             offset += label_count_stride
 */
    __pyx_t_4 = (__pyx_v_n_classes[__pyx_v_k]);
    for (__pyx_v_c = 0; __pyx_v_c < __pyx_t_4; __pyx_v_c++) {

      /* "sklearn/tree/_tree.pyx":310
 *         for k from 0 <= k < n_outputs:
 *             for c from 0 <= c < n_classes[k]:
 *                 dest[offset + c] = label_count_total[offset + c]             # <<<<<<<<<<<<<<
 *             offset += label_count_stride
 * 
 */
      (__pyx_v_dest[(__pyx_v_offset + __pyx_v_c)]) = (__pyx_v_label_count_total[(__pyx_v_offset + __pyx_v_c)]);
    }

    /* "sklearn/tree/_tree.pyx":311
 *             for c from 0 <= c < n_classes[k]:
 *                 dest[offset + c] = label_count_total[offset + c]
 *             offset += label_count_stride             # <<<<<<<<<<<<<<
 * 
 * 
 */
    __pyx_v_offset = (__pyx_v_offset + __pyx_v_label_count_stride);
  }

}

/* "sklearn/tree/_tree.pyx":328
 *         cross-entropy = - \sum_{k=0}^{K-1} pmk log(pmk)
 *     """
 *     cdef double node_impurity(self) nogil:             # <<<<<<<<<<<<<<
 *         """Evaluate the impurity of the current node, i.e. the impurity of
 *            samples[start:end]."""
 */

static double __pyx_f_7sklearn_4tree_5_tree_7Entropy_node_impurity(struct __pyx_obj_7sklearn_4tree_5_tree_Entropy *__pyx_v_self) {
  double __pyx_v_weighted_n_node_samples;
  __pyx_t_7sklearn_4tree_5_tree_SIZE_t __pyx_v_n_outputs;
  __pyx_t_7sklearn_4tree_5_tree_SIZE_t *__pyx_v_n_classes;
  __pyx_t_7sklearn_4tree_5_tree_SIZE_t __pyx_v_label_count_stride;
  double *__pyx_v_label_count_total;
  double __pyx_v_entropy;
  double __pyx_v_total;
  double __pyx_v_tmp;
  __pyx_t_7sklearn_4tree_5_tree_SIZE_t __pyx_v_k;
  __pyx_t_7sklearn_4tree_5_tree_SIZE_t __pyx_v_c;
  __pyx_t_7sklearn_4tree_5_tree_SIZE_t __pyx_v_offset;
  double __pyx_r;
  double __pyx_t_1;
  __pyx_t_7sklearn_4tree_5_tree_SIZE_t __pyx_t_2;
  __pyx_t_7sklearn_4tree_5_tree_SIZE_t *__pyx_t_3;
  double *__pyx_t_4;
  __pyx_t_7sklearn_4tree_5_tree_SIZE_t __pyx_t_5;
  int __pyx_t_6;

  /* "sklearn/tree/_tree.pyx":331
 *         """Evaluate the impurity of the current node, i.e. the impurity of
 *            samples[start:end]."""
 *         cdef double weighted_n_node_samples = self.weighted_n_node_samples             # <<<<<<<<<<<<<<
 * 
 *         cdef SIZE_t n_outputs = self.n_outputs
 */
  __pyx_t_1 = __pyx_v_self->__pyx_base.__pyx_base.weighted_n_node_samples;
  __pyx_v_weighted_n_node_samples = __pyx_t_1;

  /* "sklearn/tree/_tree.pyx":333
 *         cdef double weighted_n_node_samples = self.weighted_n_node_samples
 * 
 *         cdef SIZE_t n_outputs = self.n_outputs             # <<<<<<<<<<<<<<
 *         cdef SIZE_t* n_classes = self.n_classes
 *         cdef SIZE_t label_count_stride = self.label_count_stride
 */
  __pyx_t_2 = __pyx_v_self->__pyx_base.__pyx_base.n_outputs;
  __pyx_v_n_outputs = __pyx_t_2;

  /* "sklearn/tree/_tree.pyx":334
 * 
 *         cdef SIZE_t n_outputs = self.n_outputs
 *         cdef SIZE_t* n_classes = self.n_classes             # <<<<<<<<<<<<<<
 *         cdef SIZE_t label_count_stride = self.label_count_stride
 *         cdef double* label_count_total = self.label_count_total
 */
  __pyx_t_3 = __pyx_v_self->__pyx_base.n_classes;
  __pyx_v_n_classes = __pyx_t_3;

  /* "sklearn/tree/_tree.pyx":335
 *         cdef SIZE_t n_outputs = self.n_outputs
 *         cdef SIZE_t* n_classes = self.n_classes
 *         cdef SIZE_t label_count_stride = self.label_count_stride             # <<<<<<<<<<<<<<
 *         cdef double* label_count_total = self.label_count_total
 * 
 */
  __pyx_t_2 = __pyx_v_self->__pyx_base.label_count_stride;
  __pyx_v_label_count_stride = __pyx_t_2;

  /* "sklearn/tree/_tree.pyx":336
 *         cdef SIZE_t* n_classes = self.n_classes
 *         cdef SIZE_t label_count_stride = self.label_count_stride
 *         cdef double* label_count_total = self.label_count_total             # <<<<<<<<<<<<<<
 * 
 *         cdef double entropy = 0.0
 */
  __pyx_t_4 = __pyx_v_self->__pyx_base.label_count_total;
  __pyx_v_label_count_total = __pyx_t_4;

  /* "sklearn/tree/_tree.pyx":338
 *         cdef double* label_count_total = self.label_count_total
 * 
 *         cdef double entropy = 0.0             # <<<<<<<<<<<<<<
 *         cdef double total = 0.0
 *         cdef double tmp
 */
  __pyx_v_entropy = 0.0;

  /* "sklearn/tree/_tree.pyx":339
 * 
 *         cdef double entropy = 0.0
 *         cdef double total = 0.0             # <<<<<<<<<<<<<<
 *         cdef double tmp
 *         cdef SIZE_t k
 */
  __pyx_v_total = 0.0;

  /* "sklearn/tree/_tree.pyx":343
 *         cdef SIZE_t k
 *         cdef SIZE_t c
 *         cdef SIZE_t offset = 0             # <<<<<<<<<<<<<<
 * 
 *         for k from 0 <= k < n_outputs:
 */
  __pyx_v_offset = 0;

  /* "sklearn/tree/_tree.pyx":345
 *         cdef SIZE_t offset = 0
 * 
 *         for k from 0 <= k < n_outputs:             # <<<<<<<<<<<<<<
 *             entropy = 0.0
 * 
 */
  __pyx_t_2 = __pyx_v_n_outputs;
  for (__pyx_v_k = 0; __pyx_v_k < __pyx_t_2; __pyx_v_k++) {

    /* "sklearn/tree/_tree.pyx":346
 * 
 *         for k from 0 <= k < n_outputs:
 *             entropy = 0.0             # <<<<<<<<<<<<<<
 * 
 *             for c from 0 <= c < n_classes[k]:
 */
    __pyx_v_entropy = 0.0;

    /* "sklearn/tree/_tree.pyx":348
 *             entropy = 0.0
 * 
 *             for c from 0 <= c < n_classes[k]:             # <<<<<<<<<<<<<<
 *                 tmp = label_count_total[offset + c]
 *                 if tmp > 0.0:
 */
    __pyx_t_5 = (__pyx_v_n_classes[__pyx_v_k]);
    for (__pyx_v_c = 0; __pyx_v_c < __pyx_t_5; __pyx_v_c++) {

      /* "sklearn/tree/_tree.pyx":349
 * 
 *             for c from 0 <= c < n_classes[k]:
 *                 tmp = label_count_total[offset + c]             # <<<<<<<<<<<<<<
 *                 if tmp > 0.0:
 *                     tmp /= weighted_n_node_samples
 */
      __pyx_v_tmp = (__pyx_v_label_count_total[(__pyx_v_offset + __pyx_v_c)]);

      /* "sklearn/tree/_tree.pyx":350
 *             for c from 0 <= c < n_classes[k]:
 *                 tmp = label_count_total[offset + c]
 *                 if tmp > 0.0:             # <<<<<<<<<<<<<<
 *                     tmp /= weighted_n_node_samples
 *                     entropy -= tmp * log(tmp)
 */
      __pyx_t_6 = ((__pyx_v_tmp > 0.0) != 0);
      if (__pyx_t_6) {

        /* "sklearn/tree/_tree.pyx":351
 *                 tmp = label_count_total[offset + c]
 *                 if tmp > 0.0:
 *                     tmp /= weighted_n_node_samples             # <<<<<<<<<<<<<<
 *                     entropy -= tmp * log(tmp)
 * 
 */
        __pyx_v_tmp = (__pyx_v_tmp / __pyx_v_weighted_n_node_samples);

        /* "sklearn/tree/_tree.pyx":352
 *                 if tmp > 0.0:
 *                     tmp /= weighted_n_node_samples
 *                     entropy -= tmp * log(tmp)             # <<<<<<<<<<<<<<
 * 
 *             total += entropy
 */
        __pyx_v_entropy = (__pyx_v_entropy - (__pyx_v_tmp * __pyx_f_7sklearn_4tree_5_tree_log(__pyx_v_tmp)));
        goto __pyx_L7;
      }
      __pyx_L7:;
    }

    /* "sklearn/tree/_tree.pyx":354
 *                     entropy -= tmp * log(tmp)
 * 
 *             total += entropy             # <<<<<<<<<<<<<<
 *             offset += label_count_stride
 * 
 */
    __pyx_v_total = (__pyx_v_total + __pyx_v_entropy);

    /* "sklearn/tree/_tree.pyx":355
 * 
 *             total += entropy
 *             offset += label_count_stride             # <<<<<<<<<<<<<<
 * 
 *         return total / n_outputs
 */
    __pyx_v_offset = (__pyx_v_offset + __pyx_v_label_count_stride);
  }

  /* "sklearn/tree/_tree.pyx":357
 *             offset += label_count_stride
 * 
 *         return total / n_outputs             # <<<<<<<<<<<<<<
 * 
 *     cdef double children_impurity(self) nogil:
 */
  __pyx_r = (__pyx_v_total / __pyx_v_n_outputs);
  goto __pyx_L0;

  __pyx_r = 0;
  __pyx_L0:;
  return __pyx_r;
}

/* "sklearn/tree/_tree.pyx":359
 *         return total / n_outputs
 * 
 *     cdef double children_impurity(self) nogil:             # <<<<<<<<<<<<<<
 *         """Evaluate the impurity in children nodes, i.e. the impurity of
 *            samples[start:pos] + the impurity of samples[pos:end]."""
 */

static double __pyx_f_7sklearn_4tree_5_tree_7Entropy_children_impurity(struct __pyx_obj_7sklearn_4tree_5_tree_Entropy *__pyx_v_self) {
  double __pyx_v_weighted_n_node_samples;
  double __pyx_v_weighted_n_left;
  double __pyx_v_weighted_n_right;
  __pyx_t_7sklearn_4tree_5_tree_SIZE_t __pyx_v_n_outputs;
  __pyx_t_7sklearn_4tree_5_tree_SIZE_t *__pyx_v_n_classes;
  __pyx_t_7sklearn_4tree_5_tree_SIZE_t __pyx_v_label_count_stride;
  double *__pyx_v_label_count_left;
  double *__pyx_v_label_count_right;
  double __pyx_v_entropy_left;
  double __pyx_v_entropy_right;
  double __pyx_v_total;
  double __pyx_v_tmp;
  __pyx_t_7sklearn_4tree_5_tree_SIZE_t __pyx_v_k;
  __pyx_t_7sklearn_4tree_5_tree_SIZE_t __pyx_v_c;
  __pyx_t_7sklearn_4tree_5_tree_SIZE_t __pyx_v_offset;
  double __pyx_r;
  double __pyx_t_1;
  __pyx_t_7sklearn_4tree_5_tree_SIZE_t __pyx_t_2;
  __pyx_t_7sklearn_4tree_5_tree_SIZE_t *__pyx_t_3;
  double *__pyx_t_4;
  __pyx_t_7sklearn_4tree_5_tree_SIZE_t __pyx_t_5;
  int __pyx_t_6;

  /* "sklearn/tree/_tree.pyx":362
 *         """Evaluate the impurity in children nodes, i.e. the impurity of
 *            samples[start:pos] + the impurity of samples[pos:end]."""
 *         cdef double weighted_n_node_samples = self.weighted_n_node_samples             # <<<<<<<<<<<<<<
 *         cdef double weighted_n_left = self.weighted_n_left
 *         cdef double weighted_n_right = self.weighted_n_right
 */
  __pyx_t_1 = __pyx_v_self->__pyx_base.__pyx_base.weighted_n_node_samples;
  __pyx_v_weighted_n_node_samples = __pyx_t_1;

  /* "sklearn/tree/_tree.pyx":363
 *            samples[start:pos] + the impurity of samples[pos:end]."""
 *         cdef double weighted_n_node_samples = self.weighted_n_node_samples
 *         cdef double weighted_n_left = self.weighted_n_left             # <<<<<<<<<<<<<<
 *         cdef double weighted_n_right = self.weighted_n_right
 * 
 */
  __pyx_t_1 = __pyx_v_self->__pyx_base.__pyx_base.weighted_n_left;
  __pyx_v_weighted_n_left = __pyx_t_1;

  /* "sklearn/tree/_tree.pyx":364
 *         cdef double weighted_n_node_samples = self.weighted_n_node_samples
 *         cdef double weighted_n_left = self.weighted_n_left
 *         cdef double weighted_n_right = self.weighted_n_right             # <<<<<<<<<<<<<<
 * 
 *         cdef SIZE_t n_outputs = self.n_outputs
 */
  __pyx_t_1 = __pyx_v_self->__pyx_base.__pyx_base.weighted_n_right;
  __pyx_v_weighted_n_right = __pyx_t_1;

  /* "sklearn/tree/_tree.pyx":366
 *         cdef double weighted_n_right = self.weighted_n_right
 * 
 *         cdef SIZE_t n_outputs = self.n_outputs             # <<<<<<<<<<<<<<
 *         cdef SIZE_t* n_classes = self.n_classes
 *         cdef SIZE_t label_count_stride = self.label_count_stride
 */
  __pyx_t_2 = __pyx_v_self->__pyx_base.__pyx_base.n_outputs;
  __pyx_v_n_outputs = __pyx_t_2;

  /* "sklearn/tree/_tree.pyx":367
 * 
 *         cdef SIZE_t n_outputs = self.n_outputs
 *         cdef SIZE_t* n_classes = self.n_classes             # <<<<<<<<<<<<<<
 *         cdef SIZE_t label_count_stride = self.label_count_stride
 *         cdef double* label_count_left = self.label_count_left
 */
  __pyx_t_3 = __pyx_v_self->__pyx_base.n_classes;
  __pyx_v_n_classes = __pyx_t_3;

  /* "sklearn/tree/_tree.pyx":368
 *         cdef SIZE_t n_outputs = self.n_outputs
 *         cdef SIZE_t* n_classes = self.n_classes
 *         cdef SIZE_t label_count_stride = self.label_count_stride             # <<<<<<<<<<<<<<
 *         cdef double* label_count_left = self.label_count_left
 *         cdef double* label_count_right = self.label_count_right
 */
  __pyx_t_2 = __pyx_v_self->__pyx_base.label_count_stride;
  __pyx_v_label_count_stride = __pyx_t_2;

  /* "sklearn/tree/_tree.pyx":369
 *         cdef SIZE_t* n_classes = self.n_classes
 *         cdef SIZE_t label_count_stride = self.label_count_stride
 *         cdef double* label_count_left = self.label_count_left             # <<<<<<<<<<<<<<
 *         cdef double* label_count_right = self.label_count_right
 * 
 */
  __pyx_t_4 = __pyx_v_self->__pyx_base.label_count_left;
  __pyx_v_label_count_left = __pyx_t_4;

  /* "sklearn/tree/_tree.pyx":370
 *         cdef SIZE_t label_count_stride = self.label_count_stride
 *         cdef double* label_count_left = self.label_count_left
 *         cdef double* label_count_right = self.label_count_right             # <<<<<<<<<<<<<<
 * 
 *         cdef double entropy_left = 0.0
 */
  __pyx_t_4 = __pyx_v_self->__pyx_base.label_count_right;
  __pyx_v_label_count_right = __pyx_t_4;

  /* "sklearn/tree/_tree.pyx":372
 *         cdef double* label_count_right = self.label_count_right
 * 
 *         cdef double entropy_left = 0.0             # <<<<<<<<<<<<<<
 *         cdef double entropy_right = 0.0
 *         cdef double total = 0.0
 */
  __pyx_v_entropy_left = 0.0;

  /* "sklearn/tree/_tree.pyx":373
 * 
 *         cdef double entropy_left = 0.0
 *         cdef double entropy_right = 0.0             # <<<<<<<<<<<<<<
 *         cdef double total = 0.0
 *         cdef double tmp
 */
  __pyx_v_entropy_right = 0.0;

  /* "sklearn/tree/_tree.pyx":374
 *         cdef double entropy_left = 0.0
 *         cdef double entropy_right = 0.0
 *         cdef double total = 0.0             # <<<<<<<<<<<<<<
 *         cdef double tmp
 *         cdef SIZE_t k
 */
  __pyx_v_total = 0.0;

  /* "sklearn/tree/_tree.pyx":378
 *         cdef SIZE_t k
 *         cdef SIZE_t c
 *         cdef SIZE_t offset = 0             # <<<<<<<<<<<<<<
 * 
 *         for k from 0 <= k < n_outputs:
 */
  __pyx_v_offset = 0;

  /* "sklearn/tree/_tree.pyx":380
 *         cdef SIZE_t offset = 0
 * 
 *         for k from 0 <= k < n_outputs:             # <<<<<<<<<<<<<<
 *             entropy_left = 0.0
 *             entropy_right = 0.0
 */
  __pyx_t_2 = __pyx_v_n_outputs;
  for (__pyx_v_k = 0; __pyx_v_k < __pyx_t_2; __pyx_v_k++) {

    /* "sklearn/tree/_tree.pyx":381
 * 
 *         for k from 0 <= k < n_outputs:
 *             entropy_left = 0.0             # <<<<<<<<<<<<<<
 *             entropy_right = 0.0
 * 
 */
    __pyx_v_entropy_left = 0.0;

    /* "sklearn/tree/_tree.pyx":382
 *         for k from 0 <= k < n_outputs:
 *             entropy_left = 0.0
 *             entropy_right = 0.0             # <<<<<<<<<<<<<<
 * 
 *             for c from 0 <= c < n_classes[k]:
 */
    __pyx_v_entropy_right = 0.0;

    /* "sklearn/tree/_tree.pyx":384
 *             entropy_right = 0.0
 * 
 *             for c from 0 <= c < n_classes[k]:             # <<<<<<<<<<<<<<
 *                 tmp = label_count_left[offset + c]
 *                 if tmp > 0.0:
 */
    __pyx_t_5 = (__pyx_v_n_classes[__pyx_v_k]);
    for (__pyx_v_c = 0; __pyx_v_c < __pyx_t_5; __pyx_v_c++) {

      /* "sklearn/tree/_tree.pyx":385
 * 
 *             for c from 0 <= c < n_classes[k]:
 *                 tmp = label_count_left[offset + c]             # <<<<<<<<<<<<<<
 *                 if tmp > 0.0:
 *                     tmp /= weighted_n_left
 */
      __pyx_v_tmp = (__pyx_v_label_count_left[(__pyx_v_offset + __pyx_v_c)]);

      /* "sklearn/tree/_tree.pyx":386
 *             for c from 0 <= c < n_classes[k]:
 *                 tmp = label_count_left[offset + c]
 *                 if tmp > 0.0:             # <<<<<<<<<<<<<<
 *                     tmp /= weighted_n_left
 *                     entropy_left -= tmp * log(tmp)
 */
      __pyx_t_6 = ((__pyx_v_tmp > 0.0) != 0);
      if (__pyx_t_6) {

        /* "sklearn/tree/_tree.pyx":387
 *                 tmp = label_count_left[offset + c]
 *                 if tmp > 0.0:
 *                     tmp /= weighted_n_left             # <<<<<<<<<<<<<<
 *                     entropy_left -= tmp * log(tmp)
 * 
 */
        __pyx_v_tmp = (__pyx_v_tmp / __pyx_v_weighted_n_left);

        /* "sklearn/tree/_tree.pyx":388
 *                 if tmp > 0.0:
 *                     tmp /= weighted_n_left
 *                     entropy_left -= tmp * log(tmp)             # <<<<<<<<<<<<<<
 * 
 *                 tmp = label_count_right[offset + c]
 */
        __pyx_v_entropy_left = (__pyx_v_entropy_left - (__pyx_v_tmp * __pyx_f_7sklearn_4tree_5_tree_log(__pyx_v_tmp)));
        goto __pyx_L7;
      }
      __pyx_L7:;

      /* "sklearn/tree/_tree.pyx":390
 *                     entropy_left -= tmp * log(tmp)
 * 
 *                 tmp = label_count_right[offset + c]             # <<<<<<<<<<<<<<
 *                 if tmp > 0.0:
 *                     tmp /= weighted_n_right
 */
      __pyx_v_tmp = (__pyx_v_label_count_right[(__pyx_v_offset + __pyx_v_c)]);

      /* "sklearn/tree/_tree.pyx":391
 * 
 *                 tmp = label_count_right[offset + c]
 *                 if tmp > 0.0:             # <<<<<<<<<<<<<<
 *                     tmp /= weighted_n_right
 *                     entropy_right -= tmp * log(tmp)
 */
      __pyx_t_6 = ((__pyx_v_tmp > 0.0) != 0);
      if (__pyx_t_6) {

        /* "sklearn/tree/_tree.pyx":392
 *                 tmp = label_count_right[offset + c]
 *                 if tmp > 0.0:
 *                     tmp /= weighted_n_right             # <<<<<<<<<<<<<<
 *                     entropy_right -= tmp * log(tmp)
 * 
 */
        __pyx_v_tmp = (__pyx_v_tmp / __pyx_v_weighted_n_right);

        /* "sklearn/tree/_tree.pyx":393
 *                 if tmp > 0.0:
 *                     tmp /= weighted_n_right
 *                     entropy_right -= tmp * log(tmp)             # <<<<<<<<<<<<<<
 * 
 *             total += weighted_n_left * entropy_left
 */
        __pyx_v_entropy_right = (__pyx_v_entropy_right - (__pyx_v_tmp * __pyx_f_7sklearn_4tree_5_tree_log(__pyx_v_tmp)));
        goto __pyx_L8;
      }
      __pyx_L8:;
    }

    /* "sklearn/tree/_tree.pyx":395
 *                     entropy_right -= tmp * log(tmp)
 * 
 *             total += weighted_n_left * entropy_left             # <<<<<<<<<<<<<<
 *             total += weighted_n_right * entropy_right
 *             offset += label_count_stride
 */
    __pyx_v_total = (__pyx_v_total + (__pyx_v_weighted_n_left * __pyx_v_entropy_left));

    /* "sklearn/tree/_tree.pyx":396
 * 
 *             total += weighted_n_left * entropy_left
 *             total += weighted_n_right * entropy_right             # <<<<<<<<<<<<<<
 *             offset += label_count_stride
 * 
 */
    __pyx_v_total = (__pyx_v_total + (__pyx_v_weighted_n_right * __pyx_v_entropy_right));

    /* "sklearn/tree/_tree.pyx":397
 *             total += weighted_n_left * entropy_left
 *             total += weighted_n_right * entropy_right
 *             offset += label_count_stride             # <<<<<<<<<<<<<<
 * 
 *         return total / (weighted_n_node_samples * n_outputs)
 */
    __pyx_v_offset = (__pyx_v_offset + __pyx_v_label_count_stride);
  }

  /* "sklearn/tree/_tree.pyx":399
 *             offset += label_count_stride
 * 
 *         return total / (weighted_n_node_samples * n_outputs)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = (__pyx_v_total / (__pyx_v_weighted_n_node_samples * __pyx_v_n_outputs));
  goto __pyx_L0;

  __pyx_r = 0;
  __pyx_L0:;
  return __pyx_r;
}

/* "sklearn/tree/_tree.pyx":417
 *               = 1 - \sum_{k=0}^{K-1} pmk ** 2
 *     """
 *     cdef double node_impurity(self) nogil:             # <<<<<<<<<<<<<<
 *         """Evaluate the impurity of the current node, i.e. the impurity of
 *            samples[start:end]."""
 */

static double __pyx_f_7sklearn_4tree_5_tree_4Gini_node_impurity(struct __pyx_obj_7sklearn_4tree_5_tree_Gini *__pyx_v_self) {
  double __pyx_v_weighted_n_node_samples;
  __pyx_t_7sklearn_4tree_5_tree_SIZE_t __pyx_v_n_outputs;
  __pyx_t_7sklearn_4tree_5_tree_SIZE_t *__pyx_v_n_classes;
  __pyx_t_7sklearn_4tree_5_tree_SIZE_t __pyx_v_label_count_stride;
  double *__pyx_v_label_count_total;
  double __pyx_v_gini;
  double __pyx_v_total;
  double __pyx_v_tmp;
  __pyx_t_7sklearn_4tree_5_tree_SIZE_t __pyx_v_k;
  __pyx_t_7sklearn_4tree_5_tree_SIZE_t __pyx_v_c;
  __pyx_t_7sklearn_4tree_5_tree_SIZE_t __pyx_v_offset;
  double __pyx_r;
  double __pyx_t_1;
  __pyx_t_7sklearn_4tree_5_tree_SIZE_t __pyx_t_2;
  __pyx_t_7sklearn_4tree_5_tree_SIZE_t *__pyx_t_3;
  double *__pyx_t_4;
  __pyx_t_7sklearn_4tree_5_tree_SIZE_t __pyx_t_5;

  /* "sklearn/tree/_tree.pyx":420
 *         """Evaluate the impurity of the current node, i.e. the impurity of
 *            samples[start:end]."""
 *         cdef double weighted_n_node_samples = self.weighted_n_node_samples             # <<<<<<<<<<<<<<
 * 
 *         cdef SIZE_t n_outputs = self.n_outputs
 */
  __pyx_t_1 = __pyx_v_self->__pyx_base.__pyx_base.weighted_n_node_samples;
  __pyx_v_weighted_n_node_samples = __pyx_t_1;

  /* "sklearn/tree/_tree.pyx":422
 *         cdef double weighted_n_node_samples = self.weighted_n_node_samples
 * 
 *         cdef SIZE_t n_outputs = self.n_outputs             # <<<<<<<<<<<<<<
 *         cdef SIZE_t* n_classes = self.n_classes
 *         cdef SIZE_t label_count_stride = self.label_count_stride
 */
  __pyx_t_2 = __pyx_v_self->__pyx_base.__pyx_base.n_outputs;
  __pyx_v_n_outputs = __pyx_t_2;

  /* "sklearn/tree/_tree.pyx":423
 * 
 *         cdef SIZE_t n_outputs = self.n_outputs
 *         cdef SIZE_t* n_classes = self.n_classes             # <<<<<<<<<<<<<<
 *         cdef SIZE_t label_count_stride = self.label_count_stride
 *         cdef double* label_count_total = self.label_count_total
 */
  __pyx_t_3 = __pyx_v_self->__pyx_base.n_classes;
  __pyx_v_n_classes = __pyx_t_3;

  /* "sklearn/tree/_tree.pyx":424
 *         cdef SIZE_t n_outputs = self.n_outputs
 *         cdef SIZE_t* n_classes = self.n_classes
 *         cdef SIZE_t label_count_stride = self.label_count_stride             # <<<<<<<<<<<<<<
 *         cdef double* label_count_total = self.label_count_total
 * 
 */
  __pyx_t_2 = __pyx_v_self->__pyx_base.label_count_stride;
  __pyx_v_label_count_stride = __pyx_t_2;

  /* "sklearn/tree/_tree.pyx":425
 *         cdef SIZE_t* n_classes = self.n_classes
 *         cdef SIZE_t label_count_stride = self.label_count_stride
 *         cdef double* label_count_total = self.label_count_total             # <<<<<<<<<<<<<<
 * 
 *         cdef double gini = 0.0
 */
  __pyx_t_4 = __pyx_v_self->__pyx_base.label_count_total;
  __pyx_v_label_count_total = __pyx_t_4;

  /* "sklearn/tree/_tree.pyx":427
 *         cdef double* label_count_total = self.label_count_total
 * 
 *         cdef double gini = 0.0             # <<<<<<<<<<<<<<
 *         cdef double total = 0.0
 *         cdef double tmp
 */
  __pyx_v_gini = 0.0;

  /* "sklearn/tree/_tree.pyx":428
 * 
 *         cdef double gini = 0.0
 *         cdef double total = 0.0             # <<<<<<<<<<<<<<
 *         cdef double tmp
 *         cdef SIZE_t k
 */
  __pyx_v_total = 0.0;

  /* "sklearn/tree/_tree.pyx":432
 *         cdef SIZE_t k
 *         cdef SIZE_t c
 *         cdef SIZE_t offset = 0             # <<<<<<<<<<<<<<
 * 
 *         for k from 0 <= k < n_outputs:
 */
  __pyx_v_offset = 0;

  /* "sklearn/tree/_tree.pyx":434
 *         cdef SIZE_t offset = 0
 * 
 *         for k from 0 <= k < n_outputs:             # <<<<<<<<<<<<<<
 *             gini = 0.0
 * 
 */
  __pyx_t_2 = __pyx_v_n_outputs;
  for (__pyx_v_k = 0; __pyx_v_k < __pyx_t_2; __pyx_v_k++) {

    /* "sklearn/tree/_tree.pyx":435
 * 
 *         for k from 0 <= k < n_outputs:
 *             gini = 0.0             # <<<<<<<<<<<<<<
 * 
 *             for c from 0 <= c < n_classes[k]:
 */
    __pyx_v_gini = 0.0;

    /* "sklearn/tree/_tree.pyx":437
 *             gini = 0.0
 * 
 *             for c from 0 <= c < n_classes[k]:             # <<<<<<<<<<<<<<
 *                 tmp = label_count_total[offset + c]
 *                 gini += tmp * tmp
 */
    __pyx_t_5 = (__pyx_v_n_classes[__pyx_v_k]);
    for (__pyx_v_c = 0; __pyx_v_c < __pyx_t_5; __pyx_v_c++) {

      /* "sklearn/tree/_tree.pyx":438
 * 
 *             for c from 0 <= c < n_classes[k]:
 *                 tmp = label_count_total[offset + c]             # <<<<<<<<<<<<<<
 *                 gini += tmp * tmp
 * 
 */
      __pyx_v_tmp = (__pyx_v_label_count_total[(__pyx_v_offset + __pyx_v_c)]);

      /* "sklearn/tree/_tree.pyx":439
 *             for c from 0 <= c < n_classes[k]:
 *                 tmp = label_count_total[offset + c]
 *                 gini += tmp * tmp             # <<<<<<<<<<<<<<
 * 
 *             gini = 1.0 - gini / (weighted_n_node_samples *
 */
      __pyx_v_gini = (__pyx_v_gini + (__pyx_v_tmp * __pyx_v_tmp));
    }

    /* "sklearn/tree/_tree.pyx":442
 * 
 *             gini = 1.0 - gini / (weighted_n_node_samples *
 *                                  weighted_n_node_samples)             # <<<<<<<<<<<<<<
 * 
 *             total += gini
 */
    __pyx_v_gini = (1.0 - (__pyx_v_gini / (__pyx_v_weighted_n_node_samples * __pyx_v_weighted_n_node_samples)));

    /* "sklearn/tree/_tree.pyx":444
 *                                  weighted_n_node_samples)
 * 
 *             total += gini             # <<<<<<<<<<<<<<
 *             offset += label_count_stride
 * 
 */
    __pyx_v_total = (__pyx_v_total + __pyx_v_gini);

    /* "sklearn/tree/_tree.pyx":445
 * 
 *             total += gini
 *             offset += label_count_stride             # <<<<<<<<<<<<<<
 * 
 *         return total / n_outputs
 */
    __pyx_v_offset = (__pyx_v_offset + __pyx_v_label_count_stride);
  }

  /* "sklearn/tree/_tree.pyx":447
 *             offset += label_count_stride
 * 
 *         return total / n_outputs             # <<<<<<<<<<<<<<
 * 
 *     cdef double children_impurity(self) nogil:
 */
  __pyx_r = (__pyx_v_total / __pyx_v_n_outputs);
  goto __pyx_L0;

  __pyx_r = 0;
  __pyx_L0:;
  return __pyx_r;
}

/* "sklearn/tree/_tree.pyx":449
 *         return total / n_outputs
 * 
 *     cdef double children_impurity(self) nogil:             # <<<<<<<<<<<<<<
 *         """Evaluate the impurity in children nodes, i.e. the impurity of
 *            samples[start:pos] + the impurity of samples[pos:end]."""
 */

static double __pyx_f_7sklearn_4tree_5_tree_4Gini_children_impurity(struct __pyx_obj_7sklearn_4tree_5_tree_Gini *__pyx_v_self) {
  double __pyx_v_weighted_n_node_samples;
  double __pyx_v_weighted_n_left;
  double __pyx_v_weighted_n_right;
  __pyx_t_7sklearn_4tree_5_tree_SIZE_t __pyx_v_n_outputs;
  __pyx_t_7sklearn_4tree_5_tree_SIZE_t *__pyx_v_n_classes;
  __pyx_t_7sklearn_4tree_5_tree_SIZE_t __pyx_v_label_count_stride;
  double *__pyx_v_label_count_left;
  double *__pyx_v_label_count_right;
  double __pyx_v_gini_left;
  double __pyx_v_gini_right;
  double __pyx_v_total;
  double __pyx_v_tmp;
  __pyx_t_7sklearn_4tree_5_tree_SIZE_t __pyx_v_k;
  __pyx_t_7sklearn_4tree_5_tree_SIZE_t __pyx_v_c;
  __pyx_t_7sklearn_4tree_5_tree_SIZE_t __pyx_v_offset;
  double __pyx_r;
  double __pyx_t_1;
  __pyx_t_7sklearn_4tree_5_tree_SIZE_t __pyx_t_2;
  __pyx_t_7sklearn_4tree_5_tree_SIZE_t *__pyx_t_3;
  double *__pyx_t_4;
  __pyx_t_7sklearn_4tree_5_tree_SIZE_t __pyx_t_5;

  /* "sklearn/tree/_tree.pyx":452
 *         """Evaluate the impurity in children nodes, i.e. the impurity of
 *            samples[start:pos] + the impurity of samples[pos:end]."""
 *         cdef double weighted_n_node_samples = self.weighted_n_node_samples             # <<<<<<<<<<<<<<
 *         cdef double weighted_n_left = self.weighted_n_left
 *         cdef double weighted_n_right = self.weighted_n_right
 */
  __pyx_t_1 = __pyx_v_self->__pyx_base.__pyx_base.weighted_n_node_samples;
  __pyx_v_weighted_n_node_samples = __pyx_t_1;

  /* "sklearn/tree/_tree.pyx":453
 *            samples[start:pos] + the impurity of samples[pos:end]."""
 *         cdef double weighted_n_node_samples = self.weighted_n_node_samples
 *         cdef double weighted_n_left = self.weighted_n_left             # <<<<<<<<<<<<<<
 *         cdef double weighted_n_right = self.weighted_n_right
 * 
 */
  __pyx_t_1 = __pyx_v_self->__pyx_base.__pyx_base.weighted_n_left;
  __pyx_v_weighted_n_left = __pyx_t_1;

  /* "sklearn/tree/_tree.pyx":454
 *         cdef double weighted_n_node_samples = self.weighted_n_node_samples
 *         cdef double weighted_n_left = self.weighted_n_left
 *         cdef double weighted_n_right = self.weighted_n_right             # <<<<<<<<<<<<<<
 * 
 *         cdef SIZE_t n_outputs = self.n_outputs
 */
  __pyx_t_1 = __pyx_v_self->__pyx_base.__pyx_base.weighted_n_right;
  __pyx_v_weighted_n_right = __pyx_t_1;

  /* "sklearn/tree/_tree.pyx":456
 *         cdef double weighted_n_right = self.weighted_n_right
 * 
 *         cdef SIZE_t n_outputs = self.n_outputs             # <<<<<<<<<<<<<<
 *         cdef SIZE_t* n_classes = self.n_classes
 *         cdef SIZE_t label_count_stride = self.label_count_stride
 */
  __pyx_t_2 = __pyx_v_self->__pyx_base.__pyx_base.n_outputs;
  __pyx_v_n_outputs = __pyx_t_2;

  /* "sklearn/tree/_tree.pyx":457
 * 
 *         cdef SIZE_t n_outputs = self.n_outputs
 *         cdef SIZE_t* n_classes = self.n_classes             # <<<<<<<<<<<<<<
 *         cdef SIZE_t label_count_stride = self.label_count_stride
 *         cdef double* label_count_left = self.label_count_left
 */
  __pyx_t_3 = __pyx_v_self->__pyx_base.n_classes;
  __pyx_v_n_classes = __pyx_t_3;

  /* "sklearn/tree/_tree.pyx":458
 *         cdef SIZE_t n_outputs = self.n_outputs
 *         cdef SIZE_t* n_classes = self.n_classes
 *         cdef SIZE_t label_count_stride = self.label_count_stride             # <<<<<<<<<<<<<<
 *         cdef double* label_count_left = self.label_count_left
 *         cdef double* label_count_right = self.label_count_right
 */
  __pyx_t_2 = __pyx_v_self->__pyx_base.label_count_stride;
  __pyx_v_label_count_stride = __pyx_t_2;

  /* "sklearn/tree/_tree.pyx":459
 *         cdef SIZE_t* n_classes = self.n_classes
 *         cdef SIZE_t label_count_stride = self.label_count_stride
 *         cdef double* label_count_left = self.label_count_left             # <<<<<<<<<<<<<<
 *         cdef double* label_count_right = self.label_count_right
 * 
 */
  __pyx_t_4 = __pyx_v_self->__pyx_base.label_count_left;
  __pyx_v_label_count_left = __pyx_t_4;

  /* "sklearn/tree/_tree.pyx":460
 *         cdef SIZE_t label_count_stride = self.label_count_stride
 *         cdef double* label_count_left = self.label_count_left
 *         cdef double* label_count_right = self.label_count_right             # <<<<<<<<<<<<<<
 * 
 *         cdef double gini_left = 0.0
 */
  __pyx_t_4 = __pyx_v_self->__pyx_base.label_count_right;
  __pyx_v_label_count_right = __pyx_t_4;

  /* "sklearn/tree/_tree.pyx":462
 *         cdef double* label_count_right = self.label_count_right
 * 
 *         cdef double gini_left = 0.0             # <<<<<<<<<<<<<<
 *         cdef double gini_right = 0.0
 *         cdef double total = 0.0
 */
  __pyx_v_gini_left = 0.0;

  /* "sklearn/tree/_tree.pyx":463
 * 
 *         cdef double gini_left = 0.0
 *         cdef double gini_right = 0.0             # <<<<<<<<<<<<<<
 *         cdef double total = 0.0
 *         cdef double tmp
 */
  __pyx_v_gini_right = 0.0;

  /* "sklearn/tree/_tree.pyx":464
 *         cdef double gini_left = 0.0
 *         cdef double gini_right = 0.0
 *         cdef double total = 0.0             # <<<<<<<<<<<<<<
 *         cdef double tmp
 *         cdef SIZE_t k
 */
  __pyx_v_total = 0.0;

  /* "sklearn/tree/_tree.pyx":468
 *         cdef SIZE_t k
 *         cdef SIZE_t c
 *         cdef SIZE_t offset = 0             # <<<<<<<<<<<<<<
 * 
 *         for k from 0 <= k < n_outputs:
 */
  __pyx_v_offset = 0;

  /* "sklearn/tree/_tree.pyx":470
 *         cdef SIZE_t offset = 0
 * 
 *         for k from 0 <= k < n_outputs:             # <<<<<<<<<<<<<<
 *             gini_left = 0.0
 *             gini_right = 0.0
 */
  __pyx_t_2 = __pyx_v_n_outputs;
  for (__pyx_v_k = 0; __pyx_v_k < __pyx_t_2; __pyx_v_k++) {

    /* "sklearn/tree/_tree.pyx":471
 * 
 *         for k from 0 <= k < n_outputs:
 *             gini_left = 0.0             # <<<<<<<<<<<<<<
 *             gini_right = 0.0
 * 
 */
    __pyx_v_gini_left = 0.0;

    /* "sklearn/tree/_tree.pyx":472
 *         for k from 0 <= k < n_outputs:
 *             gini_left = 0.0
 *             gini_right = 0.0             # <<<<<<<<<<<<<<
 * 
 *             for c from 0 <= c < n_classes[k]:
 */
    __pyx_v_gini_right = 0.0;

    /* "sklearn/tree/_tree.pyx":474
 *             gini_right = 0.0
 * 
 *             for c from 0 <= c < n_classes[k]:             # <<<<<<<<<<<<<<
 *                 tmp = label_count_left[offset + c]
 *                 gini_left += tmp * tmp
 */
    __pyx_t_5 = (__pyx_v_n_classes[__pyx_v_k]);
    for (__pyx_v_c = 0; __pyx_v_c < __pyx_t_5; __pyx_v_c++) {

      /* "sklearn/tree/_tree.pyx":475
 * 
 *             for c from 0 <= c < n_classes[k]:
 *                 tmp = label_count_left[offset + c]             # <<<<<<<<<<<<<<
 *                 gini_left += tmp * tmp
 *                 tmp = label_count_right[offset + c]
 */
      __pyx_v_tmp = (__pyx_v_label_count_left[(__pyx_v_offset + __pyx_v_c)]);

      /* "sklearn/tree/_tree.pyx":476
 *             for c from 0 <= c < n_classes[k]:
 *                 tmp = label_count_left[offset + c]
 *                 gini_left += tmp * tmp             # <<<<<<<<<<<<<<
 *                 tmp = label_count_right[offset + c]
 *                 gini_right += tmp * tmp
 */
      __pyx_v_gini_left = (__pyx_v_gini_left + (__pyx_v_tmp * __pyx_v_tmp));

      /* "sklearn/tree/_tree.pyx":477
 *                 tmp = label_count_left[offset + c]
 *                 gini_left += tmp * tmp
 *                 tmp = label_count_right[offset + c]             # <<<<<<<<<<<<<<
 *                 gini_right += tmp * tmp
 * 
 */
      __pyx_v_tmp = (__pyx_v_label_count_right[(__pyx_v_offset + __pyx_v_c)]);

      /* "sklearn/tree/_tree.pyx":478
 *                 gini_left += tmp * tmp
 *                 tmp = label_count_right[offset + c]
 *                 gini_right += tmp * tmp             # <<<<<<<<<<<<<<
 * 
 *             gini_left = 1.0 - gini_left / (weighted_n_left *
 */
      __pyx_v_gini_right = (__pyx_v_gini_right + (__pyx_v_tmp * __pyx_v_tmp));
    }

    /* "sklearn/tree/_tree.pyx":481
 * 
 *             gini_left = 1.0 - gini_left / (weighted_n_left *
 *                                            weighted_n_left)             # <<<<<<<<<<<<<<
 *             gini_right = 1.0 - gini_right / (weighted_n_right *
 *                                              weighted_n_right)
 */
    __pyx_v_gini_left = (1.0 - (__pyx_v_gini_left / (__pyx_v_weighted_n_left * __pyx_v_weighted_n_left)));

    /* "sklearn/tree/_tree.pyx":483
 *                                            weighted_n_left)
 *             gini_right = 1.0 - gini_right / (weighted_n_right *
 *                                              weighted_n_right)             # <<<<<<<<<<<<<<
 * 
 *             total += weighted_n_left * gini_left
 */
    __pyx_v_gini_right = (1.0 - (__pyx_v_gini_right / (__pyx_v_weighted_n_right * __pyx_v_weighted_n_right)));

    /* "sklearn/tree/_tree.pyx":485
 *                                              weighted_n_right)
 * 
 *             total += weighted_n_left * gini_left             # <<<<<<<<<<<<<<
 *             total += weighted_n_right * gini_right
 *             offset += label_count_stride
 */
    __pyx_v_total = (__pyx_v_total + (__pyx_v_weighted_n_left * __pyx_v_gini_left));

    /* "sklearn/tree/_tree.pyx":486
 * 
 *             total += weighted_n_left * gini_left
 *             total += weighted_n_right * gini_right             # <<<<<<<<<<<<<<
 *             offset += label_count_stride
 * 
 */
    __pyx_v_total = (__pyx_v_total + (__pyx_v_weighted_n_right * __pyx_v_gini_right));

    /* "sklearn/tree/_tree.pyx":487
 *             total += weighted_n_left * gini_left
 *             total += weighted_n_right * gini_right
 *             offset += label_count_stride             # <<<<<<<<<<<<<<
 * 
 *         return total / (weighted_n_node_samples * n_outputs)
 */
    __pyx_v_offset = (__pyx_v_offset + __pyx_v_label_count_stride);
  }

  /* "sklearn/tree/_tree.pyx":489
 *             offset += label_count_stride
 * 
 *         return total / (weighted_n_node_samples * n_outputs)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = (__pyx_v_total / (__pyx_v_weighted_n_node_samples * __pyx_v_n_outputs));
  goto __pyx_L0;

  __pyx_r = 0;
  __pyx_L0:;
  return __pyx_r;
}

/* Python wrapper */
static int __pyx_pw_7sklearn_4tree_5_tree_19RegressionCriterion_1__cinit__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static int __pyx_pw_7sklearn_4tree_5_tree_19RegressionCriterion_1__cinit__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  __pyx_t_7sklearn_4tree_5_tree_SIZE_t __pyx_v_n_outputs;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__cinit__ (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s__n_outputs,0};
    PyObject* values[1] = {0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s__n_outputs)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "__cinit__") < 0)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 510; __pyx_clineno = __LINE__; goto __pyx_L3_error;}
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 1) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
    }
    __pyx_v_n_outputs = __Pyx_PyInt_from_py_Py_intptr_t(values[0]); if (unlikely((__pyx_v_n_outputs == (npy_intp)-1) && PyErr_Occurred())) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 510; __pyx_clineno = __LINE__; goto __pyx_L3_error;}
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__cinit__", 1, 1, 1, PyTuple_GET_SIZE(__pyx_args)); {__pyx_filename = __pyx_f[0]; __pyx_lineno = 510; __pyx_clineno = __LINE__; goto __pyx_L3_error;}
  __pyx_L3_error:;
  __Pyx_AddTraceback("sklearn.tree._tree.RegressionCriterion.__cinit__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return -1;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_7sklearn_4tree_5_tree_19RegressionCriterion___cinit__(((struct __pyx_obj_7sklearn_4tree_5_tree_RegressionCriterion *)__pyx_v_self), __pyx_v_n_outputs);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "sklearn/tree/_tree.pyx":510
 *     cdef double* var_right
 * 
 *     def __cinit__(self, SIZE_t n_outputs):             # <<<<<<<<<<<<<<
 *         # Default values
 *         self.y = NULL
 */

static int __pyx_pf_7sklearn_4tree_5_tree_19RegressionCriterion___cinit__(struct __pyx_obj_7sklearn_4tree_5_tree_RegressionCriterion *__pyx_v_self, __pyx_t_7sklearn_4tree_5_tree_SIZE_t __pyx_v_n_outputs) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  int __pyx_t_3;
  int __pyx_t_4;
  int __pyx_t_5;
  int __pyx_t_6;
  int __pyx_t_7;
  int __pyx_t_8;
  int __pyx_t_9;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__cinit__", 0);

  /* "sklearn/tree/_tree.pyx":512
 *     def __cinit__(self, SIZE_t n_outputs):
 *         # Default values
 *         self.y = NULL             # <<<<<<<<<<<<<<
 *         self.y_stride = 0
 *         self.sample_weight = NULL
 */
  __pyx_v_self->__pyx_base.y = NULL;

  /* "sklearn/tree/_tree.pyx":513
 *         # Default values
 *         self.y = NULL
 *         self.y_stride = 0             # <<<<<<<<<<<<<<
 *         self.sample_weight = NULL
 * 
 */
  __pyx_v_self->__pyx_base.y_stride = 0;

  /* "sklearn/tree/_tree.pyx":514
 *         self.y = NULL
 *         self.y_stride = 0
 *         self.sample_weight = NULL             # <<<<<<<<<<<<<<
 * 
 *         self.samples = NULL
 */
  __pyx_v_self->__pyx_base.sample_weight = NULL;

  /* "sklearn/tree/_tree.pyx":516
 *         self.sample_weight = NULL
 * 
 *         self.samples = NULL             # <<<<<<<<<<<<<<
 *         self.start = 0
 *         self.pos = 0
 */
  __pyx_v_self->__pyx_base.samples = NULL;

  /* "sklearn/tree/_tree.pyx":517
 * 
 *         self.samples = NULL
 *         self.start = 0             # <<<<<<<<<<<<<<
 *         self.pos = 0
 *         self.end = 0
 */
  __pyx_v_self->__pyx_base.start = 0;

  /* "sklearn/tree/_tree.pyx":518
 *         self.samples = NULL
 *         self.start = 0
 *         self.pos = 0             # <<<<<<<<<<<<<<
 *         self.end = 0
 * 
 */
  __pyx_v_self->__pyx_base.pos = 0;

  /* "sklearn/tree/_tree.pyx":519
 *         self.start = 0
 *         self.pos = 0
 *         self.end = 0             # <<<<<<<<<<<<<<
 * 
 *         self.n_outputs = n_outputs
 */
  __pyx_v_self->__pyx_base.end = 0;

  /* "sklearn/tree/_tree.pyx":521
 *         self.end = 0
 * 
 *         self.n_outputs = n_outputs             # <<<<<<<<<<<<<<
 *         self.n_node_samples = 0
 *         self.weighted_n_node_samples = 0.0
 */
  __pyx_v_self->__pyx_base.n_outputs = __pyx_v_n_outputs;

  /* "sklearn/tree/_tree.pyx":522
 * 
 *         self.n_outputs = n_outputs
 *         self.n_node_samples = 0             # <<<<<<<<<<<<<<
 *         self.weighted_n_node_samples = 0.0
 *         self.weighted_n_left = 0.0
 */
  __pyx_v_self->__pyx_base.n_node_samples = 0;

  /* "sklearn/tree/_tree.pyx":523
 *         self.n_outputs = n_outputs
 *         self.n_node_samples = 0
 *         self.weighted_n_node_samples = 0.0             # <<<<<<<<<<<<<<
 *         self.weighted_n_left = 0.0
 *         self.weighted_n_right = 0.0
 */
  __pyx_v_self->__pyx_base.weighted_n_node_samples = 0.0;

  /* "sklearn/tree/_tree.pyx":524
 *         self.n_node_samples = 0
 *         self.weighted_n_node_samples = 0.0
 *         self.weighted_n_left = 0.0             # <<<<<<<<<<<<<<
 *         self.weighted_n_right = 0.0
 * 
 */
  __pyx_v_self->__pyx_base.weighted_n_left = 0.0;

  /* "sklearn/tree/_tree.pyx":525
 *         self.weighted_n_node_samples = 0.0
 *         self.weighted_n_left = 0.0
 *         self.weighted_n_right = 0.0             # <<<<<<<<<<<<<<
 * 
 *         # Allocate accumulators
 */
  __pyx_v_self->__pyx_base.weighted_n_right = 0.0;

  /* "sklearn/tree/_tree.pyx":528
 * 
 *         # Allocate accumulators
 *         self.mean_left = <double*> calloc(n_outputs, sizeof(double))             # <<<<<<<<<<<<<<
 *         self.mean_right = <double*> calloc(n_outputs, sizeof(double))
 *         self.mean_total = <double*> calloc(n_outputs, sizeof(double))
 */
  __pyx_v_self->mean_left = ((double *)calloc(__pyx_v_n_outputs, (sizeof(double))));

  /* "sklearn/tree/_tree.pyx":529
 *         # Allocate accumulators
 *         self.mean_left = <double*> calloc(n_outputs, sizeof(double))
 *         self.mean_right = <double*> calloc(n_outputs, sizeof(double))             # <<<<<<<<<<<<<<
 *         self.mean_total = <double*> calloc(n_outputs, sizeof(double))
 *         self.sq_sum_left = <double*> calloc(n_outputs, sizeof(double))
 */
  __pyx_v_self->mean_right = ((double *)calloc(__pyx_v_n_outputs, (sizeof(double))));

  /* "sklearn/tree/_tree.pyx":530
 *         self.mean_left = <double*> calloc(n_outputs, sizeof(double))
 *         self.mean_right = <double*> calloc(n_outputs, sizeof(double))
 *         self.mean_total = <double*> calloc(n_outputs, sizeof(double))             # <<<<<<<<<<<<<<
 *         self.sq_sum_left = <double*> calloc(n_outputs, sizeof(double))
 *         self.sq_sum_right = <double*> calloc(n_outputs, sizeof(double))
 */
  __pyx_v_self->mean_total = ((double *)calloc(__pyx_v_n_outputs, (sizeof(double))));

  /* "sklearn/tree/_tree.pyx":531
 *         self.mean_right = <double*> calloc(n_outputs, sizeof(double))
 *         self.mean_total = <double*> calloc(n_outputs, sizeof(double))
 *         self.sq_sum_left = <double*> calloc(n_outputs, sizeof(double))             # <<<<<<<<<<<<<<
 *         self.sq_sum_right = <double*> calloc(n_outputs, sizeof(double))
 *         self.sq_sum_total = <double*> calloc(n_outputs, sizeof(double))
 */
  __pyx_v_self->sq_sum_left = ((double *)calloc(__pyx_v_n_outputs, (sizeof(double))));

  /* "sklearn/tree/_tree.pyx":532
 *         self.mean_total = <double*> calloc(n_outputs, sizeof(double))
 *         self.sq_sum_left = <double*> calloc(n_outputs, sizeof(double))
 *         self.sq_sum_right = <double*> calloc(n_outputs, sizeof(double))             # <<<<<<<<<<<<<<
 *         self.sq_sum_total = <double*> calloc(n_outputs, sizeof(double))
 *         self.var_left = <double*> calloc(n_outputs, sizeof(double))
 */
  __pyx_v_self->sq_sum_right = ((double *)calloc(__pyx_v_n_outputs, (sizeof(double))));

  /* "sklearn/tree/_tree.pyx":533
 *         self.sq_sum_left = <double*> calloc(n_outputs, sizeof(double))
 *         self.sq_sum_right = <double*> calloc(n_outputs, sizeof(double))
 *         self.sq_sum_total = <double*> calloc(n_outputs, sizeof(double))             # <<<<<<<<<<<<<<
 *         self.var_left = <double*> calloc(n_outputs, sizeof(double))
 *         self.var_right = <double*> calloc(n_outputs, sizeof(double))
 */
  __pyx_v_self->sq_sum_total = ((double *)calloc(__pyx_v_n_outputs, (sizeof(double))));

  /* "sklearn/tree/_tree.pyx":534
 *         self.sq_sum_right = <double*> calloc(n_outputs, sizeof(double))
 *         self.sq_sum_total = <double*> calloc(n_outputs, sizeof(double))
 *         self.var_left = <double*> calloc(n_outputs, sizeof(double))             # <<<<<<<<<<<<<<
 *         self.var_right = <double*> calloc(n_outputs, sizeof(double))
 * 
 */
  __pyx_v_self->var_left = ((double *)calloc(__pyx_v_n_outputs, (sizeof(double))));

  /* "sklearn/tree/_tree.pyx":535
 *         self.sq_sum_total = <double*> calloc(n_outputs, sizeof(double))
 *         self.var_left = <double*> calloc(n_outputs, sizeof(double))
 *         self.var_right = <double*> calloc(n_outputs, sizeof(double))             # <<<<<<<<<<<<<<
 * 
 *         # Check for allocation errors
 */
  __pyx_v_self->var_right = ((double *)calloc(__pyx_v_n_outputs, (sizeof(double))));

  /* "sklearn/tree/_tree.pyx":538
 * 
 *         # Check for allocation errors
 *         if ((self.mean_left == NULL) or             # <<<<<<<<<<<<<<
 *             (self.mean_right == NULL) or
 *             (self.mean_total == NULL) or
 */
  __pyx_t_1 = ((__pyx_v_self->mean_left == NULL) != 0);
  if (!__pyx_t_1) {

    /* "sklearn/tree/_tree.pyx":539
 *         # Check for allocation errors
 *         if ((self.mean_left == NULL) or
 *             (self.mean_right == NULL) or             # <<<<<<<<<<<<<<
 *             (self.mean_total == NULL) or
 *             (self.sq_sum_left == NULL) or
 */
    __pyx_t_2 = ((__pyx_v_self->mean_right == NULL) != 0);
    if (!__pyx_t_2) {

      /* "sklearn/tree/_tree.pyx":540
 *         if ((self.mean_left == NULL) or
 *             (self.mean_right == NULL) or
 *             (self.mean_total == NULL) or             # <<<<<<<<<<<<<<
 *             (self.sq_sum_left == NULL) or
 *             (self.sq_sum_right == NULL) or
 */
      __pyx_t_3 = ((__pyx_v_self->mean_total == NULL) != 0);
      if (!__pyx_t_3) {

        /* "sklearn/tree/_tree.pyx":541
 *             (self.mean_right == NULL) or
 *             (self.mean_total == NULL) or
 *             (self.sq_sum_left == NULL) or             # <<<<<<<<<<<<<<
 *             (self.sq_sum_right == NULL) or
 *             (self.sq_sum_total == NULL) or
 */
        __pyx_t_4 = ((__pyx_v_self->sq_sum_left == NULL) != 0);
        if (!__pyx_t_4) {

          /* "sklearn/tree/_tree.pyx":542
 *             (self.mean_total == NULL) or
 *             (self.sq_sum_left == NULL) or
 *             (self.sq_sum_right == NULL) or             # <<<<<<<<<<<<<<
 *             (self.sq_sum_total == NULL) or
 *             (self.var_left == NULL) or
 */
          __pyx_t_5 = ((__pyx_v_self->sq_sum_right == NULL) != 0);
          if (!__pyx_t_5) {

            /* "sklearn/tree/_tree.pyx":543
 *             (self.sq_sum_left == NULL) or
 *             (self.sq_sum_right == NULL) or
 *             (self.sq_sum_total == NULL) or             # <<<<<<<<<<<<<<
 *             (self.var_left == NULL) or
 *             (self.var_right == NULL)):
 */
            __pyx_t_6 = ((__pyx_v_self->sq_sum_total == NULL) != 0);
            if (!__pyx_t_6) {

              /* "sklearn/tree/_tree.pyx":544
 *             (self.sq_sum_right == NULL) or
 *             (self.sq_sum_total == NULL) or
 *             (self.var_left == NULL) or             # <<<<<<<<<<<<<<
 *             (self.var_right == NULL)):
 *             free(self.mean_left)
 */
              __pyx_t_7 = ((__pyx_v_self->var_left == NULL) != 0);
              if (!__pyx_t_7) {

                /* "sklearn/tree/_tree.pyx":545
 *             (self.sq_sum_total == NULL) or
 *             (self.var_left == NULL) or
 *             (self.var_right == NULL)):             # <<<<<<<<<<<<<<
 *             free(self.mean_left)
 *             free(self.mean_right)
 */
                __pyx_t_8 = ((__pyx_v_self->var_right == NULL) != 0);
                __pyx_t_9 = __pyx_t_8;
              } else {
                __pyx_t_9 = __pyx_t_7;
              }
              __pyx_t_7 = __pyx_t_9;
            } else {
              __pyx_t_7 = __pyx_t_6;
            }
            __pyx_t_6 = __pyx_t_7;
          } else {
            __pyx_t_6 = __pyx_t_5;
          }
          __pyx_t_5 = __pyx_t_6;
        } else {
          __pyx_t_5 = __pyx_t_4;
        }
        __pyx_t_4 = __pyx_t_5;
      } else {
        __pyx_t_4 = __pyx_t_3;
      }
      __pyx_t_3 = __pyx_t_4;
    } else {
      __pyx_t_3 = __pyx_t_2;
    }
    __pyx_t_2 = __pyx_t_3;
  } else {
    __pyx_t_2 = __pyx_t_1;
  }
  if (__pyx_t_2) {

    /* "sklearn/tree/_tree.pyx":546
 *             (self.var_left == NULL) or
 *             (self.var_right == NULL)):
 *             free(self.mean_left)             # <<<<<<<<<<<<<<
 *             free(self.mean_right)
 *             free(self.mean_total)
 */
    free(__pyx_v_self->mean_left);

    /* "sklearn/tree/_tree.pyx":547
 *             (self.var_right == NULL)):
 *             free(self.mean_left)
 *             free(self.mean_right)             # <<<<<<<<<<<<<<
 *             free(self.mean_total)
 *             free(self.sq_sum_left)
 */
    free(__pyx_v_self->mean_right);

    /* "sklearn/tree/_tree.pyx":548
 *             free(self.mean_left)
 *             free(self.mean_right)
 *             free(self.mean_total)             # <<<<<<<<<<<<<<
 *             free(self.sq_sum_left)
 *             free(self.sq_sum_right)
 */
    free(__pyx_v_self->mean_total);

    /* "sklearn/tree/_tree.pyx":549
 *             free(self.mean_right)
 *             free(self.mean_total)
 *             free(self.sq_sum_left)             # <<<<<<<<<<<<<<
 *             free(self.sq_sum_right)
 *             free(self.sq_sum_total)
 */
    free(__pyx_v_self->sq_sum_left);

    /* "sklearn/tree/_tree.pyx":550
 *             free(self.mean_total)
 *             free(self.sq_sum_left)
 *             free(self.sq_sum_right)             # <<<<<<<<<<<<<<
 *             free(self.sq_sum_total)
 *             free(self.var_left)
 */
    free(__pyx_v_self->sq_sum_right);

    /* "sklearn/tree/_tree.pyx":551
 *             free(self.sq_sum_left)
 *             free(self.sq_sum_right)
 *             free(self.sq_sum_total)             # <<<<<<<<<<<<<<
 *             free(self.var_left)
 *             free(self.var_right)
 */
    free(__pyx_v_self->sq_sum_total);

    /* "sklearn/tree/_tree.pyx":552
 *             free(self.sq_sum_right)
 *             free(self.sq_sum_total)
 *             free(self.var_left)             # <<<<<<<<<<<<<<
 *             free(self.var_right)
 *             raise MemoryError()
 */
    free(__pyx_v_self->var_left);

    /* "sklearn/tree/_tree.pyx":553
 *             free(self.sq_sum_total)
 *             free(self.var_left)
 *             free(self.var_right)             # <<<<<<<<<<<<<<
 *             raise MemoryError()
 * 
 */
    free(__pyx_v_self->var_right);

    /* "sklearn/tree/_tree.pyx":554
 *             free(self.var_left)
 *             free(self.var_right)
 *             raise MemoryError()             # <<<<<<<<<<<<<<
 * 
 *     def __dealloc__(self):
 */
    PyErr_NoMemory(); {__pyx_filename = __pyx_f[0]; __pyx_lineno = 554; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    goto __pyx_L3;
  }
  __pyx_L3:;

  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_AddTraceback("sklearn.tree._tree.RegressionCriterion.__cinit__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static void __pyx_pw_7sklearn_4tree_5_tree_19RegressionCriterion_3__dealloc__(PyObject *__pyx_v_self); /*proto*/
static void __pyx_pw_7sklearn_4tree_5_tree_19RegressionCriterion_3__dealloc__(PyObject *__pyx_v_self) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__dealloc__ (wrapper)", 0);
  __pyx_pf_7sklearn_4tree_5_tree_19RegressionCriterion_2__dealloc__(((struct __pyx_obj_7sklearn_4tree_5_tree_RegressionCriterion *)__pyx_v_self));
  __Pyx_RefNannyFinishContext();
}

/* "sklearn/tree/_tree.pyx":556
 *             raise MemoryError()
 * 
 *     def __dealloc__(self):             # <<<<<<<<<<<<<<
 *         """Destructor."""
 *         free(self.mean_left)
 */

static void __pyx_pf_7sklearn_4tree_5_tree_19RegressionCriterion_2__dealloc__(struct __pyx_obj_7sklearn_4tree_5_tree_RegressionCriterion *__pyx_v_self) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__dealloc__", 0);

  /* "sklearn/tree/_tree.pyx":558
 *     def __dealloc__(self):
 *         """Destructor."""
 *         free(self.mean_left)             # <<<<<<<<<<<<<<
 *         free(self.mean_right)
 *         free(self.mean_total)
 */
  free(__pyx_v_self->mean_left);

  /* "sklearn/tree/_tree.pyx":559
 *         """Destructor."""
 *         free(self.mean_left)
 *         free(self.mean_right)             # <<<<<<<<<<<<<<
 *         free(self.mean_total)
 *         free(self.sq_sum_left)
 */
  free(__pyx_v_self->mean_right);

  /* "sklearn/tree/_tree.pyx":560
 *         free(self.mean_left)
 *         free(self.mean_right)
 *         free(self.mean_total)             # <<<<<<<<<<<<<<
 *         free(self.sq_sum_left)
 *         free(self.sq_sum_right)
 */
  free(__pyx_v_self->mean_total);

  /* "sklearn/tree/_tree.pyx":561
 *         free(self.mean_right)
 *         free(self.mean_total)
 *         free(self.sq_sum_left)             # <<<<<<<<<<<<<<
 *         free(self.sq_sum_right)
 *         free(self.sq_sum_total)
 */
  free(__pyx_v_self->sq_sum_left);

  /* "sklearn/tree/_tree.pyx":562
 *         free(self.mean_total)
 *         free(self.sq_sum_left)
 *         free(self.sq_sum_right)             # <<<<<<<<<<<<<<
 *         free(self.sq_sum_total)
 *         free(self.var_left)
 */
  free(__pyx_v_self->sq_sum_right);

  /* "sklearn/tree/_tree.pyx":563
 *         free(self.sq_sum_left)
 *         free(self.sq_sum_right)
 *         free(self.sq_sum_total)             # <<<<<<<<<<<<<<
 *         free(self.var_left)
 *         free(self.var_right)
 */
  free(__pyx_v_self->sq_sum_total);

  /* "sklearn/tree/_tree.pyx":564
 *         free(self.sq_sum_right)
 *         free(self.sq_sum_total)
 *         free(self.var_left)             # <<<<<<<<<<<<<<
 *         free(self.var_right)
 * 
 */
  free(__pyx_v_self->var_left);

  /* "sklearn/tree/_tree.pyx":565
 *         free(self.sq_sum_total)
 *         free(self.var_left)
 *         free(self.var_right)             # <<<<<<<<<<<<<<
 * 
 *     def __reduce__(self):
 */
  free(__pyx_v_self->var_right);

  __Pyx_RefNannyFinishContext();
}

/* Python wrapper */
static PyObject *__pyx_pw_7sklearn_4tree_5_tree_19RegressionCriterion_5__reduce__(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static PyObject *__pyx_pw_7sklearn_4tree_5_tree_19RegressionCriterion_5__reduce__(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__reduce__ (wrapper)", 0);
  __pyx_r = __pyx_pf_7sklearn_4tree_5_tree_19RegressionCriterion_4__reduce__(((struct __pyx_obj_7sklearn_4tree_5_tree_RegressionCriterion *)__pyx_v_self));
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "sklearn/tree/_tree.pyx":567
 *         free(self.var_right)
 * 
 *     def __reduce__(self):             # <<<<<<<<<<<<<<
 *         return (RegressionCriterion, (self.n_outputs,), self.__getstate__())
 * 
 */

static PyObject *__pyx_pf_7sklearn_4tree_5_tree_19RegressionCriterion_4__reduce__(struct __pyx_obj_7sklearn_4tree_5_tree_RegressionCriterion *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__reduce__", 0);

  /* "sklearn/tree/_tree.pyx":568
 * 
 *     def __reduce__(self):
 *         return (RegressionCriterion, (self.n_outputs,), self.__getstate__())             # <<<<<<<<<<<<<<
 * 
 *     def __getstate__(self):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyInt_to_py_Py_intptr_t(__pyx_v_self->__pyx_base.n_outputs); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 568; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = PyTuple_New(1); if (unlikely(!__pyx_t_2)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 568; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_2);
  PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s____getstate__); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 568; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_3 = PyObject_Call(__pyx_t_1, ((PyObject *)__pyx_empty_tuple), NULL); if (unlikely(!__pyx_t_3)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 568; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = PyTuple_New(3); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 568; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_INCREF(((PyObject *)((PyObject*)__pyx_ptype_7sklearn_4tree_5_tree_RegressionCriterion)));
  PyTuple_SET_ITEM(__pyx_t_1, 0, ((PyObject *)((PyObject*)__pyx_ptype_7sklearn_4tree_5_tree_RegressionCriterion)));
  __Pyx_GIVEREF(((PyObject *)((PyObject*)__pyx_ptype_7sklearn_4tree_5_tree_RegressionCriterion)));
  PyTuple_SET_ITEM(__pyx_t_1, 1, ((PyObject *)__pyx_t_2));
  __Pyx_GIVEREF(((PyObject *)__pyx_t_2));
  PyTuple_SET_ITEM(__pyx_t_1, 2, __pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_3);
  __pyx_t_2 = 0;
  __pyx_t_3 = 0;
  __pyx_r = ((PyObject *)__pyx_t_1);
  __pyx_t_1 = 0;
  goto __pyx_L0;

  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("sklearn.tree._tree.RegressionCriterion.__reduce__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_7sklearn_4tree_5_tree_19RegressionCriterion_7__getstate__(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static PyObject *__pyx_pw_7sklearn_4tree_5_tree_19RegressionCriterion_7__getstate__(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__getstate__ (wrapper)", 0);
  __pyx_r = __pyx_pf_7sklearn_4tree_5_tree_19RegressionCriterion_6__getstate__(((struct __pyx_obj_7sklearn_4tree_5_tree_RegressionCriterion *)__pyx_v_self));
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "sklearn/tree/_tree.pyx":570
 *         return (RegressionCriterion, (self.n_outputs,), self.__getstate__())
 * 
 *     def __getstate__(self):             # <<<<<<<<<<<<<<
 *         return {}
 * 
 */

static PyObject *__pyx_pf_7sklearn_4tree_5_tree_19RegressionCriterion_6__getstate__(CYTHON_UNUSED struct __pyx_obj_7sklearn_4tree_5_tree_RegressionCriterion *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__getstate__", 0);

  /* "sklearn/tree/_tree.pyx":571
 * 
 *     def __getstate__(self):
 *         return {}             # <<<<<<<<<<<<<<
 * 
 *     def __setstate__(self, d):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = PyDict_New(); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 571; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(((PyObject *)__pyx_t_1));
  __pyx_r = ((PyObject *)__pyx_t_1);
  __pyx_t_1 = 0;
  goto __pyx_L0;

  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("sklearn.tree._tree.RegressionCriterion.__getstate__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_7sklearn_4tree_5_tree_19RegressionCriterion_9__setstate__(PyObject *__pyx_v_self, PyObject *__pyx_v_d); /*proto*/
static PyObject *__pyx_pw_7sklearn_4tree_5_tree_19RegressionCriterion_9__setstate__(PyObject *__pyx_v_self, PyObject *__pyx_v_d) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__setstate__ (wrapper)", 0);
  __pyx_r = __pyx_pf_7sklearn_4tree_5_tree_19RegressionCriterion_8__setstate__(((struct __pyx_obj_7sklearn_4tree_5_tree_RegressionCriterion *)__pyx_v_self), ((PyObject *)__pyx_v_d));
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "sklearn/tree/_tree.pyx":573
 *         return {}
 * 
 *     def __setstate__(self, d):             # <<<<<<<<<<<<<<
 *         pass
 * 
 */

static PyObject *__pyx_pf_7sklearn_4tree_5_tree_19RegressionCriterion_8__setstate__(CYTHON_UNUSED struct __pyx_obj_7sklearn_4tree_5_tree_RegressionCriterion *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v_d) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__setstate__", 0);

  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "sklearn/tree/_tree.pyx":576
 *         pass
 * 
 *     cdef void init(self, DOUBLE_t* y,             # <<<<<<<<<<<<<<
 *                          SIZE_t y_stride,
 *                          DOUBLE_t* sample_weight,
 */

static void __pyx_f_7sklearn_4tree_5_tree_19RegressionCriterion_init(struct __pyx_obj_7sklearn_4tree_5_tree_RegressionCriterion *__pyx_v_self, __pyx_t_7sklearn_4tree_5_tree_DOUBLE_t *__pyx_v_y, __pyx_t_7sklearn_4tree_5_tree_SIZE_t __pyx_v_y_stride, __pyx_t_7sklearn_4tree_5_tree_DOUBLE_t *__pyx_v_sample_weight, __pyx_t_7sklearn_4tree_5_tree_SIZE_t *__pyx_v_samples, __pyx_t_7sklearn_4tree_5_tree_SIZE_t __pyx_v_start, __pyx_t_7sklearn_4tree_5_tree_SIZE_t __pyx_v_end) {
  double __pyx_v_weighted_n_node_samples;
  __pyx_t_7sklearn_4tree_5_tree_SIZE_t __pyx_v_n_outputs;
  double *__pyx_v_mean_left;
  double *__pyx_v_mean_right;
  double *__pyx_v_mean_total;
  double *__pyx_v_sq_sum_left;
  double *__pyx_v_sq_sum_right;
  double *__pyx_v_sq_sum_total;
  double *__pyx_v_var_left;
  double *__pyx_v_var_right;
  __pyx_t_7sklearn_4tree_5_tree_SIZE_t __pyx_v_i;
  __pyx_t_7sklearn_4tree_5_tree_SIZE_t __pyx_v_p;
  __pyx_t_7sklearn_4tree_5_tree_SIZE_t __pyx_v_k;
  __pyx_t_7sklearn_4tree_5_tree_DOUBLE_t __pyx_v_y_ik;
  __pyx_t_7sklearn_4tree_5_tree_DOUBLE_t __pyx_v_w;
  __pyx_t_7sklearn_4tree_5_tree_SIZE_t __pyx_t_1;
  double *__pyx_t_2;
  int __pyx_t_3;
  __pyx_t_7sklearn_4tree_5_tree_SIZE_t __pyx_t_4;
  __pyx_t_7sklearn_4tree_5_tree_SIZE_t __pyx_t_5;

  /* "sklearn/tree/_tree.pyx":585
 *            children samples[start:start] and samples[start:end]."""
 *         # Initialize fields
 *         self.y = y             # <<<<<<<<<<<<<<
 *         self.y_stride = y_stride
 *         self.sample_weight = sample_weight
 */
  __pyx_v_self->__pyx_base.y = __pyx_v_y;

  /* "sklearn/tree/_tree.pyx":586
 *         # Initialize fields
 *         self.y = y
 *         self.y_stride = y_stride             # <<<<<<<<<<<<<<
 *         self.sample_weight = sample_weight
 *         self.samples = samples
 */
  __pyx_v_self->__pyx_base.y_stride = __pyx_v_y_stride;

  /* "sklearn/tree/_tree.pyx":587
 *         self.y = y
 *         self.y_stride = y_stride
 *         self.sample_weight = sample_weight             # <<<<<<<<<<<<<<
 *         self.samples = samples
 *         self.start = start
 */
  __pyx_v_self->__pyx_base.sample_weight = __pyx_v_sample_weight;

  /* "sklearn/tree/_tree.pyx":588
 *         self.y_stride = y_stride
 *         self.sample_weight = sample_weight
 *         self.samples = samples             # <<<<<<<<<<<<<<
 *         self.start = start
 *         self.end = end
 */
  __pyx_v_self->__pyx_base.samples = __pyx_v_samples;

  /* "sklearn/tree/_tree.pyx":589
 *         self.sample_weight = sample_weight
 *         self.samples = samples
 *         self.start = start             # <<<<<<<<<<<<<<
 *         self.end = end
 *         self.n_node_samples = end - start
 */
  __pyx_v_self->__pyx_base.start = __pyx_v_start;

  /* "sklearn/tree/_tree.pyx":590
 *         self.samples = samples
 *         self.start = start
 *         self.end = end             # <<<<<<<<<<<<<<
 *         self.n_node_samples = end - start
 *         cdef double weighted_n_node_samples = 0.
 */
  __pyx_v_self->__pyx_base.end = __pyx_v_end;

  /* "sklearn/tree/_tree.pyx":591
 *         self.start = start
 *         self.end = end
 *         self.n_node_samples = end - start             # <<<<<<<<<<<<<<
 *         cdef double weighted_n_node_samples = 0.
 * 
 */
  __pyx_v_self->__pyx_base.n_node_samples = (__pyx_v_end - __pyx_v_start);

  /* "sklearn/tree/_tree.pyx":592
 *         self.end = end
 *         self.n_node_samples = end - start
 *         cdef double weighted_n_node_samples = 0.             # <<<<<<<<<<<<<<
 * 
 *         # Initialize accumulators
 */
  __pyx_v_weighted_n_node_samples = 0.;

  /* "sklearn/tree/_tree.pyx":595
 * 
 *         # Initialize accumulators
 *         cdef SIZE_t n_outputs = self.n_outputs             # <<<<<<<<<<<<<<
 *         cdef double* mean_left = self.mean_left
 *         cdef double* mean_right = self.mean_right
 */
  __pyx_t_1 = __pyx_v_self->__pyx_base.n_outputs;
  __pyx_v_n_outputs = __pyx_t_1;

  /* "sklearn/tree/_tree.pyx":596
 *         # Initialize accumulators
 *         cdef SIZE_t n_outputs = self.n_outputs
 *         cdef double* mean_left = self.mean_left             # <<<<<<<<<<<<<<
 *         cdef double* mean_right = self.mean_right
 *         cdef double* mean_total = self.mean_total
 */
  __pyx_t_2 = __pyx_v_self->mean_left;
  __pyx_v_mean_left = __pyx_t_2;

  /* "sklearn/tree/_tree.pyx":597
 *         cdef SIZE_t n_outputs = self.n_outputs
 *         cdef double* mean_left = self.mean_left
 *         cdef double* mean_right = self.mean_right             # <<<<<<<<<<<<<<
 *         cdef double* mean_total = self.mean_total
 *         cdef double* sq_sum_left = self.sq_sum_left
 */
  __pyx_t_2 = __pyx_v_self->mean_right;
  __pyx_v_mean_right = __pyx_t_2;

  /* "sklearn/tree/_tree.pyx":598
 *         cdef double* mean_left = self.mean_left
 *         cdef double* mean_right = self.mean_right
 *         cdef double* mean_total = self.mean_total             # <<<<<<<<<<<<<<
 *         cdef double* sq_sum_left = self.sq_sum_left
 *         cdef double* sq_sum_right = self.sq_sum_right
 */
  __pyx_t_2 = __pyx_v_self->mean_total;
  __pyx_v_mean_total = __pyx_t_2;

  /* "sklearn/tree/_tree.pyx":599
 *         cdef double* mean_right = self.mean_right
 *         cdef double* mean_total = self.mean_total
 *         cdef double* sq_sum_left = self.sq_sum_left             # <<<<<<<<<<<<<<
 *         cdef double* sq_sum_right = self.sq_sum_right
 *         cdef double* sq_sum_total = self.sq_sum_total
 */
  __pyx_t_2 = __pyx_v_self->sq_sum_left;
  __pyx_v_sq_sum_left = __pyx_t_2;

  /* "sklearn/tree/_tree.pyx":600
 *         cdef double* mean_total = self.mean_total
 *         cdef double* sq_sum_left = self.sq_sum_left
 *         cdef double* sq_sum_right = self.sq_sum_right             # <<<<<<<<<<<<<<
 *         cdef double* sq_sum_total = self.sq_sum_total
 *         cdef double* var_left = self.var_left
 */
  __pyx_t_2 = __pyx_v_self->sq_sum_right;
  __pyx_v_sq_sum_right = __pyx_t_2;

  /* "sklearn/tree/_tree.pyx":601
 *         cdef double* sq_sum_left = self.sq_sum_left
 *         cdef double* sq_sum_right = self.sq_sum_right
 *         cdef double* sq_sum_total = self.sq_sum_total             # <<<<<<<<<<<<<<
 *         cdef double* var_left = self.var_left
 *         cdef double* var_right = self.var_right
 */
  __pyx_t_2 = __pyx_v_self->sq_sum_total;
  __pyx_v_sq_sum_total = __pyx_t_2;

  /* "sklearn/tree/_tree.pyx":602
 *         cdef double* sq_sum_right = self.sq_sum_right
 *         cdef double* sq_sum_total = self.sq_sum_total
 *         cdef double* var_left = self.var_left             # <<<<<<<<<<<<<<
 *         cdef double* var_right = self.var_right
 * 
 */
  __pyx_t_2 = __pyx_v_self->var_left;
  __pyx_v_var_left = __pyx_t_2;

  /* "sklearn/tree/_tree.pyx":603
 *         cdef double* sq_sum_total = self.sq_sum_total
 *         cdef double* var_left = self.var_left
 *         cdef double* var_right = self.var_right             # <<<<<<<<<<<<<<
 * 
 *         cdef SIZE_t i = 0
 */
  __pyx_t_2 = __pyx_v_self->var_right;
  __pyx_v_var_right = __pyx_t_2;

  /* "sklearn/tree/_tree.pyx":605
 *         cdef double* var_right = self.var_right
 * 
 *         cdef SIZE_t i = 0             # <<<<<<<<<<<<<<
 *         cdef SIZE_t p = 0
 *         cdef SIZE_t k = 0
 */
  __pyx_v_i = 0;

  /* "sklearn/tree/_tree.pyx":606
 * 
 *         cdef SIZE_t i = 0
 *         cdef SIZE_t p = 0             # <<<<<<<<<<<<<<
 *         cdef SIZE_t k = 0
 *         cdef DOUBLE_t y_ik = 0.0
 */
  __pyx_v_p = 0;

  /* "sklearn/tree/_tree.pyx":607
 *         cdef SIZE_t i = 0
 *         cdef SIZE_t p = 0
 *         cdef SIZE_t k = 0             # <<<<<<<<<<<<<<
 *         cdef DOUBLE_t y_ik = 0.0
 *         cdef DOUBLE_t w = 1.0
 */
  __pyx_v_k = 0;

  /* "sklearn/tree/_tree.pyx":608
 *         cdef SIZE_t p = 0
 *         cdef SIZE_t k = 0
 *         cdef DOUBLE_t y_ik = 0.0             # <<<<<<<<<<<<<<
 *         cdef DOUBLE_t w = 1.0
 * 
 */
  __pyx_v_y_ik = 0.0;

  /* "sklearn/tree/_tree.pyx":609
 *         cdef SIZE_t k = 0
 *         cdef DOUBLE_t y_ik = 0.0
 *         cdef DOUBLE_t w = 1.0             # <<<<<<<<<<<<<<
 * 
 *         for k from 0 <= k < n_outputs:
 */
  __pyx_v_w = 1.0;

  /* "sklearn/tree/_tree.pyx":611
 *         cdef DOUBLE_t w = 1.0
 * 
 *         for k from 0 <= k < n_outputs:             # <<<<<<<<<<<<<<
 *             mean_left[k] = 0.0
 *             mean_right[k] = 0.0
 */
  __pyx_t_1 = __pyx_v_n_outputs;
  for (__pyx_v_k = 0; __pyx_v_k < __pyx_t_1; __pyx_v_k++) {

    /* "sklearn/tree/_tree.pyx":612
 * 
 *         for k from 0 <= k < n_outputs:
 *             mean_left[k] = 0.0             # <<<<<<<<<<<<<<
 *             mean_right[k] = 0.0
 *             mean_total[k] = 0.0
 */
    (__pyx_v_mean_left[__pyx_v_k]) = 0.0;

    /* "sklearn/tree/_tree.pyx":613
 *         for k from 0 <= k < n_outputs:
 *             mean_left[k] = 0.0
 *             mean_right[k] = 0.0             # <<<<<<<<<<<<<<
 *             mean_total[k] = 0.0
 *             sq_sum_right[k] = 0.0
 */
    (__pyx_v_mean_right[__pyx_v_k]) = 0.0;

    /* "sklearn/tree/_tree.pyx":614
 *             mean_left[k] = 0.0
 *             mean_right[k] = 0.0
 *             mean_total[k] = 0.0             # <<<<<<<<<<<<<<
 *             sq_sum_right[k] = 0.0
 *             sq_sum_left[k] = 0.0
 */
    (__pyx_v_mean_total[__pyx_v_k]) = 0.0;

    /* "sklearn/tree/_tree.pyx":615
 *             mean_right[k] = 0.0
 *             mean_total[k] = 0.0
 *             sq_sum_right[k] = 0.0             # <<<<<<<<<<<<<<
 *             sq_sum_left[k] = 0.0
 *             sq_sum_total[k] = 0.0
 */
    (__pyx_v_sq_sum_right[__pyx_v_k]) = 0.0;

    /* "sklearn/tree/_tree.pyx":616
 *             mean_total[k] = 0.0
 *             sq_sum_right[k] = 0.0
 *             sq_sum_left[k] = 0.0             # <<<<<<<<<<<<<<
 *             sq_sum_total[k] = 0.0
 *             var_left[k] = 0.0
 */
    (__pyx_v_sq_sum_left[__pyx_v_k]) = 0.0;

    /* "sklearn/tree/_tree.pyx":617
 *             sq_sum_right[k] = 0.0
 *             sq_sum_left[k] = 0.0
 *             sq_sum_total[k] = 0.0             # <<<<<<<<<<<<<<
 *             var_left[k] = 0.0
 *             var_right[k] = 0.0
 */
    (__pyx_v_sq_sum_total[__pyx_v_k]) = 0.0;

    /* "sklearn/tree/_tree.pyx":618
 *             sq_sum_left[k] = 0.0
 *             sq_sum_total[k] = 0.0
 *             var_left[k] = 0.0             # <<<<<<<<<<<<<<
 *             var_right[k] = 0.0
 * 
 */
    (__pyx_v_var_left[__pyx_v_k]) = 0.0;

    /* "sklearn/tree/_tree.pyx":619
 *             sq_sum_total[k] = 0.0
 *             var_left[k] = 0.0
 *             var_right[k] = 0.0             # <<<<<<<<<<<<<<
 * 
 *         for p from start <= p < end:
 */
    (__pyx_v_var_right[__pyx_v_k]) = 0.0;
  }

  /* "sklearn/tree/_tree.pyx":621
 *             var_right[k] = 0.0
 * 
 *         for p from start <= p < end:             # <<<<<<<<<<<<<<
 *             i = samples[p]
 * 
 */
  __pyx_t_1 = __pyx_v_end;
  for (__pyx_v_p = __pyx_v_start; __pyx_v_p < __pyx_t_1; __pyx_v_p++) {

    /* "sklearn/tree/_tree.pyx":622
 * 
 *         for p from start <= p < end:
 *             i = samples[p]             # <<<<<<<<<<<<<<
 * 
 *             if sample_weight != NULL:
 */
    __pyx_v_i = (__pyx_v_samples[__pyx_v_p]);

    /* "sklearn/tree/_tree.pyx":624
 *             i = samples[p]
 * 
 *             if sample_weight != NULL:             # <<<<<<<<<<<<<<
 *                 w = sample_weight[i]
 * 
 */
    __pyx_t_3 = ((__pyx_v_sample_weight != NULL) != 0);
    if (__pyx_t_3) {

      /* "sklearn/tree/_tree.pyx":625
 * 
 *             if sample_weight != NULL:
 *                 w = sample_weight[i]             # <<<<<<<<<<<<<<
 * 
 *             for k from 0 <= k < n_outputs:
 */
      __pyx_v_w = (__pyx_v_sample_weight[__pyx_v_i]);
      goto __pyx_L7;
    }
    __pyx_L7:;

    /* "sklearn/tree/_tree.pyx":627
 *                 w = sample_weight[i]
 * 
 *             for k from 0 <= k < n_outputs:             # <<<<<<<<<<<<<<
 *                 y_ik = y[i * y_stride + k]
 *                 sq_sum_total[k] += w * y_ik * y_ik
 */
    __pyx_t_4 = __pyx_v_n_outputs;
    for (__pyx_v_k = 0; __pyx_v_k < __pyx_t_4; __pyx_v_k++) {

      /* "sklearn/tree/_tree.pyx":628
 * 
 *             for k from 0 <= k < n_outputs:
 *                 y_ik = y[i * y_stride + k]             # <<<<<<<<<<<<<<
 *                 sq_sum_total[k] += w * y_ik * y_ik
 *                 mean_total[k] += w * y_ik
 */
      __pyx_v_y_ik = (__pyx_v_y[((__pyx_v_i * __pyx_v_y_stride) + __pyx_v_k)]);

      /* "sklearn/tree/_tree.pyx":629
 *             for k from 0 <= k < n_outputs:
 *                 y_ik = y[i * y_stride + k]
 *                 sq_sum_total[k] += w * y_ik * y_ik             # <<<<<<<<<<<<<<
 *                 mean_total[k] += w * y_ik
 * 
 */
      __pyx_t_5 = __pyx_v_k;
      (__pyx_v_sq_sum_total[__pyx_t_5]) = ((__pyx_v_sq_sum_total[__pyx_t_5]) + ((__pyx_v_w * __pyx_v_y_ik) * __pyx_v_y_ik));

      /* "sklearn/tree/_tree.pyx":630
 *                 y_ik = y[i * y_stride + k]
 *                 sq_sum_total[k] += w * y_ik * y_ik
 *                 mean_total[k] += w * y_ik             # <<<<<<<<<<<<<<
 * 
 *             weighted_n_node_samples += w
 */
      __pyx_t_5 = __pyx_v_k;
      (__pyx_v_mean_total[__pyx_t_5]) = ((__pyx_v_mean_total[__pyx_t_5]) + (__pyx_v_w * __pyx_v_y_ik));
    }

    /* "sklearn/tree/_tree.pyx":632
 *                 mean_total[k] += w * y_ik
 * 
 *             weighted_n_node_samples += w             # <<<<<<<<<<<<<<
 * 
 *         self.weighted_n_node_samples = weighted_n_node_samples
 */
    __pyx_v_weighted_n_node_samples = (__pyx_v_weighted_n_node_samples + __pyx_v_w);
  }

  /* "sklearn/tree/_tree.pyx":634
 *             weighted_n_node_samples += w
 * 
 *         self.weighted_n_node_samples = weighted_n_node_samples             # <<<<<<<<<<<<<<
 * 
 *         for k from 0 <= k < n_outputs:
 */
  __pyx_v_self->__pyx_base.weighted_n_node_samples = __pyx_v_weighted_n_node_samples;

  /* "sklearn/tree/_tree.pyx":636
 *         self.weighted_n_node_samples = weighted_n_node_samples
 * 
 *         for k from 0 <= k < n_outputs:             # <<<<<<<<<<<<<<
 *             mean_total[k] /= weighted_n_node_samples
 * 
 */
  __pyx_t_1 = __pyx_v_n_outputs;
  for (__pyx_v_k = 0; __pyx_v_k < __pyx_t_1; __pyx_v_k++) {

    /* "sklearn/tree/_tree.pyx":637
 * 
 *         for k from 0 <= k < n_outputs:
 *             mean_total[k] /= weighted_n_node_samples             # <<<<<<<<<<<<<<
 * 
 *         # Reset to pos=start
 */
    __pyx_t_4 = __pyx_v_k;
    (__pyx_v_mean_total[__pyx_t_4]) = ((__pyx_v_mean_total[__pyx_t_4]) / __pyx_v_weighted_n_node_samples);
  }

  /* "sklearn/tree/_tree.pyx":640
 * 
 *         # Reset to pos=start
 *         self.reset()             # <<<<<<<<<<<<<<
 * 
 *     cdef void reset(self) nogil:
 */
  ((struct __pyx_vtabstruct_7sklearn_4tree_5_tree_RegressionCriterion *)__pyx_v_self->__pyx_base.__pyx_vtab)->__pyx_base.reset(((struct __pyx_obj_7sklearn_4tree_5_tree_Criterion *)__pyx_v_self));

}

/* "sklearn/tree/_tree.pyx":642
 *         self.reset()
 * 
 *     cdef void reset(self) nogil:             # <<<<<<<<<<<<<<
 *         """Reset the criterion at pos=start."""
 *         self.pos = self.start
 */

static void __pyx_f_7sklearn_4tree_5_tree_19RegressionCriterion_reset(struct __pyx_obj_7sklearn_4tree_5_tree_RegressionCriterion *__pyx_v_self) {
  __pyx_t_7sklearn_4tree_5_tree_SIZE_t __pyx_v_n_outputs;
  double *__pyx_v_mean_left;
  double *__pyx_v_mean_right;
  double *__pyx_v_mean_total;
  double *__pyx_v_sq_sum_left;
  double *__pyx_v_sq_sum_right;
  double *__pyx_v_sq_sum_total;
  double *__pyx_v_var_left;
  double *__pyx_v_var_right;
  double __pyx_v_weighted_n_node_samples;
  __pyx_t_7sklearn_4tree_5_tree_SIZE_t __pyx_v_k;
  __pyx_t_7sklearn_4tree_5_tree_SIZE_t __pyx_t_1;
  double __pyx_t_2;
  double *__pyx_t_3;

  /* "sklearn/tree/_tree.pyx":644
 *     cdef void reset(self) nogil:
 *         """Reset the criterion at pos=start."""
 *         self.pos = self.start             # <<<<<<<<<<<<<<
 * 
 *         self.weighted_n_left = 0.0
 */
  __pyx_t_1 = __pyx_v_self->__pyx_base.start;
  __pyx_v_self->__pyx_base.pos = __pyx_t_1;

  /* "sklearn/tree/_tree.pyx":646
 *         self.pos = self.start
 * 
 *         self.weighted_n_left = 0.0             # <<<<<<<<<<<<<<
 *         self.weighted_n_right = self.weighted_n_node_samples
 * 
 */
  __pyx_v_self->__pyx_base.weighted_n_left = 0.0;

  /* "sklearn/tree/_tree.pyx":647
 * 
 *         self.weighted_n_left = 0.0
 *         self.weighted_n_right = self.weighted_n_node_samples             # <<<<<<<<<<<<<<
 * 
 *         cdef SIZE_t n_outputs = self.n_outputs
 */
  __pyx_t_2 = __pyx_v_self->__pyx_base.weighted_n_node_samples;
  __pyx_v_self->__pyx_base.weighted_n_right = __pyx_t_2;

  /* "sklearn/tree/_tree.pyx":649
 *         self.weighted_n_right = self.weighted_n_node_samples
 * 
 *         cdef SIZE_t n_outputs = self.n_outputs             # <<<<<<<<<<<<<<
 *         cdef double* mean_left = self.mean_left
 *         cdef double* mean_right = self.mean_right
 */
  __pyx_t_1 = __pyx_v_self->__pyx_base.n_outputs;
  __pyx_v_n_outputs = __pyx_t_1;

  /* "sklearn/tree/_tree.pyx":650
 * 
 *         cdef SIZE_t n_outputs = self.n_outputs
 *         cdef double* mean_left = self.mean_left             # <<<<<<<<<<<<<<
 *         cdef double* mean_right = self.mean_right
 *         cdef double* mean_total = self.mean_total
 */
  __pyx_t_3 = __pyx_v_self->mean_left;
  __pyx_v_mean_left = __pyx_t_3;

  /* "sklearn/tree/_tree.pyx":651
 *         cdef SIZE_t n_outputs = self.n_outputs
 *         cdef double* mean_left = self.mean_left
 *         cdef double* mean_right = self.mean_right             # <<<<<<<<<<<<<<
 *         cdef double* mean_total = self.mean_total
 *         cdef double* sq_sum_left = self.sq_sum_left
 */
  __pyx_t_3 = __pyx_v_self->mean_right;
  __pyx_v_mean_right = __pyx_t_3;

  /* "sklearn/tree/_tree.pyx":652
 *         cdef double* mean_left = self.mean_left
 *         cdef double* mean_right = self.mean_right
 *         cdef double* mean_total = self.mean_total             # <<<<<<<<<<<<<<
 *         cdef double* sq_sum_left = self.sq_sum_left
 *         cdef double* sq_sum_right = self.sq_sum_right
 */
  __pyx_t_3 = __pyx_v_self->mean_total;
  __pyx_v_mean_total = __pyx_t_3;

  /* "sklearn/tree/_tree.pyx":653
 *         cdef double* mean_right = self.mean_right
 *         cdef double* mean_total = self.mean_total
 *         cdef double* sq_sum_left = self.sq_sum_left             # <<<<<<<<<<<<<<
 *         cdef double* sq_sum_right = self.sq_sum_right
 *         cdef double* sq_sum_total = self.sq_sum_total
 */
  __pyx_t_3 = __pyx_v_self->sq_sum_left;
  __pyx_v_sq_sum_left = __pyx_t_3;

  /* "sklearn/tree/_tree.pyx":654
 *         cdef double* mean_total = self.mean_total
 *         cdef double* sq_sum_left = self.sq_sum_left
 *         cdef double* sq_sum_right = self.sq_sum_right             # <<<<<<<<<<<<<<
 *         cdef double* sq_sum_total = self.sq_sum_total
 *         cdef double* var_left = self.var_left
 */
  __pyx_t_3 = __pyx_v_self->sq_sum_right;
  __pyx_v_sq_sum_right = __pyx_t_3;

  /* "sklearn/tree/_tree.pyx":655
 *         cdef double* sq_sum_left = self.sq_sum_left
 *         cdef double* sq_sum_right = self.sq_sum_right
 *         cdef double* sq_sum_total = self.sq_sum_total             # <<<<<<<<<<<<<<
 *         cdef double* var_left = self.var_left
 *         cdef double* var_right = self.var_right
 */
  __pyx_t_3 = __pyx_v_self->sq_sum_total;
  __pyx_v_sq_sum_total = __pyx_t_3;

  /* "sklearn/tree/_tree.pyx":656
 *         cdef double* sq_sum_right = self.sq_sum_right
 *         cdef double* sq_sum_total = self.sq_sum_total
 *         cdef double* var_left = self.var_left             # <<<<<<<<<<<<<<
 *         cdef double* var_right = self.var_right
 *         cdef double weighted_n_node_samples = self.weighted_n_node_samples
 */
  __pyx_t_3 = __pyx_v_self->var_left;
  __pyx_v_var_left = __pyx_t_3;

  /* "sklearn/tree/_tree.pyx":657
 *         cdef double* sq_sum_total = self.sq_sum_total
 *         cdef double* var_left = self.var_left
 *         cdef double* var_right = self.var_right             # <<<<<<<<<<<<<<
 *         cdef double weighted_n_node_samples = self.weighted_n_node_samples
 * 
 */
  __pyx_t_3 = __pyx_v_self->var_right;
  __pyx_v_var_right = __pyx_t_3;

  /* "sklearn/tree/_tree.pyx":658
 *         cdef double* var_left = self.var_left
 *         cdef double* var_right = self.var_right
 *         cdef double weighted_n_node_samples = self.weighted_n_node_samples             # <<<<<<<<<<<<<<
 * 
 *         cdef SIZE_t k = 0
 */
  __pyx_t_2 = __pyx_v_self->__pyx_base.weighted_n_node_samples;
  __pyx_v_weighted_n_node_samples = __pyx_t_2;

  /* "sklearn/tree/_tree.pyx":660
 *         cdef double weighted_n_node_samples = self.weighted_n_node_samples
 * 
 *         cdef SIZE_t k = 0             # <<<<<<<<<<<<<<
 * 
 *         for k from 0 <= k < n_outputs:
 */
  __pyx_v_k = 0;

  /* "sklearn/tree/_tree.pyx":662
 *         cdef SIZE_t k = 0
 * 
 *         for k from 0 <= k < n_outputs:             # <<<<<<<<<<<<<<
 *             mean_right[k] = mean_total[k]
 *             mean_left[k] = 0.0
 */
  __pyx_t_1 = __pyx_v_n_outputs;
  for (__pyx_v_k = 0; __pyx_v_k < __pyx_t_1; __pyx_v_k++) {

    /* "sklearn/tree/_tree.pyx":663
 * 
 *         for k from 0 <= k < n_outputs:
 *             mean_right[k] = mean_total[k]             # <<<<<<<<<<<<<<
 *             mean_left[k] = 0.0
 *             sq_sum_right[k] = sq_sum_total[k]
 */
    (__pyx_v_mean_right[__pyx_v_k]) = (__pyx_v_mean_total[__pyx_v_k]);

    /* "sklearn/tree/_tree.pyx":664
 *         for k from 0 <= k < n_outputs:
 *             mean_right[k] = mean_total[k]
 *             mean_left[k] = 0.0             # <<<<<<<<<<<<<<
 *             sq_sum_right[k] = sq_sum_total[k]
 *             sq_sum_left[k] = 0.0
 */
    (__pyx_v_mean_left[__pyx_v_k]) = 0.0;

    /* "sklearn/tree/_tree.pyx":665
 *             mean_right[k] = mean_total[k]
 *             mean_left[k] = 0.0
 *             sq_sum_right[k] = sq_sum_total[k]             # <<<<<<<<<<<<<<
 *             sq_sum_left[k] = 0.0
 *             var_left[k] = 0.0
 */
    (__pyx_v_sq_sum_right[__pyx_v_k]) = (__pyx_v_sq_sum_total[__pyx_v_k]);

    /* "sklearn/tree/_tree.pyx":666
 *             mean_left[k] = 0.0
 *             sq_sum_right[k] = sq_sum_total[k]
 *             sq_sum_left[k] = 0.0             # <<<<<<<<<<<<<<
 *             var_left[k] = 0.0
 *             var_right[k] = (sq_sum_right[k] -
 */
    (__pyx_v_sq_sum_left[__pyx_v_k]) = 0.0;

    /* "sklearn/tree/_tree.pyx":667
 *             sq_sum_right[k] = sq_sum_total[k]
 *             sq_sum_left[k] = 0.0
 *             var_left[k] = 0.0             # <<<<<<<<<<<<<<
 *             var_right[k] = (sq_sum_right[k] -
 *                             weighted_n_node_samples * (mean_right[k] *
 */
    (__pyx_v_var_left[__pyx_v_k]) = 0.0;

    /* "sklearn/tree/_tree.pyx":668
 *             sq_sum_left[k] = 0.0
 *             var_left[k] = 0.0
 *             var_right[k] = (sq_sum_right[k] -             # <<<<<<<<<<<<<<
 *                             weighted_n_node_samples * (mean_right[k] *
 *                                                        mean_right[k]))
 */
    (__pyx_v_var_right[__pyx_v_k]) = ((__pyx_v_sq_sum_right[__pyx_v_k]) - (__pyx_v_weighted_n_node_samples * ((__pyx_v_mean_right[__pyx_v_k]) * (__pyx_v_mean_right[__pyx_v_k]))));
  }

}

/* "sklearn/tree/_tree.pyx":672
 *                                                        mean_right[k]))
 * 
 *     cdef void update(self, SIZE_t new_pos) nogil:             # <<<<<<<<<<<<<<
 *         """Update the collected statistics by moving samples[pos:new_pos] from
 *            the right child to the left child."""
 */

static void __pyx_f_7sklearn_4tree_5_tree_19RegressionCriterion_update(struct __pyx_obj_7sklearn_4tree_5_tree_RegressionCriterion *__pyx_v_self, __pyx_t_7sklearn_4tree_5_tree_SIZE_t __pyx_v_new_pos) {
  __pyx_t_7sklearn_4tree_5_tree_DOUBLE_t *__pyx_v_y;
  __pyx_t_7sklearn_4tree_5_tree_SIZE_t __pyx_v_y_stride;
  __pyx_t_7sklearn_4tree_5_tree_DOUBLE_t *__pyx_v_sample_weight;
  __pyx_t_7sklearn_4tree_5_tree_SIZE_t *__pyx_v_samples;
  __pyx_t_7sklearn_4tree_5_tree_SIZE_t __pyx_v_pos;
  __pyx_t_7sklearn_4tree_5_tree_SIZE_t __pyx_v_n_outputs;
  double *__pyx_v_mean_left;
  double *__pyx_v_mean_right;
  double *__pyx_v_sq_sum_left;
  double *__pyx_v_sq_sum_right;
  double *__pyx_v_var_left;
  double *__pyx_v_var_right;
  double __pyx_v_weighted_n_left;
  double __pyx_v_weighted_n_right;
  __pyx_t_7sklearn_4tree_5_tree_SIZE_t __pyx_v_i;
  __pyx_t_7sklearn_4tree_5_tree_SIZE_t __pyx_v_p;
  __pyx_t_7sklearn_4tree_5_tree_SIZE_t __pyx_v_k;
  __pyx_t_7sklearn_4tree_5_tree_DOUBLE_t __pyx_v_w;
  __pyx_t_7sklearn_4tree_5_tree_DOUBLE_t __pyx_v_y_ik;
  __pyx_t_7sklearn_4tree_5_tree_DOUBLE_t __pyx_v_w_y_ik;
  __pyx_t_7sklearn_4tree_5_tree_DOUBLE_t *__pyx_t_1;
  __pyx_t_7sklearn_4tree_5_tree_SIZE_t __pyx_t_2;
  __pyx_t_7sklearn_4tree_5_tree_SIZE_t *__pyx_t_3;
  double *__pyx_t_4;
  double __pyx_t_5;
  int __pyx_t_6;
  __pyx_t_7sklearn_4tree_5_tree_SIZE_t __pyx_t_7;
  __pyx_t_7sklearn_4tree_5_tree_SIZE_t __pyx_t_8;

  /* "sklearn/tree/_tree.pyx":675
 *         """Update the collected statistics by moving samples[pos:new_pos] from
 *            the right child to the left child."""
 *         cdef DOUBLE_t* y = self.y             # <<<<<<<<<<<<<<
 *         cdef SIZE_t y_stride = self.y_stride
 *         cdef DOUBLE_t* sample_weight = self.sample_weight
 */
  __pyx_t_1 = __pyx_v_self->__pyx_base.y;
  __pyx_v_y = __pyx_t_1;

  /* "sklearn/tree/_tree.pyx":676
 *            the right child to the left child."""
 *         cdef DOUBLE_t* y = self.y
 *         cdef SIZE_t y_stride = self.y_stride             # <<<<<<<<<<<<<<
 *         cdef DOUBLE_t* sample_weight = self.sample_weight
 * 
 */
  __pyx_t_2 = __pyx_v_self->__pyx_base.y_stride;
  __pyx_v_y_stride = __pyx_t_2;

  /* "sklearn/tree/_tree.pyx":677
 *         cdef DOUBLE_t* y = self.y
 *         cdef SIZE_t y_stride = self.y_stride
 *         cdef DOUBLE_t* sample_weight = self.sample_weight             # <<<<<<<<<<<<<<
 * 
 *         cdef SIZE_t* samples = self.samples
 */
  __pyx_t_1 = __pyx_v_self->__pyx_base.sample_weight;
  __pyx_v_sample_weight = __pyx_t_1;

  /* "sklearn/tree/_tree.pyx":679
 *         cdef DOUBLE_t* sample_weight = self.sample_weight
 * 
 *         cdef SIZE_t* samples = self.samples             # <<<<<<<<<<<<<<
 *         cdef SIZE_t pos = self.pos
 * 
 */
  __pyx_t_3 = __pyx_v_self->__pyx_base.samples;
  __pyx_v_samples = __pyx_t_3;

  /* "sklearn/tree/_tree.pyx":680
 * 
 *         cdef SIZE_t* samples = self.samples
 *         cdef SIZE_t pos = self.pos             # <<<<<<<<<<<<<<
 * 
 *         cdef SIZE_t n_outputs = self.n_outputs
 */
  __pyx_t_2 = __pyx_v_self->__pyx_base.pos;
  __pyx_v_pos = __pyx_t_2;

  /* "sklearn/tree/_tree.pyx":682
 *         cdef SIZE_t pos = self.pos
 * 
 *         cdef SIZE_t n_outputs = self.n_outputs             # <<<<<<<<<<<<<<
 *         cdef double* mean_left = self.mean_left
 *         cdef double* mean_right = self.mean_right
 */
  __pyx_t_2 = __pyx_v_self->__pyx_base.n_outputs;
  __pyx_v_n_outputs = __pyx_t_2;

  /* "sklearn/tree/_tree.pyx":683
 * 
 *         cdef SIZE_t n_outputs = self.n_outputs
 *         cdef double* mean_left = self.mean_left             # <<<<<<<<<<<<<<
 *         cdef double* mean_right = self.mean_right
 *         cdef double* sq_sum_left = self.sq_sum_left
 */
  __pyx_t_4 = __pyx_v_self->mean_left;
  __pyx_v_mean_left = __pyx_t_4;

  /* "sklearn/tree/_tree.pyx":684
 *         cdef SIZE_t n_outputs = self.n_outputs
 *         cdef double* mean_left = self.mean_left
 *         cdef double* mean_right = self.mean_right             # <<<<<<<<<<<<<<
 *         cdef double* sq_sum_left = self.sq_sum_left
 *         cdef double* sq_sum_right = self.sq_sum_right
 */
  __pyx_t_4 = __pyx_v_self->mean_right;
  __pyx_v_mean_right = __pyx_t_4;

  /* "sklearn/tree/_tree.pyx":685
 *         cdef double* mean_left = self.mean_left
 *         cdef double* mean_right = self.mean_right
 *         cdef double* sq_sum_left = self.sq_sum_left             # <<<<<<<<<<<<<<
 *         cdef double* sq_sum_right = self.sq_sum_right
 *         cdef double* var_left = self.var_left
 */
  __pyx_t_4 = __pyx_v_self->sq_sum_left;
  __pyx_v_sq_sum_left = __pyx_t_4;

  /* "sklearn/tree/_tree.pyx":686
 *         cdef double* mean_right = self.mean_right
 *         cdef double* sq_sum_left = self.sq_sum_left
 *         cdef double* sq_sum_right = self.sq_sum_right             # <<<<<<<<<<<<<<
 *         cdef double* var_left = self.var_left
 *         cdef double* var_right = self.var_right
 */
  __pyx_t_4 = __pyx_v_self->sq_sum_right;
  __pyx_v_sq_sum_right = __pyx_t_4;

  /* "sklearn/tree/_tree.pyx":687
 *         cdef double* sq_sum_left = self.sq_sum_left
 *         cdef double* sq_sum_right = self.sq_sum_right
 *         cdef double* var_left = self.var_left             # <<<<<<<<<<<<<<
 *         cdef double* var_right = self.var_right
 * 
 */
  __pyx_t_4 = __pyx_v_self->var_left;
  __pyx_v_var_left = __pyx_t_4;

  /* "sklearn/tree/_tree.pyx":688
 *         cdef double* sq_sum_right = self.sq_sum_right
 *         cdef double* var_left = self.var_left
 *         cdef double* var_right = self.var_right             # <<<<<<<<<<<<<<
 * 
 *         cdef double weighted_n_left = self.weighted_n_left
 */
  __pyx_t_4 = __pyx_v_self->var_right;
  __pyx_v_var_right = __pyx_t_4;

  /* "sklearn/tree/_tree.pyx":690
 *         cdef double* var_right = self.var_right
 * 
 *         cdef double weighted_n_left = self.weighted_n_left             # <<<<<<<<<<<<<<
 *         cdef double weighted_n_right = self.weighted_n_right
 * 
 */
  __pyx_t_5 = __pyx_v_self->__pyx_base.weighted_n_left;
  __pyx_v_weighted_n_left = __pyx_t_5;

  /* "sklearn/tree/_tree.pyx":691
 * 
 *         cdef double weighted_n_left = self.weighted_n_left
 *         cdef double weighted_n_right = self.weighted_n_right             # <<<<<<<<<<<<<<
 * 
 *         cdef SIZE_t i
 */
  __pyx_t_5 = __pyx_v_self->__pyx_base.weighted_n_right;
  __pyx_v_weighted_n_right = __pyx_t_5;

  /* "sklearn/tree/_tree.pyx":696
 *         cdef SIZE_t p
 *         cdef SIZE_t k
 *         cdef DOUBLE_t w = 1.0             # <<<<<<<<<<<<<<
 *         cdef DOUBLE_t y_ik, w_y_ik
 * 
 */
  __pyx_v_w = 1.0;

  /* "sklearn/tree/_tree.pyx":701
 *         # Note: We assume start <= pos < new_pos <= end
 * 
 *         for p from pos <= p < new_pos:             # <<<<<<<<<<<<<<
 *             i = samples[p]
 * 
 */
  __pyx_t_2 = __pyx_v_new_pos;
  for (__pyx_v_p = __pyx_v_pos; __pyx_v_p < __pyx_t_2; __pyx_v_p++) {

    /* "sklearn/tree/_tree.pyx":702
 * 
 *         for p from pos <= p < new_pos:
 *             i = samples[p]             # <<<<<<<<<<<<<<
 * 
 *             if sample_weight != NULL:
 */
    __pyx_v_i = (__pyx_v_samples[__pyx_v_p]);

    /* "sklearn/tree/_tree.pyx":704
 *             i = samples[p]
 * 
 *             if sample_weight != NULL:             # <<<<<<<<<<<<<<
 *                 w  = sample_weight[i]
 * 
 */
    __pyx_t_6 = ((__pyx_v_sample_weight != NULL) != 0);
    if (__pyx_t_6) {

      /* "sklearn/tree/_tree.pyx":705
 * 
 *             if sample_weight != NULL:
 *                 w  = sample_weight[i]             # <<<<<<<<<<<<<<
 * 
 *             for k from 0 <= k < n_outputs:
 */
      __pyx_v_w = (__pyx_v_sample_weight[__pyx_v_i]);
      goto __pyx_L5;
    }
    __pyx_L5:;

    /* "sklearn/tree/_tree.pyx":707
 *                 w  = sample_weight[i]
 * 
 *             for k from 0 <= k < n_outputs:             # <<<<<<<<<<<<<<
 *                 y_ik = y[i * y_stride + k]
 *                 w_y_ik = w * y_ik
 */
    __pyx_t_7 = __pyx_v_n_outputs;
    for (__pyx_v_k = 0; __pyx_v_k < __pyx_t_7; __pyx_v_k++) {

      /* "sklearn/tree/_tree.pyx":708
 * 
 *             for k from 0 <= k < n_outputs:
 *                 y_ik = y[i * y_stride + k]             # <<<<<<<<<<<<<<
 *                 w_y_ik = w * y_ik
 * 
 */
      __pyx_v_y_ik = (__pyx_v_y[((__pyx_v_i * __pyx_v_y_stride) + __pyx_v_k)]);

      /* "sklearn/tree/_tree.pyx":709
 *             for k from 0 <= k < n_outputs:
 *                 y_ik = y[i * y_stride + k]
 *                 w_y_ik = w * y_ik             # <<<<<<<<<<<<<<
 * 
 *                 sq_sum_left[k] += w_y_ik * y_ik
 */
      __pyx_v_w_y_ik = (__pyx_v_w * __pyx_v_y_ik);

      /* "sklearn/tree/_tree.pyx":711
 *                 w_y_ik = w * y_ik
 * 
 *                 sq_sum_left[k] += w_y_ik * y_ik             # <<<<<<<<<<<<<<
 *                 sq_sum_right[k] -= w_y_ik * y_ik
 * 
 */
      __pyx_t_8 = __pyx_v_k;
      (__pyx_v_sq_sum_left[__pyx_t_8]) = ((__pyx_v_sq_sum_left[__pyx_t_8]) + (__pyx_v_w_y_ik * __pyx_v_y_ik));

      /* "sklearn/tree/_tree.pyx":712
 * 
 *                 sq_sum_left[k] += w_y_ik * y_ik
 *                 sq_sum_right[k] -= w_y_ik * y_ik             # <<<<<<<<<<<<<<
 * 
 *                 mean_left[k] = ((weighted_n_left * mean_left[k] + w_y_ik) /
 */
      __pyx_t_8 = __pyx_v_k;
      (__pyx_v_sq_sum_right[__pyx_t_8]) = ((__pyx_v_sq_sum_right[__pyx_t_8]) - (__pyx_v_w_y_ik * __pyx_v_y_ik));

      /* "sklearn/tree/_tree.pyx":714
 *                 sq_sum_right[k] -= w_y_ik * y_ik
 * 
 *                 mean_left[k] = ((weighted_n_left * mean_left[k] + w_y_ik) /             # <<<<<<<<<<<<<<
 *                                 (weighted_n_left + w))
 *                 mean_right[k] = ((weighted_n_right * mean_right[k] - w_y_ik) /
 */
      (__pyx_v_mean_left[__pyx_v_k]) = (((__pyx_v_weighted_n_left * (__pyx_v_mean_left[__pyx_v_k])) + __pyx_v_w_y_ik) / (__pyx_v_weighted_n_left + __pyx_v_w));

      /* "sklearn/tree/_tree.pyx":716
 *                 mean_left[k] = ((weighted_n_left * mean_left[k] + w_y_ik) /
 *                                 (weighted_n_left + w))
 *                 mean_right[k] = ((weighted_n_right * mean_right[k] - w_y_ik) /             # <<<<<<<<<<<<<<
 *                                  (weighted_n_right - w))
 * 
 */
      (__pyx_v_mean_right[__pyx_v_k]) = (((__pyx_v_weighted_n_right * (__pyx_v_mean_right[__pyx_v_k])) - __pyx_v_w_y_ik) / (__pyx_v_weighted_n_right - __pyx_v_w));
    }

    /* "sklearn/tree/_tree.pyx":719
 *                                  (weighted_n_right - w))
 * 
 *             weighted_n_left += w             # <<<<<<<<<<<<<<
 *             weighted_n_right -= w
 * 
 */
    __pyx_v_weighted_n_left = (__pyx_v_weighted_n_left + __pyx_v_w);

    /* "sklearn/tree/_tree.pyx":720
 * 
 *             weighted_n_left += w
 *             weighted_n_right -= w             # <<<<<<<<<<<<<<
 * 
 *         for k from 0 <= k < n_outputs:
 */
    __pyx_v_weighted_n_right = (__pyx_v_weighted_n_right - __pyx_v_w);
  }

  /* "sklearn/tree/_tree.pyx":722
 *             weighted_n_right -= w
 * 
 *         for k from 0 <= k < n_outputs:             # <<<<<<<<<<<<<<
 *             var_left[k] = (sq_sum_left[k] -
 *                            weighted_n_left * (mean_left[k] * mean_left[k]))
 */
  __pyx_t_2 = __pyx_v_n_outputs;
  for (__pyx_v_k = 0; __pyx_v_k < __pyx_t_2; __pyx_v_k++) {

    /* "sklearn/tree/_tree.pyx":723
 * 
 *         for k from 0 <= k < n_outputs:
 *             var_left[k] = (sq_sum_left[k] -             # <<<<<<<<<<<<<<
 *                            weighted_n_left * (mean_left[k] * mean_left[k]))
 *             var_right[k] = (sq_sum_right[k] -
 */
    (__pyx_v_var_left[__pyx_v_k]) = ((__pyx_v_sq_sum_left[__pyx_v_k]) - (__pyx_v_weighted_n_left * ((__pyx_v_mean_left[__pyx_v_k]) * (__pyx_v_mean_left[__pyx_v_k]))));

    /* "sklearn/tree/_tree.pyx":725
 *             var_left[k] = (sq_sum_left[k] -
 *                            weighted_n_left * (mean_left[k] * mean_left[k]))
 *             var_right[k] = (sq_sum_right[k] -             # <<<<<<<<<<<<<<
 *                             weighted_n_right * (mean_right[k] * mean_right[k]))
 * 
 */
    (__pyx_v_var_right[__pyx_v_k]) = ((__pyx_v_sq_sum_right[__pyx_v_k]) - (__pyx_v_weighted_n_right * ((__pyx_v_mean_right[__pyx_v_k]) * (__pyx_v_mean_right[__pyx_v_k]))));
  }

  /* "sklearn/tree/_tree.pyx":728
 *                             weighted_n_right * (mean_right[k] * mean_right[k]))
 * 
 *         self.weighted_n_left = weighted_n_left             # <<<<<<<<<<<<<<
 *         self.weighted_n_right = weighted_n_right
 * 
 */
  __pyx_v_self->__pyx_base.weighted_n_left = __pyx_v_weighted_n_left;

  /* "sklearn/tree/_tree.pyx":729
 * 
 *         self.weighted_n_left = weighted_n_left
 *         self.weighted_n_right = weighted_n_right             # <<<<<<<<<<<<<<
 * 
 *         self.pos = new_pos
 */
  __pyx_v_self->__pyx_base.weighted_n_right = __pyx_v_weighted_n_right;

  /* "sklearn/tree/_tree.pyx":731
 *         self.weighted_n_right = weighted_n_right
 * 
 *         self.pos = new_pos             # <<<<<<<<<<<<<<
 * 
 *     cdef double node_impurity(self) nogil:
 */
  __pyx_v_self->__pyx_base.pos = __pyx_v_new_pos;

}

/* "sklearn/tree/_tree.pyx":733
 *         self.pos = new_pos
 * 
 *     cdef double node_impurity(self) nogil:             # <<<<<<<<<<<<<<
 *         pass
 * 
 */

static double __pyx_f_7sklearn_4tree_5_tree_19RegressionCriterion_node_impurity(CYTHON_UNUSED struct __pyx_obj_7sklearn_4tree_5_tree_RegressionCriterion *__pyx_v_self) {
  double __pyx_r;

  __pyx_r = 0;
  return __pyx_r;
}

/* "sklearn/tree/_tree.pyx":736
 *         pass
 * 
 *     cdef double children_impurity(self) nogil:             # <<<<<<<<<<<<<<
 *         pass
 * 
 */

static double __pyx_f_7sklearn_4tree_5_tree_19RegressionCriterion_children_impurity(CYTHON_UNUSED struct __pyx_obj_7sklearn_4tree_5_tree_RegressionCriterion *__pyx_v_self) {
  double __pyx_r;

  __pyx_r = 0;
  return __pyx_r;
}

/* "sklearn/tree/_tree.pyx":739
 *         pass
 * 
 *     cdef void node_value(self, double* dest) nogil:             # <<<<<<<<<<<<<<
 *         """Compute the node value of samples[start:end] into dest."""
 *         cdef SIZE_t n_outputs = self.n_outputs
 */

static void __pyx_f_7sklearn_4tree_5_tree_19RegressionCriterion_node_value(struct __pyx_obj_7sklearn_4tree_5_tree_RegressionCriterion *__pyx_v_self, double *__pyx_v_dest) {
  __pyx_t_7sklearn_4tree_5_tree_SIZE_t __pyx_v_n_outputs;
  double *__pyx_v_mean_total;
  __pyx_t_7sklearn_4tree_5_tree_SIZE_t __pyx_v_k;
  __pyx_t_7sklearn_4tree_5_tree_SIZE_t __pyx_t_1;
  double *__pyx_t_2;

  /* "sklearn/tree/_tree.pyx":741
 *     cdef void node_value(self, double* dest) nogil:
 *         """Compute the node value of samples[start:end] into dest."""
 *         cdef SIZE_t n_outputs = self.n_outputs             # <<<<<<<<<<<<<<
 *         cdef double* mean_total = self.mean_total
 *         cdef SIZE_t k
 */
  __pyx_t_1 = __pyx_v_self->__pyx_base.n_outputs;
  __pyx_v_n_outputs = __pyx_t_1;

  /* "sklearn/tree/_tree.pyx":742
 *         """Compute the node value of samples[start:end] into dest."""
 *         cdef SIZE_t n_outputs = self.n_outputs
 *         cdef double* mean_total = self.mean_total             # <<<<<<<<<<<<<<
 *         cdef SIZE_t k
 * 
 */
  __pyx_t_2 = __pyx_v_self->mean_total;
  __pyx_v_mean_total = __pyx_t_2;

  /* "sklearn/tree/_tree.pyx":745
 *         cdef SIZE_t k
 * 
 *         for k from 0 <= k < n_outputs:             # <<<<<<<<<<<<<<
 *             dest[k] = mean_total[k]
 * 
 */
  __pyx_t_1 = __pyx_v_n_outputs;
  for (__pyx_v_k = 0; __pyx_v_k < __pyx_t_1; __pyx_v_k++) {

    /* "sklearn/tree/_tree.pyx":746
 * 
 *         for k from 0 <= k < n_outputs:
 *             dest[k] = mean_total[k]             # <<<<<<<<<<<<<<
 * 
 * cdef class MSE(RegressionCriterion):
 */
    (__pyx_v_dest[__pyx_v_k]) = (__pyx_v_mean_total[__pyx_v_k]);
  }

}

/* "sklearn/tree/_tree.pyx":753
 *         MSE = var_left + var_right
 *     """
 *     cdef double node_impurity(self) nogil:             # <<<<<<<<<<<<<<
 *         """Evaluate the impurity of the current node, i.e. the impurity of
 *            samples[start:end]."""
 */

static double __pyx_f_7sklearn_4tree_5_tree_3MSE_node_impurity(struct __pyx_obj_7sklearn_4tree_5_tree_MSE *__pyx_v_self) {
  __pyx_t_7sklearn_4tree_5_tree_SIZE_t __pyx_v_n_outputs;
  double *__pyx_v_sq_sum_total;
  double *__pyx_v_mean_total;
  double __pyx_v_weighted_n_node_samples;
  double __pyx_v_total;
  __pyx_t_7sklearn_4tree_5_tree_SIZE_t __pyx_v_k;
  double __pyx_r;
  __pyx_t_7sklearn_4tree_5_tree_SIZE_t __pyx_t_1;
  double *__pyx_t_2;
  double __pyx_t_3;

  /* "sklearn/tree/_tree.pyx":756
 *         """Evaluate the impurity of the current node, i.e. the impurity of
 *            samples[start:end]."""
 *         cdef SIZE_t n_outputs = self.n_outputs             # <<<<<<<<<<<<<<
 *         cdef double* sq_sum_total = self.sq_sum_total
 *         cdef double* mean_total = self.mean_total
 */
  __pyx_t_1 = __pyx_v_self->__pyx_base.__pyx_base.n_outputs;
  __pyx_v_n_outputs = __pyx_t_1;

  /* "sklearn/tree/_tree.pyx":757
 *            samples[start:end]."""
 *         cdef SIZE_t n_outputs = self.n_outputs
 *         cdef double* sq_sum_total = self.sq_sum_total             # <<<<<<<<<<<<<<
 *         cdef double* mean_total = self.mean_total
 *         cdef double weighted_n_node_samples = self.weighted_n_node_samples
 */
  __pyx_t_2 = __pyx_v_self->__pyx_base.sq_sum_total;
  __pyx_v_sq_sum_total = __pyx_t_2;

  /* "sklearn/tree/_tree.pyx":758
 *         cdef SIZE_t n_outputs = self.n_outputs
 *         cdef double* sq_sum_total = self.sq_sum_total
 *         cdef double* mean_total = self.mean_total             # <<<<<<<<<<<<<<
 *         cdef double weighted_n_node_samples = self.weighted_n_node_samples
 *         cdef double total = 0.0
 */
  __pyx_t_2 = __pyx_v_self->__pyx_base.mean_total;
  __pyx_v_mean_total = __pyx_t_2;

  /* "sklearn/tree/_tree.pyx":759
 *         cdef double* sq_sum_total = self.sq_sum_total
 *         cdef double* mean_total = self.mean_total
 *         cdef double weighted_n_node_samples = self.weighted_n_node_samples             # <<<<<<<<<<<<<<
 *         cdef double total = 0.0
 *         cdef SIZE_t k
 */
  __pyx_t_3 = __pyx_v_self->__pyx_base.__pyx_base.weighted_n_node_samples;
  __pyx_v_weighted_n_node_samples = __pyx_t_3;

  /* "sklearn/tree/_tree.pyx":760
 *         cdef double* mean_total = self.mean_total
 *         cdef double weighted_n_node_samples = self.weighted_n_node_samples
 *         cdef double total = 0.0             # <<<<<<<<<<<<<<
 *         cdef SIZE_t k
 * 
 */
  __pyx_v_total = 0.0;

  /* "sklearn/tree/_tree.pyx":763
 *         cdef SIZE_t k
 * 
 *         for k from 0 <= k < n_outputs:             # <<<<<<<<<<<<<<
 *             total += (sq_sum_total[k] -
 *                       weighted_n_node_samples * (mean_total[k] *
 */
  __pyx_t_1 = __pyx_v_n_outputs;
  for (__pyx_v_k = 0; __pyx_v_k < __pyx_t_1; __pyx_v_k++) {

    /* "sklearn/tree/_tree.pyx":766
 *             total += (sq_sum_total[k] -
 *                       weighted_n_node_samples * (mean_total[k] *
 *                                                  mean_total[k]))             # <<<<<<<<<<<<<<
 * 
 *         return total / n_outputs
 */
    __pyx_v_total = (__pyx_v_total + ((__pyx_v_sq_sum_total[__pyx_v_k]) - (__pyx_v_weighted_n_node_samples * ((__pyx_v_mean_total[__pyx_v_k]) * (__pyx_v_mean_total[__pyx_v_k])))));
  }

  /* "sklearn/tree/_tree.pyx":768
 *                                                  mean_total[k]))
 * 
 *         return total / n_outputs             # <<<<<<<<<<<<<<
 * 
 *     cdef double children_impurity(self) nogil:
 */
  __pyx_r = (__pyx_v_total / __pyx_v_n_outputs);
  goto __pyx_L0;

  __pyx_r = 0;
  __pyx_L0:;
  return __pyx_r;
}

/* "sklearn/tree/_tree.pyx":770
 *         return total / n_outputs
 * 
 *     cdef double children_impurity(self) nogil:             # <<<<<<<<<<<<<<
 *         """Evaluate the impurity in children nodes, i.e. the impurity of
 *            samples[start:pos] + the impurity of samples[pos:end]."""
 */

static double __pyx_f_7sklearn_4tree_5_tree_3MSE_children_impurity(struct __pyx_obj_7sklearn_4tree_5_tree_MSE *__pyx_v_self) {
  __pyx_t_7sklearn_4tree_5_tree_SIZE_t __pyx_v_n_outputs;
  double *__pyx_v_var_left;
  double *__pyx_v_var_right;
  double __pyx_v_total;
  __pyx_t_7sklearn_4tree_5_tree_SIZE_t __pyx_v_k;
  double __pyx_r;
  __pyx_t_7sklearn_4tree_5_tree_SIZE_t __pyx_t_1;
  double *__pyx_t_2;

  /* "sklearn/tree/_tree.pyx":773
 *         """Evaluate the impurity in children nodes, i.e. the impurity of
 *            samples[start:pos] + the impurity of samples[pos:end]."""
 *         cdef SIZE_t n_outputs = self.n_outputs             # <<<<<<<<<<<<<<
 *         cdef double* var_left = self.var_left
 *         cdef double* var_right = self.var_right
 */
  __pyx_t_1 = __pyx_v_self->__pyx_base.__pyx_base.n_outputs;
  __pyx_v_n_outputs = __pyx_t_1;

  /* "sklearn/tree/_tree.pyx":774
 *            samples[start:pos] + the impurity of samples[pos:end]."""
 *         cdef SIZE_t n_outputs = self.n_outputs
 *         cdef double* var_left = self.var_left             # <<<<<<<<<<<<<<
 *         cdef double* var_right = self.var_right
 *         cdef double total = 0.0
 */
  __pyx_t_2 = __pyx_v_self->__pyx_base.var_left;
  __pyx_v_var_left = __pyx_t_2;

  /* "sklearn/tree/_tree.pyx":775
 *         cdef SIZE_t n_outputs = self.n_outputs
 *         cdef double* var_left = self.var_left
 *         cdef double* var_right = self.var_right             # <<<<<<<<<<<<<<
 *         cdef double total = 0.0
 *         cdef SIZE_t k
 */
  __pyx_t_2 = __pyx_v_self->__pyx_base.var_right;
  __pyx_v_var_right = __pyx_t_2;

  /* "sklearn/tree/_tree.pyx":776
 *         cdef double* var_left = self.var_left
 *         cdef double* var_right = self.var_right
 *         cdef double total = 0.0             # <<<<<<<<<<<<<<
 *         cdef SIZE_t k
 * 
 */
  __pyx_v_total = 0.0;

  /* "sklearn/tree/_tree.pyx":779
 *         cdef SIZE_t k
 * 
 *         for k from 0 <= k < n_outputs:             # <<<<<<<<<<<<<<
 *             total += var_left[k]
 *             total += var_right[k]
 */
  __pyx_t_1 = __pyx_v_n_outputs;
  for (__pyx_v_k = 0; __pyx_v_k < __pyx_t_1; __pyx_v_k++) {

    /* "sklearn/tree/_tree.pyx":780
 * 
 *         for k from 0 <= k < n_outputs:
 *             total += var_left[k]             # <<<<<<<<<<<<<<
 *             total += var_right[k]
 * 
 */
    __pyx_v_total = (__pyx_v_total + (__pyx_v_var_left[__pyx_v_k]));

    /* "sklearn/tree/_tree.pyx":781
 *         for k from 0 <= k < n_outputs:
 *             total += var_left[k]
 *             total += var_right[k]             # <<<<<<<<<<<<<<
 * 
 *         return total / n_outputs
 */
    __pyx_v_total = (__pyx_v_total + (__pyx_v_var_right[__pyx_v_k]));
  }

  /* "sklearn/tree/_tree.pyx":783
 *             total += var_right[k]
 * 
 *         return total / n_outputs             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = (__pyx_v_total / __pyx_v_n_outputs);
  goto __pyx_L0;

  __pyx_r = 0;
  __pyx_L0:;
  return __pyx_r;
}

/* Python wrapper */
static int __pyx_pw_7sklearn_4tree_5_tree_8Splitter_1__cinit__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static int __pyx_pw_7sklearn_4tree_5_tree_8Splitter_1__cinit__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  struct __pyx_obj_7sklearn_4tree_5_tree_Criterion *__pyx_v_criterion = 0;
  __pyx_t_7sklearn_4tree_5_tree_SIZE_t __pyx_v_max_features;
  __pyx_t_7sklearn_4tree_5_tree_SIZE_t __pyx_v_min_samples_leaf;
  PyObject *__pyx_v_random_state = 0;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__cinit__ (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s__criterion,&__pyx_n_s__max_features,&__pyx_n_s__min_samples_leaf,&__pyx_n_s__random_state,0};
    PyObject* values[4] = {0,0,0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s__criterion)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s__max_features)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__cinit__", 1, 4, 4, 1); {__pyx_filename = __pyx_f[0]; __pyx_lineno = 791; __pyx_clineno = __LINE__; goto __pyx_L3_error;}
        }
        case  2:
        if (likely((values[2] = PyDict_GetItem(__pyx_kwds, __pyx_n_s__min_samples_leaf)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__cinit__", 1, 4, 4, 2); {__pyx_filename = __pyx_f[0]; __pyx_lineno = 791; __pyx_clineno = __LINE__; goto __pyx_L3_error;}
        }
        case  3:
        if (likely((values[3] = PyDict_GetItem(__pyx_kwds, __pyx_n_s__random_state)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__cinit__", 1, 4, 4, 3); {__pyx_filename = __pyx_f[0]; __pyx_lineno = 791; __pyx_clineno = __LINE__; goto __pyx_L3_error;}
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "__cinit__") < 0)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 791; __pyx_clineno = __LINE__; goto __pyx_L3_error;}
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 4) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
      values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
      values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
    }
    __pyx_v_criterion = ((struct __pyx_obj_7sklearn_4tree_5_tree_Criterion *)values[0]);
    __pyx_v_max_features = __Pyx_PyInt_from_py_Py_intptr_t(values[1]); if (unlikely((__pyx_v_max_features == (npy_intp)-1) && PyErr_Occurred())) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 792; __pyx_clineno = __LINE__; goto __pyx_L3_error;}
    __pyx_v_min_samples_leaf = __Pyx_PyInt_from_py_Py_intptr_t(values[2]); if (unlikely((__pyx_v_min_samples_leaf == (npy_intp)-1) && PyErr_Occurred())) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 793; __pyx_clineno = __LINE__; goto __pyx_L3_error;}
    __pyx_v_random_state = values[3];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__cinit__", 1, 4, 4, PyTuple_GET_SIZE(__pyx_args)); {__pyx_filename = __pyx_f[0]; __pyx_lineno = 791; __pyx_clineno = __LINE__; goto __pyx_L3_error;}
  __pyx_L3_error:;
  __Pyx_AddTraceback("sklearn.tree._tree.Splitter.__cinit__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return -1;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_criterion), __pyx_ptype_7sklearn_4tree_5_tree_Criterion, 1, "criterion", 0))) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 791; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __pyx_r = __pyx_pf_7sklearn_4tree_5_tree_8Splitter___cinit__(((struct __pyx_obj_7sklearn_4tree_5_tree_Splitter *)__pyx_v_self), __pyx_v_criterion, __pyx_v_max_features, __pyx_v_min_samples_leaf, __pyx_v_random_state);
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "sklearn/tree/_tree.pyx":791
 * 
 * cdef class Splitter:
 *     def __cinit__(self, Criterion criterion,             # <<<<<<<<<<<<<<
 *                         SIZE_t max_features,
 *                         SIZE_t min_samples_leaf,
 */

static int __pyx_pf_7sklearn_4tree_5_tree_8Splitter___cinit__(struct __pyx_obj_7sklearn_4tree_5_tree_Splitter *__pyx_v_self, struct __pyx_obj_7sklearn_4tree_5_tree_Criterion *__pyx_v_criterion, __pyx_t_7sklearn_4tree_5_tree_SIZE_t __pyx_v_max_features, __pyx_t_7sklearn_4tree_5_tree_SIZE_t __pyx_v_min_samples_leaf, PyObject *__pyx_v_random_state) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__cinit__", 0);

  /* "sklearn/tree/_tree.pyx":795
 *                         SIZE_t min_samples_leaf,
 *                         object random_state):
 *         self.criterion = criterion             # <<<<<<<<<<<<<<
 * 
 *         self.samples = NULL
 */
  __Pyx_INCREF(((PyObject *)__pyx_v_criterion));
  __Pyx_GIVEREF(((PyObject *)__pyx_v_criterion));
  __Pyx_GOTREF(__pyx_v_self->criterion);
  __Pyx_DECREF(((PyObject *)__pyx_v_self->criterion));
  __pyx_v_self->criterion = __pyx_v_criterion;

  /* "sklearn/tree/_tree.pyx":797
 *         self.criterion = criterion
 * 
 *         self.samples = NULL             # <<<<<<<<<<<<<<
 *         self.n_samples = 0
 *         self.features = NULL
 */
  __pyx_v_self->samples = NULL;

  /* "sklearn/tree/_tree.pyx":798
 * 
 *         self.samples = NULL
 *         self.n_samples = 0             # <<<<<<<<<<<<<<
 *         self.features = NULL
 *         self.n_features = 0
 */
  __pyx_v_self->n_samples = 0;

  /* "sklearn/tree/_tree.pyx":799
 *         self.samples = NULL
 *         self.n_samples = 0
 *         self.features = NULL             # <<<<<<<<<<<<<<
 *         self.n_features = 0
 * 
 */
  __pyx_v_self->features = NULL;

  /* "sklearn/tree/_tree.pyx":800
 *         self.n_samples = 0
 *         self.features = NULL
 *         self.n_features = 0             # <<<<<<<<<<<<<<
 * 
 *         self.X = None
 */
  __pyx_v_self->n_features = 0;

  /* "sklearn/tree/_tree.pyx":802
 *         self.n_features = 0
 * 
 *         self.X = None             # <<<<<<<<<<<<<<
 *         self.y = NULL
 *         self.y_stride = 0
 */
  __Pyx_INCREF(Py_None);
  __Pyx_GIVEREF(Py_None);
  __Pyx_GOTREF(__pyx_v_self->X);
  __Pyx_DECREF(((PyObject *)__pyx_v_self->X));
  __pyx_v_self->X = ((PyArrayObject *)Py_None);

  /* "sklearn/tree/_tree.pyx":803
 * 
 *         self.X = None
 *         self.y = NULL             # <<<<<<<<<<<<<<
 *         self.y_stride = 0
 *         self.sample_weight = NULL
 */
  __pyx_v_self->y = NULL;

  /* "sklearn/tree/_tree.pyx":804
 *         self.X = None
 *         self.y = NULL
 *         self.y_stride = 0             # <<<<<<<<<<<<<<
 *         self.sample_weight = NULL
 * 
 */
  __pyx_v_self->y_stride = 0;

  /* "sklearn/tree/_tree.pyx":805
 *         self.y = NULL
 *         self.y_stride = 0
 *         self.sample_weight = NULL             # <<<<<<<<<<<<<<
 * 
 *         self.max_features = max_features
 */
  __pyx_v_self->sample_weight = NULL;

  /* "sklearn/tree/_tree.pyx":807
 *         self.sample_weight = NULL
 * 
 *         self.max_features = max_features             # <<<<<<<<<<<<<<
 *         self.min_samples_leaf = min_samples_leaf
 *         self.random_state = random_state
 */
  __pyx_v_self->max_features = __pyx_v_max_features;

  /* "sklearn/tree/_tree.pyx":808
 * 
 *         self.max_features = max_features
 *         self.min_samples_leaf = min_samples_leaf             # <<<<<<<<<<<<<<
 *         self.random_state = random_state
 * 
 */
  __pyx_v_self->min_samples_leaf = __pyx_v_min_samples_leaf;

  /* "sklearn/tree/_tree.pyx":809
 *         self.max_features = max_features
 *         self.min_samples_leaf = min_samples_leaf
 *         self.random_state = random_state             # <<<<<<<<<<<<<<
 * 
 *     def __dealloc__(self):
 */
  __Pyx_INCREF(__pyx_v_random_state);
  __Pyx_GIVEREF(__pyx_v_random_state);
  __Pyx_GOTREF(__pyx_v_self->random_state);
  __Pyx_DECREF(__pyx_v_self->random_state);
  __pyx_v_self->random_state = __pyx_v_random_state;

  __pyx_r = 0;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static void __pyx_pw_7sklearn_4tree_5_tree_8Splitter_3__dealloc__(PyObject *__pyx_v_self); /*proto*/
static void __pyx_pw_7sklearn_4tree_5_tree_8Splitter_3__dealloc__(PyObject *__pyx_v_self) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__dealloc__ (wrapper)", 0);
  __pyx_pf_7sklearn_4tree_5_tree_8Splitter_2__dealloc__(((struct __pyx_obj_7sklearn_4tree_5_tree_Splitter *)__pyx_v_self));
  __Pyx_RefNannyFinishContext();
}

/* "sklearn/tree/_tree.pyx":811
 *         self.random_state = random_state
 * 
 *     def __dealloc__(self):             # <<<<<<<<<<<<<<
 *         """Destructor."""
 *         free(self.samples)
 */

static void __pyx_pf_7sklearn_4tree_5_tree_8Splitter_2__dealloc__(struct __pyx_obj_7sklearn_4tree_5_tree_Splitter *__pyx_v_self) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__dealloc__", 0);

  /* "sklearn/tree/_tree.pyx":813
 *     def __dealloc__(self):
 *         """Destructor."""
 *         free(self.samples)             # <<<<<<<<<<<<<<
 *         free(self.features)
 * 
 */
  free(__pyx_v_self->samples);

  /* "sklearn/tree/_tree.pyx":814
 *         """Destructor."""
 *         free(self.samples)
 *         free(self.features)             # <<<<<<<<<<<<<<
 * 
 *     def __getstate__(self):
 */
  free(__pyx_v_self->features);

  __Pyx_RefNannyFinishContext();
}

/* Python wrapper */
static PyObject *__pyx_pw_7sklearn_4tree_5_tree_8Splitter_5__getstate__(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static PyObject *__pyx_pw_7sklearn_4tree_5_tree_8Splitter_5__getstate__(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__getstate__ (wrapper)", 0);
  __pyx_r = __pyx_pf_7sklearn_4tree_5_tree_8Splitter_4__getstate__(((struct __pyx_obj_7sklearn_4tree_5_tree_Splitter *)__pyx_v_self));
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "sklearn/tree/_tree.pyx":816
 *         free(self.features)
 * 
 *     def __getstate__(self):             # <<<<<<<<<<<<<<
 *         return {}
 * 
 */

static PyObject *__pyx_pf_7sklearn_4tree_5_tree_8Splitter_4__getstate__(CYTHON_UNUSED struct __pyx_obj_7sklearn_4tree_5_tree_Splitter *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__getstate__", 0);

  /* "sklearn/tree/_tree.pyx":817
 * 
 *     def __getstate__(self):
 *         return {}             # <<<<<<<<<<<<<<
 * 
 *     def __setstate__(self, d):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = PyDict_New(); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 817; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(((PyObject *)__pyx_t_1));
  __pyx_r = ((PyObject *)__pyx_t_1);
  __pyx_t_1 = 0;
  goto __pyx_L0;

  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("sklearn.tree._tree.Splitter.__getstate__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_7sklearn_4tree_5_tree_8Splitter_7__setstate__(PyObject *__pyx_v_self, PyObject *__pyx_v_d); /*proto*/
static PyObject *__pyx_pw_7sklearn_4tree_5_tree_8Splitter_7__setstate__(PyObject *__pyx_v_self, PyObject *__pyx_v_d) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__setstate__ (wrapper)", 0);
  __pyx_r = __pyx_pf_7sklearn_4tree_5_tree_8Splitter_6__setstate__(((struct __pyx_obj_7sklearn_4tree_5_tree_Splitter *)__pyx_v_self), ((PyObject *)__pyx_v_d));
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "sklearn/tree/_tree.pyx":819
 *         return {}
 * 
 *     def __setstate__(self, d):             # <<<<<<<<<<<<<<
 *         pass
 * 
 */

static PyObject *__pyx_pf_7sklearn_4tree_5_tree_8Splitter_6__setstate__(CYTHON_UNUSED struct __pyx_obj_7sklearn_4tree_5_tree_Splitter *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v_d) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__setstate__", 0);

  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "sklearn/tree/_tree.pyx":822
 *         pass
 * 
 *     cdef void init(self, np.ndarray[DTYPE_t, ndim=2, mode="c"] X,             # <<<<<<<<<<<<<<
 *                          np.ndarray[DOUBLE_t, ndim=2, mode="c"] y,
 *                          DOUBLE_t* sample_weight):
 */

static void __pyx_f_7sklearn_4tree_5_tree_8Splitter_init(struct __pyx_obj_7sklearn_4tree_5_tree_Splitter *__pyx_v_self, PyArrayObject *__pyx_v_X, PyArrayObject *__pyx_v_y, __pyx_t_7sklearn_4tree_5_tree_DOUBLE_t *__pyx_v_sample_weight) {
  __pyx_t_7sklearn_4tree_5_tree_SIZE_t __pyx_v_n_samples;
  __pyx_t_7sklearn_4tree_5_tree_SIZE_t *__pyx_v_samples;
  __pyx_t_7sklearn_4tree_5_tree_SIZE_t __pyx_v_i;
  __pyx_t_7sklearn_4tree_5_tree_SIZE_t __pyx_v_j;
  __pyx_t_7sklearn_4tree_5_tree_SIZE_t __pyx_v_n_features;
  __pyx_t_7sklearn_4tree_5_tree_SIZE_t *__pyx_v_features;
  __Pyx_LocalBuf_ND __pyx_pybuffernd_X;
  __Pyx_Buffer __pyx_pybuffer_X;
  __Pyx_LocalBuf_ND __pyx_pybuffernd_y;
  __Pyx_Buffer __pyx_pybuffer_y;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  unsigned int __pyx_t_5;
  __pyx_t_7sklearn_4tree_5_tree_SIZE_t __pyx_t_6;
  int __pyx_t_7;
  int __pyx_t_8;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("init", 0);
  __pyx_pybuffer_X.pybuffer.buf = NULL;
  __pyx_pybuffer_X.refcount = 0;
  __pyx_pybuffernd_X.data = NULL;
  __pyx_pybuffernd_X.rcbuffer = &__pyx_pybuffer_X;
  __pyx_pybuffer_y.pybuffer.buf = NULL;
  __pyx_pybuffer_y.refcount = 0;
  __pyx_pybuffernd_y.data = NULL;
  __pyx_pybuffernd_y.rcbuffer = &__pyx_pybuffer_y;
  {
    __Pyx_BufFmt_StackElem __pyx_stack[1];
    if (unlikely(__Pyx_GetBufferAndValidate(&__pyx_pybuffernd_X.rcbuffer->pybuffer, (PyObject*)__pyx_v_X, &__Pyx_TypeInfo_nn___pyx_t_7sklearn_4tree_5_tree_DTYPE_t, PyBUF_FORMAT| PyBUF_C_CONTIGUOUS, 2, 0, __pyx_stack) == -1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 822; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  }
  __pyx_pybuffernd_X.diminfo[0].strides = __pyx_pybuffernd_X.rcbuffer->pybuffer.strides[0]; __pyx_pybuffernd_X.diminfo[0].shape = __pyx_pybuffernd_X.rcbuffer->pybuffer.shape[0]; __pyx_pybuffernd_X.diminfo[1].strides = __pyx_pybuffernd_X.rcbuffer->pybuffer.strides[1]; __pyx_pybuffernd_X.diminfo[1].shape = __pyx_pybuffernd_X.rcbuffer->pybuffer.shape[1];
  {
    __Pyx_BufFmt_StackElem __pyx_stack[1];
    if (unlikely(__Pyx_GetBufferAndValidate(&__pyx_pybuffernd_y.rcbuffer->pybuffer, (PyObject*)__pyx_v_y, &__Pyx_TypeInfo_nn___pyx_t_7sklearn_4tree_5_tree_DOUBLE_t, PyBUF_FORMAT| PyBUF_C_CONTIGUOUS, 2, 0, __pyx_stack) == -1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 822; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  }
  __pyx_pybuffernd_y.diminfo[0].strides = __pyx_pybuffernd_y.rcbuffer->pybuffer.strides[0]; __pyx_pybuffernd_y.diminfo[0].shape = __pyx_pybuffernd_y.rcbuffer->pybuffer.shape[0]; __pyx_pybuffernd_y.diminfo[1].strides = __pyx_pybuffernd_y.rcbuffer->pybuffer.strides[1]; __pyx_pybuffernd_y.diminfo[1].shape = __pyx_pybuffernd_y.rcbuffer->pybuffer.shape[1];

  /* "sklearn/tree/_tree.pyx":827
 *         """Initialize the splitter."""
 *         # Free old structures if any
 *         if self.samples != NULL:             # <<<<<<<<<<<<<<
 *             free(self.samples)
 *         if self.features != NULL:
 */
  __pyx_t_1 = ((__pyx_v_self->samples != NULL) != 0);
  if (__pyx_t_1) {

    /* "sklearn/tree/_tree.pyx":828
 *         # Free old structures if any
 *         if self.samples != NULL:
 *             free(self.samples)             # <<<<<<<<<<<<<<
 *         if self.features != NULL:
 *             free(self.features)
 */
    free(__pyx_v_self->samples);
    goto __pyx_L3;
  }
  __pyx_L3:;

  /* "sklearn/tree/_tree.pyx":829
 *         if self.samples != NULL:
 *             free(self.samples)
 *         if self.features != NULL:             # <<<<<<<<<<<<<<
 *             free(self.features)
 * 
 */
  __pyx_t_1 = ((__pyx_v_self->features != NULL) != 0);
  if (__pyx_t_1) {

    /* "sklearn/tree/_tree.pyx":830
 *             free(self.samples)
 *         if self.features != NULL:
 *             free(self.features)             # <<<<<<<<<<<<<<
 * 
 *         # Reset random state
 */
    free(__pyx_v_self->features);
    goto __pyx_L4;
  }
  __pyx_L4:;

  /* "sklearn/tree/_tree.pyx":833
 * 
 *         # Reset random state
 *         self.rand_r_state = self.random_state.randint(0, RAND_R_MAX)             # <<<<<<<<<<<<<<
 * 
 *         # Initialize samples and features structures
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self->random_state, __pyx_n_s__randint); if (unlikely(!__pyx_t_2)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 833; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = PyInt_FromLong(__pyx_e_7sklearn_4tree_5_tree_RAND_R_MAX); if (unlikely(!__pyx_t_3)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 833; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = PyTuple_New(2); if (unlikely(!__pyx_t_4)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 833; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_INCREF(__pyx_int_0);
  PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_int_0);
  __Pyx_GIVEREF(__pyx_int_0);
  PyTuple_SET_ITEM(__pyx_t_4, 1, __pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_3);
  __pyx_t_3 = 0;
  __pyx_t_3 = PyObject_Call(__pyx_t_2, ((PyObject *)__pyx_t_4), NULL); if (unlikely(!__pyx_t_3)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 833; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(((PyObject *)__pyx_t_4)); __pyx_t_4 = 0;
  __pyx_t_5 = __Pyx_PyInt_AsUnsignedInt(__pyx_t_3); if (unlikely((__pyx_t_5 == (unsigned int)-1) && PyErr_Occurred())) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 833; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_v_self->rand_r_state = __pyx_t_5;

  /* "sklearn/tree/_tree.pyx":836
 * 
 *         # Initialize samples and features structures
 *         cdef SIZE_t n_samples = X.shape[0]             # <<<<<<<<<<<<<<
 *         cdef SIZE_t* samples = <SIZE_t*> malloc(n_samples * sizeof(SIZE_t))
 * 
 */
  __pyx_v_n_samples = (__pyx_v_X->dimensions[0]);

  /* "sklearn/tree/_tree.pyx":837
 *         # Initialize samples and features structures
 *         cdef SIZE_t n_samples = X.shape[0]
 *         cdef SIZE_t* samples = <SIZE_t*> malloc(n_samples * sizeof(SIZE_t))             # <<<<<<<<<<<<<<
 * 
 *         cdef SIZE_t i, j
 */
  __pyx_v_samples = ((__pyx_t_7sklearn_4tree_5_tree_SIZE_t *)malloc((__pyx_v_n_samples * (sizeof(__pyx_t_7sklearn_4tree_5_tree_SIZE_t)))));

  /* "sklearn/tree/_tree.pyx":840
 * 
 *         cdef SIZE_t i, j
 *         j = 0             # <<<<<<<<<<<<<<
 * 
 *         for i from 0 <= i < n_samples:
 */
  __pyx_v_j = 0;

  /* "sklearn/tree/_tree.pyx":842
 *         j = 0
 * 
 *         for i from 0 <= i < n_samples:             # <<<<<<<<<<<<<<
 *             # Only work with positively weighted samples
 *             if sample_weight == NULL or sample_weight[i] != 0.0:
 */
  __pyx_t_6 = __pyx_v_n_samples;
  for (__pyx_v_i = 0; __pyx_v_i < __pyx_t_6; __pyx_v_i++) {

    /* "sklearn/tree/_tree.pyx":844
 *         for i from 0 <= i < n_samples:
 *             # Only work with positively weighted samples
 *             if sample_weight == NULL or sample_weight[i] != 0.0:             # <<<<<<<<<<<<<<
 *                 samples[j] = i
 *                 j += 1
 */
    __pyx_t_1 = ((__pyx_v_sample_weight == NULL) != 0);
    if (!__pyx_t_1) {
      __pyx_t_7 = (((__pyx_v_sample_weight[__pyx_v_i]) != 0.0) != 0);
      __pyx_t_8 = __pyx_t_7;
    } else {
      __pyx_t_8 = __pyx_t_1;
    }
    if (__pyx_t_8) {

      /* "sklearn/tree/_tree.pyx":845
 *             # Only work with positively weighted samples
 *             if sample_weight == NULL or sample_weight[i] != 0.0:
 *                 samples[j] = i             # <<<<<<<<<<<<<<
 *                 j += 1
 * 
 */
      (__pyx_v_samples[__pyx_v_j]) = __pyx_v_i;

      /* "sklearn/tree/_tree.pyx":846
 *             if sample_weight == NULL or sample_weight[i] != 0.0:
 *                 samples[j] = i
 *                 j += 1             # <<<<<<<<<<<<<<
 * 
 *         self.samples = samples
 */
      __pyx_v_j = (__pyx_v_j + 1);
      goto __pyx_L7;
    }
    __pyx_L7:;
  }

  /* "sklearn/tree/_tree.pyx":848
 *                 j += 1
 * 
 *         self.samples = samples             # <<<<<<<<<<<<<<
 *         self.n_samples = j
 * 
 */
  __pyx_v_self->samples = __pyx_v_samples;

  /* "sklearn/tree/_tree.pyx":849
 * 
 *         self.samples = samples
 *         self.n_samples = j             # <<<<<<<<<<<<<<
 * 
 *         cdef SIZE_t n_features = X.shape[1]
 */
  __pyx_v_self->n_samples = __pyx_v_j;

  /* "sklearn/tree/_tree.pyx":851
 *         self.n_samples = j
 * 
 *         cdef SIZE_t n_features = X.shape[1]             # <<<<<<<<<<<<<<
 *         cdef SIZE_t* features = <SIZE_t*> malloc(n_features * sizeof(SIZE_t))
 * 
 */
  __pyx_v_n_features = (__pyx_v_X->dimensions[1]);

  /* "sklearn/tree/_tree.pyx":852
 * 
 *         cdef SIZE_t n_features = X.shape[1]
 *         cdef SIZE_t* features = <SIZE_t*> malloc(n_features * sizeof(SIZE_t))             # <<<<<<<<<<<<<<
 * 
 *         for i from 0 <= i < n_features:
 */
  __pyx_v_features = ((__pyx_t_7sklearn_4tree_5_tree_SIZE_t *)malloc((__pyx_v_n_features * (sizeof(__pyx_t_7sklearn_4tree_5_tree_SIZE_t)))));

  /* "sklearn/tree/_tree.pyx":854
 *         cdef SIZE_t* features = <SIZE_t*> malloc(n_features * sizeof(SIZE_t))
 * 
 *         for i from 0 <= i < n_features:             # <<<<<<<<<<<<<<
 *             features[i] = i
 * 
 */
  __pyx_t_6 = __pyx_v_n_features;
  for (__pyx_v_i = 0; __pyx_v_i < __pyx_t_6; __pyx_v_i++) {

    /* "sklearn/tree/_tree.pyx":855
 * 
 *         for i from 0 <= i < n_features:
 *             features[i] = i             # <<<<<<<<<<<<<<
 * 
 *         self.features = features
 */
    (__pyx_v_features[__pyx_v_i]) = __pyx_v_i;
  }

  /* "sklearn/tree/_tree.pyx":857
 *             features[i] = i
 * 
 *         self.features = features             # <<<<<<<<<<<<<<
 *         self.n_features = n_features
 * 
 */
  __pyx_v_self->features = __pyx_v_features;

  /* "sklearn/tree/_tree.pyx":858
 * 
 *         self.features = features
 *         self.n_features = n_features             # <<<<<<<<<<<<<<
 * 
 *         # Initialize X, y, sample_weight
 */
  __pyx_v_self->n_features = __pyx_v_n_features;

  /* "sklearn/tree/_tree.pyx":861
 * 
 *         # Initialize X, y, sample_weight
 *         self.X = X             # <<<<<<<<<<<<<<
 *         self.y = <DOUBLE_t*> y.data
 *         self.y_stride = <SIZE_t> y.strides[0] / <SIZE_t> y.itemsize
 */
  __Pyx_INCREF(((PyObject *)__pyx_v_X));
  __Pyx_GIVEREF(((PyObject *)__pyx_v_X));
  __Pyx_GOTREF(__pyx_v_self->X);
  __Pyx_DECREF(((PyObject *)__pyx_v_self->X));
  __pyx_v_self->X = ((PyArrayObject *)__pyx_v_X);

  /* "sklearn/tree/_tree.pyx":862
 *         # Initialize X, y, sample_weight
 *         self.X = X
 *         self.y = <DOUBLE_t*> y.data             # <<<<<<<<<<<<<<
 *         self.y_stride = <SIZE_t> y.strides[0] / <SIZE_t> y.itemsize
 *         self.sample_weight = sample_weight
 */
  __pyx_v_self->y = ((__pyx_t_7sklearn_4tree_5_tree_DOUBLE_t *)__pyx_v_y->data);

  /* "sklearn/tree/_tree.pyx":863
 *         self.X = X
 *         self.y = <DOUBLE_t*> y.data
 *         self.y_stride = <SIZE_t> y.strides[0] / <SIZE_t> y.itemsize             # <<<<<<<<<<<<<<
 *         self.sample_weight = sample_weight
 * 
 */
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_y), __pyx_n_s__itemsize); if (unlikely(!__pyx_t_3)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 863; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_6 = __Pyx_PyInt_from_py_Py_intptr_t(__pyx_t_3); if (unlikely((__pyx_t_6 == (npy_intp)-1) && PyErr_Occurred())) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 863; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_v_self->y_stride = (((__pyx_t_7sklearn_4tree_5_tree_SIZE_t)(__pyx_v_y->strides[0])) / ((__pyx_t_7sklearn_4tree_5_tree_SIZE_t)__pyx_t_6));

  /* "sklearn/tree/_tree.pyx":864
 *         self.y = <DOUBLE_t*> y.data
 *         self.y_stride = <SIZE_t> y.strides[0] / <SIZE_t> y.itemsize
 *         self.sample_weight = sample_weight             # <<<<<<<<<<<<<<
 * 
 *     cdef void node_reset(self, SIZE_t start, SIZE_t end, double* impurity):
 */
  __pyx_v_self->sample_weight = __pyx_v_sample_weight;

  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  { PyObject *__pyx_type, *__pyx_value, *__pyx_tb;
    __Pyx_ErrFetch(&__pyx_type, &__pyx_value, &__pyx_tb);
    __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_X.rcbuffer->pybuffer);
    __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_y.rcbuffer->pybuffer);
  __Pyx_ErrRestore(__pyx_type, __pyx_value, __pyx_tb);}
  __Pyx_WriteUnraisable("sklearn.tree._tree.Splitter.init", __pyx_clineno, __pyx_lineno, __pyx_filename);
  goto __pyx_L2;
  __pyx_L0:;
  __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_X.rcbuffer->pybuffer);
  __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_y.rcbuffer->pybuffer);
  __pyx_L2:;
  __Pyx_RefNannyFinishContext();
}

/* "sklearn/tree/_tree.pyx":866
 *         self.sample_weight = sample_weight
 * 
 *     cdef void node_reset(self, SIZE_t start, SIZE_t end, double* impurity):             # <<<<<<<<<<<<<<
 *         """Reset splitter on node samples[start:end]."""
 *         cdef Criterion criterion = self.criterion
 */

static void __pyx_f_7sklearn_4tree_5_tree_8Splitter_node_reset(struct __pyx_obj_7sklearn_4tree_5_tree_Splitter *__pyx_v_self, __pyx_t_7sklearn_4tree_5_tree_SIZE_t __pyx_v_start, __pyx_t_7sklearn_4tree_5_tree_SIZE_t __pyx_v_end, double *__pyx_v_impurity) {
  struct __pyx_obj_7sklearn_4tree_5_tree_Criterion *__pyx_v_criterion = 0;
  __pyx_t_7sklearn_4tree_5_tree_SIZE_t *__pyx_v_samples;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __pyx_t_7sklearn_4tree_5_tree_SIZE_t *__pyx_t_2;
  __Pyx_RefNannySetupContext("node_reset", 0);

  /* "sklearn/tree/_tree.pyx":868
 *     cdef void node_reset(self, SIZE_t start, SIZE_t end, double* impurity):
 *         """Reset splitter on node samples[start:end]."""
 *         cdef Criterion criterion = self.criterion             # <<<<<<<<<<<<<<
 * 
 *         cdef SIZE_t* samples = self.samples
 */
  __pyx_t_1 = ((PyObject *)__pyx_v_self->criterion);
  __Pyx_INCREF(__pyx_t_1);
  __pyx_v_criterion = ((struct __pyx_obj_7sklearn_4tree_5_tree_Criterion *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "sklearn/tree/_tree.pyx":870
 *         cdef Criterion criterion = self.criterion
 * 
 *         cdef SIZE_t* samples = self.samples             # <<<<<<<<<<<<<<
 *         self.start = start
 *         self.end = end
 */
  __pyx_t_2 = __pyx_v_self->samples;
  __pyx_v_samples = __pyx_t_2;

  /* "sklearn/tree/_tree.pyx":871
 * 
 *         cdef SIZE_t* samples = self.samples
 *         self.start = start             # <<<<<<<<<<<<<<
 *         self.end = end
 * 
 */
  __pyx_v_self->start = __pyx_v_start;

  /* "sklearn/tree/_tree.pyx":872
 *         cdef SIZE_t* samples = self.samples
 *         self.start = start
 *         self.end = end             # <<<<<<<<<<<<<<
 * 
 *         criterion.init(self.y,
 */
  __pyx_v_self->end = __pyx_v_end;

  /* "sklearn/tree/_tree.pyx":879
 *                        samples,
 *                        start,
 *                        end)             # <<<<<<<<<<<<<<
 * 
 *         impurity[0] =  criterion.node_impurity()
 */
  ((struct __pyx_vtabstruct_7sklearn_4tree_5_tree_Criterion *)__pyx_v_criterion->__pyx_vtab)->init(__pyx_v_criterion, __pyx_v_self->y, __pyx_v_self->y_stride, __pyx_v_self->sample_weight, __pyx_v_samples, __pyx_v_start, __pyx_v_end);

  /* "sklearn/tree/_tree.pyx":881
 *                        end)
 * 
 *         impurity[0] =  criterion.node_impurity()             # <<<<<<<<<<<<<<
 * 
 *     cdef void node_split(self, SIZE_t* pos, SIZE_t* feature, double* threshold):
 */
  (__pyx_v_impurity[0]) = ((struct __pyx_vtabstruct_7sklearn_4tree_5_tree_Criterion *)__pyx_v_criterion->__pyx_vtab)->node_impurity(__pyx_v_criterion);

  __Pyx_XDECREF((PyObject *)__pyx_v_criterion);
  __Pyx_RefNannyFinishContext();
}

/* "sklearn/tree/_tree.pyx":883
 *         impurity[0] =  criterion.node_impurity()
 * 
 *     cdef void node_split(self, SIZE_t* pos, SIZE_t* feature, double* threshold):             # <<<<<<<<<<<<<<
 *         """Find a split on node samples[start:end]."""
 *         pass
 */

static void __pyx_f_7sklearn_4tree_5_tree_8Splitter_node_split(CYTHON_UNUSED struct __pyx_obj_7sklearn_4tree_5_tree_Splitter *__pyx_v_self, CYTHON_UNUSED __pyx_t_7sklearn_4tree_5_tree_SIZE_t *__pyx_v_pos, CYTHON_UNUSED __pyx_t_7sklearn_4tree_5_tree_SIZE_t *__pyx_v_feature, CYTHON_UNUSED double *__pyx_v_threshold) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("node_split", 0);

  __Pyx_RefNannyFinishContext();
}

/* "sklearn/tree/_tree.pyx":887
 *         pass
 * 
 *     cdef void node_value(self, double* dest):             # <<<<<<<<<<<<<<
 *         """Copy the value of node samples[start:end] into dest."""
 *         self.criterion.node_value(dest)
 */

static void __pyx_f_7sklearn_4tree_5_tree_8Splitter_node_value(struct __pyx_obj_7sklearn_4tree_5_tree_Splitter *__pyx_v_self, double *__pyx_v_dest) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("node_value", 0);

  /* "sklearn/tree/_tree.pyx":889
 *     cdef void node_value(self, double* dest):
 *         """Copy the value of node samples[start:end] into dest."""
 *         self.criterion.node_value(dest)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  ((struct __pyx_vtabstruct_7sklearn_4tree_5_tree_Criterion *)__pyx_v_self->criterion->__pyx_vtab)->node_value(__pyx_v_self->criterion, __pyx_v_dest);

  __Pyx_RefNannyFinishContext();
}

/* Python wrapper */
static PyObject *__pyx_pw_7sklearn_4tree_5_tree_8Splitter_9criterion_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_7sklearn_4tree_5_tree_8Splitter_9criterion_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_7sklearn_4tree_5_tree_8Splitter_9criterion___get__(((struct __pyx_obj_7sklearn_4tree_5_tree_Splitter *)__pyx_v_self));
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "sklearn/tree/_tree.pxd":59
 * cdef class Splitter:
 *     # Internal structures
 *     cdef public Criterion criterion      # Impurity criterion             # <<<<<<<<<<<<<<
 *     cdef public SIZE_t max_features      # Number of features to test
 *     cdef public SIZE_t min_samples_leaf  # Min samples in a leaf
 */

static PyObject *__pyx_pf_7sklearn_4tree_5_tree_8Splitter_9criterion___get__(struct __pyx_obj_7sklearn_4tree_5_tree_Splitter *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__", 0);
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(((PyObject *)__pyx_v_self->criterion));
  __pyx_r = ((PyObject *)__pyx_v_self->criterion);
  goto __pyx_L0;

  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static int __pyx_pw_7sklearn_4tree_5_tree_8Splitter_9criterion_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value); /*proto*/
static int __pyx_pw_7sklearn_4tree_5_tree_8Splitter_9criterion_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__ (wrapper)", 0);
  __pyx_r = __pyx_pf_7sklearn_4tree_5_tree_8Splitter_9criterion_2__set__(((struct __pyx_obj_7sklearn_4tree_5_tree_Splitter *)__pyx_v_self), ((PyObject *)__pyx_v_value));
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_7sklearn_4tree_5_tree_8Splitter_9criterion_2__set__(struct __pyx_obj_7sklearn_4tree_5_tree_Splitter *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__set__", 0);
  if (!(likely(((__pyx_v_value) == Py_None) || likely(__Pyx_TypeTest(__pyx_v_value, __pyx_ptype_7sklearn_4tree_5_tree_Criterion))))) {__pyx_filename = __pyx_f[1]; __pyx_lineno = 59; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_INCREF(__pyx_v_value);
  __Pyx_GIVEREF(__pyx_v_value);
  __Pyx_GOTREF(__pyx_v_self->criterion);
  __Pyx_DECREF(((PyObject *)__pyx_v_self->criterion));
  __pyx_v_self->criterion = ((struct __pyx_obj_7sklearn_4tree_5_tree_Criterion *)__pyx_v_value);

  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_AddTraceback("sklearn.tree._tree.Splitter.criterion.__set__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static int __pyx_pw_7sklearn_4tree_5_tree_8Splitter_9criterion_5__del__(PyObject *__pyx_v_self); /*proto*/
static int __pyx_pw_7sklearn_4tree_5_tree_8Splitter_9criterion_5__del__(PyObject *__pyx_v_self) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__del__ (wrapper)", 0);
  __pyx_r = __pyx_pf_7sklearn_4tree_5_tree_8Splitter_9criterion_4__del__(((struct __pyx_obj_7sklearn_4tree_5_tree_Splitter *)__pyx_v_self));
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_7sklearn_4tree_5_tree_8Splitter_9criterion_4__del__(struct __pyx_obj_7sklearn_4tree_5_tree_Splitter *__pyx_v_self) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__del__", 0);
  __Pyx_INCREF(Py_None);
  __Pyx_GIVEREF(Py_None);
  __Pyx_GOTREF(__pyx_v_self->criterion);
  __Pyx_DECREF(((PyObject *)__pyx_v_self->criterion));
  __pyx_v_self->criterion = ((struct __pyx_obj_7sklearn_4tree_5_tree_Criterion *)Py_None);

  __pyx_r = 0;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_7sklearn_4tree_5_tree_8Splitter_12max_features_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_7sklearn_4tree_5_tree_8Splitter_12max_features_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_7sklearn_4tree_5_tree_8Splitter_12max_features___get__(((struct __pyx_obj_7sklearn_4tree_5_tree_Splitter *)__pyx_v_self));
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "sklearn/tree/_tree.pxd":60
 *     # Internal structures
 *     cdef public Criterion criterion      # Impurity criterion
 *     cdef public SIZE_t max_features      # Number of features to test             # <<<<<<<<<<<<<<
 *     cdef public SIZE_t min_samples_leaf  # Min samples in a leaf
 * 
 */

static PyObject *__pyx_pf_7sklearn_4tree_5_tree_8Splitter_12max_features___get__(struct __pyx_obj_7sklearn_4tree_5_tree_Splitter *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyInt_to_py_Py_intptr_t(__pyx_v_self->max_features); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[1]; __pyx_lineno = 60; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("sklearn.tree._tree.Splitter.max_features.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static int __pyx_pw_7sklearn_4tree_5_tree_8Splitter_12max_features_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value); /*proto*/
static int __pyx_pw_7sklearn_4tree_5_tree_8Splitter_12max_features_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__ (wrapper)", 0);
  __pyx_r = __pyx_pf_7sklearn_4tree_5_tree_8Splitter_12max_features_2__set__(((struct __pyx_obj_7sklearn_4tree_5_tree_Splitter *)__pyx_v_self), ((PyObject *)__pyx_v_value));
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_7sklearn_4tree_5_tree_8Splitter_12max_features_2__set__(struct __pyx_obj_7sklearn_4tree_5_tree_Splitter *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __pyx_t_7sklearn_4tree_5_tree_SIZE_t __pyx_t_1;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__set__", 0);
  __pyx_t_1 = __Pyx_PyInt_from_py_Py_intptr_t(__pyx_v_value); if (unlikely((__pyx_t_1 == (npy_intp)-1) && PyErr_Occurred())) {__pyx_filename = __pyx_f[1]; __pyx_lineno = 60; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __pyx_v_self->max_features = __pyx_t_1;

  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_AddTraceback("sklearn.tree._tree.Splitter.max_features.__set__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_7sklearn_4tree_5_tree_8Splitter_16min_samples_leaf_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_7sklearn_4tree_5_tree_8Splitter_16min_samples_leaf_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_7sklearn_4tree_5_tree_8Splitter_16min_samples_leaf___get__(((struct __pyx_obj_7sklearn_4tree_5_tree_Splitter *)__pyx_v_self));
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "sklearn/tree/_tree.pxd":61
 *     cdef public Criterion criterion      # Impurity criterion
 *     cdef public SIZE_t max_features      # Number of features to test
 *     cdef public SIZE_t min_samples_leaf  # Min samples in a leaf             # <<<<<<<<<<<<<<
 * 
 *     cdef object random_state             # Random state
 */

static PyObject *__pyx_pf_7sklearn_4tree_5_tree_8Splitter_16min_samples_leaf___get__(struct __pyx_obj_7sklearn_4tree_5_tree_Splitter *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyInt_to_py_Py_intptr_t(__pyx_v_self->min_samples_leaf); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[1]; __pyx_lineno = 61; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("sklearn.tree._tree.Splitter.min_samples_leaf.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static int __pyx_pw_7sklearn_4tree_5_tree_8Splitter_16min_samples_leaf_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value); /*proto*/
static int __pyx_pw_7sklearn_4tree_5_tree_8Splitter_16min_samples_leaf_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__ (wrapper)", 0);
  __pyx_r = __pyx_pf_7sklearn_4tree_5_tree_8Splitter_16min_samples_leaf_2__set__(((struct __pyx_obj_7sklearn_4tree_5_tree_Splitter *)__pyx_v_self), ((PyObject *)__pyx_v_value));
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_7sklearn_4tree_5_tree_8Splitter_16min_samples_leaf_2__set__(struct __pyx_obj_7sklearn_4tree_5_tree_Splitter *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __pyx_t_7sklearn_4tree_5_tree_SIZE_t __pyx_t_1;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__set__", 0);
  __pyx_t_1 = __Pyx_PyInt_from_py_Py_intptr_t(__pyx_v_value); if (unlikely((__pyx_t_1 == (npy_intp)-1) && PyErr_Occurred())) {__pyx_filename = __pyx_f[1]; __pyx_lineno = 61; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __pyx_v_self->min_samples_leaf = __pyx_t_1;

  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_AddTraceback("sklearn.tree._tree.Splitter.min_samples_leaf.__set__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_7sklearn_4tree_5_tree_12BestSplitter_1__reduce__(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static PyObject *__pyx_pw_7sklearn_4tree_5_tree_12BestSplitter_1__reduce__(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__reduce__ (wrapper)", 0);
  __pyx_r = __pyx_pf_7sklearn_4tree_5_tree_12BestSplitter___reduce__(((struct __pyx_obj_7sklearn_4tree_5_tree_BestSplitter *)__pyx_v_self));
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "sklearn/tree/_tree.pyx":894
 * cdef class BestSplitter(Splitter):
 *     """Splitter for finding the best split."""
 *     def __reduce__(self):             # <<<<<<<<<<<<<<
 *         return (BestSplitter, (self.criterion,
 *                                self.max_features,
 */

static PyObject *__pyx_pf_7sklearn_4tree_5_tree_12BestSplitter___reduce__(struct __pyx_obj_7sklearn_4tree_5_tree_BestSplitter *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__reduce__", 0);

  /* "sklearn/tree/_tree.pyx":895
 *     """Splitter for finding the best split."""
 *     def __reduce__(self):
 *         return (BestSplitter, (self.criterion,             # <<<<<<<<<<<<<<
 *                                self.max_features,
 *                                self.min_samples_leaf,
 */
  __Pyx_XDECREF(__pyx_r);

  /* "sklearn/tree/_tree.pyx":896
 *     def __reduce__(self):
 *         return (BestSplitter, (self.criterion,
 *                                self.max_features,             # <<<<<<<<<<<<<<
 *                                self.min_samples_leaf,
 *                                self.random_state), self.__getstate__())
 */
  __pyx_t_1 = __Pyx_PyInt_to_py_Py_intptr_t(__pyx_v_self->__pyx_base.max_features); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 896; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_1);

  /* "sklearn/tree/_tree.pyx":897
 *         return (BestSplitter, (self.criterion,
 *                                self.max_features,
 *                                self.min_samples_leaf,             # <<<<<<<<<<<<<<
 *                                self.random_state), self.__getstate__())
 * 
 */
  __pyx_t_2 = __Pyx_PyInt_to_py_Py_intptr_t(__pyx_v_self->__pyx_base.min_samples_leaf); if (unlikely(!__pyx_t_2)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 897; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_2);

  /* "sklearn/tree/_tree.pyx":898
 *                                self.max_features,
 *                                self.min_samples_leaf,
 *                                self.random_state), self.__getstate__())             # <<<<<<<<<<<<<<
 * 
 *     cdef void node_split(self, SIZE_t* pos, SIZE_t* feature, double* threshold):
 */
  __pyx_t_3 = PyTuple_New(4); if (unlikely(!__pyx_t_3)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 895; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_INCREF(((PyObject *)__pyx_v_self->__pyx_base.criterion));
  PyTuple_SET_ITEM(__pyx_t_3, 0, ((PyObject *)__pyx_v_self->__pyx_base.criterion));
  __Pyx_GIVEREF(((PyObject *)__pyx_v_self->__pyx_base.criterion));
  PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  PyTuple_SET_ITEM(__pyx_t_3, 2, __pyx_t_2);
  __Pyx_GIVEREF(__pyx_t_2);
  __Pyx_INCREF(__pyx_v_self->__pyx_base.random_state);
  PyTuple_SET_ITEM(__pyx_t_3, 3, __pyx_v_self->__pyx_base.random_state);
  __Pyx_GIVEREF(__pyx_v_self->__pyx_base.random_state);
  __pyx_t_1 = 0;
  __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s____getstate__); if (unlikely(!__pyx_t_2)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 898; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_1 = PyObject_Call(__pyx_t_2, ((PyObject *)__pyx_empty_tuple), NULL); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 898; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = PyTuple_New(3); if (unlikely(!__pyx_t_2)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 895; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_INCREF(((PyObject *)((PyObject*)__pyx_ptype_7sklearn_4tree_5_tree_BestSplitter)));
  PyTuple_SET_ITEM(__pyx_t_2, 0, ((PyObject *)((PyObject*)__pyx_ptype_7sklearn_4tree_5_tree_BestSplitter)));
  __Pyx_GIVEREF(((PyObject *)((PyObject*)__pyx_ptype_7sklearn_4tree_5_tree_BestSplitter)));
  PyTuple_SET_ITEM(__pyx_t_2, 1, ((PyObject *)__pyx_t_3));
  __Pyx_GIVEREF(((PyObject *)__pyx_t_3));
  PyTuple_SET_ITEM(__pyx_t_2, 2, __pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __pyx_t_3 = 0;
  __pyx_t_1 = 0;
  __pyx_r = ((PyObject *)__pyx_t_2);
  __pyx_t_2 = 0;
  goto __pyx_L0;

  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("sklearn.tree._tree.BestSplitter.__reduce__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "sklearn/tree/_tree.pyx":900
 *                                self.random_state), self.__getstate__())
 * 
 *     cdef void node_split(self, SIZE_t* pos, SIZE_t* feature, double* threshold):             # <<<<<<<<<<<<<<
 *         """Find the best split on node samples[start:end]."""
 *         # Find the best split
 */

static void __pyx_f_7sklearn_4tree_5_tree_12BestSplitter_node_split(struct __pyx_obj_7sklearn_4tree_5_tree_BestSplitter *__pyx_v_self, __pyx_t_7sklearn_4tree_5_tree_SIZE_t *__pyx_v_pos, __pyx_t_7sklearn_4tree_5_tree_SIZE_t *__pyx_v_feature, double *__pyx_v_threshold) {
  struct __pyx_obj_7sklearn_4tree_5_tree_Criterion *__pyx_v_criterion = 0;
  __pyx_t_7sklearn_4tree_5_tree_SIZE_t *__pyx_v_samples;
  __pyx_t_7sklearn_4tree_5_tree_SIZE_t __pyx_v_start;
  __pyx_t_7sklearn_4tree_5_tree_SIZE_t __pyx_v_end;
  __pyx_t_7sklearn_4tree_5_tree_SIZE_t *__pyx_v_features;
  __pyx_t_7sklearn_4tree_5_tree_SIZE_t __pyx_v_n_features;
  PyArrayObject *__pyx_v_X = 0;
  __pyx_t_7sklearn_4tree_5_tree_SIZE_t __pyx_v_max_features;
  __pyx_t_7sklearn_4tree_5_tree_SIZE_t __pyx_v_min_samples_leaf;
  unsigned int *__pyx_v_random_state;
  double __pyx_v_best_impurity;
  __pyx_t_7sklearn_4tree_5_tree_SIZE_t __pyx_v_best_pos;
  __pyx_t_7sklearn_4tree_5_tree_SIZE_t __pyx_v_best_feature;
  double __pyx_v_best_threshold;
  double __pyx_v_current_impurity;
  __pyx_t_7sklearn_4tree_5_tree_SIZE_t __pyx_v_current_pos;
  __pyx_t_7sklearn_4tree_5_tree_SIZE_t __pyx_v_current_feature;
  double __pyx_v_current_threshold;
  __pyx_t_7sklearn_4tree_5_tree_SIZE_t __pyx_v_f_idx;
  __pyx_t_7sklearn_4tree_5_tree_SIZE_t __pyx_v_f_i;
  __pyx_t_7sklearn_4tree_5_tree_SIZE_t __pyx_v_f_j;
  __pyx_t_7sklearn_4tree_5_tree_SIZE_t __pyx_v_p;
  __pyx_t_7sklearn_4tree_5_tree_SIZE_t __pyx_v_tmp;
  __pyx_t_7sklearn_4tree_5_tree_SIZE_t __pyx_v_visited_features;
  CYTHON_UNUSED __pyx_t_7sklearn_4tree_5_tree_SIZE_t __pyx_v_partition_start;
  __pyx_t_7sklearn_4tree_5_tree_SIZE_t __pyx_v_partition_end;
  __Pyx_LocalBuf_ND __pyx_pybuffernd_X;
  __Pyx_Buffer __pyx_pybuffer_X;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __pyx_t_7sklearn_4tree_5_tree_SIZE_t *__pyx_t_2;
  __pyx_t_7sklearn_4tree_5_tree_SIZE_t __pyx_t_3;
  int __pyx_t_4;
  __pyx_t_7sklearn_4tree_5_tree_SIZE_t __pyx_t_5;
  __pyx_t_7sklearn_4tree_5_tree_SIZE_t __pyx_t_6;
  __pyx_t_7sklearn_4tree_5_tree_SIZE_t __pyx_t_7;
  __pyx_t_7sklearn_4tree_5_tree_SIZE_t __pyx_t_8;
  int __pyx_t_9;
  int __pyx_t_10;
  __pyx_t_7sklearn_4tree_5_tree_SIZE_t __pyx_t_11;
  __pyx_t_7sklearn_4tree_5_tree_SIZE_t __pyx_t_12;
  __pyx_t_7sklearn_4tree_5_tree_SIZE_t __pyx_t_13;
  __pyx_t_7sklearn_4tree_5_tree_SIZE_t __pyx_t_14;
  __pyx_t_7sklearn_4tree_5_tree_SIZE_t __pyx_t_15;
  __pyx_t_7sklearn_4tree_5_tree_SIZE_t __pyx_t_16;
  __pyx_t_7sklearn_4tree_5_tree_SIZE_t __pyx_t_17;
  __pyx_t_7sklearn_4tree_5_tree_SIZE_t __pyx_t_18;
  __pyx_t_7sklearn_4tree_5_tree_SIZE_t __pyx_t_19;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("node_split", 0);
  __pyx_pybuffer_X.pybuffer.buf = NULL;
  __pyx_pybuffer_X.refcount = 0;
  __pyx_pybuffernd_X.data = NULL;
  __pyx_pybuffernd_X.rcbuffer = &__pyx_pybuffer_X;

  /* "sklearn/tree/_tree.pyx":903
 *         """Find the best split on node samples[start:end]."""
 *         # Find the best split
 *         cdef Criterion criterion = self.criterion             # <<<<<<<<<<<<<<
 *         cdef SIZE_t* samples = self.samples
 *         cdef SIZE_t start = self.start
 */
  __pyx_t_1 = ((PyObject *)__pyx_v_self->__pyx_base.criterion);
  __Pyx_INCREF(__pyx_t_1);
  __pyx_v_criterion = ((struct __pyx_obj_7sklearn_4tree_5_tree_Criterion *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "sklearn/tree/_tree.pyx":904
 *         # Find the best split
 *         cdef Criterion criterion = self.criterion
 *         cdef SIZE_t* samples = self.samples             # <<<<<<<<<<<<<<
 *         cdef SIZE_t start = self.start
 *         cdef SIZE_t end = self.end
 */
  __pyx_t_2 = __pyx_v_self->__pyx_base.samples;
  __pyx_v_samples = __pyx_t_2;

  /* "sklearn/tree/_tree.pyx":905
 *         cdef Criterion criterion = self.criterion
 *         cdef SIZE_t* samples = self.samples
 *         cdef SIZE_t start = self.start             # <<<<<<<<<<<<<<
 *         cdef SIZE_t end = self.end
 * 
 */
  __pyx_t_3 = __pyx_v_self->__pyx_base.start;
  __pyx_v_start = __pyx_t_3;

  /* "sklearn/tree/_tree.pyx":906
 *         cdef SIZE_t* samples = self.samples
 *         cdef SIZE_t start = self.start
 *         cdef SIZE_t end = self.end             # <<<<<<<<<<<<<<
 * 
 *         cdef SIZE_t* features = self.features
 */
  __pyx_t_3 = __pyx_v_self->__pyx_base.end;
  __pyx_v_end = __pyx_t_3;

  /* "sklearn/tree/_tree.pyx":908
 *         cdef SIZE_t end = self.end
 * 
 *         cdef SIZE_t* features = self.features             # <<<<<<<<<<<<<<
 *         cdef SIZE_t n_features = self.n_features
 * 
 */
  __pyx_t_2 = __pyx_v_self->__pyx_base.features;
  __pyx_v_features = __pyx_t_2;

  /* "sklearn/tree/_tree.pyx":909
 * 
 *         cdef SIZE_t* features = self.features
 *         cdef SIZE_t n_features = self.n_features             # <<<<<<<<<<<<<<
 * 
 *         cdef np.ndarray[DTYPE_t, ndim=2, mode="c"] X = self.X
 */
  __pyx_t_3 = __pyx_v_self->__pyx_base.n_features;
  __pyx_v_n_features = __pyx_t_3;

  /* "sklearn/tree/_tree.pyx":911
 *         cdef SIZE_t n_features = self.n_features
 * 
 *         cdef np.ndarray[DTYPE_t, ndim=2, mode="c"] X = self.X             # <<<<<<<<<<<<<<
 *         cdef SIZE_t max_features = self.max_features
 *         cdef SIZE_t min_samples_leaf = self.min_samples_leaf
 */
  __pyx_t_1 = ((PyObject *)__pyx_v_self->__pyx_base.X);
  __Pyx_INCREF(__pyx_t_1);
  {
    __Pyx_BufFmt_StackElem __pyx_stack[1];
    if (unlikely(__Pyx_GetBufferAndValidate(&__pyx_pybuffernd_X.rcbuffer->pybuffer, (PyObject*)((PyArrayObject *)__pyx_t_1), &__Pyx_TypeInfo_nn___pyx_t_7sklearn_4tree_5_tree_DTYPE_t, PyBUF_FORMAT| PyBUF_C_CONTIGUOUS, 2, 0, __pyx_stack) == -1)) {
      __pyx_v_X = ((PyArrayObject *)Py_None); __Pyx_INCREF(Py_None); __pyx_pybuffernd_X.rcbuffer->pybuffer.buf = NULL;
      {__pyx_filename = __pyx_f[0]; __pyx_lineno = 911; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    } else {__pyx_pybuffernd_X.diminfo[0].strides = __pyx_pybuffernd_X.rcbuffer->pybuffer.strides[0]; __pyx_pybuffernd_X.diminfo[0].shape = __pyx_pybuffernd_X.rcbuffer->pybuffer.shape[0]; __pyx_pybuffernd_X.diminfo[1].strides = __pyx_pybuffernd_X.rcbuffer->pybuffer.strides[1]; __pyx_pybuffernd_X.diminfo[1].shape = __pyx_pybuffernd_X.rcbuffer->pybuffer.shape[1];
    }
  }
  __pyx_v_X = ((PyArrayObject *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "sklearn/tree/_tree.pyx":912
 * 
 *         cdef np.ndarray[DTYPE_t, ndim=2, mode="c"] X = self.X
 *         cdef SIZE_t max_features = self.max_features             # <<<<<<<<<<<<<<
 *         cdef SIZE_t min_samples_leaf = self.min_samples_leaf
 *         cdef unsigned int* random_state = &self.rand_r_state
 */
  __pyx_t_3 = __pyx_v_self->__pyx_base.max_features;
  __pyx_v_max_features = __pyx_t_3;

  /* "sklearn/tree/_tree.pyx":913
 *         cdef np.ndarray[DTYPE_t, ndim=2, mode="c"] X = self.X
 *         cdef SIZE_t max_features = self.max_features
 *         cdef SIZE_t min_samples_leaf = self.min_samples_leaf             # <<<<<<<<<<<<<<
 *         cdef unsigned int* random_state = &self.rand_r_state
 * 
 */
  __pyx_t_3 = __pyx_v_self->__pyx_base.min_samples_leaf;
  __pyx_v_min_samples_leaf = __pyx_t_3;

  /* "sklearn/tree/_tree.pyx":914
 *         cdef SIZE_t max_features = self.max_features
 *         cdef SIZE_t min_samples_leaf = self.min_samples_leaf
 *         cdef unsigned int* random_state = &self.rand_r_state             # <<<<<<<<<<<<<<
 * 
 *         cdef double best_impurity = INFINITY
 */
  __pyx_v_random_state = (&__pyx_v_self->__pyx_base.rand_r_state);

  /* "sklearn/tree/_tree.pyx":916
 *         cdef unsigned int* random_state = &self.rand_r_state
 * 
 *         cdef double best_impurity = INFINITY             # <<<<<<<<<<<<<<
 *         cdef SIZE_t best_pos = end
 *         cdef SIZE_t best_feature
 */
  __pyx_v_best_impurity = __pyx_v_7sklearn_4tree_5_tree_INFINITY;

  /* "sklearn/tree/_tree.pyx":917
 * 
 *         cdef double best_impurity = INFINITY
 *         cdef SIZE_t best_pos = end             # <<<<<<<<<<<<<<
 *         cdef SIZE_t best_feature
 *         cdef double best_threshold
 */
  __pyx_v_best_pos = __pyx_v_end;

  /* "sklearn/tree/_tree.pyx":927
 * 
 *         cdef SIZE_t f_idx, f_i, f_j, p, tmp
 *         cdef SIZE_t visited_features = 0             # <<<<<<<<<<<<<<
 * 
 *         cdef SIZE_t partition_start
 */
  __pyx_v_visited_features = 0;

  /* "sklearn/tree/_tree.pyx":932
 *         cdef SIZE_t partition_end
 * 
 *         for f_idx from 0 <= f_idx < n_features:             # <<<<<<<<<<<<<<
 *             # Draw a feature at random
 *             f_i = n_features - f_idx - 1
 */
  __pyx_t_3 = __pyx_v_n_features;
  for (__pyx_v_f_idx = 0; __pyx_v_f_idx < __pyx_t_3; __pyx_v_f_idx++) {

    /* "sklearn/tree/_tree.pyx":934
 *         for f_idx from 0 <= f_idx < n_features:
 *             # Draw a feature at random
 *             f_i = n_features - f_idx - 1             # <<<<<<<<<<<<<<
 *             f_j = rand_int(n_features - f_idx, random_state)
 * 
 */
    __pyx_v_f_i = ((__pyx_v_n_features - __pyx_v_f_idx) - 1);

    /* "sklearn/tree/_tree.pyx":935
 *             # Draw a feature at random
 *             f_i = n_features - f_idx - 1
 *             f_j = rand_int(n_features - f_idx, random_state)             # <<<<<<<<<<<<<<
 * 
 *             tmp = features[f_i]
 */
    __pyx_v_f_j = __pyx_f_7sklearn_4tree_5_tree_rand_int((__pyx_v_n_features - __pyx_v_f_idx), __pyx_v_random_state);

    /* "sklearn/tree/_tree.pyx":937
 *             f_j = rand_int(n_features - f_idx, random_state)
 * 
 *             tmp = features[f_i]             # <<<<<<<<<<<<<<
 *             features[f_i] = features[f_j]
 *             features[f_j] = tmp
 */
    __pyx_v_tmp = (__pyx_v_features[__pyx_v_f_i]);

    /* "sklearn/tree/_tree.pyx":938
 * 
 *             tmp = features[f_i]
 *             features[f_i] = features[f_j]             # <<<<<<<<<<<<<<
 *             features[f_j] = tmp
 * 
 */
    (__pyx_v_features[__pyx_v_f_i]) = (__pyx_v_features[__pyx_v_f_j]);

    /* "sklearn/tree/_tree.pyx":939
 *             tmp = features[f_i]
 *             features[f_i] = features[f_j]
 *             features[f_j] = tmp             # <<<<<<<<<<<<<<
 * 
 *             current_feature = features[f_i]
 */
    (__pyx_v_features[__pyx_v_f_j]) = __pyx_v_tmp;

    /* "sklearn/tree/_tree.pyx":941
 *             features[f_j] = tmp
 * 
 *             current_feature = features[f_i]             # <<<<<<<<<<<<<<
 * 
 *             # Sort samples along that feature
 */
    __pyx_v_current_feature = (__pyx_v_features[__pyx_v_f_i]);

    /* "sklearn/tree/_tree.pyx":944
 * 
 *             # Sort samples along that feature
 *             sort(X, current_feature, samples+start, end-start)             # <<<<<<<<<<<<<<
 * 
 *             # Evaluate all splits
 */
    __pyx_f_7sklearn_4tree_5_tree_sort(((PyArrayObject *)__pyx_v_X), __pyx_v_current_feature, (__pyx_v_samples + __pyx_v_start), (__pyx_v_end - __pyx_v_start));

    /* "sklearn/tree/_tree.pyx":947
 * 
 *             # Evaluate all splits
 *             criterion.reset()             # <<<<<<<<<<<<<<
 *             p = start
 * 
 */
    ((struct __pyx_vtabstruct_7sklearn_4tree_5_tree_Criterion *)__pyx_v_criterion->__pyx_vtab)->reset(__pyx_v_criterion);

    /* "sklearn/tree/_tree.pyx":948
 *             # Evaluate all splits
 *             criterion.reset()
 *             p = start             # <<<<<<<<<<<<<<
 * 
 *             while p < end:
 */
    __pyx_v_p = __pyx_v_start;

    /* "sklearn/tree/_tree.pyx":950
 *             p = start
 * 
 *             while p < end:             # <<<<<<<<<<<<<<
 *                 while ((p + 1 < end) and
 *                        (X[samples[p + 1], current_feature] <= X[samples[p], current_feature] + 1e-7)):
 */
    while (1) {
      __pyx_t_4 = ((__pyx_v_p < __pyx_v_end) != 0);
      if (!__pyx_t_4) break;

      /* "sklearn/tree/_tree.pyx":951
 * 
 *             while p < end:
 *                 while ((p + 1 < end) and             # <<<<<<<<<<<<<<
 *                        (X[samples[p + 1], current_feature] <= X[samples[p], current_feature] + 1e-7)):
 *                     p += 1
 */
      while (1) {
        __pyx_t_4 = (((__pyx_v_p + 1) < __pyx_v_end) != 0);
        if (__pyx_t_4) {

          /* "sklearn/tree/_tree.pyx":952
 *             while p < end:
 *                 while ((p + 1 < end) and
 *                        (X[samples[p + 1], current_feature] <= X[samples[p], current_feature] + 1e-7)):             # <<<<<<<<<<<<<<
 *                     p += 1
 * 
 */
          __pyx_t_5 = (__pyx_v_samples[(__pyx_v_p + 1)]);
          __pyx_t_6 = __pyx_v_current_feature;
          __pyx_t_7 = (__pyx_v_samples[__pyx_v_p]);
          __pyx_t_8 = __pyx_v_current_feature;
          __pyx_t_9 = (((*__Pyx_BufPtrCContig2d(__pyx_t_7sklearn_4tree_5_tree_DTYPE_t *, __pyx_pybuffernd_X.rcbuffer->pybuffer.buf, __pyx_t_5, __pyx_pybuffernd_X.diminfo[0].strides, __pyx_t_6, __pyx_pybuffernd_X.diminfo[1].strides)) <= ((*__Pyx_BufPtrCContig2d(__pyx_t_7sklearn_4tree_5_tree_DTYPE_t *, __pyx_pybuffernd_X.rcbuffer->pybuffer.buf, __pyx_t_7, __pyx_pybuffernd_X.diminfo[0].strides, __pyx_t_8, __pyx_pybuffernd_X.diminfo[1].strides)) + 1e-7)) != 0);
          __pyx_t_10 = __pyx_t_9;
        } else {
          __pyx_t_10 = __pyx_t_4;
        }
        if (!__pyx_t_10) break;

        /* "sklearn/tree/_tree.pyx":953
 *                 while ((p + 1 < end) and
 *                        (X[samples[p + 1], current_feature] <= X[samples[p], current_feature] + 1e-7)):
 *                     p += 1             # <<<<<<<<<<<<<<
 * 
 *                 # p + 1 >= end or X[samples[p + 1], current_feature] > X[samples[p], current_feature]
 */
        __pyx_v_p = (__pyx_v_p + 1);
      }

      /* "sklearn/tree/_tree.pyx":956
 * 
 *                 # p + 1 >= end or X[samples[p + 1], current_feature] > X[samples[p], current_feature]
 *                 p += 1             # <<<<<<<<<<<<<<
 *                 # p >= end or X[samples[p], current_feature] > X[samples[p - 1], current_feature]
 * 
 */
      __pyx_v_p = (__pyx_v_p + 1);

      /* "sklearn/tree/_tree.pyx":959
 *                 # p >= end or X[samples[p], current_feature] > X[samples[p - 1], current_feature]
 * 
 *                 if p < end:             # <<<<<<<<<<<<<<
 *                     current_pos = p
 * 
 */
      __pyx_t_10 = ((__pyx_v_p < __pyx_v_end) != 0);
      if (__pyx_t_10) {

        /* "sklearn/tree/_tree.pyx":960
 * 
 *                 if p < end:
 *                     current_pos = p             # <<<<<<<<<<<<<<
 * 
 *                     # Reject if min_samples_leaf is not guaranteed
 */
        __pyx_v_current_pos = __pyx_v_p;

        /* "sklearn/tree/_tree.pyx":963
 * 
 *                     # Reject if min_samples_leaf is not guaranteed
 *                     if (((current_pos - start) < min_samples_leaf) or             # <<<<<<<<<<<<<<
 *                         ((end - current_pos) < min_samples_leaf)):
 *                        continue
 */
        __pyx_t_10 = (((__pyx_v_current_pos - __pyx_v_start) < __pyx_v_min_samples_leaf) != 0);
        if (!__pyx_t_10) {

          /* "sklearn/tree/_tree.pyx":964
 *                     # Reject if min_samples_leaf is not guaranteed
 *                     if (((current_pos - start) < min_samples_leaf) or
 *                         ((end - current_pos) < min_samples_leaf)):             # <<<<<<<<<<<<<<
 *                        continue
 * 
 */
          __pyx_t_4 = (((__pyx_v_end - __pyx_v_current_pos) < __pyx_v_min_samples_leaf) != 0);
          __pyx_t_9 = __pyx_t_4;
        } else {
          __pyx_t_9 = __pyx_t_10;
        }
        if (__pyx_t_9) {

          /* "sklearn/tree/_tree.pyx":965
 *                     if (((current_pos - start) < min_samples_leaf) or
 *                         ((end - current_pos) < min_samples_leaf)):
 *                        continue             # <<<<<<<<<<<<<<
 * 
 *                     criterion.update(current_pos)
 */
          goto __pyx_L5_continue;
          goto __pyx_L10;
        }
        __pyx_L10:;

        /* "sklearn/tree/_tree.pyx":967
 *                        continue
 * 
 *                     criterion.update(current_pos)             # <<<<<<<<<<<<<<
 *                     current_impurity = criterion.children_impurity()
 * 
 */
        ((struct __pyx_vtabstruct_7sklearn_4tree_5_tree_Criterion *)__pyx_v_criterion->__pyx_vtab)->update(__pyx_v_criterion, __pyx_v_current_pos);

        /* "sklearn/tree/_tree.pyx":968
 * 
 *                     criterion.update(current_pos)
 *                     current_impurity = criterion.children_impurity()             # <<<<<<<<<<<<<<
 * 
 *                     if current_impurity < (best_impurity - 1e-7):
 */
        __pyx_v_current_impurity = ((struct __pyx_vtabstruct_7sklearn_4tree_5_tree_Criterion *)__pyx_v_criterion->__pyx_vtab)->children_impurity(__pyx_v_criterion);

        /* "sklearn/tree/_tree.pyx":970
 *                     current_impurity = criterion.children_impurity()
 * 
 *                     if current_impurity < (best_impurity - 1e-7):             # <<<<<<<<<<<<<<
 *                         best_impurity = current_impurity
 *                         best_pos = current_pos
 */
        __pyx_t_9 = ((__pyx_v_current_impurity < (__pyx_v_best_impurity - 1e-7)) != 0);
        if (__pyx_t_9) {

          /* "sklearn/tree/_tree.pyx":971
 * 
 *                     if current_impurity < (best_impurity - 1e-7):
 *                         best_impurity = current_impurity             # <<<<<<<<<<<<<<
 *                         best_pos = current_pos
 *                         best_feature = current_feature
 */
          __pyx_v_best_impurity = __pyx_v_current_impurity;

          /* "sklearn/tree/_tree.pyx":972
 *                     if current_impurity < (best_impurity - 1e-7):
 *                         best_impurity = current_impurity
 *                         best_pos = current_pos             # <<<<<<<<<<<<<<
 *                         best_feature = current_feature
 * 
 */
          __pyx_v_best_pos = __pyx_v_current_pos;

          /* "sklearn/tree/_tree.pyx":973
 *                         best_impurity = current_impurity
 *                         best_pos = current_pos
 *                         best_feature = current_feature             # <<<<<<<<<<<<<<
 * 
 *                         current_threshold = (X[samples[p - 1], current_feature] + X[samples[p], current_feature]) / 2.0
 */
          __pyx_v_best_feature = __pyx_v_current_feature;

          /* "sklearn/tree/_tree.pyx":975
 *                         best_feature = current_feature
 * 
 *                         current_threshold = (X[samples[p - 1], current_feature] + X[samples[p], current_feature]) / 2.0             # <<<<<<<<<<<<<<
 *                         if current_threshold == X[samples[p], current_feature]:
 *                             current_threshold = X[samples[p - 1], current_feature]
 */
          __pyx_t_11 = (__pyx_v_samples[(__pyx_v_p - 1)]);
          __pyx_t_12 = __pyx_v_current_feature;
          __pyx_t_13 = (__pyx_v_samples[__pyx_v_p]);
          __pyx_t_14 = __pyx_v_current_feature;
          __pyx_v_current_threshold = (((*__Pyx_BufPtrCContig2d(__pyx_t_7sklearn_4tree_5_tree_DTYPE_t *, __pyx_pybuffernd_X.rcbuffer->pybuffer.buf, __pyx_t_11, __pyx_pybuffernd_X.diminfo[0].strides, __pyx_t_12, __pyx_pybuffernd_X.diminfo[1].strides)) + (*__Pyx_BufPtrCContig2d(__pyx_t_7sklearn_4tree_5_tree_DTYPE_t *, __pyx_pybuffernd_X.rcbuffer->pybuffer.buf, __pyx_t_13, __pyx_pybuffernd_X.diminfo[0].strides, __pyx_t_14, __pyx_pybuffernd_X.diminfo[1].strides))) / 2.0);

          /* "sklearn/tree/_tree.pyx":976
 * 
 *                         current_threshold = (X[samples[p - 1], current_feature] + X[samples[p], current_feature]) / 2.0
 *                         if current_threshold == X[samples[p], current_feature]:             # <<<<<<<<<<<<<<
 *                             current_threshold = X[samples[p - 1], current_feature]
 * 
 */
          __pyx_t_15 = (__pyx_v_samples[__pyx_v_p]);
          __pyx_t_16 = __pyx_v_current_feature;
          __pyx_t_9 = ((__pyx_v_current_threshold == (*__Pyx_BufPtrCContig2d(__pyx_t_7sklearn_4tree_5_tree_DTYPE_t *, __pyx_pybuffernd_X.rcbuffer->pybuffer.buf, __pyx_t_15, __pyx_pybuffernd_X.diminfo[0].strides, __pyx_t_16, __pyx_pybuffernd_X.diminfo[1].strides))) != 0);
          if (__pyx_t_9) {

            /* "sklearn/tree/_tree.pyx":977
 *                         current_threshold = (X[samples[p - 1], current_feature] + X[samples[p], current_feature]) / 2.0
 *                         if current_threshold == X[samples[p], current_feature]:
 *                             current_threshold = X[samples[p - 1], current_feature]             # <<<<<<<<<<<<<<
 * 
 *                         best_threshold = current_threshold
 */
            __pyx_t_17 = (__pyx_v_samples[(__pyx_v_p - 1)]);
            __pyx_t_18 = __pyx_v_current_feature;
            __pyx_v_current_threshold = (*__Pyx_BufPtrCContig2d(__pyx_t_7sklearn_4tree_5_tree_DTYPE_t *, __pyx_pybuffernd_X.rcbuffer->pybuffer.buf, __pyx_t_17, __pyx_pybuffernd_X.diminfo[0].strides, __pyx_t_18, __pyx_pybuffernd_X.diminfo[1].strides));
            goto __pyx_L12;
          }
          __pyx_L12:;

          /* "sklearn/tree/_tree.pyx":979
 *                             current_threshold = X[samples[p - 1], current_feature]
 * 
 *                         best_threshold = current_threshold             # <<<<<<<<<<<<<<
 * 
 *             if best_pos == end: # No valid split was ever found
 */
          __pyx_v_best_threshold = __pyx_v_current_threshold;
          goto __pyx_L11;
        }
        __pyx_L11:;
        goto __pyx_L9;
      }
      __pyx_L9:;
      __pyx_L5_continue:;
    }

    /* "sklearn/tree/_tree.pyx":981
 *                         best_threshold = current_threshold
 * 
 *             if best_pos == end: # No valid split was ever found             # <<<<<<<<<<<<<<
 *                 continue
 * 
 */
    __pyx_t_9 = ((__pyx_v_best_pos == __pyx_v_end) != 0);
    if (__pyx_t_9) {

      /* "sklearn/tree/_tree.pyx":982
 * 
 *             if best_pos == end: # No valid split was ever found
 *                 continue             # <<<<<<<<<<<<<<
 * 
 *             # Count one more visited feature
 */
      goto __pyx_L3_continue;
      goto __pyx_L13;
    }
    __pyx_L13:;

    /* "sklearn/tree/_tree.pyx":985
 * 
 *             # Count one more visited feature
 *             visited_features += 1             # <<<<<<<<<<<<<<
 * 
 *             if visited_features >= max_features:
 */
    __pyx_v_visited_features = (__pyx_v_visited_features + 1);

    /* "sklearn/tree/_tree.pyx":987
 *             visited_features += 1
 * 
 *             if visited_features >= max_features:             # <<<<<<<<<<<<<<
 *                 break
 * 
 */
    __pyx_t_9 = ((__pyx_v_visited_features >= __pyx_v_max_features) != 0);
    if (__pyx_t_9) {

      /* "sklearn/tree/_tree.pyx":988
 * 
 *             if visited_features >= max_features:
 *                 break             # <<<<<<<<<<<<<<
 * 
 *         # Reorganize samples into samples[start:best_pos] + samples[best_pos:end]
 */
      goto __pyx_L4_break;
      goto __pyx_L14;
    }
    __pyx_L14:;
    __pyx_L3_continue:;
  }
  __pyx_L4_break:;

  /* "sklearn/tree/_tree.pyx":991
 * 
 *         # Reorganize samples into samples[start:best_pos] + samples[best_pos:end]
 *         if best_pos < end:             # <<<<<<<<<<<<<<
 *             partition_start = start
 *             partition_end = end
 */
  __pyx_t_9 = ((__pyx_v_best_pos < __pyx_v_end) != 0);
  if (__pyx_t_9) {

    /* "sklearn/tree/_tree.pyx":992
 *         # Reorganize samples into samples[start:best_pos] + samples[best_pos:end]
 *         if best_pos < end:
 *             partition_start = start             # <<<<<<<<<<<<<<
 *             partition_end = end
 *             p = start
 */
    __pyx_v_partition_start = __pyx_v_start;

    /* "sklearn/tree/_tree.pyx":993
 *         if best_pos < end:
 *             partition_start = start
 *             partition_end = end             # <<<<<<<<<<<<<<
 *             p = start
 * 
 */
    __pyx_v_partition_end = __pyx_v_end;

    /* "sklearn/tree/_tree.pyx":994
 *             partition_start = start
 *             partition_end = end
 *             p = start             # <<<<<<<<<<<<<<
 * 
 *             while p < partition_end:
 */
    __pyx_v_p = __pyx_v_start;

    /* "sklearn/tree/_tree.pyx":996
 *             p = start
 * 
 *             while p < partition_end:             # <<<<<<<<<<<<<<
 *                 if X[samples[p], best_feature] <= best_threshold:
 *                     p += 1
 */
    while (1) {
      __pyx_t_9 = ((__pyx_v_p < __pyx_v_partition_end) != 0);
      if (!__pyx_t_9) break;

      /* "sklearn/tree/_tree.pyx":997
 * 
 *             while p < partition_end:
 *                 if X[samples[p], best_feature] <= best_threshold:             # <<<<<<<<<<<<<<
 *                     p += 1
 * 
 */
      __pyx_t_3 = (__pyx_v_samples[__pyx_v_p]);
      __pyx_t_19 = __pyx_v_best_feature;
      __pyx_t_9 = (((*__Pyx_BufPtrCContig2d(__pyx_t_7sklearn_4tree_5_tree_DTYPE_t *, __pyx_pybuffernd_X.rcbuffer->pybuffer.buf, __pyx_t_3, __pyx_pybuffernd_X.diminfo[0].strides, __pyx_t_19, __pyx_pybuffernd_X.diminfo[1].strides)) <= __pyx_v_best_threshold) != 0);
      if (__pyx_t_9) {

        /* "sklearn/tree/_tree.pyx":998
 *             while p < partition_end:
 *                 if X[samples[p], best_feature] <= best_threshold:
 *                     p += 1             # <<<<<<<<<<<<<<
 * 
 *                 else:
 */
        __pyx_v_p = (__pyx_v_p + 1);
        goto __pyx_L18;
      }
      /*else*/ {

        /* "sklearn/tree/_tree.pyx":1001
 * 
 *                 else:
 *                     partition_end -= 1             # <<<<<<<<<<<<<<
 * 
 *                     tmp = samples[partition_end]
 */
        __pyx_v_partition_end = (__pyx_v_partition_end - 1);

        /* "sklearn/tree/_tree.pyx":1003
 *                     partition_end -= 1
 * 
 *                     tmp = samples[partition_end]             # <<<<<<<<<<<<<<
 *                     samples[partition_end] = samples[p]
 *                     samples[p] = tmp
 */
        __pyx_v_tmp = (__pyx_v_samples[__pyx_v_partition_end]);

        /* "sklearn/tree/_tree.pyx":1004
 * 
 *                     tmp = samples[partition_end]
 *                     samples[partition_end] = samples[p]             # <<<<<<<<<<<<<<
 *                     samples[p] = tmp
 * 
 */
        (__pyx_v_samples[__pyx_v_partition_end]) = (__pyx_v_samples[__pyx_v_p]);

        /* "sklearn/tree/_tree.pyx":1005
 *                     tmp = samples[partition_end]
 *                     samples[partition_end] = samples[p]
 *                     samples[p] = tmp             # <<<<<<<<<<<<<<
 * 
 *         # Return values
 */
        (__pyx_v_samples[__pyx_v_p]) = __pyx_v_tmp;
      }
      __pyx_L18:;
    }
    goto __pyx_L15;
  }
  __pyx_L15:;

  /* "sklearn/tree/_tree.pyx":1008
 * 
 *         # Return values
 *         pos[0] = best_pos             # <<<<<<<<<<<<<<
 *         feature[0] = best_feature
 *         threshold[0] = best_threshold
 */
  (__pyx_v_pos[0]) = __pyx_v_best_pos;

  /* "sklearn/tree/_tree.pyx":1009
 *         # Return values
 *         pos[0] = best_pos
 *         feature[0] = best_feature             # <<<<<<<<<<<<<<
 *         threshold[0] = best_threshold
 * 
 */
  (__pyx_v_feature[0]) = __pyx_v_best_feature;

  /* "sklearn/tree/_tree.pyx":1010
 *         pos[0] = best_pos
 *         feature[0] = best_feature
 *         threshold[0] = best_threshold             # <<<<<<<<<<<<<<
 * 
 * cdef void sort(np.ndarray[DTYPE_t, ndim=2, mode="c"] X, SIZE_t current_feature,
 */
  (__pyx_v_threshold[0]) = __pyx_v_best_threshold;

  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  { PyObject *__pyx_type, *__pyx_value, *__pyx_tb;
    __Pyx_ErrFetch(&__pyx_type, &__pyx_value, &__pyx_tb);
    __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_X.rcbuffer->pybuffer);
  __Pyx_ErrRestore(__pyx_type, __pyx_value, __pyx_tb);}
  __Pyx_WriteUnraisable("sklearn.tree._tree.BestSplitter.node_split", __pyx_clineno, __pyx_lineno, __pyx_filename);
  goto __pyx_L2;
  __pyx_L0:;
  __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_X.rcbuffer->pybuffer);
  __pyx_L2:;
  __Pyx_XDECREF((PyObject *)__pyx_v_criterion);
  __Pyx_XDECREF((PyObject *)__pyx_v_X);
  __Pyx_RefNannyFinishContext();
}

/* "sklearn/tree/_tree.pyx":1012
 *         threshold[0] = best_threshold
 * 
 * cdef void sort(np.ndarray[DTYPE_t, ndim=2, mode="c"] X, SIZE_t current_feature,             # <<<<<<<<<<<<<<
 *                SIZE_t* samples, SIZE_t length):
 *     """In-place sorting of samples[start:end] using
 */

static void __pyx_f_7sklearn_4tree_5_tree_sort(PyArrayObject *__pyx_v_X, __pyx_t_7sklearn_4tree_5_tree_SIZE_t __pyx_v_current_feature, __pyx_t_7sklearn_4tree_5_tree_SIZE_t *__pyx_v_samples, __pyx_t_7sklearn_4tree_5_tree_SIZE_t __pyx_v_length) {
  __pyx_t_7sklearn_4tree_5_tree_SIZE_t __pyx_v_tmp;
  __pyx_t_7sklearn_4tree_5_tree_DOUBLE_t __pyx_v_tmp_value;
  __pyx_t_7sklearn_4tree_5_tree_SIZE_t __pyx_v_n;
  __pyx_t_7sklearn_4tree_5_tree_SIZE_t __pyx_v_parent;
  __pyx_t_7sklearn_4tree_5_tree_SIZE_t __pyx_v_index;
  __pyx_t_7sklearn_4tree_5_tree_SIZE_t __pyx_v_child;
  __Pyx_LocalBuf_ND __pyx_pybuffernd_X;
  __Pyx_Buffer __pyx_pybuffer_X;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  __pyx_t_7sklearn_4tree_5_tree_SIZE_t __pyx_t_2;
  __pyx_t_7sklearn_4tree_5_tree_SIZE_t __pyx_t_3;
  __pyx_t_7sklearn_4tree_5_tree_SIZE_t __pyx_t_4;
  __pyx_t_7sklearn_4tree_5_tree_SIZE_t __pyx_t_5;
  __pyx_t_7sklearn_4tree_5_tree_SIZE_t __pyx_t_6;
  __pyx_t_7sklearn_4tree_5_tree_SIZE_t __pyx_t_7;
  int __pyx_t_8;
  int __pyx_t_9;
  __pyx_t_7sklearn_4tree_5_tree_SIZE_t __pyx_t_10;
  __pyx_t_7sklearn_4tree_5_tree_SIZE_t __pyx_t_11;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("sort", 0);
  __pyx_pybuffer_X.pybuffer.buf = NULL;
  __pyx_pybuffer_X.refcount = 0;
  __pyx_pybuffernd_X.data = NULL;
  __pyx_pybuffernd_X.rcbuffer = &__pyx_pybuffer_X;
  {
    __Pyx_BufFmt_StackElem __pyx_stack[1];
    if (unlikely(__Pyx_GetBufferAndValidate(&__pyx_pybuffernd_X.rcbuffer->pybuffer, (PyObject*)__pyx_v_X, &__Pyx_TypeInfo_nn___pyx_t_7sklearn_4tree_5_tree_DTYPE_t, PyBUF_FORMAT| PyBUF_C_CONTIGUOUS, 2, 0, __pyx_stack) == -1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1012; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  }
  __pyx_pybuffernd_X.diminfo[0].strides = __pyx_pybuffernd_X.rcbuffer->pybuffer.strides[0]; __pyx_pybuffernd_X.diminfo[0].shape = __pyx_pybuffernd_X.rcbuffer->pybuffer.shape[0]; __pyx_pybuffernd_X.diminfo[1].strides = __pyx_pybuffernd_X.rcbuffer->pybuffer.strides[1]; __pyx_pybuffernd_X.diminfo[1].shape = __pyx_pybuffernd_X.rcbuffer->pybuffer.shape[1];

  /* "sklearn/tree/_tree.pyx":1019
 *     cdef SIZE_t tmp
 *     cdef DOUBLE_t tmp_value
 *     cdef SIZE_t n = length             # <<<<<<<<<<<<<<
 *     cdef SIZE_t parent = length / 2
 *     cdef SIZE_t index, child
 */
  __pyx_v_n = __pyx_v_length;

  /* "sklearn/tree/_tree.pyx":1020
 *     cdef DOUBLE_t tmp_value
 *     cdef SIZE_t n = length
 *     cdef SIZE_t parent = length / 2             # <<<<<<<<<<<<<<
 *     cdef SIZE_t index, child
 * 
 */
  __pyx_v_parent = (__pyx_v_length / 2);

  /* "sklearn/tree/_tree.pyx":1023
 *     cdef SIZE_t index, child
 * 
 *     while True:             # <<<<<<<<<<<<<<
 *         if parent > 0:
 *             parent -= 1
 */
  while (1) {
    if (!1) break;

    /* "sklearn/tree/_tree.pyx":1024
 * 
 *     while True:
 *         if parent > 0:             # <<<<<<<<<<<<<<
 *             parent -= 1
 *             tmp = samples[parent]
 */
    __pyx_t_1 = ((__pyx_v_parent > 0) != 0);
    if (__pyx_t_1) {

      /* "sklearn/tree/_tree.pyx":1025
 *     while True:
 *         if parent > 0:
 *             parent -= 1             # <<<<<<<<<<<<<<
 *             tmp = samples[parent]
 *         else:
 */
      __pyx_v_parent = (__pyx_v_parent - 1);

      /* "sklearn/tree/_tree.pyx":1026
 *         if parent > 0:
 *             parent -= 1
 *             tmp = samples[parent]             # <<<<<<<<<<<<<<
 *         else:
 *             n -= 1
 */
      __pyx_v_tmp = (__pyx_v_samples[__pyx_v_parent]);
      goto __pyx_L5;
    }
    /*else*/ {

      /* "sklearn/tree/_tree.pyx":1028
 *             tmp = samples[parent]
 *         else:
 *             n -= 1             # <<<<<<<<<<<<<<
 *             if n == 0:
 *                 return
 */
      __pyx_v_n = (__pyx_v_n - 1);

      /* "sklearn/tree/_tree.pyx":1029
 *         else:
 *             n -= 1
 *             if n == 0:             # <<<<<<<<<<<<<<
 *                 return
 *             tmp = samples[n]
 */
      __pyx_t_1 = ((__pyx_v_n == 0) != 0);
      if (__pyx_t_1) {

        /* "sklearn/tree/_tree.pyx":1030
 *             n -= 1
 *             if n == 0:
 *                 return             # <<<<<<<<<<<<<<
 *             tmp = samples[n]
 *             samples[n] = samples[0]
 */
        goto __pyx_L0;
        goto __pyx_L6;
      }
      __pyx_L6:;

      /* "sklearn/tree/_tree.pyx":1031
 *             if n == 0:
 *                 return
 *             tmp = samples[n]             # <<<<<<<<<<<<<<
 *             samples[n] = samples[0]
 * 
 */
      __pyx_v_tmp = (__pyx_v_samples[__pyx_v_n]);

      /* "sklearn/tree/_tree.pyx":1032
 *                 return
 *             tmp = samples[n]
 *             samples[n] = samples[0]             # <<<<<<<<<<<<<<
 * 
 *         tmp_value = X[tmp, current_feature]
 */
      (__pyx_v_samples[__pyx_v_n]) = (__pyx_v_samples[0]);
    }
    __pyx_L5:;

    /* "sklearn/tree/_tree.pyx":1034
 *             samples[n] = samples[0]
 * 
 *         tmp_value = X[tmp, current_feature]             # <<<<<<<<<<<<<<
 *         index = parent
 *         child = index * 2 + 1
 */
    __pyx_t_2 = __pyx_v_tmp;
    __pyx_t_3 = __pyx_v_current_feature;
    __pyx_v_tmp_value = (*__Pyx_BufPtrCContig2d(__pyx_t_7sklearn_4tree_5_tree_DTYPE_t *, __pyx_pybuffernd_X.rcbuffer->pybuffer.buf, __pyx_t_2, __pyx_pybuffernd_X.diminfo[0].strides, __pyx_t_3, __pyx_pybuffernd_X.diminfo[1].strides));

    /* "sklearn/tree/_tree.pyx":1035
 * 
 *         tmp_value = X[tmp, current_feature]
 *         index = parent             # <<<<<<<<<<<<<<
 *         child = index * 2 + 1
 * 
 */
    __pyx_v_index = __pyx_v_parent;

    /* "sklearn/tree/_tree.pyx":1036
 *         tmp_value = X[tmp, current_feature]
 *         index = parent
 *         child = index * 2 + 1             # <<<<<<<<<<<<<<
 * 
 *         while child < n:
 */
    __pyx_v_child = ((__pyx_v_index * 2) + 1);

    /* "sklearn/tree/_tree.pyx":1038
 *         child = index * 2 + 1
 * 
 *         while child < n:             # <<<<<<<<<<<<<<
 *             if ((child + 1 < n) and
 *                 (X[samples[child + 1], current_feature] > X[samples[child], current_feature])):
 */
    while (1) {
      __pyx_t_1 = ((__pyx_v_child < __pyx_v_n) != 0);
      if (!__pyx_t_1) break;

      /* "sklearn/tree/_tree.pyx":1039
 * 
 *         while child < n:
 *             if ((child + 1 < n) and             # <<<<<<<<<<<<<<
 *                 (X[samples[child + 1], current_feature] > X[samples[child], current_feature])):
 *                 child += 1
 */
      __pyx_t_1 = (((__pyx_v_child + 1) < __pyx_v_n) != 0);
      if (__pyx_t_1) {

        /* "sklearn/tree/_tree.pyx":1040
 *         while child < n:
 *             if ((child + 1 < n) and
 *                 (X[samples[child + 1], current_feature] > X[samples[child], current_feature])):             # <<<<<<<<<<<<<<
 *                 child += 1
 * 
 */
        __pyx_t_4 = (__pyx_v_samples[(__pyx_v_child + 1)]);
        __pyx_t_5 = __pyx_v_current_feature;
        __pyx_t_6 = (__pyx_v_samples[__pyx_v_child]);
        __pyx_t_7 = __pyx_v_current_feature;
        __pyx_t_8 = (((*__Pyx_BufPtrCContig2d(__pyx_t_7sklearn_4tree_5_tree_DTYPE_t *, __pyx_pybuffernd_X.rcbuffer->pybuffer.buf, __pyx_t_4, __pyx_pybuffernd_X.diminfo[0].strides, __pyx_t_5, __pyx_pybuffernd_X.diminfo[1].strides)) > (*__Pyx_BufPtrCContig2d(__pyx_t_7sklearn_4tree_5_tree_DTYPE_t *, __pyx_pybuffernd_X.rcbuffer->pybuffer.buf, __pyx_t_6, __pyx_pybuffernd_X.diminfo[0].strides, __pyx_t_7, __pyx_pybuffernd_X.diminfo[1].strides))) != 0);
        __pyx_t_9 = __pyx_t_8;
      } else {
        __pyx_t_9 = __pyx_t_1;
      }
      if (__pyx_t_9) {

        /* "sklearn/tree/_tree.pyx":1041
 *             if ((child + 1 < n) and
 *                 (X[samples[child + 1], current_feature] > X[samples[child], current_feature])):
 *                 child += 1             # <<<<<<<<<<<<<<
 * 
 *             if X[samples[child], current_feature] > tmp_value:
 */
        __pyx_v_child = (__pyx_v_child + 1);
        goto __pyx_L9;
      }
      __pyx_L9:;

      /* "sklearn/tree/_tree.pyx":1043
 *                 child += 1
 * 
 *             if X[samples[child], current_feature] > tmp_value:             # <<<<<<<<<<<<<<
 *                 samples[index] = samples[child]
 *                 index = child
 */
      __pyx_t_10 = (__pyx_v_samples[__pyx_v_child]);
      __pyx_t_11 = __pyx_v_current_feature;
      __pyx_t_9 = (((*__Pyx_BufPtrCContig2d(__pyx_t_7sklearn_4tree_5_tree_DTYPE_t *, __pyx_pybuffernd_X.rcbuffer->pybuffer.buf, __pyx_t_10, __pyx_pybuffernd_X.diminfo[0].strides, __pyx_t_11, __pyx_pybuffernd_X.diminfo[1].strides)) > __pyx_v_tmp_value) != 0);
      if (__pyx_t_9) {

        /* "sklearn/tree/_tree.pyx":1044
 * 
 *             if X[samples[child], current_feature] > tmp_value:
 *                 samples[index] = samples[child]             # <<<<<<<<<<<<<<
 *                 index = child
 *                 child = index * 2 + 1
 */
        (__pyx_v_samples[__pyx_v_index]) = (__pyx_v_samples[__pyx_v_child]);

        /* "sklearn/tree/_tree.pyx":1045
 *             if X[samples[child], current_feature] > tmp_value:
 *                 samples[index] = samples[child]
 *                 index = child             # <<<<<<<<<<<<<<
 *                 child = index * 2 + 1
 * 
 */
        __pyx_v_index = __pyx_v_child;

        /* "sklearn/tree/_tree.pyx":1046
 *                 samples[index] = samples[child]
 *                 index = child
 *                 child = index * 2 + 1             # <<<<<<<<<<<<<<
 * 
 *             else:
 */
        __pyx_v_child = ((__pyx_v_index * 2) + 1);
        goto __pyx_L10;
      }
      /*else*/ {

        /* "sklearn/tree/_tree.pyx":1049
 * 
 *             else:
 *                 break             # <<<<<<<<<<<<<<
 * 
 *         samples[index] = tmp
 */
        goto __pyx_L8_break;
      }
      __pyx_L10:;
    }
    __pyx_L8_break:;

    /* "sklearn/tree/_tree.pyx":1051
 *                 break
 * 
 *         samples[index] = tmp             # <<<<<<<<<<<<<<
 * 
 * 
 */
    (__pyx_v_samples[__pyx_v_index]) = __pyx_v_tmp;
  }

  goto __pyx_L0;
  __pyx_L1_error:;
  { PyObject *__pyx_type, *__pyx_value, *__pyx_tb;
    __Pyx_ErrFetch(&__pyx_type, &__pyx_value, &__pyx_tb);
    __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_X.rcbuffer->pybuffer);
  __Pyx_ErrRestore(__pyx_type, __pyx_value, __pyx_tb);}
  __Pyx_WriteUnraisable("sklearn.tree._tree.sort", __pyx_clineno, __pyx_lineno, __pyx_filename);
  goto __pyx_L2;
  __pyx_L0:;
  __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_X.rcbuffer->pybuffer);
  __pyx_L2:;
  __Pyx_RefNannyFinishContext();
}

/* Python wrapper */
static PyObject *__pyx_pw_7sklearn_4tree_5_tree_14RandomSplitter_1__reduce__(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static PyObject *__pyx_pw_7sklearn_4tree_5_tree_14RandomSplitter_1__reduce__(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__reduce__ (wrapper)", 0);
  __pyx_r = __pyx_pf_7sklearn_4tree_5_tree_14RandomSplitter___reduce__(((struct __pyx_obj_7sklearn_4tree_5_tree_RandomSplitter *)__pyx_v_self));
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "sklearn/tree/_tree.pyx":1056
 * cdef class RandomSplitter(Splitter):
 *     """Splitter for finding the best random split."""
 *     def __reduce__(self):             # <<<<<<<<<<<<<<
 *         return (RandomSplitter, (self.criterion,
 *                                  self.max_features,
 */

static PyObject *__pyx_pf_7sklearn_4tree_5_tree_14RandomSplitter___reduce__(struct __pyx_obj_7sklearn_4tree_5_tree_RandomSplitter *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__reduce__", 0);

  /* "sklearn/tree/_tree.pyx":1057
 *     """Splitter for finding the best random split."""
 *     def __reduce__(self):
 *         return (RandomSplitter, (self.criterion,             # <<<<<<<<<<<<<<
 *                                  self.max_features,
 *                                  self.min_samples_leaf,
 */
  __Pyx_XDECREF(__pyx_r);

  /* "sklearn/tree/_tree.pyx":1058
 *     def __reduce__(self):
 *         return (RandomSplitter, (self.criterion,
 *                                  self.max_features,             # <<<<<<<<<<<<<<
 *                                  self.min_samples_leaf,
 *                                  self.random_state), self.__getstate__())
 */
  __pyx_t_1 = __Pyx_PyInt_to_py_Py_intptr_t(__pyx_v_self->__pyx_base.max_features); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1058; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_1);

  /* "sklearn/tree/_tree.pyx":1059
 *         return (RandomSplitter, (self.criterion,
 *                                  self.max_features,
 *                                  self.min_samples_leaf,             # <<<<<<<<<<<<<<
 *                                  self.random_state), self.__getstate__())
 * 
 */
  __pyx_t_2 = __Pyx_PyInt_to_py_Py_intptr_t(__pyx_v_self->__pyx_base.min_samples_leaf); if (unlikely(!__pyx_t_2)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1059; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_2);

  /* "sklearn/tree/_tree.pyx":1060
 *                                  self.max_features,
 *                                  self.min_samples_leaf,
 *                                  self.random_state), self.__getstate__())             # <<<<<<<<<<<<<<
 * 
 *     cdef void node_split(self, SIZE_t* pos, SIZE_t* feature, double* threshold):
 */
  __pyx_t_3 = PyTuple_New(4); if (unlikely(!__pyx_t_3)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1057; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_INCREF(((PyObject *)__pyx_v_self->__pyx_base.criterion));
  PyTuple_SET_ITEM(__pyx_t_3, 0, ((PyObject *)__pyx_v_self->__pyx_base.criterion));
  __Pyx_GIVEREF(((PyObject *)__pyx_v_self->__pyx_base.criterion));
  PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  PyTuple_SET_ITEM(__pyx_t_3, 2, __pyx_t_2);
  __Pyx_GIVEREF(__pyx_t_2);
  __Pyx_INCREF(__pyx_v_self->__pyx_base.random_state);
  PyTuple_SET_ITEM(__pyx_t_3, 3, __pyx_v_self->__pyx_base.random_state);
  __Pyx_GIVEREF(__pyx_v_self->__pyx_base.random_state);
  __pyx_t_1 = 0;
  __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s____getstate__); if (unlikely(!__pyx_t_2)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1060; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_1 = PyObject_Call(__pyx_t_2, ((PyObject *)__pyx_empty_tuple), NULL); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1060; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = PyTuple_New(3); if (unlikely(!__pyx_t_2)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1057; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_INCREF(((PyObject *)((PyObject*)__pyx_ptype_7sklearn_4tree_5_tree_RandomSplitter)));
  PyTuple_SET_ITEM(__pyx_t_2, 0, ((PyObject *)((PyObject*)__pyx_ptype_7sklearn_4tree_5_tree_RandomSplitter)));
  __Pyx_GIVEREF(((PyObject *)((PyObject*)__pyx_ptype_7sklearn_4tree_5_tree_RandomSplitter)));
  PyTuple_SET_ITEM(__pyx_t_2, 1, ((PyObject *)__pyx_t_3));
  __Pyx_GIVEREF(((PyObject *)__pyx_t_3));
  PyTuple_SET_ITEM(__pyx_t_2, 2, __pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __pyx_t_3 = 0;
  __pyx_t_1 = 0;
  __pyx_r = ((PyObject *)__pyx_t_2);
  __pyx_t_2 = 0;
  goto __pyx_L0;

  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("sklearn.tree._tree.RandomSplitter.__reduce__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "sklearn/tree/_tree.pyx":1062
 *                                  self.random_state), self.__getstate__())
 * 
 *     cdef void node_split(self, SIZE_t* pos, SIZE_t* feature, double* threshold):             # <<<<<<<<<<<<<<
 *         """Find the best random split on node samples[start:end]."""
 *         # Draw random splits and pick the best
 */

static void __pyx_f_7sklearn_4tree_5_tree_14RandomSplitter_node_split(struct __pyx_obj_7sklearn_4tree_5_tree_RandomSplitter *__pyx_v_self, __pyx_t_7sklearn_4tree_5_tree_SIZE_t *__pyx_v_pos, __pyx_t_7sklearn_4tree_5_tree_SIZE_t *__pyx_v_feature, double *__pyx_v_threshold) {
  struct __pyx_obj_7sklearn_4tree_5_tree_Criterion *__pyx_v_criterion = 0;
  __pyx_t_7sklearn_4tree_5_tree_SIZE_t *__pyx_v_samples;
  __pyx_t_7sklearn_4tree_5_tree_SIZE_t __pyx_v_start;
  __pyx_t_7sklearn_4tree_5_tree_SIZE_t __pyx_v_end;
  __pyx_t_7sklearn_4tree_5_tree_SIZE_t *__pyx_v_features;
  __pyx_t_7sklearn_4tree_5_tree_SIZE_t __pyx_v_n_features;
  PyArrayObject *__pyx_v_X = 0;
  __pyx_t_7sklearn_4tree_5_tree_SIZE_t __pyx_v_max_features;
  __pyx_t_7sklearn_4tree_5_tree_SIZE_t __pyx_v_min_samples_leaf;
  unsigned int *__pyx_v_random_state;
  double __pyx_v_best_impurity;
  __pyx_t_7sklearn_4tree_5_tree_SIZE_t __pyx_v_best_pos;
  __pyx_t_7sklearn_4tree_5_tree_SIZE_t __pyx_v_best_feature;
  double __pyx_v_best_threshold;
  double __pyx_v_current_impurity;
  __pyx_t_7sklearn_4tree_5_tree_SIZE_t __pyx_v_current_pos;
  __pyx_t_7sklearn_4tree_5_tree_SIZE_t __pyx_v_current_feature;
  double __pyx_v_current_threshold;
  __pyx_t_7sklearn_4tree_5_tree_SIZE_t __pyx_v_f_idx;
  __pyx_t_7sklearn_4tree_5_tree_SIZE_t __pyx_v_f_i;
  __pyx_t_7sklearn_4tree_5_tree_SIZE_t __pyx_v_f_j;
  __pyx_t_7sklearn_4tree_5_tree_SIZE_t __pyx_v_p;
  __pyx_t_7sklearn_4tree_5_tree_SIZE_t __pyx_v_tmp;
  __pyx_t_7sklearn_4tree_5_tree_SIZE_t __pyx_v_visited_features;
  __pyx_t_7sklearn_4tree_5_tree_DTYPE_t __pyx_v_min_feature_value;
  __pyx_t_7sklearn_4tree_5_tree_DTYPE_t __pyx_v_max_feature_value;
  __pyx_t_7sklearn_4tree_5_tree_DTYPE_t __pyx_v_current_feature_value;
  CYTHON_UNUSED __pyx_t_7sklearn_4tree_5_tree_SIZE_t __pyx_v_partition_start;
  __pyx_t_7sklearn_4tree_5_tree_SIZE_t __pyx_v_partition_end;
  __Pyx_LocalBuf_ND __pyx_pybuffernd_X;
  __Pyx_Buffer __pyx_pybuffer_X;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __pyx_t_7sklearn_4tree_5_tree_SIZE_t *__pyx_t_2;
  __pyx_t_7sklearn_4tree_5_tree_SIZE_t __pyx_t_3;
  __pyx_t_7sklearn_4tree_5_tree_SIZE_t __pyx_t_4;
  __pyx_t_7sklearn_4tree_5_tree_SIZE_t __pyx_t_5;
  __pyx_t_7sklearn_4tree_5_tree_DTYPE_t __pyx_t_6;
  __pyx_t_7sklearn_4tree_5_tree_SIZE_t __pyx_t_7;
  __pyx_t_7sklearn_4tree_5_tree_SIZE_t __pyx_t_8;
  __pyx_t_7sklearn_4tree_5_tree_SIZE_t __pyx_t_9;
  int __pyx_t_10;
  __pyx_t_7sklearn_4tree_5_tree_SIZE_t __pyx_t_11;
  int __pyx_t_12;
  int __pyx_t_13;
  __pyx_t_7sklearn_4tree_5_tree_SIZE_t __pyx_t_14;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("node_split", 0);
  __pyx_pybuffer_X.pybuffer.buf = NULL;
  __pyx_pybuffer_X.refcount = 0;
  __pyx_pybuffernd_X.data = NULL;
  __pyx_pybuffernd_X.rcbuffer = &__pyx_pybuffer_X;

  /* "sklearn/tree/_tree.pyx":1065
 *         """Find the best random split on node samples[start:end]."""
 *         # Draw random splits and pick the best
 *         cdef Criterion criterion = self.criterion             # <<<<<<<<<<<<<<
 *         cdef SIZE_t* samples = self.samples
 *         cdef SIZE_t start = self.start
 */
  __pyx_t_1 = ((PyObject *)__pyx_v_self->__pyx_base.criterion);
  __Pyx_INCREF(__pyx_t_1);
  __pyx_v_criterion = ((struct __pyx_obj_7sklearn_4tree_5_tree_Criterion *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "sklearn/tree/_tree.pyx":1066
 *         # Draw random splits and pick the best
 *         cdef Criterion criterion = self.criterion
 *         cdef SIZE_t* samples = self.samples             # <<<<<<<<<<<<<<
 *         cdef SIZE_t start = self.start
 *         cdef SIZE_t end = self.end
 */
  __pyx_t_2 = __pyx_v_self->__pyx_base.samples;
  __pyx_v_samples = __pyx_t_2;

  /* "sklearn/tree/_tree.pyx":1067
 *         cdef Criterion criterion = self.criterion
 *         cdef SIZE_t* samples = self.samples
 *         cdef SIZE_t start = self.start             # <<<<<<<<<<<<<<
 *         cdef SIZE_t end = self.end
 * 
 */
  __pyx_t_3 = __pyx_v_self->__pyx_base.start;
  __pyx_v_start = __pyx_t_3;

  /* "sklearn/tree/_tree.pyx":1068
 *         cdef SIZE_t* samples = self.samples
 *         cdef SIZE_t start = self.start
 *         cdef SIZE_t end = self.end             # <<<<<<<<<<<<<<
 * 
 *         cdef SIZE_t* features = self.features
 */
  __pyx_t_3 = __pyx_v_self->__pyx_base.end;
  __pyx_v_end = __pyx_t_3;

  /* "sklearn/tree/_tree.pyx":1070
 *         cdef SIZE_t end = self.end
 * 
 *         cdef SIZE_t* features = self.features             # <<<<<<<<<<<<<<
 *         cdef SIZE_t n_features = self.n_features
 * 
 */
  __pyx_t_2 = __pyx_v_self->__pyx_base.features;
  __pyx_v_features = __pyx_t_2;

  /* "sklearn/tree/_tree.pyx":1071
 * 
 *         cdef SIZE_t* features = self.features
 *         cdef SIZE_t n_features = self.n_features             # <<<<<<<<<<<<<<
 * 
 *         cdef np.ndarray[DTYPE_t, ndim=2, mode="c"] X = self.X
 */
  __pyx_t_3 = __pyx_v_self->__pyx_base.n_features;
  __pyx_v_n_features = __pyx_t_3;

  /* "sklearn/tree/_tree.pyx":1073
 *         cdef SIZE_t n_features = self.n_features
 * 
 *         cdef np.ndarray[DTYPE_t, ndim=2, mode="c"] X = self.X             # <<<<<<<<<<<<<<
 *         cdef SIZE_t max_features = self.max_features
 *         cdef SIZE_t min_samples_leaf = self.min_samples_leaf
 */
  __pyx_t_1 = ((PyObject *)__pyx_v_self->__pyx_base.X);
  __Pyx_INCREF(__pyx_t_1);
  {
    __Pyx_BufFmt_StackElem __pyx_stack[1];
    if (unlikely(__Pyx_GetBufferAndValidate(&__pyx_pybuffernd_X.rcbuffer->pybuffer, (PyObject*)((PyArrayObject *)__pyx_t_1), &__Pyx_TypeInfo_nn___pyx_t_7sklearn_4tree_5_tree_DTYPE_t, PyBUF_FORMAT| PyBUF_C_CONTIGUOUS, 2, 0, __pyx_stack) == -1)) {
      __pyx_v_X = ((PyArrayObject *)Py_None); __Pyx_INCREF(Py_None); __pyx_pybuffernd_X.rcbuffer->pybuffer.buf = NULL;
      {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1073; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    } else {__pyx_pybuffernd_X.diminfo[0].strides = __pyx_pybuffernd_X.rcbuffer->pybuffer.strides[0]; __pyx_pybuffernd_X.diminfo[0].shape = __pyx_pybuffernd_X.rcbuffer->pybuffer.shape[0]; __pyx_pybuffernd_X.diminfo[1].strides = __pyx_pybuffernd_X.rcbuffer->pybuffer.strides[1]; __pyx_pybuffernd_X.diminfo[1].shape = __pyx_pybuffernd_X.rcbuffer->pybuffer.shape[1];
    }
  }
  __pyx_v_X = ((PyArrayObject *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "sklearn/tree/_tree.pyx":1074
 * 
 *         cdef np.ndarray[DTYPE_t, ndim=2, mode="c"] X = self.X
 *         cdef SIZE_t max_features = self.max_features             # <<<<<<<<<<<<<<
 *         cdef SIZE_t min_samples_leaf = self.min_samples_leaf
 *         cdef unsigned int* random_state = &self.rand_r_state
 */
  __pyx_t_3 = __pyx_v_self->__pyx_base.max_features;
  __pyx_v_max_features = __pyx_t_3;

  /* "sklearn/tree/_tree.pyx":1075
 *         cdef np.ndarray[DTYPE_t, ndim=2, mode="c"] X = self.X
 *         cdef SIZE_t max_features = self.max_features
 *         cdef SIZE_t min_samples_leaf = self.min_samples_leaf             # <<<<<<<<<<<<<<
 *         cdef unsigned int* random_state = &self.rand_r_state
 * 
 */
  __pyx_t_3 = __pyx_v_self->__pyx_base.min_samples_leaf;
  __pyx_v_min_samples_leaf = __pyx_t_3;

  /* "sklearn/tree/_tree.pyx":1076
 *         cdef SIZE_t max_features = self.max_features
 *         cdef SIZE_t min_samples_leaf = self.min_samples_leaf
 *         cdef unsigned int* random_state = &self.rand_r_state             # <<<<<<<<<<<<<<
 * 
 *         cdef double best_impurity = INFINITY
 */
  __pyx_v_random_state = (&__pyx_v_self->__pyx_base.rand_r_state);

  /* "sklearn/tree/_tree.pyx":1078
 *         cdef unsigned int* random_state = &self.rand_r_state
 * 
 *         cdef double best_impurity = INFINITY             # <<<<<<<<<<<<<<
 *         cdef SIZE_t best_pos = end
 *         cdef SIZE_t best_feature
 */
  __pyx_v_best_impurity = __pyx_v_7sklearn_4tree_5_tree_INFINITY;

  /* "sklearn/tree/_tree.pyx":1079
 * 
 *         cdef double best_impurity = INFINITY
 *         cdef SIZE_t best_pos = end             # <<<<<<<<<<<<<<
 *         cdef SIZE_t best_feature
 *         cdef double best_threshold
 */
  __pyx_v_best_pos = __pyx_v_end;

  /* "sklearn/tree/_tree.pyx":1089
 * 
 *         cdef SIZE_t f_idx, f_i, f_j, p, tmp
 *         cdef SIZE_t visited_features = 0             # <<<<<<<<<<<<<<
 *         cdef DTYPE_t min_feature_value
 *         cdef DTYPE_t max_feature_value
 */
  __pyx_v_visited_features = 0;

  /* "sklearn/tree/_tree.pyx":1097
 *         cdef SIZE_t partition_end
 * 
 *         for f_idx from 0 <= f_idx < n_features:             # <<<<<<<<<<<<<<
 *             # Draw a feature at random
 *             f_i = n_features - f_idx - 1
 */
  __pyx_t_3 = __pyx_v_n_features;
  for (__pyx_v_f_idx = 0; __pyx_v_f_idx < __pyx_t_3; __pyx_v_f_idx++) {

    /* "sklearn/tree/_tree.pyx":1099
 *         for f_idx from 0 <= f_idx < n_features:
 *             # Draw a feature at random
 *             f_i = n_features - f_idx - 1             # <<<<<<<<<<<<<<
 *             f_j = rand_int(n_features - f_idx, random_state)
 * 
 */
    __pyx_v_f_i = ((__pyx_v_n_features - __pyx_v_f_idx) - 1);

    /* "sklearn/tree/_tree.pyx":1100
 *             # Draw a feature at random
 *             f_i = n_features - f_idx - 1
 *             f_j = rand_int(n_features - f_idx, random_state)             # <<<<<<<<<<<<<<
 * 
 *             tmp = features[f_i]
 */
    __pyx_v_f_j = __pyx_f_7sklearn_4tree_5_tree_rand_int((__pyx_v_n_features - __pyx_v_f_idx), __pyx_v_random_state);

    /* "sklearn/tree/_tree.pyx":1102
 *             f_j = rand_int(n_features - f_idx, random_state)
 * 
 *             tmp = features[f_i]             # <<<<<<<<<<<<<<
 *             features[f_i] = features[f_j]
 *             features[f_j] = tmp
 */
    __pyx_v_tmp = (__pyx_v_features[__pyx_v_f_i]);

    /* "sklearn/tree/_tree.pyx":1103
 * 
 *             tmp = features[f_i]
 *             features[f_i] = features[f_j]             # <<<<<<<<<<<<<<
 *             features[f_j] = tmp
 * 
 */
    (__pyx_v_features[__pyx_v_f_i]) = (__pyx_v_features[__pyx_v_f_j]);

    /* "sklearn/tree/_tree.pyx":1104
 *             tmp = features[f_i]
 *             features[f_i] = features[f_j]
 *             features[f_j] = tmp             # <<<<<<<<<<<<<<
 * 
 *             current_feature = features[f_i]
 */
    (__pyx_v_features[__pyx_v_f_j]) = __pyx_v_tmp;

    /* "sklearn/tree/_tree.pyx":1106
 *             features[f_j] = tmp
 * 
 *             current_feature = features[f_i]             # <<<<<<<<<<<<<<
 * 
 *             # Find min, max
 */
    __pyx_v_current_feature = (__pyx_v_features[__pyx_v_f_i]);

    /* "sklearn/tree/_tree.pyx":1109
 * 
 *             # Find min, max
 *             min_feature_value = max_feature_value = X[samples[start], current_feature]             # <<<<<<<<<<<<<<
 * 
 *             for p from start < p < end:
 */
    __pyx_t_4 = (__pyx_v_samples[__pyx_v_start]);
    __pyx_t_5 = __pyx_v_current_feature;
    __pyx_t_6 = (*__Pyx_BufPtrCContig2d(__pyx_t_7sklearn_4tree_5_tree_DTYPE_t *, __pyx_pybuffernd_X.rcbuffer->pybuffer.buf, __pyx_t_4, __pyx_pybuffernd_X.diminfo[0].strides, __pyx_t_5, __pyx_pybuffernd_X.diminfo[1].strides));
    __pyx_v_min_feature_value = __pyx_t_6;
    __pyx_v_max_feature_value = __pyx_t_6;

    /* "sklearn/tree/_tree.pyx":1111
 *             min_feature_value = max_feature_value = X[samples[start], current_feature]
 * 
 *             for p from start < p < end:             # <<<<<<<<<<<<<<
 *                 current_feature_value = X[samples[p], current_feature]
 * 
 */
    __pyx_t_7 = __pyx_v_end;
    for (__pyx_v_p = __pyx_v_start+1; __pyx_v_p < __pyx_t_7; __pyx_v_p++) {

      /* "sklearn/tree/_tree.pyx":1112
 * 
 *             for p from start < p < end:
 *                 current_feature_value = X[samples[p], current_feature]             # <<<<<<<<<<<<<<
 * 
 *                 if current_feature_value < min_feature_value:
 */
      __pyx_t_8 = (__pyx_v_samples[__pyx_v_p]);
      __pyx_t_9 = __pyx_v_current_feature;
      __pyx_v_current_feature_value = (*__Pyx_BufPtrCContig2d(__pyx_t_7sklearn_4tree_5_tree_DTYPE_t *, __pyx_pybuffernd_X.rcbuffer->pybuffer.buf, __pyx_t_8, __pyx_pybuffernd_X.diminfo[0].strides, __pyx_t_9, __pyx_pybuffernd_X.diminfo[1].strides));

      /* "sklearn/tree/_tree.pyx":1114
 *                 current_feature_value = X[samples[p], current_feature]
 * 
 *                 if current_feature_value < min_feature_value:             # <<<<<<<<<<<<<<
 *                     min_feature_value = current_feature_value
 *                 elif current_feature_value > max_feature_value:
 */
      __pyx_t_10 = ((__pyx_v_current_feature_value < __pyx_v_min_feature_value) != 0);
      if (__pyx_t_10) {

        /* "sklearn/tree/_tree.pyx":1115
 * 
 *                 if current_feature_value < min_feature_value:
 *                     min_feature_value = current_feature_value             # <<<<<<<<<<<<<<
 *                 elif current_feature_value > max_feature_value:
 *                     max_feature_value = current_feature_value
 */
        __pyx_v_min_feature_value = __pyx_v_current_feature_value;
        goto __pyx_L7;
      }

      /* "sklearn/tree/_tree.pyx":1116
 *                 if current_feature_value < min_feature_value:
 *                     min_feature_value = current_feature_value
 *                 elif current_feature_value > max_feature_value:             # <<<<<<<<<<<<<<
 *                     max_feature_value = current_feature_value
 * 
 */
      __pyx_t_10 = ((__pyx_v_current_feature_value > __pyx_v_max_feature_value) != 0);
      if (__pyx_t_10) {

        /* "sklearn/tree/_tree.pyx":1117
 *                     min_feature_value = current_feature_value
 *                 elif current_feature_value > max_feature_value:
 *                     max_feature_value = current_feature_value             # <<<<<<<<<<<<<<
 * 
 *             if min_feature_value == max_feature_value:
 */
        __pyx_v_max_feature_value = __pyx_v_current_feature_value;
        goto __pyx_L7;
      }
      __pyx_L7:;
    }

    /* "sklearn/tree/_tree.pyx":1119
 *                     max_feature_value = current_feature_value
 * 
 *             if min_feature_value == max_feature_value:             # <<<<<<<<<<<<<<
 *                 continue
 * 
 */
    __pyx_t_10 = ((__pyx_v_min_feature_value == __pyx_v_max_feature_value) != 0);
    if (__pyx_t_10) {

      /* "sklearn/tree/_tree.pyx":1120
 * 
 *             if min_feature_value == max_feature_value:
 *                 continue             # <<<<<<<<<<<<<<
 * 
 *             # Draw a random threshold
 */
      goto __pyx_L3_continue;
      goto __pyx_L8;
    }
    __pyx_L8:;

    /* "sklearn/tree/_tree.pyx":1124
 *             # Draw a random threshold
 *             current_threshold = (min_feature_value +
 *                                  rand_double(random_state) * (max_feature_value - min_feature_value))             # <<<<<<<<<<<<<<
 * 
 *             if current_threshold == max_feature_value:
 */
    __pyx_v_current_threshold = (__pyx_v_min_feature_value + (__pyx_f_7sklearn_4tree_5_tree_rand_double(__pyx_v_random_state) * (__pyx_v_max_feature_value - __pyx_v_min_feature_value)));

    /* "sklearn/tree/_tree.pyx":1126
 *                                  rand_double(random_state) * (max_feature_value - min_feature_value))
 * 
 *             if current_threshold == max_feature_value:             # <<<<<<<<<<<<<<
 *                 current_threshold = min_feature_value
 * 
 */
    __pyx_t_10 = ((__pyx_v_current_threshold == __pyx_v_max_feature_value) != 0);
    if (__pyx_t_10) {

      /* "sklearn/tree/_tree.pyx":1127
 * 
 *             if current_threshold == max_feature_value:
 *                 current_threshold = min_feature_value             # <<<<<<<<<<<<<<
 * 
 *             # Partition
 */
      __pyx_v_current_threshold = __pyx_v_min_feature_value;
      goto __pyx_L9;
    }
    __pyx_L9:;

    /* "sklearn/tree/_tree.pyx":1130
 * 
 *             # Partition
 *             partition_start = start             # <<<<<<<<<<<<<<
 *             partition_end = end
 *             p = start
 */
    __pyx_v_partition_start = __pyx_v_start;

    /* "sklearn/tree/_tree.pyx":1131
 *             # Partition
 *             partition_start = start
 *             partition_end = end             # <<<<<<<<<<<<<<
 *             p = start
 * 
 */
    __pyx_v_partition_end = __pyx_v_end;

    /* "sklearn/tree/_tree.pyx":1132
 *             partition_start = start
 *             partition_end = end
 *             p = start             # <<<<<<<<<<<<<<
 * 
 *             while p < partition_end:
 */
    __pyx_v_p = __pyx_v_start;

    /* "sklearn/tree/_tree.pyx":1134
 *             p = start
 * 
 *             while p < partition_end:             # <<<<<<<<<<<<<<
 *                 if X[samples[p], current_feature] <= current_threshold:
 *                     p += 1
 */
    while (1) {
      __pyx_t_10 = ((__pyx_v_p < __pyx_v_partition_end) != 0);
      if (!__pyx_t_10) break;

      /* "sklearn/tree/_tree.pyx":1135
 * 
 *             while p < partition_end:
 *                 if X[samples[p], current_feature] <= current_threshold:             # <<<<<<<<<<<<<<
 *                     p += 1
 * 
 */
      __pyx_t_7 = (__pyx_v_samples[__pyx_v_p]);
      __pyx_t_11 = __pyx_v_current_feature;
      __pyx_t_10 = (((*__Pyx_BufPtrCContig2d(__pyx_t_7sklearn_4tree_5_tree_DTYPE_t *, __pyx_pybuffernd_X.rcbuffer->pybuffer.buf, __pyx_t_7, __pyx_pybuffernd_X.diminfo[0].strides, __pyx_t_11, __pyx_pybuffernd_X.diminfo[1].strides)) <= __pyx_v_current_threshold) != 0);
      if (__pyx_t_10) {

        /* "sklearn/tree/_tree.pyx":1136
 *             while p < partition_end:
 *                 if X[samples[p], current_feature] <= current_threshold:
 *                     p += 1             # <<<<<<<<<<<<<<
 * 
 *                 else:
 */
        __pyx_v_p = (__pyx_v_p + 1);
        goto __pyx_L12;
      }
      /*else*/ {

        /* "sklearn/tree/_tree.pyx":1139
 * 
 *                 else:
 *                     partition_end -= 1             # <<<<<<<<<<<<<<
 * 
 *                     tmp = samples[partition_end]
 */
        __pyx_v_partition_end = (__pyx_v_partition_end - 1);

        /* "sklearn/tree/_tree.pyx":1141
 *                     partition_end -= 1
 * 
 *                     tmp = samples[partition_end]             # <<<<<<<<<<<<<<
 *                     samples[partition_end] = samples[p]
 *                     samples[p] = tmp
 */
        __pyx_v_tmp = (__pyx_v_samples[__pyx_v_partition_end]);

        /* "sklearn/tree/_tree.pyx":1142
 * 
 *                     tmp = samples[partition_end]
 *                     samples[partition_end] = samples[p]             # <<<<<<<<<<<<<<
 *                     samples[p] = tmp
 * 
 */
        (__pyx_v_samples[__pyx_v_partition_end]) = (__pyx_v_samples[__pyx_v_p]);

        /* "sklearn/tree/_tree.pyx":1143
 *                     tmp = samples[partition_end]
 *                     samples[partition_end] = samples[p]
 *                     samples[p] = tmp             # <<<<<<<<<<<<<<
 * 
 *             current_pos = partition_end
 */
        (__pyx_v_samples[__pyx_v_p]) = __pyx_v_tmp;
      }
      __pyx_L12:;
    }

    /* "sklearn/tree/_tree.pyx":1145
 *                     samples[p] = tmp
 * 
 *             current_pos = partition_end             # <<<<<<<<<<<<<<
 * 
 *             # Reject if min_samples_leaf is not guaranteed
 */
    __pyx_v_current_pos = __pyx_v_partition_end;

    /* "sklearn/tree/_tree.pyx":1148
 * 
 *             # Reject if min_samples_leaf is not guaranteed
 *             if (((current_pos - start) < min_samples_leaf) or             # <<<<<<<<<<<<<<
 *                 ((end - current_pos) < min_samples_leaf)):
 *                continue
 */
    __pyx_t_10 = (((__pyx_v_current_pos - __pyx_v_start) < __pyx_v_min_samples_leaf) != 0);
    if (!__pyx_t_10) {

      /* "sklearn/tree/_tree.pyx":1149
 *             # Reject if min_samples_leaf is not guaranteed
 *             if (((current_pos - start) < min_samples_leaf) or
 *                 ((end - current_pos) < min_samples_leaf)):             # <<<<<<<<<<<<<<
 *                continue
 * 
 */
      __pyx_t_12 = (((__pyx_v_end - __pyx_v_current_pos) < __pyx_v_min_samples_leaf) != 0);
      __pyx_t_13 = __pyx_t_12;
    } else {
      __pyx_t_13 = __pyx_t_10;
    }
    if (__pyx_t_13) {

      /* "sklearn/tree/_tree.pyx":1150
 *             if (((current_pos - start) < min_samples_leaf) or
 *                 ((end - current_pos) < min_samples_leaf)):
 *                continue             # <<<<<<<<<<<<<<
 * 
 *             # Evaluate split
 */
      goto __pyx_L3_continue;
      goto __pyx_L13;
    }
    __pyx_L13:;

    /* "sklearn/tree/_tree.pyx":1153
 * 
 *             # Evaluate split
 *             criterion.reset()             # <<<<<<<<<<<<<<
 *             criterion.update(current_pos)
 *             current_impurity = criterion.children_impurity()
 */
    ((struct __pyx_vtabstruct_7sklearn_4tree_5_tree_Criterion *)__pyx_v_criterion->__pyx_vtab)->reset(__pyx_v_criterion);

    /* "sklearn/tree/_tree.pyx":1154
 *             # Evaluate split
 *             criterion.reset()
 *             criterion.update(current_pos)             # <<<<<<<<<<<<<<
 *             current_impurity = criterion.children_impurity()
 * 
 */
    ((struct __pyx_vtabstruct_7sklearn_4tree_5_tree_Criterion *)__pyx_v_criterion->__pyx_vtab)->update(__pyx_v_criterion, __pyx_v_current_pos);

    /* "sklearn/tree/_tree.pyx":1155
 *             criterion.reset()
 *             criterion.update(current_pos)
 *             current_impurity = criterion.children_impurity()             # <<<<<<<<<<<<<<
 * 
 *             if current_impurity < best_impurity:
 */
    __pyx_v_current_impurity = ((struct __pyx_vtabstruct_7sklearn_4tree_5_tree_Criterion *)__pyx_v_criterion->__pyx_vtab)->children_impurity(__pyx_v_criterion);

    /* "sklearn/tree/_tree.pyx":1157
 *             current_impurity = criterion.children_impurity()
 * 
 *             if current_impurity < best_impurity:             # <<<<<<<<<<<<<<
 *                 best_impurity = current_impurity
 *                 best_pos = current_pos
 */
    __pyx_t_13 = ((__pyx_v_current_impurity < __pyx_v_best_impurity) != 0);
    if (__pyx_t_13) {

      /* "sklearn/tree/_tree.pyx":1158
 * 
 *             if current_impurity < best_impurity:
 *                 best_impurity = current_impurity             # <<<<<<<<<<<<<<
 *                 best_pos = current_pos
 *                 best_feature = current_feature
 */
      __pyx_v_best_impurity = __pyx_v_current_impurity;

      /* "sklearn/tree/_tree.pyx":1159
 *             if current_impurity < best_impurity:
 *                 best_impurity = current_impurity
 *                 best_pos = current_pos             # <<<<<<<<<<<<<<
 *                 best_feature = current_feature
 *                 best_threshold = current_threshold
 */
      __pyx_v_best_pos = __pyx_v_current_pos;

      /* "sklearn/tree/_tree.pyx":1160
 *                 best_impurity = current_impurity
 *                 best_pos = current_pos
 *                 best_feature = current_feature             # <<<<<<<<<<<<<<
 *                 best_threshold = current_threshold
 * 
 */
      __pyx_v_best_feature = __pyx_v_current_feature;

      /* "sklearn/tree/_tree.pyx":1161
 *                 best_pos = current_pos
 *                 best_feature = current_feature
 *                 best_threshold = current_threshold             # <<<<<<<<<<<<<<
 * 
 *             # Count one more visited feature
 */
      __pyx_v_best_threshold = __pyx_v_current_threshold;
      goto __pyx_L14;
    }
    __pyx_L14:;

    /* "sklearn/tree/_tree.pyx":1164
 * 
 *             # Count one more visited feature
 *             visited_features += 1             # <<<<<<<<<<<<<<
 * 
 *             if visited_features >= max_features:
 */
    __pyx_v_visited_features = (__pyx_v_visited_features + 1);

    /* "sklearn/tree/_tree.pyx":1166
 *             visited_features += 1
 * 
 *             if visited_features >= max_features:             # <<<<<<<<<<<<<<
 *                 break
 * 
 */
    __pyx_t_13 = ((__pyx_v_visited_features >= __pyx_v_max_features) != 0);
    if (__pyx_t_13) {

      /* "sklearn/tree/_tree.pyx":1167
 * 
 *             if visited_features >= max_features:
 *                 break             # <<<<<<<<<<<<<<
 * 
 *         # Reorganize samples into samples[start:best_pos] + samples[best_pos:end]
 */
      goto __pyx_L4_break;
      goto __pyx_L15;
    }
    __pyx_L15:;
    __pyx_L3_continue:;
  }
  __pyx_L4_break:;

  /* "sklearn/tree/_tree.pyx":1170
 * 
 *         # Reorganize samples into samples[start:best_pos] + samples[best_pos:end]
 *         if best_pos < end and current_feature != best_feature:             # <<<<<<<<<<<<<<
 *             partition_start = start
 *             partition_end = end
 */
  __pyx_t_13 = ((__pyx_v_best_pos < __pyx_v_end) != 0);
  if (__pyx_t_13) {
    __pyx_t_10 = ((__pyx_v_current_feature != __pyx_v_best_feature) != 0);
    __pyx_t_12 = __pyx_t_10;
  } else {
    __pyx_t_12 = __pyx_t_13;
  }
  if (__pyx_t_12) {

    /* "sklearn/tree/_tree.pyx":1171
 *         # Reorganize samples into samples[start:best_pos] + samples[best_pos:end]
 *         if best_pos < end and current_feature != best_feature:
 *             partition_start = start             # <<<<<<<<<<<<<<
 *             partition_end = end
 *             p = start
 */
    __pyx_v_partition_start = __pyx_v_start;

    /* "sklearn/tree/_tree.pyx":1172
 *         if best_pos < end and current_feature != best_feature:
 *             partition_start = start
 *             partition_end = end             # <<<<<<<<<<<<<<
 *             p = start
 * 
 */
    __pyx_v_partition_end = __pyx_v_end;

    /* "sklearn/tree/_tree.pyx":1173
 *             partition_start = start
 *             partition_end = end
 *             p = start             # <<<<<<<<<<<<<<
 * 
 *             while p < partition_end:
 */
    __pyx_v_p = __pyx_v_start;

    /* "sklearn/tree/_tree.pyx":1175
 *             p = start
 * 
 *             while p < partition_end:             # <<<<<<<<<<<<<<
 *                 if X[samples[p], best_feature] <= best_threshold:
 *                     p += 1
 */
    while (1) {
      __pyx_t_12 = ((__pyx_v_p < __pyx_v_partition_end) != 0);
      if (!__pyx_t_12) break;

      /* "sklearn/tree/_tree.pyx":1176
 * 
 *             while p < partition_end:
 *                 if X[samples[p], best_feature] <= best_threshold:             # <<<<<<<<<<<<<<
 *                     p += 1
 * 
 */
      __pyx_t_3 = (__pyx_v_samples[__pyx_v_p]);
      __pyx_t_14 = __pyx_v_best_feature;
      __pyx_t_12 = (((*__Pyx_BufPtrCContig2d(__pyx_t_7sklearn_4tree_5_tree_DTYPE_t *, __pyx_pybuffernd_X.rcbuffer->pybuffer.buf, __pyx_t_3, __pyx_pybuffernd_X.diminfo[0].strides, __pyx_t_14, __pyx_pybuffernd_X.diminfo[1].strides)) <= __pyx_v_best_threshold) != 0);
      if (__pyx_t_12) {

        /* "sklearn/tree/_tree.pyx":1177
 *             while p < partition_end:
 *                 if X[samples[p], best_feature] <= best_threshold:
 *                     p += 1             # <<<<<<<<<<<<<<
 * 
 *                 else:
 */
        __pyx_v_p = (__pyx_v_p + 1);
        goto __pyx_L19;
      }
      /*else*/ {

        /* "sklearn/tree/_tree.pyx":1180
 * 
 *                 else:
 *                     partition_end -= 1             # <<<<<<<<<<<<<<
 * 
 *                     tmp = samples[partition_end]
 */
        __pyx_v_partition_end = (__pyx_v_partition_end - 1);

        /* "sklearn/tree/_tree.pyx":1182
 *                     partition_end -= 1
 * 
 *                     tmp = samples[partition_end]             # <<<<<<<<<<<<<<
 *                     samples[partition_end] = samples[p]
 *                     samples[p] = tmp
 */
        __pyx_v_tmp = (__pyx_v_samples[__pyx_v_partition_end]);

        /* "sklearn/tree/_tree.pyx":1183
 * 
 *                     tmp = samples[partition_end]
 *                     samples[partition_end] = samples[p]             # <<<<<<<<<<<<<<
 *                     samples[p] = tmp
 * 
 */
        (__pyx_v_samples[__pyx_v_partition_end]) = (__pyx_v_samples[__pyx_v_p]);

        /* "sklearn/tree/_tree.pyx":1184
 *                     tmp = samples[partition_end]
 *                     samples[partition_end] = samples[p]
 *                     samples[p] = tmp             # <<<<<<<<<<<<<<
 * 
 *         # Return values
 */
        (__pyx_v_samples[__pyx_v_p]) = __pyx_v_tmp;
      }
      __pyx_L19:;
    }
    goto __pyx_L16;
  }
  __pyx_L16:;

  /* "sklearn/tree/_tree.pyx":1187
 * 
 *         # Return values
 *         pos[0] = best_pos             # <<<<<<<<<<<<<<
 *         feature[0] = best_feature
 *         threshold[0] = best_threshold
 */
  (__pyx_v_pos[0]) = __pyx_v_best_pos;

  /* "sklearn/tree/_tree.pyx":1188
 *         # Return values
 *         pos[0] = best_pos
 *         feature[0] = best_feature             # <<<<<<<<<<<<<<
 *         threshold[0] = best_threshold
 * 
 */
  (__pyx_v_feature[0]) = __pyx_v_best_feature;

  /* "sklearn/tree/_tree.pyx":1189
 *         pos[0] = best_pos
 *         feature[0] = best_feature
 *         threshold[0] = best_threshold             # <<<<<<<<<<<<<<
 * 
 * 
 */
  (__pyx_v_threshold[0]) = __pyx_v_best_threshold;

  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  { PyObject *__pyx_type, *__pyx_value, *__pyx_tb;
    __Pyx_ErrFetch(&__pyx_type, &__pyx_value, &__pyx_tb);
    __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_X.rcbuffer->pybuffer);
  __Pyx_ErrRestore(__pyx_type, __pyx_value, __pyx_tb);}
  __Pyx_WriteUnraisable("sklearn.tree._tree.RandomSplitter.node_split", __pyx_clineno, __pyx_lineno, __pyx_filename);
  goto __pyx_L2;
  __pyx_L0:;
  __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_X.rcbuffer->pybuffer);
  __pyx_L2:;
  __Pyx_XDECREF((PyObject *)__pyx_v_criterion);
  __Pyx_XDECREF((PyObject *)__pyx_v_X);
  __Pyx_RefNannyFinishContext();
}

/* Python wrapper */
static PyObject *__pyx_pw_7sklearn_4tree_5_tree_4Tree_9n_classes_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_7sklearn_4tree_5_tree_4Tree_9n_classes_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_7sklearn_4tree_5_tree_4Tree_9n_classes___get__(((struct __pyx_obj_7sklearn_4tree_5_tree_Tree *)__pyx_v_self));
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "sklearn/tree/_tree.pyx":1244
 *     # Wrap for outside world
 *     property n_classes:
 *         def __get__(self):             # <<<<<<<<<<<<<<
 *             return sizet_ptr_to_ndarray(self.n_classes, self.n_outputs)
 * 
 */

static PyObject *__pyx_pf_7sklearn_4tree_5_tree_4Tree_9n_classes___get__(struct __pyx_obj_7sklearn_4tree_5_tree_Tree *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "sklearn/tree/_tree.pyx":1245
 *     property n_classes:
 *         def __get__(self):
 *             return sizet_ptr_to_ndarray(self.n_classes, self.n_outputs)             # <<<<<<<<<<<<<<
 * 
 *     property children_left:
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = ((PyObject *)__pyx_f_7sklearn_4tree_5_tree_sizet_ptr_to_ndarray(__pyx_v_self->n_classes, __pyx_v_self->n_outputs)); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1245; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("sklearn.tree._tree.Tree.n_classes.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_7sklearn_4tree_5_tree_4Tree_13children_left_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_7sklearn_4tree_5_tree_4Tree_13children_left_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_7sklearn_4tree_5_tree_4Tree_13children_left___get__(((struct __pyx_obj_7sklearn_4tree_5_tree_Tree *)__pyx_v_self));
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "sklearn/tree/_tree.pyx":1248
 * 
 *     property children_left:
 *         def __get__(self):             # <<<<<<<<<<<<<<
 *             return sizet_ptr_to_ndarray(self.children_left, self.node_count)
 * 
 */

static PyObject *__pyx_pf_7sklearn_4tree_5_tree_4Tree_13children_left___get__(struct __pyx_obj_7sklearn_4tree_5_tree_Tree *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "sklearn/tree/_tree.pyx":1249
 *     property children_left:
 *         def __get__(self):
 *             return sizet_ptr_to_ndarray(self.children_left, self.node_count)             # <<<<<<<<<<<<<<
 * 
 *     property children_right:
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = ((PyObject *)__pyx_f_7sklearn_4tree_5_tree_sizet_ptr_to_ndarray(__pyx_v_self->children_left, __pyx_v_self->node_count)); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1249; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("sklearn.tree._tree.Tree.children_left.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_7sklearn_4tree_5_tree_4Tree_14children_right_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_7sklearn_4tree_5_tree_4Tree_14children_right_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_7sklearn_4tree_5_tree_4Tree_14children_right___get__(((struct __pyx_obj_7sklearn_4tree_5_tree_Tree *)__pyx_v_self));
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "sklearn/tree/_tree.pyx":1252
 * 
 *     property children_right:
 *         def __get__(self):             # <<<<<<<<<<<<<<
 *             return sizet_ptr_to_ndarray(self.children_right, self.node_count)
 * 
 */

static PyObject *__pyx_pf_7sklearn_4tree_5_tree_4Tree_14children_right___get__(struct __pyx_obj_7sklearn_4tree_5_tree_Tree *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "sklearn/tree/_tree.pyx":1253
 *     property children_right:
 *         def __get__(self):
 *             return sizet_ptr_to_ndarray(self.children_right, self.node_count)             # <<<<<<<<<<<<<<
 * 
 *     property feature:
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = ((PyObject *)__pyx_f_7sklearn_4tree_5_tree_sizet_ptr_to_ndarray(__pyx_v_self->children_right, __pyx_v_self->node_count)); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1253; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("sklearn.tree._tree.Tree.children_right.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_7sklearn_4tree_5_tree_4Tree_7feature_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_7sklearn_4tree_5_tree_4Tree_7feature_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_7sklearn_4tree_5_tree_4Tree_7feature___get__(((struct __pyx_obj_7sklearn_4tree_5_tree_Tree *)__pyx_v_self));
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "sklearn/tree/_tree.pyx":1256
 * 
 *     property feature:
 *         def __get__(self):             # <<<<<<<<<<<<<<
 *             return sizet_ptr_to_ndarray(self.feature, self.node_count)
 * 
 */

static PyObject *__pyx_pf_7sklearn_4tree_5_tree_4Tree_7feature___get__(struct __pyx_obj_7sklearn_4tree_5_tree_Tree *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "sklearn/tree/_tree.pyx":1257
 *     property feature:
 *         def __get__(self):
 *             return sizet_ptr_to_ndarray(self.feature, self.node_count)             # <<<<<<<<<<<<<<
 * 
 *     property threshold:
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = ((PyObject *)__pyx_f_7sklearn_4tree_5_tree_sizet_ptr_to_ndarray(__pyx_v_self->feature, __pyx_v_self->node_count)); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1257; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("sklearn.tree._tree.Tree.feature.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_7sklearn_4tree_5_tree_4Tree_9threshold_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_7sklearn_4tree_5_tree_4Tree_9threshold_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_7sklearn_4tree_5_tree_4Tree_9threshold___get__(((struct __pyx_obj_7sklearn_4tree_5_tree_Tree *)__pyx_v_self));
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "sklearn/tree/_tree.pyx":1260
 * 
 *     property threshold:
 *         def __get__(self):             # <<<<<<<<<<<<<<
 *             return double_ptr_to_ndarray(self.threshold, self.node_count)
 * 
 */

static PyObject *__pyx_pf_7sklearn_4tree_5_tree_4Tree_9threshold___get__(struct __pyx_obj_7sklearn_4tree_5_tree_Tree *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "sklearn/tree/_tree.pyx":1261
 *     property threshold:
 *         def __get__(self):
 *             return double_ptr_to_ndarray(self.threshold, self.node_count)             # <<<<<<<<<<<<<<
 * 
 *     property value:
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = ((PyObject *)__pyx_f_7sklearn_4tree_5_tree_double_ptr_to_ndarray(__pyx_v_self->threshold, __pyx_v_self->node_count)); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1261; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("sklearn.tree._tree.Tree.threshold.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_7sklearn_4tree_5_tree_4Tree_5value_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_7sklearn_4tree_5_tree_4Tree_5value_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_7sklearn_4tree_5_tree_4Tree_5value___get__(((struct __pyx_obj_7sklearn_4tree_5_tree_Tree *)__pyx_v_self));
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "sklearn/tree/_tree.pyx":1264
 * 
 *     property value:
 *         def __get__(self):             # <<<<<<<<<<<<<<
 *             cdef np.npy_intp shape[3]
 * 
 */

static PyObject *__pyx_pf_7sklearn_4tree_5_tree_4Tree_5value___get__(struct __pyx_obj_7sklearn_4tree_5_tree_Tree *__pyx_v_self) {
  npy_intp __pyx_v_shape[3];
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "sklearn/tree/_tree.pyx":1267
 *             cdef np.npy_intp shape[3]
 * 
 *             shape[0] = <np.npy_intp> self.node_count             # <<<<<<<<<<<<<<
 *             shape[1] = <np.npy_intp> self.n_outputs
 *             shape[2] = <np.npy_intp> self.max_n_classes
 */
  (__pyx_v_shape[0]) = ((npy_intp)__pyx_v_self->node_count);

  /* "sklearn/tree/_tree.pyx":1268
 * 
 *             shape[0] = <np.npy_intp> self.node_count
 *             shape[1] = <np.npy_intp> self.n_outputs             # <<<<<<<<<<<<<<
 *             shape[2] = <np.npy_intp> self.max_n_classes
 * 
 */
  (__pyx_v_shape[1]) = ((npy_intp)__pyx_v_self->n_outputs);

  /* "sklearn/tree/_tree.pyx":1269
 *             shape[0] = <np.npy_intp> self.node_count
 *             shape[1] = <np.npy_intp> self.n_outputs
 *             shape[2] = <np.npy_intp> self.max_n_classes             # <<<<<<<<<<<<<<
 * 
 *             return np.PyArray_SimpleNewFromData(
 */
  (__pyx_v_shape[2]) = ((npy_intp)__pyx_v_self->max_n_classes);

  /* "sklearn/tree/_tree.pyx":1271
 *             shape[2] = <np.npy_intp> self.max_n_classes
 * 
 *             return np.PyArray_SimpleNewFromData(             # <<<<<<<<<<<<<<
 *                 3, shape, np.NPY_DOUBLE, self.value)
 * 
 */
  __Pyx_XDECREF(__pyx_r);

  /* "sklearn/tree/_tree.pyx":1272
 * 
 *             return np.PyArray_SimpleNewFromData(
 *                 3, shape, np.NPY_DOUBLE, self.value)             # <<<<<<<<<<<<<<
 * 
 *     property impurity:
 */
  __pyx_t_1 = PyArray_SimpleNewFromData(3, __pyx_v_shape, NPY_DOUBLE, __pyx_v_self->value); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1271; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("sklearn.tree._tree.Tree.value.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_7sklearn_4tree_5_tree_4Tree_8impurity_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_7sklearn_4tree_5_tree_4Tree_8impurity_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_7sklearn_4tree_5_tree_4Tree_8impurity___get__(((struct __pyx_obj_7sklearn_4tree_5_tree_Tree *)__pyx_v_self));
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "sklearn/tree/_tree.pyx":1275
 * 
 *     property impurity:
 *         def __get__(self):             # <<<<<<<<<<<<<<
 *             return double_ptr_to_ndarray(self.impurity, self.node_count)
 * 
 */

static PyObject *__pyx_pf_7sklearn_4tree_5_tree_4Tree_8impurity___get__(struct __pyx_obj_7sklearn_4tree_5_tree_Tree *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "sklearn/tree/_tree.pyx":1276
 *     property impurity:
 *         def __get__(self):
 *             return double_ptr_to_ndarray(self.impurity, self.node_count)             # <<<<<<<<<<<<<<
 * 
 *     property n_node_samples:
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = ((PyObject *)__pyx_f_7sklearn_4tree_5_tree_double_ptr_to_ndarray(__pyx_v_self->impurity, __pyx_v_self->node_count)); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1276; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("sklearn.tree._tree.Tree.impurity.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_7sklearn_4tree_5_tree_4Tree_14n_node_samples_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_7sklearn_4tree_5_tree_4Tree_14n_node_samples_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_7sklearn_4tree_5_tree_4Tree_14n_node_samples___get__(((struct __pyx_obj_7sklearn_4tree_5_tree_Tree *)__pyx_v_self));
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "sklearn/tree/_tree.pyx":1279
 * 
 *     property n_node_samples:
 *         def __get__(self):             # <<<<<<<<<<<<<<
 *             return sizet_ptr_to_ndarray(self.n_node_samples, self.node_count)
 * 
 */

static PyObject *__pyx_pf_7sklearn_4tree_5_tree_4Tree_14n_node_samples___get__(struct __pyx_obj_7sklearn_4tree_5_tree_Tree *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "sklearn/tree/_tree.pyx":1280
 *     property n_node_samples:
 *         def __get__(self):
 *             return sizet_ptr_to_ndarray(self.n_node_samples, self.node_count)             # <<<<<<<<<<<<<<
 * 
 *     def __cinit__(self, int n_features, np.ndarray[SIZE_t, ndim=1] n_classes,
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = ((PyObject *)__pyx_f_7sklearn_4tree_5_tree_sizet_ptr_to_ndarray(__pyx_v_self->n_node_samples, __pyx_v_self->node_count)); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1280; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("sklearn.tree._tree.Tree.n_node_samples.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static int __pyx_pw_7sklearn_4tree_5_tree_4Tree_1__cinit__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static int __pyx_pw_7sklearn_4tree_5_tree_4Tree_1__cinit__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  int __pyx_v_n_features;
  PyArrayObject *__pyx_v_n_classes = 0;
  int __pyx_v_n_outputs;
  struct __pyx_obj_7sklearn_4tree_5_tree_Splitter *__pyx_v_splitter = 0;
  __pyx_t_7sklearn_4tree_5_tree_SIZE_t __pyx_v_max_depth;
  __pyx_t_7sklearn_4tree_5_tree_SIZE_t __pyx_v_min_samples_split;
  __pyx_t_7sklearn_4tree_5_tree_SIZE_t __pyx_v_min_samples_leaf;
  PyObject *__pyx_v_random_state = 0;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__cinit__ (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s__n_features,&__pyx_n_s__n_classes,&__pyx_n_s__n_outputs,&__pyx_n_s__splitter,&__pyx_n_s__max_depth,&__pyx_n_s__min_samples_split,&__pyx_n_s__min_samples_leaf,&__pyx_n_s__random_state,0};
    PyObject* values[8] = {0,0,0,0,0,0,0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  8: values[7] = PyTuple_GET_ITEM(__pyx_args, 7);
        case  7: values[6] = PyTuple_GET_ITEM(__pyx_args, 6);
        case  6: values[5] = PyTuple_GET_ITEM(__pyx_args, 5);
        case  5: values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s__n_features)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s__n_classes)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__cinit__", 1, 8, 8, 1); {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1282; __pyx_clineno = __LINE__; goto __pyx_L3_error;}
        }
        case  2:
        if (likely((values[2] = PyDict_GetItem(__pyx_kwds, __pyx_n_s__n_outputs)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__cinit__", 1, 8, 8, 2); {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1282; __pyx_clineno = __LINE__; goto __pyx_L3_error;}
        }
        case  3:
        if (likely((values[3] = PyDict_GetItem(__pyx_kwds, __pyx_n_s__splitter)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__cinit__", 1, 8, 8, 3); {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1282; __pyx_clineno = __LINE__; goto __pyx_L3_error;}
        }
        case  4:
        if (likely((values[4] = PyDict_GetItem(__pyx_kwds, __pyx_n_s__max_depth)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__cinit__", 1, 8, 8, 4); {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1282; __pyx_clineno = __LINE__; goto __pyx_L3_error;}
        }
        case  5:
        if (likely((values[5] = PyDict_GetItem(__pyx_kwds, __pyx_n_s__min_samples_split)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__cinit__", 1, 8, 8, 5); {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1282; __pyx_clineno = __LINE__; goto __pyx_L3_error;}
        }
        case  6:
        if (likely((values[6] = PyDict_GetItem(__pyx_kwds, __pyx_n_s__min_samples_leaf)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__cinit__", 1, 8, 8, 6); {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1282; __pyx_clineno = __LINE__; goto __pyx_L3_error;}
        }
        case  7:
        if (likely((values[7] = PyDict_GetItem(__pyx_kwds, __pyx_n_s__random_state)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__cinit__", 1, 8, 8, 7); {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1282; __pyx_clineno = __LINE__; goto __pyx_L3_error;}
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "__cinit__") < 0)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1282; __pyx_clineno = __LINE__; goto __pyx_L3_error;}
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 8) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
      values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
      values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
      values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
      values[5] = PyTuple_GET_ITEM(__pyx_args, 5);
      values[6] = PyTuple_GET_ITEM(__pyx_args, 6);
      values[7] = PyTuple_GET_ITEM(__pyx_args, 7);
    }
    __pyx_v_n_features = __Pyx_PyInt_AsInt(values[0]); if (unlikely((__pyx_v_n_features == (int)-1) && PyErr_Occurred())) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1282; __pyx_clineno = __LINE__; goto __pyx_L3_error;}
    __pyx_v_n_classes = ((PyArrayObject *)values[1]);
    __pyx_v_n_outputs = __Pyx_PyInt_AsInt(values[2]); if (unlikely((__pyx_v_n_outputs == (int)-1) && PyErr_Occurred())) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1283; __pyx_clineno = __LINE__; goto __pyx_L3_error;}
    __pyx_v_splitter = ((struct __pyx_obj_7sklearn_4tree_5_tree_Splitter *)values[3]);
    __pyx_v_max_depth = __Pyx_PyInt_from_py_Py_intptr_t(values[4]); if (unlikely((__pyx_v_max_depth == (npy_intp)-1) && PyErr_Occurred())) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1283; __pyx_clineno = __LINE__; goto __pyx_L3_error;}
    __pyx_v_min_samples_split = __Pyx_PyInt_from_py_Py_intptr_t(values[5]); if (unlikely((__pyx_v_min_samples_split == (npy_intp)-1) && PyErr_Occurred())) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1284; __pyx_clineno = __LINE__; goto __pyx_L3_error;}
    __pyx_v_min_samples_leaf = __Pyx_PyInt_from_py_Py_intptr_t(values[6]); if (unlikely((__pyx_v_min_samples_leaf == (npy_intp)-1) && PyErr_Occurred())) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1284; __pyx_clineno = __LINE__; goto __pyx_L3_error;}
    __pyx_v_random_state = values[7];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__cinit__", 1, 8, 8, PyTuple_GET_SIZE(__pyx_args)); {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1282; __pyx_clineno = __LINE__; goto __pyx_L3_error;}
  __pyx_L3_error:;
  __Pyx_AddTraceback("sklearn.tree._tree.Tree.__cinit__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return -1;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_n_classes), __pyx_ptype_5numpy_ndarray, 1, "n_classes", 0))) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1282; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_splitter), __pyx_ptype_7sklearn_4tree_5_tree_Splitter, 1, "splitter", 0))) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1283; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __pyx_r = __pyx_pf_7sklearn_4tree_5_tree_4Tree___cinit__(((struct __pyx_obj_7sklearn_4tree_5_tree_Tree *)__pyx_v_self), __pyx_v_n_features, __pyx_v_n_classes, __pyx_v_n_outputs, __pyx_v_splitter, __pyx_v_max_depth, __pyx_v_min_samples_split, __pyx_v_min_samples_leaf, __pyx_v_random_state);
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "sklearn/tree/_tree.pyx":1282
 *             return sizet_ptr_to_ndarray(self.n_node_samples, self.node_count)
 * 
 *     def __cinit__(self, int n_features, np.ndarray[SIZE_t, ndim=1] n_classes,             # <<<<<<<<<<<<<<
 *                         int n_outputs, Splitter splitter, SIZE_t max_depth,
 *                         SIZE_t min_samples_split, SIZE_t min_samples_leaf,
 */

static int __pyx_pf_7sklearn_4tree_5_tree_4Tree___cinit__(struct __pyx_obj_7sklearn_4tree_5_tree_Tree *__pyx_v_self, int __pyx_v_n_features, PyArrayObject *__pyx_v_n_classes, int __pyx_v_n_outputs, struct __pyx_obj_7sklearn_4tree_5_tree_Splitter *__pyx_v_splitter, __pyx_t_7sklearn_4tree_5_tree_SIZE_t __pyx_v_max_depth, __pyx_t_7sklearn_4tree_5_tree_SIZE_t __pyx_v_min_samples_split, __pyx_t_7sklearn_4tree_5_tree_SIZE_t __pyx_v_min_samples_leaf, PyObject *__pyx_v_random_state) {
  __pyx_t_7sklearn_4tree_5_tree_SIZE_t __pyx_v_k;
  __Pyx_LocalBuf_ND __pyx_pybuffernd_n_classes;
  __Pyx_Buffer __pyx_pybuffer_n_classes;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  __pyx_t_7sklearn_4tree_5_tree_SIZE_t __pyx_t_5;
  int __pyx_t_6;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__cinit__", 0);
  __pyx_pybuffer_n_classes.pybuffer.buf = NULL;
  __pyx_pybuffer_n_classes.refcount = 0;
  __pyx_pybuffernd_n_classes.data = NULL;
  __pyx_pybuffernd_n_classes.rcbuffer = &__pyx_pybuffer_n_classes;
  {
    __Pyx_BufFmt_StackElem __pyx_stack[1];
    if (unlikely(__Pyx_GetBufferAndValidate(&__pyx_pybuffernd_n_classes.rcbuffer->pybuffer, (PyObject*)__pyx_v_n_classes, &__Pyx_TypeInfo_nn___pyx_t_7sklearn_4tree_5_tree_SIZE_t, PyBUF_FORMAT| PyBUF_STRIDES, 1, 0, __pyx_stack) == -1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1282; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  }
  __pyx_pybuffernd_n_classes.diminfo[0].strides = __pyx_pybuffernd_n_classes.rcbuffer->pybuffer.strides[0]; __pyx_pybuffernd_n_classes.diminfo[0].shape = __pyx_pybuffernd_n_classes.rcbuffer->pybuffer.shape[0];

  /* "sklearn/tree/_tree.pyx":1288
 *         """Constructor."""
 *         # Input/Output layout
 *         self.n_features = n_features             # <<<<<<<<<<<<<<
 *         self.n_outputs = n_outputs
 *         self.n_classes = <SIZE_t*> malloc(n_outputs * sizeof(SIZE_t))
 */
  __pyx_v_self->n_features = __pyx_v_n_features;

  /* "sklearn/tree/_tree.pyx":1289
 *         # Input/Output layout
 *         self.n_features = n_features
 *         self.n_outputs = n_outputs             # <<<<<<<<<<<<<<
 *         self.n_classes = <SIZE_t*> malloc(n_outputs * sizeof(SIZE_t))
 * 
 */
  __pyx_v_self->n_outputs = __pyx_v_n_outputs;

  /* "sklearn/tree/_tree.pyx":1290
 *         self.n_features = n_features
 *         self.n_outputs = n_outputs
 *         self.n_classes = <SIZE_t*> malloc(n_outputs * sizeof(SIZE_t))             # <<<<<<<<<<<<<<
 * 
 *         if self.n_classes == NULL:
 */
  __pyx_v_self->n_classes = ((__pyx_t_7sklearn_4tree_5_tree_SIZE_t *)malloc((__pyx_v_n_outputs * (sizeof(__pyx_t_7sklearn_4tree_5_tree_SIZE_t)))));

  /* "sklearn/tree/_tree.pyx":1292
 *         self.n_classes = <SIZE_t*> malloc(n_outputs * sizeof(SIZE_t))
 * 
 *         if self.n_classes == NULL:             # <<<<<<<<<<<<<<
 *             raise MemoryError()
 * 
 */
  __pyx_t_1 = ((__pyx_v_self->n_classes == NULL) != 0);
  if (__pyx_t_1) {

    /* "sklearn/tree/_tree.pyx":1293
 * 
 *         if self.n_classes == NULL:
 *             raise MemoryError()             # <<<<<<<<<<<<<<
 * 
 *         self.max_n_classes = np.max(n_classes)
 */
    PyErr_NoMemory(); {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1293; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    goto __pyx_L3;
  }
  __pyx_L3:;

  /* "sklearn/tree/_tree.pyx":1295
 *             raise MemoryError()
 * 
 *         self.max_n_classes = np.max(n_classes)             # <<<<<<<<<<<<<<
 *         self.value_stride = self.n_outputs * self.max_n_classes
 * 
 */
  __pyx_t_2 = __Pyx_GetModuleGlobalName(__pyx_n_s__np); if (unlikely(!__pyx_t_2)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1295; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s__max); if (unlikely(!__pyx_t_3)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1295; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = PyTuple_New(1); if (unlikely(!__pyx_t_2)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1295; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_INCREF(((PyObject *)__pyx_v_n_classes));
  PyTuple_SET_ITEM(__pyx_t_2, 0, ((PyObject *)__pyx_v_n_classes));
  __Pyx_GIVEREF(((PyObject *)__pyx_v_n_classes));
  __pyx_t_4 = PyObject_Call(__pyx_t_3, ((PyObject *)__pyx_t_2), NULL); if (unlikely(!__pyx_t_4)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1295; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(((PyObject *)__pyx_t_2)); __pyx_t_2 = 0;
  __pyx_t_5 = __Pyx_PyInt_from_py_Py_intptr_t(__pyx_t_4); if (unlikely((__pyx_t_5 == (npy_intp)-1) && PyErr_Occurred())) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1295; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_v_self->max_n_classes = __pyx_t_5;

  /* "sklearn/tree/_tree.pyx":1296
 * 
 *         self.max_n_classes = np.max(n_classes)
 *         self.value_stride = self.n_outputs * self.max_n_classes             # <<<<<<<<<<<<<<
 * 
 *         cdef SIZE_t k
 */
  __pyx_v_self->value_stride = (__pyx_v_self->n_outputs * __pyx_v_self->max_n_classes);

  /* "sklearn/tree/_tree.pyx":1300
 *         cdef SIZE_t k
 * 
 *         for k from 0 <= k < n_outputs:             # <<<<<<<<<<<<<<
 *             self.n_classes[k] = n_classes[k]
 * 
 */
  __pyx_t_6 = __pyx_v_n_outputs;
  for (__pyx_v_k = 0; __pyx_v_k < __pyx_t_6; __pyx_v_k++) {

    /* "sklearn/tree/_tree.pyx":1301
 * 
 *         for k from 0 <= k < n_outputs:
 *             self.n_classes[k] = n_classes[k]             # <<<<<<<<<<<<<<
 * 
 *         # Parameters
 */
    __pyx_t_5 = __pyx_v_k;
    (__pyx_v_self->n_classes[__pyx_v_k]) = (*__Pyx_BufPtrStrided1d(__pyx_t_7sklearn_4tree_5_tree_SIZE_t *, __pyx_pybuffernd_n_classes.rcbuffer->pybuffer.buf, __pyx_t_5, __pyx_pybuffernd_n_classes.diminfo[0].strides));
  }

  /* "sklearn/tree/_tree.pyx":1304
 * 
 *         # Parameters
 *         self.splitter = splitter             # <<<<<<<<<<<<<<
 *         self.max_depth = max_depth
 *         self.min_samples_split = min_samples_split
 */
  __Pyx_INCREF(((PyObject *)__pyx_v_splitter));
  __Pyx_GIVEREF(((PyObject *)__pyx_v_splitter));
  __Pyx_GOTREF(__pyx_v_self->splitter);
  __Pyx_DECREF(((PyObject *)__pyx_v_self->splitter));
  __pyx_v_self->splitter = __pyx_v_splitter;

  /* "sklearn/tree/_tree.pyx":1305
 *         # Parameters
 *         self.splitter = splitter
 *         self.max_depth = max_depth             # <<<<<<<<<<<<<<
 *         self.min_samples_split = min_samples_split
 *         self.min_samples_leaf = min_samples_leaf
 */
  __pyx_v_self->max_depth = __pyx_v_max_depth;

  /* "sklearn/tree/_tree.pyx":1306
 *         self.splitter = splitter
 *         self.max_depth = max_depth
 *         self.min_samples_split = min_samples_split             # <<<<<<<<<<<<<<
 *         self.min_samples_leaf = min_samples_leaf
 *         self.random_state = random_state
 */
  __pyx_v_self->min_samples_split = __pyx_v_min_samples_split;

  /* "sklearn/tree/_tree.pyx":1307
 *         self.max_depth = max_depth
 *         self.min_samples_split = min_samples_split
 *         self.min_samples_leaf = min_samples_leaf             # <<<<<<<<<<<<<<
 *         self.random_state = random_state
 * 
 */
  __pyx_v_self->min_samples_leaf = __pyx_v_min_samples_leaf;

  /* "sklearn/tree/_tree.pyx":1308
 *         self.min_samples_split = min_samples_split
 *         self.min_samples_leaf = min_samples_leaf
 *         self.random_state = random_state             # <<<<<<<<<<<<<<
 * 
 *         # Inner structures
 */
  __Pyx_INCREF(__pyx_v_random_state);
  __Pyx_GIVEREF(__pyx_v_random_state);
  __Pyx_GOTREF(__pyx_v_self->random_state);
  __Pyx_DECREF(__pyx_v_self->random_state);
  __pyx_v_self->random_state = __pyx_v_random_state;

  /* "sklearn/tree/_tree.pyx":1311
 * 
 *         # Inner structures
 *         self.node_count = 0             # <<<<<<<<<<<<<<
 *         self.capacity = 0
 *         self.children_left = NULL
 */
  __pyx_v_self->node_count = 0;

  /* "sklearn/tree/_tree.pyx":1312
 *         # Inner structures
 *         self.node_count = 0
 *         self.capacity = 0             # <<<<<<<<<<<<<<
 *         self.children_left = NULL
 *         self.children_right = NULL
 */
  __pyx_v_self->capacity = 0;

  /* "sklearn/tree/_tree.pyx":1313
 *         self.node_count = 0
 *         self.capacity = 0
 *         self.children_left = NULL             # <<<<<<<<<<<<<<
 *         self.children_right = NULL
 *         self.feature = NULL
 */
  __pyx_v_self->children_left = NULL;

  /* "sklearn/tree/_tree.pyx":1314
 *         self.capacity = 0
 *         self.children_left = NULL
 *         self.children_right = NULL             # <<<<<<<<<<<<<<
 *         self.feature = NULL
 *         self.threshold = NULL
 */
  __pyx_v_self->children_right = NULL;

  /* "sklearn/tree/_tree.pyx":1315
 *         self.children_left = NULL
 *         self.children_right = NULL
 *         self.feature = NULL             # <<<<<<<<<<<<<<
 *         self.threshold = NULL
 *         self.value = NULL
 */
  __pyx_v_self->feature = NULL;

  /* "sklearn/tree/_tree.pyx":1316
 *         self.children_right = NULL
 *         self.feature = NULL
 *         self.threshold = NULL             # <<<<<<<<<<<<<<
 *         self.value = NULL
 *         self.impurity = NULL
 */
  __pyx_v_self->threshold = NULL;

  /* "sklearn/tree/_tree.pyx":1317
 *         self.feature = NULL
 *         self.threshold = NULL
 *         self.value = NULL             # <<<<<<<<<<<<<<
 *         self.impurity = NULL
 *         self.n_node_samples = NULL
 */
  __pyx_v_self->value = NULL;

  /* "sklearn/tree/_tree.pyx":1318
 *         self.threshold = NULL
 *         self.value = NULL
 *         self.impurity = NULL             # <<<<<<<<<<<<<<
 *         self.n_node_samples = NULL
 * 
 */
  __pyx_v_self->impurity = NULL;

  /* "sklearn/tree/_tree.pyx":1319
 *         self.value = NULL
 *         self.impurity = NULL
 *         self.n_node_samples = NULL             # <<<<<<<<<<<<<<
 * 
 *     def __dealloc__(self):
 */
  __pyx_v_self->n_node_samples = NULL;

  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  { PyObject *__pyx_type, *__pyx_value, *__pyx_tb;
    __Pyx_ErrFetch(&__pyx_type, &__pyx_value, &__pyx_tb);
    __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_n_classes.rcbuffer->pybuffer);
  __Pyx_ErrRestore(__pyx_type, __pyx_value, __pyx_tb);}
  __Pyx_AddTraceback("sklearn.tree._tree.Tree.__cinit__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  goto __pyx_L2;
  __pyx_L0:;
  __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_n_classes.rcbuffer->pybuffer);
  __pyx_L2:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static void __pyx_pw_7sklearn_4tree_5_tree_4Tree_3__dealloc__(PyObject *__pyx_v_self); /*proto*/
static void __pyx_pw_7sklearn_4tree_5_tree_4Tree_3__dealloc__(PyObject *__pyx_v_self) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__dealloc__ (wrapper)", 0);
  __pyx_pf_7sklearn_4tree_5_tree_4Tree_2__dealloc__(((struct __pyx_obj_7sklearn_4tree_5_tree_Tree *)__pyx_v_self));
  __Pyx_RefNannyFinishContext();
}

/* "sklearn/tree/_tree.pyx":1321
 *         self.n_node_samples = NULL
 * 
 *     def __dealloc__(self):             # <<<<<<<<<<<<<<
 *         """Destructor."""
 *         # Free all inner structures
 */

static void __pyx_pf_7sklearn_4tree_5_tree_4Tree_2__dealloc__(struct __pyx_obj_7sklearn_4tree_5_tree_Tree *__pyx_v_self) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__dealloc__", 0);

  /* "sklearn/tree/_tree.pyx":1324
 *         """Destructor."""
 *         # Free all inner structures
 *         free(self.n_classes)             # <<<<<<<<<<<<<<
 *         free(self.children_left)
 *         free(self.children_right)
 */
  free(__pyx_v_self->n_classes);

  /* "sklearn/tree/_tree.pyx":1325
 *         # Free all inner structures
 *         free(self.n_classes)
 *         free(self.children_left)             # <<<<<<<<<<<<<<
 *         free(self.children_right)
 *         free(self.feature)
 */
  free(__pyx_v_self->children_left);

  /* "sklearn/tree/_tree.pyx":1326
 *         free(self.n_classes)
 *         free(self.children_left)
 *         free(self.children_right)             # <<<<<<<<<<<<<<
 *         free(self.feature)
 *         free(self.threshold)
 */
  free(__pyx_v_self->children_right);

  /* "sklearn/tree/_tree.pyx":1327
 *         free(self.children_left)
 *         free(self.children_right)
 *         free(self.feature)             # <<<<<<<<<<<<<<
 *         free(self.threshold)
 *         free(self.value)
 */
  free(__pyx_v_self->feature);

  /* "sklearn/tree/_tree.pyx":1328
 *         free(self.children_right)
 *         free(self.feature)
 *         free(self.threshold)             # <<<<<<<<<<<<<<
 *         free(self.value)
 *         free(self.impurity)
 */
  free(__pyx_v_self->threshold);

  /* "sklearn/tree/_tree.pyx":1329
 *         free(self.feature)
 *         free(self.threshold)
 *         free(self.value)             # <<<<<<<<<<<<<<
 *         free(self.impurity)
 *         free(self.n_node_samples)
 */
  free(__pyx_v_self->value);

  /* "sklearn/tree/_tree.pyx":1330
 *         free(self.threshold)
 *         free(self.value)
 *         free(self.impurity)             # <<<<<<<<<<<<<<
 *         free(self.n_node_samples)
 * 
 */
  free(__pyx_v_self->impurity);

  /* "sklearn/tree/_tree.pyx":1331
 *         free(self.value)
 *         free(self.impurity)
 *         free(self.n_node_samples)             # <<<<<<<<<<<<<<
 * 
 *     def __reduce__(self):
 */
  free(__pyx_v_self->n_node_samples);

  __Pyx_RefNannyFinishContext();
}

/* Python wrapper */
static PyObject *__pyx_pw_7sklearn_4tree_5_tree_4Tree_5__reduce__(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static char __pyx_doc_7sklearn_4tree_5_tree_4Tree_4__reduce__[] = "Reduce re-implementation, for pickling.";
static PyObject *__pyx_pw_7sklearn_4tree_5_tree_4Tree_5__reduce__(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__reduce__ (wrapper)", 0);
  __pyx_r = __pyx_pf_7sklearn_4tree_5_tree_4Tree_4__reduce__(((struct __pyx_obj_7sklearn_4tree_5_tree_Tree *)__pyx_v_self));
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "sklearn/tree/_tree.pyx":1333
 *         free(self.n_node_samples)
 * 
 *     def __reduce__(self):             # <<<<<<<<<<<<<<
 *         """Reduce re-implementation, for pickling."""
 *         return (Tree, (self.n_features,
 */

static PyObject *__pyx_pf_7sklearn_4tree_5_tree_4Tree_4__reduce__(struct __pyx_obj_7sklearn_4tree_5_tree_Tree *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__reduce__", 0);

  /* "sklearn/tree/_tree.pyx":1335
 *     def __reduce__(self):
 *         """Reduce re-implementation, for pickling."""
 *         return (Tree, (self.n_features,             # <<<<<<<<<<<<<<
 *                        sizet_ptr_to_ndarray(self.n_classes, self.n_outputs),
 *                        self.n_outputs,
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyInt_to_py_Py_intptr_t(__pyx_v_self->n_features); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1335; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_1);

  /* "sklearn/tree/_tree.pyx":1336
 *         """Reduce re-implementation, for pickling."""
 *         return (Tree, (self.n_features,
 *                        sizet_ptr_to_ndarray(self.n_classes, self.n_outputs),             # <<<<<<<<<<<<<<
 *                        self.n_outputs,
 *                        self.splitter,
 */
  __pyx_t_2 = ((PyObject *)__pyx_f_7sklearn_4tree_5_tree_sizet_ptr_to_ndarray(__pyx_v_self->n_classes, __pyx_v_self->n_outputs)); if (unlikely(!__pyx_t_2)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1336; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_2);

  /* "sklearn/tree/_tree.pyx":1337
 *         return (Tree, (self.n_features,
 *                        sizet_ptr_to_ndarray(self.n_classes, self.n_outputs),
 *                        self.n_outputs,             # <<<<<<<<<<<<<<
 *                        self.splitter,
 *                        self.max_depth,
 */
  __pyx_t_3 = __Pyx_PyInt_to_py_Py_intptr_t(__pyx_v_self->n_outputs); if (unlikely(!__pyx_t_3)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1337; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_3);

  /* "sklearn/tree/_tree.pyx":1339
 *                        self.n_outputs,
 *                        self.splitter,
 *                        self.max_depth,             # <<<<<<<<<<<<<<
 *                        self.min_samples_split,
 *                        self.min_samples_leaf,
 */
  __pyx_t_4 = __Pyx_PyInt_to_py_Py_intptr_t(__pyx_v_self->max_depth); if (unlikely(!__pyx_t_4)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1339; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_4);

  /* "sklearn/tree/_tree.pyx":1340
 *                        self.splitter,
 *                        self.max_depth,
 *                        self.min_samples_split,             # <<<<<<<<<<<<<<
 *                        self.min_samples_leaf,
 *                        self.random_state), self.__getstate__())
 */
  __pyx_t_5 = __Pyx_PyInt_to_py_Py_intptr_t(__pyx_v_self->min_samples_split); if (unlikely(!__pyx_t_5)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1340; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_5);

  /* "sklearn/tree/_tree.pyx":1341
 *                        self.max_depth,
 *                        self.min_samples_split,
 *                        self.min_samples_leaf,             # <<<<<<<<<<<<<<
 *                        self.random_state), self.__getstate__())
 * 
 */
  __pyx_t_6 = __Pyx_PyInt_to_py_Py_intptr_t(__pyx_v_self->min_samples_leaf); if (unlikely(!__pyx_t_6)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1341; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_6);

  /* "sklearn/tree/_tree.pyx":1342
 *                        self.min_samples_split,
 *                        self.min_samples_leaf,
 *                        self.random_state), self.__getstate__())             # <<<<<<<<<<<<<<
 * 
 *     def __getstate__(self):
 */
  __pyx_t_7 = PyTuple_New(8); if (unlikely(!__pyx_t_7)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1335; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_7);
  PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  PyTuple_SET_ITEM(__pyx_t_7, 1, __pyx_t_2);
  __Pyx_GIVEREF(__pyx_t_2);
  PyTuple_SET_ITEM(__pyx_t_7, 2, __pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_3);
  __Pyx_INCREF(((PyObject *)__pyx_v_self->splitter));
  PyTuple_SET_ITEM(__pyx_t_7, 3, ((PyObject *)__pyx_v_self->splitter));
  __Pyx_GIVEREF(((PyObject *)__pyx_v_self->splitter));
  PyTuple_SET_ITEM(__pyx_t_7, 4, __pyx_t_4);
  __Pyx_GIVEREF(__pyx_t_4);
  PyTuple_SET_ITEM(__pyx_t_7, 5, __pyx_t_5);
  __Pyx_GIVEREF(__pyx_t_5);
  PyTuple_SET_ITEM(__pyx_t_7, 6, __pyx_t_6);
  __Pyx_GIVEREF(__pyx_t_6);
  __Pyx_INCREF(__pyx_v_self->random_state);
  PyTuple_SET_ITEM(__pyx_t_7, 7, __pyx_v_self->random_state);
  __Pyx_GIVEREF(__pyx_v_self->random_state);
  __pyx_t_1 = 0;
  __pyx_t_2 = 0;
  __pyx_t_3 = 0;
  __pyx_t_4 = 0;
  __pyx_t_5 = 0;
  __pyx_t_6 = 0;
  __pyx_t_6 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s____getstate__); if (unlikely(!__pyx_t_6)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1342; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_6);
  __pyx_t_5 = PyObject_Call(__pyx_t_6, ((PyObject *)__pyx_empty_tuple), NULL); if (unlikely(!__pyx_t_5)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1342; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  __pyx_t_6 = PyTuple_New(3); if (unlikely(!__pyx_t_6)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1335; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_6);
  __Pyx_INCREF(((PyObject *)((PyObject*)__pyx_ptype_7sklearn_4tree_5_tree_Tree)));
  PyTuple_SET_ITEM(__pyx_t_6, 0, ((PyObject *)((PyObject*)__pyx_ptype_7sklearn_4tree_5_tree_Tree)));
  __Pyx_GIVEREF(((PyObject *)((PyObject*)__pyx_ptype_7sklearn_4tree_5_tree_Tree)));
  PyTuple_SET_ITEM(__pyx_t_6, 1, ((PyObject *)__pyx_t_7));
  __Pyx_GIVEREF(((PyObject *)__pyx_t_7));
  PyTuple_SET_ITEM(__pyx_t_6, 2, __pyx_t_5);
  __Pyx_GIVEREF(__pyx_t_5);
  __pyx_t_7 = 0;
  __pyx_t_5 = 0;
  __pyx_r = ((PyObject *)__pyx_t_6);
  __pyx_t_6 = 0;
  goto __pyx_L0;

  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_AddTraceback("sklearn.tree._tree.Tree.__reduce__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_7sklearn_4tree_5_tree_4Tree_7__getstate__(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static char __pyx_doc_7sklearn_4tree_5_tree_4Tree_6__getstate__[] = "Getstate re-implementation, for pickling.";
static PyObject *__pyx_pw_7sklearn_4tree_5_tree_4Tree_7__getstate__(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__getstate__ (wrapper)", 0);
  __pyx_r = __pyx_pf_7sklearn_4tree_5_tree_4Tree_6__getstate__(((struct __pyx_obj_7sklearn_4tree_5_tree_Tree *)__pyx_v_self));
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "sklearn/tree/_tree.pyx":1344
 *                        self.random_state), self.__getstate__())
 * 
 *     def __getstate__(self):             # <<<<<<<<<<<<<<
 *         """Getstate re-implementation, for pickling."""
 *         d = {}
 */

static PyObject *__pyx_pf_7sklearn_4tree_5_tree_4Tree_6__getstate__(struct __pyx_obj_7sklearn_4tree_5_tree_Tree *__pyx_v_self) {
  PyObject *__pyx_v_d = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__getstate__", 0);

  /* "sklearn/tree/_tree.pyx":1346
 *     def __getstate__(self):
 *         """Getstate re-implementation, for pickling."""
 *         d = {}             # <<<<<<<<<<<<<<
 * 
 *         d["node_count"] = self.node_count
 */
  __pyx_t_1 = PyDict_New(); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1346; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(((PyObject *)__pyx_t_1));
  __pyx_v_d = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "sklearn/tree/_tree.pyx":1348
 *         d = {}
 * 
 *         d["node_count"] = self.node_count             # <<<<<<<<<<<<<<
 *         d["capacity"] = self.capacity
 *         d["children_left"] = sizet_ptr_to_ndarray(self.children_left, self.capacity)
 */
  __pyx_t_1 = __Pyx_PyInt_to_py_Py_intptr_t(__pyx_v_self->node_count); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1348; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(((PyObject *)__pyx_v_d), ((PyObject *)__pyx_n_s__node_count), __pyx_t_1) < 0) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1348; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "sklearn/tree/_tree.pyx":1349
 * 
 *         d["node_count"] = self.node_count
 *         d["capacity"] = self.capacity             # <<<<<<<<<<<<<<
 *         d["children_left"] = sizet_ptr_to_ndarray(self.children_left, self.capacity)
 *         d["children_right"] = sizet_ptr_to_ndarray(self.children_right, self.capacity)
 */
  __pyx_t_1 = __Pyx_PyInt_to_py_Py_intptr_t(__pyx_v_self->capacity); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1349; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(((PyObject *)__pyx_v_d), ((PyObject *)__pyx_n_s__capacity), __pyx_t_1) < 0) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1349; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "sklearn/tree/_tree.pyx":1350
 *         d["node_count"] = self.node_count
 *         d["capacity"] = self.capacity
 *         d["children_left"] = sizet_ptr_to_ndarray(self.children_left, self.capacity)             # <<<<<<<<<<<<<<
 *         d["children_right"] = sizet_ptr_to_ndarray(self.children_right, self.capacity)
 *         d["feature"] = sizet_ptr_to_ndarray(self.feature, self.capacity)
 */
  __pyx_t_1 = ((PyObject *)__pyx_f_7sklearn_4tree_5_tree_sizet_ptr_to_ndarray(__pyx_v_self->children_left, __pyx_v_self->capacity)); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1350; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(((PyObject *)__pyx_v_d), ((PyObject *)__pyx_n_s__children_left), __pyx_t_1) < 0) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1350; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "sklearn/tree/_tree.pyx":1351
 *         d["capacity"] = self.capacity
 *         d["children_left"] = sizet_ptr_to_ndarray(self.children_left, self.capacity)
 *         d["children_right"] = sizet_ptr_to_ndarray(self.children_right, self.capacity)             # <<<<<<<<<<<<<<
 *         d["feature"] = sizet_ptr_to_ndarray(self.feature, self.capacity)
 *         d["threshold"] = double_ptr_to_ndarray(self.threshold, self.capacity)
 */
  __pyx_t_1 = ((PyObject *)__pyx_f_7sklearn_4tree_5_tree_sizet_ptr_to_ndarray(__pyx_v_self->children_right, __pyx_v_self->capacity)); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1351; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(((PyObject *)__pyx_v_d), ((PyObject *)__pyx_n_s__children_right), __pyx_t_1) < 0) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1351; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "sklearn/tree/_tree.pyx":1352
 *         d["children_left"] = sizet_ptr_to_ndarray(self.children_left, self.capacity)
 *         d["children_right"] = sizet_ptr_to_ndarray(self.children_right, self.capacity)
 *         d["feature"] = sizet_ptr_to_ndarray(self.feature, self.capacity)             # <<<<<<<<<<<<<<
 *         d["threshold"] = double_ptr_to_ndarray(self.threshold, self.capacity)
 *         d["value"] = double_ptr_to_ndarray(self.value, self.capacity * self.value_stride)
 */
  __pyx_t_1 = ((PyObject *)__pyx_f_7sklearn_4tree_5_tree_sizet_ptr_to_ndarray(__pyx_v_self->feature, __pyx_v_self->capacity)); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1352; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(((PyObject *)__pyx_v_d), ((PyObject *)__pyx_n_s__feature), __pyx_t_1) < 0) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1352; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "sklearn/tree/_tree.pyx":1353
 *         d["children_right"] = sizet_ptr_to_ndarray(self.children_right, self.capacity)
 *         d["feature"] = sizet_ptr_to_ndarray(self.feature, self.capacity)
 *         d["threshold"] = double_ptr_to_ndarray(self.threshold, self.capacity)             # <<<<<<<<<<<<<<
 *         d["value"] = double_ptr_to_ndarray(self.value, self.capacity * self.value_stride)
 *         d["impurity"] = double_ptr_to_ndarray(self.impurity, self.capacity)
 */
  __pyx_t_1 = ((PyObject *)__pyx_f_7sklearn_4tree_5_tree_double_ptr_to_ndarray(__pyx_v_self->threshold, __pyx_v_self->capacity)); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1353; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(((PyObject *)__pyx_v_d), ((PyObject *)__pyx_n_s__threshold), __pyx_t_1) < 0) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1353; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "sklearn/tree/_tree.pyx":1354
 *         d["feature"] = sizet_ptr_to_ndarray(self.feature, self.capacity)
 *         d["threshold"] = double_ptr_to_ndarray(self.threshold, self.capacity)
 *         d["value"] = double_ptr_to_ndarray(self.value, self.capacity * self.value_stride)             # <<<<<<<<<<<<<<
 *         d["impurity"] = double_ptr_to_ndarray(self.impurity, self.capacity)
 *         d["n_node_samples"] = sizet_ptr_to_ndarray(self.n_node_samples, self.capacity)
 */
  __pyx_t_1 = ((PyObject *)__pyx_f_7sklearn_4tree_5_tree_double_ptr_to_ndarray(__pyx_v_self->value, (__pyx_v_self->capacity * __pyx_v_self->value_stride))); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1354; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(((PyObject *)__pyx_v_d), ((PyObject *)__pyx_n_s__value), __pyx_t_1) < 0) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1354; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "sklearn/tree/_tree.pyx":1355
 *         d["threshold"] = double_ptr_to_ndarray(self.threshold, self.capacity)
 *         d["value"] = double_ptr_to_ndarray(self.value, self.capacity * self.value_stride)
 *         d["impurity"] = double_ptr_to_ndarray(self.impurity, self.capacity)             # <<<<<<<<<<<<<<
 *         d["n_node_samples"] = sizet_ptr_to_ndarray(self.n_node_samples, self.capacity)
 * 
 */
  __pyx_t_1 = ((PyObject *)__pyx_f_7sklearn_4tree_5_tree_double_ptr_to_ndarray(__pyx_v_self->impurity, __pyx_v_self->capacity)); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1355; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(((PyObject *)__pyx_v_d), ((PyObject *)__pyx_n_s__impurity), __pyx_t_1) < 0) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1355; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "sklearn/tree/_tree.pyx":1356
 *         d["value"] = double_ptr_to_ndarray(self.value, self.capacity * self.value_stride)
 *         d["impurity"] = double_ptr_to_ndarray(self.impurity, self.capacity)
 *         d["n_node_samples"] = sizet_ptr_to_ndarray(self.n_node_samples, self.capacity)             # <<<<<<<<<<<<<<
 * 
 *         return d
 */
  __pyx_t_1 = ((PyObject *)__pyx_f_7sklearn_4tree_5_tree_sizet_ptr_to_ndarray(__pyx_v_self->n_node_samples, __pyx_v_self->capacity)); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1356; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(((PyObject *)__pyx_v_d), ((PyObject *)__pyx_n_s__n_node_samples), __pyx_t_1) < 0) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1356; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "sklearn/tree/_tree.pyx":1358
 *         d["n_node_samples"] = sizet_ptr_to_ndarray(self.n_node_samples, self.capacity)
 * 
 *         return d             # <<<<<<<<<<<<<<
 * 
 *     def __setstate__(self, d):
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(((PyObject *)__pyx_v_d));
  __pyx_r = ((PyObject *)__pyx_v_d);
  goto __pyx_L0;

  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("sklearn.tree._tree.Tree.__getstate__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_d);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_7sklearn_4tree_5_tree_4Tree_9__setstate__(PyObject *__pyx_v_self, PyObject *__pyx_v_d); /*proto*/
static char __pyx_doc_7sklearn_4tree_5_tree_4Tree_8__setstate__[] = "Setstate re-implementation, for unpickling.";
static PyObject *__pyx_pw_7sklearn_4tree_5_tree_4Tree_9__setstate__(PyObject *__pyx_v_self, PyObject *__pyx_v_d) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__setstate__ (wrapper)", 0);
  __pyx_r = __pyx_pf_7sklearn_4tree_5_tree_4Tree_8__setstate__(((struct __pyx_obj_7sklearn_4tree_5_tree_Tree *)__pyx_v_self), ((PyObject *)__pyx_v_d));
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "sklearn/tree/_tree.pyx":1360
 *         return d
 * 
 *     def __setstate__(self, d):             # <<<<<<<<<<<<<<
 *         """Setstate re-implementation, for unpickling."""
 *         self._resize(d["capacity"])
 */

static PyObject *__pyx_pf_7sklearn_4tree_5_tree_4Tree_8__setstate__(struct __pyx_obj_7sklearn_4tree_5_tree_Tree *__pyx_v_self, PyObject *__pyx_v_d) {
  __pyx_t_7sklearn_4tree_5_tree_SIZE_t *__pyx_v_children_left;
  __pyx_t_7sklearn_4tree_5_tree_SIZE_t *__pyx_v_children_right;
  __pyx_t_7sklearn_4tree_5_tree_SIZE_t *__pyx_v_feature;
  double *__pyx_v_threshold;
  double *__pyx_v_value;
  double *__pyx_v_impurity;
  __pyx_t_7sklearn_4tree_5_tree_SIZE_t *__pyx_v_n_node_samples;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __pyx_t_7sklearn_4tree_5_tree_SIZE_t __pyx_t_2;
  struct __pyx_opt_args_7sklearn_4tree_5_tree_4Tree__resize __pyx_t_3;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__setstate__", 0);

  /* "sklearn/tree/_tree.pyx":1362
 *     def __setstate__(self, d):
 *         """Setstate re-implementation, for unpickling."""
 *         self._resize(d["capacity"])             # <<<<<<<<<<<<<<
 *         self.node_count = d["node_count"]
 * 
 */
  __pyx_t_1 = PyObject_GetItem(__pyx_v_d, ((PyObject *)__pyx_n_s__capacity)); if (!__pyx_t_1) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1362; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyInt_from_py_Py_intptr_t(__pyx_t_1); if (unlikely((__pyx_t_2 == (npy_intp)-1) && PyErr_Occurred())) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1362; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_3.__pyx_n = 1;
  __pyx_t_3.capacity = __pyx_t_2;
  ((struct __pyx_vtabstruct_7sklearn_4tree_5_tree_Tree *)__pyx_v_self->__pyx_vtab)->_resize(__pyx_v_self, &__pyx_t_3); 

  /* "sklearn/tree/_tree.pyx":1363
 *         """Setstate re-implementation, for unpickling."""
 *         self._resize(d["capacity"])
 *         self.node_count = d["node_count"]             # <<<<<<<<<<<<<<
 * 
 *         cdef SIZE_t* children_left = <SIZE_t*> (<np.ndarray> d["children_left"]).data
 */
  __pyx_t_1 = PyObject_GetItem(__pyx_v_d, ((PyObject *)__pyx_n_s__node_count)); if (!__pyx_t_1) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1363; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyInt_from_py_Py_intptr_t(__pyx_t_1); if (unlikely((__pyx_t_2 == (npy_intp)-1) && PyErr_Occurred())) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1363; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v_self->node_count = __pyx_t_2;

  /* "sklearn/tree/_tree.pyx":1365
 *         self.node_count = d["node_count"]
 * 
 *         cdef SIZE_t* children_left = <SIZE_t*> (<np.ndarray> d["children_left"]).data             # <<<<<<<<<<<<<<
 *         cdef SIZE_t* children_right =  <SIZE_t*> (<np.ndarray> d["children_right"]).data
 *         cdef SIZE_t* feature = <SIZE_t*> (<np.ndarray> d["feature"]).data
 */
  __pyx_t_1 = PyObject_GetItem(__pyx_v_d, ((PyObject *)__pyx_n_s__children_left)); if (!__pyx_t_1) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1365; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_children_left = ((__pyx_t_7sklearn_4tree_5_tree_SIZE_t *)((PyArrayObject *)__pyx_t_1)->data);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "sklearn/tree/_tree.pyx":1366
 * 
 *         cdef SIZE_t* children_left = <SIZE_t*> (<np.ndarray> d["children_left"]).data
 *         cdef SIZE_t* children_right =  <SIZE_t*> (<np.ndarray> d["children_right"]).data             # <<<<<<<<<<<<<<
 *         cdef SIZE_t* feature = <SIZE_t*> (<np.ndarray> d["feature"]).data
 *         cdef double* threshold = <double*> (<np.ndarray> d["threshold"]).data
 */
  __pyx_t_1 = PyObject_GetItem(__pyx_v_d, ((PyObject *)__pyx_n_s__children_right)); if (!__pyx_t_1) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1366; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_children_right = ((__pyx_t_7sklearn_4tree_5_tree_SIZE_t *)((PyArrayObject *)__pyx_t_1)->data);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "sklearn/tree/_tree.pyx":1367
 *         cdef SIZE_t* children_left = <SIZE_t*> (<np.ndarray> d["children_left"]).data
 *         cdef SIZE_t* children_right =  <SIZE_t*> (<np.ndarray> d["children_right"]).data
 *         cdef SIZE_t* feature = <SIZE_t*> (<np.ndarray> d["feature"]).data             # <<<<<<<<<<<<<<
 *         cdef double* threshold = <double*> (<np.ndarray> d["threshold"]).data
 *         cdef double* value = <double*> (<np.ndarray> d["value"]).data
 */
  __pyx_t_1 = PyObject_GetItem(__pyx_v_d, ((PyObject *)__pyx_n_s__feature)); if (!__pyx_t_1) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1367; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_feature = ((__pyx_t_7sklearn_4tree_5_tree_SIZE_t *)((PyArrayObject *)__pyx_t_1)->data);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "sklearn/tree/_tree.pyx":1368
 *         cdef SIZE_t* children_right =  <SIZE_t*> (<np.ndarray> d["children_right"]).data
 *         cdef SIZE_t* feature = <SIZE_t*> (<np.ndarray> d["feature"]).data
 *         cdef double* threshold = <double*> (<np.ndarray> d["threshold"]).data             # <<<<<<<<<<<<<<
 *         cdef double* value = <double*> (<np.ndarray> d["value"]).data
 *         cdef double* impurity = <double*> (<np.ndarray> d["impurity"]).data
 */
  __pyx_t_1 = PyObject_GetItem(__pyx_v_d, ((PyObject *)__pyx_n_s__threshold)); if (!__pyx_t_1) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1368; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_threshold = ((double *)((PyArrayObject *)__pyx_t_1)->data);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "sklearn/tree/_tree.pyx":1369
 *         cdef SIZE_t* feature = <SIZE_t*> (<np.ndarray> d["feature"]).data
 *         cdef double* threshold = <double*> (<np.ndarray> d["threshold"]).data
 *         cdef double* value = <double*> (<np.ndarray> d["value"]).data             # <<<<<<<<<<<<<<
 *         cdef double* impurity = <double*> (<np.ndarray> d["impurity"]).data
 *         cdef SIZE_t* n_node_samples = <SIZE_t*> (<np.ndarray> d["n_node_samples"]).data
 */
  __pyx_t_1 = PyObject_GetItem(__pyx_v_d, ((PyObject *)__pyx_n_s__value)); if (!__pyx_t_1) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1369; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_value = ((double *)((PyArrayObject *)__pyx_t_1)->data);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "sklearn/tree/_tree.pyx":1370
 *         cdef double* threshold = <double*> (<np.ndarray> d["threshold"]).data
 *         cdef double* value = <double*> (<np.ndarray> d["value"]).data
 *         cdef double* impurity = <double*> (<np.ndarray> d["impurity"]).data             # <<<<<<<<<<<<<<
 *         cdef SIZE_t* n_node_samples = <SIZE_t*> (<np.ndarray> d["n_node_samples"]).data
 * 
 */
  __pyx_t_1 = PyObject_GetItem(__pyx_v_d, ((PyObject *)__pyx_n_s__impurity)); if (!__pyx_t_1) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1370; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_impurity = ((double *)((PyArrayObject *)__pyx_t_1)->data);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "sklearn/tree/_tree.pyx":1371
 *         cdef double* value = <double*> (<np.ndarray> d["value"]).data
 *         cdef double* impurity = <double*> (<np.ndarray> d["impurity"]).data
 *         cdef SIZE_t* n_node_samples = <SIZE_t*> (<np.ndarray> d["n_node_samples"]).data             # <<<<<<<<<<<<<<
 * 
 *         memcpy(self.children_left, children_left, self.capacity * sizeof(SIZE_t))
 */
  __pyx_t_1 = PyObject_GetItem(__pyx_v_d, ((PyObject *)__pyx_n_s__n_node_samples)); if (!__pyx_t_1) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1371; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_n_node_samples = ((__pyx_t_7sklearn_4tree_5_tree_SIZE_t *)((PyArrayObject *)__pyx_t_1)->data);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "sklearn/tree/_tree.pyx":1373
 *         cdef SIZE_t* n_node_samples = <SIZE_t*> (<np.ndarray> d["n_node_samples"]).data
 * 
 *         memcpy(self.children_left, children_left, self.capacity * sizeof(SIZE_t))             # <<<<<<<<<<<<<<
 *         memcpy(self.children_right, children_right, self.capacity * sizeof(SIZE_t))
 *         memcpy(self.feature, feature, self.capacity * sizeof(SIZE_t))
 */
  memcpy(__pyx_v_self->children_left, __pyx_v_children_left, (__pyx_v_self->capacity * (sizeof(__pyx_t_7sklearn_4tree_5_tree_SIZE_t))));

  /* "sklearn/tree/_tree.pyx":1374
 * 
 *         memcpy(self.children_left, children_left, self.capacity * sizeof(SIZE_t))
 *         memcpy(self.children_right, children_right, self.capacity * sizeof(SIZE_t))             # <<<<<<<<<<<<<<
 *         memcpy(self.feature, feature, self.capacity * sizeof(SIZE_t))
 *         memcpy(self.threshold, threshold, self.capacity * sizeof(double))
 */
  memcpy(__pyx_v_self->children_right, __pyx_v_children_right, (__pyx_v_self->capacity * (sizeof(__pyx_t_7sklearn_4tree_5_tree_SIZE_t))));

  /* "sklearn/tree/_tree.pyx":1375
 *         memcpy(self.children_left, children_left, self.capacity * sizeof(SIZE_t))
 *         memcpy(self.children_right, children_right, self.capacity * sizeof(SIZE_t))
 *         memcpy(self.feature, feature, self.capacity * sizeof(SIZE_t))             # <<<<<<<<<<<<<<
 *         memcpy(self.threshold, threshold, self.capacity * sizeof(double))
 *         memcpy(self.value, value, self.capacity * self.value_stride * sizeof(double))
 */
  memcpy(__pyx_v_self->feature, __pyx_v_feature, (__pyx_v_self->capacity * (sizeof(__pyx_t_7sklearn_4tree_5_tree_SIZE_t))));

  /* "sklearn/tree/_tree.pyx":1376
 *         memcpy(self.children_right, children_right, self.capacity * sizeof(SIZE_t))
 *         memcpy(self.feature, feature, self.capacity * sizeof(SIZE_t))
 *         memcpy(self.threshold, threshold, self.capacity * sizeof(double))             # <<<<<<<<<<<<<<
 *         memcpy(self.value, value, self.capacity * self.value_stride * sizeof(double))
 *         memcpy(self.impurity, impurity, self.capacity * sizeof(double))
 */
  memcpy(__pyx_v_self->threshold, __pyx_v_threshold, (__pyx_v_self->capacity * (sizeof(double))));

  /* "sklearn/tree/_tree.pyx":1377
 *         memcpy(self.feature, feature, self.capacity * sizeof(SIZE_t))
 *         memcpy(self.threshold, threshold, self.capacity * sizeof(double))
 *         memcpy(self.value, value, self.capacity * self.value_stride * sizeof(double))             # <<<<<<<<<<<<<<
 *         memcpy(self.impurity, impurity, self.capacity * sizeof(double))
 *         memcpy(self.n_node_samples, n_node_samples, self.capacity * sizeof(SIZE_t))
 */
  memcpy(__pyx_v_self->value, __pyx_v_value, ((__pyx_v_self->capacity * __pyx_v_self->value_stride) * (sizeof(double))));

  /* "sklearn/tree/_tree.pyx":1378
 *         memcpy(self.threshold, threshold, self.capacity * sizeof(double))
 *         memcpy(self.value, value, self.capacity * self.value_stride * sizeof(double))
 *         memcpy(self.impurity, impurity, self.capacity * sizeof(double))             # <<<<<<<<<<<<<<
 *         memcpy(self.n_node_samples, n_node_samples, self.capacity * sizeof(SIZE_t))
 * 
 */
  memcpy(__pyx_v_self->impurity, __pyx_v_impurity, (__pyx_v_self->capacity * (sizeof(double))));

  /* "sklearn/tree/_tree.pyx":1379
 *         memcpy(self.value, value, self.capacity * self.value_stride * sizeof(double))
 *         memcpy(self.impurity, impurity, self.capacity * sizeof(double))
 *         memcpy(self.n_node_samples, n_node_samples, self.capacity * sizeof(SIZE_t))             # <<<<<<<<<<<<<<
 * 
 *     cdef void _resize(self, int capacity=-1):
 */
  memcpy(__pyx_v_self->n_node_samples, __pyx_v_n_node_samples, (__pyx_v_self->capacity * (sizeof(__pyx_t_7sklearn_4tree_5_tree_SIZE_t))));

  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("sklearn.tree._tree.Tree.__setstate__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "sklearn/tree/_tree.pyx":1381
 *         memcpy(self.n_node_samples, n_node_samples, self.capacity * sizeof(SIZE_t))
 * 
 *     cdef void _resize(self, int capacity=-1):             # <<<<<<<<<<<<<<
 *         """Resize all inner arrays to `capacity`, if `capacity` < 0, then
 *            double the size of the inner arrays."""
 */

static void __pyx_f_7sklearn_4tree_5_tree_4Tree__resize(struct __pyx_obj_7sklearn_4tree_5_tree_Tree *__pyx_v_self, struct __pyx_opt_args_7sklearn_4tree_5_tree_4Tree__resize *__pyx_optional_args) {
  int __pyx_v_capacity = ((int)-1);
  __pyx_t_7sklearn_4tree_5_tree_SIZE_t *__pyx_v_tmp_children_left;
  __pyx_t_7sklearn_4tree_5_tree_SIZE_t *__pyx_v_tmp_children_right;
  __pyx_t_7sklearn_4tree_5_tree_SIZE_t *__pyx_v_tmp_feature;
  double *__pyx_v_tmp_threshold;
  double *__pyx_v_tmp_value;
  double *__pyx_v_tmp_impurity;
  __pyx_t_7sklearn_4tree_5_tree_SIZE_t *__pyx_v_tmp_n_node_samples;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  int __pyx_t_3;
  int __pyx_t_4;
  int __pyx_t_5;
  int __pyx_t_6;
  int __pyx_t_7;
  int __pyx_t_8;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_resize", 0);
  if (__pyx_optional_args) {
    if (__pyx_optional_args->__pyx_n > 0) {
      __pyx_v_capacity = __pyx_optional_args->capacity;
    }
  }

  /* "sklearn/tree/_tree.pyx":1384
 *         """Resize all inner arrays to `capacity`, if `capacity` < 0, then
 *            double the size of the inner arrays."""
 *         if capacity == self.capacity:             # <<<<<<<<<<<<<<
 *             return
 * 
 */
  __pyx_t_1 = ((__pyx_v_capacity == __pyx_v_self->capacity) != 0);
  if (__pyx_t_1) {

    /* "sklearn/tree/_tree.pyx":1385
 *            double the size of the inner arrays."""
 *         if capacity == self.capacity:
 *             return             # <<<<<<<<<<<<<<
 * 
 *         if capacity < 0:
 */
    goto __pyx_L0;
    goto __pyx_L3;
  }
  __pyx_L3:;

  /* "sklearn/tree/_tree.pyx":1387
 *             return
 * 
 *         if capacity < 0:             # <<<<<<<<<<<<<<
 *             if self.capacity <= 0:
 *                 capacity = 3 # default initial value
 */
  __pyx_t_1 = ((__pyx_v_capacity < 0) != 0);
  if (__pyx_t_1) {

    /* "sklearn/tree/_tree.pyx":1388
 * 
 *         if capacity < 0:
 *             if self.capacity <= 0:             # <<<<<<<<<<<<<<
 *                 capacity = 3 # default initial value
 *             else:
 */
    __pyx_t_1 = ((__pyx_v_self->capacity <= 0) != 0);
    if (__pyx_t_1) {

      /* "sklearn/tree/_tree.pyx":1389
 *         if capacity < 0:
 *             if self.capacity <= 0:
 *                 capacity = 3 # default initial value             # <<<<<<<<<<<<<<
 *             else:
 *                 capacity = 2 * self.capacity
 */
      __pyx_v_capacity = 3;
      goto __pyx_L5;
    }
    /*else*/ {

      /* "sklearn/tree/_tree.pyx":1391
 *                 capacity = 3 # default initial value
 *             else:
 *                 capacity = 2 * self.capacity             # <<<<<<<<<<<<<<
 * 
 *         self.capacity = capacity
 */
      __pyx_v_capacity = (2 * __pyx_v_self->capacity);
    }
    __pyx_L5:;
    goto __pyx_L4;
  }
  __pyx_L4:;

  /* "sklearn/tree/_tree.pyx":1393
 *                 capacity = 2 * self.capacity
 * 
 *         self.capacity = capacity             # <<<<<<<<<<<<<<
 * 
 *         cdef SIZE_t* tmp_children_left = <SIZE_t*> realloc(self.children_left, capacity * sizeof(SIZE_t))
 */
  __pyx_v_self->capacity = __pyx_v_capacity;

  /* "sklearn/tree/_tree.pyx":1395
 *         self.capacity = capacity
 * 
 *         cdef SIZE_t* tmp_children_left = <SIZE_t*> realloc(self.children_left, capacity * sizeof(SIZE_t))             # <<<<<<<<<<<<<<
 *         if tmp_children_left != NULL:
 *             self.children_left = tmp_children_left
 */
  __pyx_v_tmp_children_left = ((__pyx_t_7sklearn_4tree_5_tree_SIZE_t *)realloc(__pyx_v_self->children_left, (__pyx_v_capacity * (sizeof(__pyx_t_7sklearn_4tree_5_tree_SIZE_t)))));

  /* "sklearn/tree/_tree.pyx":1396
 * 
 *         cdef SIZE_t* tmp_children_left = <SIZE_t*> realloc(self.children_left, capacity * sizeof(SIZE_t))
 *         if tmp_children_left != NULL:             # <<<<<<<<<<<<<<
 *             self.children_left = tmp_children_left
 * 
 */
  __pyx_t_1 = ((__pyx_v_tmp_children_left != NULL) != 0);
  if (__pyx_t_1) {

    /* "sklearn/tree/_tree.pyx":1397
 *         cdef SIZE_t* tmp_children_left = <SIZE_t*> realloc(self.children_left, capacity * sizeof(SIZE_t))
 *         if tmp_children_left != NULL:
 *             self.children_left = tmp_children_left             # <<<<<<<<<<<<<<
 * 
 *         cdef SIZE_t* tmp_children_right = <SIZE_t*> realloc(self.children_right, capacity * sizeof(SIZE_t))
 */
    __pyx_v_self->children_left = __pyx_v_tmp_children_left;
    goto __pyx_L6;
  }
  __pyx_L6:;

  /* "sklearn/tree/_tree.pyx":1399
 *             self.children_left = tmp_children_left
 * 
 *         cdef SIZE_t* tmp_children_right = <SIZE_t*> realloc(self.children_right, capacity * sizeof(SIZE_t))             # <<<<<<<<<<<<<<
 *         if tmp_children_right != NULL:
 *             self.children_right = tmp_children_right
 */
  __pyx_v_tmp_children_right = ((__pyx_t_7sklearn_4tree_5_tree_SIZE_t *)realloc(__pyx_v_self->children_right, (__pyx_v_capacity * (sizeof(__pyx_t_7sklearn_4tree_5_tree_SIZE_t)))));

  /* "sklearn/tree/_tree.pyx":1400
 * 
 *         cdef SIZE_t* tmp_children_right = <SIZE_t*> realloc(self.children_right, capacity * sizeof(SIZE_t))
 *         if tmp_children_right != NULL:             # <<<<<<<<<<<<<<
 *             self.children_right = tmp_children_right
 * 
 */
  __pyx_t_1 = ((__pyx_v_tmp_children_right != NULL) != 0);
  if (__pyx_t_1) {

    /* "sklearn/tree/_tree.pyx":1401
 *         cdef SIZE_t* tmp_children_right = <SIZE_t*> realloc(self.children_right, capacity * sizeof(SIZE_t))
 *         if tmp_children_right != NULL:
 *             self.children_right = tmp_children_right             # <<<<<<<<<<<<<<
 * 
 *         cdef SIZE_t* tmp_feature = <SIZE_t*> realloc(self.feature, capacity * sizeof(SIZE_t))
 */
    __pyx_v_self->children_right = __pyx_v_tmp_children_right;
    goto __pyx_L7;
  }
  __pyx_L7:;

  /* "sklearn/tree/_tree.pyx":1403
 *             self.children_right = tmp_children_right
 * 
 *         cdef SIZE_t* tmp_feature = <SIZE_t*> realloc(self.feature, capacity * sizeof(SIZE_t))             # <<<<<<<<<<<<<<
 *         if tmp_feature != NULL:
 *             self.feature = tmp_feature
 */
  __pyx_v_tmp_feature = ((__pyx_t_7sklearn_4tree_5_tree_SIZE_t *)realloc(__pyx_v_self->feature, (__pyx_v_capacity * (sizeof(__pyx_t_7sklearn_4tree_5_tree_SIZE_t)))));

  /* "sklearn/tree/_tree.pyx":1404
 * 
 *         cdef SIZE_t* tmp_feature = <SIZE_t*> realloc(self.feature, capacity * sizeof(SIZE_t))
 *         if tmp_feature != NULL:             # <<<<<<<<<<<<<<
 *             self.feature = tmp_feature
 * 
 */
  __pyx_t_1 = ((__pyx_v_tmp_feature != NULL) != 0);
  if (__pyx_t_1) {

    /* "sklearn/tree/_tree.pyx":1405
 *         cdef SIZE_t* tmp_feature = <SIZE_t*> realloc(self.feature, capacity * sizeof(SIZE_t))
 *         if tmp_feature != NULL:
 *             self.feature = tmp_feature             # <<<<<<<<<<<<<<
 * 
 *         cdef double* tmp_threshold = <double*> realloc(self.threshold, capacity * sizeof(double))
 */
    __pyx_v_self->feature = __pyx_v_tmp_feature;
    goto __pyx_L8;
  }
  __pyx_L8:;

  /* "sklearn/tree/_tree.pyx":1407
 *             self.feature = tmp_feature
 * 
 *         cdef double* tmp_threshold = <double*> realloc(self.threshold, capacity * sizeof(double))             # <<<<<<<<<<<<<<
 *         if tmp_threshold != NULL:
 *             self.threshold = tmp_threshold
 */
  __pyx_v_tmp_threshold = ((double *)realloc(__pyx_v_self->threshold, (__pyx_v_capacity * (sizeof(double)))));

  /* "sklearn/tree/_tree.pyx":1408
 * 
 *         cdef double* tmp_threshold = <double*> realloc(self.threshold, capacity * sizeof(double))
 *         if tmp_threshold != NULL:             # <<<<<<<<<<<<<<
 *             self.threshold = tmp_threshold
 * 
 */
  __pyx_t_1 = ((__pyx_v_tmp_threshold != NULL) != 0);
  if (__pyx_t_1) {

    /* "sklearn/tree/_tree.pyx":1409
 *         cdef double* tmp_threshold = <double*> realloc(self.threshold, capacity * sizeof(double))
 *         if tmp_threshold != NULL:
 *             self.threshold = tmp_threshold             # <<<<<<<<<<<<<<
 * 
 *         cdef double* tmp_value = <double*> realloc(self.value, capacity * self.value_stride * sizeof(double))
 */
    __pyx_v_self->threshold = __pyx_v_tmp_threshold;
    goto __pyx_L9;
  }
  __pyx_L9:;

  /* "sklearn/tree/_tree.pyx":1411
 *             self.threshold = tmp_threshold
 * 
 *         cdef double* tmp_value = <double*> realloc(self.value, capacity * self.value_stride * sizeof(double))             # <<<<<<<<<<<<<<
 *         if tmp_value != NULL:
 *             self.value = tmp_value
 */
  __pyx_v_tmp_value = ((double *)realloc(__pyx_v_self->value, ((__pyx_v_capacity * __pyx_v_self->value_stride) * (sizeof(double)))));

  /* "sklearn/tree/_tree.pyx":1412
 * 
 *         cdef double* tmp_value = <double*> realloc(self.value, capacity * self.value_stride * sizeof(double))
 *         if tmp_value != NULL:             # <<<<<<<<<<<<<<
 *             self.value = tmp_value
 * 
 */
  __pyx_t_1 = ((__pyx_v_tmp_value != NULL) != 0);
  if (__pyx_t_1) {

    /* "sklearn/tree/_tree.pyx":1413
 *         cdef double* tmp_value = <double*> realloc(self.value, capacity * self.value_stride * sizeof(double))
 *         if tmp_value != NULL:
 *             self.value = tmp_value             # <<<<<<<<<<<<<<
 * 
 *         cdef double* tmp_impurity = <double*> realloc(self.impurity, capacity * sizeof(double))
 */
    __pyx_v_self->value = __pyx_v_tmp_value;
    goto __pyx_L10;
  }
  __pyx_L10:;

  /* "sklearn/tree/_tree.pyx":1415
 *             self.value = tmp_value
 * 
 *         cdef double* tmp_impurity = <double*> realloc(self.impurity, capacity * sizeof(double))             # <<<<<<<<<<<<<<
 *         if tmp_impurity != NULL:
 *             self.impurity = tmp_impurity
 */
  __pyx_v_tmp_impurity = ((double *)realloc(__pyx_v_self->impurity, (__pyx_v_capacity * (sizeof(double)))));

  /* "sklearn/tree/_tree.pyx":1416
 * 
 *         cdef double* tmp_impurity = <double*> realloc(self.impurity, capacity * sizeof(double))
 *         if tmp_impurity != NULL:             # <<<<<<<<<<<<<<
 *             self.impurity = tmp_impurity
 * 
 */
  __pyx_t_1 = ((__pyx_v_tmp_impurity != NULL) != 0);
  if (__pyx_t_1) {

    /* "sklearn/tree/_tree.pyx":1417
 *         cdef double* tmp_impurity = <double*> realloc(self.impurity, capacity * sizeof(double))
 *         if tmp_impurity != NULL:
 *             self.impurity = tmp_impurity             # <<<<<<<<<<<<<<
 * 
 *         cdef SIZE_t* tmp_n_node_samples = <SIZE_t*> realloc(self.n_node_samples, capacity * sizeof(SIZE_t))
 */
    __pyx_v_self->impurity = __pyx_v_tmp_impurity;
    goto __pyx_L11;
  }
  __pyx_L11:;

  /* "sklearn/tree/_tree.pyx":1419
 *             self.impurity = tmp_impurity
 * 
 *         cdef SIZE_t* tmp_n_node_samples = <SIZE_t*> realloc(self.n_node_samples, capacity * sizeof(SIZE_t))             # <<<<<<<<<<<<<<
 *         if tmp_n_node_samples != NULL:
 *             self.n_node_samples = tmp_n_node_samples
 */
  __pyx_v_tmp_n_node_samples = ((__pyx_t_7sklearn_4tree_5_tree_SIZE_t *)realloc(__pyx_v_self->n_node_samples, (__pyx_v_capacity * (sizeof(__pyx_t_7sklearn_4tree_5_tree_SIZE_t)))));

  /* "sklearn/tree/_tree.pyx":1420
 * 
 *         cdef SIZE_t* tmp_n_node_samples = <SIZE_t*> realloc(self.n_node_samples, capacity * sizeof(SIZE_t))
 *         if tmp_n_node_samples != NULL:             # <<<<<<<<<<<<<<
 *             self.n_node_samples = tmp_n_node_samples
 * 
 */
  __pyx_t_1 = ((__pyx_v_tmp_n_node_samples != NULL) != 0);
  if (__pyx_t_1) {

    /* "sklearn/tree/_tree.pyx":1421
 *         cdef SIZE_t* tmp_n_node_samples = <SIZE_t*> realloc(self.n_node_samples, capacity * sizeof(SIZE_t))
 *         if tmp_n_node_samples != NULL:
 *             self.n_node_samples = tmp_n_node_samples             # <<<<<<<<<<<<<<
 * 
 *         if ((tmp_children_left == NULL) or
 */
    __pyx_v_self->n_node_samples = __pyx_v_tmp_n_node_samples;
    goto __pyx_L12;
  }
  __pyx_L12:;

  /* "sklearn/tree/_tree.pyx":1423
 *             self.n_node_samples = tmp_n_node_samples
 * 
 *         if ((tmp_children_left == NULL) or             # <<<<<<<<<<<<<<
 *             (tmp_children_right == NULL) or
 *             (tmp_feature == NULL) or
 */
  __pyx_t_1 = ((__pyx_v_tmp_children_left == NULL) != 0);
  if (!__pyx_t_1) {

    /* "sklearn/tree/_tree.pyx":1424
 * 
 *         if ((tmp_children_left == NULL) or
 *             (tmp_children_right == NULL) or             # <<<<<<<<<<<<<<
 *             (tmp_feature == NULL) or
 *             (tmp_threshold == NULL) or
 */
    __pyx_t_2 = ((__pyx_v_tmp_children_right == NULL) != 0);
    if (!__pyx_t_2) {

      /* "sklearn/tree/_tree.pyx":1425
 *         if ((tmp_children_left == NULL) or
 *             (tmp_children_right == NULL) or
 *             (tmp_feature == NULL) or             # <<<<<<<<<<<<<<
 *             (tmp_threshold == NULL) or
 *             (tmp_value == NULL) or
 */
      __pyx_t_3 = ((__pyx_v_tmp_feature == NULL) != 0);
      if (!__pyx_t_3) {

        /* "sklearn/tree/_tree.pyx":1426
 *             (tmp_children_right == NULL) or
 *             (tmp_feature == NULL) or
 *             (tmp_threshold == NULL) or             # <<<<<<<<<<<<<<
 *             (tmp_value == NULL) or
 *             (tmp_impurity == NULL) or
 */
        __pyx_t_4 = ((__pyx_v_tmp_threshold == NULL) != 0);
        if (!__pyx_t_4) {

          /* "sklearn/tree/_tree.pyx":1427
 *             (tmp_feature == NULL) or
 *             (tmp_threshold == NULL) or
 *             (tmp_value == NULL) or             # <<<<<<<<<<<<<<
 *             (tmp_impurity == NULL) or
 *             (tmp_n_node_samples == NULL)):
 */
          __pyx_t_5 = ((__pyx_v_tmp_value == NULL) != 0);
          if (!__pyx_t_5) {

            /* "sklearn/tree/_tree.pyx":1428
 *             (tmp_threshold == NULL) or
 *             (tmp_value == NULL) or
 *             (tmp_impurity == NULL) or             # <<<<<<<<<<<<<<
 *             (tmp_n_node_samples == NULL)):
 *             raise MemoryError()
 */
            __pyx_t_6 = ((__pyx_v_tmp_impurity == NULL) != 0);
            if (!__pyx_t_6) {

              /* "sklearn/tree/_tree.pyx":1429
 *             (tmp_value == NULL) or
 *             (tmp_impurity == NULL) or
 *             (tmp_n_node_samples == NULL)):             # <<<<<<<<<<<<<<
 *             raise MemoryError()
 * 
 */
              __pyx_t_7 = ((__pyx_v_tmp_n_node_samples == NULL) != 0);
              __pyx_t_8 = __pyx_t_7;
            } else {
              __pyx_t_8 = __pyx_t_6;
            }
            __pyx_t_6 = __pyx_t_8;
          } else {
            __pyx_t_6 = __pyx_t_5;
          }
          __pyx_t_5 = __pyx_t_6;
        } else {
          __pyx_t_5 = __pyx_t_4;
        }
        __pyx_t_4 = __pyx_t_5;
      } else {
        __pyx_t_4 = __pyx_t_3;
      }
      __pyx_t_3 = __pyx_t_4;
    } else {
      __pyx_t_3 = __pyx_t_2;
    }
    __pyx_t_2 = __pyx_t_3;
  } else {
    __pyx_t_2 = __pyx_t_1;
  }
  if (__pyx_t_2) {

    /* "sklearn/tree/_tree.pyx":1430
 *             (tmp_impurity == NULL) or
 *             (tmp_n_node_samples == NULL)):
 *             raise MemoryError()             # <<<<<<<<<<<<<<
 * 
 *         # if capacity smaller than node_count, adjust the counter
 */
    PyErr_NoMemory(); {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1430; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    goto __pyx_L13;
  }
  __pyx_L13:;

  /* "sklearn/tree/_tree.pyx":1433
 * 
 *         # if capacity smaller than node_count, adjust the counter
 *         if capacity < self.node_count:             # <<<<<<<<<<<<<<
 *             self.node_count = capacity
 * 
 */
  __pyx_t_2 = ((__pyx_v_capacity < __pyx_v_self->node_count) != 0);
  if (__pyx_t_2) {

    /* "sklearn/tree/_tree.pyx":1434
 *         # if capacity smaller than node_count, adjust the counter
 *         if capacity < self.node_count:
 *             self.node_count = capacity             # <<<<<<<<<<<<<<
 * 
 *     cdef SIZE_t _add_node(self, SIZE_t parent,
 */
    __pyx_v_self->node_count = __pyx_v_capacity;
    goto __pyx_L14;
  }
  __pyx_L14:;

  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_WriteUnraisable("sklearn.tree._tree.Tree._resize", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
}

/* "sklearn/tree/_tree.pyx":1436
 *             self.node_count = capacity
 * 
 *     cdef SIZE_t _add_node(self, SIZE_t parent,             # <<<<<<<<<<<<<<
 *                                 bint is_left,
 *                                 bint is_leaf,
 */

static __pyx_t_7sklearn_4tree_5_tree_SIZE_t __pyx_f_7sklearn_4tree_5_tree_4Tree__add_node(struct __pyx_obj_7sklearn_4tree_5_tree_Tree *__pyx_v_self, __pyx_t_7sklearn_4tree_5_tree_SIZE_t __pyx_v_parent, int __pyx_v_is_left, int __pyx_v_is_leaf, __pyx_t_7sklearn_4tree_5_tree_SIZE_t __pyx_v_feature, double __pyx_v_threshold, double __pyx_v_impurity, __pyx_t_7sklearn_4tree_5_tree_SIZE_t __pyx_v_n_node_samples) {
  __pyx_t_7sklearn_4tree_5_tree_SIZE_t __pyx_v_node_id;
  __pyx_t_7sklearn_4tree_5_tree_SIZE_t __pyx_r;
  __Pyx_RefNannyDeclarations
  __pyx_t_7sklearn_4tree_5_tree_SIZE_t __pyx_t_1;
  int __pyx_t_2;
  __Pyx_RefNannySetupContext("_add_node", 0);

  /* "sklearn/tree/_tree.pyx":1445
 *         """Add a node to the tree. The new node registers itself as
 *            the child of its parent. """
 *         cdef SIZE_t node_id = self.node_count             # <<<<<<<<<<<<<<
 * 
 *         if node_id >= self.capacity:
 */
  __pyx_t_1 = __pyx_v_self->node_count;
  __pyx_v_node_id = __pyx_t_1;

  /* "sklearn/tree/_tree.pyx":1447
 *         cdef SIZE_t node_id = self.node_count
 * 
 *         if node_id >= self.capacity:             # <<<<<<<<<<<<<<
 *             self._resize()
 * 
 */
  __pyx_t_2 = ((__pyx_v_node_id >= __pyx_v_self->capacity) != 0);
  if (__pyx_t_2) {

    /* "sklearn/tree/_tree.pyx":1448
 * 
 *         if node_id >= self.capacity:
 *             self._resize()             # <<<<<<<<<<<<<<
 * 
 *         self.impurity[node_id] = impurity
 */
    ((struct __pyx_vtabstruct_7sklearn_4tree_5_tree_Tree *)__pyx_v_self->__pyx_vtab)->_resize(__pyx_v_self, NULL);
    goto __pyx_L3;
  }
  __pyx_L3:;

  /* "sklearn/tree/_tree.pyx":1450
 *             self._resize()
 * 
 *         self.impurity[node_id] = impurity             # <<<<<<<<<<<<<<
 *         self.n_node_samples[node_id] = n_node_samples
 * 
 */
  (__pyx_v_self->impurity[__pyx_v_node_id]) = __pyx_v_impurity;

  /* "sklearn/tree/_tree.pyx":1451
 * 
 *         self.impurity[node_id] = impurity
 *         self.n_node_samples[node_id] = n_node_samples             # <<<<<<<<<<<<<<
 * 
 *         if parent != _TREE_UNDEFINED:
 */
  (__pyx_v_self->n_node_samples[__pyx_v_node_id]) = __pyx_v_n_node_samples;

  /* "sklearn/tree/_tree.pyx":1453
 *         self.n_node_samples[node_id] = n_node_samples
 * 
 *         if parent != _TREE_UNDEFINED:             # <<<<<<<<<<<<<<
 *             if is_left:
 *                 self.children_left[parent] = node_id
 */
  __pyx_t_2 = ((__pyx_v_parent != __pyx_v_7sklearn_4tree_5_tree__TREE_UNDEFINED) != 0);
  if (__pyx_t_2) {

    /* "sklearn/tree/_tree.pyx":1454
 * 
 *         if parent != _TREE_UNDEFINED:
 *             if is_left:             # <<<<<<<<<<<<<<
 *                 self.children_left[parent] = node_id
 *             else:
 */
    __pyx_t_2 = (__pyx_v_is_left != 0);
    if (__pyx_t_2) {

      /* "sklearn/tree/_tree.pyx":1455
 *         if parent != _TREE_UNDEFINED:
 *             if is_left:
 *                 self.children_left[parent] = node_id             # <<<<<<<<<<<<<<
 *             else:
 *                 self.children_right[parent] = node_id
 */
      (__pyx_v_self->children_left[__pyx_v_parent]) = __pyx_v_node_id;
      goto __pyx_L5;
    }
    /*else*/ {

      /* "sklearn/tree/_tree.pyx":1457
 *                 self.children_left[parent] = node_id
 *             else:
 *                 self.children_right[parent] = node_id             # <<<<<<<<<<<<<<
 * 
 *         if not is_leaf:
 */
      (__pyx_v_self->children_right[__pyx_v_parent]) = __pyx_v_node_id;
    }
    __pyx_L5:;
    goto __pyx_L4;
  }
  __pyx_L4:;

  /* "sklearn/tree/_tree.pyx":1459
 *                 self.children_right[parent] = node_id
 * 
 *         if not is_leaf:             # <<<<<<<<<<<<<<
 *             # children_left and children_right will be set later
 *             self.feature[node_id] = feature
 */
  __pyx_t_2 = ((!(__pyx_v_is_leaf != 0)) != 0);
  if (__pyx_t_2) {

    /* "sklearn/tree/_tree.pyx":1461
 *         if not is_leaf:
 *             # children_left and children_right will be set later
 *             self.feature[node_id] = feature             # <<<<<<<<<<<<<<
 *             self.threshold[node_id] = threshold
 * 
 */
    (__pyx_v_self->feature[__pyx_v_node_id]) = __pyx_v_feature;

    /* "sklearn/tree/_tree.pyx":1462
 *             # children_left and children_right will be set later
 *             self.feature[node_id] = feature
 *             self.threshold[node_id] = threshold             # <<<<<<<<<<<<<<
 * 
 *         else:
 */
    (__pyx_v_self->threshold[__pyx_v_node_id]) = __pyx_v_threshold;
    goto __pyx_L6;
  }
  /*else*/ {

    /* "sklearn/tree/_tree.pyx":1465
 * 
 *         else:
 *             self.children_left[node_id] = _TREE_LEAF             # <<<<<<<<<<<<<<
 *             self.children_right[node_id] = _TREE_LEAF
 *             self.feature[node_id] = _TREE_UNDEFINED
 */
    (__pyx_v_self->children_left[__pyx_v_node_id]) = __pyx_v_7sklearn_4tree_5_tree__TREE_LEAF;

    /* "sklearn/tree/_tree.pyx":1466
 *         else:
 *             self.children_left[node_id] = _TREE_LEAF
 *             self.children_right[node_id] = _TREE_LEAF             # <<<<<<<<<<<<<<
 *             self.feature[node_id] = _TREE_UNDEFINED
 *             self.threshold[node_id] = _TREE_UNDEFINED
 */
    (__pyx_v_self->children_right[__pyx_v_node_id]) = __pyx_v_7sklearn_4tree_5_tree__TREE_LEAF;

    /* "sklearn/tree/_tree.pyx":1467
 *             self.children_left[node_id] = _TREE_LEAF
 *             self.children_right[node_id] = _TREE_LEAF
 *             self.feature[node_id] = _TREE_UNDEFINED             # <<<<<<<<<<<<<<
 *             self.threshold[node_id] = _TREE_UNDEFINED
 * 
 */
    (__pyx_v_self->feature[__pyx_v_node_id]) = __pyx_v_7sklearn_4tree_5_tree__TREE_UNDEFINED;

    /* "sklearn/tree/_tree.pyx":1468
 *             self.children_right[node_id] = _TREE_LEAF
 *             self.feature[node_id] = _TREE_UNDEFINED
 *             self.threshold[node_id] = _TREE_UNDEFINED             # <<<<<<<<<<<<<<
 * 
 *         self.node_count += 1
 */
    (__pyx_v_self->threshold[__pyx_v_node_id]) = __pyx_v_7sklearn_4tree_5_tree__TREE_UNDEFINED;
  }
  __pyx_L6:;

  /* "sklearn/tree/_tree.pyx":1470
 *             self.threshold[node_id] = _TREE_UNDEFINED
 * 
 *         self.node_count += 1             # <<<<<<<<<<<<<<
 * 
 *         return node_id
 */
  __pyx_v_self->node_count = (__pyx_v_self->node_count + 1);

  /* "sklearn/tree/_tree.pyx":1472
 *         self.node_count += 1
 * 
 *         return node_id             # <<<<<<<<<<<<<<
 * 
 *     cpdef build(self, np.ndarray X,
 */
  __pyx_r = __pyx_v_node_id;
  goto __pyx_L0;

  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "sklearn/tree/_tree.pyx":1474
 *         return node_id
 * 
 *     cpdef build(self, np.ndarray X,             # <<<<<<<<<<<<<<
 *                       np.ndarray y,
 *                       np.ndarray sample_weight=None):
 */

static PyObject *__pyx_pw_7sklearn_4tree_5_tree_4Tree_11build(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyObject *__pyx_f_7sklearn_4tree_5_tree_4Tree_build(struct __pyx_obj_7sklearn_4tree_5_tree_Tree *__pyx_v_self, PyArrayObject *__pyx_v_X, PyArrayObject *__pyx_v_y, int __pyx_skip_dispatch, struct __pyx_opt_args_7sklearn_4tree_5_tree_4Tree_build *__pyx_optional_args) {

  /* "sklearn/tree/_tree.pyx":1476
 *     cpdef build(self, np.ndarray X,
 *                       np.ndarray y,
 *                       np.ndarray sample_weight=None):             # <<<<<<<<<<<<<<
 *         """Build a decision tree from the training set (X, y)."""
 *         # Prepare data before recursive partitioning
 */
  PyArrayObject *__pyx_v_sample_weight = ((PyArrayObject *)Py_None);
  __pyx_t_7sklearn_4tree_5_tree_DOUBLE_t *__pyx_v_sample_weight_ptr;
  int __pyx_v_init_capacity;
  struct __pyx_obj_7sklearn_4tree_5_tree_Splitter *__pyx_v_splitter = 0;
  __pyx_t_7sklearn_4tree_5_tree_SIZE_t __pyx_v_stack_n_values;
  __pyx_t_7sklearn_4tree_5_tree_SIZE_t __pyx_v_stack_capacity;
  __pyx_t_7sklearn_4tree_5_tree_SIZE_t *__pyx_v_stack;
  __pyx_t_7sklearn_4tree_5_tree_SIZE_t __pyx_v_start;
  __pyx_t_7sklearn_4tree_5_tree_SIZE_t __pyx_v_end;
  __pyx_t_7sklearn_4tree_5_tree_SIZE_t __pyx_v_depth;
  __pyx_t_7sklearn_4tree_5_tree_SIZE_t __pyx_v_parent;
  int __pyx_v_is_left;
  __pyx_t_7sklearn_4tree_5_tree_SIZE_t __pyx_v_n_node_samples;
  __pyx_t_7sklearn_4tree_5_tree_SIZE_t __pyx_v_pos;
  __pyx_t_7sklearn_4tree_5_tree_SIZE_t __pyx_v_feature;
  double __pyx_v_threshold;
  double __pyx_v_impurity;
  int __pyx_v_is_leaf;
  __pyx_t_7sklearn_4tree_5_tree_SIZE_t __pyx_v_node_id;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_t_4;
  int __pyx_t_5;
  int __pyx_t_6;
  PyObject *__pyx_t_7 = NULL;
  struct __pyx_opt_args_7sklearn_4tree_5_tree_4Tree__resize __pyx_t_8;
  __pyx_t_7sklearn_4tree_5_tree_SIZE_t __pyx_t_9;
  int __pyx_t_10;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("build", 0);
  if (__pyx_optional_args) {
    if (__pyx_optional_args->__pyx_n > 0) {
      __pyx_v_sample_weight = __pyx_optional_args->sample_weight;
    }
  }
  __Pyx_INCREF((PyObject *)__pyx_v_X);
  __Pyx_INCREF((PyObject *)__pyx_v_y);
  __Pyx_INCREF((PyObject *)__pyx_v_sample_weight);

  /* "sklearn/tree/_tree.pyx":1474
 *         return node_id
 * 
 *     cpdef build(self, np.ndarray X,             # <<<<<<<<<<<<<<
 *                       np.ndarray y,
 *                       np.ndarray sample_weight=None):
 */
  /* Check if called by wrapper */
  if (unlikely(__pyx_skip_dispatch)) ;
  /* Check if overridden in Python */
  else if (unlikely(Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0)) {
    __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s__build); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1474; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    __Pyx_GOTREF(__pyx_t_1);
    if (!PyCFunction_Check(__pyx_t_1) || (PyCFunction_GET_FUNCTION(__pyx_t_1) != (PyCFunction)__pyx_pw_7sklearn_4tree_5_tree_4Tree_11build)) {
      __Pyx_XDECREF(__pyx_r);
      __pyx_t_2 = PyTuple_New(3); if (unlikely(!__pyx_t_2)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1474; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_INCREF(((PyObject *)__pyx_v_X));
      PyTuple_SET_ITEM(__pyx_t_2, 0, ((PyObject *)__pyx_v_X));
      __Pyx_GIVEREF(((PyObject *)__pyx_v_X));
      __Pyx_INCREF(((PyObject *)__pyx_v_y));
      PyTuple_SET_ITEM(__pyx_t_2, 1, ((PyObject *)__pyx_v_y));
      __Pyx_GIVEREF(((PyObject *)__pyx_v_y));
      __Pyx_INCREF(((PyObject *)__pyx_v_sample_weight));
      PyTuple_SET_ITEM(__pyx_t_2, 2, ((PyObject *)__pyx_v_sample_weight));
      __Pyx_GIVEREF(((PyObject *)__pyx_v_sample_weight));
      __pyx_t_3 = PyObject_Call(__pyx_t_1, ((PyObject *)__pyx_t_2), NULL); if (unlikely(!__pyx_t_3)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1474; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_DECREF(((PyObject *)__pyx_t_2)); __pyx_t_2 = 0;
      __pyx_r = __pyx_t_3;
      __pyx_t_3 = 0;
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      goto __pyx_L0;
    }
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  }

  /* "sklearn/tree/_tree.pyx":1479
 *         """Build a decision tree from the training set (X, y)."""
 *         # Prepare data before recursive partitioning
 *         if X.dtype != DTYPE or not X.flags.contiguous:             # <<<<<<<<<<<<<<
 *             X = np.asarray(X, dtype=DTYPE, order="C")
 * 
 */
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_X), __pyx_n_s__dtype); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1479; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_3 = __Pyx_GetModuleGlobalName(__pyx_n_s__DTYPE); if (unlikely(!__pyx_t_3)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1479; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_2 = PyObject_RichCompare(__pyx_t_1, __pyx_t_3, Py_NE); __Pyx_XGOTREF(__pyx_t_2); if (unlikely(!__pyx_t_2)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1479; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_4 = __Pyx_PyObject_IsTrue(__pyx_t_2); if (unlikely(__pyx_t_4 < 0)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1479; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (!__pyx_t_4) {
    __pyx_t_2 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_X), __pyx_n_s__flags); if (unlikely(!__pyx_t_2)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1479; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s__contiguous); if (unlikely(!__pyx_t_3)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1479; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_t_5 = __Pyx_PyObject_IsTrue(__pyx_t_3); if (unlikely(__pyx_t_5 < 0)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1479; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_t_6 = (!__pyx_t_5);
    __pyx_t_5 = __pyx_t_6;
  } else {
    __pyx_t_5 = __pyx_t_4;
  }
  if (__pyx_t_5) {

    /* "sklearn/tree/_tree.pyx":1480
 *         # Prepare data before recursive partitioning
 *         if X.dtype != DTYPE or not X.flags.contiguous:
 *             X = np.asarray(X, dtype=DTYPE, order="C")             # <<<<<<<<<<<<<<
 * 
 *         if y.dtype != DOUBLE or not y.flags.contiguous:
 */
    __pyx_t_3 = __Pyx_GetModuleGlobalName(__pyx_n_s__np); if (unlikely(!__pyx_t_3)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1480; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s__asarray); if (unlikely(!__pyx_t_2)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1480; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_t_3 = PyTuple_New(1); if (unlikely(!__pyx_t_3)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1480; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_INCREF(((PyObject *)__pyx_v_X));
    PyTuple_SET_ITEM(__pyx_t_3, 0, ((PyObject *)__pyx_v_X));
    __Pyx_GIVEREF(((PyObject *)__pyx_v_X));
    __pyx_t_1 = PyDict_New(); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1480; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    __Pyx_GOTREF(((PyObject *)__pyx_t_1));
    __pyx_t_7 = __Pyx_GetModuleGlobalName(__pyx_n_s__DTYPE); if (unlikely(!__pyx_t_7)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1480; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    __Pyx_GOTREF(__pyx_t_7);
    if (PyDict_SetItem(__pyx_t_1, ((PyObject *)__pyx_n_s__dtype), __pyx_t_7) < 0) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1480; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    if (PyDict_SetItem(__pyx_t_1, ((PyObject *)__pyx_n_s__order), ((PyObject *)__pyx_n_s__C)) < 0) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1480; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    __pyx_t_7 = PyObject_Call(__pyx_t_2, ((PyObject *)__pyx_t_3), ((PyObject *)__pyx_t_1)); if (unlikely(!__pyx_t_7)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1480; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    __Pyx_GOTREF(__pyx_t_7);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF(((PyObject *)__pyx_t_3)); __pyx_t_3 = 0;
    __Pyx_DECREF(((PyObject *)__pyx_t_1)); __pyx_t_1 = 0;
    if (!(likely(((__pyx_t_7) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_7, __pyx_ptype_5numpy_ndarray))))) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1480; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    __Pyx_DECREF(((PyObject *)__pyx_v_X));
    __pyx_v_X = ((PyArrayObject *)__pyx_t_7);
    __pyx_t_7 = 0;
    goto __pyx_L3;
  }
  __pyx_L3:;

  /* "sklearn/tree/_tree.pyx":1482
 *             X = np.asarray(X, dtype=DTYPE, order="C")
 * 
 *         if y.dtype != DOUBLE or not y.flags.contiguous:             # <<<<<<<<<<<<<<
 *             y = np.asarray(y, dtype=DOUBLE, order="C")
 * 
 */
  __pyx_t_7 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_y), __pyx_n_s__dtype); if (unlikely(!__pyx_t_7)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1482; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_7);
  __pyx_t_1 = __Pyx_GetModuleGlobalName(__pyx_n_s__DOUBLE); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1482; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_3 = PyObject_RichCompare(__pyx_t_7, __pyx_t_1, Py_NE); __Pyx_XGOTREF(__pyx_t_3); if (unlikely(!__pyx_t_3)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1482; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_5 = __Pyx_PyObject_IsTrue(__pyx_t_3); if (unlikely(__pyx_t_5 < 0)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1482; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (!__pyx_t_5) {
    __pyx_t_3 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_y), __pyx_n_s__flags); if (unlikely(!__pyx_t_3)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1482; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s__contiguous); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1482; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_t_4 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely(__pyx_t_4 < 0)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1482; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_6 = (!__pyx_t_4);
    __pyx_t_4 = __pyx_t_6;
  } else {
    __pyx_t_4 = __pyx_t_5;
  }
  if (__pyx_t_4) {

    /* "sklearn/tree/_tree.pyx":1483
 * 
 *         if y.dtype != DOUBLE or not y.flags.contiguous:
 *             y = np.asarray(y, dtype=DOUBLE, order="C")             # <<<<<<<<<<<<<<
 * 
 *         cdef DOUBLE_t* sample_weight_ptr = NULL
 */
    __pyx_t_1 = __Pyx_GetModuleGlobalName(__pyx_n_s__np); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1483; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s__asarray); if (unlikely(!__pyx_t_3)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1483; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_1 = PyTuple_New(1); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1483; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(((PyObject *)__pyx_v_y));
    PyTuple_SET_ITEM(__pyx_t_1, 0, ((PyObject *)__pyx_v_y));
    __Pyx_GIVEREF(((PyObject *)__pyx_v_y));
    __pyx_t_7 = PyDict_New(); if (unlikely(!__pyx_t_7)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1483; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    __Pyx_GOTREF(((PyObject *)__pyx_t_7));
    __pyx_t_2 = __Pyx_GetModuleGlobalName(__pyx_n_s__DOUBLE); if (unlikely(!__pyx_t_2)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1483; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    __Pyx_GOTREF(__pyx_t_2);
    if (PyDict_SetItem(__pyx_t_7, ((PyObject *)__pyx_n_s__dtype), __pyx_t_2) < 0) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1483; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    if (PyDict_SetItem(__pyx_t_7, ((PyObject *)__pyx_n_s__order), ((PyObject *)__pyx_n_s__C)) < 0) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1483; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    __pyx_t_2 = PyObject_Call(__pyx_t_3, ((PyObject *)__pyx_t_1), ((PyObject *)__pyx_t_7)); if (unlikely(!__pyx_t_2)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1483; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_DECREF(((PyObject *)__pyx_t_1)); __pyx_t_1 = 0;
    __Pyx_DECREF(((PyObject *)__pyx_t_7)); __pyx_t_7 = 0;
    if (!(likely(((__pyx_t_2) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_2, __pyx_ptype_5numpy_ndarray))))) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1483; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    __Pyx_DECREF(((PyObject *)__pyx_v_y));
    __pyx_v_y = ((PyArrayObject *)__pyx_t_2);
    __pyx_t_2 = 0;
    goto __pyx_L4;
  }
  __pyx_L4:;

  /* "sklearn/tree/_tree.pyx":1485
 *             y = np.asarray(y, dtype=DOUBLE, order="C")
 * 
 *         cdef DOUBLE_t* sample_weight_ptr = NULL             # <<<<<<<<<<<<<<
 *         if sample_weight is not None:
 *             if ((sample_weight.dtype != DOUBLE) or
 */
  __pyx_v_sample_weight_ptr = NULL;

  /* "sklearn/tree/_tree.pyx":1486
 * 
 *         cdef DOUBLE_t* sample_weight_ptr = NULL
 *         if sample_weight is not None:             # <<<<<<<<<<<<<<
 *             if ((sample_weight.dtype != DOUBLE) or
 *                 (not sample_weight.flags.contiguous)):
 */
  __pyx_t_4 = (((PyObject *)__pyx_v_sample_weight) != Py_None);
  __pyx_t_5 = (__pyx_t_4 != 0);
  if (__pyx_t_5) {

    /* "sklearn/tree/_tree.pyx":1487
 *         cdef DOUBLE_t* sample_weight_ptr = NULL
 *         if sample_weight is not None:
 *             if ((sample_weight.dtype != DOUBLE) or             # <<<<<<<<<<<<<<
 *                 (not sample_weight.flags.contiguous)):
 *                 sample_weight = np.asarray(sample_weight,
 */
    __pyx_t_2 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_sample_weight), __pyx_n_s__dtype); if (unlikely(!__pyx_t_2)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1487; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_7 = __Pyx_GetModuleGlobalName(__pyx_n_s__DOUBLE); if (unlikely(!__pyx_t_7)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1487; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    __Pyx_GOTREF(__pyx_t_7);
    __pyx_t_1 = PyObject_RichCompare(__pyx_t_2, __pyx_t_7, Py_NE); __Pyx_XGOTREF(__pyx_t_1); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1487; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    __pyx_t_5 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely(__pyx_t_5 < 0)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1487; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    if (!__pyx_t_5) {

      /* "sklearn/tree/_tree.pyx":1488
 *         if sample_weight is not None:
 *             if ((sample_weight.dtype != DOUBLE) or
 *                 (not sample_weight.flags.contiguous)):             # <<<<<<<<<<<<<<
 *                 sample_weight = np.asarray(sample_weight,
 *                                            dtype=DOUBLE, order="C")
 */
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_sample_weight), __pyx_n_s__flags); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1488; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
      __Pyx_GOTREF(__pyx_t_1);
      __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s__contiguous); if (unlikely(!__pyx_t_7)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1488; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
      __Pyx_GOTREF(__pyx_t_7);
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      __pyx_t_4 = __Pyx_PyObject_IsTrue(__pyx_t_7); if (unlikely(__pyx_t_4 < 0)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1488; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
      __pyx_t_6 = (!__pyx_t_4);
      __pyx_t_4 = __pyx_t_6;
    } else {
      __pyx_t_4 = __pyx_t_5;
    }
    if (__pyx_t_4) {

      /* "sklearn/tree/_tree.pyx":1489
 *             if ((sample_weight.dtype != DOUBLE) or
 *                 (not sample_weight.flags.contiguous)):
 *                 sample_weight = np.asarray(sample_weight,             # <<<<<<<<<<<<<<
 *                                            dtype=DOUBLE, order="C")
 *             sample_weight_ptr = <DOUBLE_t*> sample_weight.data
 */
      __pyx_t_7 = __Pyx_GetModuleGlobalName(__pyx_n_s__np); if (unlikely(!__pyx_t_7)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1489; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
      __Pyx_GOTREF(__pyx_t_7);
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_t_7, __pyx_n_s__asarray); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1489; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
      __pyx_t_7 = PyTuple_New(1); if (unlikely(!__pyx_t_7)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1489; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
      __Pyx_GOTREF(__pyx_t_7);
      __Pyx_INCREF(((PyObject *)__pyx_v_sample_weight));
      PyTuple_SET_ITEM(__pyx_t_7, 0, ((PyObject *)__pyx_v_sample_weight));
      __Pyx_GIVEREF(((PyObject *)__pyx_v_sample_weight));
      __pyx_t_2 = PyDict_New(); if (unlikely(!__pyx_t_2)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1489; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
      __Pyx_GOTREF(((PyObject *)__pyx_t_2));

      /* "sklearn/tree/_tree.pyx":1490
 *                 (not sample_weight.flags.contiguous)):
 *                 sample_weight = np.asarray(sample_weight,
 *                                            dtype=DOUBLE, order="C")             # <<<<<<<<<<<<<<
 *             sample_weight_ptr = <DOUBLE_t*> sample_weight.data
 * 
 */
      __pyx_t_3 = __Pyx_GetModuleGlobalName(__pyx_n_s__DOUBLE); if (unlikely(!__pyx_t_3)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1490; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
      __Pyx_GOTREF(__pyx_t_3);
      if (PyDict_SetItem(__pyx_t_2, ((PyObject *)__pyx_n_s__dtype), __pyx_t_3) < 0) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1489; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      if (PyDict_SetItem(__pyx_t_2, ((PyObject *)__pyx_n_s__order), ((PyObject *)__pyx_n_s__C)) < 0) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1489; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
      __pyx_t_3 = PyObject_Call(__pyx_t_1, ((PyObject *)__pyx_t_7), ((PyObject *)__pyx_t_2)); if (unlikely(!__pyx_t_3)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1489; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      __Pyx_DECREF(((PyObject *)__pyx_t_7)); __pyx_t_7 = 0;
      __Pyx_DECREF(((PyObject *)__pyx_t_2)); __pyx_t_2 = 0;
      if (!(likely(((__pyx_t_3) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_3, __pyx_ptype_5numpy_ndarray))))) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1489; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
      __Pyx_DECREF(((PyObject *)__pyx_v_sample_weight));
      __pyx_v_sample_weight = ((PyArrayObject *)__pyx_t_3);
      __pyx_t_3 = 0;
      goto __pyx_L6;
    }
    __pyx_L6:;

    /* "sklearn/tree/_tree.pyx":1491
 *                 sample_weight = np.asarray(sample_weight,
 *                                            dtype=DOUBLE, order="C")
 *             sample_weight_ptr = <DOUBLE_t*> sample_weight.data             # <<<<<<<<<<<<<<
 * 
 *         # Initial capacity
 */
    __pyx_v_sample_weight_ptr = ((__pyx_t_7sklearn_4tree_5_tree_DOUBLE_t *)__pyx_v_sample_weight->data);
    goto __pyx_L5;
  }
  __pyx_L5:;

  /* "sklearn/tree/_tree.pyx":1496
 *         cdef int init_capacity
 * 
 *         if self.max_depth <= 10:             # <<<<<<<<<<<<<<
 *             init_capacity = (2 ** (self.max_depth + 1)) - 1
 *         else:
 */
  __pyx_t_4 = ((__pyx_v_self->max_depth <= 10) != 0);
  if (__pyx_t_4) {

    /* "sklearn/tree/_tree.pyx":1497
 * 
 *         if self.max_depth <= 10:
 *             init_capacity = (2 ** (self.max_depth + 1)) - 1             # <<<<<<<<<<<<<<
 *         else:
 *             init_capacity = 2047
 */
    __pyx_v_init_capacity = (__Pyx_pow_long(2, (__pyx_v_self->max_depth + 1)) - 1);
    goto __pyx_L7;
  }
  /*else*/ {

    /* "sklearn/tree/_tree.pyx":1499
 *             init_capacity = (2 ** (self.max_depth + 1)) - 1
 *         else:
 *             init_capacity = 2047             # <<<<<<<<<<<<<<
 * 
 *         self._resize(init_capacity)
 */
    __pyx_v_init_capacity = 2047;
  }
  __pyx_L7:;

  /* "sklearn/tree/_tree.pyx":1501
 *             init_capacity = 2047
 * 
 *         self._resize(init_capacity)             # <<<<<<<<<<<<<<
 * 
 *         # Recursive partition (without actual recursion)
 */
  __pyx_t_8.__pyx_n = 1;
  __pyx_t_8.capacity = __pyx_v_init_capacity;
  ((struct __pyx_vtabstruct_7sklearn_4tree_5_tree_Tree *)__pyx_v_self->__pyx_vtab)->_resize(__pyx_v_self, &__pyx_t_8); 

  /* "sklearn/tree/_tree.pyx":1504
 * 
 *         # Recursive partition (without actual recursion)
 *         cdef Splitter splitter = self.splitter             # <<<<<<<<<<<<<<
 *         splitter.init(X, y, sample_weight_ptr)
 * 
 */
  __pyx_t_3 = ((PyObject *)__pyx_v_self->splitter);
  __Pyx_INCREF(__pyx_t_3);
  __pyx_v_splitter = ((struct __pyx_obj_7sklearn_4tree_5_tree_Splitter *)__pyx_t_3);
  __pyx_t_3 = 0;

  /* "sklearn/tree/_tree.pyx":1505
 *         # Recursive partition (without actual recursion)
 *         cdef Splitter splitter = self.splitter
 *         splitter.init(X, y, sample_weight_ptr)             # <<<<<<<<<<<<<<
 * 
 *         cdef SIZE_t stack_n_values = 5
 */
  ((struct __pyx_vtabstruct_7sklearn_4tree_5_tree_Splitter *)__pyx_v_splitter->__pyx_vtab)->init(__pyx_v_splitter, __pyx_v_X, __pyx_v_y, __pyx_v_sample_weight_ptr);

  /* "sklearn/tree/_tree.pyx":1507
 *         splitter.init(X, y, sample_weight_ptr)
 * 
 *         cdef SIZE_t stack_n_values = 5             # <<<<<<<<<<<<<<
 *         cdef SIZE_t stack_capacity = 50
 *         cdef SIZE_t* stack = <SIZE_t*> malloc(stack_capacity * sizeof(SIZE_t))
 */
  __pyx_v_stack_n_values = 5;

  /* "sklearn/tree/_tree.pyx":1508
 * 
 *         cdef SIZE_t stack_n_values = 5
 *         cdef SIZE_t stack_capacity = 50             # <<<<<<<<<<<<<<
 *         cdef SIZE_t* stack = <SIZE_t*> malloc(stack_capacity * sizeof(SIZE_t))
 * 
 */
  __pyx_v_stack_capacity = 50;

  /* "sklearn/tree/_tree.pyx":1509
 *         cdef SIZE_t stack_n_values = 5
 *         cdef SIZE_t stack_capacity = 50
 *         cdef SIZE_t* stack = <SIZE_t*> malloc(stack_capacity * sizeof(SIZE_t))             # <<<<<<<<<<<<<<
 * 
 *         stack[0] = 0                    # start
 */
  __pyx_v_stack = ((__pyx_t_7sklearn_4tree_5_tree_SIZE_t *)malloc((__pyx_v_stack_capacity * (sizeof(__pyx_t_7sklearn_4tree_5_tree_SIZE_t)))));

  /* "sklearn/tree/_tree.pyx":1511
 *         cdef SIZE_t* stack = <SIZE_t*> malloc(stack_capacity * sizeof(SIZE_t))
 * 
 *         stack[0] = 0                    # start             # <<<<<<<<<<<<<<
 *         stack[1] = splitter.n_samples   # end
 *         stack[2] = 0                    # depth
 */
  (__pyx_v_stack[0]) = 0;

  /* "sklearn/tree/_tree.pyx":1512
 * 
 *         stack[0] = 0                    # start
 *         stack[1] = splitter.n_samples   # end             # <<<<<<<<<<<<<<
 *         stack[2] = 0                    # depth
 *         stack[3] = _TREE_UNDEFINED      # parent
 */
  __pyx_t_9 = __pyx_v_splitter->n_samples;
  (__pyx_v_stack[1]) = __pyx_t_9;

  /* "sklearn/tree/_tree.pyx":1513
 *         stack[0] = 0                    # start
 *         stack[1] = splitter.n_samples   # end
 *         stack[2] = 0                    # depth             # <<<<<<<<<<<<<<
 *         stack[3] = _TREE_UNDEFINED      # parent
 *         stack[4] = 0                    # is_left
 */
  (__pyx_v_stack[2]) = 0;

  /* "sklearn/tree/_tree.pyx":1514
 *         stack[1] = splitter.n_samples   # end
 *         stack[2] = 0                    # depth
 *         stack[3] = _TREE_UNDEFINED      # parent             # <<<<<<<<<<<<<<
 *         stack[4] = 0                    # is_left
 * 
 */
  (__pyx_v_stack[3]) = __pyx_v_7sklearn_4tree_5_tree__TREE_UNDEFINED;

  /* "sklearn/tree/_tree.pyx":1515
 *         stack[2] = 0                    # depth
 *         stack[3] = _TREE_UNDEFINED      # parent
 *         stack[4] = 0                    # is_left             # <<<<<<<<<<<<<<
 * 
 *         cdef SIZE_t start
 */
  (__pyx_v_stack[4]) = 0;

  /* "sklearn/tree/_tree.pyx":1532
 *         cdef SIZE_t node_id
 * 
 *         while stack_n_values > 0:             # <<<<<<<<<<<<<<
 *             stack_n_values -= 5
 * 
 */
  while (1) {
    __pyx_t_4 = ((__pyx_v_stack_n_values > 0) != 0);
    if (!__pyx_t_4) break;

    /* "sklearn/tree/_tree.pyx":1533
 * 
 *         while stack_n_values > 0:
 *             stack_n_values -= 5             # <<<<<<<<<<<<<<
 * 
 *             start = stack[stack_n_values]
 */
    __pyx_v_stack_n_values = (__pyx_v_stack_n_values - 5);

    /* "sklearn/tree/_tree.pyx":1535
 *             stack_n_values -= 5
 * 
 *             start = stack[stack_n_values]             # <<<<<<<<<<<<<<
 *             end = stack[stack_n_values + 1]
 *             depth = stack[stack_n_values + 2]
 */
    __pyx_v_start = (__pyx_v_stack[__pyx_v_stack_n_values]);

    /* "sklearn/tree/_tree.pyx":1536
 * 
 *             start = stack[stack_n_values]
 *             end = stack[stack_n_values + 1]             # <<<<<<<<<<<<<<
 *             depth = stack[stack_n_values + 2]
 *             parent = stack[stack_n_values + 3]
 */
    __pyx_v_end = (__pyx_v_stack[(__pyx_v_stack_n_values + 1)]);

    /* "sklearn/tree/_tree.pyx":1537
 *             start = stack[stack_n_values]
 *             end = stack[stack_n_values + 1]
 *             depth = stack[stack_n_values + 2]             # <<<<<<<<<<<<<<
 *             parent = stack[stack_n_values + 3]
 *             is_left = stack[stack_n_values + 4]
 */
    __pyx_v_depth = (__pyx_v_stack[(__pyx_v_stack_n_values + 2)]);

    /* "sklearn/tree/_tree.pyx":1538
 *             end = stack[stack_n_values + 1]
 *             depth = stack[stack_n_values + 2]
 *             parent = stack[stack_n_values + 3]             # <<<<<<<<<<<<<<
 *             is_left = stack[stack_n_values + 4]
 * 
 */
    __pyx_v_parent = (__pyx_v_stack[(__pyx_v_stack_n_values + 3)]);

    /* "sklearn/tree/_tree.pyx":1539
 *             depth = stack[stack_n_values + 2]
 *             parent = stack[stack_n_values + 3]
 *             is_left = stack[stack_n_values + 4]             # <<<<<<<<<<<<<<
 * 
 *             n_node_samples = end - start
 */
    __pyx_v_is_left = (__pyx_v_stack[(__pyx_v_stack_n_values + 4)]);

    /* "sklearn/tree/_tree.pyx":1541
 *             is_left = stack[stack_n_values + 4]
 * 
 *             n_node_samples = end - start             # <<<<<<<<<<<<<<
 *             is_leaf = ((depth >= self.max_depth) or
 *                        (n_node_samples < self.min_samples_split) or
 */
    __pyx_v_n_node_samples = (__pyx_v_end - __pyx_v_start);

    /* "sklearn/tree/_tree.pyx":1542
 * 
 *             n_node_samples = end - start
 *             is_leaf = ((depth >= self.max_depth) or             # <<<<<<<<<<<<<<
 *                        (n_node_samples < self.min_samples_split) or
 *                        (n_node_samples < 2 * self.min_samples_leaf))
 */
    __pyx_t_4 = (__pyx_v_depth >= __pyx_v_self->max_depth);
    if (!__pyx_t_4) {

      /* "sklearn/tree/_tree.pyx":1543
 *             n_node_samples = end - start
 *             is_leaf = ((depth >= self.max_depth) or
 *                        (n_node_samples < self.min_samples_split) or             # <<<<<<<<<<<<<<
 *                        (n_node_samples < 2 * self.min_samples_leaf))
 * 
 */
      __pyx_t_5 = (__pyx_v_n_node_samples < __pyx_v_self->min_samples_split);
      if (!__pyx_t_5) {

        /* "sklearn/tree/_tree.pyx":1544
 *             is_leaf = ((depth >= self.max_depth) or
 *                        (n_node_samples < self.min_samples_split) or
 *                        (n_node_samples < 2 * self.min_samples_leaf))             # <<<<<<<<<<<<<<
 * 
 *             splitter.node_reset(start, end, &impurity)
 */
        __pyx_t_6 = (__pyx_v_n_node_samples < (2 * __pyx_v_self->min_samples_leaf));
        __pyx_t_10 = __pyx_t_6;
      } else {
        __pyx_t_10 = __pyx_t_5;
      }
      __pyx_t_5 = __pyx_t_10;
    } else {
      __pyx_t_5 = __pyx_t_4;
    }
    __pyx_v_is_leaf = __pyx_t_5;

    /* "sklearn/tree/_tree.pyx":1546
 *                        (n_node_samples < 2 * self.min_samples_leaf))
 * 
 *             splitter.node_reset(start, end, &impurity)             # <<<<<<<<<<<<<<
 *             is_leaf = is_leaf or (impurity < 1e-7)
 * 
 */
    ((struct __pyx_vtabstruct_7sklearn_4tree_5_tree_Splitter *)__pyx_v_splitter->__pyx_vtab)->node_reset(__pyx_v_splitter, __pyx_v_start, __pyx_v_end, (&__pyx_v_impurity));

    /* "sklearn/tree/_tree.pyx":1547
 * 
 *             splitter.node_reset(start, end, &impurity)
 *             is_leaf = is_leaf or (impurity < 1e-7)             # <<<<<<<<<<<<<<
 * 
 *             if not is_leaf:
 */
    if (!__pyx_v_is_leaf) {
      __pyx_t_5 = (__pyx_v_impurity < 1e-7);
      __pyx_t_4 = __pyx_t_5;
    } else {
      __pyx_t_4 = __pyx_v_is_leaf;
    }
    __pyx_v_is_leaf = __pyx_t_4;

    /* "sklearn/tree/_tree.pyx":1549
 *             is_leaf = is_leaf or (impurity < 1e-7)
 * 
 *             if not is_leaf:             # <<<<<<<<<<<<<<
 *                 splitter.node_split(&pos, &feature, &threshold)
 *                 is_leaf = is_leaf or (pos >= end)
 */
    __pyx_t_4 = ((!(__pyx_v_is_leaf != 0)) != 0);
    if (__pyx_t_4) {

      /* "sklearn/tree/_tree.pyx":1550
 * 
 *             if not is_leaf:
 *                 splitter.node_split(&pos, &feature, &threshold)             # <<<<<<<<<<<<<<
 *                 is_leaf = is_leaf or (pos >= end)
 * 
 */
      ((struct __pyx_vtabstruct_7sklearn_4tree_5_tree_Splitter *)__pyx_v_splitter->__pyx_vtab)->node_split(__pyx_v_splitter, (&__pyx_v_pos), (&__pyx_v_feature), (&__pyx_v_threshold));

      /* "sklearn/tree/_tree.pyx":1551
 *             if not is_leaf:
 *                 splitter.node_split(&pos, &feature, &threshold)
 *                 is_leaf = is_leaf or (pos >= end)             # <<<<<<<<<<<<<<
 * 
 *             node_id = self._add_node(parent, is_left, is_leaf, feature,
 */
      if (!__pyx_v_is_leaf) {
        __pyx_t_4 = (__pyx_v_pos >= __pyx_v_end);
        __pyx_t_5 = __pyx_t_4;
      } else {
        __pyx_t_5 = __pyx_v_is_leaf;
      }
      __pyx_v_is_leaf = __pyx_t_5;
      goto __pyx_L10;
    }
    __pyx_L10:;

    /* "sklearn/tree/_tree.pyx":1554
 * 
 *             node_id = self._add_node(parent, is_left, is_leaf, feature,
 *                                      threshold, impurity, n_node_samples)             # <<<<<<<<<<<<<<
 * 
 *             if is_leaf:
 */
    __pyx_v_node_id = ((struct __pyx_vtabstruct_7sklearn_4tree_5_tree_Tree *)__pyx_v_self->__pyx_vtab)->_add_node(__pyx_v_self, __pyx_v_parent, __pyx_v_is_left, __pyx_v_is_leaf, __pyx_v_feature, __pyx_v_threshold, __pyx_v_impurity, __pyx_v_n_node_samples);

    /* "sklearn/tree/_tree.pyx":1556
 *                                      threshold, impurity, n_node_samples)
 * 
 *             if is_leaf:             # <<<<<<<<<<<<<<
 *                 # Don't store value for internal nodes
 *                 splitter.node_value(self.value + node_id * self.value_stride)
 */
    __pyx_t_5 = (__pyx_v_is_leaf != 0);
    if (__pyx_t_5) {

      /* "sklearn/tree/_tree.pyx":1558
 *             if is_leaf:
 *                 # Don't store value for internal nodes
 *                 splitter.node_value(self.value + node_id * self.value_stride)             # <<<<<<<<<<<<<<
 * 
 *             else:
 */
      ((struct __pyx_vtabstruct_7sklearn_4tree_5_tree_Splitter *)__pyx_v_splitter->__pyx_vtab)->node_value(__pyx_v_splitter, (__pyx_v_self->value + (__pyx_v_node_id * __pyx_v_self->value_stride)));
      goto __pyx_L11;
    }
    /*else*/ {

      /* "sklearn/tree/_tree.pyx":1561
 * 
 *             else:
 *                 if stack_n_values + 10 > stack_capacity:             # <<<<<<<<<<<<<<
 *                     stack_capacity *= 2
 *                     stack = <SIZE_t*> realloc(stack, stack_capacity * sizeof(SIZE_t))
 */
      __pyx_t_5 = (((__pyx_v_stack_n_values + 10) > __pyx_v_stack_capacity) != 0);
      if (__pyx_t_5) {

        /* "sklearn/tree/_tree.pyx":1562
 *             else:
 *                 if stack_n_values + 10 > stack_capacity:
 *                     stack_capacity *= 2             # <<<<<<<<<<<<<<
 *                     stack = <SIZE_t*> realloc(stack, stack_capacity * sizeof(SIZE_t))
 * 
 */
        __pyx_v_stack_capacity = (__pyx_v_stack_capacity * 2);

        /* "sklearn/tree/_tree.pyx":1563
 *                 if stack_n_values + 10 > stack_capacity:
 *                     stack_capacity *= 2
 *                     stack = <SIZE_t*> realloc(stack, stack_capacity * sizeof(SIZE_t))             # <<<<<<<<<<<<<<
 * 
 *                 # Stack right child
 */
        __pyx_v_stack = ((__pyx_t_7sklearn_4tree_5_tree_SIZE_t *)realloc(__pyx_v_stack, (__pyx_v_stack_capacity * (sizeof(__pyx_t_7sklearn_4tree_5_tree_SIZE_t)))));
        goto __pyx_L12;
      }
      __pyx_L12:;

      /* "sklearn/tree/_tree.pyx":1566
 * 
 *                 # Stack right child
 *                 stack[stack_n_values] = pos             # <<<<<<<<<<<<<<
 *                 stack[stack_n_values + 1] = end
 *                 stack[stack_n_values + 2] = depth + 1
 */
      (__pyx_v_stack[__pyx_v_stack_n_values]) = __pyx_v_pos;

      /* "sklearn/tree/_tree.pyx":1567
 *                 # Stack right child
 *                 stack[stack_n_values] = pos
 *                 stack[stack_n_values + 1] = end             # <<<<<<<<<<<<<<
 *                 stack[stack_n_values + 2] = depth + 1
 *                 stack[stack_n_values + 3] = node_id
 */
      (__pyx_v_stack[(__pyx_v_stack_n_values + 1)]) = __pyx_v_end;

      /* "sklearn/tree/_tree.pyx":1568
 *                 stack[stack_n_values] = pos
 *                 stack[stack_n_values + 1] = end
 *                 stack[stack_n_values + 2] = depth + 1             # <<<<<<<<<<<<<<
 *                 stack[stack_n_values + 3] = node_id
 *                 stack[stack_n_values + 4] = 0
 */
      (__pyx_v_stack[(__pyx_v_stack_n_values + 2)]) = (__pyx_v_depth + 1);

      /* "sklearn/tree/_tree.pyx":1569
 *                 stack[stack_n_values + 1] = end
 *                 stack[stack_n_values + 2] = depth + 1
 *                 stack[stack_n_values + 3] = node_id             # <<<<<<<<<<<<<<
 *                 stack[stack_n_values + 4] = 0
 *                 stack_n_values += 5
 */
      (__pyx_v_stack[(__pyx_v_stack_n_values + 3)]) = __pyx_v_node_id;

      /* "sklearn/tree/_tree.pyx":1570
 *                 stack[stack_n_values + 2] = depth + 1
 *                 stack[stack_n_values + 3] = node_id
 *                 stack[stack_n_values + 4] = 0             # <<<<<<<<<<<<<<
 *                 stack_n_values += 5
 * 
 */
      (__pyx_v_stack[(__pyx_v_stack_n_values + 4)]) = 0;

      /* "sklearn/tree/_tree.pyx":1571
 *                 stack[stack_n_values + 3] = node_id
 *                 stack[stack_n_values + 4] = 0
 *                 stack_n_values += 5             # <<<<<<<<<<<<<<
 * 
 *                 # Stack left child
 */
      __pyx_v_stack_n_values = (__pyx_v_stack_n_values + 5);

      /* "sklearn/tree/_tree.pyx":1574
 * 
 *                 # Stack left child
 *                 stack[stack_n_values] = start             # <<<<<<<<<<<<<<
 *                 stack[stack_n_values + 1] = pos
 *                 stack[stack_n_values + 2] = depth + 1
 */
      (__pyx_v_stack[__pyx_v_stack_n_values]) = __pyx_v_start;

      /* "sklearn/tree/_tree.pyx":1575
 *                 # Stack left child
 *                 stack[stack_n_values] = start
 *                 stack[stack_n_values + 1] = pos             # <<<<<<<<<<<<<<
 *                 stack[stack_n_values + 2] = depth + 1
 *                 stack[stack_n_values + 3] = node_id
 */
      (__pyx_v_stack[(__pyx_v_stack_n_values + 1)]) = __pyx_v_pos;

      /* "sklearn/tree/_tree.pyx":1576
 *                 stack[stack_n_values] = start
 *                 stack[stack_n_values + 1] = pos
 *                 stack[stack_n_values + 2] = depth + 1             # <<<<<<<<<<<<<<
 *                 stack[stack_n_values + 3] = node_id
 *                 stack[stack_n_values + 4] = 1
 */
      (__pyx_v_stack[(__pyx_v_stack_n_values + 2)]) = (__pyx_v_depth + 1);

      /* "sklearn/tree/_tree.pyx":1577
 *                 stack[stack_n_values + 1] = pos
 *                 stack[stack_n_values + 2] = depth + 1
 *                 stack[stack_n_values + 3] = node_id             # <<<<<<<<<<<<<<
 *                 stack[stack_n_values + 4] = 1
 *                 stack_n_values += 5
 */
      (__pyx_v_stack[(__pyx_v_stack_n_values + 3)]) = __pyx_v_node_id;

      /* "sklearn/tree/_tree.pyx":1578
 *                 stack[stack_n_values + 2] = depth + 1
 *                 stack[stack_n_values + 3] = node_id
 *                 stack[stack_n_values + 4] = 1             # <<<<<<<<<<<<<<
 *                 stack_n_values += 5
 * 
 */
      (__pyx_v_stack[(__pyx_v_stack_n_values + 4)]) = 1;

      /* "sklearn/tree/_tree.pyx":1579
 *                 stack[stack_n_values + 3] = node_id
 *                 stack[stack_n_values + 4] = 1
 *                 stack_n_values += 5             # <<<<<<<<<<<<<<
 * 
 *         self._resize(self.node_count)
 */
      __pyx_v_stack_n_values = (__pyx_v_stack_n_values + 5);
    }
    __pyx_L11:;
  }

  /* "sklearn/tree/_tree.pyx":1581
 *                 stack_n_values += 5
 * 
 *         self._resize(self.node_count)             # <<<<<<<<<<<<<<
 *         free(stack)
 * 
 */
  __pyx_t_8.__pyx_n = 1;
  __pyx_t_8.capacity = __pyx_v_self->node_count;
  ((struct __pyx_vtabstruct_7sklearn_4tree_5_tree_Tree *)__pyx_v_self->__pyx_vtab)->_resize(__pyx_v_self, &__pyx_t_8); 

  /* "sklearn/tree/_tree.pyx":1582
 * 
 *         self._resize(self.node_count)
 *         free(stack)             # <<<<<<<<<<<<<<
 * 
 *     cpdef predict(self, np.ndarray[DTYPE_t, ndim=2] X):
 */
  free(__pyx_v_stack);

  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_AddTraceback("sklearn.tree._tree.Tree.build", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_splitter);
  __Pyx_XDECREF((PyObject *)__pyx_v_X);
  __Pyx_XDECREF((PyObject *)__pyx_v_y);
  __Pyx_XDECREF((PyObject *)__pyx_v_sample_weight);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_7sklearn_4tree_5_tree_4Tree_11build(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_7sklearn_4tree_5_tree_4Tree_10build[] = "Build a decision tree from the training set (X, y).";
static PyObject *__pyx_pw_7sklearn_4tree_5_tree_4Tree_11build(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyArrayObject *__pyx_v_X = 0;
  PyArrayObject *__pyx_v_y = 0;
  PyArrayObject *__pyx_v_sample_weight = 0;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("build (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s__X,&__pyx_n_s__y,&__pyx_n_s__sample_weight,0};
    PyObject* values[3] = {0,0,0};

    /* "sklearn/tree/_tree.pyx":1476
 *     cpdef build(self, np.ndarray X,
 *                       np.ndarray y,
 *                       np.ndarray sample_weight=None):             # <<<<<<<<<<<<<<
 *         """Build a decision tree from the training set (X, y)."""
 *         # Prepare data before recursive partitioning
 */
    values[2] = (PyObject *)((PyArrayObject *)Py_None);
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s__X)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s__y)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("build", 0, 2, 3, 1); {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1474; __pyx_clineno = __LINE__; goto __pyx_L3_error;}
        }
        case  2:
        if (kw_args > 0) {
          PyObject* value = PyDict_GetItem(__pyx_kwds, __pyx_n_s__sample_weight);
          if (value) { values[2] = value; kw_args--; }
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "build") < 0)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1474; __pyx_clineno = __LINE__; goto __pyx_L3_error;}
      }
    } else {
      switch (PyTuple_GET_SIZE(__pyx_args)) {
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_X = ((PyArrayObject *)values[0]);
    __pyx_v_y = ((PyArrayObject *)values[1]);
    __pyx_v_sample_weight = ((PyArrayObject *)values[2]);
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("build", 0, 2, 3, PyTuple_GET_SIZE(__pyx_args)); {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1474; __pyx_clineno = __LINE__; goto __pyx_L3_error;}
  __pyx_L3_error:;
  __Pyx_AddTraceback("sklearn.tree._tree.Tree.build", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_X), __pyx_ptype_5numpy_ndarray, 1, "X", 0))) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1474; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_y), __pyx_ptype_5numpy_ndarray, 1, "y", 0))) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1475; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_sample_weight), __pyx_ptype_5numpy_ndarray, 1, "sample_weight", 0))) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1476; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __pyx_r = __pyx_pf_7sklearn_4tree_5_tree_4Tree_10build(((struct __pyx_obj_7sklearn_4tree_5_tree_Tree *)__pyx_v_self), __pyx_v_X, __pyx_v_y, __pyx_v_sample_weight);
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "sklearn/tree/_tree.pyx":1474
 *         return node_id
 * 
 *     cpdef build(self, np.ndarray X,             # <<<<<<<<<<<<<<
 *                       np.ndarray y,
 *                       np.ndarray sample_weight=None):
 */

static PyObject *__pyx_pf_7sklearn_4tree_5_tree_4Tree_10build(struct __pyx_obj_7sklearn_4tree_5_tree_Tree *__pyx_v_self, PyArrayObject *__pyx_v_X, PyArrayObject *__pyx_v_y, PyArrayObject *__pyx_v_sample_weight) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  struct __pyx_opt_args_7sklearn_4tree_5_tree_4Tree_build __pyx_t_2;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("build", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_2.__pyx_n = 1;
  __pyx_t_2.sample_weight = __pyx_v_sample_weight;
  __pyx_t_1 = ((struct __pyx_vtabstruct_7sklearn_4tree_5_tree_Tree *)__pyx_v_self->__pyx_vtab)->build(__pyx_v_self, __pyx_v_X, __pyx_v_y, 1, &__pyx_t_2); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1474; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("sklearn.tree._tree.Tree.build", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "sklearn/tree/_tree.pyx":1584
 *         free(stack)
 * 
 *     cpdef predict(self, np.ndarray[DTYPE_t, ndim=2] X):             # <<<<<<<<<<<<<<
 *         """Predict target for X."""
 *         cdef SIZE_t* children_left = self.children_left
 */

static PyObject *__pyx_pw_7sklearn_4tree_5_tree_4Tree_13predict(PyObject *__pyx_v_self, PyObject *__pyx_v_X); /*proto*/
static PyObject *__pyx_f_7sklearn_4tree_5_tree_4Tree_predict(struct __pyx_obj_7sklearn_4tree_5_tree_Tree *__pyx_v_self, PyArrayObject *__pyx_v_X, int __pyx_skip_dispatch) {
  __pyx_t_7sklearn_4tree_5_tree_SIZE_t *__pyx_v_children_left;
  __pyx_t_7sklearn_4tree_5_tree_SIZE_t *__pyx_v_children_right;
  __pyx_t_7sklearn_4tree_5_tree_SIZE_t *__pyx_v_feature;
  double *__pyx_v_threshold;
  double *__pyx_v_value;
  __pyx_t_7sklearn_4tree_5_tree_SIZE_t __pyx_v_n_samples;
  __pyx_t_7sklearn_4tree_5_tree_SIZE_t *__pyx_v_n_classes;
  __pyx_t_7sklearn_4tree_5_tree_SIZE_t __pyx_v_n_outputs;
  __pyx_t_7sklearn_4tree_5_tree_SIZE_t __pyx_v_max_n_classes;
  __pyx_t_7sklearn_4tree_5_tree_SIZE_t __pyx_v_value_stride;
  __pyx_t_7sklearn_4tree_5_tree_SIZE_t __pyx_v_node_id;
  __pyx_t_7sklearn_4tree_5_tree_SIZE_t __pyx_v_offset;
  __pyx_t_7sklearn_4tree_5_tree_SIZE_t __pyx_v_i;
  __pyx_t_7sklearn_4tree_5_tree_SIZE_t __pyx_v_k;
  __pyx_t_7sklearn_4tree_5_tree_SIZE_t __pyx_v_c;
  PyArrayObject *__pyx_v_out = 0;
  PyArrayObject *__pyx_v_out_multi = 0;
  __Pyx_LocalBuf_ND __pyx_pybuffernd_X;
  __Pyx_Buffer __pyx_pybuffer_X;
  __Pyx_LocalBuf_ND __pyx_pybuffernd_out;
  __Pyx_Buffer __pyx_pybuffer_out;
  __Pyx_LocalBuf_ND __pyx_pybuffernd_out_multi;
  __Pyx_Buffer __pyx_pybuffer_out_multi;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  __pyx_t_7sklearn_4tree_5_tree_SIZE_t *__pyx_t_4;
  double *__pyx_t_5;
  __pyx_t_7sklearn_4tree_5_tree_SIZE_t __pyx_t_6;
  int __pyx_t_7;
  PyObject *__pyx_t_8 = NULL;
  PyObject *__pyx_t_9 = NULL;
  PyArrayObject *__pyx_t_10 = NULL;
  int __pyx_t_11;
  PyObject *__pyx_t_12 = NULL;
  PyObject *__pyx_t_13 = NULL;
  PyObject *__pyx_t_14 = NULL;
  __pyx_t_7sklearn_4tree_5_tree_SIZE_t __pyx_t_15;
  __pyx_t_7sklearn_4tree_5_tree_SIZE_t __pyx_t_16;
  __pyx_t_7sklearn_4tree_5_tree_SIZE_t __pyx_t_17;
  __pyx_t_7sklearn_4tree_5_tree_SIZE_t __pyx_t_18;
  __pyx_t_7sklearn_4tree_5_tree_SIZE_t __pyx_t_19;
  PyArrayObject *__pyx_t_20 = NULL;
  __pyx_t_7sklearn_4tree_5_tree_SIZE_t __pyx_t_21;
  __pyx_t_7sklearn_4tree_5_tree_SIZE_t __pyx_t_22;
  __pyx_t_7sklearn_4tree_5_tree_SIZE_t __pyx_t_23;
  __pyx_t_7sklearn_4tree_5_tree_SIZE_t __pyx_t_24;
  __pyx_t_7sklearn_4tree_5_tree_SIZE_t __pyx_t_25;
  __pyx_t_7sklearn_4tree_5_tree_SIZE_t __pyx_t_26;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("predict", 0);
  __pyx_pybuffer_out.pybuffer.buf = NULL;
  __pyx_pybuffer_out.refcount = 0;
  __pyx_pybuffernd_out.data = NULL;
  __pyx_pybuffernd_out.rcbuffer = &__pyx_pybuffer_out;
  __pyx_pybuffer_out_multi.pybuffer.buf = NULL;
  __pyx_pybuffer_out_multi.refcount = 0;
  __pyx_pybuffernd_out_multi.data = NULL;
  __pyx_pybuffernd_out_multi.rcbuffer = &__pyx_pybuffer_out_multi;
  __pyx_pybuffer_X.pybuffer.buf = NULL;
  __pyx_pybuffer_X.refcount = 0;
  __pyx_pybuffernd_X.data = NULL;
  __pyx_pybuffernd_X.rcbuffer = &__pyx_pybuffer_X;
  {
    __Pyx_BufFmt_StackElem __pyx_stack[1];
    if (unlikely(__Pyx_GetBufferAndValidate(&__pyx_pybuffernd_X.rcbuffer->pybuffer, (PyObject*)__pyx_v_X, &__Pyx_TypeInfo_nn___pyx_t_7sklearn_4tree_5_tree_DTYPE_t, PyBUF_FORMAT| PyBUF_STRIDES, 2, 0, __pyx_stack) == -1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1584; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  }
  __pyx_pybuffernd_X.diminfo[0].strides = __pyx_pybuffernd_X.rcbuffer->pybuffer.strides[0]; __pyx_pybuffernd_X.diminfo[0].shape = __pyx_pybuffernd_X.rcbuffer->pybuffer.shape[0]; __pyx_pybuffernd_X.diminfo[1].strides = __pyx_pybuffernd_X.rcbuffer->pybuffer.strides[1]; __pyx_pybuffernd_X.diminfo[1].shape = __pyx_pybuffernd_X.rcbuffer->pybuffer.shape[1];
  /* Check if called by wrapper */
  if (unlikely(__pyx_skip_dispatch)) ;
  /* Check if overridden in Python */
  else if (unlikely(Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0)) {
    __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s__predict); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1584; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    __Pyx_GOTREF(__pyx_t_1);
    if (!PyCFunction_Check(__pyx_t_1) || (PyCFunction_GET_FUNCTION(__pyx_t_1) != (PyCFunction)__pyx_pw_7sklearn_4tree_5_tree_4Tree_13predict)) {
      __Pyx_XDECREF(__pyx_r);
      __pyx_t_2 = PyTuple_New(1); if (unlikely(!__pyx_t_2)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1584; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_INCREF(((PyObject *)__pyx_v_X));
      PyTuple_SET_ITEM(__pyx_t_2, 0, ((PyObject *)__pyx_v_X));
      __Pyx_GIVEREF(((PyObject *)__pyx_v_X));
      __pyx_t_3 = PyObject_Call(__pyx_t_1, ((PyObject *)__pyx_t_2), NULL); if (unlikely(!__pyx_t_3)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1584; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_DECREF(((PyObject *)__pyx_t_2)); __pyx_t_2 = 0;
      __pyx_r = __pyx_t_3;
      __pyx_t_3 = 0;
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      goto __pyx_L0;
    }
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  }

  /* "sklearn/tree/_tree.pyx":1586
 *     cpdef predict(self, np.ndarray[DTYPE_t, ndim=2] X):
 *         """Predict target for X."""
 *         cdef SIZE_t* children_left = self.children_left             # <<<<<<<<<<<<<<
 *         cdef SIZE_t* children_right = self.children_right
 *         cdef SIZE_t* feature = self.feature
 */
  __pyx_t_4 = __pyx_v_self->children_left;
  __pyx_v_children_left = __pyx_t_4;

  /* "sklearn/tree/_tree.pyx":1587
 *         """Predict target for X."""
 *         cdef SIZE_t* children_left = self.children_left
 *         cdef SIZE_t* children_right = self.children_right             # <<<<<<<<<<<<<<
 *         cdef SIZE_t* feature = self.feature
 *         cdef double* threshold = self.threshold
 */
  __pyx_t_4 = __pyx_v_self->children_right;
  __pyx_v_children_right = __pyx_t_4;

  /* "sklearn/tree/_tree.pyx":1588
 *         cdef SIZE_t* children_left = self.children_left
 *         cdef SIZE_t* children_right = self.children_right
 *         cdef SIZE_t* feature = self.feature             # <<<<<<<<<<<<<<
 *         cdef double* threshold = self.threshold
 *         cdef double* value = self.value
 */
  __pyx_t_4 = __pyx_v_self->feature;
  __pyx_v_feature = __pyx_t_4;

  /* "sklearn/tree/_tree.pyx":1589
 *         cdef SIZE_t* children_right = self.children_right
 *         cdef SIZE_t* feature = self.feature
 *         cdef double* threshold = self.threshold             # <<<<<<<<<<<<<<
 *         cdef double* value = self.value
 * 
 */
  __pyx_t_5 = __pyx_v_self->threshold;
  __pyx_v_threshold = __pyx_t_5;

  /* "sklearn/tree/_tree.pyx":1590
 *         cdef SIZE_t* feature = self.feature
 *         cdef double* threshold = self.threshold
 *         cdef double* value = self.value             # <<<<<<<<<<<<<<
 * 
 *         cdef SIZE_t n_samples = X.shape[0]
 */
  __pyx_t_5 = __pyx_v_self->value;
  __pyx_v_value = __pyx_t_5;

  /* "sklearn/tree/_tree.pyx":1592
 *         cdef double* value = self.value
 * 
 *         cdef SIZE_t n_samples = X.shape[0]             # <<<<<<<<<<<<<<
 *         cdef SIZE_t* n_classes = self.n_classes
 *         cdef SIZE_t n_outputs = self.n_outputs
 */
  __pyx_v_n_samples = (__pyx_v_X->dimensions[0]);

  /* "sklearn/tree/_tree.pyx":1593
 * 
 *         cdef SIZE_t n_samples = X.shape[0]
 *         cdef SIZE_t* n_classes = self.n_classes             # <<<<<<<<<<<<<<
 *         cdef SIZE_t n_outputs = self.n_outputs
 *         cdef SIZE_t max_n_classes = self.max_n_classes
 */
  __pyx_t_4 = __pyx_v_self->n_classes;
  __pyx_v_n_classes = __pyx_t_4;

  /* "sklearn/tree/_tree.pyx":1594
 *         cdef SIZE_t n_samples = X.shape[0]
 *         cdef SIZE_t* n_classes = self.n_classes
 *         cdef SIZE_t n_outputs = self.n_outputs             # <<<<<<<<<<<<<<
 *         cdef SIZE_t max_n_classes = self.max_n_classes
 *         cdef SIZE_t value_stride = self.value_stride
 */
  __pyx_t_6 = __pyx_v_self->n_outputs;
  __pyx_v_n_outputs = __pyx_t_6;

  /* "sklearn/tree/_tree.pyx":1595
 *         cdef SIZE_t* n_classes = self.n_classes
 *         cdef SIZE_t n_outputs = self.n_outputs
 *         cdef SIZE_t max_n_classes = self.max_n_classes             # <<<<<<<<<<<<<<
 *         cdef SIZE_t value_stride = self.value_stride
 * 
 */
  __pyx_t_6 = __pyx_v_self->max_n_classes;
  __pyx_v_max_n_classes = __pyx_t_6;

  /* "sklearn/tree/_tree.pyx":1596
 *         cdef SIZE_t n_outputs = self.n_outputs
 *         cdef SIZE_t max_n_classes = self.max_n_classes
 *         cdef SIZE_t value_stride = self.value_stride             # <<<<<<<<<<<<<<
 * 
 *         cdef SIZE_t node_id = 0
 */
  __pyx_t_6 = __pyx_v_self->value_stride;
  __pyx_v_value_stride = __pyx_t_6;

  /* "sklearn/tree/_tree.pyx":1598
 *         cdef SIZE_t value_stride = self.value_stride
 * 
 *         cdef SIZE_t node_id = 0             # <<<<<<<<<<<<<<
 *         cdef SIZE_t offset
 *         cdef SIZE_t i
 */
  __pyx_v_node_id = 0;

  /* "sklearn/tree/_tree.pyx":1607
 *         cdef np.ndarray[np.float64_t, ndim=3] out_multi
 * 
 *         if n_outputs == 1:             # <<<<<<<<<<<<<<
 *             out = np.zeros((n_samples, max_n_classes), dtype=np.float64)
 * 
 */
  __pyx_t_7 = ((__pyx_v_n_outputs == 1) != 0);
  if (__pyx_t_7) {

    /* "sklearn/tree/_tree.pyx":1608
 * 
 *         if n_outputs == 1:
 *             out = np.zeros((n_samples, max_n_classes), dtype=np.float64)             # <<<<<<<<<<<<<<
 * 
 *             for i from 0 <= i < n_samples:
 */
    __pyx_t_1 = __Pyx_GetModuleGlobalName(__pyx_n_s__np); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1608; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s__zeros); if (unlikely(!__pyx_t_3)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1608; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_1 = __Pyx_PyInt_to_py_Py_intptr_t(__pyx_v_n_samples); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1608; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_2 = __Pyx_PyInt_to_py_Py_intptr_t(__pyx_v_max_n_classes); if (unlikely(!__pyx_t_2)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1608; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_8 = PyTuple_New(2); if (unlikely(!__pyx_t_8)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1608; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    __Pyx_GOTREF(__pyx_t_8);
    PyTuple_SET_ITEM(__pyx_t_8, 0, __pyx_t_1);
    __Pyx_GIVEREF(__pyx_t_1);
    PyTuple_SET_ITEM(__pyx_t_8, 1, __pyx_t_2);
    __Pyx_GIVEREF(__pyx_t_2);
    __pyx_t_1 = 0;
    __pyx_t_2 = 0;
    __pyx_t_2 = PyTuple_New(1); if (unlikely(!__pyx_t_2)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1608; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    __Pyx_GOTREF(__pyx_t_2);
    PyTuple_SET_ITEM(__pyx_t_2, 0, ((PyObject *)__pyx_t_8));
    __Pyx_GIVEREF(((PyObject *)__pyx_t_8));
    __pyx_t_8 = 0;
    __pyx_t_8 = PyDict_New(); if (unlikely(!__pyx_t_8)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1608; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    __Pyx_GOTREF(((PyObject *)__pyx_t_8));
    __pyx_t_1 = __Pyx_GetModuleGlobalName(__pyx_n_s__np); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1608; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_9 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s__float64); if (unlikely(!__pyx_t_9)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1608; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    __Pyx_GOTREF(__pyx_t_9);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    if (PyDict_SetItem(__pyx_t_8, ((PyObject *)__pyx_n_s__dtype), __pyx_t_9) < 0) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1608; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
    __pyx_t_9 = PyObject_Call(__pyx_t_3, ((PyObject *)__pyx_t_2), ((PyObject *)__pyx_t_8)); if (unlikely(!__pyx_t_9)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1608; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    __Pyx_GOTREF(__pyx_t_9);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_DECREF(((PyObject *)__pyx_t_2)); __pyx_t_2 = 0;
    __Pyx_DECREF(((PyObject *)__pyx_t_8)); __pyx_t_8 = 0;
    if (!(likely(((__pyx_t_9) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_9, __pyx_ptype_5numpy_ndarray))))) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1608; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    __pyx_t_10 = ((PyArrayObject *)__pyx_t_9);
    {
      __Pyx_BufFmt_StackElem __pyx_stack[1];
      __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_out.rcbuffer->pybuffer);
      __pyx_t_11 = __Pyx_GetBufferAndValidate(&__pyx_pybuffernd_out.rcbuffer->pybuffer, (PyObject*)__pyx_t_10, &__Pyx_TypeInfo_nn___pyx_t_5numpy_float64_t, PyBUF_FORMAT| PyBUF_STRIDES| PyBUF_WRITABLE, 2, 0, __pyx_stack);
      if (unlikely(__pyx_t_11 < 0)) {
        PyErr_Fetch(&__pyx_t_12, &__pyx_t_13, &__pyx_t_14);
        if (unlikely(__Pyx_GetBufferAndValidate(&__pyx_pybuffernd_out.rcbuffer->pybuffer, (PyObject*)__pyx_v_out, &__Pyx_TypeInfo_nn___pyx_t_5numpy_float64_t, PyBUF_FORMAT| PyBUF_STRIDES| PyBUF_WRITABLE, 2, 0, __pyx_stack) == -1)) {
          Py_XDECREF(__pyx_t_12); Py_XDECREF(__pyx_t_13); Py_XDECREF(__pyx_t_14);
          __Pyx_RaiseBufferFallbackError();
        } else {
          PyErr_Restore(__pyx_t_12, __pyx_t_13, __pyx_t_14);
        }
      }
      __pyx_pybuffernd_out.diminfo[0].strides = __pyx_pybuffernd_out.rcbuffer->pybuffer.strides[0]; __pyx_pybuffernd_out.diminfo[0].shape = __pyx_pybuffernd_out.rcbuffer->pybuffer.shape[0]; __pyx_pybuffernd_out.diminfo[1].strides = __pyx_pybuffernd_out.rcbuffer->pybuffer.strides[1]; __pyx_pybuffernd_out.diminfo[1].shape = __pyx_pybuffernd_out.rcbuffer->pybuffer.shape[1];
      if (unlikely(__pyx_t_11 < 0)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1608; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    }
    __pyx_t_10 = 0;
    __pyx_v_out = ((PyArrayObject *)__pyx_t_9);
    __pyx_t_9 = 0;

    /* "sklearn/tree/_tree.pyx":1610
 *             out = np.zeros((n_samples, max_n_classes), dtype=np.float64)
 * 
 *             for i from 0 <= i < n_samples:             # <<<<<<<<<<<<<<
 *                 node_id = 0
 * 
 */
    __pyx_t_6 = __pyx_v_n_samples;
    for (__pyx_v_i = 0; __pyx_v_i < __pyx_t_6; __pyx_v_i++) {

      /* "sklearn/tree/_tree.pyx":1611
 * 
 *             for i from 0 <= i < n_samples:
 *                 node_id = 0             # <<<<<<<<<<<<<<
 * 
 *                 # While node_id not a leaf
 */
      __pyx_v_node_id = 0;

      /* "sklearn/tree/_tree.pyx":1614
 * 
 *                 # While node_id not a leaf
 *                 while children_left[node_id] != _TREE_LEAF:             # <<<<<<<<<<<<<<
 *                     # ... and children_right[node_id] != _TREE_LEAF:
 *                     if X[i, feature[node_id]] <= threshold[node_id]:
 */
      while (1) {
        __pyx_t_7 = (((__pyx_v_children_left[__pyx_v_node_id]) != __pyx_v_7sklearn_4tree_5_tree__TREE_LEAF) != 0);
        if (!__pyx_t_7) break;

        /* "sklearn/tree/_tree.pyx":1616
 *                 while children_left[node_id] != _TREE_LEAF:
 *                     # ... and children_right[node_id] != _TREE_LEAF:
 *                     if X[i, feature[node_id]] <= threshold[node_id]:             # <<<<<<<<<<<<<<
 *                         node_id = children_left[node_id]
 *                     else:
 */
        __pyx_t_15 = __pyx_v_i;
        __pyx_t_16 = (__pyx_v_feature[__pyx_v_node_id]);
        __pyx_t_7 = (((*__Pyx_BufPtrStrided2d(__pyx_t_7sklearn_4tree_5_tree_DTYPE_t *, __pyx_pybuffernd_X.rcbuffer->pybuffer.buf, __pyx_t_15, __pyx_pybuffernd_X.diminfo[0].strides, __pyx_t_16, __pyx_pybuffernd_X.diminfo[1].strides)) <= (__pyx_v_threshold[__pyx_v_node_id])) != 0);
        if (__pyx_t_7) {

          /* "sklearn/tree/_tree.pyx":1617
 *                     # ... and children_right[node_id] != _TREE_LEAF:
 *                     if X[i, feature[node_id]] <= threshold[node_id]:
 *                         node_id = children_left[node_id]             # <<<<<<<<<<<<<<
 *                     else:
 *                         node_id = children_right[node_id]
 */
          __pyx_v_node_id = (__pyx_v_children_left[__pyx_v_node_id]);
          goto __pyx_L8;
        }
        /*else*/ {

          /* "sklearn/tree/_tree.pyx":1619
 *                         node_id = children_left[node_id]
 *                     else:
 *                         node_id = children_right[node_id]             # <<<<<<<<<<<<<<
 * 
 *                 offset = node_id * value_stride
 */
          __pyx_v_node_id = (__pyx_v_children_right[__pyx_v_node_id]);
        }
        __pyx_L8:;
      }

      /* "sklearn/tree/_tree.pyx":1621
 *                         node_id = children_right[node_id]
 * 
 *                 offset = node_id * value_stride             # <<<<<<<<<<<<<<
 * 
 *                 for c from 0 <= c < n_classes[0]:
 */
      __pyx_v_offset = (__pyx_v_node_id * __pyx_v_value_stride);

      /* "sklearn/tree/_tree.pyx":1623
 *                 offset = node_id * value_stride
 * 
 *                 for c from 0 <= c < n_classes[0]:             # <<<<<<<<<<<<<<
 *                     out[i, c] = value[offset + c]
 * 
 */
      __pyx_t_17 = (__pyx_v_n_classes[0]);
      for (__pyx_v_c = 0; __pyx_v_c < __pyx_t_17; __pyx_v_c++) {

        /* "sklearn/tree/_tree.pyx":1624
 * 
 *                 for c from 0 <= c < n_classes[0]:
 *                     out[i, c] = value[offset + c]             # <<<<<<<<<<<<<<
 * 
 *             return out
 */
        __pyx_t_18 = __pyx_v_i;
        __pyx_t_19 = __pyx_v_c;
        *__Pyx_BufPtrStrided2d(__pyx_t_5numpy_float64_t *, __pyx_pybuffernd_out.rcbuffer->pybuffer.buf, __pyx_t_18, __pyx_pybuffernd_out.diminfo[0].strides, __pyx_t_19, __pyx_pybuffernd_out.diminfo[1].strides) = (__pyx_v_value[(__pyx_v_offset + __pyx_v_c)]);
      }
    }

    /* "sklearn/tree/_tree.pyx":1626
 *                     out[i, c] = value[offset + c]
 * 
 *             return out             # <<<<<<<<<<<<<<
 * 
 *         else: # n_outputs > 1
 */
    __Pyx_XDECREF(__pyx_r);
    __Pyx_INCREF(((PyObject *)__pyx_v_out));
    __pyx_r = ((PyObject *)__pyx_v_out);
    goto __pyx_L0;
    goto __pyx_L3;
  }
  /*else*/ {

    /* "sklearn/tree/_tree.pyx":1629
 * 
 *         else: # n_outputs > 1
 *             out_multi = np.zeros((n_samples,             # <<<<<<<<<<<<<<
 *                                   n_outputs,
 *                                   max_n_classes), dtype=np.float64)
 */
    __pyx_t_9 = __Pyx_GetModuleGlobalName(__pyx_n_s__np); if (unlikely(!__pyx_t_9)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1629; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    __Pyx_GOTREF(__pyx_t_9);
    __pyx_t_8 = __Pyx_PyObject_GetAttrStr(__pyx_t_9, __pyx_n_s__zeros); if (unlikely(!__pyx_t_8)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1629; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    __Pyx_GOTREF(__pyx_t_8);
    __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
    __pyx_t_9 = __Pyx_PyInt_to_py_Py_intptr_t(__pyx_v_n_samples); if (unlikely(!__pyx_t_9)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1629; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    __Pyx_GOTREF(__pyx_t_9);

    /* "sklearn/tree/_tree.pyx":1630
 *         else: # n_outputs > 1
 *             out_multi = np.zeros((n_samples,
 *                                   n_outputs,             # <<<<<<<<<<<<<<
 *                                   max_n_classes), dtype=np.float64)
 * 
 */
    __pyx_t_2 = __Pyx_PyInt_to_py_Py_intptr_t(__pyx_v_n_outputs); if (unlikely(!__pyx_t_2)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1630; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    __Pyx_GOTREF(__pyx_t_2);

    /* "sklearn/tree/_tree.pyx":1631
 *             out_multi = np.zeros((n_samples,
 *                                   n_outputs,
 *                                   max_n_classes), dtype=np.float64)             # <<<<<<<<<<<<<<
 * 
 *             for i from 0 <= i < n_samples:
 */
    __pyx_t_3 = __Pyx_PyInt_to_py_Py_intptr_t(__pyx_v_max_n_classes); if (unlikely(!__pyx_t_3)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1631; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_1 = PyTuple_New(3); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1629; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    __Pyx_GOTREF(__pyx_t_1);
    PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_t_9);
    __Pyx_GIVEREF(__pyx_t_9);
    PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_t_2);
    __Pyx_GIVEREF(__pyx_t_2);
    PyTuple_SET_ITEM(__pyx_t_1, 2, __pyx_t_3);
    __Pyx_GIVEREF(__pyx_t_3);
    __pyx_t_9 = 0;
    __pyx_t_2 = 0;
    __pyx_t_3 = 0;
    __pyx_t_3 = PyTuple_New(1); if (unlikely(!__pyx_t_3)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1629; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    __Pyx_GOTREF(__pyx_t_3);
    PyTuple_SET_ITEM(__pyx_t_3, 0, ((PyObject *)__pyx_t_1));
    __Pyx_GIVEREF(((PyObject *)__pyx_t_1));
    __pyx_t_1 = 0;

    /* "sklearn/tree/_tree.pyx":1629
 * 
 *         else: # n_outputs > 1
 *             out_multi = np.zeros((n_samples,             # <<<<<<<<<<<<<<
 *                                   n_outputs,
 *                                   max_n_classes), dtype=np.float64)
 */
    __pyx_t_1 = PyDict_New(); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1629; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    __Pyx_GOTREF(((PyObject *)__pyx_t_1));

    /* "sklearn/tree/_tree.pyx":1631
 *             out_multi = np.zeros((n_samples,
 *                                   n_outputs,
 *                                   max_n_classes), dtype=np.float64)             # <<<<<<<<<<<<<<
 * 
 *             for i from 0 <= i < n_samples:
 */
    __pyx_t_2 = __Pyx_GetModuleGlobalName(__pyx_n_s__np); if (unlikely(!__pyx_t_2)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1631; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_9 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s__float64); if (unlikely(!__pyx_t_9)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1631; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    __Pyx_GOTREF(__pyx_t_9);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    if (PyDict_SetItem(__pyx_t_1, ((PyObject *)__pyx_n_s__dtype), __pyx_t_9) < 0) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1629; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
    __pyx_t_9 = PyObject_Call(__pyx_t_8, ((PyObject *)__pyx_t_3), ((PyObject *)__pyx_t_1)); if (unlikely(!__pyx_t_9)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1629; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    __Pyx_GOTREF(__pyx_t_9);
    __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
    __Pyx_DECREF(((PyObject *)__pyx_t_3)); __pyx_t_3 = 0;
    __Pyx_DECREF(((PyObject *)__pyx_t_1)); __pyx_t_1 = 0;
    if (!(likely(((__pyx_t_9) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_9, __pyx_ptype_5numpy_ndarray))))) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1629; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    __pyx_t_20 = ((PyArrayObject *)__pyx_t_9);
    {
      __Pyx_BufFmt_StackElem __pyx_stack[1];
      __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_out_multi.rcbuffer->pybuffer);
      __pyx_t_11 = __Pyx_GetBufferAndValidate(&__pyx_pybuffernd_out_multi.rcbuffer->pybuffer, (PyObject*)__pyx_t_20, &__Pyx_TypeInfo_nn___pyx_t_5numpy_float64_t, PyBUF_FORMAT| PyBUF_STRIDES| PyBUF_WRITABLE, 3, 0, __pyx_stack);
      if (unlikely(__pyx_t_11 < 0)) {
        PyErr_Fetch(&__pyx_t_14, &__pyx_t_13, &__pyx_t_12);
        if (unlikely(__Pyx_GetBufferAndValidate(&__pyx_pybuffernd_out_multi.rcbuffer->pybuffer, (PyObject*)__pyx_v_out_multi, &__Pyx_TypeInfo_nn___pyx_t_5numpy_float64_t, PyBUF_FORMAT| PyBUF_STRIDES| PyBUF_WRITABLE, 3, 0, __pyx_stack) == -1)) {
          Py_XDECREF(__pyx_t_14); Py_XDECREF(__pyx_t_13); Py_XDECREF(__pyx_t_12);
          __Pyx_RaiseBufferFallbackError();
        } else {
          PyErr_Restore(__pyx_t_14, __pyx_t_13, __pyx_t_12);
        }
      }
      __pyx_pybuffernd_out_multi.diminfo[0].strides = __pyx_pybuffernd_out_multi.rcbuffer->pybuffer.strides[0]; __pyx_pybuffernd_out_multi.diminfo[0].shape = __pyx_pybuffernd_out_multi.rcbuffer->pybuffer.shape[0]; __pyx_pybuffernd_out_multi.diminfo[1].strides = __pyx_pybuffernd_out_multi.rcbuffer->pybuffer.strides[1]; __pyx_pybuffernd_out_multi.diminfo[1].shape = __pyx_pybuffernd_out_multi.rcbuffer->pybuffer.shape[1]; __pyx_pybuffernd_out_multi.diminfo[2].strides = __pyx_pybuffernd_out_multi.rcbuffer->pybuffer.strides[2]; __pyx_pybuffernd_out_multi.diminfo[2].shape = __pyx_pybuffernd_out_multi.rcbuffer->pybuffer.shape[2];
      if (unlikely(__pyx_t_11 < 0)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1629; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    }
    __pyx_t_20 = 0;
    __pyx_v_out_multi = ((PyArrayObject *)__pyx_t_9);
    __pyx_t_9 = 0;

    /* "sklearn/tree/_tree.pyx":1633
 *                                   max_n_classes), dtype=np.float64)
 * 
 *             for i from 0 <= i < n_samples:             # <<<<<<<<<<<<<<
 *                 node_id = 0
 * 
 */
    __pyx_t_6 = __pyx_v_n_samples;
    for (__pyx_v_i = 0; __pyx_v_i < __pyx_t_6; __pyx_v_i++) {

      /* "sklearn/tree/_tree.pyx":1634
 * 
 *             for i from 0 <= i < n_samples:
 *                 node_id = 0             # <<<<<<<<<<<<<<
 * 
 *                 # While node_id not a leaf
 */
      __pyx_v_node_id = 0;

      /* "sklearn/tree/_tree.pyx":1637
 * 
 *                 # While node_id not a leaf
 *                 while children_left[node_id] != _TREE_LEAF:             # <<<<<<<<<<<<<<
 *                     # ... and children_right[node_id] != _TREE_LEAF:
 *                     if X[i, feature[node_id]] <= threshold[node_id]:
 */
      while (1) {
        __pyx_t_7 = (((__pyx_v_children_left[__pyx_v_node_id]) != __pyx_v_7sklearn_4tree_5_tree__TREE_LEAF) != 0);
        if (!__pyx_t_7) break;

        /* "sklearn/tree/_tree.pyx":1639
 *                 while children_left[node_id] != _TREE_LEAF:
 *                     # ... and children_right[node_id] != _TREE_LEAF:
 *                     if X[i, feature[node_id]] <= threshold[node_id]:             # <<<<<<<<<<<<<<
 *                         node_id = children_left[node_id]
 *                     else:
 */
        __pyx_t_17 = __pyx_v_i;
        __pyx_t_21 = (__pyx_v_feature[__pyx_v_node_id]);
        __pyx_t_7 = (((*__Pyx_BufPtrStrided2d(__pyx_t_7sklearn_4tree_5_tree_DTYPE_t *, __pyx_pybuffernd_X.rcbuffer->pybuffer.buf, __pyx_t_17, __pyx_pybuffernd_X.diminfo[0].strides, __pyx_t_21, __pyx_pybuffernd_X.diminfo[1].strides)) <= (__pyx_v_threshold[__pyx_v_node_id])) != 0);
        if (__pyx_t_7) {

          /* "sklearn/tree/_tree.pyx":1640
 *                     # ... and children_right[node_id] != _TREE_LEAF:
 *                     if X[i, feature[node_id]] <= threshold[node_id]:
 *                         node_id = children_left[node_id]             # <<<<<<<<<<<<<<
 *                     else:
 *                         node_id = children_right[node_id]
 */
          __pyx_v_node_id = (__pyx_v_children_left[__pyx_v_node_id]);
          goto __pyx_L15;
        }
        /*else*/ {

          /* "sklearn/tree/_tree.pyx":1642
 *                         node_id = children_left[node_id]
 *                     else:
 *                         node_id = children_right[node_id]             # <<<<<<<<<<<<<<
 * 
 *                 offset = node_id * value_stride
 */
          __pyx_v_node_id = (__pyx_v_children_right[__pyx_v_node_id]);
        }
        __pyx_L15:;
      }

      /* "sklearn/tree/_tree.pyx":1644
 *                         node_id = children_right[node_id]
 * 
 *                 offset = node_id * value_stride             # <<<<<<<<<<<<<<
 * 
 *                 for k from 0 <= k < n_outputs:
 */
      __pyx_v_offset = (__pyx_v_node_id * __pyx_v_value_stride);

      /* "sklearn/tree/_tree.pyx":1646
 *                 offset = node_id * value_stride
 * 
 *                 for k from 0 <= k < n_outputs:             # <<<<<<<<<<<<<<
 *                     for c from 0 <= c < n_classes[k]:
 *                         out_multi[i, k, c] = value[offset + c]
 */
      __pyx_t_22 = __pyx_v_n_outputs;
      for (__pyx_v_k = 0; __pyx_v_k < __pyx_t_22; __pyx_v_k++) {

        /* "sklearn/tree/_tree.pyx":1647
 * 
 *                 for k from 0 <= k < n_outputs:
 *                     for c from 0 <= c < n_classes[k]:             # <<<<<<<<<<<<<<
 *                         out_multi[i, k, c] = value[offset + c]
 *                     offset += max_n_classes
 */
        __pyx_t_23 = (__pyx_v_n_classes[__pyx_v_k]);
        for (__pyx_v_c = 0; __pyx_v_c < __pyx_t_23; __pyx_v_c++) {

          /* "sklearn/tree/_tree.pyx":1648
 *                 for k from 0 <= k < n_outputs:
 *                     for c from 0 <= c < n_classes[k]:
 *                         out_multi[i, k, c] = value[offset + c]             # <<<<<<<<<<<<<<
 *                     offset += max_n_classes
 * 
 */
          __pyx_t_24 = __pyx_v_i;
          __pyx_t_25 = __pyx_v_k;
          __pyx_t_26 = __pyx_v_c;
          *__Pyx_BufPtrStrided3d(__pyx_t_5numpy_float64_t *, __pyx_pybuffernd_out_multi.rcbuffer->pybuffer.buf, __pyx_t_24, __pyx_pybuffernd_out_multi.diminfo[0].strides, __pyx_t_25, __pyx_pybuffernd_out_multi.diminfo[1].strides, __pyx_t_26, __pyx_pybuffernd_out_multi.diminfo[2].strides) = (__pyx_v_value[(__pyx_v_offset + __pyx_v_c)]);
        }

        /* "sklearn/tree/_tree.pyx":1649
 *                     for c from 0 <= c < n_classes[k]:
 *                         out_multi[i, k, c] = value[offset + c]
 *                     offset += max_n_classes             # <<<<<<<<<<<<<<
 * 
 *             return out_multi
 */
        __pyx_v_offset = (__pyx_v_offset + __pyx_v_max_n_classes);
      }
    }

    /* "sklearn/tree/_tree.pyx":1651
 *                     offset += max_n_classes
 * 
 *             return out_multi             # <<<<<<<<<<<<<<
 * 
 *     cpdef apply(self, np.ndarray[DTYPE_t, ndim=2] X):
 */
    __Pyx_XDECREF(__pyx_r);
    __Pyx_INCREF(((PyObject *)__pyx_v_out_multi));
    __pyx_r = ((PyObject *)__pyx_v_out_multi);
    goto __pyx_L0;
  }
  __pyx_L3:;

  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_XDECREF(__pyx_t_9);
  { PyObject *__pyx_type, *__pyx_value, *__pyx_tb;
    __Pyx_ErrFetch(&__pyx_type, &__pyx_value, &__pyx_tb);
    __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_X.rcbuffer->pybuffer);
    __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_out.rcbuffer->pybuffer);
    __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_out_multi.rcbuffer->pybuffer);
  __Pyx_ErrRestore(__pyx_type, __pyx_value, __pyx_tb);}
  __Pyx_AddTraceback("sklearn.tree._tree.Tree.predict", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  goto __pyx_L2;
  __pyx_L0:;
  __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_X.rcbuffer->pybuffer);
  __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_out.rcbuffer->pybuffer);
  __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_out_multi.rcbuffer->pybuffer);
  __pyx_L2:;
  __Pyx_XDECREF((PyObject *)__pyx_v_out);
  __Pyx_XDECREF((PyObject *)__pyx_v_out_multi);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_7sklearn_4tree_5_tree_4Tree_13predict(PyObject *__pyx_v_self, PyObject *__pyx_v_X); /*proto*/
static char __pyx_doc_7sklearn_4tree_5_tree_4Tree_12predict[] = "Predict target for X.";
static PyObject *__pyx_pw_7sklearn_4tree_5_tree_4Tree_13predict(PyObject *__pyx_v_self, PyObject *__pyx_v_X) {
  CYTHON_UNUSED int __pyx_lineno = 0;
  CYTHON_UNUSED const char *__pyx_filename = NULL;
  CYTHON_UNUSED int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("predict (wrapper)", 0);
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_X), __pyx_ptype_5numpy_ndarray, 1, "X", 0))) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1584; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __pyx_r = __pyx_pf_7sklearn_4tree_5_tree_4Tree_12predict(((struct __pyx_obj_7sklearn_4tree_5_tree_Tree *)__pyx_v_self), ((PyArrayObject *)__pyx_v_X));
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "sklearn/tree/_tree.pyx":1584
 *         free(stack)
 * 
 *     cpdef predict(self, np.ndarray[DTYPE_t, ndim=2] X):             # <<<<<<<<<<<<<<
 *         """Predict target for X."""
 *         cdef SIZE_t* children_left = self.children_left
 */

static PyObject *__pyx_pf_7sklearn_4tree_5_tree_4Tree_12predict(struct __pyx_obj_7sklearn_4tree_5_tree_Tree *__pyx_v_self, PyArrayObject *__pyx_v_X) {
  __Pyx_LocalBuf_ND __pyx_pybuffernd_X;
  __Pyx_Buffer __pyx_pybuffer_X;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("predict", 0);
  __pyx_pybuffer_X.pybuffer.buf = NULL;
  __pyx_pybuffer_X.refcount = 0;
  __pyx_pybuffernd_X.data = NULL;
  __pyx_pybuffernd_X.rcbuffer = &__pyx_pybuffer_X;
  {
    __Pyx_BufFmt_StackElem __pyx_stack[1];
    if (unlikely(__Pyx_GetBufferAndValidate(&__pyx_pybuffernd_X.rcbuffer->pybuffer, (PyObject*)__pyx_v_X, &__Pyx_TypeInfo_nn___pyx_t_7sklearn_4tree_5_tree_DTYPE_t, PyBUF_FORMAT| PyBUF_STRIDES, 2, 0, __pyx_stack) == -1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1584; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  }
  __pyx_pybuffernd_X.diminfo[0].strides = __pyx_pybuffernd_X.rcbuffer->pybuffer.strides[0]; __pyx_pybuffernd_X.diminfo[0].shape = __pyx_pybuffernd_X.rcbuffer->pybuffer.shape[0]; __pyx_pybuffernd_X.diminfo[1].strides = __pyx_pybuffernd_X.rcbuffer->pybuffer.strides[1]; __pyx_pybuffernd_X.diminfo[1].shape = __pyx_pybuffernd_X.rcbuffer->pybuffer.shape[1];
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = ((struct __pyx_vtabstruct_7sklearn_4tree_5_tree_Tree *)__pyx_v_self->__pyx_vtab)->predict(__pyx_v_self, ((PyArrayObject *)__pyx_v_X), 1); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1584; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  { PyObject *__pyx_type, *__pyx_value, *__pyx_tb;
    __Pyx_ErrFetch(&__pyx_type, &__pyx_value, &__pyx_tb);
    __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_X.rcbuffer->pybuffer);
  __Pyx_ErrRestore(__pyx_type, __pyx_value, __pyx_tb);}
  __Pyx_AddTraceback("sklearn.tree._tree.Tree.predict", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  goto __pyx_L2;
  __pyx_L0:;
  __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_X.rcbuffer->pybuffer);
  __pyx_L2:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "sklearn/tree/_tree.pyx":1653
 *             return out_multi
 * 
 *     cpdef apply(self, np.ndarray[DTYPE_t, ndim=2] X):             # <<<<<<<<<<<<<<
 *         """Finds the terminal region (=leaf node) for each sample in X."""
 *         cdef SIZE_t* children_left = self.children_left
 */

static PyObject *__pyx_pw_7sklearn_4tree_5_tree_4Tree_15apply(PyObject *__pyx_v_self, PyObject *__pyx_v_X); /*proto*/
static PyObject *__pyx_f_7sklearn_4tree_5_tree_4Tree_apply(struct __pyx_obj_7sklearn_4tree_5_tree_Tree *__pyx_v_self, PyArrayObject *__pyx_v_X, int __pyx_skip_dispatch) {
  __pyx_t_7sklearn_4tree_5_tree_SIZE_t *__pyx_v_children_left;
  __pyx_t_7sklearn_4tree_5_tree_SIZE_t *__pyx_v_children_right;
  __pyx_t_7sklearn_4tree_5_tree_SIZE_t *__pyx_v_feature;
  double *__pyx_v_threshold;
  __pyx_t_7sklearn_4tree_5_tree_SIZE_t __pyx_v_n_samples;
  __pyx_t_7sklearn_4tree_5_tree_SIZE_t __pyx_v_node_id;
  __pyx_t_7sklearn_4tree_5_tree_SIZE_t __pyx_v_i;
  PyArrayObject *__pyx_v_out = 0;
  __Pyx_LocalBuf_ND __pyx_pybuffernd_X;
  __Pyx_Buffer __pyx_pybuffer_X;
  __Pyx_LocalBuf_ND __pyx_pybuffernd_out;
  __Pyx_Buffer __pyx_pybuffer_out;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  __pyx_t_7sklearn_4tree_5_tree_SIZE_t *__pyx_t_4;
  double *__pyx_t_5;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  PyArrayObject *__pyx_t_8 = NULL;
  int __pyx_t_9;
  PyObject *__pyx_t_10 = NULL;
  PyObject *__pyx_t_11 = NULL;
  PyObject *__pyx_t_12 = NULL;
  __pyx_t_7sklearn_4tree_5_tree_SIZE_t __pyx_t_13;
  int __pyx_t_14;
  __pyx_t_7sklearn_4tree_5_tree_SIZE_t __pyx_t_15;
  __pyx_t_7sklearn_4tree_5_tree_SIZE_t __pyx_t_16;
  __pyx_t_7sklearn_4tree_5_tree_SIZE_t __pyx_t_17;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("apply", 0);
  __pyx_pybuffer_out.pybuffer.buf = NULL;
  __pyx_pybuffer_out.refcount = 0;
  __pyx_pybuffernd_out.data = NULL;
  __pyx_pybuffernd_out.rcbuffer = &__pyx_pybuffer_out;
  __pyx_pybuffer_X.pybuffer.buf = NULL;
  __pyx_pybuffer_X.refcount = 0;
  __pyx_pybuffernd_X.data = NULL;
  __pyx_pybuffernd_X.rcbuffer = &__pyx_pybuffer_X;
  {
    __Pyx_BufFmt_StackElem __pyx_stack[1];
    if (unlikely(__Pyx_GetBufferAndValidate(&__pyx_pybuffernd_X.rcbuffer->pybuffer, (PyObject*)__pyx_v_X, &__Pyx_TypeInfo_nn___pyx_t_7sklearn_4tree_5_tree_DTYPE_t, PyBUF_FORMAT| PyBUF_STRIDES, 2, 0, __pyx_stack) == -1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1653; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  }
  __pyx_pybuffernd_X.diminfo[0].strides = __pyx_pybuffernd_X.rcbuffer->pybuffer.strides[0]; __pyx_pybuffernd_X.diminfo[0].shape = __pyx_pybuffernd_X.rcbuffer->pybuffer.shape[0]; __pyx_pybuffernd_X.diminfo[1].strides = __pyx_pybuffernd_X.rcbuffer->pybuffer.strides[1]; __pyx_pybuffernd_X.diminfo[1].shape = __pyx_pybuffernd_X.rcbuffer->pybuffer.shape[1];
  /* Check if called by wrapper */
  if (unlikely(__pyx_skip_dispatch)) ;
  /* Check if overridden in Python */
  else if (unlikely(Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0)) {
    __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s__apply); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1653; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    __Pyx_GOTREF(__pyx_t_1);
    if (!PyCFunction_Check(__pyx_t_1) || (PyCFunction_GET_FUNCTION(__pyx_t_1) != (PyCFunction)__pyx_pw_7sklearn_4tree_5_tree_4Tree_15apply)) {
      __Pyx_XDECREF(__pyx_r);
      __pyx_t_2 = PyTuple_New(1); if (unlikely(!__pyx_t_2)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1653; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_INCREF(((PyObject *)__pyx_v_X));
      PyTuple_SET_ITEM(__pyx_t_2, 0, ((PyObject *)__pyx_v_X));
      __Pyx_GIVEREF(((PyObject *)__pyx_v_X));
      __pyx_t_3 = PyObject_Call(__pyx_t_1, ((PyObject *)__pyx_t_2), NULL); if (unlikely(!__pyx_t_3)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1653; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_DECREF(((PyObject *)__pyx_t_2)); __pyx_t_2 = 0;
      __pyx_r = __pyx_t_3;
      __pyx_t_3 = 0;
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      goto __pyx_L0;
    }
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  }

  /* "sklearn/tree/_tree.pyx":1655
 *     cpdef apply(self, np.ndarray[DTYPE_t, ndim=2] X):
 *         """Finds the terminal region (=leaf node) for each sample in X."""
 *         cdef SIZE_t* children_left = self.children_left             # <<<<<<<<<<<<<<
 *         cdef SIZE_t* children_right = self.children_right
 *         cdef SIZE_t* feature = self.feature
 */
  __pyx_t_4 = __pyx_v_self->children_left;
  __pyx_v_children_left = __pyx_t_4;

  /* "sklearn/tree/_tree.pyx":1656
 *         """Finds the terminal region (=leaf node) for each sample in X."""
 *         cdef SIZE_t* children_left = self.children_left
 *         cdef SIZE_t* children_right = self.children_right             # <<<<<<<<<<<<<<
 *         cdef SIZE_t* feature = self.feature
 *         cdef double* threshold = self.threshold
 */
  __pyx_t_4 = __pyx_v_self->children_right;
  __pyx_v_children_right = __pyx_t_4;

  /* "sklearn/tree/_tree.pyx":1657
 *         cdef SIZE_t* children_left = self.children_left
 *         cdef SIZE_t* children_right = self.children_right
 *         cdef SIZE_t* feature = self.feature             # <<<<<<<<<<<<<<
 *         cdef double* threshold = self.threshold
 * 
 */
  __pyx_t_4 = __pyx_v_self->feature;
  __pyx_v_feature = __pyx_t_4;

  /* "sklearn/tree/_tree.pyx":1658
 *         cdef SIZE_t* children_right = self.children_right
 *         cdef SIZE_t* feature = self.feature
 *         cdef double* threshold = self.threshold             # <<<<<<<<<<<<<<
 * 
 *         cdef SIZE_t n_samples = X.shape[0]
 */
  __pyx_t_5 = __pyx_v_self->threshold;
  __pyx_v_threshold = __pyx_t_5;

  /* "sklearn/tree/_tree.pyx":1660
 *         cdef double* threshold = self.threshold
 * 
 *         cdef SIZE_t n_samples = X.shape[0]             # <<<<<<<<<<<<<<
 *         cdef SIZE_t node_id = 0
 *         cdef SIZE_t i = 0
 */
  __pyx_v_n_samples = (__pyx_v_X->dimensions[0]);

  /* "sklearn/tree/_tree.pyx":1661
 * 
 *         cdef SIZE_t n_samples = X.shape[0]
 *         cdef SIZE_t node_id = 0             # <<<<<<<<<<<<<<
 *         cdef SIZE_t i = 0
 * 
 */
  __pyx_v_node_id = 0;

  /* "sklearn/tree/_tree.pyx":1662
 *         cdef SIZE_t n_samples = X.shape[0]
 *         cdef SIZE_t node_id = 0
 *         cdef SIZE_t i = 0             # <<<<<<<<<<<<<<
 * 
 *         cdef np.ndarray[np.int32_t, ndim=1] out
 */
  __pyx_v_i = 0;

  /* "sklearn/tree/_tree.pyx":1665
 * 
 *         cdef np.ndarray[np.int32_t, ndim=1] out
 *         out = np.zeros((n_samples,), dtype=np.int32)             # <<<<<<<<<<<<<<
 * 
 *         for i from 0 <= i < n_samples:
 */
  __pyx_t_1 = __Pyx_GetModuleGlobalName(__pyx_n_s__np); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1665; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s__zeros); if (unlikely(!__pyx_t_3)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1665; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_PyInt_to_py_Py_intptr_t(__pyx_v_n_samples); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1665; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = PyTuple_New(1); if (unlikely(!__pyx_t_2)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1665; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_2);
  PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __pyx_t_1 = 0;
  __pyx_t_1 = PyTuple_New(1); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1665; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_1);
  PyTuple_SET_ITEM(__pyx_t_1, 0, ((PyObject *)__pyx_t_2));
  __Pyx_GIVEREF(((PyObject *)__pyx_t_2));
  __pyx_t_2 = 0;
  __pyx_t_2 = PyDict_New(); if (unlikely(!__pyx_t_2)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1665; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(((PyObject *)__pyx_t_2));
  __pyx_t_6 = __Pyx_GetModuleGlobalName(__pyx_n_s__np); if (unlikely(!__pyx_t_6)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1665; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_6);
  __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_t_6, __pyx_n_s__int32); if (unlikely(!__pyx_t_7)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1665; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  if (PyDict_SetItem(__pyx_t_2, ((PyObject *)__pyx_n_s__dtype), __pyx_t_7) < 0) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1665; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  __pyx_t_7 = PyObject_Call(__pyx_t_3, ((PyObject *)__pyx_t_1), ((PyObject *)__pyx_t_2)); if (unlikely(!__pyx_t_7)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1665; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(((PyObject *)__pyx_t_1)); __pyx_t_1 = 0;
  __Pyx_DECREF(((PyObject *)__pyx_t_2)); __pyx_t_2 = 0;
  if (!(likely(((__pyx_t_7) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_7, __pyx_ptype_5numpy_ndarray))))) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1665; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __pyx_t_8 = ((PyArrayObject *)__pyx_t_7);
  {
    __Pyx_BufFmt_StackElem __pyx_stack[1];
    __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_out.rcbuffer->pybuffer);
    __pyx_t_9 = __Pyx_GetBufferAndValidate(&__pyx_pybuffernd_out.rcbuffer->pybuffer, (PyObject*)__pyx_t_8, &__Pyx_TypeInfo_nn___pyx_t_5numpy_int32_t, PyBUF_FORMAT| PyBUF_STRIDES| PyBUF_WRITABLE, 1, 0, __pyx_stack);
    if (unlikely(__pyx_t_9 < 0)) {
      PyErr_Fetch(&__pyx_t_10, &__pyx_t_11, &__pyx_t_12);
      if (unlikely(__Pyx_GetBufferAndValidate(&__pyx_pybuffernd_out.rcbuffer->pybuffer, (PyObject*)__pyx_v_out, &__Pyx_TypeInfo_nn___pyx_t_5numpy_int32_t, PyBUF_FORMAT| PyBUF_STRIDES| PyBUF_WRITABLE, 1, 0, __pyx_stack) == -1)) {
        Py_XDECREF(__pyx_t_10); Py_XDECREF(__pyx_t_11); Py_XDECREF(__pyx_t_12);
        __Pyx_RaiseBufferFallbackError();
      } else {
        PyErr_Restore(__pyx_t_10, __pyx_t_11, __pyx_t_12);
      }
    }
    __pyx_pybuffernd_out.diminfo[0].strides = __pyx_pybuffernd_out.rcbuffer->pybuffer.strides[0]; __pyx_pybuffernd_out.diminfo[0].shape = __pyx_pybuffernd_out.rcbuffer->pybuffer.shape[0];
    if (unlikely(__pyx_t_9 < 0)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1665; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  }
  __pyx_t_8 = 0;
  __pyx_v_out = ((PyArrayObject *)__pyx_t_7);
  __pyx_t_7 = 0;

  /* "sklearn/tree/_tree.pyx":1667
 *         out = np.zeros((n_samples,), dtype=np.int32)
 * 
 *         for i from 0 <= i < n_samples:             # <<<<<<<<<<<<<<
 *             node_id = 0
 * 
 */
  __pyx_t_13 = __pyx_v_n_samples;
  for (__pyx_v_i = 0; __pyx_v_i < __pyx_t_13; __pyx_v_i++) {

    /* "sklearn/tree/_tree.pyx":1668
 * 
 *         for i from 0 <= i < n_samples:
 *             node_id = 0             # <<<<<<<<<<<<<<
 * 
 *             # While node_id not a leaf
 */
    __pyx_v_node_id = 0;

    /* "sklearn/tree/_tree.pyx":1671
 * 
 *             # While node_id not a leaf
 *             while children_left[node_id] != _TREE_LEAF:             # <<<<<<<<<<<<<<
 *                 # ... and children_right[node_id] != _TREE_LEAF:
 *                 if X[i, feature[node_id]] <= threshold[node_id]:
 */
    while (1) {
      __pyx_t_14 = (((__pyx_v_children_left[__pyx_v_node_id]) != __pyx_v_7sklearn_4tree_5_tree__TREE_LEAF) != 0);
      if (!__pyx_t_14) break;

      /* "sklearn/tree/_tree.pyx":1673
 *             while children_left[node_id] != _TREE_LEAF:
 *                 # ... and children_right[node_id] != _TREE_LEAF:
 *                 if X[i, feature[node_id]] <= threshold[node_id]:             # <<<<<<<<<<<<<<
 *                     node_id = children_left[node_id]
 *                 else:
 */
      __pyx_t_15 = __pyx_v_i;
      __pyx_t_16 = (__pyx_v_feature[__pyx_v_node_id]);
      __pyx_t_14 = (((*__Pyx_BufPtrStrided2d(__pyx_t_7sklearn_4tree_5_tree_DTYPE_t *, __pyx_pybuffernd_X.rcbuffer->pybuffer.buf, __pyx_t_15, __pyx_pybuffernd_X.diminfo[0].strides, __pyx_t_16, __pyx_pybuffernd_X.diminfo[1].strides)) <= (__pyx_v_threshold[__pyx_v_node_id])) != 0);
      if (__pyx_t_14) {

        /* "sklearn/tree/_tree.pyx":1674
 *                 # ... and children_right[node_id] != _TREE_LEAF:
 *                 if X[i, feature[node_id]] <= threshold[node_id]:
 *                     node_id = children_left[node_id]             # <<<<<<<<<<<<<<
 *                 else:
 *                     node_id = children_right[node_id]
 */
        __pyx_v_node_id = (__pyx_v_children_left[__pyx_v_node_id]);
        goto __pyx_L7;
      }
      /*else*/ {

        /* "sklearn/tree/_tree.pyx":1676
 *                     node_id = children_left[node_id]
 *                 else:
 *                     node_id = children_right[node_id]             # <<<<<<<<<<<<<<
 * 
 *             out[i] = node_id
 */
        __pyx_v_node_id = (__pyx_v_children_right[__pyx_v_node_id]);
      }
      __pyx_L7:;
    }

    /* "sklearn/tree/_tree.pyx":1678
 *                     node_id = children_right[node_id]
 * 
 *             out[i] = node_id             # <<<<<<<<<<<<<<
 * 
 *         return out
 */
    __pyx_t_17 = __pyx_v_i;
    *__Pyx_BufPtrStrided1d(__pyx_t_5numpy_int32_t *, __pyx_pybuffernd_out.rcbuffer->pybuffer.buf, __pyx_t_17, __pyx_pybuffernd_out.diminfo[0].strides) = __pyx_v_node_id;
  }

  /* "sklearn/tree/_tree.pyx":1680
 *             out[i] = node_id
 * 
 *         return out             # <<<<<<<<<<<<<<
 * 
 *     cpdef compute_feature_importances(self, normalize=True):
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(((PyObject *)__pyx_v_out));
  __pyx_r = ((PyObject *)__pyx_v_out);
  goto __pyx_L0;

  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  { PyObject *__pyx_type, *__pyx_value, *__pyx_tb;
    __Pyx_ErrFetch(&__pyx_type, &__pyx_value, &__pyx_tb);
    __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_X.rcbuffer->pybuffer);
    __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_out.rcbuffer->pybuffer);
  __Pyx_ErrRestore(__pyx_type, __pyx_value, __pyx_tb);}
  __Pyx_AddTraceback("sklearn.tree._tree.Tree.apply", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  goto __pyx_L2;
  __pyx_L0:;
  __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_X.rcbuffer->pybuffer);
  __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_out.rcbuffer->pybuffer);
  __pyx_L2:;
  __Pyx_XDECREF((PyObject *)__pyx_v_out);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_7sklearn_4tree_5_tree_4Tree_15apply(PyObject *__pyx_v_self, PyObject *__pyx_v_X); /*proto*/
static char __pyx_doc_7sklearn_4tree_5_tree_4Tree_14apply[] = "Finds the terminal region (=leaf node) for each sample in X.";
static PyObject *__pyx_pw_7sklearn_4tree_5_tree_4Tree_15apply(PyObject *__pyx_v_self, PyObject *__pyx_v_X) {
  CYTHON_UNUSED int __pyx_lineno = 0;
  CYTHON_UNUSED const char *__pyx_filename = NULL;
  CYTHON_UNUSED int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("apply (wrapper)", 0);
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_X), __pyx_ptype_5numpy_ndarray, 1, "X", 0))) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1653; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __pyx_r = __pyx_pf_7sklearn_4tree_5_tree_4Tree_14apply(((struct __pyx_obj_7sklearn_4tree_5_tree_Tree *)__pyx_v_self), ((PyArrayObject *)__pyx_v_X));
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "sklearn/tree/_tree.pyx":1653
 *             return out_multi
 * 
 *     cpdef apply(self, np.ndarray[DTYPE_t, ndim=2] X):             # <<<<<<<<<<<<<<
 *         """Finds the terminal region (=leaf node) for each sample in X."""
 *         cdef SIZE_t* children_left = self.children_left
 */

static PyObject *__pyx_pf_7sklearn_4tree_5_tree_4Tree_14apply(struct __pyx_obj_7sklearn_4tree_5_tree_Tree *__pyx_v_self, PyArrayObject *__pyx_v_X) {
  __Pyx_LocalBuf_ND __pyx_pybuffernd_X;
  __Pyx_Buffer __pyx_pybuffer_X;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("apply", 0);
  __pyx_pybuffer_X.pybuffer.buf = NULL;
  __pyx_pybuffer_X.refcount = 0;
  __pyx_pybuffernd_X.data = NULL;
  __pyx_pybuffernd_X.rcbuffer = &__pyx_pybuffer_X;
  {
    __Pyx_BufFmt_StackElem __pyx_stack[1];
    if (unlikely(__Pyx_GetBufferAndValidate(&__pyx_pybuffernd_X.rcbuffer->pybuffer, (PyObject*)__pyx_v_X, &__Pyx_TypeInfo_nn___pyx_t_7sklearn_4tree_5_tree_DTYPE_t, PyBUF_FORMAT| PyBUF_STRIDES, 2, 0, __pyx_stack) == -1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1653; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  }
  __pyx_pybuffernd_X.diminfo[0].strides = __pyx_pybuffernd_X.rcbuffer->pybuffer.strides[0]; __pyx_pybuffernd_X.diminfo[0].shape = __pyx_pybuffernd_X.rcbuffer->pybuffer.shape[0]; __pyx_pybuffernd_X.diminfo[1].strides = __pyx_pybuffernd_X.rcbuffer->pybuffer.strides[1]; __pyx_pybuffernd_X.diminfo[1].shape = __pyx_pybuffernd_X.rcbuffer->pybuffer.shape[1];
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = ((struct __pyx_vtabstruct_7sklearn_4tree_5_tree_Tree *)__pyx_v_self->__pyx_vtab)->apply(__pyx_v_self, ((PyArrayObject *)__pyx_v_X), 1); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1653; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  { PyObject *__pyx_type, *__pyx_value, *__pyx_tb;
    __Pyx_ErrFetch(&__pyx_type, &__pyx_value, &__pyx_tb);
    __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_X.rcbuffer->pybuffer);
  __Pyx_ErrRestore(__pyx_type, __pyx_value, __pyx_tb);}
  __Pyx_AddTraceback("sklearn.tree._tree.Tree.apply", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  goto __pyx_L2;
  __pyx_L0:;
  __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_X.rcbuffer->pybuffer);
  __pyx_L2:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "sklearn/tree/_tree.pyx":1682
 *         return out
 * 
 *     cpdef compute_feature_importances(self, normalize=True):             # <<<<<<<<<<<<<<
 *         """Computes the importance of each feature (aka variable)."""
 *         cdef SIZE_t* children_left = self.children_left
 */

static PyObject *__pyx_pw_7sklearn_4tree_5_tree_4Tree_17compute_feature_importances(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyObject *__pyx_f_7sklearn_4tree_5_tree_4Tree_compute_feature_importances(struct __pyx_obj_7sklearn_4tree_5_tree_Tree *__pyx_v_self, int __pyx_skip_dispatch, struct __pyx_opt_args_7sklearn_4tree_5_tree_4Tree_compute_feature_importances *__pyx_optional_args) {
  PyObject *__pyx_v_normalize = __pyx_k_1;
  __pyx_t_7sklearn_4tree_5_tree_SIZE_t *__pyx_v_children_left;
  __pyx_t_7sklearn_4tree_5_tree_SIZE_t *__pyx_v_children_right;
  __pyx_t_7sklearn_4tree_5_tree_SIZE_t *__pyx_v_feature;
  double *__pyx_v_impurity;
  __pyx_t_7sklearn_4tree_5_tree_SIZE_t *__pyx_v_n_node_samples;
  CYTHON_UNUSED __pyx_t_7sklearn_4tree_5_tree_SIZE_t __pyx_v_n_features;
  __pyx_t_7sklearn_4tree_5_tree_SIZE_t __pyx_v_node_count;
  __pyx_t_7sklearn_4tree_5_tree_SIZE_t __pyx_v_n_left;
  __pyx_t_7sklearn_4tree_5_tree_SIZE_t __pyx_v_n_right;
  __pyx_t_7sklearn_4tree_5_tree_SIZE_t __pyx_v_node;
  PyArrayObject *__pyx_v_importances = 0;
  double __pyx_v_normalizer;
  __Pyx_LocalBuf_ND __pyx_pybuffernd_importances;
  __Pyx_Buffer __pyx_pybuffer_importances;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  __pyx_t_7sklearn_4tree_5_tree_SIZE_t *__pyx_t_4;
  double *__pyx_t_5;
  __pyx_t_7sklearn_4tree_5_tree_SIZE_t __pyx_t_6;
  PyArrayObject *__pyx_t_7 = NULL;
  int __pyx_t_8;
  PyObject *__pyx_t_9 = NULL;
  PyObject *__pyx_t_10 = NULL;
  PyObject *__pyx_t_11 = NULL;
  int __pyx_t_12;
  __pyx_t_7sklearn_4tree_5_tree_SIZE_t __pyx_t_13;
  double __pyx_t_14;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("compute_feature_importances", 0);
  if (__pyx_optional_args) {
    if (__pyx_optional_args->__pyx_n > 0) {
      __pyx_v_normalize = __pyx_optional_args->normalize;
    }
  }
  __pyx_pybuffer_importances.pybuffer.buf = NULL;
  __pyx_pybuffer_importances.refcount = 0;
  __pyx_pybuffernd_importances.data = NULL;
  __pyx_pybuffernd_importances.rcbuffer = &__pyx_pybuffer_importances;
  /* Check if called by wrapper */
  if (unlikely(__pyx_skip_dispatch)) ;
  /* Check if overridden in Python */
  else if (unlikely(Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0)) {
    __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_2); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1682; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    __Pyx_GOTREF(__pyx_t_1);
    if (!PyCFunction_Check(__pyx_t_1) || (PyCFunction_GET_FUNCTION(__pyx_t_1) != (PyCFunction)__pyx_pw_7sklearn_4tree_5_tree_4Tree_17compute_feature_importances)) {
      __Pyx_XDECREF(__pyx_r);
      __pyx_t_2 = PyTuple_New(1); if (unlikely(!__pyx_t_2)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1682; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_INCREF(__pyx_v_normalize);
      PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_v_normalize);
      __Pyx_GIVEREF(__pyx_v_normalize);
      __pyx_t_3 = PyObject_Call(__pyx_t_1, ((PyObject *)__pyx_t_2), NULL); if (unlikely(!__pyx_t_3)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1682; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_DECREF(((PyObject *)__pyx_t_2)); __pyx_t_2 = 0;
      __pyx_r = __pyx_t_3;
      __pyx_t_3 = 0;
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      goto __pyx_L0;
    }
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  }

  /* "sklearn/tree/_tree.pyx":1684
 *     cpdef compute_feature_importances(self, normalize=True):
 *         """Computes the importance of each feature (aka variable)."""
 *         cdef SIZE_t* children_left = self.children_left             # <<<<<<<<<<<<<<
 *         cdef SIZE_t* children_right = self.children_right
 *         cdef SIZE_t* feature = self.feature
 */
  __pyx_t_4 = __pyx_v_self->children_left;
  __pyx_v_children_left = __pyx_t_4;

  /* "sklearn/tree/_tree.pyx":1685
 *         """Computes the importance of each feature (aka variable)."""
 *         cdef SIZE_t* children_left = self.children_left
 *         cdef SIZE_t* children_right = self.children_right             # <<<<<<<<<<<<<<
 *         cdef SIZE_t* feature = self.feature
 *         cdef double* impurity = self.impurity
 */
  __pyx_t_4 = __pyx_v_self->children_right;
  __pyx_v_children_right = __pyx_t_4;

  /* "sklearn/tree/_tree.pyx":1686
 *         cdef SIZE_t* children_left = self.children_left
 *         cdef SIZE_t* children_right = self.children_right
 *         cdef SIZE_t* feature = self.feature             # <<<<<<<<<<<<<<
 *         cdef double* impurity = self.impurity
 *         cdef SIZE_t* n_node_samples = self.n_node_samples
 */
  __pyx_t_4 = __pyx_v_self->feature;
  __pyx_v_feature = __pyx_t_4;

  /* "sklearn/tree/_tree.pyx":1687
 *         cdef SIZE_t* children_right = self.children_right
 *         cdef SIZE_t* feature = self.feature
 *         cdef double* impurity = self.impurity             # <<<<<<<<<<<<<<
 *         cdef SIZE_t* n_node_samples = self.n_node_samples
 * 
 */
  __pyx_t_5 = __pyx_v_self->impurity;
  __pyx_v_impurity = __pyx_t_5;

  /* "sklearn/tree/_tree.pyx":1688
 *         cdef SIZE_t* feature = self.feature
 *         cdef double* impurity = self.impurity
 *         cdef SIZE_t* n_node_samples = self.n_node_samples             # <<<<<<<<<<<<<<
 * 
 *         cdef SIZE_t n_features = self.n_features
 */
  __pyx_t_4 = __pyx_v_self->n_node_samples;
  __pyx_v_n_node_samples = __pyx_t_4;

  /* "sklearn/tree/_tree.pyx":1690
 *         cdef SIZE_t* n_node_samples = self.n_node_samples
 * 
 *         cdef SIZE_t n_features = self.n_features             # <<<<<<<<<<<<<<
 *         cdef SIZE_t node_count = self.node_count
 * 
 */
  __pyx_t_6 = __pyx_v_self->n_features;
  __pyx_v_n_features = __pyx_t_6;

  /* "sklearn/tree/_tree.pyx":1691
 * 
 *         cdef SIZE_t n_features = self.n_features
 *         cdef SIZE_t node_count = self.node_count             # <<<<<<<<<<<<<<
 * 
 *         cdef SIZE_t n_left
 */
  __pyx_t_6 = __pyx_v_self->node_count;
  __pyx_v_node_count = __pyx_t_6;

  /* "sklearn/tree/_tree.pyx":1698
 * 
 *         cdef np.ndarray[np.float64_t, ndim=1] importances
 *         importances = np.zeros((self.n_features,))             # <<<<<<<<<<<<<<
 * 
 *         for node from 0 <= node < node_count:
 */
  __pyx_t_1 = __Pyx_GetModuleGlobalName(__pyx_n_s__np); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1698; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s__zeros); if (unlikely(!__pyx_t_3)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1698; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_PyInt_to_py_Py_intptr_t(__pyx_v_self->n_features); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1698; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = PyTuple_New(1); if (unlikely(!__pyx_t_2)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1698; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_2);
  PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __pyx_t_1 = 0;
  __pyx_t_1 = PyTuple_New(1); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1698; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_1);
  PyTuple_SET_ITEM(__pyx_t_1, 0, ((PyObject *)__pyx_t_2));
  __Pyx_GIVEREF(((PyObject *)__pyx_t_2));
  __pyx_t_2 = 0;
  __pyx_t_2 = PyObject_Call(__pyx_t_3, ((PyObject *)__pyx_t_1), NULL); if (unlikely(!__pyx_t_2)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1698; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(((PyObject *)__pyx_t_1)); __pyx_t_1 = 0;
  if (!(likely(((__pyx_t_2) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_2, __pyx_ptype_5numpy_ndarray))))) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1698; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __pyx_t_7 = ((PyArrayObject *)__pyx_t_2);
  {
    __Pyx_BufFmt_StackElem __pyx_stack[1];
    __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_importances.rcbuffer->pybuffer);
    __pyx_t_8 = __Pyx_GetBufferAndValidate(&__pyx_pybuffernd_importances.rcbuffer->pybuffer, (PyObject*)__pyx_t_7, &__Pyx_TypeInfo_nn___pyx_t_5numpy_float64_t, PyBUF_FORMAT| PyBUF_STRIDES| PyBUF_WRITABLE, 1, 0, __pyx_stack);
    if (unlikely(__pyx_t_8 < 0)) {
      PyErr_Fetch(&__pyx_t_9, &__pyx_t_10, &__pyx_t_11);
      if (unlikely(__Pyx_GetBufferAndValidate(&__pyx_pybuffernd_importances.rcbuffer->pybuffer, (PyObject*)__pyx_v_importances, &__Pyx_TypeInfo_nn___pyx_t_5numpy_float64_t, PyBUF_FORMAT| PyBUF_STRIDES| PyBUF_WRITABLE, 1, 0, __pyx_stack) == -1)) {
        Py_XDECREF(__pyx_t_9); Py_XDECREF(__pyx_t_10); Py_XDECREF(__pyx_t_11);
        __Pyx_RaiseBufferFallbackError();
      } else {
        PyErr_Restore(__pyx_t_9, __pyx_t_10, __pyx_t_11);
      }
    }
    __pyx_pybuffernd_importances.diminfo[0].strides = __pyx_pybuffernd_importances.rcbuffer->pybuffer.strides[0]; __pyx_pybuffernd_importances.diminfo[0].shape = __pyx_pybuffernd_importances.rcbuffer->pybuffer.shape[0];
    if (unlikely(__pyx_t_8 < 0)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1698; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  }
  __pyx_t_7 = 0;
  __pyx_v_importances = ((PyArrayObject *)__pyx_t_2);
  __pyx_t_2 = 0;

  /* "sklearn/tree/_tree.pyx":1700
 *         importances = np.zeros((self.n_features,))
 * 
 *         for node from 0 <= node < node_count:             # <<<<<<<<<<<<<<
 *             if children_left[node] != _TREE_LEAF:
 *                 # ... and children_right[node] != _TREE_LEAF:
 */
  __pyx_t_6 = __pyx_v_node_count;
  for (__pyx_v_node = 0; __pyx_v_node < __pyx_t_6; __pyx_v_node++) {

    /* "sklearn/tree/_tree.pyx":1701
 * 
 *         for node from 0 <= node < node_count:
 *             if children_left[node] != _TREE_LEAF:             # <<<<<<<<<<<<<<
 *                 # ... and children_right[node] != _TREE_LEAF:
 *                 n_left = n_node_samples[children_left[node]]
 */
    __pyx_t_12 = (((__pyx_v_children_left[__pyx_v_node]) != __pyx_v_7sklearn_4tree_5_tree__TREE_LEAF) != 0);
    if (__pyx_t_12) {

      /* "sklearn/tree/_tree.pyx":1703
 *             if children_left[node] != _TREE_LEAF:
 *                 # ... and children_right[node] != _TREE_LEAF:
 *                 n_left = n_node_samples[children_left[node]]             # <<<<<<<<<<<<<<
 *                 n_right = n_node_samples[children_right[node]]
 * 
 */
      __pyx_v_n_left = (__pyx_v_n_node_samples[(__pyx_v_children_left[__pyx_v_node])]);

      /* "sklearn/tree/_tree.pyx":1704
 *                 # ... and children_right[node] != _TREE_LEAF:
 *                 n_left = n_node_samples[children_left[node]]
 *                 n_right = n_node_samples[children_right[node]]             # <<<<<<<<<<<<<<
 * 
 *                 importances[feature[node]] += \
 */
      __pyx_v_n_right = (__pyx_v_n_node_samples[(__pyx_v_children_right[__pyx_v_node])]);

      /* "sklearn/tree/_tree.pyx":1706
 *                 n_right = n_node_samples[children_right[node]]
 * 
 *                 importances[feature[node]] += \             # <<<<<<<<<<<<<<
 *                     n_node_samples[node] * impurity[node] \
 *                         - n_left * impurity[children_left[node]] \
 */
      __pyx_t_13 = (__pyx_v_feature[__pyx_v_node]);
      *__Pyx_BufPtrStrided1d(__pyx_t_5numpy_float64_t *, __pyx_pybuffernd_importances.rcbuffer->pybuffer.buf, __pyx_t_13, __pyx_pybuffernd_importances.diminfo[0].strides) += ((((__pyx_v_n_node_samples[__pyx_v_node]) * (__pyx_v_impurity[__pyx_v_node])) - (__pyx_v_n_left * (__pyx_v_impurity[(__pyx_v_children_left[__pyx_v_node])]))) - (__pyx_v_n_right * (__pyx_v_impurity[(__pyx_v_children_right[__pyx_v_node])])));
      goto __pyx_L5;
    }
    __pyx_L5:;
  }

  /* "sklearn/tree/_tree.pyx":1711
 *                         - n_right * impurity[children_right[node]]
 * 
 *         importances = importances / self.n_node_samples[0]             # <<<<<<<<<<<<<<
 *         cdef double normalizer
 * 
 */
  __pyx_t_2 = __Pyx_PyInt_to_py_Py_intptr_t((__pyx_v_self->n_node_samples[0])); if (unlikely(!__pyx_t_2)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1711; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_1 = __Pyx_PyNumber_Divide(((PyObject *)__pyx_v_importances), __pyx_t_2); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1711; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (!(likely(((__pyx_t_1) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_1, __pyx_ptype_5numpy_ndarray))))) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1711; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __pyx_t_7 = ((PyArrayObject *)__pyx_t_1);
  {
    __Pyx_BufFmt_StackElem __pyx_stack[1];
    __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_importances.rcbuffer->pybuffer);
    __pyx_t_8 = __Pyx_GetBufferAndValidate(&__pyx_pybuffernd_importances.rcbuffer->pybuffer, (PyObject*)__pyx_t_7, &__Pyx_TypeInfo_nn___pyx_t_5numpy_float64_t, PyBUF_FORMAT| PyBUF_STRIDES| PyBUF_WRITABLE, 1, 0, __pyx_stack);
    if (unlikely(__pyx_t_8 < 0)) {
      PyErr_Fetch(&__pyx_t_11, &__pyx_t_10, &__pyx_t_9);
      if (unlikely(__Pyx_GetBufferAndValidate(&__pyx_pybuffernd_importances.rcbuffer->pybuffer, (PyObject*)__pyx_v_importances, &__Pyx_TypeInfo_nn___pyx_t_5numpy_float64_t, PyBUF_FORMAT| PyBUF_STRIDES| PyBUF_WRITABLE, 1, 0, __pyx_stack) == -1)) {
        Py_XDECREF(__pyx_t_11); Py_XDECREF(__pyx_t_10); Py_XDECREF(__pyx_t_9);
        __Pyx_RaiseBufferFallbackError();
      } else {
        PyErr_Restore(__pyx_t_11, __pyx_t_10, __pyx_t_9);
      }
    }
    __pyx_pybuffernd_importances.diminfo[0].strides = __pyx_pybuffernd_importances.rcbuffer->pybuffer.strides[0]; __pyx_pybuffernd_importances.diminfo[0].shape = __pyx_pybuffernd_importances.rcbuffer->pybuffer.shape[0];
    if (unlikely(__pyx_t_8 < 0)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1711; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  }
  __pyx_t_7 = 0;
  __Pyx_DECREF(((PyObject *)__pyx_v_importances));
  __pyx_v_importances = ((PyArrayObject *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "sklearn/tree/_tree.pyx":1714
 *         cdef double normalizer
 * 
 *         if normalize:             # <<<<<<<<<<<<<<
 *             normalizer = np.sum(importances)
 * 
 */
  __pyx_t_12 = __Pyx_PyObject_IsTrue(__pyx_v_normalize); if (unlikely(__pyx_t_12 < 0)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1714; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  if (__pyx_t_12) {

    /* "sklearn/tree/_tree.pyx":1715
 * 
 *         if normalize:
 *             normalizer = np.sum(importances)             # <<<<<<<<<<<<<<
 * 
 *             if normalizer > 0.0:
 */
    __pyx_t_1 = __Pyx_GetModuleGlobalName(__pyx_n_s__np); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1715; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s__sum); if (unlikely(!__pyx_t_2)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1715; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_1 = PyTuple_New(1); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1715; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(((PyObject *)__pyx_v_importances));
    PyTuple_SET_ITEM(__pyx_t_1, 0, ((PyObject *)__pyx_v_importances));
    __Pyx_GIVEREF(((PyObject *)__pyx_v_importances));
    __pyx_t_3 = PyObject_Call(__pyx_t_2, ((PyObject *)__pyx_t_1), NULL); if (unlikely(!__pyx_t_3)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1715; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF(((PyObject *)__pyx_t_1)); __pyx_t_1 = 0;
    __pyx_t_14 = __pyx_PyFloat_AsDouble(__pyx_t_3); if (unlikely((__pyx_t_14 == (double)-1) && PyErr_Occurred())) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1715; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_v_normalizer = __pyx_t_14;

    /* "sklearn/tree/_tree.pyx":1717
 *             normalizer = np.sum(importances)
 * 
 *             if normalizer > 0.0:             # <<<<<<<<<<<<<<
 *                 # Avoid dividing by zero (e.g., when root is pure)
 *                 importances /= normalizer
 */
    __pyx_t_12 = ((__pyx_v_normalizer > 0.0) != 0);
    if (__pyx_t_12) {

      /* "sklearn/tree/_tree.pyx":1719
 *             if normalizer > 0.0:
 *                 # Avoid dividing by zero (e.g., when root is pure)
 *                 importances /= normalizer             # <<<<<<<<<<<<<<
 * 
 *         return importances
 */
      __pyx_t_3 = PyFloat_FromDouble(__pyx_v_normalizer); if (unlikely(!__pyx_t_3)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1719; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
      __Pyx_GOTREF(__pyx_t_3);
      __pyx_t_1 = __Pyx_PyNumber_InPlaceDivide(((PyObject *)__pyx_v_importances), __pyx_t_3); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1719; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      if (!(likely(((__pyx_t_1) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_1, __pyx_ptype_5numpy_ndarray))))) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1719; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
      __pyx_t_7 = ((PyArrayObject *)__pyx_t_1);
      {
        __Pyx_BufFmt_StackElem __pyx_stack[1];
        __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_importances.rcbuffer->pybuffer);
        __pyx_t_8 = __Pyx_GetBufferAndValidate(&__pyx_pybuffernd_importances.rcbuffer->pybuffer, (PyObject*)__pyx_t_7, &__Pyx_TypeInfo_nn___pyx_t_5numpy_float64_t, PyBUF_FORMAT| PyBUF_STRIDES| PyBUF_WRITABLE, 1, 0, __pyx_stack);
        if (unlikely(__pyx_t_8 < 0)) {
          PyErr_Fetch(&__pyx_t_9, &__pyx_t_10, &__pyx_t_11);
          if (unlikely(__Pyx_GetBufferAndValidate(&__pyx_pybuffernd_importances.rcbuffer->pybuffer, (PyObject*)__pyx_v_importances, &__Pyx_TypeInfo_nn___pyx_t_5numpy_float64_t, PyBUF_FORMAT| PyBUF_STRIDES| PyBUF_WRITABLE, 1, 0, __pyx_stack) == -1)) {
            Py_XDECREF(__pyx_t_9); Py_XDECREF(__pyx_t_10); Py_XDECREF(__pyx_t_11);
            __Pyx_RaiseBufferFallbackError();
          } else {
            PyErr_Restore(__pyx_t_9, __pyx_t_10, __pyx_t_11);
          }
        }
        __pyx_pybuffernd_importances.diminfo[0].strides = __pyx_pybuffernd_importances.rcbuffer->pybuffer.strides[0]; __pyx_pybuffernd_importances.diminfo[0].shape = __pyx_pybuffernd_importances.rcbuffer->pybuffer.shape[0];
        if (unlikely(__pyx_t_8 < 0)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1719; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
      }
      __pyx_t_7 = 0;
      __Pyx_DECREF(((PyObject *)__pyx_v_importances));
      __pyx_v_importances = ((PyArrayObject *)__pyx_t_1);
      __pyx_t_1 = 0;
      goto __pyx_L7;
    }
    __pyx_L7:;
    goto __pyx_L6;
  }
  __pyx_L6:;

  /* "sklearn/tree/_tree.pyx":1721
 *                 importances /= normalizer
 * 
 *         return importances             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(((PyObject *)__pyx_v_importances));
  __pyx_r = ((PyObject *)__pyx_v_importances);
  goto __pyx_L0;

  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  { PyObject *__pyx_type, *__pyx_value, *__pyx_tb;
    __Pyx_ErrFetch(&__pyx_type, &__pyx_value, &__pyx_tb);
    __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_importances.rcbuffer->pybuffer);
  __Pyx_ErrRestore(__pyx_type, __pyx_value, __pyx_tb);}
  __Pyx_AddTraceback("sklearn.tree._tree.Tree.compute_feature_importances", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  goto __pyx_L2;
  __pyx_L0:;
  __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_importances.rcbuffer->pybuffer);
  __pyx_L2:;
  __Pyx_XDECREF((PyObject *)__pyx_v_importances);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_7sklearn_4tree_5_tree_4Tree_17compute_feature_importances(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_7sklearn_4tree_5_tree_4Tree_16compute_feature_importances[] = "Computes the importance of each feature (aka variable).";
static PyObject *__pyx_pw_7sklearn_4tree_5_tree_4Tree_17compute_feature_importances(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_normalize = 0;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("compute_feature_importances (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s__normalize,0};
    PyObject* values[1] = {0};
    values[0] = __pyx_k_1;
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (kw_args > 0) {
          PyObject* value = PyDict_GetItem(__pyx_kwds, __pyx_n_s__normalize);
          if (value) { values[0] = value; kw_args--; }
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "compute_feature_importances") < 0)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1682; __pyx_clineno = __LINE__; goto __pyx_L3_error;}
      }
    } else {
      switch (PyTuple_GET_SIZE(__pyx_args)) {
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_normalize = values[0];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("compute_feature_importances", 0, 0, 1, PyTuple_GET_SIZE(__pyx_args)); {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1682; __pyx_clineno = __LINE__; goto __pyx_L3_error;}
  __pyx_L3_error:;
  __Pyx_AddTraceback("sklearn.tree._tree.Tree.compute_feature_importances", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_7sklearn_4tree_5_tree_4Tree_16compute_feature_importances(((struct __pyx_obj_7sklearn_4tree_5_tree_Tree *)__pyx_v_self), __pyx_v_normalize);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "sklearn/tree/_tree.pyx":1682
 *         return out
 * 
 *     cpdef compute_feature_importances(self, normalize=True):             # <<<<<<<<<<<<<<
 *         """Computes the importance of each feature (aka variable)."""
 *         cdef SIZE_t* children_left = self.children_left
 */

static PyObject *__pyx_pf_7sklearn_4tree_5_tree_4Tree_16compute_feature_importances(struct __pyx_obj_7sklearn_4tree_5_tree_Tree *__pyx_v_self, PyObject *__pyx_v_normalize) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  struct __pyx_opt_args_7sklearn_4tree_5_tree_4Tree_compute_feature_importances __pyx_t_2;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("compute_feature_importances", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_2.__pyx_n = 1;
  __pyx_t_2.normalize = __pyx_v_normalize;
  __pyx_t_1 = ((struct __pyx_vtabstruct_7sklearn_4tree_5_tree_Tree *)__pyx_v_self->__pyx_vtab)->compute_feature_importances(__pyx_v_self, 1, &__pyx_t_2); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1682; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("sklearn.tree._tree.Tree.compute_feature_importances", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_7sklearn_4tree_5_tree_4Tree_10n_features_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_7sklearn_4tree_5_tree_4Tree_10n_features_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_7sklearn_4tree_5_tree_4Tree_10n_features___get__(((struct __pyx_obj_7sklearn_4tree_5_tree_Tree *)__pyx_v_self));
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "sklearn/tree/_tree.pxd":104
 * cdef class Tree:
 *     # Input/Output layout
 *     cdef public SIZE_t n_features        # Number of features in X             # <<<<<<<<<<<<<<
 *     cdef SIZE_t* n_classes               # Number of classes in y[:, k]
 *     cdef public SIZE_t n_outputs         # Number of outputs in y
 */

static PyObject *__pyx_pf_7sklearn_4tree_5_tree_4Tree_10n_features___get__(struct __pyx_obj_7sklearn_4tree_5_tree_Tree *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyInt_to_py_Py_intptr_t(__pyx_v_self->n_features); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[1]; __pyx_lineno = 104; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("sklearn.tree._tree.Tree.n_features.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static int __pyx_pw_7sklearn_4tree_5_tree_4Tree_10n_features_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value); /*proto*/
static int __pyx_pw_7sklearn_4tree_5_tree_4Tree_10n_features_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__ (wrapper)", 0);
  __pyx_r = __pyx_pf_7sklearn_4tree_5_tree_4Tree_10n_features_2__set__(((struct __pyx_obj_7sklearn_4tree_5_tree_Tree *)__pyx_v_self), ((PyObject *)__pyx_v_value));
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_7sklearn_4tree_5_tree_4Tree_10n_features_2__set__(struct __pyx_obj_7sklearn_4tree_5_tree_Tree *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __pyx_t_7sklearn_4tree_5_tree_SIZE_t __pyx_t_1;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__set__", 0);
  __pyx_t_1 = __Pyx_PyInt_from_py_Py_intptr_t(__pyx_v_value); if (unlikely((__pyx_t_1 == (npy_intp)-1) && PyErr_Occurred())) {__pyx_filename = __pyx_f[1]; __pyx_lineno = 104; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __pyx_v_self->n_features = __pyx_t_1;

  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_AddTraceback("sklearn.tree._tree.Tree.n_features.__set__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_7sklearn_4tree_5_tree_4Tree_9n_outputs_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_7sklearn_4tree_5_tree_4Tree_9n_outputs_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_7sklearn_4tree_5_tree_4Tree_9n_outputs___get__(((struct __pyx_obj_7sklearn_4tree_5_tree_Tree *)__pyx_v_self));
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "sklearn/tree/_tree.pxd":106
 *     cdef public SIZE_t n_features        # Number of features in X
 *     cdef SIZE_t* n_classes               # Number of classes in y[:, k]
 *     cdef public SIZE_t n_outputs         # Number of outputs in y             # <<<<<<<<<<<<<<
 *     cdef public SIZE_t max_n_classes     # max(n_classes)
 *     cdef public SIZE_t value_stride      # n_outputs * max_n_classes
 */

static PyObject *__pyx_pf_7sklearn_4tree_5_tree_4Tree_9n_outputs___get__(struct __pyx_obj_7sklearn_4tree_5_tree_Tree *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyInt_to_py_Py_intptr_t(__pyx_v_self->n_outputs); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[1]; __pyx_lineno = 106; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("sklearn.tree._tree.Tree.n_outputs.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static int __pyx_pw_7sklearn_4tree_5_tree_4Tree_9n_outputs_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value); /*proto*/
static int __pyx_pw_7sklearn_4tree_5_tree_4Tree_9n_outputs_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__ (wrapper)", 0);
  __pyx_r = __pyx_pf_7sklearn_4tree_5_tree_4Tree_9n_outputs_2__set__(((struct __pyx_obj_7sklearn_4tree_5_tree_Tree *)__pyx_v_self), ((PyObject *)__pyx_v_value));
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_7sklearn_4tree_5_tree_4Tree_9n_outputs_2__set__(struct __pyx_obj_7sklearn_4tree_5_tree_Tree *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __pyx_t_7sklearn_4tree_5_tree_SIZE_t __pyx_t_1;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__set__", 0);
  __pyx_t_1 = __Pyx_PyInt_from_py_Py_intptr_t(__pyx_v_value); if (unlikely((__pyx_t_1 == (npy_intp)-1) && PyErr_Occurred())) {__pyx_filename = __pyx_f[1]; __pyx_lineno = 106; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __pyx_v_self->n_outputs = __pyx_t_1;

  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_AddTraceback("sklearn.tree._tree.Tree.n_outputs.__set__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_7sklearn_4tree_5_tree_4Tree_13max_n_classes_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_7sklearn_4tree_5_tree_4Tree_13max_n_classes_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_7sklearn_4tree_5_tree_4Tree_13max_n_classes___get__(((struct __pyx_obj_7sklearn_4tree_5_tree_Tree *)__pyx_v_self));
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "sklearn/tree/_tree.pxd":107
 *     cdef SIZE_t* n_classes               # Number of classes in y[:, k]
 *     cdef public SIZE_t n_outputs         # Number of outputs in y
 *     cdef public SIZE_t max_n_classes     # max(n_classes)             # <<<<<<<<<<<<<<
 *     cdef public SIZE_t value_stride      # n_outputs * max_n_classes
 * 
 */

static PyObject *__pyx_pf_7sklearn_4tree_5_tree_4Tree_13max_n_classes___get__(struct __pyx_obj_7sklearn_4tree_5_tree_Tree *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyInt_to_py_Py_intptr_t(__pyx_v_self->max_n_classes); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[1]; __pyx_lineno = 107; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("sklearn.tree._tree.Tree.max_n_classes.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static int __pyx_pw_7sklearn_4tree_5_tree_4Tree_13max_n_classes_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value); /*proto*/
static int __pyx_pw_7sklearn_4tree_5_tree_4Tree_13max_n_classes_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__ (wrapper)", 0);
  __pyx_r = __pyx_pf_7sklearn_4tree_5_tree_4Tree_13max_n_classes_2__set__(((struct __pyx_obj_7sklearn_4tree_5_tree_Tree *)__pyx_v_self), ((PyObject *)__pyx_v_value));
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_7sklearn_4tree_5_tree_4Tree_13max_n_classes_2__set__(struct __pyx_obj_7sklearn_4tree_5_tree_Tree *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __pyx_t_7sklearn_4tree_5_tree_SIZE_t __pyx_t_1;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__set__", 0);
  __pyx_t_1 = __Pyx_PyInt_from_py_Py_intptr_t(__pyx_v_value); if (unlikely((__pyx_t_1 == (npy_intp)-1) && PyErr_Occurred())) {__pyx_filename = __pyx_f[1]; __pyx_lineno = 107; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __pyx_v_self->max_n_classes = __pyx_t_1;

  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_AddTraceback("sklearn.tree._tree.Tree.max_n_classes.__set__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_7sklearn_4tree_5_tree_4Tree_12value_stride_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_7sklearn_4tree_5_tree_4Tree_12value_stride_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_7sklearn_4tree_5_tree_4Tree_12value_stride___get__(((struct __pyx_obj_7sklearn_4tree_5_tree_Tree *)__pyx_v_self));
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "sklearn/tree/_tree.pxd":108
 *     cdef public SIZE_t n_outputs         # Number of outputs in y
 *     cdef public SIZE_t max_n_classes     # max(n_classes)
 *     cdef public SIZE_t value_stride      # n_outputs * max_n_classes             # <<<<<<<<<<<<<<
 * 
 *     # Parameters
 */

static PyObject *__pyx_pf_7sklearn_4tree_5_tree_4Tree_12value_stride___get__(struct __pyx_obj_7sklearn_4tree_5_tree_Tree *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyInt_to_py_Py_intptr_t(__pyx_v_self->value_stride); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[1]; __pyx_lineno = 108; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("sklearn.tree._tree.Tree.value_stride.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static int __pyx_pw_7sklearn_4tree_5_tree_4Tree_12value_stride_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value); /*proto*/
static int __pyx_pw_7sklearn_4tree_5_tree_4Tree_12value_stride_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__ (wrapper)", 0);
  __pyx_r = __pyx_pf_7sklearn_4tree_5_tree_4Tree_12value_stride_2__set__(((struct __pyx_obj_7sklearn_4tree_5_tree_Tree *)__pyx_v_self), ((PyObject *)__pyx_v_value));
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_7sklearn_4tree_5_tree_4Tree_12value_stride_2__set__(struct __pyx_obj_7sklearn_4tree_5_tree_Tree *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __pyx_t_7sklearn_4tree_5_tree_SIZE_t __pyx_t_1;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__set__", 0);
  __pyx_t_1 = __Pyx_PyInt_from_py_Py_intptr_t(__pyx_v_value); if (unlikely((__pyx_t_1 == (npy_intp)-1) && PyErr_Occurred())) {__pyx_filename = __pyx_f[1]; __pyx_lineno = 108; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __pyx_v_self->value_stride = __pyx_t_1;

  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_AddTraceback("sklearn.tree._tree.Tree.value_stride.__set__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_7sklearn_4tree_5_tree_4Tree_8splitter_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_7sklearn_4tree_5_tree_4Tree_8splitter_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_7sklearn_4tree_5_tree_4Tree_8splitter___get__(((struct __pyx_obj_7sklearn_4tree_5_tree_Tree *)__pyx_v_self));
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "sklearn/tree/_tree.pxd":111
 * 
 *     # Parameters
 *     cdef public Splitter splitter        # Splitting algorithm             # <<<<<<<<<<<<<<
 *     cdef public SIZE_t max_depth         # Max depth of the tree
 *     cdef public SIZE_t min_samples_split # Minimum number of samples in an internal node
 */

static PyObject *__pyx_pf_7sklearn_4tree_5_tree_4Tree_8splitter___get__(struct __pyx_obj_7sklearn_4tree_5_tree_Tree *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__", 0);
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(((PyObject *)__pyx_v_self->splitter));
  __pyx_r = ((PyObject *)__pyx_v_self->splitter);
  goto __pyx_L0;

  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static int __pyx_pw_7sklearn_4tree_5_tree_4Tree_8splitter_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value); /*proto*/
static int __pyx_pw_7sklearn_4tree_5_tree_4Tree_8splitter_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__ (wrapper)", 0);
  __pyx_r = __pyx_pf_7sklearn_4tree_5_tree_4Tree_8splitter_2__set__(((struct __pyx_obj_7sklearn_4tree_5_tree_Tree *)__pyx_v_self), ((PyObject *)__pyx_v_value));
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_7sklearn_4tree_5_tree_4Tree_8splitter_2__set__(struct __pyx_obj_7sklearn_4tree_5_tree_Tree *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__set__", 0);
  if (!(likely(((__pyx_v_value) == Py_None) || likely(__Pyx_TypeTest(__pyx_v_value, __pyx_ptype_7sklearn_4tree_5_tree_Splitter))))) {__pyx_filename = __pyx_f[1]; __pyx_lineno = 111; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_INCREF(__pyx_v_value);
  __Pyx_GIVEREF(__pyx_v_value);
  __Pyx_GOTREF(__pyx_v_self->splitter);
  __Pyx_DECREF(((PyObject *)__pyx_v_self->splitter));
  __pyx_v_self->splitter = ((struct __pyx_obj_7sklearn_4tree_5_tree_Splitter *)__pyx_v_value);

  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_AddTraceback("sklearn.tree._tree.Tree.splitter.__set__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static int __pyx_pw_7sklearn_4tree_5_tree_4Tree_8splitter_5__del__(PyObject *__pyx_v_self); /*proto*/
static int __pyx_pw_7sklearn_4tree_5_tree_4Tree_8splitter_5__del__(PyObject *__pyx_v_self) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__del__ (wrapper)", 0);
  __pyx_r = __pyx_pf_7sklearn_4tree_5_tree_4Tree_8splitter_4__del__(((struct __pyx_obj_7sklearn_4tree_5_tree_Tree *)__pyx_v_self));
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_7sklearn_4tree_5_tree_4Tree_8splitter_4__del__(struct __pyx_obj_7sklearn_4tree_5_tree_Tree *__pyx_v_self) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__del__", 0);
  __Pyx_INCREF(Py_None);
  __Pyx_GIVEREF(Py_None);
  __Pyx_GOTREF(__pyx_v_self->splitter);
  __Pyx_DECREF(((PyObject *)__pyx_v_self->splitter));
  __pyx_v_self->splitter = ((struct __pyx_obj_7sklearn_4tree_5_tree_Splitter *)Py_None);

  __pyx_r = 0;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_7sklearn_4tree_5_tree_4Tree_9max_depth_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_7sklearn_4tree_5_tree_4Tree_9max_depth_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_7sklearn_4tree_5_tree_4Tree_9max_depth___get__(((struct __pyx_obj_7sklearn_4tree_5_tree_Tree *)__pyx_v_self));
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "sklearn/tree/_tree.pxd":112
 *     # Parameters
 *     cdef public Splitter splitter        # Splitting algorithm
 *     cdef public SIZE_t max_depth         # Max depth of the tree             # <<<<<<<<<<<<<<
 *     cdef public SIZE_t min_samples_split # Minimum number of samples in an internal node
 *     cdef public SIZE_t min_samples_leaf  # Minimum number of samples in a leaf
 */

static PyObject *__pyx_pf_7sklearn_4tree_5_tree_4Tree_9max_depth___get__(struct __pyx_obj_7sklearn_4tree_5_tree_Tree *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyInt_to_py_Py_intptr_t(__pyx_v_self->max_depth); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[1]; __pyx_lineno = 112; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("sklearn.tree._tree.Tree.max_depth.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static int __pyx_pw_7sklearn_4tree_5_tree_4Tree_9max_depth_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value); /*proto*/
static int __pyx_pw_7sklearn_4tree_5_tree_4Tree_9max_depth_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__ (wrapper)", 0);
  __pyx_r = __pyx_pf_7sklearn_4tree_5_tree_4Tree_9max_depth_2__set__(((struct __pyx_obj_7sklearn_4tree_5_tree_Tree *)__pyx_v_self), ((PyObject *)__pyx_v_value));
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_7sklearn_4tree_5_tree_4Tree_9max_depth_2__set__(struct __pyx_obj_7sklearn_4tree_5_tree_Tree *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __pyx_t_7sklearn_4tree_5_tree_SIZE_t __pyx_t_1;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__set__", 0);
  __pyx_t_1 = __Pyx_PyInt_from_py_Py_intptr_t(__pyx_v_value); if (unlikely((__pyx_t_1 == (npy_intp)-1) && PyErr_Occurred())) {__pyx_filename = __pyx_f[1]; __pyx_lineno = 112; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __pyx_v_self->max_depth = __pyx_t_1;

  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_AddTraceback("sklearn.tree._tree.Tree.max_depth.__set__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_7sklearn_4tree_5_tree_4Tree_17min_samples_split_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_7sklearn_4tree_5_tree_4Tree_17min_samples_split_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_7sklearn_4tree_5_tree_4Tree_17min_samples_split___get__(((struct __pyx_obj_7sklearn_4tree_5_tree_Tree *)__pyx_v_self));
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "sklearn/tree/_tree.pxd":113
 *     cdef public Splitter splitter        # Splitting algorithm
 *     cdef public SIZE_t max_depth         # Max depth of the tree
 *     cdef public SIZE_t min_samples_split # Minimum number of samples in an internal node             # <<<<<<<<<<<<<<
 *     cdef public SIZE_t min_samples_leaf  # Minimum number of samples in a leaf
 *     cdef public object random_state      # Random state
 */

static PyObject *__pyx_pf_7sklearn_4tree_5_tree_4Tree_17min_samples_split___get__(struct __pyx_obj_7sklearn_4tree_5_tree_Tree *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyInt_to_py_Py_intptr_t(__pyx_v_self->min_samples_split); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[1]; __pyx_lineno = 113; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("sklearn.tree._tree.Tree.min_samples_split.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static int __pyx_pw_7sklearn_4tree_5_tree_4Tree_17min_samples_split_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value); /*proto*/
static int __pyx_pw_7sklearn_4tree_5_tree_4Tree_17min_samples_split_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__ (wrapper)", 0);
  __pyx_r = __pyx_pf_7sklearn_4tree_5_tree_4Tree_17min_samples_split_2__set__(((struct __pyx_obj_7sklearn_4tree_5_tree_Tree *)__pyx_v_self), ((PyObject *)__pyx_v_value));
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_7sklearn_4tree_5_tree_4Tree_17min_samples_split_2__set__(struct __pyx_obj_7sklearn_4tree_5_tree_Tree *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __pyx_t_7sklearn_4tree_5_tree_SIZE_t __pyx_t_1;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__set__", 0);
  __pyx_t_1 = __Pyx_PyInt_from_py_Py_intptr_t(__pyx_v_value); if (unlikely((__pyx_t_1 == (npy_intp)-1) && PyErr_Occurred())) {__pyx_filename = __pyx_f[1]; __pyx_lineno = 113; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __pyx_v_self->min_samples_split = __pyx_t_1;

  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_AddTraceback("sklearn.tree._tree.Tree.min_samples_split.__set__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_7sklearn_4tree_5_tree_4Tree_16min_samples_leaf_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_7sklearn_4tree_5_tree_4Tree_16min_samples_leaf_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_7sklearn_4tree_5_tree_4Tree_16min_samples_leaf___get__(((struct __pyx_obj_7sklearn_4tree_5_tree_Tree *)__pyx_v_self));
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "sklearn/tree/_tree.pxd":114
 *     cdef public SIZE_t max_depth         # Max depth of the tree
 *     cdef public SIZE_t min_samples_split # Minimum number of samples in an internal node
 *     cdef public SIZE_t min_samples_leaf  # Minimum number of samples in a leaf             # <<<<<<<<<<<<<<
 *     cdef public object random_state      # Random state
 * 
 */

static PyObject *__pyx_pf_7sklearn_4tree_5_tree_4Tree_16min_samples_leaf___get__(struct __pyx_obj_7sklearn_4tree_5_tree_Tree *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyInt_to_py_Py_intptr_t(__pyx_v_self->min_samples_leaf); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[1]; __pyx_lineno = 114; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("sklearn.tree._tree.Tree.min_samples_leaf.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static int __pyx_pw_7sklearn_4tree_5_tree_4Tree_16min_samples_leaf_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value); /*proto*/
static int __pyx_pw_7sklearn_4tree_5_tree_4Tree_16min_samples_leaf_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__ (wrapper)", 0);
  __pyx_r = __pyx_pf_7sklearn_4tree_5_tree_4Tree_16min_samples_leaf_2__set__(((struct __pyx_obj_7sklearn_4tree_5_tree_Tree *)__pyx_v_self), ((PyObject *)__pyx_v_value));
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_7sklearn_4tree_5_tree_4Tree_16min_samples_leaf_2__set__(struct __pyx_obj_7sklearn_4tree_5_tree_Tree *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __pyx_t_7sklearn_4tree_5_tree_SIZE_t __pyx_t_1;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__set__", 0);
  __pyx_t_1 = __Pyx_PyInt_from_py_Py_intptr_t(__pyx_v_value); if (unlikely((__pyx_t_1 == (npy_intp)-1) && PyErr_Occurred())) {__pyx_filename = __pyx_f[1]; __pyx_lineno = 114; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __pyx_v_self->min_samples_leaf = __pyx_t_1;

  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_AddTraceback("sklearn.tree._tree.Tree.min_samples_leaf.__set__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_7sklearn_4tree_5_tree_4Tree_12random_state_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_7sklearn_4tree_5_tree_4Tree_12random_state_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_7sklearn_4tree_5_tree_4Tree_12random_state___get__(((struct __pyx_obj_7sklearn_4tree_5_tree_Tree *)__pyx_v_self));
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "sklearn/tree/_tree.pxd":115
 *     cdef public SIZE_t min_samples_split # Minimum number of samples in an internal node
 *     cdef public SIZE_t min_samples_leaf  # Minimum number of samples in a leaf
 *     cdef public object random_state      # Random state             # <<<<<<<<<<<<<<
 * 
 *     # Inner structures
 */

static PyObject *__pyx_pf_7sklearn_4tree_5_tree_4Tree_12random_state___get__(struct __pyx_obj_7sklearn_4tree_5_tree_Tree *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__", 0);
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_self->random_state);
  __pyx_r = __pyx_v_self->random_state;
  goto __pyx_L0;

  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static int __pyx_pw_7sklearn_4tree_5_tree_4Tree_12random_state_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value); /*proto*/
static int __pyx_pw_7sklearn_4tree_5_tree_4Tree_12random_state_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__ (wrapper)", 0);
  __pyx_r = __pyx_pf_7sklearn_4tree_5_tree_4Tree_12random_state_2__set__(((struct __pyx_obj_7sklearn_4tree_5_tree_Tree *)__pyx_v_self), ((PyObject *)__pyx_v_value));
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_7sklearn_4tree_5_tree_4Tree_12random_state_2__set__(struct __pyx_obj_7sklearn_4tree_5_tree_Tree *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__", 0);
  __Pyx_INCREF(__pyx_v_value);
  __Pyx_GIVEREF(__pyx_v_value);
  __Pyx_GOTREF(__pyx_v_self->random_state);
  __Pyx_DECREF(__pyx_v_self->random_state);
  __pyx_v_self->random_state = __pyx_v_value;

  __pyx_r = 0;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static int __pyx_pw_7sklearn_4tree_5_tree_4Tree_12random_state_5__del__(PyObject *__pyx_v_self); /*proto*/
static int __pyx_pw_7sklearn_4tree_5_tree_4Tree_12random_state_5__del__(PyObject *__pyx_v_self) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__del__ (wrapper)", 0);
  __pyx_r = __pyx_pf_7sklearn_4tree_5_tree_4Tree_12random_state_4__del__(((struct __pyx_obj_7sklearn_4tree_5_tree_Tree *)__pyx_v_self));
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_7sklearn_4tree_5_tree_4Tree_12random_state_4__del__(struct __pyx_obj_7sklearn_4tree_5_tree_Tree *__pyx_v_self) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__del__", 0);
  __Pyx_INCREF(Py_None);
  __Pyx_GIVEREF(Py_None);
  __Pyx_GOTREF(__pyx_v_self->random_state);
  __Pyx_DECREF(__pyx_v_self->random_state);
  __pyx_v_self->random_state = Py_None;

  __pyx_r = 0;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_7sklearn_4tree_5_tree_4Tree_10node_count_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_7sklearn_4tree_5_tree_4Tree_10node_count_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_7sklearn_4tree_5_tree_4Tree_10node_count___get__(((struct __pyx_obj_7sklearn_4tree_5_tree_Tree *)__pyx_v_self));
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "sklearn/tree/_tree.pxd":118
 * 
 *     # Inner structures
 *     cdef public SIZE_t node_count        # Counter for node IDs             # <<<<<<<<<<<<<<
 *     cdef public SIZE_t capacity          # Capacity of tree, in terms of nodes
 *     cdef SIZE_t* children_left           # children_left[i] is the left child of node i
 */

static PyObject *__pyx_pf_7sklearn_4tree_5_tree_4Tree_10node_count___get__(struct __pyx_obj_7sklearn_4tree_5_tree_Tree *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyInt_to_py_Py_intptr_t(__pyx_v_self->node_count); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[1]; __pyx_lineno = 118; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("sklearn.tree._tree.Tree.node_count.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static int __pyx_pw_7sklearn_4tree_5_tree_4Tree_10node_count_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value); /*proto*/
static int __pyx_pw_7sklearn_4tree_5_tree_4Tree_10node_count_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__ (wrapper)", 0);
  __pyx_r = __pyx_pf_7sklearn_4tree_5_tree_4Tree_10node_count_2__set__(((struct __pyx_obj_7sklearn_4tree_5_tree_Tree *)__pyx_v_self), ((PyObject *)__pyx_v_value));
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_7sklearn_4tree_5_tree_4Tree_10node_count_2__set__(struct __pyx_obj_7sklearn_4tree_5_tree_Tree *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __pyx_t_7sklearn_4tree_5_tree_SIZE_t __pyx_t_1;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__set__", 0);
  __pyx_t_1 = __Pyx_PyInt_from_py_Py_intptr_t(__pyx_v_value); if (unlikely((__pyx_t_1 == (npy_intp)-1) && PyErr_Occurred())) {__pyx_filename = __pyx_f[1]; __pyx_lineno = 118; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __pyx_v_self->node_count = __pyx_t_1;

  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_AddTraceback("sklearn.tree._tree.Tree.node_count.__set__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_7sklearn_4tree_5_tree_4Tree_8capacity_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_7sklearn_4tree_5_tree_4Tree_8capacity_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_7sklearn_4tree_5_tree_4Tree_8capacity___get__(((struct __pyx_obj_7sklearn_4tree_5_tree_Tree *)__pyx_v_self));
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "sklearn/tree/_tree.pxd":119
 *     # Inner structures
 *     cdef public SIZE_t node_count        # Counter for node IDs
 *     cdef public SIZE_t capacity          # Capacity of tree, in terms of nodes             # <<<<<<<<<<<<<<
 *     cdef SIZE_t* children_left           # children_left[i] is the left child of node i
 *     cdef SIZE_t* children_right          # children_right[i] is the right child of node i
 */

static PyObject *__pyx_pf_7sklearn_4tree_5_tree_4Tree_8capacity___get__(struct __pyx_obj_7sklearn_4tree_5_tree_Tree *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyInt_to_py_Py_intptr_t(__pyx_v_self->capacity); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[1]; __pyx_lineno = 119; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("sklearn.tree._tree.Tree.capacity.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static int __pyx_pw_7sklearn_4tree_5_tree_4Tree_8capacity_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value); /*proto*/
static int __pyx_pw_7sklearn_4tree_5_tree_4Tree_8capacity_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__ (wrapper)", 0);
  __pyx_r = __pyx_pf_7sklearn_4tree_5_tree_4Tree_8capacity_2__set__(((struct __pyx_obj_7sklearn_4tree_5_tree_Tree *)__pyx_v_self), ((PyObject *)__pyx_v_value));
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_7sklearn_4tree_5_tree_4Tree_8capacity_2__set__(struct __pyx_obj_7sklearn_4tree_5_tree_Tree *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __pyx_t_7sklearn_4tree_5_tree_SIZE_t __pyx_t_1;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__set__", 0);
  __pyx_t_1 = __Pyx_PyInt_from_py_Py_intptr_t(__pyx_v_value); if (unlikely((__pyx_t_1 == (npy_intp)-1) && PyErr_Occurred())) {__pyx_filename = __pyx_f[1]; __pyx_lineno = 119; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __pyx_v_self->capacity = __pyx_t_1;

  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_AddTraceback("sklearn.tree._tree.Tree.capacity.__set__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "sklearn/tree/_tree.pyx":1729
 * 
 * # rand_r replacement taken from 4.4BSD C library.
 * cdef inline int our_rand_r(unsigned *seed) nogil:             # <<<<<<<<<<<<<<
 *     seed[0] = seed[0] * 1103515245 + 12345
 *     return (seed[0] % (<unsigned>RAND_R_MAX + 1))
 */

static CYTHON_INLINE int __pyx_f_7sklearn_4tree_5_tree_our_rand_r(unsigned int *__pyx_v_seed) {
  int __pyx_r;

  /* "sklearn/tree/_tree.pyx":1730
 * # rand_r replacement taken from 4.4BSD C library.
 * cdef inline int our_rand_r(unsigned *seed) nogil:
 *     seed[0] = seed[0] * 1103515245 + 12345             # <<<<<<<<<<<<<<
 *     return (seed[0] % (<unsigned>RAND_R_MAX + 1))
 * 
 */
  (__pyx_v_seed[0]) = (((__pyx_v_seed[0]) * 1103515245) + 12345);

  /* "sklearn/tree/_tree.pyx":1731
 * cdef inline int our_rand_r(unsigned *seed) nogil:
 *     seed[0] = seed[0] * 1103515245 + 12345
 *     return (seed[0] % (<unsigned>RAND_R_MAX + 1))             # <<<<<<<<<<<<<<
 * 
 * cdef inline np.ndarray int_ptr_to_ndarray(int* data, SIZE_t size):
 */
  __pyx_r = ((__pyx_v_seed[0]) % (((unsigned int)__pyx_e_7sklearn_4tree_5_tree_RAND_R_MAX) + 1));
  goto __pyx_L0;

  __pyx_r = 0;
  __pyx_L0:;
  return __pyx_r;
}

/* "sklearn/tree/_tree.pyx":1733
 *     return (seed[0] % (<unsigned>RAND_R_MAX + 1))
 * 
 * cdef inline np.ndarray int_ptr_to_ndarray(int* data, SIZE_t size):             # <<<<<<<<<<<<<<
 *     """Encapsulate data into a 1D numpy array of int's."""
 *     cdef np.npy_intp shape[1]
 */

static CYTHON_INLINE PyArrayObject *__pyx_f_7sklearn_4tree_5_tree_int_ptr_to_ndarray(int *__pyx_v_data, __pyx_t_7sklearn_4tree_5_tree_SIZE_t __pyx_v_size) {
  npy_intp __pyx_v_shape[1];
  PyArrayObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("int_ptr_to_ndarray", 0);

  /* "sklearn/tree/_tree.pyx":1736
 *     """Encapsulate data into a 1D numpy array of int's."""
 *     cdef np.npy_intp shape[1]
 *     shape[0] = <np.npy_intp> size             # <<<<<<<<<<<<<<
 *     return np.PyArray_SimpleNewFromData(1, shape, np.NPY_INT, data)
 * 
 */
  (__pyx_v_shape[0]) = ((npy_intp)__pyx_v_size);

  /* "sklearn/tree/_tree.pyx":1737
 *     cdef np.npy_intp shape[1]
 *     shape[0] = <np.npy_intp> size
 *     return np.PyArray_SimpleNewFromData(1, shape, np.NPY_INT, data)             # <<<<<<<<<<<<<<
 * 
 * cdef inline np.ndarray sizet_ptr_to_ndarray(SIZE_t* data, SIZE_t size):
 */
  __Pyx_XDECREF(((PyObject *)__pyx_r));
  __pyx_t_1 = PyArray_SimpleNewFromData(1, __pyx_v_shape, NPY_INT, __pyx_v_data); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1737; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_1);
  if (!(likely(((__pyx_t_1) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_1, __pyx_ptype_5numpy_ndarray))))) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1737; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __pyx_r = ((PyArrayObject *)__pyx_t_1);
  __pyx_t_1 = 0;
  goto __pyx_L0;

  __pyx_r = ((PyArrayObject *)Py_None); __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("sklearn.tree._tree.int_ptr_to_ndarray", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF((PyObject *)__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "sklearn/tree/_tree.pyx":1739
 *     return np.PyArray_SimpleNewFromData(1, shape, np.NPY_INT, data)
 * 
 * cdef inline np.ndarray sizet_ptr_to_ndarray(SIZE_t* data, SIZE_t size):             # <<<<<<<<<<<<<<
 *     """Encapsulate data into a 1D numpy array of intp's."""
 *     cdef np.npy_intp shape[1]
 */

static CYTHON_INLINE PyArrayObject *__pyx_f_7sklearn_4tree_5_tree_sizet_ptr_to_ndarray(__pyx_t_7sklearn_4tree_5_tree_SIZE_t *__pyx_v_data, __pyx_t_7sklearn_4tree_5_tree_SIZE_t __pyx_v_size) {
  npy_intp __pyx_v_shape[1];
  PyArrayObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("sizet_ptr_to_ndarray", 0);

  /* "sklearn/tree/_tree.pyx":1742
 *     """Encapsulate data into a 1D numpy array of intp's."""
 *     cdef np.npy_intp shape[1]
 *     shape[0] = <np.npy_intp> size             # <<<<<<<<<<<<<<
 *     return np.PyArray_SimpleNewFromData(1, shape, np.NPY_INTP, data)
 * 
 */
  (__pyx_v_shape[0]) = ((npy_intp)__pyx_v_size);

  /* "sklearn/tree/_tree.pyx":1743
 *     cdef np.npy_intp shape[1]
 *     shape[0] = <np.npy_intp> size
 *     return np.PyArray_SimpleNewFromData(1, shape, np.NPY_INTP, data)             # <<<<<<<<<<<<<<
 * 
 * cdef inline np.ndarray double_ptr_to_ndarray(double* data, SIZE_t size):
 */
  __Pyx_XDECREF(((PyObject *)__pyx_r));
  __pyx_t_1 = PyArray_SimpleNewFromData(1, __pyx_v_shape, NPY_INTP, __pyx_v_data); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1743; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_1);
  if (!(likely(((__pyx_t_1) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_1, __pyx_ptype_5numpy_ndarray))))) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1743; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __pyx_r = ((PyArrayObject *)__pyx_t_1);
  __pyx_t_1 = 0;
  goto __pyx_L0;

  __pyx_r = ((PyArrayObject *)Py_None); __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("sklearn.tree._tree.sizet_ptr_to_ndarray", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF((PyObject *)__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "sklearn/tree/_tree.pyx":1745
 *     return np.PyArray_SimpleNewFromData(1, shape, np.NPY_INTP, data)
 * 
 * cdef inline np.ndarray double_ptr_to_ndarray(double* data, SIZE_t size):             # <<<<<<<<<<<<<<
 *     """Encapsulate data into a 1D numpy array of double's."""
 *     cdef np.npy_intp shape[1]
 */

static CYTHON_INLINE PyArrayObject *__pyx_f_7sklearn_4tree_5_tree_double_ptr_to_ndarray(double *__pyx_v_data, __pyx_t_7sklearn_4tree_5_tree_SIZE_t __pyx_v_size) {
  npy_intp __pyx_v_shape[1];
  PyArrayObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("double_ptr_to_ndarray", 0);

  /* "sklearn/tree/_tree.pyx":1748
 *     """Encapsulate data into a 1D numpy array of double's."""
 *     cdef np.npy_intp shape[1]
 *     shape[0] = <np.npy_intp> size             # <<<<<<<<<<<<<<
 *     return np.PyArray_SimpleNewFromData(1, shape, np.NPY_DOUBLE, data)
 * 
 */
  (__pyx_v_shape[0]) = ((npy_intp)__pyx_v_size);

  /* "sklearn/tree/_tree.pyx":1749
 *     cdef np.npy_intp shape[1]
 *     shape[0] = <np.npy_intp> size
 *     return np.PyArray_SimpleNewFromData(1, shape, np.NPY_DOUBLE, data)             # <<<<<<<<<<<<<<
 * 
 * cdef inline SIZE_t rand_int(SIZE_t end, unsigned int* random_state) nogil:
 */
  __Pyx_XDECREF(((PyObject *)__pyx_r));
  __pyx_t_1 = PyArray_SimpleNewFromData(1, __pyx_v_shape, NPY_DOUBLE, __pyx_v_data); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1749; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_1);
  if (!(likely(((__pyx_t_1) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_1, __pyx_ptype_5numpy_ndarray))))) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1749; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __pyx_r = ((PyArrayObject *)__pyx_t_1);
  __pyx_t_1 = 0;
  goto __pyx_L0;

  __pyx_r = ((PyArrayObject *)Py_None); __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("sklearn.tree._tree.double_ptr_to_ndarray", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF((PyObject *)__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "sklearn/tree/_tree.pyx":1751
 *     return np.PyArray_SimpleNewFromData(1, shape, np.NPY_DOUBLE, data)
 * 
 * cdef inline SIZE_t rand_int(SIZE_t end, unsigned int* random_state) nogil:             # <<<<<<<<<<<<<<
 *     """Generate a random integer in [0; end)."""
 *     return our_rand_r(random_state) % end
 */

static CYTHON_INLINE __pyx_t_7sklearn_4tree_5_tree_SIZE_t __pyx_f_7sklearn_4tree_5_tree_rand_int(__pyx_t_7sklearn_4tree_5_tree_SIZE_t __pyx_v_end, unsigned int *__pyx_v_random_state) {
  __pyx_t_7sklearn_4tree_5_tree_SIZE_t __pyx_r;

  /* "sklearn/tree/_tree.pyx":1753
 * cdef inline SIZE_t rand_int(SIZE_t end, unsigned int* random_state) nogil:
 *     """Generate a random integer in [0; end)."""
 *     return our_rand_r(random_state) % end             # <<<<<<<<<<<<<<
 * 
 * cdef inline double rand_double(unsigned int* random_state) nogil:
 */
  __pyx_r = (__pyx_f_7sklearn_4tree_5_tree_our_rand_r(__pyx_v_random_state) % __pyx_v_end);
  goto __pyx_L0;

  __pyx_r = 0;
  __pyx_L0:;
  return __pyx_r;
}

/* "sklearn/tree/_tree.pyx":1755
 *     return our_rand_r(random_state) % end
 * 
 * cdef inline double rand_double(unsigned int* random_state) nogil:             # <<<<<<<<<<<<<<
 *     """Generate a random double in [0; 1)."""
 *     return <double> our_rand_r(random_state) / <double> RAND_R_MAX
 */

static CYTHON_INLINE double __pyx_f_7sklearn_4tree_5_tree_rand_double(unsigned int *__pyx_v_random_state) {
  double __pyx_r;

  /* "sklearn/tree/_tree.pyx":1757
 * cdef inline double rand_double(unsigned int* random_state) nogil:
 *     """Generate a random double in [0; 1)."""
 *     return <double> our_rand_r(random_state) / <double> RAND_R_MAX             # <<<<<<<<<<<<<<
 * 
 * cdef inline double log(double x) nogil:
 */
  __pyx_r = (((double)__pyx_f_7sklearn_4tree_5_tree_our_rand_r(__pyx_v_random_state)) / ((double)__pyx_e_7sklearn_4tree_5_tree_RAND_R_MAX));
  goto __pyx_L0;

  __pyx_r = 0;
  __pyx_L0:;
  return __pyx_r;
}

/* "sklearn/tree/_tree.pyx":1759
 *     return <double> our_rand_r(random_state) / <double> RAND_R_MAX
 * 
 * cdef inline double log(double x) nogil:             # <<<<<<<<<<<<<<
 *     return ln(x) / ln(2.0)
 */

static CYTHON_INLINE double __pyx_f_7sklearn_4tree_5_tree_log(double __pyx_v_x) {
  double __pyx_r;

  /* "sklearn/tree/_tree.pyx":1760
 * 
 * cdef inline double log(double x) nogil:
 *     return ln(x) / ln(2.0)             # <<<<<<<<<<<<<<
 */
  __pyx_r = (log(__pyx_v_x) / log(2.0));
  goto __pyx_L0;

  __pyx_r = 0;
  __pyx_L0:;
  return __pyx_r;
}

/* Python wrapper */
static CYTHON_UNUSED int __pyx_pw_5numpy_7ndarray_1__getbuffer__(PyObject *__pyx_v_self, Py_buffer *__pyx_v_info, int __pyx_v_flags); /*proto*/
static CYTHON_UNUSED int __pyx_pw_5numpy_7ndarray_1__getbuffer__(PyObject *__pyx_v_self, Py_buffer *__pyx_v_info, int __pyx_v_flags) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__getbuffer__ (wrapper)", 0);
  __pyx_r = __pyx_pf_5numpy_7ndarray___getbuffer__(((PyArrayObject *)__pyx_v_self), ((Py_buffer *)__pyx_v_info), ((int)__pyx_v_flags));
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "numpy.pxd":194
 *         # experimental exception made for __getbuffer__ and __releasebuffer__
 *         # -- the details of this may change.
 *         def __getbuffer__(ndarray self, Py_buffer* info, int flags):             # <<<<<<<<<<<<<<
 *             # This implementation of getbuffer is geared towards Cython
 *             # requirements, and does not yet fullfill the PEP.
 */

static int __pyx_pf_5numpy_7ndarray___getbuffer__(PyArrayObject *__pyx_v_self, Py_buffer *__pyx_v_info, int __pyx_v_flags) {
  int __pyx_v_copy_shape;
  int __pyx_v_i;
  int __pyx_v_ndim;
  int __pyx_v_endian_detector;
  int __pyx_v_little_endian;
  int __pyx_v_t;
  char *__pyx_v_f;
  PyArray_Descr *__pyx_v_descr = 0;
  int __pyx_v_offset;
  int __pyx_v_hasfields;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  int __pyx_t_3;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_t_5;
  int __pyx_t_6;
  int __pyx_t_7;
  PyObject *__pyx_t_8 = NULL;
  char *__pyx_t_9;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__getbuffer__", 0);
  if (__pyx_v_info != NULL) {
    __pyx_v_info->obj = Py_None; __Pyx_INCREF(Py_None);
    __Pyx_GIVEREF(__pyx_v_info->obj);
  }

  /* "numpy.pxd":200
 *             # of flags
 * 
 *             if info == NULL: return             # <<<<<<<<<<<<<<
 * 
 *             cdef int copy_shape, i, ndim
 */
  __pyx_t_1 = ((__pyx_v_info == NULL) != 0);
  if (__pyx_t_1) {
    __pyx_r = 0;
    goto __pyx_L0;
    goto __pyx_L3;
  }
  __pyx_L3:;

  /* "numpy.pxd":203
 * 
 *             cdef int copy_shape, i, ndim
 *             cdef int endian_detector = 1             # <<<<<<<<<<<<<<
 *             cdef bint little_endian = ((<char*>&endian_detector)[0] != 0)
 * 
 */
  __pyx_v_endian_detector = 1;

  /* "numpy.pxd":204
 *             cdef int copy_shape, i, ndim
 *             cdef int endian_detector = 1
 *             cdef bint little_endian = ((<char*>&endian_detector)[0] != 0)             # <<<<<<<<<<<<<<
 * 
 *             ndim = PyArray_NDIM(self)
 */
  __pyx_v_little_endian = ((((char *)(&__pyx_v_endian_detector))[0]) != 0);

  /* "numpy.pxd":206
 *             cdef bint little_endian = ((<char*>&endian_detector)[0] != 0)
 * 
 *             ndim = PyArray_NDIM(self)             # <<<<<<<<<<<<<<
 * 
 *             if sizeof(npy_intp) != sizeof(Py_ssize_t):
 */
  __pyx_v_ndim = PyArray_NDIM(__pyx_v_self);

  /* "numpy.pxd":208
 *             ndim = PyArray_NDIM(self)
 * 
 *             if sizeof(npy_intp) != sizeof(Py_ssize_t):             # <<<<<<<<<<<<<<
 *                 copy_shape = 1
 *             else:
 */
  __pyx_t_1 = (((sizeof(npy_intp)) != (sizeof(Py_ssize_t))) != 0);
  if (__pyx_t_1) {

    /* "numpy.pxd":209
 * 
 *             if sizeof(npy_intp) != sizeof(Py_ssize_t):
 *                 copy_shape = 1             # <<<<<<<<<<<<<<
 *             else:
 *                 copy_shape = 0
 */
    __pyx_v_copy_shape = 1;
    goto __pyx_L4;
  }
  /*else*/ {

    /* "numpy.pxd":211
 *                 copy_shape = 1
 *             else:
 *                 copy_shape = 0             # <<<<<<<<<<<<<<
 * 
 *             if ((flags & pybuf.PyBUF_C_CONTIGUOUS == pybuf.PyBUF_C_CONTIGUOUS)
 */
    __pyx_v_copy_shape = 0;
  }
  __pyx_L4:;

  /* "numpy.pxd":213
 *                 copy_shape = 0
 * 
 *             if ((flags & pybuf.PyBUF_C_CONTIGUOUS == pybuf.PyBUF_C_CONTIGUOUS)             # <<<<<<<<<<<<<<
 *                 and not PyArray_CHKFLAGS(self, NPY_C_CONTIGUOUS)):
 *                 raise ValueError(u"ndarray is not C contiguous")
 */
  __pyx_t_1 = (((__pyx_v_flags & PyBUF_C_CONTIGUOUS) == PyBUF_C_CONTIGUOUS) != 0);
  if (__pyx_t_1) {

    /* "numpy.pxd":214
 * 
 *             if ((flags & pybuf.PyBUF_C_CONTIGUOUS == pybuf.PyBUF_C_CONTIGUOUS)
 *                 and not PyArray_CHKFLAGS(self, NPY_C_CONTIGUOUS)):             # <<<<<<<<<<<<<<
 *                 raise ValueError(u"ndarray is not C contiguous")
 * 
 */
    __pyx_t_2 = ((!(PyArray_CHKFLAGS(__pyx_v_self, NPY_C_CONTIGUOUS) != 0)) != 0);
    __pyx_t_3 = __pyx_t_2;
  } else {
    __pyx_t_3 = __pyx_t_1;
  }
  if (__pyx_t_3) {

    /* "numpy.pxd":215
 *             if ((flags & pybuf.PyBUF_C_CONTIGUOUS == pybuf.PyBUF_C_CONTIGUOUS)
 *                 and not PyArray_CHKFLAGS(self, NPY_C_CONTIGUOUS)):
 *                 raise ValueError(u"ndarray is not C contiguous")             # <<<<<<<<<<<<<<
 * 
 *             if ((flags & pybuf.PyBUF_F_CONTIGUOUS == pybuf.PyBUF_F_CONTIGUOUS)
 */
    __pyx_t_4 = PyObject_Call(__pyx_builtin_ValueError, ((PyObject *)__pyx_k_tuple_4), NULL); if (unlikely(!__pyx_t_4)) {__pyx_filename = __pyx_f[2]; __pyx_lineno = 215; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_Raise(__pyx_t_4, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    {__pyx_filename = __pyx_f[2]; __pyx_lineno = 215; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    goto __pyx_L5;
  }
  __pyx_L5:;

  /* "numpy.pxd":217
 *                 raise ValueError(u"ndarray is not C contiguous")
 * 
 *             if ((flags & pybuf.PyBUF_F_CONTIGUOUS == pybuf.PyBUF_F_CONTIGUOUS)             # <<<<<<<<<<<<<<
 *                 and not PyArray_CHKFLAGS(self, NPY_F_CONTIGUOUS)):
 *                 raise ValueError(u"ndarray is not Fortran contiguous")
 */
  __pyx_t_3 = (((__pyx_v_flags & PyBUF_F_CONTIGUOUS) == PyBUF_F_CONTIGUOUS) != 0);
  if (__pyx_t_3) {

    /* "numpy.pxd":218
 * 
 *             if ((flags & pybuf.PyBUF_F_CONTIGUOUS == pybuf.PyBUF_F_CONTIGUOUS)
 *                 and not PyArray_CHKFLAGS(self, NPY_F_CONTIGUOUS)):             # <<<<<<<<<<<<<<
 *                 raise ValueError(u"ndarray is not Fortran contiguous")
 * 
 */
    __pyx_t_1 = ((!(PyArray_CHKFLAGS(__pyx_v_self, NPY_F_CONTIGUOUS) != 0)) != 0);
    __pyx_t_2 = __pyx_t_1;
  } else {
    __pyx_t_2 = __pyx_t_3;
  }
  if (__pyx_t_2) {

    /* "numpy.pxd":219
 *             if ((flags & pybuf.PyBUF_F_CONTIGUOUS == pybuf.PyBUF_F_CONTIGUOUS)
 *                 and not PyArray_CHKFLAGS(self, NPY_F_CONTIGUOUS)):
 *                 raise ValueError(u"ndarray is not Fortran contiguous")             # <<<<<<<<<<<<<<
 * 
 *             info.buf = PyArray_DATA(self)
 */
    __pyx_t_4 = PyObject_Call(__pyx_builtin_ValueError, ((PyObject *)__pyx_k_tuple_6), NULL); if (unlikely(!__pyx_t_4)) {__pyx_filename = __pyx_f[2]; __pyx_lineno = 219; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_Raise(__pyx_t_4, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    {__pyx_filename = __pyx_f[2]; __pyx_lineno = 219; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    goto __pyx_L6;
  }
  __pyx_L6:;

  /* "numpy.pxd":221
 *                 raise ValueError(u"ndarray is not Fortran contiguous")
 * 
 *             info.buf = PyArray_DATA(self)             # <<<<<<<<<<<<<<
 *             info.ndim = ndim
 *             if copy_shape:
 */
  __pyx_v_info->buf = PyArray_DATA(__pyx_v_self);

  /* "numpy.pxd":222
 * 
 *             info.buf = PyArray_DATA(self)
 *             info.ndim = ndim             # <<<<<<<<<<<<<<
 *             if copy_shape:
 *                 # Allocate new buffer for strides and shape info.
 */
  __pyx_v_info->ndim = __pyx_v_ndim;

  /* "numpy.pxd":223
 *             info.buf = PyArray_DATA(self)
 *             info.ndim = ndim
 *             if copy_shape:             # <<<<<<<<<<<<<<
 *                 # Allocate new buffer for strides and shape info.
 *                 # This is allocated as one block, strides first.
 */
  __pyx_t_2 = (__pyx_v_copy_shape != 0);
  if (__pyx_t_2) {

    /* "numpy.pxd":226
 *                 # Allocate new buffer for strides and shape info.
 *                 # This is allocated as one block, strides first.
 *                 info.strides = <Py_ssize_t*>stdlib.malloc(sizeof(Py_ssize_t) * <size_t>ndim * 2)             # <<<<<<<<<<<<<<
 *                 info.shape = info.strides + ndim
 *                 for i in range(ndim):
 */
    __pyx_v_info->strides = ((Py_ssize_t *)malloc((((sizeof(Py_ssize_t)) * ((size_t)__pyx_v_ndim)) * 2)));

    /* "numpy.pxd":227
 *                 # This is allocated as one block, strides first.
 *                 info.strides = <Py_ssize_t*>stdlib.malloc(sizeof(Py_ssize_t) * <size_t>ndim * 2)
 *                 info.shape = info.strides + ndim             # <<<<<<<<<<<<<<
 *                 for i in range(ndim):
 *                     info.strides[i] = PyArray_STRIDES(self)[i]
 */
    __pyx_v_info->shape = (__pyx_v_info->strides + __pyx_v_ndim);

    /* "numpy.pxd":228
 *                 info.strides = <Py_ssize_t*>stdlib.malloc(sizeof(Py_ssize_t) * <size_t>ndim * 2)
 *                 info.shape = info.strides + ndim
 *                 for i in range(ndim):             # <<<<<<<<<<<<<<
 *                     info.strides[i] = PyArray_STRIDES(self)[i]
 *                     info.shape[i] = PyArray_DIMS(self)[i]
 */
    __pyx_t_5 = __pyx_v_ndim;
    for (__pyx_t_6 = 0; __pyx_t_6 < __pyx_t_5; __pyx_t_6+=1) {
      __pyx_v_i = __pyx_t_6;

      /* "numpy.pxd":229
 *                 info.shape = info.strides + ndim
 *                 for i in range(ndim):
 *                     info.strides[i] = PyArray_STRIDES(self)[i]             # <<<<<<<<<<<<<<
 *                     info.shape[i] = PyArray_DIMS(self)[i]
 *             else:
 */
      (__pyx_v_info->strides[__pyx_v_i]) = (PyArray_STRIDES(__pyx_v_self)[__pyx_v_i]);

      /* "numpy.pxd":230
 *                 for i in range(ndim):
 *                     info.strides[i] = PyArray_STRIDES(self)[i]
 *                     info.shape[i] = PyArray_DIMS(self)[i]             # <<<<<<<<<<<<<<
 *             else:
 *                 info.strides = <Py_ssize_t*>PyArray_STRIDES(self)
 */
      (__pyx_v_info->shape[__pyx_v_i]) = (PyArray_DIMS(__pyx_v_self)[__pyx_v_i]);
    }
    goto __pyx_L7;
  }
  /*else*/ {

    /* "numpy.pxd":232
 *                     info.shape[i] = PyArray_DIMS(self)[i]
 *             else:
 *                 info.strides = <Py_ssize_t*>PyArray_STRIDES(self)             # <<<<<<<<<<<<<<
 *                 info.shape = <Py_ssize_t*>PyArray_DIMS(self)
 *             info.suboffsets = NULL
 */
    __pyx_v_info->strides = ((Py_ssize_t *)PyArray_STRIDES(__pyx_v_self));

    /* "numpy.pxd":233
 *             else:
 *                 info.strides = <Py_ssize_t*>PyArray_STRIDES(self)
 *                 info.shape = <Py_ssize_t*>PyArray_DIMS(self)             # <<<<<<<<<<<<<<
 *             info.suboffsets = NULL
 *             info.itemsize = PyArray_ITEMSIZE(self)
 */
    __pyx_v_info->shape = ((Py_ssize_t *)PyArray_DIMS(__pyx_v_self));
  }
  __pyx_L7:;

  /* "numpy.pxd":234
 *                 info.strides = <Py_ssize_t*>PyArray_STRIDES(self)
 *                 info.shape = <Py_ssize_t*>PyArray_DIMS(self)
 *             info.suboffsets = NULL             # <<<<<<<<<<<<<<
 *             info.itemsize = PyArray_ITEMSIZE(self)
 *             info.readonly = not PyArray_ISWRITEABLE(self)
 */
  __pyx_v_info->suboffsets = NULL;

  /* "numpy.pxd":235
 *                 info.shape = <Py_ssize_t*>PyArray_DIMS(self)
 *             info.suboffsets = NULL
 *             info.itemsize = PyArray_ITEMSIZE(self)             # <<<<<<<<<<<<<<
 *             info.readonly = not PyArray_ISWRITEABLE(self)
 * 
 */
  __pyx_v_info->itemsize = PyArray_ITEMSIZE(__pyx_v_self);

  /* "numpy.pxd":236
 *             info.suboffsets = NULL
 *             info.itemsize = PyArray_ITEMSIZE(self)
 *             info.readonly = not PyArray_ISWRITEABLE(self)             # <<<<<<<<<<<<<<
 * 
 *             cdef int t
 */
  __pyx_v_info->readonly = (!(PyArray_ISWRITEABLE(__pyx_v_self) != 0));

  /* "numpy.pxd":239
 * 
 *             cdef int t
 *             cdef char* f = NULL             # <<<<<<<<<<<<<<
 *             cdef dtype descr = self.descr
 *             cdef list stack
 */
  __pyx_v_f = NULL;

  /* "numpy.pxd":240
 *             cdef int t
 *             cdef char* f = NULL
 *             cdef dtype descr = self.descr             # <<<<<<<<<<<<<<
 *             cdef list stack
 *             cdef int offset
 */
  __pyx_t_4 = ((PyObject *)__pyx_v_self->descr);
  __Pyx_INCREF(__pyx_t_4);
  __pyx_v_descr = ((PyArray_Descr *)__pyx_t_4);
  __pyx_t_4 = 0;

  /* "numpy.pxd":244
 *             cdef int offset
 * 
 *             cdef bint hasfields = PyDataType_HASFIELDS(descr)             # <<<<<<<<<<<<<<
 * 
 *             if not hasfields and not copy_shape:
 */
  __pyx_v_hasfields = PyDataType_HASFIELDS(__pyx_v_descr);

  /* "numpy.pxd":246
 *             cdef bint hasfields = PyDataType_HASFIELDS(descr)
 * 
 *             if not hasfields and not copy_shape:             # <<<<<<<<<<<<<<
 *                 # do not call releasebuffer
 *                 info.obj = None
 */
  __pyx_t_2 = ((!(__pyx_v_hasfields != 0)) != 0);
  if (__pyx_t_2) {
    __pyx_t_3 = ((!(__pyx_v_copy_shape != 0)) != 0);
    __pyx_t_1 = __pyx_t_3;
  } else {
    __pyx_t_1 = __pyx_t_2;
  }
  if (__pyx_t_1) {

    /* "numpy.pxd":248
 *             if not hasfields and not copy_shape:
 *                 # do not call releasebuffer
 *                 info.obj = None             # <<<<<<<<<<<<<<
 *             else:
 *                 # need to call releasebuffer
 */
    __Pyx_INCREF(Py_None);
    __Pyx_GIVEREF(Py_None);
    __Pyx_GOTREF(__pyx_v_info->obj);
    __Pyx_DECREF(__pyx_v_info->obj);
    __pyx_v_info->obj = Py_None;
    goto __pyx_L10;
  }
  /*else*/ {

    /* "numpy.pxd":251
 *             else:
 *                 # need to call releasebuffer
 *                 info.obj = self             # <<<<<<<<<<<<<<
 * 
 *             if not hasfields:
 */
    __Pyx_INCREF(((PyObject *)__pyx_v_self));
    __Pyx_GIVEREF(((PyObject *)__pyx_v_self));
    __Pyx_GOTREF(__pyx_v_info->obj);
    __Pyx_DECREF(__pyx_v_info->obj);
    __pyx_v_info->obj = ((PyObject *)__pyx_v_self);
  }
  __pyx_L10:;

  /* "numpy.pxd":253
 *                 info.obj = self
 * 
 *             if not hasfields:             # <<<<<<<<<<<<<<
 *                 t = descr.type_num
 *                 if ((descr.byteorder == c'>' and little_endian) or
 */
  __pyx_t_1 = ((!(__pyx_v_hasfields != 0)) != 0);
  if (__pyx_t_1) {

    /* "numpy.pxd":254
 * 
 *             if not hasfields:
 *                 t = descr.type_num             # <<<<<<<<<<<<<<
 *                 if ((descr.byteorder == c'>' and little_endian) or
 *                     (descr.byteorder == c'<' and not little_endian)):
 */
    __pyx_t_5 = __pyx_v_descr->type_num;
    __pyx_v_t = __pyx_t_5;

    /* "numpy.pxd":255
 *             if not hasfields:
 *                 t = descr.type_num
 *                 if ((descr.byteorder == c'>' and little_endian) or             # <<<<<<<<<<<<<<
 *                     (descr.byteorder == c'<' and not little_endian)):
 *                     raise ValueError(u"Non-native byte order not supported")
 */
    __pyx_t_1 = ((__pyx_v_descr->byteorder == '>') != 0);
    if (__pyx_t_1) {
      __pyx_t_2 = (__pyx_v_little_endian != 0);
    } else {
      __pyx_t_2 = __pyx_t_1;
    }
    if (!__pyx_t_2) {

      /* "numpy.pxd":256
 *                 t = descr.type_num
 *                 if ((descr.byteorder == c'>' and little_endian) or
 *                     (descr.byteorder == c'<' and not little_endian)):             # <<<<<<<<<<<<<<
 *                     raise ValueError(u"Non-native byte order not supported")
 *                 if   t == NPY_BYTE:        f = "b"
 */
      __pyx_t_1 = ((__pyx_v_descr->byteorder == '<') != 0);
      if (__pyx_t_1) {
        __pyx_t_3 = ((!(__pyx_v_little_endian != 0)) != 0);
        __pyx_t_7 = __pyx_t_3;
      } else {
        __pyx_t_7 = __pyx_t_1;
      }
      __pyx_t_1 = __pyx_t_7;
    } else {
      __pyx_t_1 = __pyx_t_2;
    }
    if (__pyx_t_1) {

      /* "numpy.pxd":257
 *                 if ((descr.byteorder == c'>' and little_endian) or
 *                     (descr.byteorder == c'<' and not little_endian)):
 *                     raise ValueError(u"Non-native byte order not supported")             # <<<<<<<<<<<<<<
 *                 if   t == NPY_BYTE:        f = "b"
 *                 elif t == NPY_UBYTE:       f = "B"
 */
      __pyx_t_4 = PyObject_Call(__pyx_builtin_ValueError, ((PyObject *)__pyx_k_tuple_8), NULL); if (unlikely(!__pyx_t_4)) {__pyx_filename = __pyx_f[2]; __pyx_lineno = 257; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_Raise(__pyx_t_4, 0, 0, 0);
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      {__pyx_filename = __pyx_f[2]; __pyx_lineno = 257; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
      goto __pyx_L12;
    }
    __pyx_L12:;

    /* "numpy.pxd":274
 *                 elif t == NPY_CDOUBLE:     f = "Zd"
 *                 elif t == NPY_CLONGDOUBLE: f = "Zg"
 *                 elif t == NPY_OBJECT:      f = "O"             # <<<<<<<<<<<<<<
 *                 else:
 *                     raise ValueError(u"unknown dtype code in numpy.pxd (%d)" % t)
 */
    switch (__pyx_v_t) {

      /* "numpy.pxd":258
 *                     (descr.byteorder == c'<' and not little_endian)):
 *                     raise ValueError(u"Non-native byte order not supported")
 *                 if   t == NPY_BYTE:        f = "b"             # <<<<<<<<<<<<<<
 *                 elif t == NPY_UBYTE:       f = "B"
 *                 elif t == NPY_SHORT:       f = "h"
 */
      case NPY_BYTE:
      __pyx_v_f = __pyx_k__b;
      break;

      /* "numpy.pxd":259
 *                     raise ValueError(u"Non-native byte order not supported")
 *                 if   t == NPY_BYTE:        f = "b"
 *                 elif t == NPY_UBYTE:       f = "B"             # <<<<<<<<<<<<<<
 *                 elif t == NPY_SHORT:       f = "h"
 *                 elif t == NPY_USHORT:      f = "H"
 */
      case NPY_UBYTE:
      __pyx_v_f = __pyx_k__B;
      break;

      /* "numpy.pxd":260
 *                 if   t == NPY_BYTE:        f = "b"
 *                 elif t == NPY_UBYTE:       f = "B"
 *                 elif t == NPY_SHORT:       f = "h"             # <<<<<<<<<<<<<<
 *                 elif t == NPY_USHORT:      f = "H"
 *                 elif t == NPY_INT:         f = "i"
 */
      case NPY_SHORT:
      __pyx_v_f = __pyx_k__h;
      break;

      /* "numpy.pxd":261
 *                 elif t == NPY_UBYTE:       f = "B"
 *                 elif t == NPY_SHORT:       f = "h"
 *                 elif t == NPY_USHORT:      f = "H"             # <<<<<<<<<<<<<<
 *                 elif t == NPY_INT:         f = "i"
 *                 elif t == NPY_UINT:        f = "I"
 */
      case NPY_USHORT:
      __pyx_v_f = __pyx_k__H;
      break;

      /* "numpy.pxd":262
 *                 elif t == NPY_SHORT:       f = "h"
 *                 elif t == NPY_USHORT:      f = "H"
 *                 elif t == NPY_INT:         f = "i"             # <<<<<<<<<<<<<<
 *                 elif t == NPY_UINT:        f = "I"
 *                 elif t == NPY_LONG:        f = "l"
 */
      case NPY_INT:
      __pyx_v_f = __pyx_k__i;
      break;

      /* "numpy.pxd":263
 *                 elif t == NPY_USHORT:      f = "H"
 *                 elif t == NPY_INT:         f = "i"
 *                 elif t == NPY_UINT:        f = "I"             # <<<<<<<<<<<<<<
 *                 elif t == NPY_LONG:        f = "l"
 *                 elif t == NPY_ULONG:       f = "L"
 */
      case NPY_UINT:
      __pyx_v_f = __pyx_k__I;
      break;

      /* "numpy.pxd":264
 *                 elif t == NPY_INT:         f = "i"
 *                 elif t == NPY_UINT:        f = "I"
 *                 elif t == NPY_LONG:        f = "l"             # <<<<<<<<<<<<<<
 *                 elif t == NPY_ULONG:       f = "L"
 *                 elif t == NPY_LONGLONG:    f = "q"
 */
      case NPY_LONG:
      __pyx_v_f = __pyx_k__l;
      break;

      /* "numpy.pxd":265
 *                 elif t == NPY_UINT:        f = "I"
 *                 elif t == NPY_LONG:        f = "l"
 *                 elif t == NPY_ULONG:       f = "L"             # <<<<<<<<<<<<<<
 *                 elif t == NPY_LONGLONG:    f = "q"
 *                 elif t == NPY_ULONGLONG:   f = "Q"
 */
      case NPY_ULONG:
      __pyx_v_f = __pyx_k__L;
      break;

      /* "numpy.pxd":266
 *                 elif t == NPY_LONG:        f = "l"
 *                 elif t == NPY_ULONG:       f = "L"
 *                 elif t == NPY_LONGLONG:    f = "q"             # <<<<<<<<<<<<<<
 *                 elif t == NPY_ULONGLONG:   f = "Q"
 *                 elif t == NPY_FLOAT:       f = "f"
 */
      case NPY_LONGLONG:
      __pyx_v_f = __pyx_k__q;
      break;

      /* "numpy.pxd":267
 *                 elif t == NPY_ULONG:       f = "L"
 *                 elif t == NPY_LONGLONG:    f = "q"
 *                 elif t == NPY_ULONGLONG:   f = "Q"             # <<<<<<<<<<<<<<
 *                 elif t == NPY_FLOAT:       f = "f"
 *                 elif t == NPY_DOUBLE:      f = "d"
 */
      case NPY_ULONGLONG:
      __pyx_v_f = __pyx_k__Q;
      break;

      /* "numpy.pxd":268
 *                 elif t == NPY_LONGLONG:    f = "q"
 *                 elif t == NPY_ULONGLONG:   f = "Q"
 *                 elif t == NPY_FLOAT:       f = "f"             # <<<<<<<<<<<<<<
 *                 elif t == NPY_DOUBLE:      f = "d"
 *                 elif t == NPY_LONGDOUBLE:  f = "g"
 */
      case NPY_FLOAT:
      __pyx_v_f = __pyx_k__f;
      break;

      /* "numpy.pxd":269
 *                 elif t == NPY_ULONGLONG:   f = "Q"
 *                 elif t == NPY_FLOAT:       f = "f"
 *                 elif t == NPY_DOUBLE:      f = "d"             # <<<<<<<<<<<<<<
 *                 elif t == NPY_LONGDOUBLE:  f = "g"
 *                 elif t == NPY_CFLOAT:      f = "Zf"
 */
      case NPY_DOUBLE:
      __pyx_v_f = __pyx_k__d;
      break;

      /* "numpy.pxd":270
 *                 elif t == NPY_FLOAT:       f = "f"
 *                 elif t == NPY_DOUBLE:      f = "d"
 *                 elif t == NPY_LONGDOUBLE:  f = "g"             # <<<<<<<<<<<<<<
 *                 elif t == NPY_CFLOAT:      f = "Zf"
 *                 elif t == NPY_CDOUBLE:     f = "Zd"
 */
      case NPY_LONGDOUBLE:
      __pyx_v_f = __pyx_k__g;
      break;

      /* "numpy.pxd":271
 *                 elif t == NPY_DOUBLE:      f = "d"
 *                 elif t == NPY_LONGDOUBLE:  f = "g"
 *                 elif t == NPY_CFLOAT:      f = "Zf"             # <<<<<<<<<<<<<<
 *                 elif t == NPY_CDOUBLE:     f = "Zd"
 *                 elif t == NPY_CLONGDOUBLE: f = "Zg"
 */
      case NPY_CFLOAT:
      __pyx_v_f = __pyx_k__Zf;
      break;

      /* "numpy.pxd":272
 *                 elif t == NPY_LONGDOUBLE:  f = "g"
 *                 elif t == NPY_CFLOAT:      f = "Zf"
 *                 elif t == NPY_CDOUBLE:     f = "Zd"             # <<<<<<<<<<<<<<
 *                 elif t == NPY_CLONGDOUBLE: f = "Zg"
 *                 elif t == NPY_OBJECT:      f = "O"
 */
      case NPY_CDOUBLE:
      __pyx_v_f = __pyx_k__Zd;
      break;

      /* "numpy.pxd":273
 *                 elif t == NPY_CFLOAT:      f = "Zf"
 *                 elif t == NPY_CDOUBLE:     f = "Zd"
 *                 elif t == NPY_CLONGDOUBLE: f = "Zg"             # <<<<<<<<<<<<<<
 *                 elif t == NPY_OBJECT:      f = "O"
 *                 else:
 */
      case NPY_CLONGDOUBLE:
      __pyx_v_f = __pyx_k__Zg;
      break;

      /* "numpy.pxd":274
 *                 elif t == NPY_CDOUBLE:     f = "Zd"
 *                 elif t == NPY_CLONGDOUBLE: f = "Zg"
 *                 elif t == NPY_OBJECT:      f = "O"             # <<<<<<<<<<<<<<
 *                 else:
 *                     raise ValueError(u"unknown dtype code in numpy.pxd (%d)" % t)
 */
      case NPY_OBJECT:
      __pyx_v_f = __pyx_k__O;
      break;
      default:

      /* "numpy.pxd":276
 *                 elif t == NPY_OBJECT:      f = "O"
 *                 else:
 *                     raise ValueError(u"unknown dtype code in numpy.pxd (%d)" % t)             # <<<<<<<<<<<<<<
 *                 info.format = f
 *                 return
 */
      __pyx_t_4 = PyInt_FromLong(__pyx_v_t); if (unlikely(!__pyx_t_4)) {__pyx_filename = __pyx_f[2]; __pyx_lineno = 276; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
      __Pyx_GOTREF(__pyx_t_4);
      __pyx_t_8 = PyNumber_Remainder(((PyObject *)__pyx_kp_u_9), __pyx_t_4); if (unlikely(!__pyx_t_8)) {__pyx_filename = __pyx_f[2]; __pyx_lineno = 276; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
      __Pyx_GOTREF(((PyObject *)__pyx_t_8));
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __pyx_t_4 = PyTuple_New(1); if (unlikely(!__pyx_t_4)) {__pyx_filename = __pyx_f[2]; __pyx_lineno = 276; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
      __Pyx_GOTREF(__pyx_t_4);
      PyTuple_SET_ITEM(__pyx_t_4, 0, ((PyObject *)__pyx_t_8));
      __Pyx_GIVEREF(((PyObject *)__pyx_t_8));
      __pyx_t_8 = 0;
      __pyx_t_8 = PyObject_Call(__pyx_builtin_ValueError, ((PyObject *)__pyx_t_4), NULL); if (unlikely(!__pyx_t_8)) {__pyx_filename = __pyx_f[2]; __pyx_lineno = 276; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
      __Pyx_GOTREF(__pyx_t_8);
      __Pyx_DECREF(((PyObject *)__pyx_t_4)); __pyx_t_4 = 0;
      __Pyx_Raise(__pyx_t_8, 0, 0, 0);
      __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
      {__pyx_filename = __pyx_f[2]; __pyx_lineno = 276; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
      break;
    }

    /* "numpy.pxd":277
 *                 else:
 *                     raise ValueError(u"unknown dtype code in numpy.pxd (%d)" % t)
 *                 info.format = f             # <<<<<<<<<<<<<<
 *                 return
 *             else:
 */
    __pyx_v_info->format = __pyx_v_f;

    /* "numpy.pxd":278
 *                     raise ValueError(u"unknown dtype code in numpy.pxd (%d)" % t)
 *                 info.format = f
 *                 return             # <<<<<<<<<<<<<<
 *             else:
 *                 info.format = <char*>stdlib.malloc(_buffer_format_string_len)
 */
    __pyx_r = 0;
    goto __pyx_L0;
    goto __pyx_L11;
  }
  /*else*/ {

    /* "numpy.pxd":280
 *                 return
 *             else:
 *                 info.format = <char*>stdlib.malloc(_buffer_format_string_len)             # <<<<<<<<<<<<<<
 *                 info.format[0] = c'^' # Native data types, manual alignment
 *                 offset = 0
 */
    __pyx_v_info->format = ((char *)malloc(255));

    /* "numpy.pxd":281
 *             else:
 *                 info.format = <char*>stdlib.malloc(_buffer_format_string_len)
 *                 info.format[0] = c'^' # Native data types, manual alignment             # <<<<<<<<<<<<<<
 *                 offset = 0
 *                 f = _util_dtypestring(descr, info.format + 1,
 */
    (__pyx_v_info->format[0]) = '^';

    /* "numpy.pxd":282
 *                 info.format = <char*>stdlib.malloc(_buffer_format_string_len)
 *                 info.format[0] = c'^' # Native data types, manual alignment
 *                 offset = 0             # <<<<<<<<<<<<<<
 *                 f = _util_dtypestring(descr, info.format + 1,
 *                                       info.format + _buffer_format_string_len,
 */
    __pyx_v_offset = 0;

    /* "numpy.pxd":285
 *                 f = _util_dtypestring(descr, info.format + 1,
 *                                       info.format + _buffer_format_string_len,
 *                                       &offset)             # <<<<<<<<<<<<<<
 *                 f[0] = c'\0' # Terminate format string
 * 
 */
    __pyx_t_9 = __pyx_f_5numpy__util_dtypestring(__pyx_v_descr, (__pyx_v_info->format + 1), (__pyx_v_info->format + 255), (&__pyx_v_offset)); if (unlikely(__pyx_t_9 == NULL)) {__pyx_filename = __pyx_f[2]; __pyx_lineno = 283; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    __pyx_v_f = __pyx_t_9;

    /* "numpy.pxd":286
 *                                       info.format + _buffer_format_string_len,
 *                                       &offset)
 *                 f[0] = c'\0' # Terminate format string             # <<<<<<<<<<<<<<
 * 
 *         def __releasebuffer__(ndarray self, Py_buffer* info):
 */
    (__pyx_v_f[0]) = '\x00';
  }
  __pyx_L11:;

  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_AddTraceback("numpy.ndarray.__getbuffer__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  if (__pyx_v_info != NULL && __pyx_v_info->obj != NULL) {
    __Pyx_GOTREF(__pyx_v_info->obj);
    __Pyx_DECREF(__pyx_v_info->obj); __pyx_v_info->obj = NULL;
  }
  goto __pyx_L2;
  __pyx_L0:;
  if (__pyx_v_info != NULL && __pyx_v_info->obj == Py_None) {
    __Pyx_GOTREF(Py_None);
    __Pyx_DECREF(Py_None); __pyx_v_info->obj = NULL;
  }
  __pyx_L2:;
  __Pyx_XDECREF((PyObject *)__pyx_v_descr);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static CYTHON_UNUSED void __pyx_pw_5numpy_7ndarray_3__releasebuffer__(PyObject *__pyx_v_self, Py_buffer *__pyx_v_info); /*proto*/
static CYTHON_UNUSED void __pyx_pw_5numpy_7ndarray_3__releasebuffer__(PyObject *__pyx_v_self, Py_buffer *__pyx_v_info) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__releasebuffer__ (wrapper)", 0);
  __pyx_pf_5numpy_7ndarray_2__releasebuffer__(((PyArrayObject *)__pyx_v_self), ((Py_buffer *)__pyx_v_info));
  __Pyx_RefNannyFinishContext();
}

/* "numpy.pxd":288
 *                 f[0] = c'\0' # Terminate format string
 * 
 *         def __releasebuffer__(ndarray self, Py_buffer* info):             # <<<<<<<<<<<<<<
 *             if PyArray_HASFIELDS(self):
 *                 stdlib.free(info.format)
 */

static void __pyx_pf_5numpy_7ndarray_2__releasebuffer__(PyArrayObject *__pyx_v_self, Py_buffer *__pyx_v_info) {
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  __Pyx_RefNannySetupContext("__releasebuffer__", 0);

  /* "numpy.pxd":289
 * 
 *         def __releasebuffer__(ndarray self, Py_buffer* info):
 *             if PyArray_HASFIELDS(self):             # <<<<<<<<<<<<<<
 *                 stdlib.free(info.format)
 *             if sizeof(npy_intp) != sizeof(Py_ssize_t):
 */
  __pyx_t_1 = (PyArray_HASFIELDS(__pyx_v_self) != 0);
  if (__pyx_t_1) {

    /* "numpy.pxd":290
 *         def __releasebuffer__(ndarray self, Py_buffer* info):
 *             if PyArray_HASFIELDS(self):
 *                 stdlib.free(info.format)             # <<<<<<<<<<<<<<
 *             if sizeof(npy_intp) != sizeof(Py_ssize_t):
 *                 stdlib.free(info.strides)
 */
    free(__pyx_v_info->format);
    goto __pyx_L3;
  }
  __pyx_L3:;

  /* "numpy.pxd":291
 *             if PyArray_HASFIELDS(self):
 *                 stdlib.free(info.format)
 *             if sizeof(npy_intp) != sizeof(Py_ssize_t):             # <<<<<<<<<<<<<<
 *                 stdlib.free(info.strides)
 *                 # info.shape was stored after info.strides in the same block
 */
  __pyx_t_1 = (((sizeof(npy_intp)) != (sizeof(Py_ssize_t))) != 0);
  if (__pyx_t_1) {

    /* "numpy.pxd":292
 *                 stdlib.free(info.format)
 *             if sizeof(npy_intp) != sizeof(Py_ssize_t):
 *                 stdlib.free(info.strides)             # <<<<<<<<<<<<<<
 *                 # info.shape was stored after info.strides in the same block
 * 
 */
    free(__pyx_v_info->strides);
    goto __pyx_L4;
  }
  __pyx_L4:;

  __Pyx_RefNannyFinishContext();
}

/* "numpy.pxd":768
 * ctypedef npy_cdouble     complex_t
 * 
 * cdef inline object PyArray_MultiIterNew1(a):             # <<<<<<<<<<<<<<
 *     return PyArray_MultiIterNew(1, <void*>a)
 * 
 */

static CYTHON_INLINE PyObject *__pyx_f_5numpy_PyArray_MultiIterNew1(PyObject *__pyx_v_a) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("PyArray_MultiIterNew1", 0);

  /* "numpy.pxd":769
 * 
 * cdef inline object PyArray_MultiIterNew1(a):
 *     return PyArray_MultiIterNew(1, <void*>a)             # <<<<<<<<<<<<<<
 * 
 * cdef inline object PyArray_MultiIterNew2(a, b):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = PyArray_MultiIterNew(1, ((void *)__pyx_v_a)); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[2]; __pyx_lineno = 769; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("numpy.PyArray_MultiIterNew1", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "numpy.pxd":771
 *     return PyArray_MultiIterNew(1, <void*>a)
 * 
 * cdef inline object PyArray_MultiIterNew2(a, b):             # <<<<<<<<<<<<<<
 *     return PyArray_MultiIterNew(2, <void*>a, <void*>b)
 * 
 */

static CYTHON_INLINE PyObject *__pyx_f_5numpy_PyArray_MultiIterNew2(PyObject *__pyx_v_a, PyObject *__pyx_v_b) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("PyArray_MultiIterNew2", 0);

  /* "numpy.pxd":772
 * 
 * cdef inline object PyArray_MultiIterNew2(a, b):
 *     return PyArray_MultiIterNew(2, <void*>a, <void*>b)             # <<<<<<<<<<<<<<
 * 
 * cdef inline object PyArray_MultiIterNew3(a, b, c):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = PyArray_MultiIterNew(2, ((void *)__pyx_v_a), ((void *)__pyx_v_b)); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[2]; __pyx_lineno = 772; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("numpy.PyArray_MultiIterNew2", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "numpy.pxd":774
 *     return PyArray_MultiIterNew(2, <void*>a, <void*>b)
 * 
 * cdef inline object PyArray_MultiIterNew3(a, b, c):             # <<<<<<<<<<<<<<
 *     return PyArray_MultiIterNew(3, <void*>a, <void*>b, <void*> c)
 * 
 */

static CYTHON_INLINE PyObject *__pyx_f_5numpy_PyArray_MultiIterNew3(PyObject *__pyx_v_a, PyObject *__pyx_v_b, PyObject *__pyx_v_c) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("PyArray_MultiIterNew3", 0);

  /* "numpy.pxd":775
 * 
 * cdef inline object PyArray_MultiIterNew3(a, b, c):
 *     return PyArray_MultiIterNew(3, <void*>a, <void*>b, <void*> c)             # <<<<<<<<<<<<<<
 * 
 * cdef inline object PyArray_MultiIterNew4(a, b, c, d):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = PyArray_MultiIterNew(3, ((void *)__pyx_v_a), ((void *)__pyx_v_b), ((void *)__pyx_v_c)); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[2]; __pyx_lineno = 775; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("numpy.PyArray_MultiIterNew3", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "numpy.pxd":777
 *     return PyArray_MultiIterNew(3, <void*>a, <void*>b, <void*> c)
 * 
 * cdef inline object PyArray_MultiIterNew4(a, b, c, d):             # <<<<<<<<<<<<<<
 *     return PyArray_MultiIterNew(4, <void*>a, <void*>b, <void*>c, <void*> d)
 * 
 */

static CYTHON_INLINE PyObject *__pyx_f_5numpy_PyArray_MultiIterNew4(PyObject *__pyx_v_a, PyObject *__pyx_v_b, PyObject *__pyx_v_c, PyObject *__pyx_v_d) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("PyArray_MultiIterNew4", 0);

  /* "numpy.pxd":778
 * 
 * cdef inline object PyArray_MultiIterNew4(a, b, c, d):
 *     return PyArray_MultiIterNew(4, <void*>a, <void*>b, <void*>c, <void*> d)             # <<<<<<<<<<<<<<
 * 
 * cdef inline object PyArray_MultiIterNew5(a, b, c, d, e):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = PyArray_MultiIterNew(4, ((void *)__pyx_v_a), ((void *)__pyx_v_b), ((void *)__pyx_v_c), ((void *)__pyx_v_d)); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[2]; __pyx_lineno = 778; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("numpy.PyArray_MultiIterNew4", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "numpy.pxd":780
 *     return PyArray_MultiIterNew(4, <void*>a, <void*>b, <void*>c, <void*> d)
 * 
 * cdef inline object PyArray_MultiIterNew5(a, b, c, d, e):             # <<<<<<<<<<<<<<
 *     return PyArray_MultiIterNew(5, <void*>a, <void*>b, <void*>c, <void*> d, <void*> e)
 * 
 */

static CYTHON_INLINE PyObject *__pyx_f_5numpy_PyArray_MultiIterNew5(PyObject *__pyx_v_a, PyObject *__pyx_v_b, PyObject *__pyx_v_c, PyObject *__pyx_v_d, PyObject *__pyx_v_e) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("PyArray_MultiIterNew5", 0);

  /* "numpy.pxd":781
 * 
 * cdef inline object PyArray_MultiIterNew5(a, b, c, d, e):
 *     return PyArray_MultiIterNew(5, <void*>a, <void*>b, <void*>c, <void*> d, <void*> e)             # <<<<<<<<<<<<<<
 * 
 * cdef inline char* _util_dtypestring(dtype descr, char* f, char* end, int* offset) except NULL:
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = PyArray_MultiIterNew(5, ((void *)__pyx_v_a), ((void *)__pyx_v_b), ((void *)__pyx_v_c), ((void *)__pyx_v_d), ((void *)__pyx_v_e)); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[2]; __pyx_lineno = 781; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("numpy.PyArray_MultiIterNew5", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "numpy.pxd":783
 *     return PyArray_MultiIterNew(5, <void*>a, <void*>b, <void*>c, <void*> d, <void*> e)
 * 
 * cdef inline char* _util_dtypestring(dtype descr, char* f, char* end, int* offset) except NULL:             # <<<<<<<<<<<<<<
 *     # Recursive utility function used in __getbuffer__ to get format
 *     # string. The new location in the format string is returned.
 */

static CYTHON_INLINE char *__pyx_f_5numpy__util_dtypestring(PyArray_Descr *__pyx_v_descr, char *__pyx_v_f, char *__pyx_v_end, int *__pyx_v_offset) {
  PyArray_Descr *__pyx_v_child = 0;
  int __pyx_v_endian_detector;
  int __pyx_v_little_endian;
  PyObject *__pyx_v_fields = 0;
  PyObject *__pyx_v_childname = NULL;
  PyObject *__pyx_v_new_offset = NULL;
  PyObject *__pyx_v_t = NULL;
  char *__pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  Py_ssize_t __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *(*__pyx_t_6)(PyObject *);
  int __pyx_t_7;
  int __pyx_t_8;
  int __pyx_t_9;
  int __pyx_t_10;
  long __pyx_t_11;
  char *__pyx_t_12;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_util_dtypestring", 0);

  /* "numpy.pxd":790
 *     cdef int delta_offset
 *     cdef tuple i
 *     cdef int endian_detector = 1             # <<<<<<<<<<<<<<
 *     cdef bint little_endian = ((<char*>&endian_detector)[0] != 0)
 *     cdef tuple fields
 */
  __pyx_v_endian_detector = 1;

  /* "numpy.pxd":791
 *     cdef tuple i
 *     cdef int endian_detector = 1
 *     cdef bint little_endian = ((<char*>&endian_detector)[0] != 0)             # <<<<<<<<<<<<<<
 *     cdef tuple fields
 * 
 */
  __pyx_v_little_endian = ((((char *)(&__pyx_v_endian_detector))[0]) != 0);

  /* "numpy.pxd":794
 *     cdef tuple fields
 * 
 *     for childname in descr.names:             # <<<<<<<<<<<<<<
 *         fields = descr.fields[childname]
 *         child, new_offset = fields
 */
  if (unlikely(((PyObject *)__pyx_v_descr->names) == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not iterable");
    {__pyx_filename = __pyx_f[2]; __pyx_lineno = 794; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  }
  __pyx_t_1 = ((PyObject *)__pyx_v_descr->names); __Pyx_INCREF(__pyx_t_1); __pyx_t_2 = 0;
  for (;;) {
    if (__pyx_t_2 >= PyTuple_GET_SIZE(__pyx_t_1)) break;
    #if CYTHON_COMPILING_IN_CPYTHON
    __pyx_t_3 = PyTuple_GET_ITEM(__pyx_t_1, __pyx_t_2); __Pyx_INCREF(__pyx_t_3); __pyx_t_2++; if (unlikely(0 < 0)) {__pyx_filename = __pyx_f[2]; __pyx_lineno = 794; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    #else
    __pyx_t_3 = PySequence_ITEM(__pyx_t_1, __pyx_t_2); __pyx_t_2++; if (unlikely(!__pyx_t_3)) {__pyx_filename = __pyx_f[2]; __pyx_lineno = 794; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    #endif
    __Pyx_XDECREF(__pyx_v_childname);
    __pyx_v_childname = __pyx_t_3;
    __pyx_t_3 = 0;

    /* "numpy.pxd":795
 * 
 *     for childname in descr.names:
 *         fields = descr.fields[childname]             # <<<<<<<<<<<<<<
 *         child, new_offset = fields
 * 
 */
    __pyx_t_3 = PyObject_GetItem(__pyx_v_descr->fields, __pyx_v_childname); if (!__pyx_t_3) {__pyx_filename = __pyx_f[2]; __pyx_lineno = 795; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    __Pyx_GOTREF(__pyx_t_3);
    if (!(likely(PyTuple_CheckExact(__pyx_t_3))||((__pyx_t_3) == Py_None)||(PyErr_Format(PyExc_TypeError, "Expected tuple, got %.200s", Py_TYPE(__pyx_t_3)->tp_name), 0))) {__pyx_filename = __pyx_f[2]; __pyx_lineno = 795; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    __Pyx_XDECREF(((PyObject *)__pyx_v_fields));
    __pyx_v_fields = ((PyObject*)__pyx_t_3);
    __pyx_t_3 = 0;

    /* "numpy.pxd":796
 *     for childname in descr.names:
 *         fields = descr.fields[childname]
 *         child, new_offset = fields             # <<<<<<<<<<<<<<
 * 
 *         if (end - f) - (new_offset - offset[0]) < 15:
 */
    if (likely(PyTuple_CheckExact(((PyObject *)__pyx_v_fields)))) {
      PyObject* sequence = ((PyObject *)__pyx_v_fields);
      #if CYTHON_COMPILING_IN_CPYTHON
      Py_ssize_t size = Py_SIZE(sequence);
      #else
      Py_ssize_t size = PySequence_Size(sequence);
      #endif
      if (unlikely(size != 2)) {
        if (size > 2) __Pyx_RaiseTooManyValuesError(2);
        else if (size >= 0) __Pyx_RaiseNeedMoreValuesError(size);
        {__pyx_filename = __pyx_f[2]; __pyx_lineno = 796; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
      }
      #if CYTHON_COMPILING_IN_CPYTHON
      __pyx_t_3 = PyTuple_GET_ITEM(sequence, 0); 
      __pyx_t_4 = PyTuple_GET_ITEM(sequence, 1); 
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_4);
      #else
      __pyx_t_3 = PySequence_ITEM(sequence, 0); if (unlikely(!__pyx_t_3)) {__pyx_filename = __pyx_f[2]; __pyx_lineno = 796; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
      __Pyx_GOTREF(__pyx_t_3);
      __pyx_t_4 = PySequence_ITEM(sequence, 1); if (unlikely(!__pyx_t_4)) {__pyx_filename = __pyx_f[2]; __pyx_lineno = 796; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
      __Pyx_GOTREF(__pyx_t_4);
      #endif
    } else if (1) {
      __Pyx_RaiseNoneNotIterableError(); {__pyx_filename = __pyx_f[2]; __pyx_lineno = 796; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    } else
    {
      Py_ssize_t index = -1;
      __pyx_t_5 = PyObject_GetIter(((PyObject *)__pyx_v_fields)); if (unlikely(!__pyx_t_5)) {__pyx_filename = __pyx_f[2]; __pyx_lineno = 796; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
      __Pyx_GOTREF(__pyx_t_5);
      __pyx_t_6 = Py_TYPE(__pyx_t_5)->tp_iternext;
      index = 0; __pyx_t_3 = __pyx_t_6(__pyx_t_5); if (unlikely(!__pyx_t_3)) goto __pyx_L5_unpacking_failed;
      __Pyx_GOTREF(__pyx_t_3);
      index = 1; __pyx_t_4 = __pyx_t_6(__pyx_t_5); if (unlikely(!__pyx_t_4)) goto __pyx_L5_unpacking_failed;
      __Pyx_GOTREF(__pyx_t_4);
      if (__Pyx_IternextUnpackEndCheck(__pyx_t_6(__pyx_t_5), 2) < 0) {__pyx_filename = __pyx_f[2]; __pyx_lineno = 796; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
      __pyx_t_6 = NULL;
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      goto __pyx_L6_unpacking_done;
      __pyx_L5_unpacking_failed:;
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      __pyx_t_6 = NULL;
      if (__Pyx_IterFinish() == 0) __Pyx_RaiseNeedMoreValuesError(index);
      {__pyx_filename = __pyx_f[2]; __pyx_lineno = 796; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
      __pyx_L6_unpacking_done:;
    }
    if (!(likely(((__pyx_t_3) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_3, __pyx_ptype_5numpy_dtype))))) {__pyx_filename = __pyx_f[2]; __pyx_lineno = 796; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    __Pyx_XDECREF(((PyObject *)__pyx_v_child));
    __pyx_v_child = ((PyArray_Descr *)__pyx_t_3);
    __pyx_t_3 = 0;
    __Pyx_XDECREF(__pyx_v_new_offset);
    __pyx_v_new_offset = __pyx_t_4;
    __pyx_t_4 = 0;

    /* "numpy.pxd":798
 *         child, new_offset = fields
 * 
 *         if (end - f) - (new_offset - offset[0]) < 15:             # <<<<<<<<<<<<<<
 *             raise RuntimeError(u"Format string allocated too short, see comment in numpy.pxd")
 * 
 */
    __pyx_t_4 = PyInt_FromLong((__pyx_v_end - __pyx_v_f)); if (unlikely(!__pyx_t_4)) {__pyx_filename = __pyx_f[2]; __pyx_lineno = 798; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_3 = PyInt_FromLong((__pyx_v_offset[0])); if (unlikely(!__pyx_t_3)) {__pyx_filename = __pyx_f[2]; __pyx_lineno = 798; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_5 = PyNumber_Subtract(__pyx_v_new_offset, __pyx_t_3); if (unlikely(!__pyx_t_5)) {__pyx_filename = __pyx_f[2]; __pyx_lineno = 798; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_t_3 = PyNumber_Subtract(__pyx_t_4, __pyx_t_5); if (unlikely(!__pyx_t_3)) {__pyx_filename = __pyx_f[2]; __pyx_lineno = 798; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __pyx_t_5 = PyObject_RichCompare(__pyx_t_3, __pyx_int_15, Py_LT); __Pyx_XGOTREF(__pyx_t_5); if (unlikely(!__pyx_t_5)) {__pyx_filename = __pyx_f[2]; __pyx_lineno = 798; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_t_7 = __Pyx_PyObject_IsTrue(__pyx_t_5); if (unlikely(__pyx_t_7 < 0)) {__pyx_filename = __pyx_f[2]; __pyx_lineno = 798; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    if (__pyx_t_7) {

      /* "numpy.pxd":799
 * 
 *         if (end - f) - (new_offset - offset[0]) < 15:
 *             raise RuntimeError(u"Format string allocated too short, see comment in numpy.pxd")             # <<<<<<<<<<<<<<
 * 
 *         if ((child.byteorder == c'>' and little_endian) or
 */
      __pyx_t_5 = PyObject_Call(__pyx_builtin_RuntimeError, ((PyObject *)__pyx_k_tuple_11), NULL); if (unlikely(!__pyx_t_5)) {__pyx_filename = __pyx_f[2]; __pyx_lineno = 799; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
      __Pyx_GOTREF(__pyx_t_5);
      __Pyx_Raise(__pyx_t_5, 0, 0, 0);
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      {__pyx_filename = __pyx_f[2]; __pyx_lineno = 799; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
      goto __pyx_L7;
    }
    __pyx_L7:;

    /* "numpy.pxd":801
 *             raise RuntimeError(u"Format string allocated too short, see comment in numpy.pxd")
 * 
 *         if ((child.byteorder == c'>' and little_endian) or             # <<<<<<<<<<<<<<
 *             (child.byteorder == c'<' and not little_endian)):
 *             raise ValueError(u"Non-native byte order not supported")
 */
    __pyx_t_7 = ((__pyx_v_child->byteorder == '>') != 0);
    if (__pyx_t_7) {
      __pyx_t_8 = (__pyx_v_little_endian != 0);
    } else {
      __pyx_t_8 = __pyx_t_7;
    }
    if (!__pyx_t_8) {

      /* "numpy.pxd":802
 * 
 *         if ((child.byteorder == c'>' and little_endian) or
 *             (child.byteorder == c'<' and not little_endian)):             # <<<<<<<<<<<<<<
 *             raise ValueError(u"Non-native byte order not supported")
 *             # One could encode it in the format string and have Cython
 */
      __pyx_t_7 = ((__pyx_v_child->byteorder == '<') != 0);
      if (__pyx_t_7) {
        __pyx_t_9 = ((!(__pyx_v_little_endian != 0)) != 0);
        __pyx_t_10 = __pyx_t_9;
      } else {
        __pyx_t_10 = __pyx_t_7;
      }
      __pyx_t_7 = __pyx_t_10;
    } else {
      __pyx_t_7 = __pyx_t_8;
    }
    if (__pyx_t_7) {

      /* "numpy.pxd":803
 *         if ((child.byteorder == c'>' and little_endian) or
 *             (child.byteorder == c'<' and not little_endian)):
 *             raise ValueError(u"Non-native byte order not supported")             # <<<<<<<<<<<<<<
 *             # One could encode it in the format string and have Cython
 *             # complain instead, BUT: < and > in format strings also imply
 */
      __pyx_t_5 = PyObject_Call(__pyx_builtin_ValueError, ((PyObject *)__pyx_k_tuple_12), NULL); if (unlikely(!__pyx_t_5)) {__pyx_filename = __pyx_f[2]; __pyx_lineno = 803; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
      __Pyx_GOTREF(__pyx_t_5);
      __Pyx_Raise(__pyx_t_5, 0, 0, 0);
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      {__pyx_filename = __pyx_f[2]; __pyx_lineno = 803; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
      goto __pyx_L8;
    }
    __pyx_L8:;

    /* "numpy.pxd":813
 * 
 *         # Output padding bytes
 *         while offset[0] < new_offset:             # <<<<<<<<<<<<<<
 *             f[0] = 120 # "x"; pad byte
 *             f += 1
 */
    while (1) {
      __pyx_t_5 = PyInt_FromLong((__pyx_v_offset[0])); if (unlikely(!__pyx_t_5)) {__pyx_filename = __pyx_f[2]; __pyx_lineno = 813; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
      __Pyx_GOTREF(__pyx_t_5);
      __pyx_t_3 = PyObject_RichCompare(__pyx_t_5, __pyx_v_new_offset, Py_LT); __Pyx_XGOTREF(__pyx_t_3); if (unlikely(!__pyx_t_3)) {__pyx_filename = __pyx_f[2]; __pyx_lineno = 813; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      __pyx_t_7 = __Pyx_PyObject_IsTrue(__pyx_t_3); if (unlikely(__pyx_t_7 < 0)) {__pyx_filename = __pyx_f[2]; __pyx_lineno = 813; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      if (!__pyx_t_7) break;

      /* "numpy.pxd":814
 *         # Output padding bytes
 *         while offset[0] < new_offset:
 *             f[0] = 120 # "x"; pad byte             # <<<<<<<<<<<<<<
 *             f += 1
 *             offset[0] += 1
 */
      (__pyx_v_f[0]) = 120;

      /* "numpy.pxd":815
 *         while offset[0] < new_offset:
 *             f[0] = 120 # "x"; pad byte
 *             f += 1             # <<<<<<<<<<<<<<
 *             offset[0] += 1
 * 
 */
      __pyx_v_f = (__pyx_v_f + 1);

      /* "numpy.pxd":816
 *             f[0] = 120 # "x"; pad byte
 *             f += 1
 *             offset[0] += 1             # <<<<<<<<<<<<<<
 * 
 *         offset[0] += child.itemsize
 */
      __pyx_t_11 = 0;
      (__pyx_v_offset[__pyx_t_11]) = ((__pyx_v_offset[__pyx_t_11]) + 1);
    }

    /* "numpy.pxd":818
 *             offset[0] += 1
 * 
 *         offset[0] += child.itemsize             # <<<<<<<<<<<<<<
 * 
 *         if not PyDataType_HASFIELDS(child):
 */
    __pyx_t_11 = 0;
    (__pyx_v_offset[__pyx_t_11]) = ((__pyx_v_offset[__pyx_t_11]) + __pyx_v_child->elsize);

    /* "numpy.pxd":820
 *         offset[0] += child.itemsize
 * 
 *         if not PyDataType_HASFIELDS(child):             # <<<<<<<<<<<<<<
 *             t = child.type_num
 *             if end - f < 5:
 */
    __pyx_t_7 = ((!(PyDataType_HASFIELDS(__pyx_v_child) != 0)) != 0);
    if (__pyx_t_7) {

      /* "numpy.pxd":821
 * 
 *         if not PyDataType_HASFIELDS(child):
 *             t = child.type_num             # <<<<<<<<<<<<<<
 *             if end - f < 5:
 *                 raise RuntimeError(u"Format string allocated too short.")
 */
      __pyx_t_3 = PyInt_FromLong(__pyx_v_child->type_num); if (unlikely(!__pyx_t_3)) {__pyx_filename = __pyx_f[2]; __pyx_lineno = 821; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_XDECREF(__pyx_v_t);
      __pyx_v_t = __pyx_t_3;
      __pyx_t_3 = 0;

      /* "numpy.pxd":822
 *         if not PyDataType_HASFIELDS(child):
 *             t = child.type_num
 *             if end - f < 5:             # <<<<<<<<<<<<<<
 *                 raise RuntimeError(u"Format string allocated too short.")
 * 
 */
      __pyx_t_7 = (((__pyx_v_end - __pyx_v_f) < 5) != 0);
      if (__pyx_t_7) {

        /* "numpy.pxd":823
 *             t = child.type_num
 *             if end - f < 5:
 *                 raise RuntimeError(u"Format string allocated too short.")             # <<<<<<<<<<<<<<
 * 
 *             # Until ticket #99 is fixed, use integers to avoid warnings
 */
        __pyx_t_3 = PyObject_Call(__pyx_builtin_RuntimeError, ((PyObject *)__pyx_k_tuple_14), NULL); if (unlikely(!__pyx_t_3)) {__pyx_filename = __pyx_f[2]; __pyx_lineno = 823; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_Raise(__pyx_t_3, 0, 0, 0);
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        {__pyx_filename = __pyx_f[2]; __pyx_lineno = 823; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
        goto __pyx_L12;
      }
      __pyx_L12:;

      /* "numpy.pxd":826
 * 
 *             # Until ticket #99 is fixed, use integers to avoid warnings
 *             if   t == NPY_BYTE:        f[0] =  98 #"b"             # <<<<<<<<<<<<<<
 *             elif t == NPY_UBYTE:       f[0] =  66 #"B"
 *             elif t == NPY_SHORT:       f[0] = 104 #"h"
 */
      __pyx_t_3 = PyInt_FromLong(NPY_BYTE); if (unlikely(!__pyx_t_3)) {__pyx_filename = __pyx_f[2]; __pyx_lineno = 826; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
      __Pyx_GOTREF(__pyx_t_3);
      __pyx_t_5 = PyObject_RichCompare(__pyx_v_t, __pyx_t_3, Py_EQ); __Pyx_XGOTREF(__pyx_t_5); if (unlikely(!__pyx_t_5)) {__pyx_filename = __pyx_f[2]; __pyx_lineno = 826; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __pyx_t_7 = __Pyx_PyObject_IsTrue(__pyx_t_5); if (unlikely(__pyx_t_7 < 0)) {__pyx_filename = __pyx_f[2]; __pyx_lineno = 826; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      if (__pyx_t_7) {
        (__pyx_v_f[0]) = 98;
        goto __pyx_L13;
      }

      /* "numpy.pxd":827
 *             # Until ticket #99 is fixed, use integers to avoid warnings
 *             if   t == NPY_BYTE:        f[0] =  98 #"b"
 *             elif t == NPY_UBYTE:       f[0] =  66 #"B"             # <<<<<<<<<<<<<<
 *             elif t == NPY_SHORT:       f[0] = 104 #"h"
 *             elif t == NPY_USHORT:      f[0] =  72 #"H"
 */
      __pyx_t_5 = PyInt_FromLong(NPY_UBYTE); if (unlikely(!__pyx_t_5)) {__pyx_filename = __pyx_f[2]; __pyx_lineno = 827; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
      __Pyx_GOTREF(__pyx_t_5);
      __pyx_t_3 = PyObject_RichCompare(__pyx_v_t, __pyx_t_5, Py_EQ); __Pyx_XGOTREF(__pyx_t_3); if (unlikely(!__pyx_t_3)) {__pyx_filename = __pyx_f[2]; __pyx_lineno = 827; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      __pyx_t_7 = __Pyx_PyObject_IsTrue(__pyx_t_3); if (unlikely(__pyx_t_7 < 0)) {__pyx_filename = __pyx_f[2]; __pyx_lineno = 827; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      if (__pyx_t_7) {
        (__pyx_v_f[0]) = 66;
        goto __pyx_L13;
      }

      /* "numpy.pxd":828
 *             if   t == NPY_BYTE:        f[0] =  98 #"b"
 *             elif t == NPY_UBYTE:       f[0] =  66 #"B"
 *             elif t == NPY_SHORT:       f[0] = 104 #"h"             # <<<<<<<<<<<<<<
 *             elif t == NPY_USHORT:      f[0] =  72 #"H"
 *             elif t == NPY_INT:         f[0] = 105 #"i"
 */
      __pyx_t_3 = PyInt_FromLong(NPY_SHORT); if (unlikely(!__pyx_t_3)) {__pyx_filename = __pyx_f[2]; __pyx_lineno = 828; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
      __Pyx_GOTREF(__pyx_t_3);
      __pyx_t_5 = PyObject_RichCompare(__pyx_v_t, __pyx_t_3, Py_EQ); __Pyx_XGOTREF(__pyx_t_5); if (unlikely(!__pyx_t_5)) {__pyx_filename = __pyx_f[2]; __pyx_lineno = 828; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __pyx_t_7 = __Pyx_PyObject_IsTrue(__pyx_t_5); if (unlikely(__pyx_t_7 < 0)) {__pyx_filename = __pyx_f[2]; __pyx_lineno = 828; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      if (__pyx_t_7) {
        (__pyx_v_f[0]) = 104;
        goto __pyx_L13;
      }

      /* "numpy.pxd":829
 *             elif t == NPY_UBYTE:       f[0] =  66 #"B"
 *             elif t == NPY_SHORT:       f[0] = 104 #"h"
 *             elif t == NPY_USHORT:      f[0] =  72 #"H"             # <<<<<<<<<<<<<<
 *             elif t == NPY_INT:         f[0] = 105 #"i"
 *             elif t == NPY_UINT:        f[0] =  73 #"I"
 */
      __pyx_t_5 = PyInt_FromLong(NPY_USHORT); if (unlikely(!__pyx_t_5)) {__pyx_filename = __pyx_f[2]; __pyx_lineno = 829; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
      __Pyx_GOTREF(__pyx_t_5);
      __pyx_t_3 = PyObject_RichCompare(__pyx_v_t, __pyx_t_5, Py_EQ); __Pyx_XGOTREF(__pyx_t_3); if (unlikely(!__pyx_t_3)) {__pyx_filename = __pyx_f[2]; __pyx_lineno = 829; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      __pyx_t_7 = __Pyx_PyObject_IsTrue(__pyx_t_3); if (unlikely(__pyx_t_7 < 0)) {__pyx_filename = __pyx_f[2]; __pyx_lineno = 829; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      if (__pyx_t_7) {
        (__pyx_v_f[0]) = 72;
        goto __pyx_L13;
      }

      /* "numpy.pxd":830
 *             elif t == NPY_SHORT:       f[0] = 104 #"h"
 *             elif t == NPY_USHORT:      f[0] =  72 #"H"
 *             elif t == NPY_INT:         f[0] = 105 #"i"             # <<<<<<<<<<<<<<
 *             elif t == NPY_UINT:        f[0] =  73 #"I"
 *             elif t == NPY_LONG:        f[0] = 108 #"l"
 */
      __pyx_t_3 = PyInt_FromLong(NPY_INT); if (unlikely(!__pyx_t_3)) {__pyx_filename = __pyx_f[2]; __pyx_lineno = 830; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
      __Pyx_GOTREF(__pyx_t_3);
      __pyx_t_5 = PyObject_RichCompare(__pyx_v_t, __pyx_t_3, Py_EQ); __Pyx_XGOTREF(__pyx_t_5); if (unlikely(!__pyx_t_5)) {__pyx_filename = __pyx_f[2]; __pyx_lineno = 830; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __pyx_t_7 = __Pyx_PyObject_IsTrue(__pyx_t_5); if (unlikely(__pyx_t_7 < 0)) {__pyx_filename = __pyx_f[2]; __pyx_lineno = 830; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      if (__pyx_t_7) {
        (__pyx_v_f[0]) = 105;
        goto __pyx_L13;
      }

      /* "numpy.pxd":831
 *             elif t == NPY_USHORT:      f[0] =  72 #"H"
 *             elif t == NPY_INT:         f[0] = 105 #"i"
 *             elif t == NPY_UINT:        f[0] =  73 #"I"             # <<<<<<<<<<<<<<
 *             elif t == NPY_LONG:        f[0] = 108 #"l"
 *             elif t == NPY_ULONG:       f[0] = 76  #"L"
 */
      __pyx_t_5 = PyInt_FromLong(NPY_UINT); if (unlikely(!__pyx_t_5)) {__pyx_filename = __pyx_f[2]; __pyx_lineno = 831; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
      __Pyx_GOTREF(__pyx_t_5);
      __pyx_t_3 = PyObject_RichCompare(__pyx_v_t, __pyx_t_5, Py_EQ); __Pyx_XGOTREF(__pyx_t_3); if (unlikely(!__pyx_t_3)) {__pyx_filename = __pyx_f[2]; __pyx_lineno = 831; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      __pyx_t_7 = __Pyx_PyObject_IsTrue(__pyx_t_3); if (unlikely(__pyx_t_7 < 0)) {__pyx_filename = __pyx_f[2]; __pyx_lineno = 831; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      if (__pyx_t_7) {
        (__pyx_v_f[0]) = 73;
        goto __pyx_L13;
      }

      /* "numpy.pxd":832
 *             elif t == NPY_INT:         f[0] = 105 #"i"
 *             elif t == NPY_UINT:        f[0] =  73 #"I"
 *             elif t == NPY_LONG:        f[0] = 108 #"l"             # <<<<<<<<<<<<<<
 *             elif t == NPY_ULONG:       f[0] = 76  #"L"
 *             elif t == NPY_LONGLONG:    f[0] = 113 #"q"
 */
      __pyx_t_3 = PyInt_FromLong(NPY_LONG); if (unlikely(!__pyx_t_3)) {__pyx_filename = __pyx_f[2]; __pyx_lineno = 832; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
      __Pyx_GOTREF(__pyx_t_3);
      __pyx_t_5 = PyObject_RichCompare(__pyx_v_t, __pyx_t_3, Py_EQ); __Pyx_XGOTREF(__pyx_t_5); if (unlikely(!__pyx_t_5)) {__pyx_filename = __pyx_f[2]; __pyx_lineno = 832; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __pyx_t_7 = __Pyx_PyObject_IsTrue(__pyx_t_5); if (unlikely(__pyx_t_7 < 0)) {__pyx_filename = __pyx_f[2]; __pyx_lineno = 832; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      if (__pyx_t_7) {
        (__pyx_v_f[0]) = 108;
        goto __pyx_L13;
      }

      /* "numpy.pxd":833
 *             elif t == NPY_UINT:        f[0] =  73 #"I"
 *             elif t == NPY_LONG:        f[0] = 108 #"l"
 *             elif t == NPY_ULONG:       f[0] = 76  #"L"             # <<<<<<<<<<<<<<
 *             elif t == NPY_LONGLONG:    f[0] = 113 #"q"
 *             elif t == NPY_ULONGLONG:   f[0] = 81  #"Q"
 */
      __pyx_t_5 = PyInt_FromLong(NPY_ULONG); if (unlikely(!__pyx_t_5)) {__pyx_filename = __pyx_f[2]; __pyx_lineno = 833; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
      __Pyx_GOTREF(__pyx_t_5);
      __pyx_t_3 = PyObject_RichCompare(__pyx_v_t, __pyx_t_5, Py_EQ); __Pyx_XGOTREF(__pyx_t_3); if (unlikely(!__pyx_t_3)) {__pyx_filename = __pyx_f[2]; __pyx_lineno = 833; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      __pyx_t_7 = __Pyx_PyObject_IsTrue(__pyx_t_3); if (unlikely(__pyx_t_7 < 0)) {__pyx_filename = __pyx_f[2]; __pyx_lineno = 833; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      if (__pyx_t_7) {
        (__pyx_v_f[0]) = 76;
        goto __pyx_L13;
      }

      /* "numpy.pxd":834
 *             elif t == NPY_LONG:        f[0] = 108 #"l"
 *             elif t == NPY_ULONG:       f[0] = 76  #"L"
 *             elif t == NPY_LONGLONG:    f[0] = 113 #"q"             # <<<<<<<<<<<<<<
 *             elif t == NPY_ULONGLONG:   f[0] = 81  #"Q"
 *             elif t == NPY_FLOAT:       f[0] = 102 #"f"
 */
      __pyx_t_3 = PyInt_FromLong(NPY_LONGLONG); if (unlikely(!__pyx_t_3)) {__pyx_filename = __pyx_f[2]; __pyx_lineno = 834; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
      __Pyx_GOTREF(__pyx_t_3);
      __pyx_t_5 = PyObject_RichCompare(__pyx_v_t, __pyx_t_3, Py_EQ); __Pyx_XGOTREF(__pyx_t_5); if (unlikely(!__pyx_t_5)) {__pyx_filename = __pyx_f[2]; __pyx_lineno = 834; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __pyx_t_7 = __Pyx_PyObject_IsTrue(__pyx_t_5); if (unlikely(__pyx_t_7 < 0)) {__pyx_filename = __pyx_f[2]; __pyx_lineno = 834; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      if (__pyx_t_7) {
        (__pyx_v_f[0]) = 113;
        goto __pyx_L13;
      }

      /* "numpy.pxd":835
 *             elif t == NPY_ULONG:       f[0] = 76  #"L"
 *             elif t == NPY_LONGLONG:    f[0] = 113 #"q"
 *             elif t == NPY_ULONGLONG:   f[0] = 81  #"Q"             # <<<<<<<<<<<<<<
 *             elif t == NPY_FLOAT:       f[0] = 102 #"f"
 *             elif t == NPY_DOUBLE:      f[0] = 100 #"d"
 */
      __pyx_t_5 = PyInt_FromLong(NPY_ULONGLONG); if (unlikely(!__pyx_t_5)) {__pyx_filename = __pyx_f[2]; __pyx_lineno = 835; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
      __Pyx_GOTREF(__pyx_t_5);
      __pyx_t_3 = PyObject_RichCompare(__pyx_v_t, __pyx_t_5, Py_EQ); __Pyx_XGOTREF(__pyx_t_3); if (unlikely(!__pyx_t_3)) {__pyx_filename = __pyx_f[2]; __pyx_lineno = 835; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      __pyx_t_7 = __Pyx_PyObject_IsTrue(__pyx_t_3); if (unlikely(__pyx_t_7 < 0)) {__pyx_filename = __pyx_f[2]; __pyx_lineno = 835; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      if (__pyx_t_7) {
        (__pyx_v_f[0]) = 81;
        goto __pyx_L13;
      }

      /* "numpy.pxd":836
 *             elif t == NPY_LONGLONG:    f[0] = 113 #"q"
 *             elif t == NPY_ULONGLONG:   f[0] = 81  #"Q"
 *             elif t == NPY_FLOAT:       f[0] = 102 #"f"             # <<<<<<<<<<<<<<
 *             elif t == NPY_DOUBLE:      f[0] = 100 #"d"
 *             elif t == NPY_LONGDOUBLE:  f[0] = 103 #"g"
 */
      __pyx_t_3 = PyInt_FromLong(NPY_FLOAT); if (unlikely(!__pyx_t_3)) {__pyx_filename = __pyx_f[2]; __pyx_lineno = 836; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
      __Pyx_GOTREF(__pyx_t_3);
      __pyx_t_5 = PyObject_RichCompare(__pyx_v_t, __pyx_t_3, Py_EQ); __Pyx_XGOTREF(__pyx_t_5); if (unlikely(!__pyx_t_5)) {__pyx_filename = __pyx_f[2]; __pyx_lineno = 836; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __pyx_t_7 = __Pyx_PyObject_IsTrue(__pyx_t_5); if (unlikely(__pyx_t_7 < 0)) {__pyx_filename = __pyx_f[2]; __pyx_lineno = 836; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      if (__pyx_t_7) {
        (__pyx_v_f[0]) = 102;
        goto __pyx_L13;
      }

      /* "numpy.pxd":837
 *             elif t == NPY_ULONGLONG:   f[0] = 81  #"Q"
 *             elif t == NPY_FLOAT:       f[0] = 102 #"f"
 *             elif t == NPY_DOUBLE:      f[0] = 100 #"d"             # <<<<<<<<<<<<<<
 *             elif t == NPY_LONGDOUBLE:  f[0] = 103 #"g"
 *             elif t == NPY_CFLOAT:      f[0] = 90; f[1] = 102; f += 1 # Zf
 */
      __pyx_t_5 = PyInt_FromLong(NPY_DOUBLE); if (unlikely(!__pyx_t_5)) {__pyx_filename = __pyx_f[2]; __pyx_lineno = 837; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
      __Pyx_GOTREF(__pyx_t_5);
      __pyx_t_3 = PyObject_RichCompare(__pyx_v_t, __pyx_t_5, Py_EQ); __Pyx_XGOTREF(__pyx_t_3); if (unlikely(!__pyx_t_3)) {__pyx_filename = __pyx_f[2]; __pyx_lineno = 837; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      __pyx_t_7 = __Pyx_PyObject_IsTrue(__pyx_t_3); if (unlikely(__pyx_t_7 < 0)) {__pyx_filename = __pyx_f[2]; __pyx_lineno = 837; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      if (__pyx_t_7) {
        (__pyx_v_f[0]) = 100;
        goto __pyx_L13;
      }

      /* "numpy.pxd":838
 *             elif t == NPY_FLOAT:       f[0] = 102 #"f"
 *             elif t == NPY_DOUBLE:      f[0] = 100 #"d"
 *             elif t == NPY_LONGDOUBLE:  f[0] = 103 #"g"             # <<<<<<<<<<<<<<
 *             elif t == NPY_CFLOAT:      f[0] = 90; f[1] = 102; f += 1 # Zf
 *             elif t == NPY_CDOUBLE:     f[0] = 90; f[1] = 100; f += 1 # Zd
 */
      __pyx_t_3 = PyInt_FromLong(NPY_LONGDOUBLE); if (unlikely(!__pyx_t_3)) {__pyx_filename = __pyx_f[2]; __pyx_lineno = 838; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
      __Pyx_GOTREF(__pyx_t_3);
      __pyx_t_5 = PyObject_RichCompare(__pyx_v_t, __pyx_t_3, Py_EQ); __Pyx_XGOTREF(__pyx_t_5); if (unlikely(!__pyx_t_5)) {__pyx_filename = __pyx_f[2]; __pyx_lineno = 838; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __pyx_t_7 = __Pyx_PyObject_IsTrue(__pyx_t_5); if (unlikely(__pyx_t_7 < 0)) {__pyx_filename = __pyx_f[2]; __pyx_lineno = 838; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      if (__pyx_t_7) {
        (__pyx_v_f[0]) = 103;
        goto __pyx_L13;
      }

      /* "numpy.pxd":839
 *             elif t == NPY_DOUBLE:      f[0] = 100 #"d"
 *             elif t == NPY_LONGDOUBLE:  f[0] = 103 #"g"
 *             elif t == NPY_CFLOAT:      f[0] = 90; f[1] = 102; f += 1 # Zf             # <<<<<<<<<<<<<<
 *             elif t == NPY_CDOUBLE:     f[0] = 90; f[1] = 100; f += 1 # Zd
 *             elif t == NPY_CLONGDOUBLE: f[0] = 90; f[1] = 103; f += 1 # Zg
 */
      __pyx_t_5 = PyInt_FromLong(NPY_CFLOAT); if (unlikely(!__pyx_t_5)) {__pyx_filename = __pyx_f[2]; __pyx_lineno = 839; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
      __Pyx_GOTREF(__pyx_t_5);
      __pyx_t_3 = PyObject_RichCompare(__pyx_v_t, __pyx_t_5, Py_EQ); __Pyx_XGOTREF(__pyx_t_3); if (unlikely(!__pyx_t_3)) {__pyx_filename = __pyx_f[2]; __pyx_lineno = 839; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      __pyx_t_7 = __Pyx_PyObject_IsTrue(__pyx_t_3); if (unlikely(__pyx_t_7 < 0)) {__pyx_filename = __pyx_f[2]; __pyx_lineno = 839; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      if (__pyx_t_7) {
        (__pyx_v_f[0]) = 90;
        (__pyx_v_f[1]) = 102;
        __pyx_v_f = (__pyx_v_f + 1);
        goto __pyx_L13;
      }

      /* "numpy.pxd":840
 *             elif t == NPY_LONGDOUBLE:  f[0] = 103 #"g"
 *             elif t == NPY_CFLOAT:      f[0] = 90; f[1] = 102; f += 1 # Zf
 *             elif t == NPY_CDOUBLE:     f[0] = 90; f[1] = 100; f += 1 # Zd             # <<<<<<<<<<<<<<
 *             elif t == NPY_CLONGDOUBLE: f[0] = 90; f[1] = 103; f += 1 # Zg
 *             elif t == NPY_OBJECT:      f[0] = 79 #"O"
 */
      __pyx_t_3 = PyInt_FromLong(NPY_CDOUBLE); if (unlikely(!__pyx_t_3)) {__pyx_filename = __pyx_f[2]; __pyx_lineno = 840; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
      __Pyx_GOTREF(__pyx_t_3);
      __pyx_t_5 = PyObject_RichCompare(__pyx_v_t, __pyx_t_3, Py_EQ); __Pyx_XGOTREF(__pyx_t_5); if (unlikely(!__pyx_t_5)) {__pyx_filename = __pyx_f[2]; __pyx_lineno = 840; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __pyx_t_7 = __Pyx_PyObject_IsTrue(__pyx_t_5); if (unlikely(__pyx_t_7 < 0)) {__pyx_filename = __pyx_f[2]; __pyx_lineno = 840; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      if (__pyx_t_7) {
        (__pyx_v_f[0]) = 90;
        (__pyx_v_f[1]) = 100;
        __pyx_v_f = (__pyx_v_f + 1);
        goto __pyx_L13;
      }

      /* "numpy.pxd":841
 *             elif t == NPY_CFLOAT:      f[0] = 90; f[1] = 102; f += 1 # Zf
 *             elif t == NPY_CDOUBLE:     f[0] = 90; f[1] = 100; f += 1 # Zd
 *             elif t == NPY_CLONGDOUBLE: f[0] = 90; f[1] = 103; f += 1 # Zg             # <<<<<<<<<<<<<<
 *             elif t == NPY_OBJECT:      f[0] = 79 #"O"
 *             else:
 */
      __pyx_t_5 = PyInt_FromLong(NPY_CLONGDOUBLE); if (unlikely(!__pyx_t_5)) {__pyx_filename = __pyx_f[2]; __pyx_lineno = 841; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
      __Pyx_GOTREF(__pyx_t_5);
      __pyx_t_3 = PyObject_RichCompare(__pyx_v_t, __pyx_t_5, Py_EQ); __Pyx_XGOTREF(__pyx_t_3); if (unlikely(!__pyx_t_3)) {__pyx_filename = __pyx_f[2]; __pyx_lineno = 841; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      __pyx_t_7 = __Pyx_PyObject_IsTrue(__pyx_t_3); if (unlikely(__pyx_t_7 < 0)) {__pyx_filename = __pyx_f[2]; __pyx_lineno = 841; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      if (__pyx_t_7) {
        (__pyx_v_f[0]) = 90;
        (__pyx_v_f[1]) = 103;
        __pyx_v_f = (__pyx_v_f + 1);
        goto __pyx_L13;
      }

      /* "numpy.pxd":842
 *             elif t == NPY_CDOUBLE:     f[0] = 90; f[1] = 100; f += 1 # Zd
 *             elif t == NPY_CLONGDOUBLE: f[0] = 90; f[1] = 103; f += 1 # Zg
 *             elif t == NPY_OBJECT:      f[0] = 79 #"O"             # <<<<<<<<<<<<<<
 *             else:
 *                 raise ValueError(u"unknown dtype code in numpy.pxd (%d)" % t)
 */
      __pyx_t_3 = PyInt_FromLong(NPY_OBJECT); if (unlikely(!__pyx_t_3)) {__pyx_filename = __pyx_f[2]; __pyx_lineno = 842; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
      __Pyx_GOTREF(__pyx_t_3);
      __pyx_t_5 = PyObject_RichCompare(__pyx_v_t, __pyx_t_3, Py_EQ); __Pyx_XGOTREF(__pyx_t_5); if (unlikely(!__pyx_t_5)) {__pyx_filename = __pyx_f[2]; __pyx_lineno = 842; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __pyx_t_7 = __Pyx_PyObject_IsTrue(__pyx_t_5); if (unlikely(__pyx_t_7 < 0)) {__pyx_filename = __pyx_f[2]; __pyx_lineno = 842; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      if (__pyx_t_7) {
        (__pyx_v_f[0]) = 79;
        goto __pyx_L13;
      }
      /*else*/ {

        /* "numpy.pxd":844
 *             elif t == NPY_OBJECT:      f[0] = 79 #"O"
 *             else:
 *                 raise ValueError(u"unknown dtype code in numpy.pxd (%d)" % t)             # <<<<<<<<<<<<<<
 *             f += 1
 *         else:
 */
        __pyx_t_5 = PyNumber_Remainder(((PyObject *)__pyx_kp_u_9), __pyx_v_t); if (unlikely(!__pyx_t_5)) {__pyx_filename = __pyx_f[2]; __pyx_lineno = 844; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
        __Pyx_GOTREF(((PyObject *)__pyx_t_5));
        __pyx_t_3 = PyTuple_New(1); if (unlikely(!__pyx_t_3)) {__pyx_filename = __pyx_f[2]; __pyx_lineno = 844; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
        __Pyx_GOTREF(__pyx_t_3);
        PyTuple_SET_ITEM(__pyx_t_3, 0, ((PyObject *)__pyx_t_5));
        __Pyx_GIVEREF(((PyObject *)__pyx_t_5));
        __pyx_t_5 = 0;
        __pyx_t_5 = PyObject_Call(__pyx_builtin_ValueError, ((PyObject *)__pyx_t_3), NULL); if (unlikely(!__pyx_t_5)) {__pyx_filename = __pyx_f[2]; __pyx_lineno = 844; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
        __Pyx_GOTREF(__pyx_t_5);
        __Pyx_DECREF(((PyObject *)__pyx_t_3)); __pyx_t_3 = 0;
        __Pyx_Raise(__pyx_t_5, 0, 0, 0);
        __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
        {__pyx_filename = __pyx_f[2]; __pyx_lineno = 844; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
      }
      __pyx_L13:;

      /* "numpy.pxd":845
 *             else:
 *                 raise ValueError(u"unknown dtype code in numpy.pxd (%d)" % t)
 *             f += 1             # <<<<<<<<<<<<<<
 *         else:
 *             # Cython ignores struct boundary information ("T{...}"),
 */
      __pyx_v_f = (__pyx_v_f + 1);
      goto __pyx_L11;
    }
    /*else*/ {

      /* "numpy.pxd":849
 *             # Cython ignores struct boundary information ("T{...}"),
 *             # so don't output it
 *             f = _util_dtypestring(child, f, end, offset)             # <<<<<<<<<<<<<<
 *     return f
 * 
 */
      __pyx_t_12 = __pyx_f_5numpy__util_dtypestring(__pyx_v_child, __pyx_v_f, __pyx_v_end, __pyx_v_offset); if (unlikely(__pyx_t_12 == NULL)) {__pyx_filename = __pyx_f[2]; __pyx_lineno = 849; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
      __pyx_v_f = __pyx_t_12;
    }
    __pyx_L11:;
  }
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "numpy.pxd":850
 *             # so don't output it
 *             f = _util_dtypestring(child, f, end, offset)
 *     return f             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = __pyx_v_f;
  goto __pyx_L0;

  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("numpy._util_dtypestring", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_child);
  __Pyx_XDECREF(__pyx_v_fields);
  __Pyx_XDECREF(__pyx_v_childname);
  __Pyx_XDECREF(__pyx_v_new_offset);
  __Pyx_XDECREF(__pyx_v_t);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "numpy.pxd":965
 * 
 * 
 * cdef inline void set_array_base(ndarray arr, object base):             # <<<<<<<<<<<<<<
 *      cdef PyObject* baseptr
 *      if base is None:
 */

static CYTHON_INLINE void __pyx_f_5numpy_set_array_base(PyArrayObject *__pyx_v_arr, PyObject *__pyx_v_base) {
  PyObject *__pyx_v_baseptr;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  __Pyx_RefNannySetupContext("set_array_base", 0);

  /* "numpy.pxd":967
 * cdef inline void set_array_base(ndarray arr, object base):
 *      cdef PyObject* baseptr
 *      if base is None:             # <<<<<<<<<<<<<<
 *          baseptr = NULL
 *      else:
 */
  __pyx_t_1 = (__pyx_v_base == Py_None);
  __pyx_t_2 = (__pyx_t_1 != 0);
  if (__pyx_t_2) {

    /* "numpy.pxd":968
 *      cdef PyObject* baseptr
 *      if base is None:
 *          baseptr = NULL             # <<<<<<<<<<<<<<
 *      else:
 *          Py_INCREF(base) # important to do this before decref below!
 */
    __pyx_v_baseptr = NULL;
    goto __pyx_L3;
  }
  /*else*/ {

    /* "numpy.pxd":970
 *          baseptr = NULL
 *      else:
 *          Py_INCREF(base) # important to do this before decref below!             # <<<<<<<<<<<<<<
 *          baseptr = <PyObject*>base
 *      Py_XDECREF(arr.base)
 */
    Py_INCREF(__pyx_v_base);

    /* "numpy.pxd":971
 *      else:
 *          Py_INCREF(base) # important to do this before decref below!
 *          baseptr = <PyObject*>base             # <<<<<<<<<<<<<<
 *      Py_XDECREF(arr.base)
 *      arr.base = baseptr
 */
    __pyx_v_baseptr = ((PyObject *)__pyx_v_base);
  }
  __pyx_L3:;

  /* "numpy.pxd":972
 *          Py_INCREF(base) # important to do this before decref below!
 *          baseptr = <PyObject*>base
 *      Py_XDECREF(arr.base)             # <<<<<<<<<<<<<<
 *      arr.base = baseptr
 * 
 */
  Py_XDECREF(__pyx_v_arr->base);

  /* "numpy.pxd":973
 *          baseptr = <PyObject*>base
 *      Py_XDECREF(arr.base)
 *      arr.base = baseptr             # <<<<<<<<<<<<<<
 * 
 * cdef inline object get_array_base(ndarray arr):
 */
  __pyx_v_arr->base = __pyx_v_baseptr;

  __Pyx_RefNannyFinishContext();
}

/* "numpy.pxd":975
 *      arr.base = baseptr
 * 
 * cdef inline object get_array_base(ndarray arr):             # <<<<<<<<<<<<<<
 *     if arr.base is NULL:
 *         return None
 */

static CYTHON_INLINE PyObject *__pyx_f_5numpy_get_array_base(PyArrayObject *__pyx_v_arr) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  __Pyx_RefNannySetupContext("get_array_base", 0);

  /* "numpy.pxd":976
 * 
 * cdef inline object get_array_base(ndarray arr):
 *     if arr.base is NULL:             # <<<<<<<<<<<<<<
 *         return None
 *     else:
 */
  __pyx_t_1 = ((__pyx_v_arr->base == NULL) != 0);
  if (__pyx_t_1) {

    /* "numpy.pxd":977
 * cdef inline object get_array_base(ndarray arr):
 *     if arr.base is NULL:
 *         return None             # <<<<<<<<<<<<<<
 *     else:
 *         return <object>arr.base
 */
    __Pyx_XDECREF(__pyx_r);
    __Pyx_INCREF(Py_None);
    __pyx_r = Py_None;
    goto __pyx_L0;
    goto __pyx_L3;
  }
  /*else*/ {

    /* "numpy.pxd":979
 *         return None
 *     else:
 *         return <object>arr.base             # <<<<<<<<<<<<<<
 */
    __Pyx_XDECREF(__pyx_r);
    __Pyx_INCREF(((PyObject *)__pyx_v_arr->base));
    __pyx_r = ((PyObject *)__pyx_v_arr->base);
    goto __pyx_L0;
  }
  __pyx_L3:;

  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}
static struct __pyx_vtabstruct_7sklearn_4tree_5_tree_Tree __pyx_vtable_7sklearn_4tree_5_tree_Tree;

static PyObject *__pyx_tp_new_7sklearn_4tree_5_tree_Tree(PyTypeObject *t, PyObject *a, PyObject *k) {
  struct __pyx_obj_7sklearn_4tree_5_tree_Tree *p;
  PyObject *o;
  o = (*t->tp_alloc)(t, 0);
  if (unlikely(!o)) return 0;
  p = ((struct __pyx_obj_7sklearn_4tree_5_tree_Tree *)o);
  p->__pyx_vtab = __pyx_vtabptr_7sklearn_4tree_5_tree_Tree;
  p->splitter = ((struct __pyx_obj_7sklearn_4tree_5_tree_Splitter *)Py_None); Py_INCREF(Py_None);
  p->random_state = Py_None; Py_INCREF(Py_None);
  if (unlikely(__pyx_pw_7sklearn_4tree_5_tree_4Tree_1__cinit__(o, a, k) < 0)) {
    Py_DECREF(o); o = 0;
  }
  return o;
}

static void __pyx_tp_dealloc_7sklearn_4tree_5_tree_Tree(PyObject *o) {
  struct __pyx_obj_7sklearn_4tree_5_tree_Tree *p = (struct __pyx_obj_7sklearn_4tree_5_tree_Tree *)o;
  PyObject_GC_UnTrack(o);
  {
    PyObject *etype, *eval, *etb;
    PyErr_Fetch(&etype, &eval, &etb);
    ++Py_REFCNT(o);
    __pyx_pw_7sklearn_4tree_5_tree_4Tree_3__dealloc__(o);
    if (PyErr_Occurred()) PyErr_WriteUnraisable(o);
    --Py_REFCNT(o);
    PyErr_Restore(etype, eval, etb);
  }
  Py_CLEAR(p->splitter);
  Py_CLEAR(p->random_state);
  (*Py_TYPE(o)->tp_free)(o);
}

static int __pyx_tp_traverse_7sklearn_4tree_5_tree_Tree(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_7sklearn_4tree_5_tree_Tree *p = (struct __pyx_obj_7sklearn_4tree_5_tree_Tree *)o;
  if (p->splitter) {
    e = (*v)(((PyObject*)p->splitter), a); if (e) return e;
  }
  if (p->random_state) {
    e = (*v)(p->random_state, a); if (e) return e;
  }
  return 0;
}

static int __pyx_tp_clear_7sklearn_4tree_5_tree_Tree(PyObject *o) {
  struct __pyx_obj_7sklearn_4tree_5_tree_Tree *p = (struct __pyx_obj_7sklearn_4tree_5_tree_Tree *)o;
  PyObject* tmp;
  tmp = ((PyObject*)p->splitter);
  p->splitter = ((struct __pyx_obj_7sklearn_4tree_5_tree_Splitter *)Py_None); Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->random_state);
  p->random_state = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  return 0;
}

static PyObject *__pyx_getprop_7sklearn_4tree_5_tree_4Tree_n_classes(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_7sklearn_4tree_5_tree_4Tree_9n_classes_1__get__(o);
}

static PyObject *__pyx_getprop_7sklearn_4tree_5_tree_4Tree_children_left(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_7sklearn_4tree_5_tree_4Tree_13children_left_1__get__(o);
}

static PyObject *__pyx_getprop_7sklearn_4tree_5_tree_4Tree_children_right(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_7sklearn_4tree_5_tree_4Tree_14children_right_1__get__(o);
}

static PyObject *__pyx_getprop_7sklearn_4tree_5_tree_4Tree_feature(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_7sklearn_4tree_5_tree_4Tree_7feature_1__get__(o);
}

static PyObject *__pyx_getprop_7sklearn_4tree_5_tree_4Tree_threshold(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_7sklearn_4tree_5_tree_4Tree_9threshold_1__get__(o);
}

static PyObject *__pyx_getprop_7sklearn_4tree_5_tree_4Tree_value(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_7sklearn_4tree_5_tree_4Tree_5value_1__get__(o);
}

static PyObject *__pyx_getprop_7sklearn_4tree_5_tree_4Tree_impurity(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_7sklearn_4tree_5_tree_4Tree_8impurity_1__get__(o);
}

static PyObject *__pyx_getprop_7sklearn_4tree_5_tree_4Tree_n_node_samples(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_7sklearn_4tree_5_tree_4Tree_14n_node_samples_1__get__(o);
}

static PyObject *__pyx_getprop_7sklearn_4tree_5_tree_4Tree_n_features(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_7sklearn_4tree_5_tree_4Tree_10n_features_1__get__(o);
}

static int __pyx_setprop_7sklearn_4tree_5_tree_4Tree_n_features(PyObject *o, PyObject *v, CYTHON_UNUSED void *x) {
  if (v) {
    return __pyx_pw_7sklearn_4tree_5_tree_4Tree_10n_features_3__set__(o, v);
  }
  else {
    PyErr_SetString(PyExc_NotImplementedError, "__del__");
    return -1;
  }
}

static PyObject *__pyx_getprop_7sklearn_4tree_5_tree_4Tree_n_outputs(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_7sklearn_4tree_5_tree_4Tree_9n_outputs_1__get__(o);
}

static int __pyx_setprop_7sklearn_4tree_5_tree_4Tree_n_outputs(PyObject *o, PyObject *v, CYTHON_UNUSED void *x) {
  if (v) {
    return __pyx_pw_7sklearn_4tree_5_tree_4Tree_9n_outputs_3__set__(o, v);
  }
  else {
    PyErr_SetString(PyExc_NotImplementedError, "__del__");
    return -1;
  }
}

static PyObject *__pyx_getprop_7sklearn_4tree_5_tree_4Tree_max_n_classes(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_7sklearn_4tree_5_tree_4Tree_13max_n_classes_1__get__(o);
}

static int __pyx_setprop_7sklearn_4tree_5_tree_4Tree_max_n_classes(PyObject *o, PyObject *v, CYTHON_UNUSED void *x) {
  if (v) {
    return __pyx_pw_7sklearn_4tree_5_tree_4Tree_13max_n_classes_3__set__(o, v);
  }
  else {
    PyErr_SetString(PyExc_NotImplementedError, "__del__");
    return -1;
  }
}

static PyObject *__pyx_getprop_7sklearn_4tree_5_tree_4Tree_value_stride(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_7sklearn_4tree_5_tree_4Tree_12value_stride_1__get__(o);
}

static int __pyx_setprop_7sklearn_4tree_5_tree_4Tree_value_stride(PyObject *o, PyObject *v, CYTHON_UNUSED void *x) {
  if (v) {
    return __pyx_pw_7sklearn_4tree_5_tree_4Tree_12value_stride_3__set__(o, v);
  }
  else {
    PyErr_SetString(PyExc_NotImplementedError, "__del__");
    return -1;
  }
}

static PyObject *__pyx_getprop_7sklearn_4tree_5_tree_4Tree_splitter(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_7sklearn_4tree_5_tree_4Tree_8splitter_1__get__(o);
}

static int __pyx_setprop_7sklearn_4tree_5_tree_4Tree_splitter(PyObject *o, PyObject *v, CYTHON_UNUSED void *x) {
  if (v) {
    return __pyx_pw_7sklearn_4tree_5_tree_4Tree_8splitter_3__set__(o, v);
  }
  else {
    return __pyx_pw_7sklearn_4tree_5_tree_4Tree_8splitter_5__del__(o);
  }
}

static PyObject *__pyx_getprop_7sklearn_4tree_5_tree_4Tree_max_depth(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_7sklearn_4tree_5_tree_4Tree_9max_depth_1__get__(o);
}

static int __pyx_setprop_7sklearn_4tree_5_tree_4Tree_max_depth(PyObject *o, PyObject *v, CYTHON_UNUSED void *x) {
  if (v) {
    return __pyx_pw_7sklearn_4tree_5_tree_4Tree_9max_depth_3__set__(o, v);
  }
  else {
    PyErr_SetString(PyExc_NotImplementedError, "__del__");
    return -1;
  }
}

static PyObject *__pyx_getprop_7sklearn_4tree_5_tree_4Tree_min_samples_split(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_7sklearn_4tree_5_tree_4Tree_17min_samples_split_1__get__(o);
}

static int __pyx_setprop_7sklearn_4tree_5_tree_4Tree_min_samples_split(PyObject *o, PyObject *v, CYTHON_UNUSED void *x) {
  if (v) {
    return __pyx_pw_7sklearn_4tree_5_tree_4Tree_17min_samples_split_3__set__(o, v);
  }
  else {
    PyErr_SetString(PyExc_NotImplementedError, "__del__");
    return -1;
  }
}

static PyObject *__pyx_getprop_7sklearn_4tree_5_tree_4Tree_min_samples_leaf(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_7sklearn_4tree_5_tree_4Tree_16min_samples_leaf_1__get__(o);
}

static int __pyx_setprop_7sklearn_4tree_5_tree_4Tree_min_samples_leaf(PyObject *o, PyObject *v, CYTHON_UNUSED void *x) {
  if (v) {
    return __pyx_pw_7sklearn_4tree_5_tree_4Tree_16min_samples_leaf_3__set__(o, v);
  }
  else {
    PyErr_SetString(PyExc_NotImplementedError, "__del__");
    return -1;
  }
}

static PyObject *__pyx_getprop_7sklearn_4tree_5_tree_4Tree_random_state(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_7sklearn_4tree_5_tree_4Tree_12random_state_1__get__(o);
}

static int __pyx_setprop_7sklearn_4tree_5_tree_4Tree_random_state(PyObject *o, PyObject *v, CYTHON_UNUSED void *x) {
  if (v) {
    return __pyx_pw_7sklearn_4tree_5_tree_4Tree_12random_state_3__set__(o, v);
  }
  else {
    return __pyx_pw_7sklearn_4tree_5_tree_4Tree_12random_state_5__del__(o);
  }
}

static PyObject *__pyx_getprop_7sklearn_4tree_5_tree_4Tree_node_count(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_7sklearn_4tree_5_tree_4Tree_10node_count_1__get__(o);
}

static int __pyx_setprop_7sklearn_4tree_5_tree_4Tree_node_count(PyObject *o, PyObject *v, CYTHON_UNUSED void *x) {
  if (v) {
    return __pyx_pw_7sklearn_4tree_5_tree_4Tree_10node_count_3__set__(o, v);
  }
  else {
    PyErr_SetString(PyExc_NotImplementedError, "__del__");
    return -1;
  }
}

static PyObject *__pyx_getprop_7sklearn_4tree_5_tree_4Tree_capacity(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_7sklearn_4tree_5_tree_4Tree_8capacity_1__get__(o);
}

static int __pyx_setprop_7sklearn_4tree_5_tree_4Tree_capacity(PyObject *o, PyObject *v, CYTHON_UNUSED void *x) {
  if (v) {
    return __pyx_pw_7sklearn_4tree_5_tree_4Tree_8capacity_3__set__(o, v);
  }
  else {
    PyErr_SetString(PyExc_NotImplementedError, "__del__");
    return -1;
  }
}

static PyMethodDef __pyx_methods_7sklearn_4tree_5_tree_Tree[] = {
  {__Pyx_NAMESTR("__reduce__"), (PyCFunction)__pyx_pw_7sklearn_4tree_5_tree_4Tree_5__reduce__, METH_NOARGS, __Pyx_DOCSTR(__pyx_doc_7sklearn_4tree_5_tree_4Tree_4__reduce__)},
  {__Pyx_NAMESTR("__getstate__"), (PyCFunction)__pyx_pw_7sklearn_4tree_5_tree_4Tree_7__getstate__, METH_NOARGS, __Pyx_DOCSTR(__pyx_doc_7sklearn_4tree_5_tree_4Tree_6__getstate__)},
  {__Pyx_NAMESTR("__setstate__"), (PyCFunction)__pyx_pw_7sklearn_4tree_5_tree_4Tree_9__setstate__, METH_O, __Pyx_DOCSTR(__pyx_doc_7sklearn_4tree_5_tree_4Tree_8__setstate__)},
  {__Pyx_NAMESTR("build"), (PyCFunction)__pyx_pw_7sklearn_4tree_5_tree_4Tree_11build, METH_VARARGS|METH_KEYWORDS, __Pyx_DOCSTR(__pyx_doc_7sklearn_4tree_5_tree_4Tree_10build)},
  {__Pyx_NAMESTR("predict"), (PyCFunction)__pyx_pw_7sklearn_4tree_5_tree_4Tree_13predict, METH_O, __Pyx_DOCSTR(__pyx_doc_7sklearn_4tree_5_tree_4Tree_12predict)},
  {__Pyx_NAMESTR("apply"), (PyCFunction)__pyx_pw_7sklearn_4tree_5_tree_4Tree_15apply, METH_O, __Pyx_DOCSTR(__pyx_doc_7sklearn_4tree_5_tree_4Tree_14apply)},
  {__Pyx_NAMESTR("compute_feature_importances"), (PyCFunction)__pyx_pw_7sklearn_4tree_5_tree_4Tree_17compute_feature_importances, METH_VARARGS|METH_KEYWORDS, __Pyx_DOCSTR(__pyx_doc_7sklearn_4tree_5_tree_4Tree_16compute_feature_importances)},
  {0, 0, 0, 0}
};

static struct PyGetSetDef __pyx_getsets_7sklearn_4tree_5_tree_Tree[] = {
  {(char *)"n_classes", __pyx_getprop_7sklearn_4tree_5_tree_4Tree_n_classes, 0, 0, 0},
  {(char *)"children_left", __pyx_getprop_7sklearn_4tree_5_tree_4Tree_children_left, 0, 0, 0},
  {(char *)"children_right", __pyx_getprop_7sklearn_4tree_5_tree_4Tree_children_right, 0, 0, 0},
  {(char *)"feature", __pyx_getprop_7sklearn_4tree_5_tree_4Tree_feature, 0, 0, 0},
  {(char *)"threshold", __pyx_getprop_7sklearn_4tree_5_tree_4Tree_threshold, 0, 0, 0},
  {(char *)"value", __pyx_getprop_7sklearn_4tree_5_tree_4Tree_value, 0, 0, 0},
  {(char *)"impurity", __pyx_getprop_7sklearn_4tree_5_tree_4Tree_impurity, 0, 0, 0},
  {(char *)"n_node_samples", __pyx_getprop_7sklearn_4tree_5_tree_4Tree_n_node_samples, 0, 0, 0},
  {(char *)"n_features", __pyx_getprop_7sklearn_4tree_5_tree_4Tree_n_features, __pyx_setprop_7sklearn_4tree_5_tree_4Tree_n_features, 0, 0},
  {(char *)"n_outputs", __pyx_getprop_7sklearn_4tree_5_tree_4Tree_n_outputs, __pyx_setprop_7sklearn_4tree_5_tree_4Tree_n_outputs, 0, 0},
  {(char *)"max_n_classes", __pyx_getprop_7sklearn_4tree_5_tree_4Tree_max_n_classes, __pyx_setprop_7sklearn_4tree_5_tree_4Tree_max_n_classes, 0, 0},
  {(char *)"value_stride", __pyx_getprop_7sklearn_4tree_5_tree_4Tree_value_stride, __pyx_setprop_7sklearn_4tree_5_tree_4Tree_value_stride, 0, 0},
  {(char *)"splitter", __pyx_getprop_7sklearn_4tree_5_tree_4Tree_splitter, __pyx_setprop_7sklearn_4tree_5_tree_4Tree_splitter, 0, 0},
  {(char *)"max_depth", __pyx_getprop_7sklearn_4tree_5_tree_4Tree_max_depth, __pyx_setprop_7sklearn_4tree_5_tree_4Tree_max_depth, 0, 0},
  {(char *)"min_samples_split", __pyx_getprop_7sklearn_4tree_5_tree_4Tree_min_samples_split, __pyx_setprop_7sklearn_4tree_5_tree_4Tree_min_samples_split, 0, 0},
  {(char *)"min_samples_leaf", __pyx_getprop_7sklearn_4tree_5_tree_4Tree_min_samples_leaf, __pyx_setprop_7sklearn_4tree_5_tree_4Tree_min_samples_leaf, 0, 0},
  {(char *)"random_state", __pyx_getprop_7sklearn_4tree_5_tree_4Tree_random_state, __pyx_setprop_7sklearn_4tree_5_tree_4Tree_random_state, 0, 0},
  {(char *)"node_count", __pyx_getprop_7sklearn_4tree_5_tree_4Tree_node_count, __pyx_setprop_7sklearn_4tree_5_tree_4Tree_node_count, 0, 0},
  {(char *)"capacity", __pyx_getprop_7sklearn_4tree_5_tree_4Tree_capacity, __pyx_setprop_7sklearn_4tree_5_tree_4Tree_capacity, 0, 0},
  {0, 0, 0, 0, 0}
};

static PyTypeObject __pyx_type_7sklearn_4tree_5_tree_Tree = {
  PyVarObject_HEAD_INIT(0, 0)
  __Pyx_NAMESTR("sklearn.tree._tree.Tree"), /*tp_name*/
  sizeof(struct __pyx_obj_7sklearn_4tree_5_tree_Tree), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_7sklearn_4tree_5_tree_Tree, /*tp_dealloc*/
  0, /*tp_print*/
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #else
  0, /*reserved*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  __Pyx_DOCSTR("Struct-of-arrays representation of a binary decision tree.\n\n    The binary tree is represented as a number of parallel arrays. The i-th\n    element of each array holds information about the node `i`. Node 0 is the\n    tree's root. You can find a detailed description of all arrays in\n    `_tree.pxd`. NOTE: Some of the arrays only apply to either leaves or split\n    nodes, resp. In this case the values of nodes of the other type are\n    arbitrary!\n\n    Attributes\n    ----------\n    node_count : int\n        The number of nodes (internal nodes + leaves) in the tree.\n\n    capacity : int\n        The current capacity (i.e., size) of the arrays.\n\n    children_left : int*\n        children_left[i] holds the node id of the left child of node i.\n        For leaves, children_left[i] == TREE_LEAF. Otherwise,\n        children_left[i] > i. This child handles the case where\n        X[:, feature[i]] <= threshold[i].\n\n    children_right : int*\n        children_right[i] holds the node id of the right child of node i.\n        For leaves, children_right[i] == TREE_LEAF. Otherwise,\n        children_right[i] > i. This child handles the case where\n        X[:, feature[i]] > threshold[i].\n\n    feature : int*\n        feature[i] holds the feature to split on, for the internal node i.\n\n    threshold : double*\n        threshold[i] holds the threshold for the internal node i.\n\n    value : double*\n        Contains the constant prediction value of each node.\n\n    impurity : double*\n        impurity[i] holds the impurity (i.e., the value of the splitting\n        criterion) at node i.\n\n    n_node_samples : int*\n        n_samples[i] holds the number of training samples reaching node i.\n    "), /*tp_doc*/
  __pyx_tp_traverse_7sklearn_4tree_5_tree_Tree, /*tp_traverse*/
  __pyx_tp_clear_7sklearn_4tree_5_tree_Tree, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  __pyx_methods_7sklearn_4tree_5_tree_Tree, /*tp_methods*/
  0, /*tp_members*/
  __pyx_getsets_7sklearn_4tree_5_tree_Tree, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  0, /*tp_dictoffset*/
  0, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_7sklearn_4tree_5_tree_Tree, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  #if PY_VERSION_HEX >= 0x02060000
  0, /*tp_version_tag*/
  #endif
};
static struct __pyx_vtabstruct_7sklearn_4tree_5_tree_Splitter __pyx_vtable_7sklearn_4tree_5_tree_Splitter;

static PyObject *__pyx_tp_new_7sklearn_4tree_5_tree_Splitter(PyTypeObject *t, PyObject *a, PyObject *k) {
  struct __pyx_obj_7sklearn_4tree_5_tree_Splitter *p;
  PyObject *o;
  o = (*t->tp_alloc)(t, 0);
  if (unlikely(!o)) return 0;
  p = ((struct __pyx_obj_7sklearn_4tree_5_tree_Splitter *)o);
  p->__pyx_vtab = __pyx_vtabptr_7sklearn_4tree_5_tree_Splitter;
  p->criterion = ((struct __pyx_obj_7sklearn_4tree_5_tree_Criterion *)Py_None); Py_INCREF(Py_None);
  p->random_state = Py_None; Py_INCREF(Py_None);
  p->X = ((PyArrayObject *)Py_None); Py_INCREF(Py_None);
  if (unlikely(__pyx_pw_7sklearn_4tree_5_tree_8Splitter_1__cinit__(o, a, k) < 0)) {
    Py_DECREF(o); o = 0;
  }
  return o;
}

static void __pyx_tp_dealloc_7sklearn_4tree_5_tree_Splitter(PyObject *o) {
  struct __pyx_obj_7sklearn_4tree_5_tree_Splitter *p = (struct __pyx_obj_7sklearn_4tree_5_tree_Splitter *)o;
  PyObject_GC_UnTrack(o);
  {
    PyObject *etype, *eval, *etb;
    PyErr_Fetch(&etype, &eval, &etb);
    ++Py_REFCNT(o);
    __pyx_pw_7sklearn_4tree_5_tree_8Splitter_3__dealloc__(o);
    if (PyErr_Occurred()) PyErr_WriteUnraisable(o);
    --Py_REFCNT(o);
    PyErr_Restore(etype, eval, etb);
  }
  Py_CLEAR(p->criterion);
  Py_CLEAR(p->random_state);
  Py_CLEAR(p->X);
  (*Py_TYPE(o)->tp_free)(o);
}

static int __pyx_tp_traverse_7sklearn_4tree_5_tree_Splitter(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_7sklearn_4tree_5_tree_Splitter *p = (struct __pyx_obj_7sklearn_4tree_5_tree_Splitter *)o;
  if (p->criterion) {
    e = (*v)(((PyObject*)p->criterion), a); if (e) return e;
  }
  if (p->random_state) {
    e = (*v)(p->random_state, a); if (e) return e;
  }
  if (p->X) {
    e = (*v)(((PyObject*)p->X), a); if (e) return e;
  }
  return 0;
}

static int __pyx_tp_clear_7sklearn_4tree_5_tree_Splitter(PyObject *o) {
  struct __pyx_obj_7sklearn_4tree_5_tree_Splitter *p = (struct __pyx_obj_7sklearn_4tree_5_tree_Splitter *)o;
  PyObject* tmp;
  tmp = ((PyObject*)p->criterion);
  p->criterion = ((struct __pyx_obj_7sklearn_4tree_5_tree_Criterion *)Py_None); Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->random_state);
  p->random_state = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->X);
  p->X = ((PyArrayObject *)Py_None); Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  return 0;
}

static PyObject *__pyx_getprop_7sklearn_4tree_5_tree_8Splitter_criterion(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_7sklearn_4tree_5_tree_8Splitter_9criterion_1__get__(o);
}

static int __pyx_setprop_7sklearn_4tree_5_tree_8Splitter_criterion(PyObject *o, PyObject *v, CYTHON_UNUSED void *x) {
  if (v) {
    return __pyx_pw_7sklearn_4tree_5_tree_8Splitter_9criterion_3__set__(o, v);
  }
  else {
    return __pyx_pw_7sklearn_4tree_5_tree_8Splitter_9criterion_5__del__(o);
  }
}

static PyObject *__pyx_getprop_7sklearn_4tree_5_tree_8Splitter_max_features(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_7sklearn_4tree_5_tree_8Splitter_12max_features_1__get__(o);
}

static int __pyx_setprop_7sklearn_4tree_5_tree_8Splitter_max_features(PyObject *o, PyObject *v, CYTHON_UNUSED void *x) {
  if (v) {
    return __pyx_pw_7sklearn_4tree_5_tree_8Splitter_12max_features_3__set__(o, v);
  }
  else {
    PyErr_SetString(PyExc_NotImplementedError, "__del__");
    return -1;
  }
}

static PyObject *__pyx_getprop_7sklearn_4tree_5_tree_8Splitter_min_samples_leaf(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_7sklearn_4tree_5_tree_8Splitter_16min_samples_leaf_1__get__(o);
}

static int __pyx_setprop_7sklearn_4tree_5_tree_8Splitter_min_samples_leaf(PyObject *o, PyObject *v, CYTHON_UNUSED void *x) {
  if (v) {
    return __pyx_pw_7sklearn_4tree_5_tree_8Splitter_16min_samples_leaf_3__set__(o, v);
  }
  else {
    PyErr_SetString(PyExc_NotImplementedError, "__del__");
    return -1;
  }
}

static PyMethodDef __pyx_methods_7sklearn_4tree_5_tree_Splitter[] = {
  {__Pyx_NAMESTR("__getstate__"), (PyCFunction)__pyx_pw_7sklearn_4tree_5_tree_8Splitter_5__getstate__, METH_NOARGS, __Pyx_DOCSTR(0)},
  {__Pyx_NAMESTR("__setstate__"), (PyCFunction)__pyx_pw_7sklearn_4tree_5_tree_8Splitter_7__setstate__, METH_O, __Pyx_DOCSTR(0)},
  {0, 0, 0, 0}
};

static struct PyGetSetDef __pyx_getsets_7sklearn_4tree_5_tree_Splitter[] = {
  {(char *)"criterion", __pyx_getprop_7sklearn_4tree_5_tree_8Splitter_criterion, __pyx_setprop_7sklearn_4tree_5_tree_8Splitter_criterion, 0, 0},
  {(char *)"max_features", __pyx_getprop_7sklearn_4tree_5_tree_8Splitter_max_features, __pyx_setprop_7sklearn_4tree_5_tree_8Splitter_max_features, 0, 0},
  {(char *)"min_samples_leaf", __pyx_getprop_7sklearn_4tree_5_tree_8Splitter_min_samples_leaf, __pyx_setprop_7sklearn_4tree_5_tree_8Splitter_min_samples_leaf, 0, 0},
  {0, 0, 0, 0, 0}
};

static PyTypeObject __pyx_type_7sklearn_4tree_5_tree_Splitter = {
  PyVarObject_HEAD_INIT(0, 0)
  __Pyx_NAMESTR("sklearn.tree._tree.Splitter"), /*tp_name*/
  sizeof(struct __pyx_obj_7sklearn_4tree_5_tree_Splitter), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_7sklearn_4tree_5_tree_Splitter, /*tp_dealloc*/
  0, /*tp_print*/
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #else
  0, /*reserved*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  0, /*tp_doc*/
  __pyx_tp_traverse_7sklearn_4tree_5_tree_Splitter, /*tp_traverse*/
  __pyx_tp_clear_7sklearn_4tree_5_tree_Splitter, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  __pyx_methods_7sklearn_4tree_5_tree_Splitter, /*tp_methods*/
  0, /*tp_members*/
  __pyx_getsets_7sklearn_4tree_5_tree_Splitter, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  0, /*tp_dictoffset*/
  0, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_7sklearn_4tree_5_tree_Splitter, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  #if PY_VERSION_HEX >= 0x02060000
  0, /*tp_version_tag*/
  #endif
};
static struct __pyx_vtabstruct_7sklearn_4tree_5_tree_RandomSplitter __pyx_vtable_7sklearn_4tree_5_tree_RandomSplitter;

static PyObject *__pyx_tp_new_7sklearn_4tree_5_tree_RandomSplitter(PyTypeObject *t, PyObject *a, PyObject *k) {
  struct __pyx_obj_7sklearn_4tree_5_tree_RandomSplitter *p;
  PyObject *o = __pyx_tp_new_7sklearn_4tree_5_tree_Splitter(t, a, k);
  if (unlikely(!o)) return 0;
  p = ((struct __pyx_obj_7sklearn_4tree_5_tree_RandomSplitter *)o);
  p->__pyx_base.__pyx_vtab = (struct __pyx_vtabstruct_7sklearn_4tree_5_tree_Splitter*)__pyx_vtabptr_7sklearn_4tree_5_tree_RandomSplitter;
  return o;
}

static PyMethodDef __pyx_methods_7sklearn_4tree_5_tree_RandomSplitter[] = {
  {__Pyx_NAMESTR("__reduce__"), (PyCFunction)__pyx_pw_7sklearn_4tree_5_tree_14RandomSplitter_1__reduce__, METH_NOARGS, __Pyx_DOCSTR(0)},
  {0, 0, 0, 0}
};

static PyTypeObject __pyx_type_7sklearn_4tree_5_tree_RandomSplitter = {
  PyVarObject_HEAD_INIT(0, 0)
  __Pyx_NAMESTR("sklearn.tree._tree.RandomSplitter"), /*tp_name*/
  sizeof(struct __pyx_obj_7sklearn_4tree_5_tree_RandomSplitter), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_7sklearn_4tree_5_tree_Splitter, /*tp_dealloc*/
  0, /*tp_print*/
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #else
  0, /*reserved*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  __Pyx_DOCSTR("Splitter for finding the best random split."), /*tp_doc*/
  __pyx_tp_traverse_7sklearn_4tree_5_tree_Splitter, /*tp_traverse*/
  __pyx_tp_clear_7sklearn_4tree_5_tree_Splitter, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  __pyx_methods_7sklearn_4tree_5_tree_RandomSplitter, /*tp_methods*/
  0, /*tp_members*/
  0, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  0, /*tp_dictoffset*/
  0, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_7sklearn_4tree_5_tree_RandomSplitter, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  #if PY_VERSION_HEX >= 0x02060000
  0, /*tp_version_tag*/
  #endif
};
static struct __pyx_vtabstruct_7sklearn_4tree_5_tree_Criterion __pyx_vtable_7sklearn_4tree_5_tree_Criterion;

static PyObject *__pyx_tp_new_7sklearn_4tree_5_tree_Criterion(PyTypeObject *t, CYTHON_UNUSED PyObject *a, CYTHON_UNUSED PyObject *k) {
  struct __pyx_obj_7sklearn_4tree_5_tree_Criterion *p;
  PyObject *o;
  o = (*t->tp_alloc)(t, 0);
  if (unlikely(!o)) return 0;
  p = ((struct __pyx_obj_7sklearn_4tree_5_tree_Criterion *)o);
  p->__pyx_vtab = __pyx_vtabptr_7sklearn_4tree_5_tree_Criterion;
  return o;
}

static void __pyx_tp_dealloc_7sklearn_4tree_5_tree_Criterion(PyObject *o) {
  (*Py_TYPE(o)->tp_free)(o);
}

static PyMethodDef __pyx_methods_7sklearn_4tree_5_tree_Criterion[] = {
  {0, 0, 0, 0}
};

static PyTypeObject __pyx_type_7sklearn_4tree_5_tree_Criterion = {
  PyVarObject_HEAD_INIT(0, 0)
  __Pyx_NAMESTR("sklearn.tree._tree.Criterion"), /*tp_name*/
  sizeof(struct __pyx_obj_7sklearn_4tree_5_tree_Criterion), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_7sklearn_4tree_5_tree_Criterion, /*tp_dealloc*/
  0, /*tp_print*/
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #else
  0, /*reserved*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE, /*tp_flags*/
  __Pyx_DOCSTR("Interface for impurity criteria."), /*tp_doc*/
  0, /*tp_traverse*/
  0, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  __pyx_methods_7sklearn_4tree_5_tree_Criterion, /*tp_methods*/
  0, /*tp_members*/
  0, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  0, /*tp_dictoffset*/
  0, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_7sklearn_4tree_5_tree_Criterion, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  #if PY_VERSION_HEX >= 0x02060000
  0, /*tp_version_tag*/
  #endif
};
static struct __pyx_vtabstruct_7sklearn_4tree_5_tree_ClassificationCriterion __pyx_vtable_7sklearn_4tree_5_tree_ClassificationCriterion;

static PyObject *__pyx_tp_new_7sklearn_4tree_5_tree_ClassificationCriterion(PyTypeObject *t, PyObject *a, PyObject *k) {
  struct __pyx_obj_7sklearn_4tree_5_tree_ClassificationCriterion *p;
  PyObject *o = __pyx_tp_new_7sklearn_4tree_5_tree_Criterion(t, a, k);
  if (unlikely(!o)) return 0;
  p = ((struct __pyx_obj_7sklearn_4tree_5_tree_ClassificationCriterion *)o);
  p->__pyx_base.__pyx_vtab = (struct __pyx_vtabstruct_7sklearn_4tree_5_tree_Criterion*)__pyx_vtabptr_7sklearn_4tree_5_tree_ClassificationCriterion;
  if (unlikely(__pyx_pw_7sklearn_4tree_5_tree_23ClassificationCriterion_1__cinit__(o, a, k) < 0)) {
    Py_DECREF(o); o = 0;
  }
  return o;
}

static void __pyx_tp_dealloc_7sklearn_4tree_5_tree_ClassificationCriterion(PyObject *o) {
  {
    PyObject *etype, *eval, *etb;
    PyErr_Fetch(&etype, &eval, &etb);
    ++Py_REFCNT(o);
    __pyx_pw_7sklearn_4tree_5_tree_23ClassificationCriterion_3__dealloc__(o);
    if (PyErr_Occurred()) PyErr_WriteUnraisable(o);
    --Py_REFCNT(o);
    PyErr_Restore(etype, eval, etb);
  }
  __pyx_tp_dealloc_7sklearn_4tree_5_tree_Criterion(o);
}

static PyMethodDef __pyx_methods_7sklearn_4tree_5_tree_ClassificationCriterion[] = {
  {__Pyx_NAMESTR("__reduce__"), (PyCFunction)__pyx_pw_7sklearn_4tree_5_tree_23ClassificationCriterion_5__reduce__, METH_NOARGS, __Pyx_DOCSTR(0)},
  {__Pyx_NAMESTR("__getstate__"), (PyCFunction)__pyx_pw_7sklearn_4tree_5_tree_23ClassificationCriterion_7__getstate__, METH_NOARGS, __Pyx_DOCSTR(0)},
  {__Pyx_NAMESTR("__setstate__"), (PyCFunction)__pyx_pw_7sklearn_4tree_5_tree_23ClassificationCriterion_9__setstate__, METH_O, __Pyx_DOCSTR(0)},
  {0, 0, 0, 0}
};

static PyTypeObject __pyx_type_7sklearn_4tree_5_tree_ClassificationCriterion = {
  PyVarObject_HEAD_INIT(0, 0)
  __Pyx_NAMESTR("sklearn.tree._tree.ClassificationCriterion"), /*tp_name*/
  sizeof(struct __pyx_obj_7sklearn_4tree_5_tree_ClassificationCriterion), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_7sklearn_4tree_5_tree_ClassificationCriterion, /*tp_dealloc*/
  0, /*tp_print*/
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #else
  0, /*reserved*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE, /*tp_flags*/
  __Pyx_DOCSTR("Abstract criterion for classification."), /*tp_doc*/
  0, /*tp_traverse*/
  0, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  __pyx_methods_7sklearn_4tree_5_tree_ClassificationCriterion, /*tp_methods*/
  0, /*tp_members*/
  0, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  0, /*tp_dictoffset*/
  0, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_7sklearn_4tree_5_tree_ClassificationCriterion, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  #if PY_VERSION_HEX >= 0x02060000
  0, /*tp_version_tag*/
  #endif
};
static struct __pyx_vtabstruct_7sklearn_4tree_5_tree_Gini __pyx_vtable_7sklearn_4tree_5_tree_Gini;

static PyObject *__pyx_tp_new_7sklearn_4tree_5_tree_Gini(PyTypeObject *t, PyObject *a, PyObject *k) {
  struct __pyx_obj_7sklearn_4tree_5_tree_Gini *p;
  PyObject *o = __pyx_tp_new_7sklearn_4tree_5_tree_ClassificationCriterion(t, a, k);
  if (unlikely(!o)) return 0;
  p = ((struct __pyx_obj_7sklearn_4tree_5_tree_Gini *)o);
  p->__pyx_base.__pyx_base.__pyx_vtab = (struct __pyx_vtabstruct_7sklearn_4tree_5_tree_Criterion*)__pyx_vtabptr_7sklearn_4tree_5_tree_Gini;
  return o;
}

static PyMethodDef __pyx_methods_7sklearn_4tree_5_tree_Gini[] = {
  {0, 0, 0, 0}
};

static PyTypeObject __pyx_type_7sklearn_4tree_5_tree_Gini = {
  PyVarObject_HEAD_INIT(0, 0)
  __Pyx_NAMESTR("sklearn.tree._tree.Gini"), /*tp_name*/
  sizeof(struct __pyx_obj_7sklearn_4tree_5_tree_Gini), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_7sklearn_4tree_5_tree_ClassificationCriterion, /*tp_dealloc*/
  0, /*tp_print*/
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #else
  0, /*reserved*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE, /*tp_flags*/
  __Pyx_DOCSTR("Gini Index impurity criteria.\n\n    Let the target be a classification outcome taking values in 0, 1, ..., K-1.\n    If node m represents a region Rm with Nm observations, then let\n\n        pmk = 1/ Nm \\sum_{x_i in Rm} I(yi = k)\n\n    be the proportion of class k observations in node m.\n\n    The Gini Index is then defined as:\n\n        index = \\sum_{k=0}^{K-1} pmk (1 - pmk)\n              = 1 - \\sum_{k=0}^{K-1} pmk ** 2\n    "), /*tp_doc*/
  0, /*tp_traverse*/
  0, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  __pyx_methods_7sklearn_4tree_5_tree_Gini, /*tp_methods*/
  0, /*tp_members*/
  0, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  0, /*tp_dictoffset*/
  0, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_7sklearn_4tree_5_tree_Gini, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  #if PY_VERSION_HEX >= 0x02060000
  0, /*tp_version_tag*/
  #endif
};
static struct __pyx_vtabstruct_7sklearn_4tree_5_tree_Entropy __pyx_vtable_7sklearn_4tree_5_tree_Entropy;

static PyObject *__pyx_tp_new_7sklearn_4tree_5_tree_Entropy(PyTypeObject *t, PyObject *a, PyObject *k) {
  struct __pyx_obj_7sklearn_4tree_5_tree_Entropy *p;
  PyObject *o = __pyx_tp_new_7sklearn_4tree_5_tree_ClassificationCriterion(t, a, k);
  if (unlikely(!o)) return 0;
  p = ((struct __pyx_obj_7sklearn_4tree_5_tree_Entropy *)o);
  p->__pyx_base.__pyx_base.__pyx_vtab = (struct __pyx_vtabstruct_7sklearn_4tree_5_tree_Criterion*)__pyx_vtabptr_7sklearn_4tree_5_tree_Entropy;
  return o;
}

static PyMethodDef __pyx_methods_7sklearn_4tree_5_tree_Entropy[] = {
  {0, 0, 0, 0}
};

static PyTypeObject __pyx_type_7sklearn_4tree_5_tree_Entropy = {
  PyVarObject_HEAD_INIT(0, 0)
  __Pyx_NAMESTR("sklearn.tree._tree.Entropy"), /*tp_name*/
  sizeof(struct __pyx_obj_7sklearn_4tree_5_tree_Entropy), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_7sklearn_4tree_5_tree_ClassificationCriterion, /*tp_dealloc*/
  0, /*tp_print*/
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #else
  0, /*reserved*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE, /*tp_flags*/
  __Pyx_DOCSTR("Cross Entropy impurity criteria.\n\n    Let the target be a classification outcome taking values in 0, 1, ..., K-1.\n    If node m represents a region Rm with Nm observations, then let\n\n        pmk = 1/ Nm \\sum_{x_i in Rm} I(yi = k)\n\n    be the proportion of class k observations in node m.\n\n    The cross-entropy is then defined as\n\n        cross-entropy = - \\sum_{k=0}^{K-1} pmk log(pmk)\n    "), /*tp_doc*/
  0, /*tp_traverse*/
  0, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  __pyx_methods_7sklearn_4tree_5_tree_Entropy, /*tp_methods*/
  0, /*tp_members*/
  0, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  0, /*tp_dictoffset*/
  0, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_7sklearn_4tree_5_tree_Entropy, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  #if PY_VERSION_HEX >= 0x02060000
  0, /*tp_version_tag*/
  #endif
};
static struct __pyx_vtabstruct_7sklearn_4tree_5_tree_RegressionCriterion __pyx_vtable_7sklearn_4tree_5_tree_RegressionCriterion;

static PyObject *__pyx_tp_new_7sklearn_4tree_5_tree_RegressionCriterion(PyTypeObject *t, PyObject *a, PyObject *k) {
  struct __pyx_obj_7sklearn_4tree_5_tree_RegressionCriterion *p;
  PyObject *o = __pyx_tp_new_7sklearn_4tree_5_tree_Criterion(t, a, k);
  if (unlikely(!o)) return 0;
  p = ((struct __pyx_obj_7sklearn_4tree_5_tree_RegressionCriterion *)o);
  p->__pyx_base.__pyx_vtab = (struct __pyx_vtabstruct_7sklearn_4tree_5_tree_Criterion*)__pyx_vtabptr_7sklearn_4tree_5_tree_RegressionCriterion;
  if (unlikely(__pyx_pw_7sklearn_4tree_5_tree_19RegressionCriterion_1__cinit__(o, a, k) < 0)) {
    Py_DECREF(o); o = 0;
  }
  return o;
}

static void __pyx_tp_dealloc_7sklearn_4tree_5_tree_RegressionCriterion(PyObject *o) {
  {
    PyObject *etype, *eval, *etb;
    PyErr_Fetch(&etype, &eval, &etb);
    ++Py_REFCNT(o);
    __pyx_pw_7sklearn_4tree_5_tree_19RegressionCriterion_3__dealloc__(o);
    if (PyErr_Occurred()) PyErr_WriteUnraisable(o);
    --Py_REFCNT(o);
    PyErr_Restore(etype, eval, etb);
  }
  __pyx_tp_dealloc_7sklearn_4tree_5_tree_Criterion(o);
}

static PyMethodDef __pyx_methods_7sklearn_4tree_5_tree_RegressionCriterion[] = {
  {__Pyx_NAMESTR("__reduce__"), (PyCFunction)__pyx_pw_7sklearn_4tree_5_tree_19RegressionCriterion_5__reduce__, METH_NOARGS, __Pyx_DOCSTR(0)},
  {__Pyx_NAMESTR("__getstate__"), (PyCFunction)__pyx_pw_7sklearn_4tree_5_tree_19RegressionCriterion_7__getstate__, METH_NOARGS, __Pyx_DOCSTR(0)},
  {__Pyx_NAMESTR("__setstate__"), (PyCFunction)__pyx_pw_7sklearn_4tree_5_tree_19RegressionCriterion_9__setstate__, METH_O, __Pyx_DOCSTR(0)},
  {0, 0, 0, 0}
};

static PyTypeObject __pyx_type_7sklearn_4tree_5_tree_RegressionCriterion = {
  PyVarObject_HEAD_INIT(0, 0)
  __Pyx_NAMESTR("sklearn.tree._tree.RegressionCriterion"), /*tp_name*/
  sizeof(struct __pyx_obj_7sklearn_4tree_5_tree_RegressionCriterion), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_7sklearn_4tree_5_tree_RegressionCriterion, /*tp_dealloc*/
  0, /*tp_print*/
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #else
  0, /*reserved*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE, /*tp_flags*/
  __Pyx_DOCSTR("Abstract criterion for regression.\n\n    Computes variance of the target values left and right of the split point.\n    Computation is linear in `n_samples` by using ::\n\n        var = \\sum_i^n (y_i - y_bar) ** 2\n            = (\\sum_i^n y_i ** 2) - n_samples y_bar ** 2\n    "), /*tp_doc*/
  0, /*tp_traverse*/
  0, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  __pyx_methods_7sklearn_4tree_5_tree_RegressionCriterion, /*tp_methods*/
  0, /*tp_members*/
  0, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  0, /*tp_dictoffset*/
  0, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_7sklearn_4tree_5_tree_RegressionCriterion, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  #if PY_VERSION_HEX >= 0x02060000
  0, /*tp_version_tag*/
  #endif
};
static struct __pyx_vtabstruct_7sklearn_4tree_5_tree_MSE __pyx_vtable_7sklearn_4tree_5_tree_MSE;

static PyObject *__pyx_tp_new_7sklearn_4tree_5_tree_MSE(PyTypeObject *t, PyObject *a, PyObject *k) {
  struct __pyx_obj_7sklearn_4tree_5_tree_MSE *p;
  PyObject *o = __pyx_tp_new_7sklearn_4tree_5_tree_RegressionCriterion(t, a, k);
  if (unlikely(!o)) return 0;
  p = ((struct __pyx_obj_7sklearn_4tree_5_tree_MSE *)o);
  p->__pyx_base.__pyx_base.__pyx_vtab = (struct __pyx_vtabstruct_7sklearn_4tree_5_tree_Criterion*)__pyx_vtabptr_7sklearn_4tree_5_tree_MSE;
  return o;
}

static PyMethodDef __pyx_methods_7sklearn_4tree_5_tree_MSE[] = {
  {0, 0, 0, 0}
};

static PyTypeObject __pyx_type_7sklearn_4tree_5_tree_MSE = {
  PyVarObject_HEAD_INIT(0, 0)
  __Pyx_NAMESTR("sklearn.tree._tree.MSE"), /*tp_name*/
  sizeof(struct __pyx_obj_7sklearn_4tree_5_tree_MSE), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_7sklearn_4tree_5_tree_RegressionCriterion, /*tp_dealloc*/
  0, /*tp_print*/
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #else
  0, /*reserved*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE, /*tp_flags*/
  __Pyx_DOCSTR("Mean squared error impurity criterion.\n\n        MSE = var_left + var_right\n    "), /*tp_doc*/
  0, /*tp_traverse*/
  0, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  __pyx_methods_7sklearn_4tree_5_tree_MSE, /*tp_methods*/
  0, /*tp_members*/
  0, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  0, /*tp_dictoffset*/
  0, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_7sklearn_4tree_5_tree_MSE, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  #if PY_VERSION_HEX >= 0x02060000
  0, /*tp_version_tag*/
  #endif
};
static struct __pyx_vtabstruct_7sklearn_4tree_5_tree_BestSplitter __pyx_vtable_7sklearn_4tree_5_tree_BestSplitter;

static PyObject *__pyx_tp_new_7sklearn_4tree_5_tree_BestSplitter(PyTypeObject *t, PyObject *a, PyObject *k) {
  struct __pyx_obj_7sklearn_4tree_5_tree_BestSplitter *p;
  PyObject *o = __pyx_tp_new_7sklearn_4tree_5_tree_Splitter(t, a, k);
  if (unlikely(!o)) return 0;
  p = ((struct __pyx_obj_7sklearn_4tree_5_tree_BestSplitter *)o);
  p->__pyx_base.__pyx_vtab = (struct __pyx_vtabstruct_7sklearn_4tree_5_tree_Splitter*)__pyx_vtabptr_7sklearn_4tree_5_tree_BestSplitter;
  return o;
}

static PyMethodDef __pyx_methods_7sklearn_4tree_5_tree_BestSplitter[] = {
  {__Pyx_NAMESTR("__reduce__"), (PyCFunction)__pyx_pw_7sklearn_4tree_5_tree_12BestSplitter_1__reduce__, METH_NOARGS, __Pyx_DOCSTR(0)},
  {0, 0, 0, 0}
};

static PyTypeObject __pyx_type_7sklearn_4tree_5_tree_BestSplitter = {
  PyVarObject_HEAD_INIT(0, 0)
  __Pyx_NAMESTR("sklearn.tree._tree.BestSplitter"), /*tp_name*/
  sizeof(struct __pyx_obj_7sklearn_4tree_5_tree_BestSplitter), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_7sklearn_4tree_5_tree_Splitter, /*tp_dealloc*/
  0, /*tp_print*/
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #else
  0, /*reserved*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  __Pyx_DOCSTR("Splitter for finding the best split."), /*tp_doc*/
  __pyx_tp_traverse_7sklearn_4tree_5_tree_Splitter, /*tp_traverse*/
  __pyx_tp_clear_7sklearn_4tree_5_tree_Splitter, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  __pyx_methods_7sklearn_4tree_5_tree_BestSplitter, /*tp_methods*/
  0, /*tp_members*/
  0, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  0, /*tp_dictoffset*/
  0, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_7sklearn_4tree_5_tree_BestSplitter, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  #if PY_VERSION_HEX >= 0x02060000
  0, /*tp_version_tag*/
  #endif
};

static PyMethodDef __pyx_methods[] = {
  {0, 0, 0, 0}
};

#if PY_MAJOR_VERSION >= 3
static struct PyModuleDef __pyx_moduledef = {
  #if PY_VERSION_HEX < 0x03020000
    { PyObject_HEAD_INIT(NULL) NULL, 0, NULL },
  #else
    PyModuleDef_HEAD_INIT,
  #endif
    __Pyx_NAMESTR("_tree"),
    0, /* m_doc */
    -1, /* m_size */
    __pyx_methods /* m_methods */,
    NULL, /* m_reload */
    NULL, /* m_traverse */
    NULL, /* m_clear */
    NULL /* m_free */
};
#endif

static __Pyx_StringTabEntry __pyx_string_tab[] = {
  {&__pyx_kp_u_10, __pyx_k_10, sizeof(__pyx_k_10), 0, 1, 0, 0},
  {&__pyx_kp_u_13, __pyx_k_13, sizeof(__pyx_k_13), 0, 1, 0, 0},
  {&__pyx_n_s_2, __pyx_k_2, sizeof(__pyx_k_2), 0, 0, 1, 1},
  {&__pyx_kp_u_3, __pyx_k_3, sizeof(__pyx_k_3), 0, 1, 0, 0},
  {&__pyx_kp_u_5, __pyx_k_5, sizeof(__pyx_k_5), 0, 1, 0, 0},
  {&__pyx_kp_u_7, __pyx_k_7, sizeof(__pyx_k_7), 0, 1, 0, 0},
  {&__pyx_kp_u_9, __pyx_k_9, sizeof(__pyx_k_9), 0, 1, 0, 0},
  {&__pyx_n_s__C, __pyx_k__C, sizeof(__pyx_k__C), 0, 0, 1, 1},
  {&__pyx_n_s__DOUBLE, __pyx_k__DOUBLE, sizeof(__pyx_k__DOUBLE), 0, 0, 1, 1},
  {&__pyx_n_s__DTYPE, __pyx_k__DTYPE, sizeof(__pyx_k__DTYPE), 0, 0, 1, 1},
  {&__pyx_n_s__MemoryError, __pyx_k__MemoryError, sizeof(__pyx_k__MemoryError), 0, 0, 1, 1},
  {&__pyx_n_s__RuntimeError, __pyx_k__RuntimeError, sizeof(__pyx_k__RuntimeError), 0, 0, 1, 1},
  {&__pyx_n_s__TREE_LEAF, __pyx_k__TREE_LEAF, sizeof(__pyx_k__TREE_LEAF), 0, 0, 1, 1},
  {&__pyx_n_s__TREE_UNDEFINED, __pyx_k__TREE_UNDEFINED, sizeof(__pyx_k__TREE_UNDEFINED), 0, 0, 1, 1},
  {&__pyx_n_s__ValueError, __pyx_k__ValueError, sizeof(__pyx_k__ValueError), 0, 0, 1, 1},
  {&__pyx_n_s__X, __pyx_k__X, sizeof(__pyx_k__X), 0, 0, 1, 1},
  {&__pyx_n_s____getstate__, __pyx_k____getstate__, sizeof(__pyx_k____getstate__), 0, 0, 1, 1},
  {&__pyx_n_s____import__, __pyx_k____import__, sizeof(__pyx_k____import__), 0, 0, 1, 1},
  {&__pyx_n_s____main__, __pyx_k____main__, sizeof(__pyx_k____main__), 0, 0, 1, 1},
  {&__pyx_n_s____pyx_getbuffer, __pyx_k____pyx_getbuffer, sizeof(__pyx_k____pyx_getbuffer), 0, 0, 1, 1},
  {&__pyx_n_s____pyx_releasebuffer, __pyx_k____pyx_releasebuffer, sizeof(__pyx_k____pyx_releasebuffer), 0, 0, 1, 1},
  {&__pyx_n_s____pyx_vtable__, __pyx_k____pyx_vtable__, sizeof(__pyx_k____pyx_vtable__), 0, 0, 1, 1},
  {&__pyx_n_s____test__, __pyx_k____test__, sizeof(__pyx_k____test__), 0, 0, 1, 1},
  {&__pyx_n_s__apply, __pyx_k__apply, sizeof(__pyx_k__apply), 0, 0, 1, 1},
  {&__pyx_n_s__asarray, __pyx_k__asarray, sizeof(__pyx_k__asarray), 0, 0, 1, 1},
  {&__pyx_n_s__build, __pyx_k__build, sizeof(__pyx_k__build), 0, 0, 1, 1},
  {&__pyx_n_s__capacity, __pyx_k__capacity, sizeof(__pyx_k__capacity), 0, 0, 1, 1},
  {&__pyx_n_s__children_left, __pyx_k__children_left, sizeof(__pyx_k__children_left), 0, 0, 1, 1},
  {&__pyx_n_s__children_right, __pyx_k__children_right, sizeof(__pyx_k__children_right), 0, 0, 1, 1},
  {&__pyx_n_s__contiguous, __pyx_k__contiguous, sizeof(__pyx_k__contiguous), 0, 0, 1, 1},
  {&__pyx_n_s__criterion, __pyx_k__criterion, sizeof(__pyx_k__criterion), 0, 0, 1, 1},
  {&__pyx_n_s__dtype, __pyx_k__dtype, sizeof(__pyx_k__dtype), 0, 0, 1, 1},
  {&__pyx_n_s__feature, __pyx_k__feature, sizeof(__pyx_k__feature), 0, 0, 1, 1},
  {&__pyx_n_s__flags, __pyx_k__flags, sizeof(__pyx_k__flags), 0, 0, 1, 1},
  {&__pyx_n_s__float32, __pyx_k__float32, sizeof(__pyx_k__float32), 0, 0, 1, 1},
  {&__pyx_n_s__float64, __pyx_k__float64, sizeof(__pyx_k__float64), 0, 0, 1, 1},
  {&__pyx_n_s__impurity, __pyx_k__impurity, sizeof(__pyx_k__impurity), 0, 0, 1, 1},
  {&__pyx_n_s__inf, __pyx_k__inf, sizeof(__pyx_k__inf), 0, 0, 1, 1},
  {&__pyx_n_s__int32, __pyx_k__int32, sizeof(__pyx_k__int32), 0, 0, 1, 1},
  {&__pyx_n_s__itemsize, __pyx_k__itemsize, sizeof(__pyx_k__itemsize), 0, 0, 1, 1},
  {&__pyx_n_s__max, __pyx_k__max, sizeof(__pyx_k__max), 0, 0, 1, 1},
  {&__pyx_n_s__max_depth, __pyx_k__max_depth, sizeof(__pyx_k__max_depth), 0, 0, 1, 1},
  {&__pyx_n_s__max_features, __pyx_k__max_features, sizeof(__pyx_k__max_features), 0, 0, 1, 1},
  {&__pyx_n_s__min_samples_leaf, __pyx_k__min_samples_leaf, sizeof(__pyx_k__min_samples_leaf), 0, 0, 1, 1},
  {&__pyx_n_s__min_samples_split, __pyx_k__min_samples_split, sizeof(__pyx_k__min_samples_split), 0, 0, 1, 1},
  {&__pyx_n_s__n_classes, __pyx_k__n_classes, sizeof(__pyx_k__n_classes), 0, 0, 1, 1},
  {&__pyx_n_s__n_features, __pyx_k__n_features, sizeof(__pyx_k__n_features), 0, 0, 1, 1},
  {&__pyx_n_s__n_node_samples, __pyx_k__n_node_samples, sizeof(__pyx_k__n_node_samples), 0, 0, 1, 1},
  {&__pyx_n_s__n_outputs, __pyx_k__n_outputs, sizeof(__pyx_k__n_outputs), 0, 0, 1, 1},
  {&__pyx_n_s__node_count, __pyx_k__node_count, sizeof(__pyx_k__node_count), 0, 0, 1, 1},
  {&__pyx_n_s__normalize, __pyx_k__normalize, sizeof(__pyx_k__normalize), 0, 0, 1, 1},
  {&__pyx_n_s__np, __pyx_k__np, sizeof(__pyx_k__np), 0, 0, 1, 1},
  {&__pyx_n_s__numpy, __pyx_k__numpy, sizeof(__pyx_k__numpy), 0, 0, 1, 1},
  {&__pyx_n_s__order, __pyx_k__order, sizeof(__pyx_k__order), 0, 0, 1, 1},
  {&__pyx_n_s__predict, __pyx_k__predict, sizeof(__pyx_k__predict), 0, 0, 1, 1},
  {&__pyx_n_s__randint, __pyx_k__randint, sizeof(__pyx_k__randint), 0, 0, 1, 1},
  {&__pyx_n_s__random_state, __pyx_k__random_state, sizeof(__pyx_k__random_state), 0, 0, 1, 1},
  {&__pyx_n_s__range, __pyx_k__range, sizeof(__pyx_k__range), 0, 0, 1, 1},
  {&__pyx_n_s__sample_weight, __pyx_k__sample_weight, sizeof(__pyx_k__sample_weight), 0, 0, 1, 1},
  {&__pyx_n_s__splitter, __pyx_k__splitter, sizeof(__pyx_k__splitter), 0, 0, 1, 1},
  {&__pyx_n_s__sum, __pyx_k__sum, sizeof(__pyx_k__sum), 0, 0, 1, 1},
  {&__pyx_n_s__threshold, __pyx_k__threshold, sizeof(__pyx_k__threshold), 0, 0, 1, 1},
  {&__pyx_n_s__value, __pyx_k__value, sizeof(__pyx_k__value), 0, 0, 1, 1},
  {&__pyx_n_s__y, __pyx_k__y, sizeof(__pyx_k__y), 0, 0, 1, 1},
  {&__pyx_n_s__zeros, __pyx_k__zeros, sizeof(__pyx_k__zeros), 0, 0, 1, 1},
  {0, 0, 0, 0, 0, 0, 0}
};
static int __Pyx_InitCachedBuiltins(void) {
  __pyx_builtin_MemoryError = __Pyx_GetBuiltinName(__pyx_n_s__MemoryError); if (!__pyx_builtin_MemoryError) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 117; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __pyx_builtin_ValueError = __Pyx_GetBuiltinName(__pyx_n_s__ValueError); if (!__pyx_builtin_ValueError) {__pyx_filename = __pyx_f[2]; __pyx_lineno = 215; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __pyx_builtin_range = __Pyx_GetBuiltinName(__pyx_n_s__range); if (!__pyx_builtin_range) {__pyx_filename = __pyx_f[2]; __pyx_lineno = 228; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __pyx_builtin_RuntimeError = __Pyx_GetBuiltinName(__pyx_n_s__RuntimeError); if (!__pyx_builtin_RuntimeError) {__pyx_filename = __pyx_f[2]; __pyx_lineno = 799; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  return 0;
  __pyx_L1_error:;
  return -1;
}

static int __Pyx_InitCachedConstants(void) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__Pyx_InitCachedConstants", 0);

  /* "numpy.pxd":215
 *             if ((flags & pybuf.PyBUF_C_CONTIGUOUS == pybuf.PyBUF_C_CONTIGUOUS)
 *                 and not PyArray_CHKFLAGS(self, NPY_C_CONTIGUOUS)):
 *                 raise ValueError(u"ndarray is not C contiguous")             # <<<<<<<<<<<<<<
 * 
 *             if ((flags & pybuf.PyBUF_F_CONTIGUOUS == pybuf.PyBUF_F_CONTIGUOUS)
 */
  __pyx_k_tuple_4 = PyTuple_Pack(1, ((PyObject *)__pyx_kp_u_3)); if (unlikely(!__pyx_k_tuple_4)) {__pyx_filename = __pyx_f[2]; __pyx_lineno = 215; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_k_tuple_4);
  __Pyx_GIVEREF(((PyObject *)__pyx_k_tuple_4));

  /* "numpy.pxd":219
 *             if ((flags & pybuf.PyBUF_F_CONTIGUOUS == pybuf.PyBUF_F_CONTIGUOUS)
 *                 and not PyArray_CHKFLAGS(self, NPY_F_CONTIGUOUS)):
 *                 raise ValueError(u"ndarray is not Fortran contiguous")             # <<<<<<<<<<<<<<
 * 
 *             info.buf = PyArray_DATA(self)
 */
  __pyx_k_tuple_6 = PyTuple_Pack(1, ((PyObject *)__pyx_kp_u_5)); if (unlikely(!__pyx_k_tuple_6)) {__pyx_filename = __pyx_f[2]; __pyx_lineno = 219; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_k_tuple_6);
  __Pyx_GIVEREF(((PyObject *)__pyx_k_tuple_6));

  /* "numpy.pxd":257
 *                 if ((descr.byteorder == c'>' and little_endian) or
 *                     (descr.byteorder == c'<' and not little_endian)):
 *                     raise ValueError(u"Non-native byte order not supported")             # <<<<<<<<<<<<<<
 *                 if   t == NPY_BYTE:        f = "b"
 *                 elif t == NPY_UBYTE:       f = "B"
 */
  __pyx_k_tuple_8 = PyTuple_Pack(1, ((PyObject *)__pyx_kp_u_7)); if (unlikely(!__pyx_k_tuple_8)) {__pyx_filename = __pyx_f[2]; __pyx_lineno = 257; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_k_tuple_8);
  __Pyx_GIVEREF(((PyObject *)__pyx_k_tuple_8));

  /* "numpy.pxd":799
 * 
 *         if (end - f) - (new_offset - offset[0]) < 15:
 *             raise RuntimeError(u"Format string allocated too short, see comment in numpy.pxd")             # <<<<<<<<<<<<<<
 * 
 *         if ((child.byteorder == c'>' and little_endian) or
 */
  __pyx_k_tuple_11 = PyTuple_Pack(1, ((PyObject *)__pyx_kp_u_10)); if (unlikely(!__pyx_k_tuple_11)) {__pyx_filename = __pyx_f[2]; __pyx_lineno = 799; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_k_tuple_11);
  __Pyx_GIVEREF(((PyObject *)__pyx_k_tuple_11));

  /* "numpy.pxd":803
 *         if ((child.byteorder == c'>' and little_endian) or
 *             (child.byteorder == c'<' and not little_endian)):
 *             raise ValueError(u"Non-native byte order not supported")             # <<<<<<<<<<<<<<
 *             # One could encode it in the format string and have Cython
 *             # complain instead, BUT: < and > in format strings also imply
 */
  __pyx_k_tuple_12 = PyTuple_Pack(1, ((PyObject *)__pyx_kp_u_7)); if (unlikely(!__pyx_k_tuple_12)) {__pyx_filename = __pyx_f[2]; __pyx_lineno = 803; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_k_tuple_12);
  __Pyx_GIVEREF(((PyObject *)__pyx_k_tuple_12));

  /* "numpy.pxd":823
 *             t = child.type_num
 *             if end - f < 5:
 *                 raise RuntimeError(u"Format string allocated too short.")             # <<<<<<<<<<<<<<
 * 
 *             # Until ticket #99 is fixed, use integers to avoid warnings
 */
  __pyx_k_tuple_14 = PyTuple_Pack(1, ((PyObject *)__pyx_kp_u_13)); if (unlikely(!__pyx_k_tuple_14)) {__pyx_filename = __pyx_f[2]; __pyx_lineno = 823; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_k_tuple_14);
  __Pyx_GIVEREF(((PyObject *)__pyx_k_tuple_14));
  __Pyx_RefNannyFinishContext();
  return 0;
  __pyx_L1_error:;
  __Pyx_RefNannyFinishContext();
  return -1;
}

static int __Pyx_InitGlobals(void) {
  if (__Pyx_InitStrings(__pyx_string_tab) < 0) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1; __pyx_clineno = __LINE__; goto __pyx_L1_error;};
  __pyx_int_0 = PyInt_FromLong(0); if (unlikely(!__pyx_int_0)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1; __pyx_clineno = __LINE__; goto __pyx_L1_error;};
  __pyx_int_neg_1 = PyInt_FromLong(-1); if (unlikely(!__pyx_int_neg_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1; __pyx_clineno = __LINE__; goto __pyx_L1_error;};
  __pyx_int_neg_2 = PyInt_FromLong(-2); if (unlikely(!__pyx_int_neg_2)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1; __pyx_clineno = __LINE__; goto __pyx_L1_error;};
  __pyx_int_15 = PyInt_FromLong(15); if (unlikely(!__pyx_int_15)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1; __pyx_clineno = __LINE__; goto __pyx_L1_error;};
  return 0;
  __pyx_L1_error:;
  return -1;
}

#if PY_MAJOR_VERSION < 3
PyMODINIT_FUNC init_tree(void); /*proto*/
PyMODINIT_FUNC init_tree(void)
#else
PyMODINIT_FUNC PyInit__tree(void); /*proto*/
PyMODINIT_FUNC PyInit__tree(void)
#endif
{
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  double __pyx_t_3;
  __pyx_t_7sklearn_4tree_5_tree_SIZE_t __pyx_t_4;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannyDeclarations
  #if CYTHON_REFNANNY
  __Pyx_RefNanny = __Pyx_RefNannyImportAPI("refnanny");
  if (!__Pyx_RefNanny) {
      PyErr_Clear();
      __Pyx_RefNanny = __Pyx_RefNannyImportAPI("Cython.Runtime.refnanny");
      if (!__Pyx_RefNanny)
          Py_FatalError("failed to import 'refnanny' module");
  }
  #endif
  __Pyx_RefNannySetupContext("PyMODINIT_FUNC PyInit__tree(void)", 0);
  if ( __Pyx_check_binary_version() < 0) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __pyx_empty_tuple = PyTuple_New(0); if (unlikely(!__pyx_empty_tuple)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __pyx_empty_bytes = PyBytes_FromStringAndSize("", 0); if (unlikely(!__pyx_empty_bytes)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  #ifdef __Pyx_CyFunction_USED
  if (__Pyx_CyFunction_init() < 0) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  #endif
  #ifdef __Pyx_FusedFunction_USED
  if (__pyx_FusedFunction_init() < 0) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  #endif
  #ifdef __Pyx_Generator_USED
  if (__pyx_Generator_init() < 0) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  #endif
  /*--- Library function declarations ---*/
  /*--- Threads initialization code ---*/
  #if defined(__PYX_FORCE_INIT_THREADS) && __PYX_FORCE_INIT_THREADS
  #ifdef WITH_THREAD /* Python build with threading support? */
  PyEval_InitThreads();
  #endif
  #endif
  /*--- Module creation code ---*/
  #if PY_MAJOR_VERSION < 3
  __pyx_m = Py_InitModule4(__Pyx_NAMESTR("_tree"), __pyx_methods, 0, 0, PYTHON_API_VERSION); Py_XINCREF(__pyx_m);
  #else
  __pyx_m = PyModule_Create(&__pyx_moduledef);
  #endif
  if (unlikely(!__pyx_m)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __pyx_d = PyModule_GetDict(__pyx_m); if (unlikely(!__pyx_d)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  Py_INCREF(__pyx_d);
  #if PY_MAJOR_VERSION >= 3
  {
    PyObject *modules = PyImport_GetModuleDict(); if (unlikely(!modules)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    if (!PyDict_GetItemString(modules, "sklearn.tree._tree")) {
      if (unlikely(PyDict_SetItemString(modules, "sklearn.tree._tree", __pyx_m) < 0)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    }
  }
  #endif
  __pyx_b = PyImport_AddModule(__Pyx_NAMESTR(__Pyx_BUILTIN_MODULE_NAME)); if (unlikely(!__pyx_b)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  #if CYTHON_COMPILING_IN_PYPY
  Py_INCREF(__pyx_b);
  #endif
  if (__Pyx_SetAttrString(__pyx_m, "__builtins__", __pyx_b) < 0) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1; __pyx_clineno = __LINE__; goto __pyx_L1_error;};
  /*--- Initialize various global constants etc. ---*/
  if (unlikely(__Pyx_InitGlobals() < 0)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  #if PY_MAJOR_VERSION < 3 && (__PYX_DEFAULT_STRING_ENCODING_IS_ASCII || __PYX_DEFAULT_STRING_ENCODING_IS_DEFAULT)
  if (__Pyx_init_sys_getdefaultencoding_params() < 0) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  #endif
  if (__pyx_module_is_main_sklearn__tree___tree) {
    if (__Pyx_SetAttrString(__pyx_m, "__name__", __pyx_n_s____main__) < 0) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1; __pyx_clineno = __LINE__; goto __pyx_L1_error;};
  }
  /*--- Builtin init code ---*/
  if (unlikely(__Pyx_InitCachedBuiltins() < 0)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  /*--- Constants init code ---*/
  if (unlikely(__Pyx_InitCachedConstants() < 0)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  /*--- Global init code ---*/
  /*--- Variable export code ---*/
  /*--- Function export code ---*/
  /*--- Type init code ---*/
  __pyx_vtabptr_7sklearn_4tree_5_tree_Tree = &__pyx_vtable_7sklearn_4tree_5_tree_Tree;
  __pyx_vtable_7sklearn_4tree_5_tree_Tree._add_node = (__pyx_t_7sklearn_4tree_5_tree_SIZE_t (*)(struct __pyx_obj_7sklearn_4tree_5_tree_Tree *, __pyx_t_7sklearn_4tree_5_tree_SIZE_t, int, int, __pyx_t_7sklearn_4tree_5_tree_SIZE_t, double, double, __pyx_t_7sklearn_4tree_5_tree_SIZE_t))__pyx_f_7sklearn_4tree_5_tree_4Tree__add_node;
  __pyx_vtable_7sklearn_4tree_5_tree_Tree._resize = (void (*)(struct __pyx_obj_7sklearn_4tree_5_tree_Tree *, struct __pyx_opt_args_7sklearn_4tree_5_tree_4Tree__resize *__pyx_optional_args))__pyx_f_7sklearn_4tree_5_tree_4Tree__resize;
  __pyx_vtable_7sklearn_4tree_5_tree_Tree.build = (PyObject *(*)(struct __pyx_obj_7sklearn_4tree_5_tree_Tree *, PyArrayObject *, PyArrayObject *, int __pyx_skip_dispatch, struct __pyx_opt_args_7sklearn_4tree_5_tree_4Tree_build *__pyx_optional_args))__pyx_f_7sklearn_4tree_5_tree_4Tree_build;
  __pyx_vtable_7sklearn_4tree_5_tree_Tree.predict = (PyObject *(*)(struct __pyx_obj_7sklearn_4tree_5_tree_Tree *, PyArrayObject *, int __pyx_skip_dispatch))__pyx_f_7sklearn_4tree_5_tree_4Tree_predict;
  __pyx_vtable_7sklearn_4tree_5_tree_Tree.apply = (PyObject *(*)(struct __pyx_obj_7sklearn_4tree_5_tree_Tree *, PyArrayObject *, int __pyx_skip_dispatch))__pyx_f_7sklearn_4tree_5_tree_4Tree_apply;
  __pyx_vtable_7sklearn_4tree_5_tree_Tree.compute_feature_importances = (PyObject *(*)(struct __pyx_obj_7sklearn_4tree_5_tree_Tree *, int __pyx_skip_dispatch, struct __pyx_opt_args_7sklearn_4tree_5_tree_4Tree_compute_feature_importances *__pyx_optional_args))__pyx_f_7sklearn_4tree_5_tree_4Tree_compute_feature_importances;
  if (PyType_Ready(&__pyx_type_7sklearn_4tree_5_tree_Tree) < 0) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1196; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  if (__Pyx_SetVtable(__pyx_type_7sklearn_4tree_5_tree_Tree.tp_dict, __pyx_vtabptr_7sklearn_4tree_5_tree_Tree) < 0) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1196; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  if (__Pyx_SetAttrString(__pyx_m, "Tree", (PyObject *)&__pyx_type_7sklearn_4tree_5_tree_Tree) < 0) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1196; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __pyx_ptype_7sklearn_4tree_5_tree_Tree = &__pyx_type_7sklearn_4tree_5_tree_Tree;
  __pyx_vtabptr_7sklearn_4tree_5_tree_Splitter = &__pyx_vtable_7sklearn_4tree_5_tree_Splitter;
  __pyx_vtable_7sklearn_4tree_5_tree_Splitter.init = (void (*)(struct __pyx_obj_7sklearn_4tree_5_tree_Splitter *, PyArrayObject *, PyArrayObject *, __pyx_t_7sklearn_4tree_5_tree_DOUBLE_t *))__pyx_f_7sklearn_4tree_5_tree_8Splitter_init;
  __pyx_vtable_7sklearn_4tree_5_tree_Splitter.node_reset = (void (*)(struct __pyx_obj_7sklearn_4tree_5_tree_Splitter *, __pyx_t_7sklearn_4tree_5_tree_SIZE_t, __pyx_t_7sklearn_4tree_5_tree_SIZE_t, double *))__pyx_f_7sklearn_4tree_5_tree_8Splitter_node_reset;
  __pyx_vtable_7sklearn_4tree_5_tree_Splitter.node_split = (void (*)(struct __pyx_obj_7sklearn_4tree_5_tree_Splitter *, __pyx_t_7sklearn_4tree_5_tree_SIZE_t *, __pyx_t_7sklearn_4tree_5_tree_SIZE_t *, double *))__pyx_f_7sklearn_4tree_5_tree_8Splitter_node_split;
  __pyx_vtable_7sklearn_4tree_5_tree_Splitter.node_value = (void (*)(struct __pyx_obj_7sklearn_4tree_5_tree_Splitter *, double *))__pyx_f_7sklearn_4tree_5_tree_8Splitter_node_value;
  if (PyType_Ready(&__pyx_type_7sklearn_4tree_5_tree_Splitter) < 0) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 790; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  if (__Pyx_SetVtable(__pyx_type_7sklearn_4tree_5_tree_Splitter.tp_dict, __pyx_vtabptr_7sklearn_4tree_5_tree_Splitter) < 0) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 790; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  if (__Pyx_SetAttrString(__pyx_m, "Splitter", (PyObject *)&__pyx_type_7sklearn_4tree_5_tree_Splitter) < 0) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 790; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __pyx_ptype_7sklearn_4tree_5_tree_Splitter = &__pyx_type_7sklearn_4tree_5_tree_Splitter;
  __pyx_vtabptr_7sklearn_4tree_5_tree_RandomSplitter = &__pyx_vtable_7sklearn_4tree_5_tree_RandomSplitter;
  __pyx_vtable_7sklearn_4tree_5_tree_RandomSplitter.__pyx_base = *__pyx_vtabptr_7sklearn_4tree_5_tree_Splitter;
  __pyx_vtable_7sklearn_4tree_5_tree_RandomSplitter.__pyx_base.node_split = (void (*)(struct __pyx_obj_7sklearn_4tree_5_tree_Splitter *, __pyx_t_7sklearn_4tree_5_tree_SIZE_t *, __pyx_t_7sklearn_4tree_5_tree_SIZE_t *, double *))__pyx_f_7sklearn_4tree_5_tree_14RandomSplitter_node_split;
  __pyx_type_7sklearn_4tree_5_tree_RandomSplitter.tp_base = __pyx_ptype_7sklearn_4tree_5_tree_Splitter;
  if (PyType_Ready(&__pyx_type_7sklearn_4tree_5_tree_RandomSplitter) < 0) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1054; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  if (__Pyx_SetVtable(__pyx_type_7sklearn_4tree_5_tree_RandomSplitter.tp_dict, __pyx_vtabptr_7sklearn_4tree_5_tree_RandomSplitter) < 0) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1054; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  if (__Pyx_SetAttrString(__pyx_m, "RandomSplitter", (PyObject *)&__pyx_type_7sklearn_4tree_5_tree_RandomSplitter) < 0) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1054; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __pyx_ptype_7sklearn_4tree_5_tree_RandomSplitter = &__pyx_type_7sklearn_4tree_5_tree_RandomSplitter;
  __pyx_vtabptr_7sklearn_4tree_5_tree_Criterion = &__pyx_vtable_7sklearn_4tree_5_tree_Criterion;
  __pyx_vtable_7sklearn_4tree_5_tree_Criterion.init = (void (*)(struct __pyx_obj_7sklearn_4tree_5_tree_Criterion *, __pyx_t_7sklearn_4tree_5_tree_DOUBLE_t *, __pyx_t_7sklearn_4tree_5_tree_SIZE_t, __pyx_t_7sklearn_4tree_5_tree_DOUBLE_t *, __pyx_t_7sklearn_4tree_5_tree_SIZE_t *, __pyx_t_7sklearn_4tree_5_tree_SIZE_t, __pyx_t_7sklearn_4tree_5_tree_SIZE_t))__pyx_f_7sklearn_4tree_5_tree_9Criterion_init;
  __pyx_vtable_7sklearn_4tree_5_tree_Criterion.reset = (void (*)(struct __pyx_obj_7sklearn_4tree_5_tree_Criterion *))__pyx_f_7sklearn_4tree_5_tree_9Criterion_reset;
  __pyx_vtable_7sklearn_4tree_5_tree_Criterion.update = (void (*)(struct __pyx_obj_7sklearn_4tree_5_tree_Criterion *, __pyx_t_7sklearn_4tree_5_tree_SIZE_t))__pyx_f_7sklearn_4tree_5_tree_9Criterion_update;
  __pyx_vtable_7sklearn_4tree_5_tree_Criterion.node_impurity = (double (*)(struct __pyx_obj_7sklearn_4tree_5_tree_Criterion *))__pyx_f_7sklearn_4tree_5_tree_9Criterion_node_impurity;
  __pyx_vtable_7sklearn_4tree_5_tree_Criterion.children_impurity = (double (*)(struct __pyx_obj_7sklearn_4tree_5_tree_Criterion *))__pyx_f_7sklearn_4tree_5_tree_9Criterion_children_impurity;
  __pyx_vtable_7sklearn_4tree_5_tree_Criterion.node_value = (void (*)(struct __pyx_obj_7sklearn_4tree_5_tree_Criterion *, double *))__pyx_f_7sklearn_4tree_5_tree_9Criterion_node_value;
  if (PyType_Ready(&__pyx_type_7sklearn_4tree_5_tree_Criterion) < 0) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 52; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  if (__Pyx_SetVtable(__pyx_type_7sklearn_4tree_5_tree_Criterion.tp_dict, __pyx_vtabptr_7sklearn_4tree_5_tree_Criterion) < 0) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 52; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  if (__Pyx_SetAttrString(__pyx_m, "Criterion", (PyObject *)&__pyx_type_7sklearn_4tree_5_tree_Criterion) < 0) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 52; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __pyx_ptype_7sklearn_4tree_5_tree_Criterion = &__pyx_type_7sklearn_4tree_5_tree_Criterion;
  __pyx_vtabptr_7sklearn_4tree_5_tree_ClassificationCriterion = &__pyx_vtable_7sklearn_4tree_5_tree_ClassificationCriterion;
  __pyx_vtable_7sklearn_4tree_5_tree_ClassificationCriterion.__pyx_base = *__pyx_vtabptr_7sklearn_4tree_5_tree_Criterion;
  __pyx_vtable_7sklearn_4tree_5_tree_ClassificationCriterion.__pyx_base.init = (void (*)(struct __pyx_obj_7sklearn_4tree_5_tree_Criterion *, __pyx_t_7sklearn_4tree_5_tree_DOUBLE_t *, __pyx_t_7sklearn_4tree_5_tree_SIZE_t, __pyx_t_7sklearn_4tree_5_tree_DOUBLE_t *, __pyx_t_7sklearn_4tree_5_tree_SIZE_t *, __pyx_t_7sklearn_4tree_5_tree_SIZE_t, __pyx_t_7sklearn_4tree_5_tree_SIZE_t))__pyx_f_7sklearn_4tree_5_tree_23ClassificationCriterion_init;
  __pyx_vtable_7sklearn_4tree_5_tree_ClassificationCriterion.__pyx_base.reset = (void (*)(struct __pyx_obj_7sklearn_4tree_5_tree_Criterion *))__pyx_f_7sklearn_4tree_5_tree_23ClassificationCriterion_reset;
  __pyx_vtable_7sklearn_4tree_5_tree_ClassificationCriterion.__pyx_base.update = (void (*)(struct __pyx_obj_7sklearn_4tree_5_tree_Criterion *, __pyx_t_7sklearn_4tree_5_tree_SIZE_t))__pyx_f_7sklearn_4tree_5_tree_23ClassificationCriterion_update;
  __pyx_vtable_7sklearn_4tree_5_tree_ClassificationCriterion.__pyx_base.node_impurity = (double (*)(struct __pyx_obj_7sklearn_4tree_5_tree_Criterion *))__pyx_f_7sklearn_4tree_5_tree_23ClassificationCriterion_node_impurity;
  __pyx_vtable_7sklearn_4tree_5_tree_ClassificationCriterion.__pyx_base.children_impurity = (double (*)(struct __pyx_obj_7sklearn_4tree_5_tree_Criterion *))__pyx_f_7sklearn_4tree_5_tree_23ClassificationCriterion_children_impurity;
  __pyx_vtable_7sklearn_4tree_5_tree_ClassificationCriterion.__pyx_base.node_value = (void (*)(struct __pyx_obj_7sklearn_4tree_5_tree_Criterion *, double *))__pyx_f_7sklearn_4tree_5_tree_23ClassificationCriterion_node_value;
  __pyx_type_7sklearn_4tree_5_tree_ClassificationCriterion.tp_base = __pyx_ptype_7sklearn_4tree_5_tree_Criterion;
  if (PyType_Ready(&__pyx_type_7sklearn_4tree_5_tree_ClassificationCriterion) < 0) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 89; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  if (__Pyx_SetVtable(__pyx_type_7sklearn_4tree_5_tree_ClassificationCriterion.tp_dict, __pyx_vtabptr_7sklearn_4tree_5_tree_ClassificationCriterion) < 0) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 89; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  if (__Pyx_SetAttrString(__pyx_m, "ClassificationCriterion", (PyObject *)&__pyx_type_7sklearn_4tree_5_tree_ClassificationCriterion) < 0) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 89; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __pyx_ptype_7sklearn_4tree_5_tree_ClassificationCriterion = &__pyx_type_7sklearn_4tree_5_tree_ClassificationCriterion;
  __pyx_vtabptr_7sklearn_4tree_5_tree_Gini = &__pyx_vtable_7sklearn_4tree_5_tree_Gini;
  __pyx_vtable_7sklearn_4tree_5_tree_Gini.__pyx_base = *__pyx_vtabptr_7sklearn_4tree_5_tree_ClassificationCriterion;
  __pyx_vtable_7sklearn_4tree_5_tree_Gini.__pyx_base.__pyx_base.node_impurity = (double (*)(struct __pyx_obj_7sklearn_4tree_5_tree_Criterion *))__pyx_f_7sklearn_4tree_5_tree_4Gini_node_impurity;
  __pyx_vtable_7sklearn_4tree_5_tree_Gini.__pyx_base.__pyx_base.children_impurity = (double (*)(struct __pyx_obj_7sklearn_4tree_5_tree_Criterion *))__pyx_f_7sklearn_4tree_5_tree_4Gini_children_impurity;
  __pyx_type_7sklearn_4tree_5_tree_Gini.tp_base = __pyx_ptype_7sklearn_4tree_5_tree_ClassificationCriterion;
  if (PyType_Ready(&__pyx_type_7sklearn_4tree_5_tree_Gini) < 0) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 402; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  if (__Pyx_SetVtable(__pyx_type_7sklearn_4tree_5_tree_Gini.tp_dict, __pyx_vtabptr_7sklearn_4tree_5_tree_Gini) < 0) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 402; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  if (__Pyx_SetAttrString(__pyx_m, "Gini", (PyObject *)&__pyx_type_7sklearn_4tree_5_tree_Gini) < 0) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 402; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __pyx_ptype_7sklearn_4tree_5_tree_Gini = &__pyx_type_7sklearn_4tree_5_tree_Gini;
  __pyx_vtabptr_7sklearn_4tree_5_tree_Entropy = &__pyx_vtable_7sklearn_4tree_5_tree_Entropy;
  __pyx_vtable_7sklearn_4tree_5_tree_Entropy.__pyx_base = *__pyx_vtabptr_7sklearn_4tree_5_tree_ClassificationCriterion;
  __pyx_vtable_7sklearn_4tree_5_tree_Entropy.__pyx_base.__pyx_base.node_impurity = (double (*)(struct __pyx_obj_7sklearn_4tree_5_tree_Criterion *))__pyx_f_7sklearn_4tree_5_tree_7Entropy_node_impurity;
  __pyx_vtable_7sklearn_4tree_5_tree_Entropy.__pyx_base.__pyx_base.children_impurity = (double (*)(struct __pyx_obj_7sklearn_4tree_5_tree_Criterion *))__pyx_f_7sklearn_4tree_5_tree_7Entropy_children_impurity;
  __pyx_type_7sklearn_4tree_5_tree_Entropy.tp_base = __pyx_ptype_7sklearn_4tree_5_tree_ClassificationCriterion;
  if (PyType_Ready(&__pyx_type_7sklearn_4tree_5_tree_Entropy) < 0) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 314; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  if (__Pyx_SetVtable(__pyx_type_7sklearn_4tree_5_tree_Entropy.tp_dict, __pyx_vtabptr_7sklearn_4tree_5_tree_Entropy) < 0) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 314; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  if (__Pyx_SetAttrString(__pyx_m, "Entropy", (PyObject *)&__pyx_type_7sklearn_4tree_5_tree_Entropy) < 0) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 314; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __pyx_ptype_7sklearn_4tree_5_tree_Entropy = &__pyx_type_7sklearn_4tree_5_tree_Entropy;
  __pyx_vtabptr_7sklearn_4tree_5_tree_RegressionCriterion = &__pyx_vtable_7sklearn_4tree_5_tree_RegressionCriterion;
  __pyx_vtable_7sklearn_4tree_5_tree_RegressionCriterion.__pyx_base = *__pyx_vtabptr_7sklearn_4tree_5_tree_Criterion;
  __pyx_vtable_7sklearn_4tree_5_tree_RegressionCriterion.__pyx_base.init = (void (*)(struct __pyx_obj_7sklearn_4tree_5_tree_Criterion *, __pyx_t_7sklearn_4tree_5_tree_DOUBLE_t *, __pyx_t_7sklearn_4tree_5_tree_SIZE_t, __pyx_t_7sklearn_4tree_5_tree_DOUBLE_t *, __pyx_t_7sklearn_4tree_5_tree_SIZE_t *, __pyx_t_7sklearn_4tree_5_tree_SIZE_t, __pyx_t_7sklearn_4tree_5_tree_SIZE_t))__pyx_f_7sklearn_4tree_5_tree_19RegressionCriterion_init;
  __pyx_vtable_7sklearn_4tree_5_tree_RegressionCriterion.__pyx_base.reset = (void (*)(struct __pyx_obj_7sklearn_4tree_5_tree_Criterion *))__pyx_f_7sklearn_4tree_5_tree_19RegressionCriterion_reset;
  __pyx_vtable_7sklearn_4tree_5_tree_RegressionCriterion.__pyx_base.update = (void (*)(struct __pyx_obj_7sklearn_4tree_5_tree_Criterion *, __pyx_t_7sklearn_4tree_5_tree_SIZE_t))__pyx_f_7sklearn_4tree_5_tree_19RegressionCriterion_update;
  __pyx_vtable_7sklearn_4tree_5_tree_RegressionCriterion.__pyx_base.node_impurity = (double (*)(struct __pyx_obj_7sklearn_4tree_5_tree_Criterion *))__pyx_f_7sklearn_4tree_5_tree_19RegressionCriterion_node_impurity;
  __pyx_vtable_7sklearn_4tree_5_tree_RegressionCriterion.__pyx_base.children_impurity = (double (*)(struct __pyx_obj_7sklearn_4tree_5_tree_Criterion *))__pyx_f_7sklearn_4tree_5_tree_19RegressionCriterion_children_impurity;
  __pyx_vtable_7sklearn_4tree_5_tree_RegressionCriterion.__pyx_base.node_value = (void (*)(struct __pyx_obj_7sklearn_4tree_5_tree_Criterion *, double *))__pyx_f_7sklearn_4tree_5_tree_19RegressionCriterion_node_value;
  __pyx_type_7sklearn_4tree_5_tree_RegressionCriterion.tp_base = __pyx_ptype_7sklearn_4tree_5_tree_Criterion;
  if (PyType_Ready(&__pyx_type_7sklearn_4tree_5_tree_RegressionCriterion) < 0) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 492; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  if (__Pyx_SetVtable(__pyx_type_7sklearn_4tree_5_tree_RegressionCriterion.tp_dict, __pyx_vtabptr_7sklearn_4tree_5_tree_RegressionCriterion) < 0) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 492; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  if (__Pyx_SetAttrString(__pyx_m, "RegressionCriterion", (PyObject *)&__pyx_type_7sklearn_4tree_5_tree_RegressionCriterion) < 0) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 492; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __pyx_ptype_7sklearn_4tree_5_tree_RegressionCriterion = &__pyx_type_7sklearn_4tree_5_tree_RegressionCriterion;
  __pyx_vtabptr_7sklearn_4tree_5_tree_MSE = &__pyx_vtable_7sklearn_4tree_5_tree_MSE;
  __pyx_vtable_7sklearn_4tree_5_tree_MSE.__pyx_base = *__pyx_vtabptr_7sklearn_4tree_5_tree_RegressionCriterion;
  __pyx_vtable_7sklearn_4tree_5_tree_MSE.__pyx_base.__pyx_base.node_impurity = (double (*)(struct __pyx_obj_7sklearn_4tree_5_tree_Criterion *))__pyx_f_7sklearn_4tree_5_tree_3MSE_node_impurity;
  __pyx_vtable_7sklearn_4tree_5_tree_MSE.__pyx_base.__pyx_base.children_impurity = (double (*)(struct __pyx_obj_7sklearn_4tree_5_tree_Criterion *))__pyx_f_7sklearn_4tree_5_tree_3MSE_children_impurity;
  __pyx_type_7sklearn_4tree_5_tree_MSE.tp_base = __pyx_ptype_7sklearn_4tree_5_tree_RegressionCriterion;
  if (PyType_Ready(&__pyx_type_7sklearn_4tree_5_tree_MSE) < 0) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 748; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  if (__Pyx_SetVtable(__pyx_type_7sklearn_4tree_5_tree_MSE.tp_dict, __pyx_vtabptr_7sklearn_4tree_5_tree_MSE) < 0) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 748; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  if (__Pyx_SetAttrString(__pyx_m, "MSE", (PyObject *)&__pyx_type_7sklearn_4tree_5_tree_MSE) < 0) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 748; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __pyx_ptype_7sklearn_4tree_5_tree_MSE = &__pyx_type_7sklearn_4tree_5_tree_MSE;
  __pyx_vtabptr_7sklearn_4tree_5_tree_BestSplitter = &__pyx_vtable_7sklearn_4tree_5_tree_BestSplitter;
  __pyx_vtable_7sklearn_4tree_5_tree_BestSplitter.__pyx_base = *__pyx_vtabptr_7sklearn_4tree_5_tree_Splitter;
  __pyx_vtable_7sklearn_4tree_5_tree_BestSplitter.__pyx_base.node_split = (void (*)(struct __pyx_obj_7sklearn_4tree_5_tree_Splitter *, __pyx_t_7sklearn_4tree_5_tree_SIZE_t *, __pyx_t_7sklearn_4tree_5_tree_SIZE_t *, double *))__pyx_f_7sklearn_4tree_5_tree_12BestSplitter_node_split;
  __pyx_type_7sklearn_4tree_5_tree_BestSplitter.tp_base = __pyx_ptype_7sklearn_4tree_5_tree_Splitter;
  if (PyType_Ready(&__pyx_type_7sklearn_4tree_5_tree_BestSplitter) < 0) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 892; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  if (__Pyx_SetVtable(__pyx_type_7sklearn_4tree_5_tree_BestSplitter.tp_dict, __pyx_vtabptr_7sklearn_4tree_5_tree_BestSplitter) < 0) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 892; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  if (__Pyx_SetAttrString(__pyx_m, "BestSplitter", (PyObject *)&__pyx_type_7sklearn_4tree_5_tree_BestSplitter) < 0) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 892; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __pyx_ptype_7sklearn_4tree_5_tree_BestSplitter = &__pyx_type_7sklearn_4tree_5_tree_BestSplitter;
  /*--- Type import code ---*/
  __pyx_ptype_7cpython_4type_type = __Pyx_ImportType(__Pyx_BUILTIN_MODULE_NAME, "type", 
  #if CYTHON_COMPILING_IN_PYPY
  sizeof(PyTypeObject),
  #else
  sizeof(PyHeapTypeObject),
  #endif
  0); if (unlikely(!__pyx_ptype_7cpython_4type_type)) {__pyx_filename = __pyx_f[3]; __pyx_lineno = 9; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __pyx_ptype_5numpy_dtype = __Pyx_ImportType("numpy", "dtype", sizeof(PyArray_Descr), 0); if (unlikely(!__pyx_ptype_5numpy_dtype)) {__pyx_filename = __pyx_f[2]; __pyx_lineno = 155; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __pyx_ptype_5numpy_flatiter = __Pyx_ImportType("numpy", "flatiter", sizeof(PyArrayIterObject), 0); if (unlikely(!__pyx_ptype_5numpy_flatiter)) {__pyx_filename = __pyx_f[2]; __pyx_lineno = 165; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __pyx_ptype_5numpy_broadcast = __Pyx_ImportType("numpy", "broadcast", sizeof(PyArrayMultiIterObject), 0); if (unlikely(!__pyx_ptype_5numpy_broadcast)) {__pyx_filename = __pyx_f[2]; __pyx_lineno = 169; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __pyx_ptype_5numpy_ndarray = __Pyx_ImportType("numpy", "ndarray", sizeof(PyArrayObject), 0); if (unlikely(!__pyx_ptype_5numpy_ndarray)) {__pyx_filename = __pyx_f[2]; __pyx_lineno = 178; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __pyx_ptype_5numpy_ufunc = __Pyx_ImportType("numpy", "ufunc", sizeof(PyUFuncObject), 0); if (unlikely(!__pyx_ptype_5numpy_ufunc)) {__pyx_filename = __pyx_f[2]; __pyx_lineno = 861; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  /*--- Variable import code ---*/
  /*--- Function import code ---*/
  /*--- Execution code ---*/

  /* "sklearn/tree/_tree.pyx":23
 * from libc.math cimport log as ln
 * 
 * import numpy as np             # <<<<<<<<<<<<<<
 * cimport numpy as np
 * np.import_array()
 */
  __pyx_t_1 = __Pyx_Import(((PyObject *)__pyx_n_s__numpy), 0, -1); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 23; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s__np, __pyx_t_1) < 0) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 23; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "sklearn/tree/_tree.pyx":25
 * import numpy as np
 * cimport numpy as np
 * np.import_array()             # <<<<<<<<<<<<<<
 * 
 * cdef enum:
 */
  import_array();

  /* "sklearn/tree/_tree.pyx":38
 * # =============================================================================
 * 
 * from numpy import float32 as DTYPE             # <<<<<<<<<<<<<<
 * from numpy import float64 as DOUBLE
 * 
 */
  __pyx_t_1 = PyList_New(1); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 38; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_INCREF(((PyObject *)__pyx_n_s__float32));
  PyList_SET_ITEM(__pyx_t_1, 0, ((PyObject *)__pyx_n_s__float32));
  __Pyx_GIVEREF(((PyObject *)__pyx_n_s__float32));
  __pyx_t_2 = __Pyx_Import(((PyObject *)__pyx_n_s__numpy), ((PyObject *)__pyx_t_1), -1); if (unlikely(!__pyx_t_2)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 38; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(((PyObject *)__pyx_t_1)); __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_ImportFrom(__pyx_t_2, __pyx_n_s__float32); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 38; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s__DTYPE, __pyx_t_1) < 0) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 38; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "sklearn/tree/_tree.pyx":39
 * 
 * from numpy import float32 as DTYPE
 * from numpy import float64 as DOUBLE             # <<<<<<<<<<<<<<
 * 
 * cdef double INFINITY = np.inf
 */
  __pyx_t_2 = PyList_New(1); if (unlikely(!__pyx_t_2)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 39; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_INCREF(((PyObject *)__pyx_n_s__float64));
  PyList_SET_ITEM(__pyx_t_2, 0, ((PyObject *)__pyx_n_s__float64));
  __Pyx_GIVEREF(((PyObject *)__pyx_n_s__float64));
  __pyx_t_1 = __Pyx_Import(((PyObject *)__pyx_n_s__numpy), ((PyObject *)__pyx_t_2), -1); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 39; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(((PyObject *)__pyx_t_2)); __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_ImportFrom(__pyx_t_1, __pyx_n_s__float64); if (unlikely(!__pyx_t_2)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 39; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s__DOUBLE, __pyx_t_2) < 0) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 39; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "sklearn/tree/_tree.pyx":41
 * from numpy import float64 as DOUBLE
 * 
 * cdef double INFINITY = np.inf             # <<<<<<<<<<<<<<
 * TREE_LEAF = -1
 * TREE_UNDEFINED = -2
 */
  __pyx_t_1 = __Pyx_GetModuleGlobalName(__pyx_n_s__np); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 41; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s__inf); if (unlikely(!__pyx_t_2)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 41; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_3 = __pyx_PyFloat_AsDouble(__pyx_t_2); if (unlikely((__pyx_t_3 == (double)-1) && PyErr_Occurred())) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 41; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_v_7sklearn_4tree_5_tree_INFINITY = __pyx_t_3;

  /* "sklearn/tree/_tree.pyx":42
 * 
 * cdef double INFINITY = np.inf
 * TREE_LEAF = -1             # <<<<<<<<<<<<<<
 * TREE_UNDEFINED = -2
 * cdef SIZE_t _TREE_LEAF = TREE_LEAF
 */
  if (PyDict_SetItem(__pyx_d, __pyx_n_s__TREE_LEAF, __pyx_int_neg_1) < 0) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 42; __pyx_clineno = __LINE__; goto __pyx_L1_error;}

  /* "sklearn/tree/_tree.pyx":43
 * cdef double INFINITY = np.inf
 * TREE_LEAF = -1
 * TREE_UNDEFINED = -2             # <<<<<<<<<<<<<<
 * cdef SIZE_t _TREE_LEAF = TREE_LEAF
 * cdef SIZE_t _TREE_UNDEFINED = TREE_UNDEFINED
 */
  if (PyDict_SetItem(__pyx_d, __pyx_n_s__TREE_UNDEFINED, __pyx_int_neg_2) < 0) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 43; __pyx_clineno = __LINE__; goto __pyx_L1_error;}

  /* "sklearn/tree/_tree.pyx":44
 * TREE_LEAF = -1
 * TREE_UNDEFINED = -2
 * cdef SIZE_t _TREE_LEAF = TREE_LEAF             # <<<<<<<<<<<<<<
 * cdef SIZE_t _TREE_UNDEFINED = TREE_UNDEFINED
 * 
 */
  __pyx_t_2 = __Pyx_GetModuleGlobalName(__pyx_n_s__TREE_LEAF); if (unlikely(!__pyx_t_2)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 44; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_4 = __Pyx_PyInt_from_py_Py_intptr_t(__pyx_t_2); if (unlikely((__pyx_t_4 == (npy_intp)-1) && PyErr_Occurred())) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 44; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_v_7sklearn_4tree_5_tree__TREE_LEAF = __pyx_t_4;

  /* "sklearn/tree/_tree.pyx":45
 * TREE_UNDEFINED = -2
 * cdef SIZE_t _TREE_LEAF = TREE_LEAF
 * cdef SIZE_t _TREE_UNDEFINED = TREE_UNDEFINED             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_t_2 = __Pyx_GetModuleGlobalName(__pyx_n_s__TREE_UNDEFINED); if (unlikely(!__pyx_t_2)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 45; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_4 = __Pyx_PyInt_from_py_Py_intptr_t(__pyx_t_2); if (unlikely((__pyx_t_4 == (npy_intp)-1) && PyErr_Occurred())) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 45; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_v_7sklearn_4tree_5_tree__TREE_UNDEFINED = __pyx_t_4;

  /* "sklearn/tree/_tree.pyx":1682
 *         return out
 * 
 *     cpdef compute_feature_importances(self, normalize=True):             # <<<<<<<<<<<<<<
 *         """Computes the importance of each feature (aka variable)."""
 *         cdef SIZE_t* children_left = self.children_left
 */
  __pyx_t_2 = __Pyx_PyBool_FromLong(1); if (unlikely(!__pyx_t_2)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1682; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_k_1 = __pyx_t_2;
  __Pyx_GIVEREF(__pyx_t_2);
  __pyx_t_2 = 0;

  /* "sklearn/tree/_tree.pyx":1
 * # encoding: utf-8             # <<<<<<<<<<<<<<
 * # cython: cdivision=True
 * # cython: boundscheck=False
 */
  __pyx_t_2 = PyDict_New(); if (unlikely(!__pyx_t_2)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(((PyObject *)__pyx_t_2));
  if (PyDict_SetItem(__pyx_d, __pyx_n_s____test__, ((PyObject *)__pyx_t_2)) < 0) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_DECREF(((PyObject *)__pyx_t_2)); __pyx_t_2 = 0;

  /* "numpy.pxd":975
 *      arr.base = baseptr
 * 
 * cdef inline object get_array_base(ndarray arr):             # <<<<<<<<<<<<<<
 *     if arr.base is NULL:
 *         return None
 */
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  if (__pyx_m) {
    __Pyx_AddTraceback("init sklearn.tree._tree", __pyx_clineno, __pyx_lineno, __pyx_filename);
    Py_DECREF(__pyx_m); __pyx_m = 0;
  } else if (!PyErr_Occurred()) {
    PyErr_SetString(PyExc_ImportError, "init sklearn.tree._tree");
  }
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  #if PY_MAJOR_VERSION < 3
  return;
  #else
  return __pyx_m;
  #endif
}

/* Runtime support code */
#if CYTHON_REFNANNY
static __Pyx_RefNannyAPIStruct *__Pyx_RefNannyImportAPI(const char *modname) {
    PyObject *m = NULL, *p = NULL;
    void *r = NULL;
    m = PyImport_ImportModule((char *)modname);
    if (!m) goto end;
    p = PyObject_GetAttrString(m, (char *)"RefNannyAPI");
    if (!p) goto end;
    r = PyLong_AsVoidPtr(p);
end:
    Py_XDECREF(p);
    Py_XDECREF(m);
    return (__Pyx_RefNannyAPIStruct *)r;
}
#endif /* CYTHON_REFNANNY */

static PyObject *__Pyx_GetBuiltinName(PyObject *name) {
    PyObject* result = __Pyx_PyObject_GetAttrStr(__pyx_b, name);
    if (unlikely(!result)) {
        PyErr_Format(PyExc_NameError,
#if PY_MAJOR_VERSION >= 3
            "name '%U' is not defined", name);
#else
            "name '%s' is not defined", PyString_AS_STRING(name));
#endif
    }
    return result;
}

static void __Pyx_RaiseArgtupleInvalid(
    const char* func_name,
    int exact,
    Py_ssize_t num_min,
    Py_ssize_t num_max,
    Py_ssize_t num_found)
{
    Py_ssize_t num_expected;
    const char *more_or_less;
    if (num_found < num_min) {
        num_expected = num_min;
        more_or_less = "at least";
    } else {
        num_expected = num_max;
        more_or_less = "at most";
    }
    if (exact) {
        more_or_less = "exactly";
    }
    PyErr_Format(PyExc_TypeError,
                 "%s() takes %s %" CYTHON_FORMAT_SSIZE_T "d positional argument%s (%" CYTHON_FORMAT_SSIZE_T "d given)",
                 func_name, more_or_less, num_expected,
                 (num_expected == 1) ? "" : "s", num_found);
}

static void __Pyx_RaiseDoubleKeywordsError(
    const char* func_name,
    PyObject* kw_name)
{
    PyErr_Format(PyExc_TypeError,
        #if PY_MAJOR_VERSION >= 3
        "%s() got multiple values for keyword argument '%U'", func_name, kw_name);
        #else
        "%s() got multiple values for keyword argument '%s'", func_name,
        PyString_AsString(kw_name));
        #endif
}

static int __Pyx_ParseOptionalKeywords(
    PyObject *kwds,
    PyObject **argnames[],
    PyObject *kwds2,
    PyObject *values[],
    Py_ssize_t num_pos_args,
    const char* function_name)
{
    PyObject *key = 0, *value = 0;
    Py_ssize_t pos = 0;
    PyObject*** name;
    PyObject*** first_kw_arg = argnames + num_pos_args;
    while (PyDict_Next(kwds, &pos, &key, &value)) {
        name = first_kw_arg;
        while (*name && (**name != key)) name++;
        if (*name) {
            values[name-argnames] = value;
            continue;
        }
        name = first_kw_arg;
        #if PY_MAJOR_VERSION < 3
        if (likely(PyString_CheckExact(key)) || likely(PyString_Check(key))) {
            while (*name) {
                if ((CYTHON_COMPILING_IN_PYPY || PyString_GET_SIZE(**name) == PyString_GET_SIZE(key))
                        && _PyString_Eq(**name, key)) {
                    values[name-argnames] = value;
                    break;
                }
                name++;
            }
            if (*name) continue;
            else {
                PyObject*** argname = argnames;
                while (argname != first_kw_arg) {
                    if ((**argname == key) || (
                            (CYTHON_COMPILING_IN_PYPY || PyString_GET_SIZE(**argname) == PyString_GET_SIZE(key))
                             && _PyString_Eq(**argname, key))) {
                        goto arg_passed_twice;
                    }
                    argname++;
                }
            }
        } else
        #endif
        if (likely(PyUnicode_Check(key))) {
            while (*name) {
                int cmp = (**name == key) ? 0 :
                #if !CYTHON_COMPILING_IN_PYPY && PY_MAJOR_VERSION >= 3
                    (PyUnicode_GET_SIZE(**name) != PyUnicode_GET_SIZE(key)) ? 1 :
                #endif
                    PyUnicode_Compare(**name, key);
                if (cmp < 0 && unlikely(PyErr_Occurred())) goto bad;
                if (cmp == 0) {
                    values[name-argnames] = value;
                    break;
                }
                name++;
            }
            if (*name) continue;
            else {
                PyObject*** argname = argnames;
                while (argname != first_kw_arg) {
                    int cmp = (**argname == key) ? 0 :
                    #if !CYTHON_COMPILING_IN_PYPY && PY_MAJOR_VERSION >= 3
                        (PyUnicode_GET_SIZE(**argname) != PyUnicode_GET_SIZE(key)) ? 1 :
                    #endif
                        PyUnicode_Compare(**argname, key);
                    if (cmp < 0 && unlikely(PyErr_Occurred())) goto bad;
                    if (cmp == 0) goto arg_passed_twice;
                    argname++;
                }
            }
        } else
            goto invalid_keyword_type;
        if (kwds2) {
            if (unlikely(PyDict_SetItem(kwds2, key, value))) goto bad;
        } else {
            goto invalid_keyword;
        }
    }
    return 0;
arg_passed_twice:
    __Pyx_RaiseDoubleKeywordsError(function_name, key);
    goto bad;
invalid_keyword_type:
    PyErr_Format(PyExc_TypeError,
        "%s() keywords must be strings", function_name);
    goto bad;
invalid_keyword:
    PyErr_Format(PyExc_TypeError,
    #if PY_MAJOR_VERSION < 3
        "%s() got an unexpected keyword argument '%s'",
        function_name, PyString_AsString(key));
    #else
        "%s() got an unexpected keyword argument '%U'",
        function_name, key);
    #endif
bad:
    return -1;
}

static int __Pyx_ArgTypeTest(PyObject *obj, PyTypeObject *type, int none_allowed,
    const char *name, int exact)
{
    if (!type) {
        PyErr_Format(PyExc_SystemError, "Missing type object");
        return 0;
    }
    if (none_allowed && obj == Py_None) return 1;
    else if (exact) {
        if (Py_TYPE(obj) == type) return 1;
    }
    else {
        if (PyObject_TypeCheck(obj, type)) return 1;
    }
    PyErr_Format(PyExc_TypeError,
        "Argument '%s' has incorrect type (expected %s, got %s)",
        name, type->tp_name, Py_TYPE(obj)->tp_name);
    return 0;
}

static CYTHON_INLINE int __Pyx_IsLittleEndian(void) {
  unsigned int n = 1;
  return *(unsigned char*)(&n) != 0;
}
static void __Pyx_BufFmt_Init(__Pyx_BufFmt_Context* ctx,
                              __Pyx_BufFmt_StackElem* stack,
                              __Pyx_TypeInfo* type) {
  stack[0].field = &ctx->root;
  stack[0].parent_offset = 0;
  ctx->root.type = type;
  ctx->root.name = "buffer dtype";
  ctx->root.offset = 0;
  ctx->head = stack;
  ctx->head->field = &ctx->root;
  ctx->fmt_offset = 0;
  ctx->head->parent_offset = 0;
  ctx->new_packmode = '@';
  ctx->enc_packmode = '@';
  ctx->new_count = 1;
  ctx->enc_count = 0;
  ctx->enc_type = 0;
  ctx->is_complex = 0;
  ctx->is_valid_array = 0;
  ctx->struct_alignment = 0;
  while (type->typegroup == 'S') {
    ++ctx->head;
    ctx->head->field = type->fields;
    ctx->head->parent_offset = 0;
    type = type->fields->type;
  }
}
static int __Pyx_BufFmt_ParseNumber(const char** ts) {
    int count;
    const char* t = *ts;
    if (*t < '0' || *t > '9') {
      return -1;
    } else {
        count = *t++ - '0';
        while (*t >= '0' && *t < '9') {
            count *= 10;
            count += *t++ - '0';
        }
    }
    *ts = t;
    return count;
}
static int __Pyx_BufFmt_ExpectNumber(const char **ts) {
    int number = __Pyx_BufFmt_ParseNumber(ts);
    if (number == -1) /* First char was not a digit */
        PyErr_Format(PyExc_ValueError,\
                     "Does not understand character buffer dtype format string ('%c')", **ts);
    return number;
}
static void __Pyx_BufFmt_RaiseUnexpectedChar(char ch) {
  PyErr_Format(PyExc_ValueError,
               "Unexpected format string character: '%c'", ch);
}
static const char* __Pyx_BufFmt_DescribeTypeChar(char ch, int is_complex) {
  switch (ch) {
    case 'c': return "'char'";
    case 'b': return "'signed char'";
    case 'B': return "'unsigned char'";
    case 'h': return "'short'";
    case 'H': return "'unsigned short'";
    case 'i': return "'int'";
    case 'I': return "'unsigned int'";
    case 'l': return "'long'";
    case 'L': return "'unsigned long'";
    case 'q': return "'long long'";
    case 'Q': return "'unsigned long long'";
    case 'f': return (is_complex ? "'complex float'" : "'float'");
    case 'd': return (is_complex ? "'complex double'" : "'double'");
    case 'g': return (is_complex ? "'complex long double'" : "'long double'");
    case 'T': return "a struct";
    case 'O': return "Python object";
    case 'P': return "a pointer";
    case 's': case 'p': return "a string";
    case 0: return "end";
    default: return "unparseable format string";
  }
}
static size_t __Pyx_BufFmt_TypeCharToStandardSize(char ch, int is_complex) {
  switch (ch) {
    case '?': case 'c': case 'b': case 'B': case 's': case 'p': return 1;
    case 'h': case 'H': return 2;
    case 'i': case 'I': case 'l': case 'L': return 4;
    case 'q': case 'Q': return 8;
    case 'f': return (is_complex ? 8 : 4);
    case 'd': return (is_complex ? 16 : 8);
    case 'g': {
      PyErr_SetString(PyExc_ValueError, "Python does not define a standard format string size for long double ('g')..");
      return 0;
    }
    case 'O': case 'P': return sizeof(void*);
    default:
      __Pyx_BufFmt_RaiseUnexpectedChar(ch);
      return 0;
    }
}
static size_t __Pyx_BufFmt_TypeCharToNativeSize(char ch, int is_complex) {
  switch (ch) {
    case 'c': case 'b': case 'B': case 's': case 'p': return 1;
    case 'h': case 'H': return sizeof(short);
    case 'i': case 'I': return sizeof(int);
    case 'l': case 'L': return sizeof(long);
    #ifdef HAVE_LONG_LONG
    case 'q': case 'Q': return sizeof(PY_LONG_LONG);
    #endif
    case 'f': return sizeof(float) * (is_complex ? 2 : 1);
    case 'd': return sizeof(double) * (is_complex ? 2 : 1);
    case 'g': return sizeof(long double) * (is_complex ? 2 : 1);
    case 'O': case 'P': return sizeof(void*);
    default: {
      __Pyx_BufFmt_RaiseUnexpectedChar(ch);
      return 0;
    }
  }
}
typedef struct { char c; short x; } __Pyx_st_short;
typedef struct { char c; int x; } __Pyx_st_int;
typedef struct { char c; long x; } __Pyx_st_long;
typedef struct { char c; float x; } __Pyx_st_float;
typedef struct { char c; double x; } __Pyx_st_double;
typedef struct { char c; long double x; } __Pyx_st_longdouble;
typedef struct { char c; void *x; } __Pyx_st_void_p;
#ifdef HAVE_LONG_LONG
typedef struct { char c; PY_LONG_LONG x; } __Pyx_st_longlong;
#endif
static size_t __Pyx_BufFmt_TypeCharToAlignment(char ch, CYTHON_UNUSED int is_complex) {
  switch (ch) {
    case '?': case 'c': case 'b': case 'B': case 's': case 'p': return 1;
    case 'h': case 'H': return sizeof(__Pyx_st_short) - sizeof(short);
    case 'i': case 'I': return sizeof(__Pyx_st_int) - sizeof(int);
    case 'l': case 'L': return sizeof(__Pyx_st_long) - sizeof(long);
#ifdef HAVE_LONG_LONG
    case 'q': case 'Q': return sizeof(__Pyx_st_longlong) - sizeof(PY_LONG_LONG);
#endif
    case 'f': return sizeof(__Pyx_st_float) - sizeof(float);
    case 'd': return sizeof(__Pyx_st_double) - sizeof(double);
    case 'g': return sizeof(__Pyx_st_longdouble) - sizeof(long double);
    case 'P': case 'O': return sizeof(__Pyx_st_void_p) - sizeof(void*);
    default:
      __Pyx_BufFmt_RaiseUnexpectedChar(ch);
      return 0;
    }
}
/* These are for computing the padding at the end of the struct to align
   on the first member of the struct. This will probably the same as above,
   but we don't have any guarantees.
 */
typedef struct { short x; char c; } __Pyx_pad_short;
typedef struct { int x; char c; } __Pyx_pad_int;
typedef struct { long x; char c; } __Pyx_pad_long;
typedef struct { float x; char c; } __Pyx_pad_float;
typedef struct { double x; char c; } __Pyx_pad_double;
typedef struct { long double x; char c; } __Pyx_pad_longdouble;
typedef struct { void *x; char c; } __Pyx_pad_void_p;
#ifdef HAVE_LONG_LONG
typedef struct { PY_LONG_LONG x; char c; } __Pyx_pad_longlong;
#endif
static size_t __Pyx_BufFmt_TypeCharToPadding(char ch, CYTHON_UNUSED int is_complex) {
  switch (ch) {
    case '?': case 'c': case 'b': case 'B': case 's': case 'p': return 1;
    case 'h': case 'H': return sizeof(__Pyx_pad_short) - sizeof(short);
    case 'i': case 'I': return sizeof(__Pyx_pad_int) - sizeof(int);
    case 'l': case 'L': return sizeof(__Pyx_pad_long) - sizeof(long);
#ifdef HAVE_LONG_LONG
    case 'q': case 'Q': return sizeof(__Pyx_pad_longlong) - sizeof(PY_LONG_LONG);
#endif
    case 'f': return sizeof(__Pyx_pad_float) - sizeof(float);
    case 'd': return sizeof(__Pyx_pad_double) - sizeof(double);
    case 'g': return sizeof(__Pyx_pad_longdouble) - sizeof(long double);
    case 'P': case 'O': return sizeof(__Pyx_pad_void_p) - sizeof(void*);
    default:
      __Pyx_BufFmt_RaiseUnexpectedChar(ch);
      return 0;
    }
}
static char __Pyx_BufFmt_TypeCharToGroup(char ch, int is_complex) {
  switch (ch) {
    case 'c':
        return 'H';
    case 'b': case 'h': case 'i':
    case 'l': case 'q': case 's': case 'p':
        return 'I';
    case 'B': case 'H': case 'I': case 'L': case 'Q':
        return 'U';
    case 'f': case 'd': case 'g':
        return (is_complex ? 'C' : 'R');
    case 'O':
        return 'O';
    case 'P':
        return 'P';
    default: {
      __Pyx_BufFmt_RaiseUnexpectedChar(ch);
      return 0;
    }
  }
}
static void __Pyx_BufFmt_RaiseExpected(__Pyx_BufFmt_Context* ctx) {
  if (ctx->head == NULL || ctx->head->field == &ctx->root) {
    const char* expected;
    const char* quote;
    if (ctx->head == NULL) {
      expected = "end";
      quote = "";
    } else {
      expected = ctx->head->field->type->name;
      quote = "'";
    }
    PyErr_Format(PyExc_ValueError,
                 "Buffer dtype mismatch, expected %s%s%s but got %s",
                 quote, expected, quote,
                 __Pyx_BufFmt_DescribeTypeChar(ctx->enc_type, ctx->is_complex));
  } else {
    __Pyx_StructField* field = ctx->head->field;
    __Pyx_StructField* parent = (ctx->head - 1)->field;
    PyErr_Format(PyExc_ValueError,
                 "Buffer dtype mismatch, expected '%s' but got %s in '%s.%s'",
                 field->type->name, __Pyx_BufFmt_DescribeTypeChar(ctx->enc_type, ctx->is_complex),
                 parent->type->name, field->name);
  }
}
static int __Pyx_BufFmt_ProcessTypeChunk(__Pyx_BufFmt_Context* ctx) {
  char group;
  size_t size, offset, arraysize = 1;
  if (ctx->enc_type == 0) return 0;
  if (ctx->head->field->type->arraysize[0]) {
    int i, ndim = 0;
    if (ctx->enc_type == 's' || ctx->enc_type == 'p') {
        ctx->is_valid_array = ctx->head->field->type->ndim == 1;
        ndim = 1;
        if (ctx->enc_count != ctx->head->field->type->arraysize[0]) {
            PyErr_Format(PyExc_ValueError,
                         "Expected a dimension of size %zu, got %zu",
                         ctx->head->field->type->arraysize[0], ctx->enc_count);
            return -1;
        }
    }
    if (!ctx->is_valid_array) {
      PyErr_Format(PyExc_ValueError, "Expected %d dimensions, got %d",
                   ctx->head->field->type->ndim, ndim);
      return -1;
    }
    for (i = 0; i < ctx->head->field->type->ndim; i++) {
      arraysize *= ctx->head->field->type->arraysize[i];
    }
    ctx->is_valid_array = 0;
    ctx->enc_count = 1;
  }
  group = __Pyx_BufFmt_TypeCharToGroup(ctx->enc_type, ctx->is_complex);
  do {
    __Pyx_StructField* field = ctx->head->field;
    __Pyx_TypeInfo* type = field->type;
    if (ctx->enc_packmode == '@' || ctx->enc_packmode == '^') {
      size = __Pyx_BufFmt_TypeCharToNativeSize(ctx->enc_type, ctx->is_complex);
    } else {
      size = __Pyx_BufFmt_TypeCharToStandardSize(ctx->enc_type, ctx->is_complex);
    }
    if (ctx->enc_packmode == '@') {
      size_t align_at = __Pyx_BufFmt_TypeCharToAlignment(ctx->enc_type, ctx->is_complex);
      size_t align_mod_offset;
      if (align_at == 0) return -1;
      align_mod_offset = ctx->fmt_offset % align_at;
      if (align_mod_offset > 0) ctx->fmt_offset += align_at - align_mod_offset;
      if (ctx->struct_alignment == 0)
          ctx->struct_alignment = __Pyx_BufFmt_TypeCharToPadding(ctx->enc_type,
                                                                 ctx->is_complex);
    }
    if (type->size != size || type->typegroup != group) {
      if (type->typegroup == 'C' && type->fields != NULL) {
        size_t parent_offset = ctx->head->parent_offset + field->offset;
        ++ctx->head;
        ctx->head->field = type->fields;
        ctx->head->parent_offset = parent_offset;
        continue;
      }
      if ((type->typegroup == 'H' || group == 'H') && type->size == size) {
      } else {
          __Pyx_BufFmt_RaiseExpected(ctx);
          return -1;
      }
    }
    offset = ctx->head->parent_offset + field->offset;
    if (ctx->fmt_offset != offset) {
      PyErr_Format(PyExc_ValueError,
                   "Buffer dtype mismatch; next field is at offset %" CYTHON_FORMAT_SSIZE_T "d but %" CYTHON_FORMAT_SSIZE_T "d expected",
                   (Py_ssize_t)ctx->fmt_offset, (Py_ssize_t)offset);
      return -1;
    }
    ctx->fmt_offset += size;
    if (arraysize)
      ctx->fmt_offset += (arraysize - 1) * size;
    --ctx->enc_count; /* Consume from buffer string */
    while (1) {
      if (field == &ctx->root) {
        ctx->head = NULL;
        if (ctx->enc_count != 0) {
          __Pyx_BufFmt_RaiseExpected(ctx);
          return -1;
        }
        break; /* breaks both loops as ctx->enc_count == 0 */
      }
      ctx->head->field = ++field;
      if (field->type == NULL) {
        --ctx->head;
        field = ctx->head->field;
        continue;
      } else if (field->type->typegroup == 'S') {
        size_t parent_offset = ctx->head->parent_offset + field->offset;
        if (field->type->fields->type == NULL) continue; /* empty struct */
        field = field->type->fields;
        ++ctx->head;
        ctx->head->field = field;
        ctx->head->parent_offset = parent_offset;
        break;
      } else {
        break;
      }
    }
  } while (ctx->enc_count);
  ctx->enc_type = 0;
  ctx->is_complex = 0;
  return 0;
}
static CYTHON_INLINE PyObject *
__pyx_buffmt_parse_array(__Pyx_BufFmt_Context* ctx, const char** tsp)
{
    const char *ts = *tsp;
    int i = 0, number;
    int ndim = ctx->head->field->type->ndim;
;
    ++ts;
    if (ctx->new_count != 1) {
        PyErr_SetString(PyExc_ValueError,
                        "Cannot handle repeated arrays in format string");
        return NULL;
    }
    if (__Pyx_BufFmt_ProcessTypeChunk(ctx) == -1) return NULL;
    while (*ts && *ts != ')') {
        if (isspace(*ts))
            continue;
        number = __Pyx_BufFmt_ExpectNumber(&ts);
        if (number == -1) return NULL;
        if (i < ndim && (size_t) number != ctx->head->field->type->arraysize[i])
            return PyErr_Format(PyExc_ValueError,
                        "Expected a dimension of size %zu, got %d",
                        ctx->head->field->type->arraysize[i], number);
        if (*ts != ',' && *ts != ')')
            return PyErr_Format(PyExc_ValueError,
                                "Expected a comma in format string, got '%c'", *ts);
        if (*ts == ',') ts++;
        i++;
    }
    if (i != ndim)
        return PyErr_Format(PyExc_ValueError, "Expected %d dimension(s), got %d",
                            ctx->head->field->type->ndim, i);
    if (!*ts) {
        PyErr_SetString(PyExc_ValueError,
                        "Unexpected end of format string, expected ')'");
        return NULL;
    }
    ctx->is_valid_array = 1;
    ctx->new_count = 1;
    *tsp = ++ts;
    return Py_None;
}
static const char* __Pyx_BufFmt_CheckString(__Pyx_BufFmt_Context* ctx, const char* ts) {
  int got_Z = 0;
  while (1) {
    switch(*ts) {
      case 0:
        if (ctx->enc_type != 0 && ctx->head == NULL) {
          __Pyx_BufFmt_RaiseExpected(ctx);
          return NULL;
        }
        if (__Pyx_BufFmt_ProcessTypeChunk(ctx) == -1) return NULL;
        if (ctx->head != NULL) {
          __Pyx_BufFmt_RaiseExpected(ctx);
          return NULL;
        }
                return ts;
      case ' ':
      case 10:
      case 13:
        ++ts;
        break;
      case '<':
        if (!__Pyx_IsLittleEndian()) {
          PyErr_SetString(PyExc_ValueError, "Little-endian buffer not supported on big-endian compiler");
          return NULL;
        }
        ctx->new_packmode = '=';
        ++ts;
        break;
      case '>':
      case '!':
        if (__Pyx_IsLittleEndian()) {
          PyErr_SetString(PyExc_ValueError, "Big-endian buffer not supported on little-endian compiler");
          return NULL;
        }
        ctx->new_packmode = '=';
        ++ts;
        break;
      case '=':
      case '@':
      case '^':
        ctx->new_packmode = *ts++;
        break;
      case 'T': /* substruct */
        {
          const char* ts_after_sub;
          size_t i, struct_count = ctx->new_count;
          size_t struct_alignment = ctx->struct_alignment;
          ctx->new_count = 1;
          ++ts;
          if (*ts != '{') {
            PyErr_SetString(PyExc_ValueError, "Buffer acquisition: Expected '{' after 'T'");
            return NULL;
          }
          if (__Pyx_BufFmt_ProcessTypeChunk(ctx) == -1) return NULL;
          ctx->enc_type = 0; /* Erase processed last struct element */
          ctx->enc_count = 0;
          ctx->struct_alignment = 0;
          ++ts;
          ts_after_sub = ts;
          for (i = 0; i != struct_count; ++i) {
            ts_after_sub = __Pyx_BufFmt_CheckString(ctx, ts);
            if (!ts_after_sub) return NULL;
          }
          ts = ts_after_sub;
          if (struct_alignment) ctx->struct_alignment = struct_alignment;
        }
        break;
      case '}': /* end of substruct; either repeat or move on */
        {
          size_t alignment = ctx->struct_alignment;
          ++ts;
          if (__Pyx_BufFmt_ProcessTypeChunk(ctx) == -1) return NULL;
          ctx->enc_type = 0; /* Erase processed last struct element */
          if (alignment && ctx->fmt_offset % alignment) {
            ctx->fmt_offset += alignment - (ctx->fmt_offset % alignment);
          }
        }
        return ts;
      case 'x':
        if (__Pyx_BufFmt_ProcessTypeChunk(ctx) == -1) return NULL;
        ctx->fmt_offset += ctx->new_count;
        ctx->new_count = 1;
        ctx->enc_count = 0;
        ctx->enc_type = 0;
        ctx->enc_packmode = ctx->new_packmode;
        ++ts;
        break;
      case 'Z':
        got_Z = 1;
        ++ts;
        if (*ts != 'f' && *ts != 'd' && *ts != 'g') {
          __Pyx_BufFmt_RaiseUnexpectedChar('Z');
          return NULL;
        }        /* fall through */
      case 'c': case 'b': case 'B': case 'h': case 'H': case 'i': case 'I':
      case 'l': case 'L': case 'q': case 'Q':
      case 'f': case 'd': case 'g':
      case 'O': case 's': case 'p':
        if (ctx->enc_type == *ts && got_Z == ctx->is_complex &&
            ctx->enc_packmode == ctx->new_packmode) {
          ctx->enc_count += ctx->new_count;
        } else {
          if (__Pyx_BufFmt_ProcessTypeChunk(ctx) == -1) return NULL;
          ctx->enc_count = ctx->new_count;
          ctx->enc_packmode = ctx->new_packmode;
          ctx->enc_type = *ts;
          ctx->is_complex = got_Z;
        }
        ++ts;
        ctx->new_count = 1;
        got_Z = 0;
        break;
      case ':':
        ++ts;
        while(*ts != ':') ++ts;
        ++ts;
        break;
      case '(':
        if (!__pyx_buffmt_parse_array(ctx, &ts)) return NULL;
        break;
      default:
        {
          int number = __Pyx_BufFmt_ExpectNumber(&ts);
          if (number == -1) return NULL;
          ctx->new_count = (size_t)number;
        }
    }
  }
}
static CYTHON_INLINE void __Pyx_ZeroBuffer(Py_buffer* buf) {
  buf->buf = NULL;
  buf->obj = NULL;
  buf->strides = __Pyx_zeros;
  buf->shape = __Pyx_zeros;
  buf->suboffsets = __Pyx_minusones;
}
static CYTHON_INLINE int __Pyx_GetBufferAndValidate(
        Py_buffer* buf, PyObject* obj,  __Pyx_TypeInfo* dtype, int flags,
        int nd, int cast, __Pyx_BufFmt_StackElem* stack)
{
  if (obj == Py_None || obj == NULL) {
    __Pyx_ZeroBuffer(buf);
    return 0;
  }
  buf->buf = NULL;
  if (__Pyx_GetBuffer(obj, buf, flags) == -1) goto fail;
  if (buf->ndim != nd) {
    PyErr_Format(PyExc_ValueError,
                 "Buffer has wrong number of dimensions (expected %d, got %d)",
                 nd, buf->ndim);
    goto fail;
  }
  if (!cast) {
    __Pyx_BufFmt_Context ctx;
    __Pyx_BufFmt_Init(&ctx, stack, dtype);
    if (!__Pyx_BufFmt_CheckString(&ctx, buf->format)) goto fail;
  }
  if ((unsigned)buf->itemsize != dtype->size) {
    PyErr_Format(PyExc_ValueError,
      "Item size of buffer (%" CYTHON_FORMAT_SSIZE_T "d byte%s) does not match size of '%s' (%" CYTHON_FORMAT_SSIZE_T "d byte%s)",
      buf->itemsize, (buf->itemsize > 1) ? "s" : "",
      dtype->name, (Py_ssize_t)dtype->size, (dtype->size > 1) ? "s" : "");
    goto fail;
  }
  if (buf->suboffsets == NULL) buf->suboffsets = __Pyx_minusones;
  return 0;
fail:;
  __Pyx_ZeroBuffer(buf);
  return -1;
}
static CYTHON_INLINE void __Pyx_SafeReleaseBuffer(Py_buffer* info) {
  if (info->buf == NULL) return;
  if (info->suboffsets == __Pyx_minusones) info->suboffsets = NULL;
  __Pyx_ReleaseBuffer(info);
}

static CYTHON_INLINE void __Pyx_ErrRestore(PyObject *type, PyObject *value, PyObject *tb) {
#if CYTHON_COMPILING_IN_CPYTHON
    PyObject *tmp_type, *tmp_value, *tmp_tb;
    PyThreadState *tstate = PyThreadState_GET();
    tmp_type = tstate->curexc_type;
    tmp_value = tstate->curexc_value;
    tmp_tb = tstate->curexc_traceback;
    tstate->curexc_type = type;
    tstate->curexc_value = value;
    tstate->curexc_traceback = tb;
    Py_XDECREF(tmp_type);
    Py_XDECREF(tmp_value);
    Py_XDECREF(tmp_tb);
#else
    PyErr_Restore(type, value, tb);
#endif
}
static CYTHON_INLINE void __Pyx_ErrFetch(PyObject **type, PyObject **value, PyObject **tb) {
#if CYTHON_COMPILING_IN_CPYTHON
    PyThreadState *tstate = PyThreadState_GET();
    *type = tstate->curexc_type;
    *value = tstate->curexc_value;
    *tb = tstate->curexc_traceback;
    tstate->curexc_type = 0;
    tstate->curexc_value = 0;
    tstate->curexc_traceback = 0;
#else
    PyErr_Fetch(type, value, tb);
#endif
}

static void __Pyx_WriteUnraisable(const char *name, CYTHON_UNUSED int clineno,
                                  CYTHON_UNUSED int lineno, CYTHON_UNUSED const char *filename) {
    PyObject *old_exc, *old_val, *old_tb;
    PyObject *ctx;
    __Pyx_ErrFetch(&old_exc, &old_val, &old_tb);
    #if PY_MAJOR_VERSION < 3
    ctx = PyString_FromString(name);
    #else
    ctx = PyUnicode_FromString(name);
    #endif
    __Pyx_ErrRestore(old_exc, old_val, old_tb);
    if (!ctx) {
        PyErr_WriteUnraisable(Py_None);
    } else {
        PyErr_WriteUnraisable(ctx);
        Py_DECREF(ctx);
    }
}

static CYTHON_INLINE int __Pyx_TypeTest(PyObject *obj, PyTypeObject *type) {
    if (unlikely(!type)) {
        PyErr_Format(PyExc_SystemError, "Missing type object");
        return 0;
    }
    if (likely(PyObject_TypeCheck(obj, type)))
        return 1;
    PyErr_Format(PyExc_TypeError, "Cannot convert %.200s to %.200s",
                 Py_TYPE(obj)->tp_name, type->tp_name);
    return 0;
}

static CYTHON_INLINE PyObject *__Pyx_GetModuleGlobalName(PyObject *name) {
    PyObject *result;
#if CYTHON_COMPILING_IN_CPYTHON
    result = PyDict_GetItem(__pyx_d, name);
    if (result) {
        Py_INCREF(result);
    } else {
#else
    result = PyObject_GetItem(__pyx_d, name);
    if (!result) {
        PyErr_Clear();
#endif
        result = __Pyx_GetBuiltinName(name);
    }
    return result;
}

static void __Pyx_RaiseBufferFallbackError(void) {
  PyErr_Format(PyExc_ValueError,
     "Buffer acquisition failed on assignment; and then reacquiring the old buffer failed too!");
}

#if PY_MAJOR_VERSION < 3
static void __Pyx_Raise(PyObject *type, PyObject *value, PyObject *tb,
                        CYTHON_UNUSED PyObject *cause) {
    Py_XINCREF(type);
    if (!value || value == Py_None)
        value = NULL;
    else
        Py_INCREF(value);
    if (!tb || tb == Py_None)
        tb = NULL;
    else {
        Py_INCREF(tb);
        if (!PyTraceBack_Check(tb)) {
            PyErr_SetString(PyExc_TypeError,
                "raise: arg 3 must be a traceback or None");
            goto raise_error;
        }
    }
    #if PY_VERSION_HEX < 0x02050000
    if (PyClass_Check(type)) {
    #else
    if (PyType_Check(type)) {
    #endif
#if CYTHON_COMPILING_IN_PYPY
        if (!value) {
            Py_INCREF(Py_None);
            value = Py_None;
        }
#endif
        PyErr_NormalizeException(&type, &value, &tb);
    } else {
        if (value) {
            PyErr_SetString(PyExc_TypeError,
                "instance exception may not have a separate value");
            goto raise_error;
        }
        value = type;
        #if PY_VERSION_HEX < 0x02050000
        if (PyInstance_Check(type)) {
            type = (PyObject*) ((PyInstanceObject*)type)->in_class;
            Py_INCREF(type);
        } else {
            type = 0;
            PyErr_SetString(PyExc_TypeError,
                "raise: exception must be an old-style class or instance");
            goto raise_error;
        }
        #else
        type = (PyObject*) Py_TYPE(type);
        Py_INCREF(type);
        if (!PyType_IsSubtype((PyTypeObject *)type, (PyTypeObject *)PyExc_BaseException)) {
            PyErr_SetString(PyExc_TypeError,
                "raise: exception class must be a subclass of BaseException");
            goto raise_error;
        }
        #endif
    }
    __Pyx_ErrRestore(type, value, tb);
    return;
raise_error:
    Py_XDECREF(value);
    Py_XDECREF(type);
    Py_XDECREF(tb);
    return;
}
#else /* Python 3+ */
static void __Pyx_Raise(PyObject *type, PyObject *value, PyObject *tb, PyObject *cause) {
    PyObject* owned_instance = NULL;
    if (tb == Py_None) {
        tb = 0;
    } else if (tb && !PyTraceBack_Check(tb)) {
        PyErr_SetString(PyExc_TypeError,
            "raise: arg 3 must be a traceback or None");
        goto bad;
    }
    if (value == Py_None)
        value = 0;
    if (PyExceptionInstance_Check(type)) {
        if (value) {
            PyErr_SetString(PyExc_TypeError,
                "instance exception may not have a separate value");
            goto bad;
        }
        value = type;
        type = (PyObject*) Py_TYPE(value);
    } else if (PyExceptionClass_Check(type)) {
        PyObject *args;
        if (!value)
            args = PyTuple_New(0);
        else if (PyTuple_Check(value)) {
            Py_INCREF(value);
            args = value;
        } else
            args = PyTuple_Pack(1, value);
        if (!args)
            goto bad;
        owned_instance = PyEval_CallObject(type, args);
        Py_DECREF(args);
        if (!owned_instance)
            goto bad;
        value = owned_instance;
        if (!PyExceptionInstance_Check(value)) {
            PyErr_Format(PyExc_TypeError,
                         "calling %R should have returned an instance of "
                         "BaseException, not %R",
                         type, Py_TYPE(value));
            goto bad;
        }
    } else {
        PyErr_SetString(PyExc_TypeError,
            "raise: exception class must be a subclass of BaseException");
        goto bad;
    }
#if PY_VERSION_HEX >= 0x03030000
    if (cause) {
#else
    if (cause && cause != Py_None) {
#endif
        PyObject *fixed_cause;
        if (cause == Py_None) {
            fixed_cause = NULL;
        } else if (PyExceptionClass_Check(cause)) {
            fixed_cause = PyObject_CallObject(cause, NULL);
            if (fixed_cause == NULL)
                goto bad;
        } else if (PyExceptionInstance_Check(cause)) {
            fixed_cause = cause;
            Py_INCREF(fixed_cause);
        } else {
            PyErr_SetString(PyExc_TypeError,
                            "exception causes must derive from "
                            "BaseException");
            goto bad;
        }
        PyException_SetCause(value, fixed_cause);
    }
    PyErr_SetObject(type, value);
    if (tb) {
        PyThreadState *tstate = PyThreadState_GET();
        PyObject* tmp_tb = tstate->curexc_traceback;
        if (tb != tmp_tb) {
            Py_INCREF(tb);
            tstate->curexc_traceback = tb;
            Py_XDECREF(tmp_tb);
        }
    }
bad:
    Py_XDECREF(owned_instance);
    return;
}
#endif

static CYTHON_INLINE void __Pyx_RaiseTooManyValuesError(Py_ssize_t expected) {
    PyErr_Format(PyExc_ValueError,
                 "too many values to unpack (expected %" CYTHON_FORMAT_SSIZE_T "d)", expected);
}

static CYTHON_INLINE void __Pyx_RaiseNeedMoreValuesError(Py_ssize_t index) {
    PyErr_Format(PyExc_ValueError,
                 "need more than %" CYTHON_FORMAT_SSIZE_T "d value%s to unpack",
                 index, (index == 1) ? "" : "s");
}

static CYTHON_INLINE void __Pyx_RaiseNoneNotIterableError(void) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not iterable");
}

static CYTHON_INLINE int __Pyx_IterFinish(void) {
#if CYTHON_COMPILING_IN_CPYTHON
    PyThreadState *tstate = PyThreadState_GET();
    PyObject* exc_type = tstate->curexc_type;
    if (unlikely(exc_type)) {
        if (likely(exc_type == PyExc_StopIteration) || PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration)) {
            PyObject *exc_value, *exc_tb;
            exc_value = tstate->curexc_value;
            exc_tb = tstate->curexc_traceback;
            tstate->curexc_type = 0;
            tstate->curexc_value = 0;
            tstate->curexc_traceback = 0;
            Py_DECREF(exc_type);
            Py_XDECREF(exc_value);
            Py_XDECREF(exc_tb);
            return 0;
        } else {
            return -1;
        }
    }
    return 0;
#else
    if (unlikely(PyErr_Occurred())) {
        if (likely(PyErr_ExceptionMatches(PyExc_StopIteration))) {
            PyErr_Clear();
            return 0;
        } else {
            return -1;
        }
    }
    return 0;
#endif
}

static int __Pyx_IternextUnpackEndCheck(PyObject *retval, Py_ssize_t expected) {
    if (unlikely(retval)) {
        Py_DECREF(retval);
        __Pyx_RaiseTooManyValuesError(expected);
        return -1;
    } else {
        return __Pyx_IterFinish();
    }
    return 0;
}

static int __Pyx_SetVtable(PyObject *dict, void *vtable) {
#if PY_VERSION_HEX >= 0x02070000 && !(PY_MAJOR_VERSION==3&&PY_MINOR_VERSION==0)
    PyObject *ob = PyCapsule_New(vtable, 0, 0);
#else
    PyObject *ob = PyCObject_FromVoidPtr(vtable, 0);
#endif
    if (!ob)
        goto bad;
    if (PyDict_SetItem(dict, __pyx_n_s____pyx_vtable__, ob) < 0)
        goto bad;
    Py_DECREF(ob);
    return 0;
bad:
    Py_XDECREF(ob);
    return -1;
}

static PyObject* __Pyx_ImportFrom(PyObject* module, PyObject* name) {
    PyObject* value = __Pyx_PyObject_GetAttrStr(module, name);
    if (unlikely(!value) && PyErr_ExceptionMatches(PyExc_AttributeError)) {
        PyErr_Format(PyExc_ImportError,
        #if PY_MAJOR_VERSION < 3
            "cannot import name %.230s", PyString_AS_STRING(name));
        #else
            "cannot import name %S", name);
        #endif
    }
    return value;
}

static CYTHON_INLINE PyObject *__Pyx_PyInt_to_py_Py_intptr_t(Py_intptr_t val) {
    const Py_intptr_t neg_one = (Py_intptr_t)-1, const_zero = (Py_intptr_t)0;
    const int is_unsigned = const_zero < neg_one;
    if ((sizeof(Py_intptr_t) == sizeof(char))  ||
        (sizeof(Py_intptr_t) == sizeof(short))) {
        return PyInt_FromLong((long)val);
    } else if ((sizeof(Py_intptr_t) == sizeof(int)) ||
               (sizeof(Py_intptr_t) == sizeof(long))) {
        if (is_unsigned)
            return PyLong_FromUnsignedLong((unsigned long)val);
        else
            return PyInt_FromLong((long)val);
    } else if (sizeof(Py_intptr_t) == sizeof(PY_LONG_LONG)) {
        if (is_unsigned)
            return PyLong_FromUnsignedLongLong((unsigned PY_LONG_LONG)val);
        else
            return PyLong_FromLongLong((PY_LONG_LONG)val);
    } else {
        int one = 1; int little = (int)*(unsigned char *)&one;
        unsigned char *bytes = (unsigned char *)&val;
        return _PyLong_FromByteArray(bytes, sizeof(Py_intptr_t),
                                     little, !is_unsigned);
    }
}

static CYTHON_INLINE Py_intptr_t __Pyx_PyInt_from_py_Py_intptr_t(PyObject* x) {
    const Py_intptr_t neg_one = (Py_intptr_t)-1, const_zero = (Py_intptr_t)0;
    const int is_unsigned = const_zero < neg_one;
    if (sizeof(Py_intptr_t) == sizeof(char)) {
        if (is_unsigned)
            return (Py_intptr_t)__Pyx_PyInt_AsUnsignedChar(x);
        else
            return (Py_intptr_t)__Pyx_PyInt_AsSignedChar(x);
    } else if (sizeof(Py_intptr_t) == sizeof(short)) {
        if (is_unsigned)
            return (Py_intptr_t)__Pyx_PyInt_AsUnsignedShort(x);
        else
            return (Py_intptr_t)__Pyx_PyInt_AsSignedShort(x);
    } else if (sizeof(Py_intptr_t) == sizeof(int)) {
        if (is_unsigned)
            return (Py_intptr_t)__Pyx_PyInt_AsUnsignedInt(x);
        else
            return (Py_intptr_t)__Pyx_PyInt_AsSignedInt(x);
    } else if (sizeof(Py_intptr_t) == sizeof(long)) {
        if (is_unsigned)
            return (Py_intptr_t)__Pyx_PyInt_AsUnsignedLong(x);
        else
            return (Py_intptr_t)__Pyx_PyInt_AsSignedLong(x);
    } else if (sizeof(Py_intptr_t) == sizeof(PY_LONG_LONG)) {
        if (is_unsigned)
            return (Py_intptr_t)__Pyx_PyInt_AsUnsignedLongLong(x);
        else
            return (Py_intptr_t)__Pyx_PyInt_AsSignedLongLong(x);
    }  else {
        #if CYTHON_COMPILING_IN_PYPY && !defined(_PyLong_AsByteArray)
        PyErr_SetString(PyExc_RuntimeError,
                        "_PyLong_AsByteArray() not available in PyPy, cannot convert large numbers");
        #else
        Py_intptr_t val;
        PyObject *v = __Pyx_PyNumber_Int(x);
        #if PY_MAJOR_VERSION < 3
        if (likely(v) && !PyLong_Check(v)) {
            PyObject *tmp = v;
            v = PyNumber_Long(tmp);
            Py_DECREF(tmp);
        }
        #endif
        if (likely(v)) {
            int one = 1; int is_little = (int)*(unsigned char *)&one;
            unsigned char *bytes = (unsigned char *)&val;
            int ret = _PyLong_AsByteArray((PyLongObject *)v,
                                          bytes, sizeof(val),
                                          is_little, !is_unsigned);
            Py_DECREF(v);
            if (likely(!ret))
                return val;
        }
        #endif
        return (Py_intptr_t)-1;
    }
}

static PyObject *__Pyx_Import(PyObject *name, PyObject *from_list, int level) {
    PyObject *empty_list = 0;
    PyObject *module = 0;
    PyObject *global_dict = 0;
    PyObject *empty_dict = 0;
    PyObject *list;
    #if PY_VERSION_HEX < 0x03030000
    PyObject *py_import;
    py_import = __Pyx_PyObject_GetAttrStr(__pyx_b, __pyx_n_s____import__);
    if (!py_import)
        goto bad;
    #endif
    if (from_list)
        list = from_list;
    else {
        empty_list = PyList_New(0);
        if (!empty_list)
            goto bad;
        list = empty_list;
    }
    global_dict = PyModule_GetDict(__pyx_m);
    if (!global_dict)
        goto bad;
    empty_dict = PyDict_New();
    if (!empty_dict)
        goto bad;
    #if PY_VERSION_HEX >= 0x02050000
    {
        #if PY_MAJOR_VERSION >= 3
        if (level == -1) {
            if (strchr(__Pyx_MODULE_NAME, '.')) {
                #if PY_VERSION_HEX < 0x03030000
                PyObject *py_level = PyInt_FromLong(1);
                if (!py_level)
                    goto bad;
                module = PyObject_CallFunctionObjArgs(py_import,
                    name, global_dict, empty_dict, list, py_level, NULL);
                Py_DECREF(py_level);
                #else
                module = PyImport_ImportModuleLevelObject(
                    name, global_dict, empty_dict, list, 1);
                #endif
                if (!module) {
                    if (!PyErr_ExceptionMatches(PyExc_ImportError))
                        goto bad;
                    PyErr_Clear();
                }
            }
            level = 0; /* try absolute import on failure */
        }
        #endif
        if (!module) {
            #if PY_VERSION_HEX < 0x03030000
            PyObject *py_level = PyInt_FromLong(level);
            if (!py_level)
                goto bad;
            module = PyObject_CallFunctionObjArgs(py_import,
                name, global_dict, empty_dict, list, py_level, NULL);
            Py_DECREF(py_level);
            #else
            module = PyImport_ImportModuleLevelObject(
                name, global_dict, empty_dict, list, level);
            #endif
        }
    }
    #else
    if (level>0) {
        PyErr_SetString(PyExc_RuntimeError, "Relative import is not supported for Python <=2.4.");
        goto bad;
    }
    module = PyObject_CallFunctionObjArgs(py_import,
        name, global_dict, empty_dict, list, NULL);
    #endif
bad:
    #if PY_VERSION_HEX < 0x03030000
    Py_XDECREF(py_import);
    #endif
    Py_XDECREF(empty_list);
    Py_XDECREF(empty_dict);
    return module;
}

#if PY_MAJOR_VERSION < 3
static int __Pyx_GetBuffer(PyObject *obj, Py_buffer *view, int flags) {
  #if PY_VERSION_HEX >= 0x02060000
    if (PyObject_CheckBuffer(obj)) return PyObject_GetBuffer(obj, view, flags);
  #endif
        if (PyObject_TypeCheck(obj, __pyx_ptype_5numpy_ndarray)) return __pyx_pw_5numpy_7ndarray_1__getbuffer__(obj, view, flags);
  #if PY_VERSION_HEX < 0x02060000
    if (obj->ob_type->tp_dict) {
        PyObject *getbuffer_cobj = PyObject_GetItem(
            obj->ob_type->tp_dict, __pyx_n_s____pyx_getbuffer);
        if (getbuffer_cobj) {
            getbufferproc func = (getbufferproc) PyCObject_AsVoidPtr(getbuffer_cobj);
            Py_DECREF(getbuffer_cobj);
            if (!func)
                goto fail;
            return func(obj, view, flags);
        } else {
            PyErr_Clear();
        }
    }
  #endif
    PyErr_Format(PyExc_TypeError, "'%100s' does not have the buffer interface", Py_TYPE(obj)->tp_name);
#if PY_VERSION_HEX < 0x02060000
fail:
#endif
    return -1;
}
static void __Pyx_ReleaseBuffer(Py_buffer *view) {
    PyObject *obj = view->obj;
    if (!obj) return;
  #if PY_VERSION_HEX >= 0x02060000
    if (PyObject_CheckBuffer(obj)) {
        PyBuffer_Release(view);
        return;
    }
  #endif
        if (PyObject_TypeCheck(obj, __pyx_ptype_5numpy_ndarray)) { __pyx_pw_5numpy_7ndarray_3__releasebuffer__(obj, view); return; }
  #if PY_VERSION_HEX < 0x02060000
    if (obj->ob_type->tp_dict) {
        PyObject *releasebuffer_cobj = PyObject_GetItem(
            obj->ob_type->tp_dict, __pyx_n_s____pyx_releasebuffer);
        if (releasebuffer_cobj) {
            releasebufferproc func = (releasebufferproc) PyCObject_AsVoidPtr(releasebuffer_cobj);
            Py_DECREF(releasebuffer_cobj);
            if (!func)
                goto fail;
            func(obj, view);
            return;
        } else {
            PyErr_Clear();
        }
    }
  #endif
    goto nofail;
#if PY_VERSION_HEX < 0x02060000
fail:
#endif
    PyErr_WriteUnraisable(obj);
nofail:
    Py_DECREF(obj);
    view->obj = NULL;
}
#endif /*  PY_MAJOR_VERSION < 3 */


        static CYTHON_INLINE long __Pyx_pow_long(long b, long e) {
    long t = b;
    switch (e) {
        case 3:
            t *= b;
        case 2:
            t *= b;
        case 1:
            return t;
        case 0:
            return 1;
    }
    #if 1
    if (unlikely(e<0)) return 0;
    #endif
    t = 1;
    while (likely(e)) {
        t *= (b * (e&1)) | ((~e)&1);    /* 1 or b */
        b *= b;
        e >>= 1;
    }
    return t;
}

#if CYTHON_CCOMPLEX
  #ifdef __cplusplus
    static CYTHON_INLINE __pyx_t_float_complex __pyx_t_float_complex_from_parts(float x, float y) {
      return ::std::complex< float >(x, y);
    }
  #else
    static CYTHON_INLINE __pyx_t_float_complex __pyx_t_float_complex_from_parts(float x, float y) {
      return x + y*(__pyx_t_float_complex)_Complex_I;
    }
  #endif
#else
    static CYTHON_INLINE __pyx_t_float_complex __pyx_t_float_complex_from_parts(float x, float y) {
      __pyx_t_float_complex z;
      z.real = x;
      z.imag = y;
      return z;
    }
#endif

#if CYTHON_CCOMPLEX
#else
    static CYTHON_INLINE int __Pyx_c_eqf(__pyx_t_float_complex a, __pyx_t_float_complex b) {
       return (a.real == b.real) && (a.imag == b.imag);
    }
    static CYTHON_INLINE __pyx_t_float_complex __Pyx_c_sumf(__pyx_t_float_complex a, __pyx_t_float_complex b) {
        __pyx_t_float_complex z;
        z.real = a.real + b.real;
        z.imag = a.imag + b.imag;
        return z;
    }
    static CYTHON_INLINE __pyx_t_float_complex __Pyx_c_difff(__pyx_t_float_complex a, __pyx_t_float_complex b) {
        __pyx_t_float_complex z;
        z.real = a.real - b.real;
        z.imag = a.imag - b.imag;
        return z;
    }
    static CYTHON_INLINE __pyx_t_float_complex __Pyx_c_prodf(__pyx_t_float_complex a, __pyx_t_float_complex b) {
        __pyx_t_float_complex z;
        z.real = a.real * b.real - a.imag * b.imag;
        z.imag = a.real * b.imag + a.imag * b.real;
        return z;
    }
    static CYTHON_INLINE __pyx_t_float_complex __Pyx_c_quotf(__pyx_t_float_complex a, __pyx_t_float_complex b) {
        __pyx_t_float_complex z;
        float denom = b.real * b.real + b.imag * b.imag;
        z.real = (a.real * b.real + a.imag * b.imag) / denom;
        z.imag = (a.imag * b.real - a.real * b.imag) / denom;
        return z;
    }
    static CYTHON_INLINE __pyx_t_float_complex __Pyx_c_negf(__pyx_t_float_complex a) {
        __pyx_t_float_complex z;
        z.real = -a.real;
        z.imag = -a.imag;
        return z;
    }
    static CYTHON_INLINE int __Pyx_c_is_zerof(__pyx_t_float_complex a) {
       return (a.real == 0) && (a.imag == 0);
    }
    static CYTHON_INLINE __pyx_t_float_complex __Pyx_c_conjf(__pyx_t_float_complex a) {
        __pyx_t_float_complex z;
        z.real =  a.real;
        z.imag = -a.imag;
        return z;
    }
    #if 1
        static CYTHON_INLINE float __Pyx_c_absf(__pyx_t_float_complex z) {
          #if !defined(HAVE_HYPOT) || defined(_MSC_VER)
            return sqrtf(z.real*z.real + z.imag*z.imag);
          #else
            return hypotf(z.real, z.imag);
          #endif
        }
        static CYTHON_INLINE __pyx_t_float_complex __Pyx_c_powf(__pyx_t_float_complex a, __pyx_t_float_complex b) {
            __pyx_t_float_complex z;
            float r, lnr, theta, z_r, z_theta;
            if (b.imag == 0 && b.real == (int)b.real) {
                if (b.real < 0) {
                    float denom = a.real * a.real + a.imag * a.imag;
                    a.real = a.real / denom;
                    a.imag = -a.imag / denom;
                    b.real = -b.real;
                }
                switch ((int)b.real) {
                    case 0:
                        z.real = 1;
                        z.imag = 0;
                        return z;
                    case 1:
                        return a;
                    case 2:
                        z = __Pyx_c_prodf(a, a);
                        return __Pyx_c_prodf(a, a);
                    case 3:
                        z = __Pyx_c_prodf(a, a);
                        return __Pyx_c_prodf(z, a);
                    case 4:
                        z = __Pyx_c_prodf(a, a);
                        return __Pyx_c_prodf(z, z);
                }
            }
            if (a.imag == 0) {
                if (a.real == 0) {
                    return a;
                }
                r = a.real;
                theta = 0;
            } else {
                r = __Pyx_c_absf(a);
                theta = atan2f(a.imag, a.real);
            }
            lnr = logf(r);
            z_r = expf(lnr * b.real - theta * b.imag);
            z_theta = theta * b.real + lnr * b.imag;
            z.real = z_r * cosf(z_theta);
            z.imag = z_r * sinf(z_theta);
            return z;
        }
    #endif
#endif

#if CYTHON_CCOMPLEX
  #ifdef __cplusplus
    static CYTHON_INLINE __pyx_t_double_complex __pyx_t_double_complex_from_parts(double x, double y) {
      return ::std::complex< double >(x, y);
    }
  #else
    static CYTHON_INLINE __pyx_t_double_complex __pyx_t_double_complex_from_parts(double x, double y) {
      return x + y*(__pyx_t_double_complex)_Complex_I;
    }
  #endif
#else
    static CYTHON_INLINE __pyx_t_double_complex __pyx_t_double_complex_from_parts(double x, double y) {
      __pyx_t_double_complex z;
      z.real = x;
      z.imag = y;
      return z;
    }
#endif

#if CYTHON_CCOMPLEX
#else
    static CYTHON_INLINE int __Pyx_c_eq(__pyx_t_double_complex a, __pyx_t_double_complex b) {
       return (a.real == b.real) && (a.imag == b.imag);
    }
    static CYTHON_INLINE __pyx_t_double_complex __Pyx_c_sum(__pyx_t_double_complex a, __pyx_t_double_complex b) {
        __pyx_t_double_complex z;
        z.real = a.real + b.real;
        z.imag = a.imag + b.imag;
        return z;
    }
    static CYTHON_INLINE __pyx_t_double_complex __Pyx_c_diff(__pyx_t_double_complex a, __pyx_t_double_complex b) {
        __pyx_t_double_complex z;
        z.real = a.real - b.real;
        z.imag = a.imag - b.imag;
        return z;
    }
    static CYTHON_INLINE __pyx_t_double_complex __Pyx_c_prod(__pyx_t_double_complex a, __pyx_t_double_complex b) {
        __pyx_t_double_complex z;
        z.real = a.real * b.real - a.imag * b.imag;
        z.imag = a.real * b.imag + a.imag * b.real;
        return z;
    }
    static CYTHON_INLINE __pyx_t_double_complex __Pyx_c_quot(__pyx_t_double_complex a, __pyx_t_double_complex b) {
        __pyx_t_double_complex z;
        double denom = b.real * b.real + b.imag * b.imag;
        z.real = (a.real * b.real + a.imag * b.imag) / denom;
        z.imag = (a.imag * b.real - a.real * b.imag) / denom;
        return z;
    }
    static CYTHON_INLINE __pyx_t_double_complex __Pyx_c_neg(__pyx_t_double_complex a) {
        __pyx_t_double_complex z;
        z.real = -a.real;
        z.imag = -a.imag;
        return z;
    }
    static CYTHON_INLINE int __Pyx_c_is_zero(__pyx_t_double_complex a) {
       return (a.real == 0) && (a.imag == 0);
    }
    static CYTHON_INLINE __pyx_t_double_complex __Pyx_c_conj(__pyx_t_double_complex a) {
        __pyx_t_double_complex z;
        z.real =  a.real;
        z.imag = -a.imag;
        return z;
    }
    #if 1
        static CYTHON_INLINE double __Pyx_c_abs(__pyx_t_double_complex z) {
          #if !defined(HAVE_HYPOT) || defined(_MSC_VER)
            return sqrt(z.real*z.real + z.imag*z.imag);
          #else
            return hypot(z.real, z.imag);
          #endif
        }
        static CYTHON_INLINE __pyx_t_double_complex __Pyx_c_pow(__pyx_t_double_complex a, __pyx_t_double_complex b) {
            __pyx_t_double_complex z;
            double r, lnr, theta, z_r, z_theta;
            if (b.imag == 0 && b.real == (int)b.real) {
                if (b.real < 0) {
                    double denom = a.real * a.real + a.imag * a.imag;
                    a.real = a.real / denom;
                    a.imag = -a.imag / denom;
                    b.real = -b.real;
                }
                switch ((int)b.real) {
                    case 0:
                        z.real = 1;
                        z.imag = 0;
                        return z;
                    case 1:
                        return a;
                    case 2:
                        z = __Pyx_c_prod(a, a);
                        return __Pyx_c_prod(a, a);
                    case 3:
                        z = __Pyx_c_prod(a, a);
                        return __Pyx_c_prod(z, a);
                    case 4:
                        z = __Pyx_c_prod(a, a);
                        return __Pyx_c_prod(z, z);
                }
            }
            if (a.imag == 0) {
                if (a.real == 0) {
                    return a;
                }
                r = a.real;
                theta = 0;
            } else {
                r = __Pyx_c_abs(a);
                theta = atan2(a.imag, a.real);
            }
            lnr = log(r);
            z_r = exp(lnr * b.real - theta * b.imag);
            z_theta = theta * b.real + lnr * b.imag;
            z.real = z_r * cos(z_theta);
            z.imag = z_r * sin(z_theta);
            return z;
        }
    #endif
#endif

static CYTHON_INLINE unsigned char __Pyx_PyInt_AsUnsignedChar(PyObject* x) {
    const unsigned char neg_one = (unsigned char)-1, const_zero = 0;
    const int is_unsigned = neg_one > const_zero;
    if (sizeof(unsigned char) < sizeof(long)) {
        long val = __Pyx_PyInt_AsLong(x);
        if (unlikely(val != (long)(unsigned char)val)) {
            if (!unlikely(val == -1 && PyErr_Occurred())) {
                PyErr_SetString(PyExc_OverflowError,
                    (is_unsigned && unlikely(val < 0)) ?
                    "can't convert negative value to unsigned char" :
                    "value too large to convert to unsigned char");
            }
            return (unsigned char)-1;
        }
        return (unsigned char)val;
    }
    return (unsigned char)__Pyx_PyInt_AsUnsignedLong(x);
}

static CYTHON_INLINE unsigned short __Pyx_PyInt_AsUnsignedShort(PyObject* x) {
    const unsigned short neg_one = (unsigned short)-1, const_zero = 0;
    const int is_unsigned = neg_one > const_zero;
    if (sizeof(unsigned short) < sizeof(long)) {
        long val = __Pyx_PyInt_AsLong(x);
        if (unlikely(val != (long)(unsigned short)val)) {
            if (!unlikely(val == -1 && PyErr_Occurred())) {
                PyErr_SetString(PyExc_OverflowError,
                    (is_unsigned && unlikely(val < 0)) ?
                    "can't convert negative value to unsigned short" :
                    "value too large to convert to unsigned short");
            }
            return (unsigned short)-1;
        }
        return (unsigned short)val;
    }
    return (unsigned short)__Pyx_PyInt_AsUnsignedLong(x);
}

static CYTHON_INLINE unsigned int __Pyx_PyInt_AsUnsignedInt(PyObject* x) {
    const unsigned int neg_one = (unsigned int)-1, const_zero = 0;
    const int is_unsigned = neg_one > const_zero;
    if (sizeof(unsigned int) < sizeof(long)) {
        long val = __Pyx_PyInt_AsLong(x);
        if (unlikely(val != (long)(unsigned int)val)) {
            if (!unlikely(val == -1 && PyErr_Occurred())) {
                PyErr_SetString(PyExc_OverflowError,
                    (is_unsigned && unlikely(val < 0)) ?
                    "can't convert negative value to unsigned int" :
                    "value too large to convert to unsigned int");
            }
            return (unsigned int)-1;
        }
        return (unsigned int)val;
    }
    return (unsigned int)__Pyx_PyInt_AsUnsignedLong(x);
}

static CYTHON_INLINE char __Pyx_PyInt_AsChar(PyObject* x) {
    const char neg_one = (char)-1, const_zero = 0;
    const int is_unsigned = neg_one > const_zero;
    if (sizeof(char) < sizeof(long)) {
        long val = __Pyx_PyInt_AsLong(x);
        if (unlikely(val != (long)(char)val)) {
            if (!unlikely(val == -1 && PyErr_Occurred())) {
                PyErr_SetString(PyExc_OverflowError,
                    (is_unsigned && unlikely(val < 0)) ?
                    "can't convert negative value to char" :
                    "value too large to convert to char");
            }
            return (char)-1;
        }
        return (char)val;
    }
    return (char)__Pyx_PyInt_AsLong(x);
}

static CYTHON_INLINE short __Pyx_PyInt_AsShort(PyObject* x) {
    const short neg_one = (short)-1, const_zero = 0;
    const int is_unsigned = neg_one > const_zero;
    if (sizeof(short) < sizeof(long)) {
        long val = __Pyx_PyInt_AsLong(x);
        if (unlikely(val != (long)(short)val)) {
            if (!unlikely(val == -1 && PyErr_Occurred())) {
                PyErr_SetString(PyExc_OverflowError,
                    (is_unsigned && unlikely(val < 0)) ?
                    "can't convert negative value to short" :
                    "value too large to convert to short");
            }
            return (short)-1;
        }
        return (short)val;
    }
    return (short)__Pyx_PyInt_AsLong(x);
}

static CYTHON_INLINE int __Pyx_PyInt_AsInt(PyObject* x) {
    const int neg_one = (int)-1, const_zero = 0;
    const int is_unsigned = neg_one > const_zero;
    if (sizeof(int) < sizeof(long)) {
        long val = __Pyx_PyInt_AsLong(x);
        if (unlikely(val != (long)(int)val)) {
            if (!unlikely(val == -1 && PyErr_Occurred())) {
                PyErr_SetString(PyExc_OverflowError,
                    (is_unsigned && unlikely(val < 0)) ?
                    "can't convert negative value to int" :
                    "value too large to convert to int");
            }
            return (int)-1;
        }
        return (int)val;
    }
    return (int)__Pyx_PyInt_AsLong(x);
}

static CYTHON_INLINE signed char __Pyx_PyInt_AsSignedChar(PyObject* x) {
    const signed char neg_one = (signed char)-1, const_zero = 0;
    const int is_unsigned = neg_one > const_zero;
    if (sizeof(signed char) < sizeof(long)) {
        long val = __Pyx_PyInt_AsLong(x);
        if (unlikely(val != (long)(signed char)val)) {
            if (!unlikely(val == -1 && PyErr_Occurred())) {
                PyErr_SetString(PyExc_OverflowError,
                    (is_unsigned && unlikely(val < 0)) ?
                    "can't convert negative value to signed char" :
                    "value too large to convert to signed char");
            }
            return (signed char)-1;
        }
        return (signed char)val;
    }
    return (signed char)__Pyx_PyInt_AsSignedLong(x);
}

static CYTHON_INLINE signed short __Pyx_PyInt_AsSignedShort(PyObject* x) {
    const signed short neg_one = (signed short)-1, const_zero = 0;
    const int is_unsigned = neg_one > const_zero;
    if (sizeof(signed short) < sizeof(long)) {
        long val = __Pyx_PyInt_AsLong(x);
        if (unlikely(val != (long)(signed short)val)) {
            if (!unlikely(val == -1 && PyErr_Occurred())) {
                PyErr_SetString(PyExc_OverflowError,
                    (is_unsigned && unlikely(val < 0)) ?
                    "can't convert negative value to signed short" :
                    "value too large to convert to signed short");
            }
            return (signed short)-1;
        }
        return (signed short)val;
    }
    return (signed short)__Pyx_PyInt_AsSignedLong(x);
}

static CYTHON_INLINE signed int __Pyx_PyInt_AsSignedInt(PyObject* x) {
    const signed int neg_one = (signed int)-1, const_zero = 0;
    const int is_unsigned = neg_one > const_zero;
    if (sizeof(signed int) < sizeof(long)) {
        long val = __Pyx_PyInt_AsLong(x);
        if (unlikely(val != (long)(signed int)val)) {
            if (!unlikely(val == -1 && PyErr_Occurred())) {
                PyErr_SetString(PyExc_OverflowError,
                    (is_unsigned && unlikely(val < 0)) ?
                    "can't convert negative value to signed int" :
                    "value too large to convert to signed int");
            }
            return (signed int)-1;
        }
        return (signed int)val;
    }
    return (signed int)__Pyx_PyInt_AsSignedLong(x);
}

static CYTHON_INLINE int __Pyx_PyInt_AsLongDouble(PyObject* x) {
    const int neg_one = (int)-1, const_zero = 0;
    const int is_unsigned = neg_one > const_zero;
    if (sizeof(int) < sizeof(long)) {
        long val = __Pyx_PyInt_AsLong(x);
        if (unlikely(val != (long)(int)val)) {
            if (!unlikely(val == -1 && PyErr_Occurred())) {
                PyErr_SetString(PyExc_OverflowError,
                    (is_unsigned && unlikely(val < 0)) ?
                    "can't convert negative value to int" :
                    "value too large to convert to int");
            }
            return (int)-1;
        }
        return (int)val;
    }
    return (int)__Pyx_PyInt_AsLong(x);
}

#if CYTHON_COMPILING_IN_CPYTHON && PY_MAJOR_VERSION >= 3
#if CYTHON_USE_PYLONG_INTERNALS
#include "longintrepr.h"
#endif
#endif
static CYTHON_INLINE unsigned long __Pyx_PyInt_AsUnsignedLong(PyObject* x) {
    const unsigned long neg_one = (unsigned long)-1, const_zero = 0;
    const int is_unsigned = neg_one > const_zero;
#if PY_MAJOR_VERSION < 3
    if (likely(PyInt_Check(x))) {
        long val = PyInt_AS_LONG(x);
        if (is_unsigned && unlikely(val < 0)) {
            PyErr_SetString(PyExc_OverflowError,
                            "can't convert negative value to unsigned long");
            return (unsigned long)-1;
        }
        return (unsigned long)val;
    } else
#endif
    if (likely(PyLong_Check(x))) {
        if (is_unsigned) {
#if CYTHON_COMPILING_IN_CPYTHON && PY_MAJOR_VERSION >= 3
#if CYTHON_USE_PYLONG_INTERNALS
            if (sizeof(digit) <= sizeof(unsigned long)) {
                switch (Py_SIZE(x)) {
                    case  0: return 0;
                    case  1: return (unsigned long) ((PyLongObject*)x)->ob_digit[0];
                }
            }
#endif
#endif
            if (unlikely(Py_SIZE(x) < 0)) {
                PyErr_SetString(PyExc_OverflowError,
                                "can't convert negative value to unsigned long");
                return (unsigned long)-1;
            }
            return (unsigned long)PyLong_AsUnsignedLong(x);
        } else {
#if CYTHON_COMPILING_IN_CPYTHON && PY_MAJOR_VERSION >= 3
#if CYTHON_USE_PYLONG_INTERNALS
            if (sizeof(digit) <= sizeof(unsigned long)) {
                switch (Py_SIZE(x)) {
                    case  0: return 0;
                    case  1: return +(unsigned long) ((PyLongObject*)x)->ob_digit[0];
                    case -1: return -(unsigned long) ((PyLongObject*)x)->ob_digit[0];
                }
            }
#endif
#endif
            return (unsigned long)PyLong_AsLong(x);
        }
    } else {
        unsigned long val;
        PyObject *tmp = __Pyx_PyNumber_Int(x);
        if (!tmp) return (unsigned long)-1;
        val = __Pyx_PyInt_AsUnsignedLong(tmp);
        Py_DECREF(tmp);
        return val;
    }
}

#if CYTHON_COMPILING_IN_CPYTHON && PY_MAJOR_VERSION >= 3
#if CYTHON_USE_PYLONG_INTERNALS
#include "longintrepr.h"
#endif
#endif
static CYTHON_INLINE unsigned PY_LONG_LONG __Pyx_PyInt_AsUnsignedLongLong(PyObject* x) {
    const unsigned PY_LONG_LONG neg_one = (unsigned PY_LONG_LONG)-1, const_zero = 0;
    const int is_unsigned = neg_one > const_zero;
#if PY_MAJOR_VERSION < 3
    if (likely(PyInt_Check(x))) {
        long val = PyInt_AS_LONG(x);
        if (is_unsigned && unlikely(val < 0)) {
            PyErr_SetString(PyExc_OverflowError,
                            "can't convert negative value to unsigned PY_LONG_LONG");
            return (unsigned PY_LONG_LONG)-1;
        }
        return (unsigned PY_LONG_LONG)val;
    } else
#endif
    if (likely(PyLong_Check(x))) {
        if (is_unsigned) {
#if CYTHON_COMPILING_IN_CPYTHON && PY_MAJOR_VERSION >= 3
#if CYTHON_USE_PYLONG_INTERNALS
            if (sizeof(digit) <= sizeof(unsigned PY_LONG_LONG)) {
                switch (Py_SIZE(x)) {
                    case  0: return 0;
                    case  1: return (unsigned PY_LONG_LONG) ((PyLongObject*)x)->ob_digit[0];
                }
            }
#endif
#endif
            if (unlikely(Py_SIZE(x) < 0)) {
                PyErr_SetString(PyExc_OverflowError,
                                "can't convert negative value to unsigned PY_LONG_LONG");
                return (unsigned PY_LONG_LONG)-1;
            }
            return (unsigned PY_LONG_LONG)PyLong_AsUnsignedLongLong(x);
        } else {
#if CYTHON_COMPILING_IN_CPYTHON && PY_MAJOR_VERSION >= 3
#if CYTHON_USE_PYLONG_INTERNALS
            if (sizeof(digit) <= sizeof(unsigned PY_LONG_LONG)) {
                switch (Py_SIZE(x)) {
                    case  0: return 0;
                    case  1: return +(unsigned PY_LONG_LONG) ((PyLongObject*)x)->ob_digit[0];
                    case -1: return -(unsigned PY_LONG_LONG) ((PyLongObject*)x)->ob_digit[0];
                }
            }
#endif
#endif
            return (unsigned PY_LONG_LONG)PyLong_AsLongLong(x);
        }
    } else {
        unsigned PY_LONG_LONG val;
        PyObject *tmp = __Pyx_PyNumber_Int(x);
        if (!tmp) return (unsigned PY_LONG_LONG)-1;
        val = __Pyx_PyInt_AsUnsignedLongLong(tmp);
        Py_DECREF(tmp);
        return val;
    }
}

#if CYTHON_COMPILING_IN_CPYTHON && PY_MAJOR_VERSION >= 3
#if CYTHON_USE_PYLONG_INTERNALS
#include "longintrepr.h"
#endif
#endif
static CYTHON_INLINE long __Pyx_PyInt_AsLong(PyObject* x) {
    const long neg_one = (long)-1, const_zero = 0;
    const int is_unsigned = neg_one > const_zero;
#if PY_MAJOR_VERSION < 3
    if (likely(PyInt_Check(x))) {
        long val = PyInt_AS_LONG(x);
        if (is_unsigned && unlikely(val < 0)) {
            PyErr_SetString(PyExc_OverflowError,
                            "can't convert negative value to long");
            return (long)-1;
        }
        return (long)val;
    } else
#endif
    if (likely(PyLong_Check(x))) {
        if (is_unsigned) {
#if CYTHON_COMPILING_IN_CPYTHON && PY_MAJOR_VERSION >= 3
#if CYTHON_USE_PYLONG_INTERNALS
            if (sizeof(digit) <= sizeof(long)) {
                switch (Py_SIZE(x)) {
                    case  0: return 0;
                    case  1: return (long) ((PyLongObject*)x)->ob_digit[0];
                }
            }
#endif
#endif
            if (unlikely(Py_SIZE(x) < 0)) {
                PyErr_SetString(PyExc_OverflowError,
                                "can't convert negative value to long");
                return (long)-1;
            }
            return (long)PyLong_AsUnsignedLong(x);
        } else {
#if CYTHON_COMPILING_IN_CPYTHON && PY_MAJOR_VERSION >= 3
#if CYTHON_USE_PYLONG_INTERNALS
            if (sizeof(digit) <= sizeof(long)) {
                switch (Py_SIZE(x)) {
                    case  0: return 0;
                    case  1: return +(long) ((PyLongObject*)x)->ob_digit[0];
                    case -1: return -(long) ((PyLongObject*)x)->ob_digit[0];
                }
            }
#endif
#endif
            return (long)PyLong_AsLong(x);
        }
    } else {
        long val;
        PyObject *tmp = __Pyx_PyNumber_Int(x);
        if (!tmp) return (long)-1;
        val = __Pyx_PyInt_AsLong(tmp);
        Py_DECREF(tmp);
        return val;
    }
}

#if CYTHON_COMPILING_IN_CPYTHON && PY_MAJOR_VERSION >= 3
#if CYTHON_USE_PYLONG_INTERNALS
#include "longintrepr.h"
#endif
#endif
static CYTHON_INLINE PY_LONG_LONG __Pyx_PyInt_AsLongLong(PyObject* x) {
    const PY_LONG_LONG neg_one = (PY_LONG_LONG)-1, const_zero = 0;
    const int is_unsigned = neg_one > const_zero;
#if PY_MAJOR_VERSION < 3
    if (likely(PyInt_Check(x))) {
        long val = PyInt_AS_LONG(x);
        if (is_unsigned && unlikely(val < 0)) {
            PyErr_SetString(PyExc_OverflowError,
                            "can't convert negative value to PY_LONG_LONG");
            return (PY_LONG_LONG)-1;
        }
        return (PY_LONG_LONG)val;
    } else
#endif
    if (likely(PyLong_Check(x))) {
        if (is_unsigned) {
#if CYTHON_COMPILING_IN_CPYTHON && PY_MAJOR_VERSION >= 3
#if CYTHON_USE_PYLONG_INTERNALS
            if (sizeof(digit) <= sizeof(PY_LONG_LONG)) {
                switch (Py_SIZE(x)) {
                    case  0: return 0;
                    case  1: return (PY_LONG_LONG) ((PyLongObject*)x)->ob_digit[0];
                }
            }
#endif
#endif
            if (unlikely(Py_SIZE(x) < 0)) {
                PyErr_SetString(PyExc_OverflowError,
                                "can't convert negative value to PY_LONG_LONG");
                return (PY_LONG_LONG)-1;
            }
            return (PY_LONG_LONG)PyLong_AsUnsignedLongLong(x);
        } else {
#if CYTHON_COMPILING_IN_CPYTHON && PY_MAJOR_VERSION >= 3
#if CYTHON_USE_PYLONG_INTERNALS
            if (sizeof(digit) <= sizeof(PY_LONG_LONG)) {
                switch (Py_SIZE(x)) {
                    case  0: return 0;
                    case  1: return +(PY_LONG_LONG) ((PyLongObject*)x)->ob_digit[0];
                    case -1: return -(PY_LONG_LONG) ((PyLongObject*)x)->ob_digit[0];
                }
            }
#endif
#endif
            return (PY_LONG_LONG)PyLong_AsLongLong(x);
        }
    } else {
        PY_LONG_LONG val;
        PyObject *tmp = __Pyx_PyNumber_Int(x);
        if (!tmp) return (PY_LONG_LONG)-1;
        val = __Pyx_PyInt_AsLongLong(tmp);
        Py_DECREF(tmp);
        return val;
    }
}

#if CYTHON_COMPILING_IN_CPYTHON && PY_MAJOR_VERSION >= 3
#if CYTHON_USE_PYLONG_INTERNALS
#include "longintrepr.h"
#endif
#endif
static CYTHON_INLINE signed long __Pyx_PyInt_AsSignedLong(PyObject* x) {
    const signed long neg_one = (signed long)-1, const_zero = 0;
    const int is_unsigned = neg_one > const_zero;
#if PY_MAJOR_VERSION < 3
    if (likely(PyInt_Check(x))) {
        long val = PyInt_AS_LONG(x);
        if (is_unsigned && unlikely(val < 0)) {
            PyErr_SetString(PyExc_OverflowError,
                            "can't convert negative value to signed long");
            return (signed long)-1;
        }
        return (signed long)val;
    } else
#endif
    if (likely(PyLong_Check(x))) {
        if (is_unsigned) {
#if CYTHON_COMPILING_IN_CPYTHON && PY_MAJOR_VERSION >= 3
#if CYTHON_USE_PYLONG_INTERNALS
            if (sizeof(digit) <= sizeof(signed long)) {
                switch (Py_SIZE(x)) {
                    case  0: return 0;
                    case  1: return (signed long) ((PyLongObject*)x)->ob_digit[0];
                }
            }
#endif
#endif
            if (unlikely(Py_SIZE(x) < 0)) {
                PyErr_SetString(PyExc_OverflowError,
                                "can't convert negative value to signed long");
                return (signed long)-1;
            }
            return (signed long)PyLong_AsUnsignedLong(x);
        } else {
#if CYTHON_COMPILING_IN_CPYTHON && PY_MAJOR_VERSION >= 3
#if CYTHON_USE_PYLONG_INTERNALS
            if (sizeof(digit) <= sizeof(signed long)) {
                switch (Py_SIZE(x)) {
                    case  0: return 0;
                    case  1: return +(signed long) ((PyLongObject*)x)->ob_digit[0];
                    case -1: return -(signed long) ((PyLongObject*)x)->ob_digit[0];
                }
            }
#endif
#endif
            return (signed long)PyLong_AsLong(x);
        }
    } else {
        signed long val;
        PyObject *tmp = __Pyx_PyNumber_Int(x);
        if (!tmp) return (signed long)-1;
        val = __Pyx_PyInt_AsSignedLong(tmp);
        Py_DECREF(tmp);
        return val;
    }
}

#if CYTHON_COMPILING_IN_CPYTHON && PY_MAJOR_VERSION >= 3
#if CYTHON_USE_PYLONG_INTERNALS
#include "longintrepr.h"
#endif
#endif
static CYTHON_INLINE signed PY_LONG_LONG __Pyx_PyInt_AsSignedLongLong(PyObject* x) {
    const signed PY_LONG_LONG neg_one = (signed PY_LONG_LONG)-1, const_zero = 0;
    const int is_unsigned = neg_one > const_zero;
#if PY_MAJOR_VERSION < 3
    if (likely(PyInt_Check(x))) {
        long val = PyInt_AS_LONG(x);
        if (is_unsigned && unlikely(val < 0)) {
            PyErr_SetString(PyExc_OverflowError,
                            "can't convert negative value to signed PY_LONG_LONG");
            return (signed PY_LONG_LONG)-1;
        }
        return (signed PY_LONG_LONG)val;
    } else
#endif
    if (likely(PyLong_Check(x))) {
        if (is_unsigned) {
#if CYTHON_COMPILING_IN_CPYTHON && PY_MAJOR_VERSION >= 3
#if CYTHON_USE_PYLONG_INTERNALS
            if (sizeof(digit) <= sizeof(signed PY_LONG_LONG)) {
                switch (Py_SIZE(x)) {
                    case  0: return 0;
                    case  1: return (signed PY_LONG_LONG) ((PyLongObject*)x)->ob_digit[0];
                }
            }
#endif
#endif
            if (unlikely(Py_SIZE(x) < 0)) {
                PyErr_SetString(PyExc_OverflowError,
                                "can't convert negative value to signed PY_LONG_LONG");
                return (signed PY_LONG_LONG)-1;
            }
            return (signed PY_LONG_LONG)PyLong_AsUnsignedLongLong(x);
        } else {
#if CYTHON_COMPILING_IN_CPYTHON && PY_MAJOR_VERSION >= 3
#if CYTHON_USE_PYLONG_INTERNALS
            if (sizeof(digit) <= sizeof(signed PY_LONG_LONG)) {
                switch (Py_SIZE(x)) {
                    case  0: return 0;
                    case  1: return +(signed PY_LONG_LONG) ((PyLongObject*)x)->ob_digit[0];
                    case -1: return -(signed PY_LONG_LONG) ((PyLongObject*)x)->ob_digit[0];
                }
            }
#endif
#endif
            return (signed PY_LONG_LONG)PyLong_AsLongLong(x);
        }
    } else {
        signed PY_LONG_LONG val;
        PyObject *tmp = __Pyx_PyNumber_Int(x);
        if (!tmp) return (signed PY_LONG_LONG)-1;
        val = __Pyx_PyInt_AsSignedLongLong(tmp);
        Py_DECREF(tmp);
        return val;
    }
}

static int __Pyx_check_binary_version(void) {
    char ctversion[4], rtversion[4];
    PyOS_snprintf(ctversion, 4, "%d.%d", PY_MAJOR_VERSION, PY_MINOR_VERSION);
    PyOS_snprintf(rtversion, 4, "%s", Py_GetVersion());
    if (ctversion[0] != rtversion[0] || ctversion[2] != rtversion[2]) {
        char message[200];
        PyOS_snprintf(message, sizeof(message),
                      "compiletime version %s of module '%.100s' "
                      "does not match runtime version %s",
                      ctversion, __Pyx_MODULE_NAME, rtversion);
        #if PY_VERSION_HEX < 0x02050000
        return PyErr_Warn(NULL, message);
        #else
        return PyErr_WarnEx(NULL, message, 1);
        #endif
    }
    return 0;
}

#ifndef __PYX_HAVE_RT_ImportModule
#define __PYX_HAVE_RT_ImportModule
static PyObject *__Pyx_ImportModule(const char *name) {
    PyObject *py_name = 0;
    PyObject *py_module = 0;
    py_name = __Pyx_PyIdentifier_FromString(name);
    if (!py_name)
        goto bad;
    py_module = PyImport_Import(py_name);
    Py_DECREF(py_name);
    return py_module;
bad:
    Py_XDECREF(py_name);
    return 0;
}
#endif

#ifndef __PYX_HAVE_RT_ImportType
#define __PYX_HAVE_RT_ImportType
static PyTypeObject *__Pyx_ImportType(const char *module_name, const char *class_name,
    size_t size, int strict)
{
    PyObject *py_module = 0;
    PyObject *result = 0;
    PyObject *py_name = 0;
    char warning[200];
    Py_ssize_t basicsize;
#ifdef Py_LIMITED_API
    PyObject *py_basicsize;
#endif
    py_module = __Pyx_ImportModule(module_name);
    if (!py_module)
        goto bad;
    py_name = __Pyx_PyIdentifier_FromString(class_name);
    if (!py_name)
        goto bad;
    result = PyObject_GetAttr(py_module, py_name);
    Py_DECREF(py_name);
    py_name = 0;
    Py_DECREF(py_module);
    py_module = 0;
    if (!result)
        goto bad;
    if (!PyType_Check(result)) {
        PyErr_Format(PyExc_TypeError,
            "%s.%s is not a type object",
            module_name, class_name);
        goto bad;
    }
#ifndef Py_LIMITED_API
    basicsize = ((PyTypeObject *)result)->tp_basicsize;
#else
    py_basicsize = PyObject_GetAttrString(result, "__basicsize__");
    if (!py_basicsize)
        goto bad;
    basicsize = PyLong_AsSsize_t(py_basicsize);
    Py_DECREF(py_basicsize);
    py_basicsize = 0;
    if (basicsize == (Py_ssize_t)-1 && PyErr_Occurred())
        goto bad;
#endif
    if (!strict && (size_t)basicsize > size) {
        PyOS_snprintf(warning, sizeof(warning),
            "%s.%s size changed, may indicate binary incompatibility",
            module_name, class_name);
        #if PY_VERSION_HEX < 0x02050000
        if (PyErr_Warn(NULL, warning) < 0) goto bad;
        #else
        if (PyErr_WarnEx(NULL, warning, 0) < 0) goto bad;
        #endif
    }
    else if ((size_t)basicsize != size) {
        PyErr_Format(PyExc_ValueError,
            "%s.%s has the wrong size, try recompiling",
            module_name, class_name);
        goto bad;
    }
    return (PyTypeObject *)result;
bad:
    Py_XDECREF(py_module);
    Py_XDECREF(result);
    return NULL;
}
#endif

static int __pyx_bisect_code_objects(__Pyx_CodeObjectCacheEntry* entries, int count, int code_line) {
    int start = 0, mid = 0, end = count - 1;
    if (end >= 0 && code_line > entries[end].code_line) {
        return count;
    }
    while (start < end) {
        mid = (start + end) / 2;
        if (code_line < entries[mid].code_line) {
            end = mid;
        } else if (code_line > entries[mid].code_line) {
             start = mid + 1;
        } else {
            return mid;
        }
    }
    if (code_line <= entries[mid].code_line) {
        return mid;
    } else {
        return mid + 1;
    }
}
static PyCodeObject *__pyx_find_code_object(int code_line) {
    PyCodeObject* code_object;
    int pos;
    if (unlikely(!code_line) || unlikely(!__pyx_code_cache.entries)) {
        return NULL;
    }
    pos = __pyx_bisect_code_objects(__pyx_code_cache.entries, __pyx_code_cache.count, code_line);
    if (unlikely(pos >= __pyx_code_cache.count) || unlikely(__pyx_code_cache.entries[pos].code_line != code_line)) {
        return NULL;
    }
    code_object = __pyx_code_cache.entries[pos].code_object;
    Py_INCREF(code_object);
    return code_object;
}
static void __pyx_insert_code_object(int code_line, PyCodeObject* code_object) {
    int pos, i;
    __Pyx_CodeObjectCacheEntry* entries = __pyx_code_cache.entries;
    if (unlikely(!code_line)) {
        return;
    }
    if (unlikely(!entries)) {
        entries = (__Pyx_CodeObjectCacheEntry*)PyMem_Malloc(64*sizeof(__Pyx_CodeObjectCacheEntry));
        if (likely(entries)) {
            __pyx_code_cache.entries = entries;
            __pyx_code_cache.max_count = 64;
            __pyx_code_cache.count = 1;
            entries[0].code_line = code_line;
            entries[0].code_object = code_object;
            Py_INCREF(code_object);
        }
        return;
    }
    pos = __pyx_bisect_code_objects(__pyx_code_cache.entries, __pyx_code_cache.count, code_line);
    if ((pos < __pyx_code_cache.count) && unlikely(__pyx_code_cache.entries[pos].code_line == code_line)) {
        PyCodeObject* tmp = entries[pos].code_object;
        entries[pos].code_object = code_object;
        Py_DECREF(tmp);
        return;
    }
    if (__pyx_code_cache.count == __pyx_code_cache.max_count) {
        int new_max = __pyx_code_cache.max_count + 64;
        entries = (__Pyx_CodeObjectCacheEntry*)PyMem_Realloc(
            __pyx_code_cache.entries, new_max*sizeof(__Pyx_CodeObjectCacheEntry));
        if (unlikely(!entries)) {
            return;
        }
        __pyx_code_cache.entries = entries;
        __pyx_code_cache.max_count = new_max;
    }
    for (i=__pyx_code_cache.count; i>pos; i--) {
        entries[i] = entries[i-1];
    }
    entries[pos].code_line = code_line;
    entries[pos].code_object = code_object;
    __pyx_code_cache.count++;
    Py_INCREF(code_object);
}

#include "compile.h"
#include "frameobject.h"
#include "traceback.h"
static PyCodeObject* __Pyx_CreateCodeObjectForTraceback(
            const char *funcname, int c_line,
            int py_line, const char *filename) {
    PyCodeObject *py_code = 0;
    PyObject *py_srcfile = 0;
    PyObject *py_funcname = 0;
    #if PY_MAJOR_VERSION < 3
    py_srcfile = PyString_FromString(filename);
    #else
    py_srcfile = PyUnicode_FromString(filename);
    #endif
    if (!py_srcfile) goto bad;
    if (c_line) {
        #if PY_MAJOR_VERSION < 3
        py_funcname = PyString_FromFormat( "%s (%s:%d)", funcname, __pyx_cfilenm, c_line);
        #else
        py_funcname = PyUnicode_FromFormat( "%s (%s:%d)", funcname, __pyx_cfilenm, c_line);
        #endif
    }
    else {
        #if PY_MAJOR_VERSION < 3
        py_funcname = PyString_FromString(funcname);
        #else
        py_funcname = PyUnicode_FromString(funcname);
        #endif
    }
    if (!py_funcname) goto bad;
    py_code = __Pyx_PyCode_New(
        0,            /*int argcount,*/
        0,            /*int kwonlyargcount,*/
        0,            /*int nlocals,*/
        0,            /*int stacksize,*/
        0,            /*int flags,*/
        __pyx_empty_bytes, /*PyObject *code,*/
        __pyx_empty_tuple, /*PyObject *consts,*/
        __pyx_empty_tuple, /*PyObject *names,*/
        __pyx_empty_tuple, /*PyObject *varnames,*/
        __pyx_empty_tuple, /*PyObject *freevars,*/
        __pyx_empty_tuple, /*PyObject *cellvars,*/
        py_srcfile,   /*PyObject *filename,*/
        py_funcname,  /*PyObject *name,*/
        py_line,      /*int firstlineno,*/
        __pyx_empty_bytes  /*PyObject *lnotab*/
    );
    Py_DECREF(py_srcfile);
    Py_DECREF(py_funcname);
    return py_code;
bad:
    Py_XDECREF(py_srcfile);
    Py_XDECREF(py_funcname);
    return NULL;
}
static void __Pyx_AddTraceback(const char *funcname, int c_line,
                               int py_line, const char *filename) {
    PyCodeObject *py_code = 0;
    PyObject *py_globals = 0;
    PyFrameObject *py_frame = 0;
    py_code = __pyx_find_code_object(c_line ? c_line : py_line);
    if (!py_code) {
        py_code = __Pyx_CreateCodeObjectForTraceback(
            funcname, c_line, py_line, filename);
        if (!py_code) goto bad;
        __pyx_insert_code_object(c_line ? c_line : py_line, py_code);
    }
    py_globals = PyModule_GetDict(__pyx_m);
    if (!py_globals) goto bad;
    py_frame = PyFrame_New(
        PyThreadState_GET(), /*PyThreadState *tstate,*/
        py_code,             /*PyCodeObject *code,*/
        py_globals,          /*PyObject *globals,*/
        0                    /*PyObject *locals*/
    );
    if (!py_frame) goto bad;
    py_frame->f_lineno = py_line;
    PyTraceBack_Here(py_frame);
bad:
    Py_XDECREF(py_code);
    Py_XDECREF(py_frame);
}

static int __Pyx_InitStrings(__Pyx_StringTabEntry *t) {
    while (t->p) {
        #if PY_MAJOR_VERSION < 3
        if (t->is_unicode) {
            *t->p = PyUnicode_DecodeUTF8(t->s, t->n - 1, NULL);
        } else if (t->intern) {
            *t->p = PyString_InternFromString(t->s);
        } else {
            *t->p = PyString_FromStringAndSize(t->s, t->n - 1);
        }
        #else  /* Python 3+ has unicode identifiers */
        if (t->is_unicode | t->is_str) {
            if (t->intern) {
                *t->p = PyUnicode_InternFromString(t->s);
            } else if (t->encoding) {
                *t->p = PyUnicode_Decode(t->s, t->n - 1, t->encoding, NULL);
            } else {
                *t->p = PyUnicode_FromStringAndSize(t->s, t->n - 1);
            }
        } else {
            *t->p = PyBytes_FromStringAndSize(t->s, t->n - 1);
        }
        #endif
        if (!*t->p)
            return -1;
        ++t;
    }
    return 0;
}

static CYTHON_INLINE PyObject* __Pyx_PyUnicode_FromString(char* c_str) {
    return __Pyx_PyUnicode_FromStringAndSize(c_str, strlen(c_str));
}
static CYTHON_INLINE char* __Pyx_PyObject_AsString(PyObject* o) {
    Py_ssize_t ignore;
    return __Pyx_PyObject_AsStringAndSize(o, &ignore);
}
static CYTHON_INLINE char* __Pyx_PyObject_AsStringAndSize(PyObject* o, Py_ssize_t *length) {
#if __PYX_DEFAULT_STRING_ENCODING_IS_ASCII || __PYX_DEFAULT_STRING_ENCODING_IS_DEFAULT
    if (
#if PY_MAJOR_VERSION < 3 && __PYX_DEFAULT_STRING_ENCODING_IS_ASCII
            __Pyx_sys_getdefaultencoding_not_ascii &&
#endif
            PyUnicode_Check(o)) {
#if PY_VERSION_HEX < 0x03030000
        char* defenc_c;
        PyObject* defenc = _PyUnicode_AsDefaultEncodedString(o, NULL);
        if (!defenc) return NULL;
        defenc_c = PyBytes_AS_STRING(defenc);
#if __PYX_DEFAULT_STRING_ENCODING_IS_ASCII
        {
            char* end = defenc_c + PyBytes_GET_SIZE(defenc);
            char* c;
            for (c = defenc_c; c < end; c++) {
                if ((unsigned char) (*c) >= 128) {
                    PyUnicode_AsASCIIString(o);
                    return NULL;
                }
            }
        }
#endif /*__PYX_DEFAULT_STRING_ENCODING_IS_ASCII*/
        *length = PyBytes_GET_SIZE(defenc);
        return defenc_c;
#else /* PY_VERSION_HEX < 0x03030000 */
        if (PyUnicode_READY(o) == -1) return NULL;
#if __PYX_DEFAULT_STRING_ENCODING_IS_ASCII
        if (PyUnicode_IS_ASCII(o)) {
            *length = PyUnicode_GET_DATA_SIZE(o);
            return PyUnicode_AsUTF8(o);
        } else {
            PyUnicode_AsASCIIString(o);
            return NULL;
        }
#else /* __PYX_DEFAULT_STRING_ENCODING_IS_ASCII */
        return PyUnicode_AsUTF8AndSize(o, length);
#endif /* __PYX_DEFAULT_STRING_ENCODING_IS_ASCII */
#endif /* PY_VERSION_HEX < 0x03030000 */
    } else
#endif /* __PYX_DEFAULT_STRING_ENCODING_IS_ASCII  || __PYX_DEFAULT_STRING_ENCODING_IS_DEFAULT */
    {
        char* result;
        int r = PyBytes_AsStringAndSize(o, &result, length);
        if (r < 0) {
            return NULL;
        } else {
            return result;
        }
    }
}
static CYTHON_INLINE int __Pyx_PyObject_IsTrue(PyObject* x) {
   int is_true = x == Py_True;
   if (is_true | (x == Py_False) | (x == Py_None)) return is_true;
   else return PyObject_IsTrue(x);
}
static CYTHON_INLINE PyObject* __Pyx_PyNumber_Int(PyObject* x) {
  PyNumberMethods *m;
  const char *name = NULL;
  PyObject *res = NULL;
#if PY_MAJOR_VERSION < 3
  if (PyInt_Check(x) || PyLong_Check(x))
#else
  if (PyLong_Check(x))
#endif
    return Py_INCREF(x), x;
  m = Py_TYPE(x)->tp_as_number;
#if PY_MAJOR_VERSION < 3
  if (m && m->nb_int) {
    name = "int";
    res = PyNumber_Int(x);
  }
  else if (m && m->nb_long) {
    name = "long";
    res = PyNumber_Long(x);
  }
#else
  if (m && m->nb_int) {
    name = "int";
    res = PyNumber_Long(x);
  }
#endif
  if (res) {
#if PY_MAJOR_VERSION < 3
    if (!PyInt_Check(res) && !PyLong_Check(res)) {
#else
    if (!PyLong_Check(res)) {
#endif
      PyErr_Format(PyExc_TypeError,
                   "__%s__ returned non-%s (type %.200s)",
                   name, name, Py_TYPE(res)->tp_name);
      Py_DECREF(res);
      return NULL;
    }
  }
  else if (!PyErr_Occurred()) {
    PyErr_SetString(PyExc_TypeError,
                    "an integer is required");
  }
  return res;
}
static CYTHON_INLINE Py_ssize_t __Pyx_PyIndex_AsSsize_t(PyObject* b) {
  Py_ssize_t ival;
  PyObject* x = PyNumber_Index(b);
  if (!x) return -1;
  ival = PyInt_AsSsize_t(x);
  Py_DECREF(x);
  return ival;
}
static CYTHON_INLINE PyObject * __Pyx_PyInt_FromSize_t(size_t ival) {
#if PY_VERSION_HEX < 0x02050000
   if (ival <= LONG_MAX)
       return PyInt_FromLong((long)ival);
   else {
       unsigned char *bytes = (unsigned char *) &ival;
       int one = 1; int little = (int)*(unsigned char*)&one;
       return _PyLong_FromByteArray(bytes, sizeof(size_t), little, 0);
   }
#else
   return PyInt_FromSize_t(ival);
#endif
}
static CYTHON_INLINE size_t __Pyx_PyInt_AsSize_t(PyObject* x) {
   unsigned PY_LONG_LONG val = __Pyx_PyInt_AsUnsignedLongLong(x);
   if (unlikely(val != (unsigned PY_LONG_LONG)(size_t)val)) {
       if ((val != (unsigned PY_LONG_LONG)-1) || !PyErr_Occurred())
           PyErr_SetString(PyExc_OverflowError,
                           "value too large to convert to size_t");
       return (size_t)-1;
   }
   return (size_t)val;
}


#endif /* Py_PYTHON_H */
